/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.Mode;
import inversion_applet.Kernel;
import inversion_applet.Utilities;
import inversion_applet.FrequenciesDataModel;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.io.File;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.URL;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.JLabel;
import java.security.AccessController;
import java.io.FilePermission;

/************************************************************************
* This class contains all of the data for the applet.  By centralising
* the data in this applet, it becomes easier and more transparent for
* the different parts to communicate.
************************************************************************/
public class DataCenter {

  /**********************************************************************
  * A constructor.
  **********************************************************************/
  public DataCenter(boolean isApplet) {
    // variables relating to rotational inversion:
    n_gridR = 10;
    RLambda = 1e-3;
    RTheta = 0.1;
    RThetaM = 0.1;
    RDelta = 0.2;
    RNoiseLevel = 1.0;
    RNoiseSelected = false;
    targetRotaLoaded = false;
    rotationInversionUpToDate = true;

    // variables relating to structural inversion:
    kernelType = 0;
    n_gridS = 10;
    n_surf = 6;
    SLambda = 1e-3;
    Beta1 = 1.0;
    Beta2 = 1.0;
    STheta1 = 0.1;
    STheta2 = 0.1;
    SDelta1 = 0.2;
    SDelta2 = 0.2;
    SNoiseLevel = 1.0;
    SNoiseSelected = false;
    targetStructLoaded = false;
    structInversionUpToDate = true;
    meanDensityEstimateUpToDate = true;
    acousticTimeEstimateUpToDate = true;
    smallSeparationEstimateUpToDate = true;

    // variables relating to mean density estimate:
    kernelTypeME = 0;
    n_surfME = 0;
    nminME = 0;
    expME = 4.9;
    MEBeta = 1e-6;
    METheta = 1e-2;
    MENoiseLevel = 1.0;
    MENoiseSelected = false;
    
    // variables relating to acoustic radius estimate:
    kernelTypeAR = 0;
    n_surfAR = 0;
    nminAR = 0;
    ARBeta = 1e-6;
    ARTheta = 1e-2;
    ARNoiseLevel = 1.0;
    ARNoiseSelected = false;

    // variables relating to acoustic radius estimate:
    n_surfSSep = 0;
    nminSSep = 0;
    SSBeta = 1e-2;
    SSTheta = 1e-8;
    SSNoiseLevel = 1.0;
    SSNoiseSelected = false;

    // other variables:
    modelLoaded = false;
    DF = 0.0;

    // set current directory:
    if (!isApplet)
      currentDirectory = new File(System.getProperty("user.dir"));

    // set isApplet to true if in restricted mode:
    if (isApplet) {
      try {
        AccessController.checkPermission(new FilePermission("*","read"));
        this.isApplet = false;
      } catch(Exception e) {
        this.isApplet = true;
      }
    } else {
      this.isApplet = false;
    }

    // initialise gravitational constant, solar mass, and solar radius
    BufferedReader in;
    String line;
    if (!isApplet) {
      File f = new File(System.getProperty("user.home"),".InversionKit");
      if (f.exists()) {
        try {
          in = Utilities.OpenFile(f.getAbsolutePath());
          line = Utilities.trim(in.readLine());
          G = Utilities.parseDouble(line);
          line = Utilities.trim(in.readLine());
          SolarMass = Utilities.parseDouble(line);
          line = Utilities.trim(in.readLine());
          SolarRadius = Utilities.parseDouble(line);
          in.close();
          return;
        } catch(Exception e) {
          // go on to reading default values in /resource_files/constants
        }
      }
    }
    URL url = getClass().getResource("/resource_files/constants");
    try {
      in = new BufferedReader(new InputStreamReader(url.openStream()));
      line = Utilities.trim(in.readLine());
      G = Utilities.parseDouble(line);
      line = Utilities.trim(in.readLine());
      SolarMass = Utilities.parseDouble(line);
      line = Utilities.trim(in.readLine());
      SolarRadius = Utilities.parseDouble(line);
      in.close();
    } catch (Exception e) {
      // if all else fails, manually insert default values
      // (this should not happen)
      G = 6.67232e-08; // gravitational constant (in cm^3.g^-1.s^-2)
      SolarMass = 1.989e+33; // solar mass (in g)
      SolarRadius = 6.9599e+10; // solar radius (in cm)
    }
  }

  /**********************************************************************
  * This tells whether the program is running as an applet in restricted
  * mode (isApplet == true) or as an unrestricted applet or a 
  * stand-alone application (isApplet == false).
  **********************************************************************/
  public boolean isApplet;

  /**********************************************************************
  * This keeps track of the current directory, which is used when opening
  * and saving files.
  **********************************************************************/
  public File currentDirectory;

  ///////////////////////////////////////////////////////////////////////
  // Information on pulsations modes:
  ///////////////////////////////////////////////////////////////////////

  /** ArrayList of modes (see Mode class). */
  public ArrayList<Mode> modes = new ArrayList<Mode>();
  /** ArrayList of eigenfunctions and kernels (see kernel class). */
  public ArrayList<Kernel> kernels = new ArrayList<Kernel>();
  /** Number of frequencies AND associated error bars. */
  public int nfreq = 0;
  /** Number of rotational splittings. */
  public int nRS = 0;
  /** Large frequency separation used in the echelle diagram */
  public double DF;
  /** This keeps track of the kernels table on the kernels page,
   *  so that actions on other tabs may update this table. */
  public JTable kernelTable;
  /** This keeps track of the frequency data table on the frequencies page,
   *  so that actions on other tabs may update this table. */
  public JTable frequenciesTable;
  /** This is to keep track of the frequencies data model, so as to
   *  be able to calculate new "observed" frequencies using theoretical
   *  frequency shifts or rotational splittings */
  public FrequenciesDataModel frequenciesData;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables relating to the model:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  
  /** This variable is set to 'true' when a stellar model is loaded and
   *  'false' otherwise. */
  public boolean modelLoaded;
  /** Underlying grid (normalised) */
  public double[] grid;
  /** The stellar radius (in cm) */
  public double StellarRadius;
  /** The stellar mass (in g) */
  public double StellarMass;
  /** Reference value for the pressure (= GM^2/R^4). 
   *  Pref*pressure[] = pressure (in g.cm^-1.s^-2 or dyn.cm^-2) */
  public double Pref;
  /** Reference value for the density (= M/R^3).
   *  RhoRef*rho[] = density (in g.cm^-3) */
  public double RhoRef;
  /** Reference value for frequencies (= sqrt(GM/R^3), in cyclic.muHz) */
  public double FreqRef;
  /** Normalised second derivative of density profile, at the center */
  public double ddrho;
  /** Dynamical Time */
  public double TauRef;
  /** Normalised value for the acoustic radius.
   *  TauRef*Tau = acoustic radius (in s) */
  public double Tau;
  /** Index of nearest point to photometric radius */
  public int indSurface;

  /** Gamma1 profile. Gamma1 is the adiabatic exponent. */
  public double[] Gamma1;
  /** Normalised density profile */
  public double[] rho;
  /** Normalised radial derivative of density profile */
  public double[] drho;
  /** Normalised pressure profile*/
  public double[] pressure;
  /** Normalised squared sound velocity profile */
  public double[] c2;
  /** Normalised squared Brunt-Vaisala frequency */
  public double[] N2;
  /** Normalised cumulative mass m(r)/M_total */
  public double[] mass;
  /** Normalised gravity profile */
  public double[] gravity;
  /** Normalised radial derivative of the gravity profile */
  public double[] dgravity;
  /** Relative difference between the local and average densities, 
   *  divided by the normalised grid squared */
  public double[] diffrho_r2;
  /** Variables which intervene in JCD's version of the pulsation equations */
  public double[] U, A, Vg, eta;
  /** integrand which intervenes in the small frequency separation
   *  and is used to estimate the stellar age */
  public double[] agesmallsep2;
  /** modified age indicator: variable representing the integral */
  public double agesmallsep;

  /////////////////////////////////////////////////////////////////////////////
  /** This rescales a model using a specified mass and radius ratio. */
  /////////////////////////////////////////////////////////////////////////////
  public void rescale(double Mratio, double Rratio) {
    StellarRadius*=Rratio;
    StellarMass*=Mratio;
    Pref*=(Math.pow(Mratio,2.0)/Math.pow(Rratio,4.0));
    RhoRef*=(Mratio/Math.pow(Rratio,3.0));
    FreqRef*=Math.sqrt(Mratio/Math.pow(Rratio,3.0));
    TauRef*=Math.sqrt(Math.pow(Rratio,3.0)/Mratio);
  }

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Physical constants:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** The gravitational constant (in cm^3.g^-1.s^-2) */
  public double G;
  /** A label in the ConstantsPage with the current value of G */
  public JLabel GravitationalCurrentValue;
  /** The solar mass in g */
  public double SolarMass;
  /** The solar radius in cm */
  public double SolarRadius;

  /** Solar large frequency separation (in microHz) */
  public static final double SolarDF = 135.1;
  /** Frequency at maximum power for solar pulsation spectrum */
  public static final double SolarNuMax = 3090;
  /** Solar effective temperature */
  public static final double SolarTeff = 5777.54;
  /** The solar luminosity in erg/s */
  public static final double Lsun = 3.846e33;
  /** Stefan-Boltzmann constant (ac/4) in erg.cm-2.K-4.s-1 */
  public static final double Sigma = 5.67051e-5;
  /** Natural logarithm of 10 (useful for finding log base 10 of something) */
  public static final double ln10 = 2.302585092994046;

  /////////////////////////////////////////////////////////////////////////////
  /** This modifies the value of G and updates model reference values and
   *  profiles accordingly. */
  /////////////////////////////////////////////////////////////////////////////
  public void modifyG(double NewG) {
    double Gratio = NewG/G;
    G = NewG;
    GravitationalCurrentValue.setText(String.format("Current value: %.7e (cm\u00B3.g\u207B\u00B9.s\u207B\u00B2)",G));

    if (!modelLoaded) return;
    Pref*=(Gratio);
    FreqRef*=Math.sqrt(Gratio);
    TauRef/=Math.sqrt(Gratio);
    int ngrid = grid.length;
    for(int i=0; i<ngrid; i++) {
      pressure[i]/=Gratio;
      c2[i]/=Gratio;
      // The gravity profile and associated quantites (dgravity, N2)
      // should not be modified, since G factors out of its
      // non-dimensional form.
    }
  }

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables relating to the rotation inversion:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable states whether or not the rotational inversion is up
  *   to date or not (i.e. that a suitable list of frequency shifts and
  *   kernels  exists and agrees with current variable contents). */
  public boolean rotationInversionUpToDate;
  /** This contains the grid used in the inversion.  It doesn't need
  *   to contain more than 200 points and should be spaced such that
  *   c(r)Delta(r) = cnst */
  public double[] gridRInversion;
  /** The number of points used in the above grid */
  public int n_gridR;
  /** This regulates the trade-off between an exact fit to data and
  *   regularisation of the solution in the RLS method. */
  public double RLambda;
  /** This regulates the trade-off between fitting to the data points
  *   data points and regularising the solution in the SOLA method. */
  public double RTheta;
  /** This regulates the width of the Gaussian target functions
  *   in the SOLA method. */
  public double RDelta;   
  /** This regulates the trade-off between fitting to the data points
  *   data points and regularising the solution in the MOLA method. */
  public double RThetaM;
  /** If this parameter is set to true, than artiticial noise will
  *   be added to the data and used in the inversion. */
  public boolean RNoiseSelected;
  /** This regulates the noise level */
  public double RNoiseLevel;
  /** This contains artificial noise to be added to the data */
  public double[] RNoise;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables relating to the structural inversion:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable states whether or not the structural inversion
  *   is up to date or not (i.e. that a suitable list of frequency shifts
  *   and kernels exists and agrees with current variable contents). */
  public boolean structInversionUpToDate;
  /** This contains the grid used in the inversion.  It doesn't need
  *   to contain more than 200 points and should be spaced such that
  *   c(r)Delta(r) = cnst */
  public double[] gridSInversion;
  /** The number of points used in the above grid */
  public int n_gridS;
  /** Number of Legendre polynomials to be used in modelling surface terms */
  public int n_surf;
  /** This regulates the trade-off between an exact fit to data and
  *   regularisation of the solution in the RLS method. */
  public double SLambda;
  /** This regulates the trade-off between fitting to the data points
  *   data points and regularising the solution in the SOLA method. */
  public double STheta1;
  public double STheta2;
  /** This regulates the width of the Gaussian target functions
  *   in the SOLA method. */
  public double SDelta1;
  public double SDelta2;
  /** This regulates the balance between the obtaining "nice" averaging
  *   kernels and reducing cross-term kernels (see e.g.
  *   Rabello-Soares et al., 1999) */
  public double Beta1;
  public double Beta2;
  /** If this parameter is set to true, than artiticial noise will
  *   be added to the data and used in the inversion. */
  public boolean SNoiseSelected;
  /** This regulates the noise level */
  public double SNoiseLevel;
  /** This contains artificial noise to be added to the data */
  public double[] SNoise;

  /** The following string tells type of kernels to use */
  public int kernelType;
  /** The list of different kernel types */
  public static final String[] kernelTypes = {"\u03C1,c\u00B2 kernels",
                                              "\u03C1,\u0393\u2081 kernels",
                                              "\u03C1,c\u00B2 kernels - constant mass",
                                              "\u03C1,\u0393\u2081 kernels - constant mass"};
  /** The names of the different structural functions involved in the inversions */
  public static final String[][] structNames = {{"\u03B4\u03C1/\u03C1","\u03B4c\u00B2/c\u00B2"},
                                                {"\u03B4\u03C1/\u03C1","\u03B4\u0393\u2081/\u0393\u2081"},
                                                {"\u03B4\u03C1/\u03C1","\u03B4c\u00B2/c\u00B2"},
                                                {"\u03B4\u03C1/\u03C1","\u03B4\u0393\u2081/\u0393\u2081"}};

  /** This keeps track of whether an extra constraint which keeps the
  *   mass of the star constant is needed or not. 0 means no contraints,
  *   1 means a constraint on the first structural inversion, 2 means
  *   a contraint on the second structural inversion, and 3 means a
  *   constraint on both inversions*/
  public int constantMass;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables relating to the mean density estimate
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable states whether or not the mean density estimate
  *   is up to date or not (i.e. that a suitable list of frequency shifts
  *   and kernels exists and agrees with current variable contents). */
  public boolean meanDensityEstimateUpToDate;
  /** Number of Legendre polynomials to be used in modelling surface terms */
  public int n_surfME;
  /** Minimal radial order used when applying differential scaling law */
  public int nminME;
  /** Exponent used in Kjeldsen et al. (2008) surface-correcting power law */
  public double expME;
  /** This regulates the trade-off between fitting to the data points
  *   data points and regularising the solution in the SOLA method. */
  public double METheta;
  /** This regulates the balance between the obtaining "nice" averaging
  *   kernels and reducing cross-term kernels (see e.g.
  *   Rabello-Soares et al., 1999) */
  public double MEBeta;
  /** If this parameter is set to true, than artiticial noise will
  *   be added to the data and used in the inversion. */
  public boolean MENoiseSelected;
  /** This regulates the noise level */
  public double MENoiseLevel;
  /** This contains artificial noise to be added to the data */
  public double[] MENoise;
  /** The list of different kernel types */
  public static final String[] kernelTypesME = {"\u03C1,\u0393\u2081 kernels",
                                                "\u03C1,c\u00B2 kernels"};
  /** The following string tells type of kernels to use */
  public int kernelTypeME;
  /** This gives the exponent for the scale factor for the second
   *  structural function, assuming that the radius is fixed. */
  public double scalePowerME;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables relating to the small separation estimate
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable states whether or not the stellar age estimate (small separation)
  *   is up to date or not (i.e. that a suitable list of frequency shifts
  *   and kernels exists and agrees with current variable contents). */
  public boolean smallSeparationEstimateUpToDate;
  /** Number of Legendre polynomials to be used in modelling surface terms */
  public int n_surfSSep;
  /** Minimal radial order used when applying differential scaling law */
  public int nminSSep;
  /** This regulates the trade-off between fitting to the data points
  *   data points and regularising the solution in the SOLA method. */
  public double SSTheta;
  /** This regulates the balance between the obtaining "nice" averaging
  *   kernels and reducing cross-term kernels (see e.g.
  *   Rabello-Soares et al., 1999) */
  public double SSBeta;
  /** If this parameter is set to true, than artiticial noise will
  *   be added to the data and used in the inversion. */
  public boolean SSNoiseSelected;
  /** This regulates the noise level */
  public double SSNoiseLevel;
  /** This contains artificial noise to be added to the data */
  public double[] SSNoise;
  /** This gives the exponent for the scale factor for the second
   *  structural function, assuming that the radius is fixed. */
  public double scalePowerSSep;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables relating to the acoustic radius estimate
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  /** This variable states whether or not the acoustic radius estimate
  *   is up to date or not (i.e. that a suitable list of frequency shifts
  *   and kernels exists and agrees with current variable contents). */
  public boolean acousticTimeEstimateUpToDate;
  /** Number of Legendre polynomials to be used in modelling surface terms */
  public int n_surfAR;
  /** Minimal radial order used when applying differential scaling law */
  public int nminAR;
  /** This regulates the trade-off between fitting to the data points
  *   data points and regularising the solution in the SOLA method. */
  public double ARTheta;
  /** The list of different kernel types */
  public static final String[] kernelTypesAR = {"\u03C1,\u0393\u2081 kernels",
                                                "\u03C1,c\u00B2 kernels"};
  /** The following string tells type of kernels to use */
  public int kernelTypeAR;
  /** This regulates the balance between the obtaining "nice" averaging
  *   kernels and reducing cross-term kernels (see e.g.
  *   Rabello-Soares et al., 1999) */
  public double ARBeta;
  /** If this parameter is set to true, than artiticial noise will
  *   be added to the data and used in the inversion. */
  public boolean ARNoiseSelected;
  /** This regulates the noise level */
  public double ARNoiseLevel;
  /** This contains artificial noise to be added to the data */
  public double[] ARNoise;
  /** This gives the exponent for the scale factor for the second
   *  structural function, assuming that the radius is fixed. */
  public double scalePowerAR;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Target functions:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable is set to 'true' when a targetRota function is
  *   loaded. */
  public boolean targetRotaLoaded;
  /** This contains an optional target rota function which can be compared
  *   with the inversion results, or used to calculate theoretical
  *   frequency shifts */
  public double[] targetRota;
  /** x grid corresponding to the target rota function */
  public double[] gridTargetRota;
  /** This variable is set to 'true' when a target structural functions are
  *   loaded. */
  public boolean targetStructLoaded;
  /** This contains an optional target struct1 function which can be compared
  *   with the inversion results, or used to calculate theoretical
  *   frequency shifts */
  public double[] targetStruct1;
  /** This contains an option target struct2 function which can be compared
  *   with the inversion results, or used to calculate theoretical
  *   frequency shifts */
  public double[] targetStruct2;
  /** x grid corresponding to the target structural functions */
  public double[] gridTargetStruct;
  /**Target rho */
  public double[] targetrho;
  /**Target Gamma1 */
  public double[] targetGamma1;
  /**Target c2 */
  public double[] targetc2;
  /**Target u0 */
  public double[] targetu0;
  /**Target meandensity*/
  public double targetRhoRef;
  /**Target acoustic radius*/
  public double targetTau;
  /**Target small separation age indicator*/
  public double targetagesmallsep;
  /**Target age indicator*/
  public double targetageindic;
  /**Target frequencies*/
  public double[] targetfreq;
  /**Target adimensional frequencies*/
  public double[] targetfreqDimLess;
  /**Target radial orders of the frequencies*/
  public int[] targetn;
  /**Target degrees of the frequencies*/
  public int[] targetl;
  /**This variable is set to 'true' when the target frequencies are loaded*/
  public boolean targetFreqLoaded;
  

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Various utility methods:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /******************************************************************************
  * Find Delta F, using a least-squares method based on the radial orders.  If
  * the radial orders have not been calculated, then default to the alternative
  * method findDFsimple().
  ******************************************************************************/
  public void findDF() {
    DF = 0.0;
    if (modes.size()<2) return;
    int i, lmax = -1;

    for(i=0; i<modes.size(); i++) {
      if (modes.get(i).l == null) continue;
      if (modes.get(i).FreqObs == null) continue;
      if (lmax < modes.get(i).l.intValue()) lmax = modes.get(i).l.intValue();
    }

    // there are no valid (l,nu) pairs - give up:
    if (lmax == -1) return;

    int[] larray = new int[lmax+1];
    for(i=0;i<=lmax;i++) larray[i] = 0;
    for(i=0;i<modes.size(); i++) { 
      if (modes.get(i).kernel != null) larray[modes.get(i).l.intValue()]++;
    }

    boolean applyAlternativeMethod = true;
    for(i=0;i<=lmax;i++) { 
      if (larray[i]>1) {
        applyAlternativeMethod = false;
        break;
      }
    }

    // an insufficient number of radial orders are available:
    // apply alternative method
    if (applyAlternativeMethod) {
      findDFsimple();
      return;
    }

    // apply the main method:
    // find Delta F for each l value, using a least-squares approach:
    double[] matrix = new double[3];
    double[] v      = new double[2];
    double ntotal = 0; // total number of data (used for weighting purposes)
    for(int l=0; l<=lmax; l++) {
      if (larray[l] < 2) continue;
      matrix[0]=matrix[1]=matrix[2]=0.0;
      v[0]=v[1]=0.0;
      for(i=0; i<modes.size(); i++) {
        if (modes.get(i).kernel == null) continue;
        if (modes.get(i).l.intValue() != l) continue;
        matrix[0]+=(double)modes.get(i).kernel.n*modes.get(i).kernel.n;
        matrix[1]+=(double)modes.get(i).kernel.n;
        matrix[2]+=1.0;
        v[0]+=modes.get(i).FreqObs*(double)modes.get(i).kernel.n;
        v[1]+=modes.get(i).FreqObs;
      }
      DF+=((double)larray[l]-1)*(v[0]*matrix[2]-v[1]*matrix[1])
         /(matrix[0]*matrix[2]-matrix[1]*matrix[1]);
      ntotal+=((double)(larray[l]-1));
    }
    // calculate weighted average of DF for each l:
    DF/=ntotal;
  }

  /******************************************************************************
  * Alternative method for finding Delta F.  In this approach, temporary (and
  * false) radial orders are assigned to the l=0 modes.  Delta F is then
  * calculated based on this.
  ******************************************************************************/
  public void findDFsimple() {
    int i, istart, ifinish;

    DF = 0.0;
    // sort modes according to ell and frequency:
    int[] indexFreq = sortModes();
    frequenciesTable.updateUI();

    // count number of ell=0 modes:
    istart = 0;
    while(istart<modes.size()) {
      if (modes.get(istart).l != null) {
        if ((modes.get(istart).l.intValue() == 0)&&(modes.get(istart).FreqObs != null)) {
          break;
        }
      }
      istart++;
    }

    // no radial modes (incl. frequency): give up
    if (istart==modes.size()) return; 

    ifinish = istart;
    while(ifinish<modes.size()) {
      if (modes.get(ifinish).l == null) break; // this shouldn't happen
      if (modes.get(ifinish).l.intValue() != 0) break;
      if (modes.get(ifinish).FreqObs == null) break;
      ifinish++;
    }

    // insufficient number of radial modes (incl. frequency): give up
    if ((ifinish-istart) < 2) return; 

    // find minimum separation between ell=0 modes:
    double diff, mindiff = modes.get(istart+1).FreqObs-modes.get(istart).FreqObs;
    for(i=istart+2;i<ifinish;i++) {
      diff = modes.get(i).FreqObs - modes.get(i-1).FreqObs;
      if (diff < mindiff) mindiff = diff;
    }

    // assign temporary n values:
    double[] nvalue = new double[ifinish];
    nvalue[istart] = 0.0;
    for(i=istart+1;i<ifinish;i++) {
      diff = (modes.get(i).FreqObs - modes.get(i-1).FreqObs)/mindiff;
      nvalue[i] = nvalue[i-1] + (double)Math.round(diff);
    }

    // find DF through least-squares method: 
    double[] matrix = new double[3];
    double[] v      = new double[2];
    for(i=istart; i<ifinish; i++) {
      matrix[0]+=nvalue[i]*nvalue[i];
      matrix[1]+=nvalue[i];
      matrix[2]+=1.0;
      v[0]+=modes.get(i).FreqObs*nvalue[i];
      v[1]+=modes.get(i).FreqObs;
    }

    // get the final result:
    DF = (v[0]*matrix[2]-v[1]*matrix[1])
       / (matrix[0]*matrix[2]-matrix[1]*matrix[1]);
  }

  /******************************************************************************
  * This method finds the matches between the observed modes and the theoretical
  * ones.  It clears any previous matches before carrying out this operation.
  * It also updates the UI of both kernelTable and frequenciesTable.
  ******************************************************************************/
  public void findMatch() {
    int i,j;
    double diffmin, diff;
    for(i=0; i<modes.size(); i++)   modes.get(i).kernel = null;
    for(i=0; i<kernels.size(); i++) kernels.get(i).mode = null;
    // sweep through observd frequencies
    for(i=0; i<modes.size(); i++) {
      if (modes.get(i).l == null) continue;
      if (modes.get(i).FreqObs == null) continue;
      diffmin = Double.MAX_VALUE;
      for(j=0; j<kernels.size(); j++) {
        if (!modes.get(i).l.equals(kernels.get(j).l)) continue;
        diff = Math.abs(modes.get(i).FreqObs-kernels.get(j).FreqTheo);
        if (diff < diffmin) {
          diffmin = diff;
          // erase previous match (if need be):
          if (modes.get(i).kernel != null) {
            modes.get(i).kernel.mode = null;
            modes.get(i).kernel = null;
          }
          // make new match (if possible):
          if (kernels.get(j).mode == null) {
            modes.get(i).kernel = kernels.get(j);
            kernels.get(j).mode = modes.get(i);
          } else {
            diff = Math.abs(kernels.get(j).mode.FreqObs-kernels.get(j).FreqTheo);
            if (diff > diffmin) {
              kernels.get(j).mode.kernel = null;
              modes.get(i).kernel = kernels.get(j);
              kernels.get(j).mode = modes.get(i);
            }
          }
        }
      }
    }
    //testMatch();
    kernelTable.updateUI();
    frequenciesTable.updateUI();
    rotationInversionUpToDate = false;
    structInversionUpToDate = false;
    meanDensityEstimateUpToDate = false;
    acousticTimeEstimateUpToDate = false;
    smallSeparationEstimateUpToDate = false;
  }

  /******************************************************************************
  * This is a private internal test of the consistency of the kernels and modes
  * matching.  This is only to be used for the purposes of debugging.
  ******************************************************************************/
  private void testMatch() {
    int i, j, count = 0;
    for(i=0; i<modes.size(); i++) {
      if (modes.get(i).kernel != null) count++;
    }
    System.out.println("n modes -> kernels: "+count);
    for(i=0; i<modes.size(); i++) {
      if (modes.get(i).kernel != null) {
        if (modes.get(i).kernel.mode != modes.get(i)) {
          System.out.println("mode warning");
        }
      }
    }
    count = 0;
    for(i=0; i<kernels.size(); i++) {
      if (kernels.get(i).mode != null) count++;
    }
    System.out.println("n kernels -> modes: "+count);
    for(i=0; i<kernels.size(); i++) {
      if (kernels.get(i).mode != null) {
        if (kernels.get(i).mode.kernel != kernels.get(i)) {
          System.out.println("kernel warning");
        }
      }
    }
  }

  /******************************************************************************
  * This method clears the matches between the observed modes and the theoretical
  * ones.  It also updates the UI of both kernelTable and frequenciesTable.
  ******************************************************************************/
  public void clearMatch() {
    int i;
    for(i=0; i<modes.size(); i++)   modes.get(i).kernel = null;
    for(i=0; i<kernels.size(); i++) kernels.get(i).mode = null;
    kernelTable.updateUI();
    frequenciesTable.updateUI();
    rotationInversionUpToDate = false;
    structInversionUpToDate = false;
    meanDensityEstimateUpToDate = false;
    acousticTimeEstimateUpToDate = false;
    smallSeparationEstimateUpToDate = false;
  }

  /******************************************************************************
  * This finds duplicate kernels by searching for kernels with the same (l,n)
  * values.  This can be done efficiently by scanning through a sorted list of
  * kernels.  Duplicate kernels are kept track of thanks to the "modes.isDuplicate"
  * array.  Only the mode for which the observed frequency is closest to the
  * theoretical result is kept.
  ******************************************************************************/
  public void findDuplicates() {
    if (modes.size() < 1) return;
    int[] index = sortModes();
    int i, j, k, ibest;
    double diff, diffBest;
    i = 0;
    while(i<(modes.size()-1)) {
      if (modes.get(index[i]).kernel == null) {
        modes.get(index[i]).isDuplicate = false;
        i++;
        continue;
      }
      j=i+1;
      while(j<modes.size()) {
        if (modes.get(index[j]).l == null) break;
        if (modes.get(index[j]).kernel == null) break;
        if (Utilities.compareIntegers(modes.get(index[i]).l,modes.get(index[j]).l)!=0) break;
        if (modes.get(index[i]).kernel.n != modes.get(index[j]).kernel.n) break;
        j++;
      }
      ibest = i;
      diffBest = Math.abs(modes.get(index[i]).kernel.FreqTheo-modes.get(index[i]).FreqObs);
      modes.get(index[i]).isDuplicate = true;
      for(k=i+1; k<j; k++) {
        diff = Math.abs(modes.get(index[k]).kernel.FreqTheo-modes.get(index[k]).FreqObs);
        modes.get(index[k]).isDuplicate = true;
        if (diff < diffBest) {
          ibest = k;
          diffBest = diff;
        }
      }
      modes.get(index[ibest]).isDuplicate = false;
      i=j;
    }
    frequenciesTable.updateUI();
    rotationInversionUpToDate = false;
    structInversionUpToDate = false;
    meanDensityEstimateUpToDate = false;
    acousticTimeEstimateUpToDate = false;
    smallSeparationEstimateUpToDate = false;
  }

  /******************************************************************************
  * This clears the isDuplicate fields in the modes.
  ******************************************************************************/
  public void clearDuplicates() {
    for(int i=0; i<modes.size(); i++) modes.get(i).isDuplicate = false;
    frequenciesTable.updateUI();
    rotationInversionUpToDate = false;
    structInversionUpToDate = false;
    meanDensityEstimateUpToDate = false;
    acousticTimeEstimateUpToDate = false;
    smallSeparationEstimateUpToDate = false;
  }

  /******************************************************************************
  * This method clears the modes and matches between the observed and theoretical
  * modes.  It also updates the UI of both kernelTable and frequenciesTable.
  ******************************************************************************/
  public void clearModes() {
    for(int i=0; i<kernels.size(); i++) kernels.get(i).mode = null;
    modes.clear();
    kernelTable.clearSelection();
    kernelTable.updateUI();
    frequenciesTable.clearSelection();
    frequenciesTable.updateUI();
    rotationInversionUpToDate = false;
    structInversionUpToDate = false;
    meanDensityEstimateUpToDate = false;
    acousticTimeEstimateUpToDate = false;
    smallSeparationEstimateUpToDate = false;
  }

  /******************************************************************************
  * This method clears the kernels and matches between the observed and theoretical
  * modes.  It also updates the UI of both kernelTable and frequenciesTable.
  ******************************************************************************/
  public void clearKernels() {
    for(int i=0; i<modes.size(); i++) {
      modes.get(i).kernel = null;
      modes.get(i).isDuplicate = false;
    }
    kernels.clear();
    kernelTable.clearSelection();
    kernelTable.updateUI();
    frequenciesTable.clearSelection();
    frequenciesTable.updateUI();
    rotationInversionUpToDate = false;
    structInversionUpToDate = false;
    meanDensityEstimateUpToDate = false;
    acousticTimeEstimateUpToDate = false;
    smallSeparationEstimateUpToDate = false;
  }

  /******************************************************************************
  * This method produces an index array such that modes[index[i]] is sorted
  * according to l and n (or freq if n is unknown).  The algorithm is based 
  * on the heapsort method.  It is copied and adapted from:
  *
  * "Numerical Recipes in Fortran: The Art of Scientific Computing"
  * Second Edition.
  * William H. Press, Saul A. Teukolsky, William T. Vetterling, Brian P. Flannery
  * Cambridge University Press, 1992.
  ******************************************************************************/

  private int[] sortModes() {
    int n = modes.size();
    int[] index = new int[n];

    // Initialise the index array
    for(int i=0; i<n; i++) index[i] = i;
    if (n < 2) return index;
    
    // The index l will be decremented from its initial value down to 0
    // during the "hiring" (heap creation) phase.  Once it reaches 0, the
    // index ir will be decremented from its initial value down to 0 during
    // the "retirement-and-promotion" (heap selection) phase.

    int l = n/2;
    int ir = n-1;
    int i, j, iindex;
    while(true) {
      if (l > 0) {                    // Still in hiring phase
        l--;
        iindex = index[l];
      } else {                        // In retirement-and-promotion phase.
        iindex = index[ir];           // Clear a space at end of array.
        index[ir] = index[0];         // Retire the top of the heap into it
        ir--;                         // Decrease the size of the corporation.
        if(ir == 0) {                 // Done with the last promotion.
          index[0] = iindex;          // The least competent worker of all!
          return index;
        }
      }
      i=l;                            // Whether in the hiring phase or
                                      // promotion phase, we here set up to
                                      // sift down element rra to its proper
                                      // level.
      j=l+l+1;
      while (j <= ir) {
        if (j < ir) {
          if (Mode.compareModes(modes.get(index[j]),modes.get(index[j+1]))) j++;
                                      // Compare to the better underling
        }
        if (Mode.compareModes(modes.get(iindex),modes.get(index[j]))) {
          index[i]=index[j];          // Demote ra(iindex)
          i=j;
          j=j+j+1;
        } else {                      // This is rra's level.  Set j to
                                      // terminate the sift-down.
          j=ir+1;
        }
      }
      index[i] = iindex;              // Put rra into its slot.
    }
  }
  
  /******************************************************************************
  * This method sorts the observations according to l and frequency, using the
  * heapsort method.  The heapsort algorithm is copied and adapted from:
  *
  * "Numerical Recipes in Fortran: The Art of Scientific Computing"
  * Second Edition.
  * William H. Press, Saul A. Teukolsky, William T. Vetterling, Brian P. Flannery
  * Cambridge University Press, 1992.
  ******************************************************************************/
  public void sortObs() {
    int n = modes.size();
    if (n < 2) return;

    // The index l will be decremented from its initial value down to 0
    // during the "hiring" (heap creation) phase.  Once it reaches 0, the
    // index ir will be decremented from its initial value down to 0 during
    // the "retirement-and-promotion" (heap selection) phase.

    int l = n/2;
    int ir = n-1;
    int i, j;
    Mode amode;
    while(true) {
      if (l > 0) {                    // Still in hiring phase
        l--;
        amode = modes.get(l);
      } else {                        // In retirement-and-promotion phase.
        amode = modes.get(ir);        // Clear a space at end of array.
        modes.set(ir,modes.get(0));   // Retire the top of the heap into it
        ir--;                         // Decrease the size of the corporation.
        if(ir == 0) {                 // Done with the last promotion.
          modes.set(0,amode);         // The least competent worker of all!
          break;
        }
      }
      i=l;                            // Whether in the hiring phase or
                                      // promotion phase, we here set up to
                                      // sift down element rra to its proper
                                      // level.
      j=l+l+1;
      while (j <= ir) {
        if (j < ir) {
          if (Mode.compareObs(modes.get(j),modes.get(j+1))) j++;
                                      // Compare to the better underling
        }
        if (Mode.compareObs(amode,modes.get(j))) {
          modes.set(i,modes.get(j));  // Demote ra(iindex)
          i=j;
          j=j+j+1;
        } else {                      // This is rra's level.  Set j to
                                      // terminate the sift-down.
          j=ir+1;
        }
      }
      modes.set(i,amode);             // Put rra into its slot.
    }
  }

  /******************************************************************************
  * This method sorts the observations according to l and frequency, using the
  * heapsort method.  The heapsort algorithm is copied and adapted from:
  *
  * "Numerical Recipes in Fortran: The Art of Scientific Computing"
  * Second Edition.
  * William H. Press, Saul A. Teukolsky, William T. Vetterling, Brian P. Flannery
  * Cambridge University Press, 1992.
  ******************************************************************************/
  public void sortKernels() {
    int n = kernels.size();
    if (n < 2) return;

    // The index l will be decremented from its initial value down to 0
    // during the "hiring" (heap creation) phase.  Once it reaches 0, the
    // index ir will be decremented from its initial value down to 0 during
    // the "retirement-and-promotion" (heap selection) phase.

    int l = n/2;
    int ir = n-1;
    int i, j;
    Kernel aKernel;
    while(true) {
      if (l > 0) {                    // Still in hiring phase
        l--;
        aKernel = kernels.get(l);
      } else {                        // In retirement-and-promotion phase.
        aKernel = kernels.get(ir);        // Clear a space at end of array.
        kernels.set(ir,kernels.get(0));   // Retire the top of the heap into it
        ir--;                         // Decrease the size of the corporation.
        if(ir == 0) {                 // Done with the last promotion.
          kernels.set(0,aKernel);         // The least competent worker of all!
          break;
        }
      }
      i=l;                            // Whether in the hiring phase or
                                      // promotion phase, we here set up to
                                      // sift down element rra to its proper
                                      // level.
      j=l+l+1;
      while (j <= ir) {
        if (j < ir) {
          if (Kernel.compareKernels(kernels.get(j),kernels.get(j+1))) j++;
                                      // Compare to the better underling
        }
        if (Kernel.compareKernels(aKernel,kernels.get(j))) {
          kernels.set(i,kernels.get(j));  // Demote ra(iindex)
          i=j;
          j=j+j+1;
        } else {                      // This is rra's level.  Set j to
                                      // terminate the sift-down.
          j=ir+1;
        }
      }
      kernels.set(i,aKernel);             // Put rra into its slot.
    }
  }
}
