/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) 2009, Daniel Reese, Sergei Zharkov, 2014
*
****************************************************************************/
package inversion_applet;

import graph.IPlot;
import graph.Plot;
import inversion_applet.DataCenter;
import inversion_applet.Utilities;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.SwingConstants;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.JTextArea;
import javax.swing.JSlider;
import javax.swing.JButton;
import javax.swing.JCheckBox;

public class EchellePage extends JPanel {
  private DataCenter dataCenter;

  // this contains all the controls:
  private JPanel     buttons;

  // this contains the refresh and optimise buttons:
  private JPanel     buttonBox;
  private JButton    RefreshButton;
  private JButton    OptimiseButton;
  private JCheckBox  errorCheckBox;

  // this contains the controls and labels which adjust DF and the offset
  private JPanel     DfPanel;
  private JPanel     DfTitles;
  private JPanel     DfMiniPanel;
  private JButton    ReduceButton;
  private JButton    IncreaseButton;
  private JTextField DfTextField;
  private JSlider    DfSlider;
  private JSlider    OffsetSlider;
  private double     DfCenter;
  private double     Offset;

  // this will contain the legend
  private JPanel     LegendPanel;

  // this is the plot with the echelle diagram:
  private IPlot      EchelleDiagram;

  // this sets the colors of the observed and theoretical frequencies:
  private final Color ColorObs  = Color.blue;
  private final Color ColorTheo = Color.red;

  /////////////////////////////////////////////////////////////////////////////
  /** A constructor.
   *
   *  @param dataCenter the data center shared by all of the components of
   *                    of this program */
  /////////////////////////////////////////////////////////////////////////////
  public EchellePage(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This is what sets up the buttons and the layout */
  /////////////////////////////////////////////////////////////////////////////
  public void initProgram() {
    initButtonBox();
    initDfControls();
    initLegend();

    EchelleDiagram = new IPlot("Echelle Diagram",
                               "Reduced frequency",
                               "Frequency",24);

    EchelleDiagram.setIDLTitles("!17Echelle Diagram",
                                "!17Reduced frequency",
                                "!17Frequency");

    EchelleDiagram.setIOenabled(!dataCenter.isApplet);

    setLayout(new BorderLayout());
    buttons = new JPanel();
    buttons.setLayout(new FlowLayout(FlowLayout.CENTER));
    buttons.add(buttonBox);
    buttons.add(DfTitles);
    buttons.add(DfPanel);
    buttons.add(LegendPanel);
    
    add("Center",EchelleDiagram);
    add("South",buttons);
  }

  //////////////////////////////////////////////////////////////////////////////
  // This initialises the Refresh and Optimise buttons
  //////////////////////////////////////////////////////////////////////////////
  private void initButtonBox() {
    buttonBox = new JPanel();
    buttonBox.setLayout(new GridLayout(3,1));
    RefreshButton = new JButton("Refresh");
    RefreshButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (dataCenter.DF == 0.0) {
          dataCenter.findDF();
          setDfCenter();
          DfTextField.setText(String.format("%.4f",dataCenter.DF));
        }
        updatePlot();
      }
    });
    OptimiseButton = new JButton("Optimise \u0394f");
    OptimiseButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.findDF();
        setDfCenter();
        DfTextField.setText(String.format("%.4f",dataCenter.DF));
        updatePlot();
      }
    });
    errorCheckBox = new JCheckBox("Show error bars");
    errorCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (dataCenter.DF == 0.0) {
          dataCenter.findDF();
          setDfCenter();
          DfTextField.setText(String.format("%.4f",dataCenter.DF));
        }
        updatePlot();
      }
    });
    buttonBox.add(RefreshButton);
    buttonBox.add(OptimiseButton);
    buttonBox.add(errorCheckBox);
  }

  //////////////////////////////////////////////////////////////////////////////
  // this initialises the different controls on the large frequency separation:
  //////////////////////////////////////////////////////////////////////////////
  private void initDfControls() {
    DfMiniPanel = new JPanel();
    DfMiniPanel.setLayout(new GridLayout(1,3));

    ReduceButton = new JButton("\u25C4");
    ReduceButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.DF/=1.02;
        setDfCenter();
        DfTextField.setText(String.format("%.4f",dataCenter.DF));
        updatePlot();
      }
    });

    DfTextField = new JTextField("0.0000");
    DfTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double DF = Utilities.readDouble(DfTextField.getText(),
                    "value of \u0394f",0.0,1e6,DfTextField);
        if (DF == null) {
          DfTextField.setText(String.format("%.4f",dataCenter.DF));
        } else {
          dataCenter.DF = DF;
          setDfCenter();
          updatePlot();
        }
      }
    });

    IncreaseButton = new JButton("\u25BA");
    IncreaseButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.DF*=1.02;
        setDfCenter();
        DfTextField.setText(String.format("%.4f",dataCenter.DF));
        updatePlot();
      }
    });

    DfMiniPanel.add(ReduceButton);
    DfMiniPanel.add(DfTextField);
    DfMiniPanel.add(IncreaseButton);
    
    // initialise DfCenter and DfSlider:
    DfSlider = new JSlider(JSlider.HORIZONTAL,-100,100,0);
    DfCenter = 0.0;
    DfSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        SliderToDF();
        DfTextField.setText(String.format("%.4f",dataCenter.DF));
        updatePlot();
      }
    });

    // initialise Offset and slider:
    OffsetSlider = new JSlider(JSlider.HORIZONTAL,-100,100,0);
    Offset = 0.0;
    OffsetSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        SliderToOffset();
        updatePlot();
      }
    });


    DfTitles = new JPanel(new GridLayout(3,1,0,10));
    DfTitles.add(new JLabel("\u0394f",SwingConstants.CENTER));
    DfTitles.add(new JLabel("Magnitude:",SwingConstants.CENTER));
    DfTitles.add(new JLabel("Offset:",SwingConstants.CENTER));

    DfPanel = new JPanel(new GridLayout(3,1));
    DfPanel.add(DfMiniPanel);
    DfPanel.add(DfSlider);
    DfPanel.add(OffsetSlider);
  }

  //////////////////////////////////////////////////////////////////////////////
  // this initialises the legend:
  //////////////////////////////////////////////////////////////////////////////
  private void initLegend() {
    LegendPanel = new JPanel(new GridLayout(3,1,0,10));
    LegendPanel.add(new JLabel("<html><u>Legend</u></html>",SwingConstants.CENTER));
    LegendPanel.add(new JLabel("<html>+&nbsp;\u2113=0&nbsp;&nbsp;&nbsp;&nbsp;\u2715&nbsp;\u2113=1&nbsp;&nbsp;&nbsp;&nbsp;<font color='blue'>Observations</font></html>"));
    LegendPanel.add(new JLabel("<html>\u25A2&nbsp;\u2113=2&nbsp;&nbsp;&nbsp;&nbsp;\u25CA&nbsp;\u2113=3&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<font color='red'>Theoretical</font></html>"));
  }

  //////////////////////////////////////////////////////////////////////////////
  // this calculates the large separation from the value on the DfSlider
  //////////////////////////////////////////////////////////////////////////////
  private void SliderToDF() {
    dataCenter.DF = DfCenter*Math.pow(1.3,DfSlider.getValue()/100.0);
  }

  //////////////////////////////////////////////////////////////////////////////
  // this calculates the central value needed on the slider to produce the
  // the current DF, taking into account the slider's position.
  //////////////////////////////////////////////////////////////////////////////
  private void setDfCenter() {
    DfCenter = dataCenter.DF/Math.pow(1.3,DfSlider.getValue()/100.0);
  }

  //////////////////////////////////////////////////////////////////////////////
  // this calculates the Offset from the OffsetSlider.
  //////////////////////////////////////////////////////////////////////////////
  private void SliderToOffset() {
    Offset = OffsetSlider.getValue()/200.0;
  }

  //////////////////////////////////////////////////////////////////////////////
  // this updates the plot with the echelle diagram
  //////////////////////////////////////////////////////////////////////////////
  private void updatePlot() {
    plotFreqObs();
    plotFreqTheo();

    EchelleDiagram.findRange();
    EchelleDiagram.setXmin((Offset-0.02)*dataCenter.DF);
    EchelleDiagram.setXmax((Offset+1.02)*dataCenter.DF);
    EchelleDiagram.repaint();

  }

  //////////////////////////////////////////////////////////////////////////////
  // this plots the observed frequencies
  //////////////////////////////////////////////////////////////////////////////
  private void plotFreqObs() {
    int i, ll;

    // count the number of frequencies according to their l values:
    int[] nl = new int[4];
    for(i=0;i<dataCenter.modes.size();i++) {
      if (dataCenter.modes.get(i).FreqObs == null) continue;
      if (dataCenter.modes.get(i).l == null) continue;
      if (dataCenter.modes.get(i).l.intValue() > 3) continue;
      if (dataCenter.modes.get(i).l.intValue() < 0) continue;
      nl[dataCenter.modes.get(i).l.intValue()]++;
    }

    // test to see whether there is at least one (l,nu) pair
    // with l between 0 and 3 (inclusive)
    boolean cannotPlot = true;
    for(i=0;i<4;i++) {
      if (nl[i]>0) {
        cannotPlot = false;
        break;
      }
    }

    if (cannotPlot) {
      EchelleDiagram.nullify();
      return;
    }

    // create arrays with the appropriate dimensions:
    double[][] x = new double[4][];
    double[][] y = new double[4][];
    double[][] error = new double[4][];
    for(ll=0;ll<4;ll++) {
      x[ll] = new double[nl[ll]];
      y[ll] = new double[nl[ll]];
      error[ll] = new double[nl[ll]];
      nl[ll] = 0;
    }

    // find the different points in the Echelle Diagram:
    for(i=0;i<dataCenter.modes.size();i++) {
      if (dataCenter.modes.get(i).FreqObs == null) continue;
      if (dataCenter.modes.get(i).l == null) continue;
      if (dataCenter.modes.get(i).l.intValue() > 3) continue;
      if (dataCenter.modes.get(i).l.intValue() < 0) continue;
      ll = dataCenter.modes.get(i).l.intValue();
      y[ll][nl[ll]] = dataCenter.modes.get(i).FreqObs.doubleValue();
      if (dataCenter.DF != 0.0) {
        x[ll][nl[ll]] = dataCenter.modes.get(i).FreqObs.doubleValue()%dataCenter.DF;
      } else {
        x[ll][nl[ll]] = dataCenter.modes.get(i).FreqObs.doubleValue();
      }
      if (dataCenter.modes.get(i).ErrorFreq != null) {
        error[ll][nl[ll]] = dataCenter.modes.get(i).ErrorFreq.doubleValue();
      } else {
        error[ll][nl[ll]] = 0.0;
      }
      nl[ll]++;
    }

    EchelleDiagram.replaceDataSet(0,x[0],y[0],Plot.noLine,Plot.plus,ColorObs);
    EchelleDiagram.replaceDataSet(1,x[1],y[1],Plot.noLine,Plot.times,ColorObs);
    EchelleDiagram.replaceDataSet(2,x[2],y[2],Plot.noLine,Plot.square,ColorObs);
    EchelleDiagram.replaceDataSet(3,x[3],y[3],Plot.noLine,Plot.diamond,ColorObs);

    // make multiple copies of the data sets so as to create the illusion
    // of periodicity:
    for(ll=0;ll<4;ll++) {
      for(i=0;i<x[ll].length;i++) x[ll][i]-=dataCenter.DF;
    }

    EchelleDiagram.replaceDataSet(4,x[0],y[0],Plot.noLine,Plot.plus,ColorObs);
    EchelleDiagram.replaceDataSet(5,x[1],y[1],Plot.noLine,Plot.times,ColorObs);
    EchelleDiagram.replaceDataSet(6,x[2],y[2],Plot.noLine,Plot.square,ColorObs);
    EchelleDiagram.replaceDataSet(7,x[3],y[3],Plot.noLine,Plot.diamond,ColorObs);

    for(ll=0;ll<4;ll++) {
      for(i=0;i<x[ll].length;i++) x[ll][i]+=(2*dataCenter.DF);
    }

    EchelleDiagram.replaceDataSet(8,x[0],y[0],Plot.noLine,Plot.plus,ColorObs);
    EchelleDiagram.replaceDataSet(9,x[1],y[1],Plot.noLine,Plot.times,ColorObs);
    EchelleDiagram.replaceDataSet(10,x[2],y[2],Plot.noLine,Plot.square,ColorObs);
    EchelleDiagram.replaceDataSet(11,x[3],y[3],Plot.noLine,Plot.diamond,ColorObs);

    for(i=0; i<12; i++) {
      EchelleDiagram.replaceLeftError(i,error[i%4]);
      EchelleDiagram.replaceRightError(i,error[i%4]);
      EchelleDiagram.setHErrorValue(i,errorCheckBox.isSelected());
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  // this plots the theoretical frequencies
  //////////////////////////////////////////////////////////////////////////////
  private void plotFreqTheo() {
    int i, ll;

    // count the number of frequencies according to their l values:
    int[] nl = new int[4];
    for(i=0;i<dataCenter.modes.size();i++) {
      if (dataCenter.modes.get(i).kernel == null) continue;
      if (dataCenter.modes.get(i).l.intValue() > 3) continue;
      if (dataCenter.modes.get(i).l.intValue() < 0) continue;
      nl[dataCenter.modes.get(i).l.intValue()]++;
    }

    // test to see whether there is at least one (l,nu) pair
    // with l between 0 and 3 (inclusive)
    boolean cannotPlot = true;
    for(i=0;i<4;i++) {
      if (nl[i]>0) {
        cannotPlot = false;
        break;
      }
    }

    if (cannotPlot) {
      for(i=12;i<24;i++) EchelleDiagram.nullify(i);
      return;
    }

    // create arrays with the appropriate dimensions:
    double[][] x = new double[4][];
    double[][] y = new double[4][];
    for(ll=0;ll<4;ll++) {
      x[ll] = new double[nl[ll]];
      y[ll] = new double[nl[ll]];
      nl[ll] = 0;
    }

    // find the different points in the Echelle Diagram:
    for(i=0;i<dataCenter.modes.size();i++) {
      if (dataCenter.modes.get(i).kernel == null) continue;
      if (dataCenter.modes.get(i).l.intValue() > 3) continue;
      if (dataCenter.modes.get(i).l.intValue() < 0) continue;
      ll = dataCenter.modes.get(i).l.intValue();
      y[ll][nl[ll]] = dataCenter.modes.get(i).kernel.FreqTheo;
      if (dataCenter.DF != 0.0) {
        x[ll][nl[ll]] = dataCenter.modes.get(i).kernel.FreqTheo%dataCenter.DF;
      } else {
        x[ll][nl[ll]] = dataCenter.modes.get(i).kernel.FreqTheo;
      }
      nl[ll]++;
    }

    EchelleDiagram.replaceDataSet(12,x[0],y[0],Plot.noLine,Plot.plus,ColorTheo);
    EchelleDiagram.replaceDataSet(13,x[1],y[1],Plot.noLine,Plot.times,ColorTheo);
    EchelleDiagram.replaceDataSet(14,x[2],y[2],Plot.noLine,Plot.square,ColorTheo);
    EchelleDiagram.replaceDataSet(15,x[3],y[3],Plot.noLine,Plot.diamond,ColorTheo);

    // make multiple copies of the data sets so as to create the illusion
    // of periodicity:
    for(ll=0;ll<4;ll++) {
      for(i=0;i<x[ll].length;i++) x[ll][i]-=dataCenter.DF;
    }

    EchelleDiagram.replaceDataSet(16,x[0],y[0],Plot.noLine,Plot.plus,ColorTheo);
    EchelleDiagram.replaceDataSet(17,x[1],y[1],Plot.noLine,Plot.times,ColorTheo);
    EchelleDiagram.replaceDataSet(18,x[2],y[2],Plot.noLine,Plot.square,ColorTheo);
    EchelleDiagram.replaceDataSet(19,x[3],y[3],Plot.noLine,Plot.diamond,ColorTheo);

    for(ll=0;ll<4;ll++) {
      for(i=0;i<x[ll].length;i++) x[ll][i]+=(2*dataCenter.DF);
    }

    EchelleDiagram.replaceDataSet(20,x[0],y[0],Plot.noLine,Plot.plus,ColorTheo);
    EchelleDiagram.replaceDataSet(21,x[1],y[1],Plot.noLine,Plot.times,ColorTheo);
    EchelleDiagram.replaceDataSet(22,x[2],y[2],Plot.noLine,Plot.square,ColorTheo);
    EchelleDiagram.replaceDataSet(23,x[3],y[3],Plot.noLine,Plot.diamond,ColorTheo);

    EchelleDiagram.findRange();
    EchelleDiagram.setXmin((Offset-0.02)*dataCenter.DF);
    EchelleDiagram.setXmax((Offset+1.02)*dataCenter.DF);
    EchelleDiagram.repaint();

  }
}
