/**********************************************************************
* Some revisions were made to this file, following modifications
* to NativeBinaryInputStream.java.  To get an original version, see:
*   http://docjar.com/docs/api/org/fudaa/dodico/fortran/FortranBinaryInputStream.html
*
* D. Reese, 24/9/2009
**********************************************************************/

/*
 * @file         FortranBinaryInputStream.java
 * @creation     1998-08-19
 * @modification $Date: 2002/11/25 09:59:25 $
 * @license      GNU General Public License 2
 * @copyright    (c)1998-2001 CETMEF 2 bd Gambetta F-60231 Compiegne
 * @mail         devel@fudaa.org
 */

package org.fudaa.dodico.fortran;

import org.fudaa.dodico.fortran.NativeBinaryInputStream;
import java.io.*;
import java.util.*;
import java.nio.ByteOrder;

/**
 * Une classe facilitant la lecture de fichiers binaires ecrits par Fortran
 * L'equivalence d'intructions entre Java et Fortran se fera de la maniere
 * suivante :<BR><PRE>
 * (en considerant i=integer/int, f=real/float, d=double precision/double et
 * s=character*()/String)
 * 1) Pour un fichier a acces sequentiel :
 * Fortran
 * open  (unit=10,file='fichier.bin',access='sequentiel',form='unformatted')
 * read  (unit=10) {i},{f},{d},{s}
 * ...
 * close (unit=10)
 * Java
 * FortranBinaryInputStream in=
 * new FortranBinaryInputStream(new FileInputStream("fichier.bin"),true);
 * in.readRecord();
 * i=in.readInteger();
 * f=in.readReal();
 * d=in.readDoublePrecision();
 * s=in.readCharacter(s.length());
 * ...
 * in.close();
 * 2) Pour un fichier a acces direct :
 * Fortran
 * open  (unit=10,file='fichier.bin',access='direct',recl=30,form='unformatted')
 * read  (unit=10,rec=1) {i},{f},{d},{s}
 * ...
 * close (unit=10)
 * Java
 * FortranBinaryInputStream in=
 * new FortranBinaryInputStream(new FileInputStream("fichier.bin"),false);
 * in.setRecordLength(30);
 * in.readRecord();
 * i=in.readInteger();
 * f=in.readReal();
 * d=in.readDoublePrecision();
 * s=in.readCharacter(s.length());
 * ...
 * in.close();
 * </PRE>
 *
 * @version      $Revision: 1.9 $ $Date: 2002/11/25 09:59:25 $ by $Author: deniger $
 * @author       Bertrand Marchand 
 */
public class FortranBinaryInputStream extends NativeBinaryInputStream {

  private boolean sequential_;
  private int     recordLength_;
  private long    currentPos_;
  private long    nextPos_;

  /**
   * Creation en precisant si le fichier binaire est a access sequentiel ou non.
   * Si sequentiel, alors certaines donnees ( generees par la meme instruction write 
   * de fortran) sont entourees par 4 octets qui correspondent a la longueur
   * de l'enregistrement. La variable systeme ("os.arch") permet de determiner
   * l'architecture a utiliser.
   * @param _in InputStream
   * @param _sequential <B>true</B> le fichier est binaire a acces
   * <I>Sequential</I>. <B>false</B> le fichier est binaire a acces <I>Direct</I>
   */
  public FortranBinaryInputStream(InputStream _in, boolean _sequential)
   throws IOException{
    super(_in);
    sequential_  =_sequential;
    recordLength_=0;
    nextPos_     =0;
    currentPos_  =nextPos_;
  }

  /**
   * Autre constructeur dans lequel on precise en plus l'endianness
   * que l'on veut.
   */
  public FortranBinaryInputStream(InputStream _in, boolean _sequential, ByteOrder byteOrder)
   throws IOException{
    super(_in, byteOrder);
    sequential_  =_sequential;
    recordLength_=0;
    nextPos_     =0;
    currentPos_  =nextPos_;
  }

  /**
   * Autre constructeur dans lequel on precise si on veut l'endianness 
   * natif de la machine ou son oppose.
   */
  public FortranBinaryInputStream(InputStream _in, boolean _sequential, boolean isNativeEndian)
   throws IOException{
    super(_in, isNativeEndian);
    sequential_  =_sequential;
    recordLength_=0;
    nextPos_     =0;
    currentPos_  =nextPos_;
  }
  
  /**
   * ATTENTION Different du cas acces direct. 
   * Renvoie la taille de l'enregistrement lu par le flux. 
   */
  public int getSequentialRecordLength()
  {
    return recordLength_;
  }
  
  /**
   * Affectation de la longueur des enregistrements (pour les fichiers a acces
   * <I>Direct</I>
   * @param _length Longueur d'enregistrement en longworld
   *                (1 longworld=4 octets)
   */
  public void setRecordLength(int _length) { recordLength_=_length*4; }

  /**
   * Retourne la longueur des enregistrements (pour les fichiers a acces
   * <I>Direct</I>
   * @return Longueur d'enregistrement en longworld (1 longworld=4 octets)
   */
  public int getRecordLength() { return recordLength_/4; }

  /**
   * Lecture d'un champ chaine de caracteres "<I>character</I>" Fortran
   * @param _lgString Longueur de la chaine a lire
   * @return String correspondant a la chaine de caracteres
   */
  public String readCharacter(int _lgString) throws IOException {
    byte[] buf=new byte[_lgString];
    currentPos_+=_lgString;
    read(buf);
    return new String(buf);
  }

  /**
   * Lecture d'un champ entier "<I>integer</I>" Fortran
   * @return int correspondant au integer
   */
  public int readInteger() throws IOException {
    currentPos_+=4;
    return super.readInt_32();
  }

  /**
   * Lecture d'un champ reel "<I>real</I>" Fortran
   * @return float correspondant au real
   */
  public float readReal() throws IOException {
    currentPos_+=4;
    return super.readFloat_32();
  }

  /**
   * Lecture d'un champ reel en double precision "<I>double precision</I>"
   * Fortran
   * @return double correspondant au double precision
   */
  public double readDoublePrecision() throws IOException {
    currentPos_+=8;
    return super.readFloat_64();
  }
  
  /**
   * Renvoie en nb d'octets la position courante. 
   */
  public long positionCourante()
  {
    return currentPos_;
  }

  /**
   * Renvoie en nb d'octets de la position suivante. 
   */
  public long positionSuivante()
  {
    return nextPos_;
  }

  /**
   * Lecture des champs de l'enregistrement. L'enregistrement doit etre lu
   * avant la lecture des champs.
   * Un enregistrement correspond a une instruction READ du Fortran
   */
  public int readRecord() throws IOException {
    skip(nextPos_-currentPos_);
    currentPos_=nextPos_;

    if (sequential_) {
      recordLength_=readInteger();
      nextPos_+=(long)recordLength_+8;
    }
    else {
      nextPos_+=(long)recordLength_;
    }
    return recordLength_;
  }
}
