/**********************************************************************
* Some revisions were made to this file, following modifications
* to NativeBinaryOutputStream.java.  To get an original version, see:
*   http://docjar.com/docs/api/org/fudaa/dodico/fortran/FortranBinaryOutputStream.html
*
* D. Reese, 24/9/2009
**********************************************************************/

/*
 * @file         FortranBinaryOutputStream.java
 * @creation     1998-08-19
 * @modification $Date: 2002/11/25 09:59:25 $
 * @license      GNU General Public License 2
 * @copyright    (c)1998-2001 CETMEF 2 bd Gambetta F-60231 Compiegne
 * @mail         devel@fudaa.org
 */

package org.fudaa.dodico.fortran;

import org.fudaa.dodico.fortran.NativeBinaryOutputStream;
import java.io.*;
import java.util.*;
import java.nio.ByteOrder;

/**
 * Une classe facilitant l'ecriture de fichiers binaires lus par Fortran.
 * ATTENTION les methodes heritees et non redefinies dans cette 
 * donneront des resultats faux.
 * L'equivalence d'intructions entre Java et Fortran se fera de la maniere
 * suivante :<BR><PRE>
 * (en considerant i=integer/int, f=real/float, d=double precision/double et
 * s=character*()/String)
 * 1) Pour un fichier a acces sequentiel :
 *   Fortran
 *     open  (unit=10,file='fichier.bin',access='sequentiel',form='unformatted')
 *     write (unit=10) {i},{f},{d},{s}
 *     ...
 *     close (unit=10)
 *   Java
 *     FortranBinaryOutputStream out=
 *      new FortranBinaryOutputStream(new FileOutputStream("fichier.bin"),true);
 *     out.writeInteger(i);
 *     out.writeReal(f);
 *     out.writeDoublePrecision(d);
 *     out.writeCharacter(s);
 *     out.writeRecord();
 *     ...
 *     out.close();
 * 2) Pour un fichier a acces direct :
 *   Fortran
 *     open(unit=10,file='fichier.bin',access='direct',recl=30,form='unformatted')
 *     write (unit=10,rec=1) {i},{f},{d},{s}
 *     ...
 *     close (unit=10)
 *   Java
 *     FortranBinaryOutputStream out=
 *      new FortranBinaryOutputStream(new FileOutputStream("fichier.bin"),false);
 *     out.setRecordLength(30);
 *     out.writeInteger(i);
 *     out.writeReal(f);
 *     out.writeDoublePrecision(d);
 *     out.writeCharacter(s);
 *     out.writeRecord();
 *     ...
 *     out.close();
 * </PRE>
 *
 * @version      $Id: FortranBinaryOutputStream.java,v 1.8 2002/11/25 09:59:25 deniger Exp $
 * @author       Bertrand Marchand 
 */
public class FortranBinaryOutputStream extends NativeBinaryOutputStream {
  private NativeBinaryOutputStream bufStream_;
  private ByteArrayOutputStream    arrayStream_;
  private boolean                  sequential_;
  private int                      recordLength_;

  
  /**
   * Creation en precisant si le fichier binaire est a access sequentiel ou non
   * @param _out OutputStream
   * @param _sequential <B>true</B> le fichier est binaire a acces
   * <I>Sequential</I>. <B>false</B> le fichier est binaire a acces <I>Direct</I>
   */
  public FortranBinaryOutputStream(OutputStream _out, boolean _sequential)
   throws IOException{
    super(_out);
    arrayStream_ =new ByteArrayOutputStream();
    bufStream_   =new NativeBinaryOutputStream(arrayStream_);
    sequential_  =_sequential;
    recordLength_=0;
  }

  /**
   * Autre constructeur dans lequel on precise en plus l'endianness
   * que l'on veut.
   */
  public FortranBinaryOutputStream(OutputStream _out, boolean _sequential, ByteOrder byteOrder)
   throws IOException{
    super(_out, byteOrder);
    arrayStream_ =new ByteArrayOutputStream();
    bufStream_   =new NativeBinaryOutputStream(arrayStream_,byteOrder);
    sequential_  =_sequential;
    recordLength_=0;
  }

  /**
   * Autre constructeur dans lequel on precise si on veut l'endianness 
   * natif de la machine ou son oppose.
   */
  public FortranBinaryOutputStream(OutputStream _out, boolean _sequential, boolean isNativeEndian)
   throws IOException{
    super(_out, isNativeEndian);
    arrayStream_ =new ByteArrayOutputStream();
    bufStream_   =new NativeBinaryOutputStream(arrayStream_,isNativeEndian);
    sequential_  =_sequential;
    recordLength_=0;
  }

  /**
   * Affectation de la longueur des enregistrements (pour les fichiers a acces
   * <I>Direct</I>)
   * @param _length Longueur d'enregistrement en longworld
   *                (1 longworld=4 octets)
   */
  public void setRecordLength(int _length) { recordLength_=_length; }

  /**
   * Retourne la longueur des enregistrements (pour les fichiers a acces
   * <I>Direct</I>
   * @return Longueur d'enregistrement en longworld (1 longworld=4 octets)
   */
  public int getRecordLength() { return recordLength_; }

  /**
   * Ecriture d'un champ chaine de caracteres "<I>character</I>" Fortran
   * @param _s string correspondant a la chaine de caracteres
   */
  public void writeCharacter(String _s) throws IOException {
    bufStream_.writeBytes(_s);
  }

  /**
   * Ecriture d'un champ entier "<I>integer</I>" Fortran
   * @param _i int correspondant au integer
   */
  public void writeInteger(int _i) throws IOException {
    bufStream_.writeInt_32(_i);
  }

  /**
   * Ecriture d'un champ reel "<I>real</I>" Fortran
   * @param _f float correspondant au real
   */
  public void writeReal(float _f) throws IOException {
    bufStream_.writeFloat_32(_f);
  }

  /**
   * Ecriture d'un champ reel en double precision "<I>double precision</I>"
   * Fortran
   * @param _d double correspondant au double precision
   */
  public void writeDoublePrecision(double _d) throws IOException {
    bufStream_.writeFloat_64(_d);
  }

  /**
   * Ecriture des champs de l'enregistrement. L'enregistrement doit etre ecrit
   * pour que les champs soient egalement ecrits sur le fichiers.
   * Un enregistrement correspond a une instruction WRITE du Fortran
   */
  public void writeRecord() throws IOException {
    if (sequential_) {
      int t=arrayStream_.size();
      writeInt_32(t);
      arrayStream_.writeTo(this);
      writeInt_32(t);
    }
    else {
      byte[] buf=arrayStream_.toByteArray();
      write(buf,0,Math.min(buf.length,4*recordLength_));
      for (int i=buf.length; i<4*recordLength_; i++) write(0);
    }

    arrayStream_.reset();
  }

  /**
   * Fermeture du fichier
   */
  public void close() throws IOException {
    super       .close();
    arrayStream_.close();
    bufStream_  .close();
  }
}
