/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import inversion_applet.Utilities;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JTable;
import javax.swing.JLabel;
import javax.swing.table.AbstractTableModel;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.File;
import java.io.BufferedReader;
import java.util.StringTokenizer;

public class FrequenciesDataModel extends AbstractTableModel {

  protected JLabel numberFreq = new JLabel();
  protected JLabel numberRS = new JLabel();

  // the data with the frequency shifts is in dataCenter;
  private DataCenter dataCenter;
 
  // These contain the parent components, which intervenes
  // when using a JPaneOption object.
  private JComponent parentComponent;

  
  // the names of the columns:
  private String[] columns = {"\u2113","n",
                              "<html>\u03BD<sub>obs</sub> (\u03BCHz)</html>",
                              "<html>\u03BD<sub>theo</sub> (\u03BCHz)</html>",
                              "<html>\u03C3\u03BD<sub>obs</sub> (\u03BCHz)</html>",
                              "Rot. Split.","\u03C3(R.S.)","Dupl."};

  // the classes of the columns:
  private Class[] columnClasses = new Class[] 
    {Integer.class, Integer.class, Double.class, Double.class,
     Double.class, Double.class, Double.class, Boolean.class};
 
  /////////////////////////////////////////////////////////
  // a constructor:
  /////////////////////////////////////////////////////////
  protected FrequenciesDataModel(DataCenter dataCenter,
                               JComponent parentComponent) {
    this.dataCenter = dataCenter;
    this.parentComponent = parentComponent;
    this.updateNumbers();
  }
  
  // this updates the number of DF data in numberDF:
  private void updateNumbers() {
    dataCenter.nfreq = dataCenter.nRS = 0;
    for(int i=0; i<dataCenter.modes.size();i++) {
      if ((dataCenter.modes.get(i).l==null)||(dataCenter.modes.get(i).FreqObs==null)) continue;
      if (dataCenter.modes.get(i).l < 0) continue;
      if (dataCenter.modes.get(i).ErrorFreq!=null) dataCenter.nfreq++;
      if ((dataCenter.modes.get(i).RS!=null)&&(dataCenter.modes.get(i).ErrorRS!=null)) dataCenter.nRS++;
    }
    numberFreq.setText(String.format(" Number of frequencies: %d",
                     dataCenter.nfreq));
    numberRS.setText(String.format(" Number of R.S.: %d",
                     dataCenter.nRS));
  }

  /////////////////////////////////////////////////////////
  // load a file:
  /////////////////////////////////////////////////////////
  protected void loadFile(String filename){
    BufferedReader f;
    String line;
    Boolean skipLine = false;
    Integer l;
    Double FreqObs, ErrorFreq, RS, ErrorRS;
    try {
      f = Utilities.OpenFile(filename);
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    parentComponent,  // the parent component
                    "Unable to open file \""+filename+"\"",
                    "File not found",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    try {
      while (true) {
        line = f.readLine();     // read a line of text
        if (line == null) break; // test for the end of the buffer
        line = Utilities.trim(line);       // remove comments
        StringTokenizer t = new StringTokenizer(line," \t,");
        if ((t.countTokens() >= 3) && (t.countTokens() < 5)) {
          l  = new Integer(t.nextToken());
          FreqObs = new Double(Utilities.parseDouble(t.nextToken()));
          ErrorFreq = new Double(Utilities.parseDouble(t.nextToken()));
          if ((l.intValue() >= 0)&&(!FreqObs.isNaN())) {
            if (ErrorFreq.isNaN()) ErrorFreq = null;
            dataCenter.modes.add(new Mode(l,FreqObs,ErrorFreq));
          } else {
            skipLine = true;
          }
        } else if(t.countTokens() >= 5) {
          l  = new Integer(t.nextToken());
          FreqObs = new Double(Utilities.parseDouble(t.nextToken()));
          ErrorFreq = new Double(Utilities.parseDouble(t.nextToken()));
          RS = new Double(Utilities.parseDouble(t.nextToken()));
          ErrorRS = new Double(Utilities.parseDouble(t.nextToken()));
          if ((l.intValue() >= 0)&&(!FreqObs.isNaN())) {
            if (ErrorFreq.isNaN()) ErrorFreq = null;
            if (RS.isNaN()) RS = null;
            if (ErrorRS.isNaN()) ErrorRS = null;
            dataCenter.modes.add(new Mode(l,FreqObs,ErrorFreq,RS,ErrorRS));
          } else {
            skipLine = true;
          }
        } else {
          if (t.countTokens() != 0) skipLine = true;
        }
      }
      f.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    parentComponent,  // the parent component
                    "Format not recognised",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    if (skipLine) {
      JOptionPane.showMessageDialog(
                    parentComponent,  // the parent component
                    "Some lines with either the wrong\n"+
                    "number of entries or with \u2113 < 0\n"+
                    "were skipped.",
                    "Warning",
                    JOptionPane.WARNING_MESSAGE
                  );
    }
    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    updateNumbers();
  }

  /////////////////////////////////////////////////////////
  // write a file:
  /////////////////////////////////////////////////////////
  protected void writeFile(String filename) {
    File f = new File(filename);
    PrintWriter out;
    try {
      out = new PrintWriter(new FileWriter(f));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        parentComponent,
        "Unable to write to file \""+filename+"\"",
        "Failure",
        JOptionPane.WARNING_MESSAGE
      );
      return;
    }
    out.println("# l    nu (in microHz)    sigma_nu (in microHz)    Rotational Splittings    sigma_{R.S.}");
    for(int i = 0; i < dataCenter.modes.size(); i++) {
      out.printf("%3d ",dataCenter.modes.get(i).l);
      if (dataCenter.modes.get(i).FreqObs==null) {
        out.printf("null                ");
      } else {
        out.printf("%19.12e ",dataCenter.modes.get(i).FreqObs);
      }
      if (dataCenter.modes.get(i).ErrorFreq==null) {
        out.printf("null                ");
      } else {
        out.printf("%19.12e ",dataCenter.modes.get(i).ErrorFreq);
      }
      if (dataCenter.modes.get(i).RS==null) {
        out.printf("null                ");
      } else {
        out.printf("%19.12e ",dataCenter.modes.get(i).RS);
      }
      if (dataCenter.modes.get(i).ErrorRS==null) {
        out.printf("null%n");
      } else {
        out.printf("%19.12e%n",dataCenter.modes.get(i).ErrorRS);
      }
    }
    out.close();
  }


  /////////////////////////////////////////////////////////
  // the number of columns:
  /////////////////////////////////////////////////////////
  public int getColumnCount() {return 8;}
  
  /////////////////////////////////////////////////////////
  // the number of rows:
  /////////////////////////////////////////////////////////
  public int getRowCount() {return dataCenter.modes.size();}
 
  /////////////////////////////////////////////////////////
  // return the value of a cell:
  /////////////////////////////////////////////////////////
  public Object getValueAt(int row, int col) {
    if (col == 0) return dataCenter.modes.get(row).l;
    if (col == 1) {
      if (dataCenter.modes.get(row).kernel == null) {
        return null;
      } else {
        return dataCenter.modes.get(row).kernel.n;
      }
    }
    if (col == 2) return dataCenter.modes.get(row).FreqObs;
    if (col == 3) {
      if (dataCenter.modes.get(row).kernel == null) {
        return null;
      } else {
        return dataCenter.modes.get(row).kernel.FreqTheo;
      }
    }
    if (col == 4) return dataCenter.modes.get(row).ErrorFreq;
    if (col == 5) return dataCenter.modes.get(row).RS;
    if (col == 6) return dataCenter.modes.get(row).ErrorRS;
    return dataCenter.modes.get(row).isDuplicate;
  }

  /////////////////////////////////////////////////////////
  // return the name of a column:
  /////////////////////////////////////////////////////////
  public String getColumnName(int column) {return columns[column];}
  
  /////////////////////////////////////////////////////////
  // return the class of a column:
  /////////////////////////////////////////////////////////
  public Class getColumnClass(int c) {return columnClasses[c];}
  
  /////////////////////////////////////////////////////////
  // this makes some of the cells editable:
  /////////////////////////////////////////////////////////
  public boolean isCellEditable(int row, int col) {
    if ((col == 1)||(col == 3)||(col == 7)) return false;
    return true;
  }

  /////////////////////////////////////////////////////////
  /** set the value of a particular cell: */
  /////////////////////////////////////////////////////////
  public void setValueAt(Object obj, int row, int col) {
    boolean changeKernel = false;
    if (col == 0) {
      if (dataCenter.modes.get(row).l == null) {
        changeKernel = !(obj==null);
      } else {
        changeKernel = !dataCenter.modes.get(row).l.equals(obj);
      }
      dataCenter.modes.get(row).l=(Integer)obj;
    }
    if (col == 2) {
      if (dataCenter.modes.get(row).FreqObs == null) {
        changeKernel = !(obj==null);
      } else {
        changeKernel = !dataCenter.modes.get(row).FreqObs.equals(obj);
      }
      dataCenter.modes.get(row).FreqObs=(Double)obj;
    }
    if (col == 4) dataCenter.modes.get(row).ErrorFreq=(Double)obj;
    if (col == 5) dataCenter.modes.get(row).RS=(Double)obj;
    if (col == 6) dataCenter.modes.get(row).ErrorRS=(Double)obj;

    // show that the current kernel is not up-to-date
    if (changeKernel) {
      if (dataCenter.modes.get(row).kernel != null) {
        dataCenter.modes.get(row).kernel.mode = null;
        dataCenter.modes.get(row).kernel = null;
      }
    }

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.kernelTable.updateUI();
    //dataCenter.frequenciesTable.updateUI(); // this causes a null pointer exception
    updateNumbers();
  }
  
  /////////////////////////////////////////////////////////
  // add a blank row:
  /////////////////////////////////////////////////////////
  public void addBlankRow() {
    dataCenter.modes.add(new Mode());
  }

  /////////////////////////////////////////////////////////
  /** remove a row: */
  /////////////////////////////////////////////////////////
  public void removeRow(int i) {
    if (i >= dataCenter.modes.size()) return;
    if (dataCenter.modes.get(i).kernel != null) {
      dataCenter.modes.get(i).kernel.mode = null;
    }
    dataCenter.modes.remove(i);
    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.kernelTable.updateUI();
    updateNumbers();
  }

  /////////////////////////////////////////////////////////
  /** This sorts the frequency shifts according to
   *  l and freq. */
  /////////////////////////////////////////////////////////
  public void sort() { dataCenter.sortObs(); }

  /////////////////////////////////////////////////////////
  // Calculate theoretical frequency shifts
  /////////////////////////////////////////////////////////
  public void calculateFreq(JComponent c){
    if (!dataCenter.targetStructLoaded) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "No structural target functions loaded.  Unable to proceed.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    boolean noKernel = true;
    for(int i=0; i<dataCenter.modes.size(); i++) {
      if (dataCenter.modes.get(i).kernel == null) continue;
      noKernel = false;
      break;
    }

    if (noKernel) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "No kernels loaded.  Unable to proceed.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    // check to see if the span covered by the target grid is
    // large enough:
    if ((dataCenter.gridTargetStruct[0]>dataCenter.grid[0])||
        (dataCenter.gridTargetStruct[dataCenter.gridTargetStruct.length-1]<
         dataCenter.grid[dataCenter.grid.length-1])) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "Warning: the target functions will be extrapolated to\n"+
                    "match the grid on which the kernels are defined.",
                    "Grid warning",
                    JOptionPane.WARNING_MESSAGE
                  );
    }

    // interpolate structural target functions onto the same grid as the kernels
    double[] targetStruct1 = Numerics.interpolate1D_r2(
                             dataCenter.gridTargetStruct,
                             dataCenter.grid,
                             dataCenter.targetStruct1,false);
    double[] targetStruct2 = Numerics.interpolate1D_r2(
                             dataCenter.gridTargetStruct,
                             dataCenter.grid,
                             dataCenter.targetStruct2,false);
    double[] rho = dataCenter.rho;
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);

    if (dataCenter.constantMass!=0) {
      JOptionPane.showMessageDialog(
        c,
        "Target function(s) may have to be adjusted to\n"+
        "ensure that the stellar mass remains constant",
        "Notification",
        JOptionPane.INFORMATION_MESSAGE
      );
    }
        
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==3)) {
      double num = 0, den = 0;
      for(int i=0; i<weights.length; i++) {
        num+=weights[i]*dataCenter.grid[i]*dataCenter.grid[i]*rho[i]*targetStruct1[i];
        den+=weights[i]*dataCenter.grid[i]*dataCenter.grid[i]*rho[i];
      }
      double A = num/den;
      for(int i=0; i<weights.length; i++) targetStruct1[i]-=A;
      for(int i=0; i<dataCenter.targetStruct1.length; i++)
        dataCenter.targetStruct1[i]-=A;
    }
            
    if ((dataCenter.constantMass==2)||(dataCenter.constantMass==3)) {
      double num = 0, den = 0;
      for(int i=0; i<weights.length; i++) {
        num+=weights[i]*dataCenter.grid[i]*dataCenter.grid[i]*rho[i]*targetStruct2[i];
        den+=weights[i]*dataCenter.grid[i]*dataCenter.grid[i]*rho[i];
      }
      double A = num/den;
      for(int i=0; i<weights.length; i++) targetStruct2[i]-=A;
      for(int i=0; i<dataCenter.targetStruct2.length; i++)
        dataCenter.targetStruct2[i]-=A;
    }

    double sum;
    for(int i=0; i<dataCenter.modes.size(); i++) {
      if (dataCenter.modes.get(i).kernel == null) continue;
      sum = 0.0;
      if ((dataCenter.kernelType==0)||(dataCenter.kernelType==2)) {
        for(int j=0; j<targetStruct1.length; j++) {
          sum+=weights[j]*targetStruct1[j]*dataCenter.modes.get(i).kernel.rhoc2Kernel[j];
          sum+=weights[j]*targetStruct2[j]*dataCenter.modes.get(i).kernel.c2rhoKernel[j];
        }
      } else {
        for(int j=0; j<targetStruct1.length; j++) {
          sum+=weights[j]*targetStruct1[j]*dataCenter.modes.get(i).kernel.rhoGamma1Kernel[j];
          sum+=weights[j]*targetStruct2[j]*dataCenter.modes.get(i).kernel.c2rhoKernel[j];
        }
      }
      if (!Double.isNaN(sum)) {
        // Only change observed frequencies, leave observational errors unchanged.
        dataCenter.modes.get(i).FreqObs = dataCenter.modes.get(i).kernel.FreqTheo*(1.0+sum);
      }
    }
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.kernelTable.updateUI();
    updateNumbers();
  }

  /////////////////////////////////////////////////////////
  // Calculate theoretical rotational splittings
  /////////////////////////////////////////////////////////
  public void calculateRS(JComponent c){
    // I'll assume that if targetRota != null, than gridTargetRota != null
    if (!dataCenter.targetRotaLoaded) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "No target \u03A9 function loaded.  Unable to proceed.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    boolean noKernel = true;
    for(int i=0; i<dataCenter.modes.size(); i++) {
      if (dataCenter.modes.get(i).kernel == null) continue;
      if (dataCenter.modes.get(i).l.intValue() < 1) continue;
      noKernel = false;
      break;
    }

    if (noKernel) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "No kernels loaded.  Unable to proceed.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    // check to see if the span covered by the target grid is
    // large enough:
    if ((dataCenter.gridTargetRota[0]>dataCenter.grid[0])||
        (dataCenter.gridTargetRota[dataCenter.gridTargetRota.length-1]<
         dataCenter.grid[dataCenter.grid.length-1])) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "Warning: the target \u03A9 function will be extrapolated\n"+
                    "so as to match the domain over which the rotational\n"+
                    "kernels are defined.",
                    "Grid warning",
                    JOptionPane.WARNING_MESSAGE
                  );
    }
         
    // interpolate rotaTarget onto the same grid as the kernels
    double[] rotaTarget = Numerics.interpolate1D_r2(dataCenter.gridTargetRota,
                          dataCenter.grid, dataCenter.targetRota,false);
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);
    for (int j=0; j<rotaTarget.length; j++) weights[j]*=rotaTarget[j];

    double sum;
    for(int i=0; i<dataCenter.modes.size(); i++) {
      if (dataCenter.modes.get(i).kernel == null) continue;
      if (dataCenter.modes.get(i).l.intValue() < 1) continue;

      // calculate rotational splitting for m = 1 (prograde convention):
      sum = 0.0;
      for(int j=0; j<rotaTarget.length; j++)
        sum+=weights[j]*dataCenter.modes.get(i).kernel.rotaKernel[j];
      if (!Double.isNaN(sum)) {
        dataCenter.modes.get(i).RS = sum;
        dataCenter.modes.get(i).ErrorRS = sum*0.01;
      }
    }
    dataCenter.rotationInversionUpToDate = false;
    updateNumbers();
  }

  /////////////////////////////////////////////////////////
  // This replaces all of the error bars on the frequencies
  // by ErrorFreq.
  /////////////////////////////////////////////////////////
  public void replaceErrorFreq(Double ErrorFreq) {
    for(int i=0; i< dataCenter.modes.size(); i++) {
      dataCenter.modes.get(i).ErrorFreq = null;
      if (dataCenter.modes.get(i).FreqObs == null) continue;
      dataCenter.modes.get(i).ErrorFreq = ErrorFreq;
    }
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
  }

  /////////////////////////////////////////////////////////
  // This replaces all of the error bars on the rotational
  // splittings by ErrorRS.
  /////////////////////////////////////////////////////////
  public void replaceErrorRS(Double ErrorRS) {
    for(int i=0; i< dataCenter.modes.size(); i++) {
      dataCenter.modes.get(i).ErrorRS = null;
      if (dataCenter.modes.get(i).l == null) continue;
      if (dataCenter.modes.get(i).l < 1) continue;
      if (dataCenter.modes.get(i).RS == null) continue;
      dataCenter.modes.get(i).ErrorRS = ErrorRS;
    }
    dataCenter.rotationInversionUpToDate = false;
    updateNumbers();
  }

  /////////////////////////////////////////////////////////
  // This generates a grid of frequencies from initial, final
  // and increment values for both the frequencies and l values.
  // All of the error bars on the frequencies are set to ErrorFreq.
  /////////////////////////////////////////////////////////
  public void generateFrequencies(int lStart, int lStop, int lIncr,
                double freqStart, double freqStop, double freqIncr,
                Double ErrorFreq) {
    double freq;
    if (lIncr > 0) {
      if (freqIncr > 0) {
        for(int l=lStart; l<=lStop; l+=lIncr) {
          for(freq=freqStart; freq<=freqStop; freq+=freqIncr) 
            dataCenter.modes.add(new Mode(new Integer(l), new Double(freq), ErrorFreq));
        }
      } else {
        for(int l=lStart; l<=lStop; l+=lIncr) {
          for(freq=freqStart; freq>=freqStop; freq+=freqIncr) 
            dataCenter.modes.add(new Mode(new Integer(l), new Double(freq), ErrorFreq));
        }
      }
    } else {
      if (freqIncr > 0) {
        for(int l=lStart; l>=lStop; l+=lIncr) {
          for(freq=freqStart; freq<=freqStop; freq+=freqIncr) 
            dataCenter.modes.add(new Mode(new Integer(l), new Double(freq), ErrorFreq));
        }
      } else {
        for(int l=lStart; l>=lStop; l+=lIncr) {
          for(freq=freqStart; freq>=freqStop; freq+=freqIncr) 
            dataCenter.modes.add(new Mode(new Integer(l), new Double(freq), ErrorFreq));
        }
      }
    }
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    updateNumbers();
  }
}
