/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
/////////////////////////////////////////////////////////////////////////////
/*
 * @(#)OptionPaneDemo.java      1.9 04/07/26
 *
 * Copyright (c) 2004 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * -Redistribution of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 * -Redistribution in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MIDROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility.
 */

/*
 * @(#)OptionPaneDemo.java      1.9 04/07/26
 */

/////////////////////////////////////////////////////////////////////////////
package graph;

import graph.Plot;
import java.io.File;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.BoxLayout;
import javax.swing.WindowConstants;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.lang.Cloneable;

/**********************************************************************
* The purpose of this class is to create interactive plots in Java.
* It uses the class graph.Plot to make the actual plot, but also adds
* various buttons on the side to perform different jobs and allows the
* the user to zoom in on different regions by using the mouse.
**********************************************************************/

public class IPlot extends JComponent implements Cloneable
{
  // This panel contains a number buttons to be used with the plot:
  private JPanel buttonBox = new JPanel();
  
  // Buttons on the side of the plot:
  private JButton IDLButton;
  private JButton TXTButton;
  private JButton DuplicateButton;
  private JButton AutoScaleButton;

  // Check boxes on the side of the plot:
  private JCheckBox XlogCheckBox;
  private JCheckBox YlogCheckBox;

  // This keeps track of whether input/output buttons are enabled/disabled
  private boolean IOenabled = true;
  
  // This contains the plot itself:
  private Plot graphicsArea;

  /////////////////////////////////////////////////////////////////////////////
  // This sets up the interactive plot.  I'm assuming the actual plot has
  // already been initialised.
  /////////////////////////////////////////////////////////////////////////////
  private void initIPlot() {
    initOutputButtons();
    initAutoScaleButton();
    initGraphicsArea();
    initDuplicateButton();
    initCheckBoxes();

    buttonBox.setLayout(new BoxLayout(buttonBox, BoxLayout.Y_AXIS));
    buttonBox.add(IDLButton);
    buttonBox.add(TXTButton);
    buttonBox.add(AutoScaleButton);
    buttonBox.add(DuplicateButton);
    buttonBox.add(XlogCheckBox);
    buttonBox.add(YlogCheckBox);
    setLayout(new BorderLayout());
    add("Center", graphicsArea);
    add("West", buttonBox);
  }

  /**************************************************************************
  * This enables or disables all input/output related buttons, depending on
  * the value of flag.
  *
  * @param flag this decides whether input/output buttons should be enabled
  *             or disabled:
  *             <ul>
  *               <li> true: enable input/output buttons </li>
  *               <li> false: disable input/output buttons </li>
  *             </ul>
  **************************************************************************/
  public void setIOenabled(boolean flag) {
    IDLButton.setEnabled(flag);
    TXTButton.setEnabled(flag);
    IOenabled = flag;
  }

  /////////////////////////////////////////////////////////////////////////////
  // This enables the graphics area to respond to mouse events by adding a
  // mouse listener and a mouse motion listener.
  /////////////////////////////////////////////////////////////////////////////
  private void initGraphicsArea() {
    MyMouseListener l = new MyMouseListener();
    graphicsArea.addMouseListener(l);
    graphicsArea.addMouseMotionListener(l);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This retrieves the graphicsArea (so it can remain private).
  /////////////////////////////////////////////////////////////////////////////
  public Plot getGraphicsArea() { return graphicsArea; }

  /////////////////////////////////////////////////////////////////////////////
  // Private class which describes the mouse listener/mouse motion listener
  // that goes with graphicsArea.
  /////////////////////////////////////////////////////////////////////////////
  private class MyMouseListener implements MouseListener, MouseMotionListener {

    // This keeps track of whether the mouse is
    // being dragged (and has stayed within
    // graphicsArea):
    private boolean drag = false;
    
    // This keeps track of whether or not there
    // is a previous rectangle that needs to be
    // erased.
    private boolean prevRect = false;

    // This keeps track of whether the plot
    // is being shifted (and has stayed within
    // graphicsArea):
    private boolean shift = false;
    
    // (x0,y0) = start of drag (in window coordinates)
    // (x1,y1) = end of drag (in window coordinates)
    private int x0, y0, x1, y1;

    // (xx0,yy0) = initial position when shifting (in real coordinates)
    // (xl0,yl0) = log of the above coordinates
    private double xx0, yy0, xl0, yl0;

    /** Unused method which still needs to be
    *   implemented. */
    public void mouseClicked(MouseEvent e) {
    }

    /** Unused method which still needs to be
    *   implemented. */
    public void mouseEntered(MouseEvent e) {
    }

    /** Unused method which still needs to be
    *   implemented. */
    public void mouseMoved(MouseEvent e) {
    }

    /** Method which erases the previous rectangle
    *   and draws a new one when the mouse is being
    *   dragged across the plot. */
    public void mouseDragged(MouseEvent e) {
      if (drag) {
        if (prevRect) {
          // this erases the previous rectangle:
          graphicsArea.drawRect(x0,y0,x1,y1);
        }
        prevRect = true;
    
        // this draws the new rectangle:
        x1 = e.getX();
        y1 = e.getY();
        graphicsArea.drawRect(x0,y0,x1,y1);
      } else if (shift) {
        x1 = e.getX();
        y1 = e.getY();
        graphicsArea.shift(xx0,yy0,xl0,yl0,x1,y1);
        graphicsArea.repaint();
      }
    }

    /** Method which erases the last rectangle
    *   when the mouse leaves the plot area,
    *   and cancels the zoom. */
    public void mouseExited(MouseEvent e) {
      if (drag&&prevRect) 
        graphicsArea.drawRect(x0,y0,x1,y1);
      drag = false;
      prevRect = false;
      shift = false;
    }

    /** Method which keeps track of the point
    *   from where the mouse is being dragged */
    public void mousePressed(MouseEvent e) {
      if (e.getButton() == MouseEvent.BUTTON1) {
        drag = true;
        prevRect = false;
        x0 = e.getX();
        y0 = e.getY();
      } else if (e.getButton() == MouseEvent.BUTTON2) {
        shift = true;
        xx0 = graphicsArea.xValue(e.getX());
        yy0 = graphicsArea.yValue(e.getY());
        xl0 = Math.log(xx0);
        yl0 = Math.log(yy0);
      }
    }

    /** Method which performs the zoom-in once
    *   the mouse button has been released. */
    public void mouseReleased(MouseEvent e) {
      if (e.getButton() == MouseEvent.BUTTON1) {

        // To avoid inverting the range, we put the
        // appropriate values in the following
        // variables and use them instead:
        int xleft, xright, ytop, ybottom;

        if (drag) {
          x1 = e.getX();
          y1 = e.getY();
          if (x0==x1) return;
          if (y0==y1) return;
          xleft=x0<x1?x0:x1;
          xright=x0<x1?x1:x0;
          ytop=y0<y1?y0:y1;
          ybottom=y0<y1?y1:y0;
          graphicsArea.zoom(xleft,xright,ybottom,ytop);
          graphicsArea.repaint();
        }
        drag = false;
        prevRect = false;
      } else if (e.getButton() == MouseEvent.BUTTON2) {
        shift = false;
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the autoscale button:
  /////////////////////////////////////////////////////////////////////////////
  private void initAutoScaleButton() {
    AutoScaleButton = createButton("AS");
    AutoScaleButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        graphicsArea.findRange();
        repaint();
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the buttons which write IDL or TXT files:
  /////////////////////////////////////////////////////////////////////////////
  private void initOutputButtons() {
    IDLButton = createButton("IDL");
    TXTButton = createButton("TXT");
    ActionListener l = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        JButton source = (JButton)e.getSource();
        File aFile;
        int result;

        // open in current directory:
        JFileChooser fc = new JFileChooser(".");

        // show the filechooser
        result = fc.showSaveDialog(IDLButton);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
           aFile = fc.getSelectedFile();

          // if file exists, check before overwriting:
          if (aFile.exists()) {
            result = JOptionPane.showConfirmDialog(IDLButton,
                 "Are you sure you want to overwrite "+aFile.getName()+"?",
                 "File exists",JOptionPane.YES_NO_OPTION);
          } else {
            result = JOptionPane.YES_OPTION;
          }
        
          if(result == JOptionPane.YES_OPTION) {
            if (source == IDLButton) {
              graphicsArea.writeIDLfile(aFile,source);
            } else {
              graphicsArea.writeTXTfile(aFile,source);
            }
          }
        }
      }
    };
    IDLButton.addActionListener(l);
    TXTButton.addActionListener(l);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise XlogCheckBox and YlogCheckBox:
  /////////////////////////////////////////////////////////////////////////////
  private void initCheckBoxes() {
    XlogCheckBox = new JCheckBox("x-log",graphicsArea.getXlog());
    XlogCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        graphicsArea.setXlog(XlogCheckBox.isSelected());
        graphicsArea.correctXrange();
        repaint();
      }
    });

    YlogCheckBox = new JCheckBox("y-log",graphicsArea.getYlog());
    YlogCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        graphicsArea.setYlog(YlogCheckBox.isSelected());
        graphicsArea.correctYrange();
        repaint();
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialises the DuplicateButton, which produces a duplicate of the
  // IPlot in a separate window.
  /////////////////////////////////////////////////////////////////////////////
  private void initDuplicateButton() {
    DuplicateButton = createButton("DUPL");
    DuplicateButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        JFrame f = new JFrame(graphicsArea.title);
        IPlot newIPlot = new IPlot(graphicsArea.clone());
        f.add("Center", newIPlot);
        f.setSize(500,400);
        newIPlot.initIPlot();
        newIPlot.setIOenabled(IOenabled);
        newIPlot.repaint();
        f.setVisible(true);
        f.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This clones an IPlot. */
  /////////////////////////////////////////////////////////////////////////////
  public IPlot clone() {
    return new IPlot(this.graphicsArea.clone());
  }

  /////////////////////////////////////////////////////////////////////////////
  /** A constructor which creates an IPlot from a Plot.
  *
  *   @param aPlot a Plot object from which to construct an IPlot
  */
  /////////////////////////////////////////////////////////////////////////////
  public IPlot(Plot aPlot) {
    this.graphicsArea = aPlot;
  }

    
  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public IPlot(double[] x, double[] y, int ls, int ps, Color aColor) {
    graphicsArea = new Plot(x,y,ls,ps,aColor);
    initIPlot();
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public IPlot(double[] x, double[] y, int ls, int ps) {
  
    // Note: it is best to use the correct Plot constructor, rather than
    // doing "this(x,y,ls,ps,Color.black)" because this way any modifications
    // to Plot(x,y,ls,ps) is then taken into account.

    graphicsArea = new Plot(x,y,ls,ps);
    initIPlot();
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public IPlot(double[] x, double[] y, int ls, Color aColor) {
    graphicsArea = new Plot(x,y,ls,aColor);
    initIPlot();
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   */
  /////////////////////////////////////////////////////////////////////////////
  public IPlot(double[] x, double[] y, int ls) {
    graphicsArea = new Plot(x,y,ls);
    initIPlot();
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   */
  /////////////////////////////////////////////////////////////////////////////
  public IPlot(double[] x, double[] y) {
    graphicsArea = new Plot(x,y);
    initIPlot();
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** A constructor based on the different titles and number of data sets.
   *
   * @param title  title above the plot
   * @param xtitle title for the x-axis
   * @param ytitle title for the y-axis
   * @param nDataSets number of data sets (i.e. x[], y[] pairs)
   */
  /////////////////////////////////////////////////////////////////////////////
  public IPlot(String title, String xtitle, String ytitle, int nDataSets) {
    graphicsArea = new Plot(title,xtitle,ytitle,nDataSets);
    initIPlot();
  }

  /////////////////////////////////////////////////////////////////////////////
  /** A constructor based on the different titles.
   *
   * @param title  title above the plot
   * @param xtitle title for the x-axis
   * @param ytitle title for the y-axis
   */
  /////////////////////////////////////////////////////////////////////////////
  public IPlot(String title, String xtitle, String ytitle) {
    graphicsArea = new Plot(title,xtitle,ytitle);
    initIPlot();
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the titles to be modified.
   *
   * @param title  new title for the plot
   * @param xtitle new title for the x-axis
   * @param ytitle new title for the y-axis
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setTitles(String title, String xtitle, String ytitle) {
    graphicsArea.setTitles(title,xtitle,ytitle);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the IDL titles to be modified
   *
   * @param title  title above the plot in the IDL program
   * @param xtitle title for the x-axis in the IDL program
   * @param ytitle title for the y-axis in the IDL program
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setIDLTitles(String title, String xtitle, String ytitle) {
    graphicsArea.setIDLTitles(title,xtitle,ytitle);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the x-axis titles to be modified.  Both the normal and the IDL
   *  x-axis titles are modified.
   *
   *  @param xtitle  new title for the x-axis
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setXtitle(String xtitle) {
    graphicsArea.setXtitle(xtitle);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the IDL x-axis title to be modified.
   *
   *  @param xtitle  new IDL title for the x-axis
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setIDLXtitle(String xtitle) {
    graphicsArea.setIDLXtitle(xtitle);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle, choice of 
   *  symbols and color.
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y, int ls, int ps, Color aColor) {
    graphicsArea.addDataSet(ds,x,y,ls,ps,aColor);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of symbols.
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y, int ls, int ps) {
    graphicsArea.addDataSet(ds,x,y,ls,ps);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of color.
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y, int ls, Color aColor) {
    graphicsArea.addDataSet(ds,x,y,ls,aColor);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle.
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y, int ls) {
    graphicsArea.addDataSet(ds,x,y,ls);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y).
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y) {
    graphicsArea.addDataSet(ds,x,y);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle, choice of 
   *  symbols and color.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls, int ps, Color aColor) {
    graphicsArea.addDataSet(x,y,ls,ps,aColor);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of symbols.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls, int ps) {
    graphicsArea.addDataSet(x,y,ls,ps);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of color.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls, Color aColor) {
    graphicsArea.addDataSet(x,y,ls,aColor);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls) {
    graphicsArea.addDataSet(x,y,ls);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y).
   *
   * @param  x      array of x values
   * @param  y      array of y values
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y) {
    graphicsArea.addDataSet(x,y);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This removes data set number ds from arrays x and y.
   *
   * @param ds index of the data set to be removed (0 <= ds < nDataSets)
   */
  /////////////////////////////////////////////////////////////////////////////
  public void removeDataSet(int ds) {
    graphicsArea.removeDataSet(ds);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This removes all data sets.
   */
  /////////////////////////////////////////////////////////////////////////////
  public void clearDataSets() {
    graphicsArea.clearDataSets();
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls,
   * symbol choice ps, and color choice aColor.
   *
   * @param ds     data set that will be replaced (0 <= ds < nDataSets)
   * @param x      new array of x values
   * @param y      new array of y values
   * @param ls     new linestyle
   * @param ps     new choice of symbols
   * @param aColor new choice of color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls,
                             int ps, Color aColor) {
    graphicsArea.replaceDataSet(ds,x,y,ls,ps,aColor);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls,
   * and color choice aColor.
   *
   * @param ds     data set that will be replaced (0 <= ds < nDataSets)
   * @param x      new array of x values
   * @param y      new array of y values
   * @param ls     new linestyle
   * @param aColor new choice of color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls,
                             Color aColor) {
    graphicsArea.replaceDataSet(ds,x,y,ls,aColor);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls and
   * symbol choice ps.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param x  new array of x values
   * @param y  new array of y values
   * @param ls new linestyle
   * @param ps new choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls, int ps) {
    graphicsArea.replaceDataSet(ds,x,y,ls,ps);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y and linestyle ls.
   *
   * @param ds  data set that will be replaced (0 <= ds < nDataSets)
   * @param x   new array of x values
   * @param y   new array of y values
   * @param ls  new linestyle
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls) {
    graphicsArea.replaceDataSet(ds,x,y,ls);
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param x  new array of x values
   * @param y  new array of y values
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y) {
    graphicsArea.replaceDataSet(ds,x,y);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces the lower x error bar for data set ds.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param leftError new lower x error bar array
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceLeftError(int ds, double[] leftError) {
    graphicsArea.replaceLeftError(ds,leftError);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces the upper x error bar for data set ds.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param rightError new upper x error bar array
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceRightError(int ds, double[] rightError) {
    graphicsArea.replaceRightError(ds,rightError);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces the lower y error bar for data set ds.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param lowerError new lower y error bar array
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceLowerError(int ds, double[] lowerError) {
    graphicsArea.replaceLowerError(ds,lowerError);
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces the upper y error bar for data set ds.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param leftError new upper y error bar array
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceUpperError(int ds, double[] upperError) {
    graphicsArea.replaceUpperError(ds,upperError);
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This sets hErrorArray(ds) with a new boolean value.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param hErrorValue new boolean value
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setHErrorValue(int ds, boolean hErrorValue) {
    graphicsArea.setHErrorValue(ds,hErrorValue);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This sets vErrorArray(ds) with a new boolean value.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param vErrorValue new boolean value
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setVErrorValue(int ds, boolean vErrorValue) {
    graphicsArea.setVErrorValue(ds,vErrorValue);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of extendXrange */
  /////////////////////////////////////////////////////////////////////////////
  public void setExtendXrange(boolean value) {graphicsArea.setExtendXrange(value);}
 
  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of extendYrange */
  /////////////////////////////////////////////////////////////////////////////
  public void setExtendYrange(boolean value) {graphicsArea.setExtendYrange(value);}

  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of reverseXrange */
  /////////////////////////////////////////////////////////////////////////////
  public void setReverseXrange(boolean value) {graphicsArea.setReverseXrange(value);}
 
  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of reverseYrange */
  /////////////////////////////////////////////////////////////////////////////
  public void setReverseYrange(boolean value) {graphicsArea.setReverseYrange(value);}

  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of Xlog */
  /////////////////////////////////////////////////////////////////////////////
  public void setXlog(boolean value) {
    graphicsArea.setXlog(value);
    XlogCheckBox.setSelected(value);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of Ylog */
  /////////////////////////////////////////////////////////////////////////////
  public void setYlog(boolean value) {
    graphicsArea.setYlog(value);
    YlogCheckBox.setSelected(value);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This gets the number of data sets */
  /////////////////////////////////////////////////////////////////////////////
  public int getNDataSets() {return graphicsArea.getNDataSets();}

  /////////////////////////////////////////////////////////////////////////////
  /** This gets the x values of a particular data set. */
  /////////////////////////////////////////////////////////////////////////////
  public double[] getDataSetX(int i) {return graphicsArea.getDataSetX(i);}

  /////////////////////////////////////////////////////////////////////////////
  /** This gets the y values of a particular data set. */
  /////////////////////////////////////////////////////////////////////////////
  public double[] getDataSetY(int i) {return graphicsArea.getDataSetY(i);}

  /////////////////////////////////////////////////////////////////////////////
  /** This gets the value of Xlog */
  /////////////////////////////////////////////////////////////////////////////
  public boolean getXlog() {return graphicsArea.getXlog();}
 
  /////////////////////////////////////////////////////////////////////////////
  /** This gets the value of Ylog */
  /////////////////////////////////////////////////////////////////////////////
  public boolean getYlog() {return graphicsArea.getYlog();}

  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for xmin, xmax, ymin and ymax. */
  /////////////////////////////////////////////////////////////////////////////
  public void findRange() {
    graphicsArea.findRange();
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for xmin, xmax, ymin and ymax, for a
   *  list of data sets.
   * 
   *  @param selection list containing the indices of the data sets for which
   *                   to adjust the values of xmin, xmax, ymin and ymax */
  /////////////////////////////////////////////////////////////////////////////
  public void findRange(int[] selection) {
    graphicsArea.findRange(selection);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This reverses the x-axis by swapping xmin and xmax */
  /////////////////////////////////////////////////////////////////////////////
  public void reverseXaxis() {
    graphicsArea.reverseXaxis();
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This reverses the y-axis by swapping ymin and ymax */
  /////////////////////////////////////////////////////////////////////////////
  public void reverseYaxis() {
    graphicsArea.reverseYaxis();
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set xmin
   *
   *  @param xmin new value for xmin */
  /////////////////////////////////////////////////////////////////////////////
  public void setXmin(double xmin) {graphicsArea.setXmin(xmin);}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set xmax
   *
   *  @param xmax new value for xmax */
  /////////////////////////////////////////////////////////////////////////////
  public void setXmax(double xmax) {graphicsArea.setXmax(xmax);}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set ymin
   *
   *  @param ymin new value for ymin */
  /////////////////////////////////////////////////////////////////////////////
  public void setYmin(double ymin) {graphicsArea.setYmin(ymin);}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set ymax
   *
   *  @param ymax new value for ymax */
  /////////////////////////////////////////////////////////////////////////////
  public void setYmax(double ymax) {graphicsArea.setYmax(ymax);}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to obtain xmin.
   *
   *  @return the value of xmin. */
  /////////////////////////////////////////////////////////////////////////////
  public double getXmin() {return graphicsArea.getXmin();}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to obtain xmax.
   *
   *  @return the value of xmax. */
  /////////////////////////////////////////////////////////////////////////////
  public double getXmax() {return graphicsArea.getXmax();}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to obtain ymin
   *
   *  @return the value of ymin. */
  /////////////////////////////////////////////////////////////////////////////
  public double getYmin() {return graphicsArea.getYmin();}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to obtain ymax
   *
   *  @return the value of ymax. */
  /////////////////////////////////////////////////////////////////////////////
  public double getYmax() {return graphicsArea.getYmax();}

  /////////////////////////////////////////////////////////////////////////////
  /** This repaints the graphics area. */
  /////////////////////////////////////////////////////////////////////////////
  public void repaint() {
    graphicsArea.repaint();
  }

  /////////////////////////////////////////////////////////////////////////////
  // This creates buttons of the right size for IPlot.  I don't know how
  // it works, but simply copied it from JDK SwingSet2 demo program.
  // I modified it to include s as a button title.
  //
  // @param s title of the button
  /////////////////////////////////////////////////////////////////////////////
  private JButton createButton(String s) {
    JButton b = new JButton(s) {
      public Dimension getMaximumSize() {
        int width = Short.MAX_VALUE;
        int height = super.getMaximumSize().height;
        return new Dimension(width, height);
      }
    };
    return b;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This removes all data sets. */
  /////////////////////////////////////////////////////////////////////////////
  public void clear() {
    graphicsArea.clear();
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This nullifies a plot by setting all the arrays contained in
  *   graphicsArea.x and graphicsArea.y to null */
  /////////////////////////////////////////////////////////////////////////////
  public void nullify() {
    graphicsArea.nullify();
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This nullifies a specific data set in a plot by setting its arrays 
  *   contained in graphicsArea.x and graphicsArea.y to null
  *
  *   @param i index of the data set to be nullified */
  /////////////////////////////////////////////////////////////////////////////
  public void nullify(int i) {
    graphicsArea.nullify(i);
  }
}
