/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.Utilities;
import inversion_applet.Mode;

/************************************************************************
* This class contains all theoretical variables associated with a
* calculated mode.  By creating such a structure, it is more
* straightforward keeping track of theoretical modes which have been
* added or removed, due to user manipulations.
************************************************************************/
public class Kernel {

  ///////////////////////////////////////////////////////////////////////
  // Identification:
  ///////////////////////////////////////////////////////////////////////
  /** Harmonic degree of the pulsation mode. */
  public int l;
  /** Radial order of the pulsation mode. */
  public int n;
  /** Associated observed mode. */
  public Mode mode;
 
  ///////////////////////////////////////////////////////////////////////
  // Theorectical eigenmode results:
  ///////////////////////////////////////////////////////////////////////
  /** Index which indicates which equations were solved to obtain the mode
    * (see Pulsations to understand the different possible values) */
  public int equations;
  /** Theoretical frequency (in muHz). */
  public double FreqTheo;
  /** Variational frequency (in muHz). */
  public double FreqVar;
  /** Frequency normalised by sqrt(GM/R^3). */
  public double FreqDimless;
  /** Scaled horizontal displacement. First index = kernel number, Second index =
   *  grid position.  */
  public double[] hDisp;
  /** central value of scaled horizontal displacement. */
  public double hDisp_center;
  /** central quadratic behaviour of scaled horizontal displacement. (= 1/2 of second derivative) */
  public double hDisp_center_pp;
  /** Scaled vertical displacement. First index = kernel number, Second index =
   *  grid position.  */
  public double[] vDisp;
  /** central value of scaled vertical displacement. */
  public double vDisp_center;
  /** central quadratic behaviour of scaled vertical displacement. (= 1/2 of second derivative) */
  public double vDisp_center_pp;
  /** delta P/P_0, where delta P is the lagrangian pressure perturbation */
  public double[] dP_P;
  /** Eulerian pressure perturbation */
  public double[] P;
  /** Eulerian pressure perturbation */
  public double[] rho;
  /** Eulerian pertubation to the gravity potential */
  public double[] phi;
  /** Radial derivative of the eulerian pertubation to the gravity potential */
  public double[] phip;
  /** Y_1^a function from Takata (2006), see Eq. (67) */
  public double[] Y1;
  /** Y_2^a function from Takata (2006), see Eq. (68) */
  public double[] Y2;

  ///////////////////////////////////////////////////////////////////////
  // Various derived quantities:
  ///////////////////////////////////////////////////////////////////////
  /** Unnormalised mode inertia.  When multipled by the square of the 
   *  frequency, this becomes the kinetic energy. This intervenes in
   *  the different kernels and removes any dependance on amplitude. */
  public double IK;
  /** Mode energy.  This intervenes in the surface terms which appear
   *  in structure inversions. */
  public double Enl;
  /** Mean radius defined from rotation kernel */
  public double rRota;
  /** Mean radius defined from rotation kernel divided by r^2 */
  public double rRotaR2;

  ///////////////////////////////////////////////////////////////////////
  // Various kernels:
  ///////////////////////////////////////////////////////////////////////
  /** Rotation kernel. */
  public double[] rotaKernel;
  /** Integrated rotation kernel (int_0^r rotaKernel) */
  public double[] intRotaKernel;
  /** Integrated rotation kernel/r^2 (int_0^r rotaKernel/r^2) */
  public double[] intRotaKernelR2;
  /** {c^2,rho} kernel. */
  public double[] c2rhoKernel;
  /** {rho,c^2} kernel. */
  public double[] rhoc2Kernel;
  /** {rho,Gamma1} kernel. */
  public double[] rhoGamma1Kernel;
  /** Limit of c2rhoKernel/(r^2.rho_0) as r goes to zero. */
  public double center_c2rhoKernel;
  /** Limit of rhoc2Kernel/(r^2.rho_0) as r goes to zero. */
  public double center_rhoc2Kernel;
  /** Limit of rhoGamma1Kernel/(r^2.rho_0) as r goes to zero. */
  public double center_rhoGamma1Kernel;

  /**********************************************************************
  * A simple constructor.
  **********************************************************************/
  public Kernel(int l) {
    this.n = 1;
    this.l = l;
    this.rotaKernel = null;
    this.rhoc2Kernel = null;
    this.FreqVar = Double.NaN;
    this.mode = null;
  }

  /**********************************************************************
  * A more elaborate constructor.
  **********************************************************************/
  public Kernel(int n, int l) {
    this.n = n;
    this.l = l;
    this.rotaKernel = null;
    this.rhoc2Kernel = null;
    this.FreqVar = Double.NaN;
    this.mode = null;
  }

  /**********************************************************************
  * This is method returns "true" if Mode m1 comes before Mode
  * m2 in terms of their (l,n) values.
  **********************************************************************/
  public static boolean compareKernels(Kernel k1, Kernel k2) {
    if (k1.l < k2.l) return true;
    if (k1.l > k2.l) return false;
    if (k1.n < k2.n) return true;
    if (k1.n > k2.n) return false;
    return (k1.FreqTheo < k2.FreqTheo);
  }
}
