/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import graph.IPlot;
import graph.Plot;
import inversion_applet.KernelDataModel;
import inversion_applet.Pulsations;
import inversion_applet.DataCenter;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingConstants;
import javax.swing.filechooser.FileFilter;
import javax.swing.table.DefaultTableCellRenderer;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.File;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileInputStream;
import java.nio.ByteOrder;
import org.netlib.blas.Dgbmv;
import org.netlib.lapack.Dgbsv;
import org.fudaa.dodico.fortran.FortranBinaryInputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.StringTokenizer;

public class KernelPage extends JPanel {
  private DataCenter dataCenter;

  private JPanel buttonBox;

  private JButton findMatchButton;
  private JButton clearMatchButton;

  private JButton sortButton;
  private JButton removeRowsButton;
  private JButton clearButton;
  private JButton writeKernelInfoButton;
  private JButton plotKernelsButton;
  private JButton plotKernelsTogetherButton;
  private JButton plotRatiosButton;
  private JButton plotRatiosR2Button;

  private JButton calculateKernelsButton;
  private JSpinner OrderSpinner;
  private SpinnerNumberModel OrderSpinnerModel;
  private JRadioButton LagrangeEquations;
  private JRadioButton EulerEquations;
  private JRadioButton JCDEquations;
  private JTable kernelTable;
  private KernelDataModel kernelData;
  private JScrollPane tableScroller;

  private JButton readEFunctionsButton;
  private JPanel EFunctionPanel;
  private JLabel EFunctionTypeLabel = new JLabel("<html><font color='blue'>&nbsp;Eigenfunction format</font></html>");
  private JLabel Endianness = new JLabel("<html><font color='blue'>&nbsp;Endianness</font></html>");
  private JLabel nfmodeLabel = new JLabel("<html><font color='blue'>&nbsp;Value of nfmode</font></html>");
  private JLabel NativeEndianLabel;
  private JRadioButton AMDEtype;
  private JRadioButton FAMDEtype;
  private JRadioButton FILOUtype;
  private JRadioButton[] nfRadioButtons;
  private JRadioButton NativeEndian;
  private JRadioButton OppositeEndian;
  private ButtonGroup EFunctionGroup;
  private ButtonGroup EndianGroup;
  private int nfmode;
  private String EFunctionType;
  private boolean isNative;

  private static final int maxPlots = 10;
  private Pulsations pulsations;  // this calculates pulsation modes

  // sequence of ascii codes used to terminate the header in an
  // FAMDE file:
  private static final int[] termSeq = new int[] {12,10,12,10,12,10};

  // maximal number of grid points in eigenfunctions (this is useful
  // for testing an AMDE file to make sure it actually is in AMDE
  // format):
  private static final int maxNgrid= 300000;

  /********************************************************************
  * A constructor.
  ********************************************************************/
  public KernelPage(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
  }

  /********************************************************************
  * This creates the table with the information on the theoretically
  * calculated pulsation frequencies, modes and kernels.
  ********************************************************************/
  private void createTable() {
    // Create the table data model:
    kernelData = new KernelDataModel(dataCenter,buttonBox);

    // Create the table
    kernelTable = new JTable(kernelData);
    dataCenter.kernelTable = this.kernelTable; 
      // this is necessary, so that the table can be updated, even from
      // the frequency data page.
    
    // Change the cell renderers:
    Utilities.IntegerRenderer ir = new Utilities.IntegerRenderer(Color.WHITE,Color.decode("#B8CFE5"));
    kernelTable.getColumn("n").setCellRenderer(ir);
    kernelTable.getColumn("\u2113").setCellRenderer(ir);
    Utilities.DoubleRenderer dr1 = new Utilities.DoubleRenderer(3,Color.WHITE,Color.decode("#B8CFE5"));
    Utilities.DoubleRenderer dr2 = new Utilities.DoubleRenderer(3,Color.YELLOW,Color.decode("#B8CFE5"));
    kernelTable.getColumn("<html>\u03BD<sub>theo</sub> (\u03BCHz)</html>").setCellRenderer(dr1);
    kernelTable.getColumn("<html>\u03BD<sub>obs</sub> (\u03BCHz)</html>").setCellRenderer(dr2);
    kernelTable.getColumn("<html>\u03BD<sub>var</sub>-\u03BD<sub>theo</sub> (\u03BCHz)</html>").setCellRenderer(dr1);
    EquationRenderer er = new EquationRenderer();
    kernelTable.getColumn("Source").setCellRenderer(er);

    // Create a scroll pane for the table
    tableScroller = new JScrollPane(kernelTable);
    return;
  }

  /////////////////////////////////////////////////////////////////////
  /** This sets up the graphical user interface */
  /////////////////////////////////////////////////////////////////////
  public void init() {
    pulsations = new Pulsations(dataCenter);
    createTable();
    initMatchButtons();
    initSortButton();
    initClearButtons();
    initOrderSpinner();
    initCalculateKernelsButton();
    initPlotKernelsButton();
    initEquationRadioButtons();
    initEndians();
    initNfRadioButtons();
    initReadEFunctionsButton();
    initEFunctionTypes();
    initEFunctionPanel();

    buttonBox = new JPanel();
    buttonBox.setLayout(new GridBagLayout());
    GridBagConstraints c = new GridBagConstraints();
    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridx = 0;
    c.gridy = 0;
    c.gridwidth = 3;
    buttonBox.add(findMatchButton,c);
    c.gridy++;
    buttonBox.add(clearMatchButton,c);
    c.gridy++;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    buttonBox.add(sortButton,c);
    c.gridy++;
    buttonBox.add(removeRowsButton,c);
    c.gridy++;
    buttonBox.add(clearButton,c);
    c.gridy++;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    buttonBox.add(writeKernelInfoButton,c);
    c.gridy++;
    buttonBox.add(plotKernelsButton,c);
    c.gridy++;
    buttonBox.add(plotKernelsTogetherButton,c);
    c.gridy++;
    buttonBox.add(plotRatiosButton,c);
    c.gridy++;
    buttonBox.add(plotRatiosR2Button,c);
    c.gridy++;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    buttonBox.add(calculateKernelsButton,c);
    c.gridy++;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    buttonBox.add(new JLabel("Order of the numerical scheme"),c);
    c.gridy++;
    buttonBox.add(new JLabel("(recommended value = 4):"),c);
    c.gridy++;
    buttonBox.add(OrderSpinner,c);
    c.gridy++;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    buttonBox.add(new JLabel("Use the following variables"),c);
    c.gridy++;
    buttonBox.add(new JLabel("in the pulsation equations:"),c);
    c.gridwidth = 1;
    c.gridy++;
    c.gridx = 0;
    buttonBox.add(JCDEquations,c);
    c.gridx = 1;
    buttonBox.add(LagrangeEquations,c);
    c.gridx = 2;
    buttonBox.add(EulerEquations,c);
    c.gridwidth = 3;
    c.gridy++;
    c.gridx = 0;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    buttonBox.add(readEFunctionsButton,c);

    setLayout(new BorderLayout());
    add("Center", tableScroller);
    add("East",buttonBox);
  }

  ///////////////////////////////////////////////////////////
  // this initialises the find and clear match buttons
  ///////////////////////////////////////////////////////////
  private void initMatchButtons() {
    findMatchButton = Utilities.createButton("Find match");
    findMatchButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.findMatch();
        dataCenter.findDuplicates();
      }
    });

    clearMatchButton = Utilities.createButton("Clear match");
    clearMatchButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.clearMatch();
        dataCenter.clearDuplicates();
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // this initialises sortButton
  ///////////////////////////////////////////////////////////
  private void initSortButton() {
    sortButton = Utilities.createButton("Sort");
    sortButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        kernelData.sort();       // sort the data
        kernelTable.updateUI();  // update the table
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // this initialises removeRowsButton and clearButton
  ///////////////////////////////////////////////////////////
  private void initClearButtons() {
    removeRowsButton = Utilities.createButton("Remove selected row(s)");
    removeRowsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // find out which rows are selected
        int[] sr = kernelTable.getSelectedRows();

        // This stops any ongoing editing in a cell:
        kernelTable.removeEditor();

        // This clears the selection (so as to avoid having
        // non-existent rows selected):
        kernelTable.clearSelection();

        if (sr.length == dataCenter.kernels.size()) {
          // if the entire table is selected, just clear
          // the data (this is more rapid and sets the
          // correct value for dataCenter.DFloaded):
            dataCenter.clearKernels();
        } else {
          // Remove the selected rows (the loop is backwards
          // to avoid problems with modified indices):
          for(int i = sr.length-1; i >= 0; i--) 
            kernelData.removeRow(sr[i]);
          // this updates the table:
          kernelTable.updateUI();
          dataCenter.findDuplicates();
        }
      }
    });

    clearButton = Utilities.createButton("Clear kernels");
    clearButton.addActionListener(new ActionListener(){
      public void actionPerformed(ActionEvent e) {
        kernelTable.removeEditor();
        dataCenter.clearKernels();
      }
    });
  }

  /////////////////////////////////////////////////////////////////////
  // This initialises the (pulsation) order text field
  /////////////////////////////////////////////////////////////////////
  private void initOrderSpinner() {
    OrderSpinnerModel = new SpinnerNumberModel(2*pulsations.order,2,20,2);
    OrderSpinner = new JSpinner(OrderSpinnerModel);
  }

  /////////////////////////////////////////////////////////////////////
  // initialise calculateKernelsButton:
  /////////////////////////////////////////////////////////////////////
  private void initCalculateKernelsButton() {
    calculateKernelsButton = Utilities.createButton("Calculate kernels");
    calculateKernelsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!dataCenter.modelLoaded) {
          JOptionPane.showMessageDialog(
            calculateKernelsButton,
            "No model loaded, please load the model first.",
            "No model",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }

        pulsations.order = OrderSpinnerModel.getNumber().intValue()/2;
        OrderSpinnerModel.setValue(2*pulsations.order);  // in case there's an odd number

        if ((dataCenter.nfreq+dataCenter.nRS) == 0) {
          JOptionPane.showMessageDialog(
            calculateKernelsButton,
            "No frequencies loaded, please load frequencies first.",
            "No frequencies",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }

        ArrayList<Integer> domains = Numerics.findDomains(dataCenter.grid);
        int minDomainSize = Numerics.findMinDomainSize(domains);
        if (pulsations.order > minDomainSize) {
          JOptionPane.showMessageDialog(
            calculateKernelsButton,
            "The order of the pulsation calculations is too high,\n"+
            "compared to the size of some of your stellar domains.\n"+
            "The order should be no greater than "+minDomainSize+" for this\n"+
            "particular stellar model",
            "Numerical order too high",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }

        pulsations.findPulsations();
        pulsations.buildRotaKernels();
        pulsations.buildStructKernels();
        pulsations.applyVariational();

        // update some status variables in the dataCenter.
        dataCenter.rotationInversionUpToDate = false;
        dataCenter.structInversionUpToDate = false;
        dataCenter.meanDensityEstimateUpToDate = false;
        dataCenter.acousticTimeEstimateUpToDate = false;
        dataCenter.smallSeparationEstimateUpToDate = false;
        dataCenter.findDuplicates();
        
        // this updates the table
        kernelTable.updateUI();
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // This initialises the equation radio buttons
  ///////////////////////////////////////////////////////////
  private void initEquationRadioButtons() {
    LagrangeEquations = new JRadioButton("\u03B4P/P");
    LagrangeEquations.addActionListener(new ActionListener(){
      public void actionPerformed(ActionEvent e) {
        pulsations.equations = Pulsations.Lagrange;
      }
    });

    EulerEquations = new JRadioButton("P', \u03C1'");
    EulerEquations.addActionListener(new ActionListener(){
      public void actionPerformed(ActionEvent e) {
        pulsations.equations = Pulsations.Euler;
      }
    });

    JCDEquations = new JRadioButton("JCD");
    JCDEquations.addActionListener(new ActionListener(){
      public void actionPerformed(ActionEvent e) {
        pulsations.equations = Pulsations.JCD;
      }
    });

    ButtonGroup equationGroup = new ButtonGroup();
    equationGroup.add(LagrangeEquations);
    equationGroup.add(EulerEquations);
    equationGroup.add(JCDEquations);

    if (pulsations.equations == Pulsations.Lagrange) LagrangeEquations.setSelected(true);
    if (pulsations.equations == Pulsations.Euler)    EulerEquations.setSelected(true);
    if (pulsations.equations == Pulsations.JCD)      JCDEquations.setSelected(true);
  }

  ///////////////////////////////////////////////////////////
  // this initialises various buttons related to plotting
  // the kernels and various tests based on the rotational
  // splittings.
  ///////////////////////////////////////////////////////////
  private void initPlotKernelsButton() {
    writeKernelInfoButton = Utilities.createButton("Write kernel info.");
    writeKernelInfoButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showSaveDialog(writeKernelInfoButton);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            result = JOptionPane.showConfirmDialog(writeKernelInfoButton,
                 "Are you sure you want to overwrite "+aFile.getName()+"?",
                 "File exists",JOptionPane.YES_NO_OPTION);
          } else {
            result = JOptionPane.YES_OPTION;
          }

          if(result == JOptionPane.YES_OPTION)
            writeKernelInfo(aFile.getPath());

          if (aFile.exists())
            dataCenter.currentDirectory = aFile.getParentFile();  
        }
      }
    });

    plotKernelsButton  = Utilities.createButton("Plot kernels separately");
    plotKernelsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // find out which rows are selected
        int[] sr = kernelTable.getSelectedRows();

        // this plots the kernels:
        int nPlots = sr.length>maxPlots?maxPlots:sr.length;
        for(int i=0; i < nPlots; i++) plotKernel(sr[i]);

        // If too many rows are selected.  The program does not plot them
        // all and notifies the user:
        if (nPlots < sr.length) {
          JOptionPane.showMessageDialog(plotKernelsButton,
            "Too many rows are selected: I have\n"+
            "only plotted the first "+maxPlots+".",
            "Warning",
            JOptionPane.WARNING_MESSAGE
          );
        }
      }
    });

    plotKernelsTogetherButton  = Utilities.createButton("Plot kernels together");
    plotKernelsTogetherButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        plotKernels(kernelTable.getSelectedRows());
      }
    });

    plotRatiosButton = Utilities.createButton("Plot ratios");
    plotRatiosButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        plotRatios(kernelTable.getSelectedRows());
      }
    });

    plotRatiosR2Button = Utilities.createButton("Plot R\u00B2 ratios");
    plotRatiosR2Button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        plotRatiosR2(kernelTable.getSelectedRows());
      }
    });
  }

  /////////////////////////////////////////////////////////
  // Write a file with theoretical and corresponding observed
  // frequencies, variational errors on the theoretical
  // frequencies, and the method used for obtaining the
  // theoretical frequencies.
  /////////////////////////////////////////////////////////
  protected void writeKernelInfo(String filename) {
    File f = new File(filename);
    PrintWriter out;
    try {
      out = new PrintWriter(new FileWriter(f));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        writeKernelInfoButton,
        "Unable to write to file \""+filename+"\"",
        "Failure",
        JOptionPane.WARNING_MESSAGE
      );
      return;
    }

    out.printf("#%2s %3s %22s %22s %22s %7s%n",
               "l","n","nu_theo (muHz)","nu_obs (muHz)","nu_var-nu_theo (muHz)","Source");
    for(int i = 0; i < dataCenter.kernels.size(); i++) {
      out.printf("%3d ",dataCenter.kernels.get(i).l);
      out.printf("%3d ",dataCenter.kernels.get(i).n);
      out.printf("%22.15e ",dataCenter.kernels.get(i).FreqTheo);
      if (dataCenter.kernels.get(i).mode != null) {
        out.printf("%22.15e ",dataCenter.kernels.get(i).mode.FreqObs);
      } else {
        out.printf("          0.0          ");
      }
      out.printf("%22.15e ",dataCenter.kernels.get(i).FreqVar-dataCenter.kernels.get(i).FreqTheo);
      out.printf("%7s%n",pulsations.equationName[dataCenter.kernels.get(i).equations]);
    }
    out.close();
  }

  ///////////////////////////////////////////////////////////
  // This plots the different functions associated with an
  // eigenmode:
  ///////////////////////////////////////////////////////////
  private void plotKernel(int l) {
    int nn = dataCenter.kernels.get(l).n;
    int ll = dataCenter.kernels.get(l).l;
    String title = "The n="+nn+", \u2113="+ll+" pulsation mode";
    
    JTabbedPane tabbedPane = new JTabbedPane();
    IPlot onePlot;

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).vDisp);
    onePlot.setTitles("Vertical displacement","Radial grid, r/R","\u03BEr");
    onePlot.setIDLTitles("Vertical displacement","Radial grid, r/R","!7n!17!Dr!N");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03BE<sub>r</sub></font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).hDisp);
    onePlot.setTitles("Horizontal displacement","Radial grid, r/R","\u03BEh");
    onePlot.setIDLTitles("Horizontal displacement","Radial grid, r/R","!7n!17!Dh!N");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03BE<sub>h</sub></font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).dP_P);
    onePlot.setTitles("Pressure perturbation","Radial grid, r/R","\u03B4P/P");
    onePlot.setIDLTitles("Pressure perturbation","Radial grid, r/R","!7d!17P/P");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03B4P/P</font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).P);
    onePlot.setTitles("Pressure perturbation","Radial grid, r/R","P'");
    onePlot.setIDLTitles("Pressure perturbation","Radial grid, r/R","!17P'");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>P'</font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).rho);
    onePlot.setTitles("Density perturbation","Radial grid, r/R","\u03C1'");
    onePlot.setIDLTitles("Density perturbation","Radial grid, r/R","!7q'");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03C1'</font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).phi);
    onePlot.setTitles("Gravitational potential variation","Radial grid, r/R","\u03C6'");
    onePlot.setIDLTitles("Gravitational potential variation","Radial grid, r/R","!7u'");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03C6'</font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).phip);
    onePlot.setTitles("Gravity variation","Radial grid, r/R","d(\u03C6')/dr");
    onePlot.setIDLTitles("Gravity variation","Radial grid, r/R","!17d(!7u'!17)/dr");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>d(\u03C6')/dr</font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).Y1);
    onePlot.setTitles("Y\u2081 (Takata 2006, Eq. 67)","Radial grid, r/R","Y\u2081");
    onePlot.setIDLTitles("Y!D1!N (Takata 2006, Eq. 67)","Radial grid, r/R","Y!D1!N");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>Y\u2081</font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).Y2);
    onePlot.setTitles("Y\u2082 (Takata 2006, Eq. 68)","Radial grid, r/R","Y\u2082");
    onePlot.setIDLTitles("Y!D2!N (Takata 2006, Eq. 68)","Radial grid, r/R","Y!D2!N");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>Y\u2082</font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).rotaKernel);
    onePlot.setTitles("Rotation kernel", "Radius, r/R", "Profile");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>K<sub>\u03A9</sub></font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).intRotaKernel);
    onePlot.setTitles("\u222BRotation kernel", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!9i!17Rotation kernel", "Radius, r/R", "Profile");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>\u222BK<sub>\u03A9</sub></font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).intRotaKernelR2);
    onePlot.setTitles("\u222BK\u03A9/r\u00B2", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!9i!17K!7!DX!N!17/r!E2!N", "Radius, r/R", "Profile");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>\u222BK<sub>\u03A9</sub>/r\u00B2</font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).c2rhoKernel);
    onePlot.setTitles("c\u00B2,\u03C1 or \u0393\u2081,\u03C1 kernel", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!17!Nc!E2!N,!7q!17 or !7C!17!I1!N,!7q!17 kernel", "Radius, r/R", "Profile");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>K<sub>c\u00B2,\u03C1</sub> or K<sub>\u0393\u2081,\u03C1</sub></font></html>",onePlot);

    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).rhoc2Kernel);
    onePlot.setTitles("\u03C1,c\u00B2 kernel", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!7!Nq!17,c!E2!N kernel", "Radius, r/R", "Profile");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>K<sub>\u03C1,c\u00B2</sub></font></html>",onePlot);

    int ngrid = dataCenter.grid.length;
    double[] aux = new double[ngrid];
    for(int i=0; i<ngrid; i++) aux[i] = 2.0*Math.PI*Math.pow(dataCenter.grid[i],2.0)*dataCenter.rho[i];
    onePlot = new IPlot(dataCenter.grid,dataCenter.kernels.get(l).rhoGamma1Kernel);
    onePlot.addDataSet(dataCenter.grid,aux,Plot.dotted,Plot.noSymbol,Color.blue);
    onePlot.setTitles("\u03C1,\u0393\u2081 kernel", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!7!Nq!17,!7!NC!17!I1!N kernel", "Radius, r/R", "Profile");
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>K<sub>\u03C1,\u0393\u2081</sub></font></html>",onePlot);

    JPanel labelBox = new JPanel();
    labelBox.setLayout(new BoxLayout(labelBox, BoxLayout.Y_AXIS));
    labelBox.add(new JLabel(String.format("<html><font color='blue'>&nbsp;Mode %d/%d</font></html>",
                 l+1, dataCenter.kernels.size())));
    labelBox.add(new JLabel(String.format("    \u03BD=%.2f \u03BCHz",
                 dataCenter.kernels.get(l).FreqTheo)));
    labelBox.add(new JLabel(String.format("<html>&nbsp;&nbsp;&nbsp;&nbsp;\u03BD/(GM/R<sup>3</sup>)<sup>1/2</sup>=%.4f</html>",
                 dataCenter.kernels.get(l).FreqDimless)));
    labelBox.add(new JLabel(String.format("<html>&nbsp;&nbsp;&nbsp;&nbsp;\u03BD<sub>var</sub>-\u03BD<sub>theo</sub>=%.2e</html>",
                 dataCenter.kernels.get(l).FreqVar - dataCenter.kernels.get(l).FreqTheo)));
    labelBox.add(new JLabel(String.format("    n=%d",dataCenter.kernels.get(l).n)));
    labelBox.add(new JLabel(String.format("    \u2113=%d",dataCenter.kernels.get(l).l)));
    labelBox.add(new JLabel(String.format("<html>&nbsp;&nbsp;&nbsp;&nbsp;E<sub>n\u2113</sub>=%.2e</html>",
                 dataCenter.kernels.get(l).Enl)));

    JFrame f = new JFrame(title);
    f.setLayout(new BorderLayout());
    f.add("Center",tabbedPane);
    f.add("East",labelBox);
    f.setSize(700,600);
    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    f.setVisible(true);    
  }

  ///////////////////////////////////////////////////////////
  // This plots selected rotation kernels on the same plot.
  ///////////////////////////////////////////////////////////
  private void plotKernels(int[] selection) {
    if (selection.length == 0) {
      JOptionPane.showMessageDialog(plotKernelsButton,
        "Please select some mode(s) first.",
        "Error",
        JOptionPane.ERROR_MESSAGE
        );
      return;
    }

    int i;

    // make a color array to be used in the plots:
    Color[] colors = new Color[selection.length];
    double cnst = 2.0*Math.PI/6.0;
    for(i=0; i<selection.length; i++) {
      colors[i] = new Color((float)(0.5+0.5*Math.cos(cnst*(double)(i-4))),
                            (float)(0.5+0.5*Math.cos(cnst*(double)(i-2))),
                            (float)(0.5+0.5*Math.cos(cnst*(double)i)));
    }

    JTabbedPane tabbedPane = new JTabbedPane();
    IPlot onePlot;

    // create plots
    onePlot = new IPlot("Vertical displacement","Radial grid, r/R","\u03BEr");
    onePlot.setIDLTitles("Vertical displacement","Radial grid, r/R","!7n!17!Dr!N");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).vDisp,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03BE<sub>r</sub></font></html>",onePlot);

    onePlot = new IPlot("Horizontal displacement","Radial grid, r/R","\u03BEh");
    onePlot.setIDLTitles("Horizontal displacement","Radial grid, r/R","!7n!17!Dh!N");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).hDisp,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03BE<sub>h</sub></font></html>",onePlot);

    onePlot = new IPlot("Pressure perturbation","Radial grid, r/R","\u03B4P/P");
    onePlot.setIDLTitles("Pressure perturbation","Radial grid, r/R","!7d!17P/P");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).dP_P,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03B4P/P</font></html>",onePlot);

    onePlot = new IPlot("Pressure perturbation","Radial grid, r/R","P'");
    onePlot.setIDLTitles("Pressure perturbation","Radial grid, r/R","!17P'");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).P,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>P'</font></html>",onePlot);

    onePlot = new IPlot("Pressure perturbation","Radial grid, r/R","\u03C1'");
    onePlot.setIDLTitles("Pressure perturbation","Radial grid, r/R","!7q'");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).rho,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03C1'</font></html>",onePlot);

    onePlot = new IPlot("Gravitational potential variation","Radial grid, r/R","\u03C6'");
    onePlot.setIDLTitles("Gravitational potential variation","Radial grid, r/R","!7u'");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).phi,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>\u03C6'</font></html>",onePlot);

    onePlot = new IPlot("Gravity variation","Radial grid, r/R","d(\u03C6')/dr");
    onePlot.setIDLTitles("Gravity variation","Radial grid, r/R","!17d(!7u'!17)/dr");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).phip,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>d(\u03C6')/dr</font></html>",onePlot);

    onePlot = new IPlot("Y\u2081 (Takata 2006, Eq. 67)","Radial grid, r/R","Y\u2081");
    onePlot.setIDLTitles("Y!D1!N (Takata 2006, Eq. 67)","Radial grid, r/R","Y!D1!N");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).Y1,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>Y\u2081</font></html>",onePlot);

    onePlot = new IPlot("Y\u2082 (Takata 2006, Eq. 68)","Radial grid, r/R","Y\u2082");
    onePlot.setIDLTitles("Y!D2!N (Takata 2006, Eq. 68)","Radial grid, r/R","Y!D2!N");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).Y2,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='blue'>Y\u2082</font></html>",onePlot);

    onePlot = new IPlot("Rotation kernel", "Radius, r/R", "Profile");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).rotaKernel,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>K<sub>\u03A9</sub></font></html>",onePlot);

    onePlot = new IPlot("\u222BRotation kernel", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!9i!17Rotation kernel", "Radius, r/R", "Profile");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).intRotaKernel,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>\u222BK<sub>\u03A9</sub></font></html>",onePlot);

    onePlot = new IPlot("\u222BK\u03A9/r\u00B2", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!9i!17K!7!DX!N!17/r!E2!N", "Radius, r/R", "Profile");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).intRotaKernelR2,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>\u222BK<sub>\u03A9</sub>/r\u00B2</font></html>",onePlot);

    onePlot = new IPlot("c\u00B2,\u03C1 or \u0393\u2081,\u03C1 kernel", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!17!Nc!E2!N,!7q!17 or !7C!17!I1!N,!7q!17 kernel", "Radius, r/R", "Profile");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).c2rhoKernel,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>K<sub>c\u00B2,\u03C1</sub> or K<sub>\u0393\u2081,\u03C1</sub></font></html>",onePlot);

    onePlot = new IPlot("\u03C1,c\u00B2 kernel", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!7!Nq!17,c!E2!N kernel", "Radius, r/R", "Profile");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).rhoc2Kernel,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>K<sub>\u03C1,c\u00B2</sub></font></html>",onePlot);

    onePlot = new IPlot("\u03C1,\u0393\u2081 kernel", "Radius, r/R", "Profile");
    onePlot.setIDLTitles("!7!Nq!17,!7!NC!17!I1!N kernel", "Radius, r/R", "Profile");
    for(i=0; i<selection.length; i++) {
      onePlot.addDataSet(dataCenter.grid,dataCenter.kernels.get(selection[i]).rhoGamma1Kernel,
                         Plot.solid,Plot.noSymbol,colors[i]);
    }
    int ngrid = dataCenter.grid.length;
    double[] aux = new double[ngrid];
    for(i=0; i<ngrid; i++) aux[i] = 2.0*Math.PI*Math.pow(dataCenter.grid[i],2.0)*dataCenter.rho[i];
    onePlot.addDataSet(dataCenter.grid,aux,Plot.dotted,Plot.noSymbol,Color.blue);
    onePlot.setIOenabled(!dataCenter.isApplet);
    onePlot.findRange();
    tabbedPane.add("<html><font color='red'>K<sub>\u03C1,\u0393\u2081</sub></font></html>",onePlot);

    JFrame f = new JFrame(selection.length+" eigenmodes");
    f.setLayout(new BorderLayout());
    f.add("Center",tabbedPane);
    f.setSize(700,600);
    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    f.setVisible(true);    
  }

  ///////////////////////////////////////////////////////////
  // This plots ratios of the antiderivatives of rotation
  // kernels for all pairs of modes in the selection.
  ///////////////////////////////////////////////////////////
  private void plotRatios(int[] selection) {
    if (selection.length < 2) {
      JOptionPane.showMessageDialog(
        plotRatiosButton,
        "Please select at least two modes first.",
        "Error",
        JOptionPane.ERROR_MESSAGE
        );
      return;
    }

    if (selection.length > 7) {
      int nPlots = selection.length*(selection.length-1);
      int result = JOptionPane.showConfirmDialog(
          plotRatiosButton,
          "You have selected "+selection.length+" modes.\n"+
          "This will result in "+nPlots+" plots.\n"+
          "Are you sure you want to proceed?",
          "Many plots warning",
          JOptionPane.YES_NO_OPTION
      );
      if (result == JOptionPane.NO_OPTION) return;
    }

    int i, j, ii, jj, k, ngrid = dataCenter.grid.length;
    double[] aux = new double[ngrid];
    IPlot aPlot;
    double cnst;
    String title;
    JTabbedPane tb = new JTabbedPane();
    double IK1, IK2, xi1, xi2, xi1_pp, xi2_pp, eta1, eta2, eta1_pp, eta2_pp, ll;

    for(i=0; i<selection.length; i++) {
      ii = selection[i];
      if (dataCenter.kernels.get(ii).rotaKernel == null) continue;
      for(j=0; j<selection.length; j++) {
        jj = selection[j];
        if (dataCenter.kernels.get(jj).rotaKernel == null) continue;
        if (i==j) continue;
        aPlot = new IPlot("Ratio of \u222B K\u03A9", "Radius, r/R", "Ratio");
        aPlot.setIDLTitles("!17Ratio of !9i!17K!D!7X!N!17", "Radius, r/R", "Ratio");
        cnst = dataCenter.kernels.get(jj).intRotaKernel[ngrid-1]
             / dataCenter.kernels.get(ii).intRotaKernel[ngrid-1];
        for(k=1;k<ngrid;k++) {
          aux[k] = cnst * dataCenter.kernels.get(ii).intRotaKernel[k]
                        / dataCenter.kernels.get(jj).intRotaKernel[k];
        }
        // treat center separately
        if ((dataCenter.kernels.get(ii).equations == Pulsations.FromFile)||
            (dataCenter.kernels.get(jj).equations == Pulsations.FromFile)||
            (dataCenter.kernels.get(ii).l != dataCenter.kernels.get(jj).l)) {
          // it could be anything, ranging from 0 to infinity:
          aux[0] = Double.NaN;
        } else {
          // only calculate internal ratio when you can:
          IK1     = dataCenter.kernels.get(ii).IK;
          IK2     = dataCenter.kernels.get(jj).IK;
          xi1     = dataCenter.kernels.get(ii).vDisp_center;
          xi2     = dataCenter.kernels.get(jj).vDisp_center;
          xi1_pp  = dataCenter.kernels.get(ii).vDisp_center_pp;
          xi2_pp  = dataCenter.kernels.get(jj).vDisp_center_pp;
          eta1    = dataCenter.kernels.get(ii).hDisp_center;
          eta2    = dataCenter.kernels.get(jj).hDisp_center;
          eta1_pp = dataCenter.kernels.get(ii).hDisp_center_pp;
          eta2_pp = dataCenter.kernels.get(jj).hDisp_center_pp;
          ll      = (double)dataCenter.kernels.get(ii).l;
          if (1==dataCenter.kernels.get(ii).l) {
            aux[0] = cnst*(IK2*(xi1_pp-eta1_pp)*(xi1_pp-eta1_pp))
                         /(IK1*(xi2_pp-eta2_pp)*(xi2_pp-eta2_pp));
          } else {
            aux[0] = cnst* IK2*((xi1-eta1)*(xi1-eta1)+(ll*ll+ll-2.0)*eta1*eta1)
                         /(IK1*((xi2-eta2)*(xi2-eta2)+(ll*ll+ll-2.0)*eta2*eta2));
          }
        }

        aPlot.addDataSet(dataCenter.grid,aux);
        aPlot.setIOenabled(!dataCenter.isApplet);
        aPlot.setExtendXrange(true);
        aPlot.findRange();

        title = "("+dataCenter.kernels.get(ii).n+","
                   +dataCenter.kernels.get(ii).l+")/("
                   +dataCenter.kernels.get(jj).n+","
                   +dataCenter.kernels.get(jj).l+")";
        tb.add(title,aPlot);
      }
    }

    JFrame f = new JFrame("Ratio of \u222B K\u03A9");
    f.setLayout(new BorderLayout());
    f.add("Center",tb);
    f.setSize(700,600);
    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    f.setVisible(true);    

  }

  ///////////////////////////////////////////////////////////
  // This plots ratios of the antiderivatives of rotation
  // kernels for all pairs of modes in the selection.
  ///////////////////////////////////////////////////////////
  private void plotRatiosR2(int[] selection) {
    if (selection.length < 2) {
      JOptionPane.showMessageDialog(
        plotRatiosR2Button,
        "Please select at least two modes first.",
        "Error",
        JOptionPane.ERROR_MESSAGE
        );
      return;
    }

    if (selection.length > 7) {
      int nPlots = selection.length*(selection.length-1);
      int result = JOptionPane.showConfirmDialog(
          plotRatiosR2Button,
          "You have selected "+selection.length+" modes.\n"+
          "This will result in "+nPlots+" plots.\n"+
          "Are you sure you want to proceed?",
          "Many plots warning",
          JOptionPane.YES_NO_OPTION
      );
      if (result == JOptionPane.NO_OPTION) return;
    }

    int i, j, ii, jj, k, ngrid = dataCenter.grid.length;
    double[] aux = new double[ngrid];
    IPlot aPlot;
    double cnst;
    String title;
    JTabbedPane tb = new JTabbedPane();
    double IK1, IK2, xi1, xi2, xi1_pp, xi2_pp, eta1, eta2, eta1_pp, eta2_pp, ll;

    for(i=0; i<selection.length; i++) {
      ii = selection[i];
      if (dataCenter.kernels.get(ii).rotaKernel == null) continue;
      for(j=0; j<selection.length; j++) {
        jj = selection[j];
        if (dataCenter.kernels.get(jj).rotaKernel == null) continue;
        if (i==j) continue;
        aPlot = new IPlot("Ratio of \u222B K\u03A9/r\u00B2", "Radius, r/R", "Ratio");
        aPlot.setIDLTitles("!17Ratio of !9i!17K!D!7X!N!17/r!E2!N", "Radius, r/R", "Ratio");
        cnst = dataCenter.kernels.get(jj).intRotaKernelR2[ngrid-1]
             / dataCenter.kernels.get(ii).intRotaKernelR2[ngrid-1];
        for(k=1;k<ngrid;k++) {
          aux[k] = cnst * dataCenter.kernels.get(ii).intRotaKernelR2[k]
                        / dataCenter.kernels.get(jj).intRotaKernelR2[k];
        }
        // treat center separately
        if ((dataCenter.kernels.get(ii).equations == Pulsations.FromFile)||
            (dataCenter.kernels.get(jj).equations == Pulsations.FromFile)||
            (dataCenter.kernels.get(ii).l != dataCenter.kernels.get(jj).l)) {
          // it could be anything, ranging from 0 to infinity:
          aux[0] = Double.NaN;
        } else {
          IK1     = dataCenter.kernels.get(ii).IK;
          IK2     = dataCenter.kernels.get(jj).IK;
          xi1     = dataCenter.kernels.get(ii).vDisp_center;
          xi2     = dataCenter.kernels.get(jj).vDisp_center;
          xi1_pp  = dataCenter.kernels.get(ii).vDisp_center_pp;
          xi2_pp  = dataCenter.kernels.get(jj).vDisp_center_pp;
          eta1    = dataCenter.kernels.get(ii).hDisp_center;
          eta2    = dataCenter.kernels.get(jj).hDisp_center;
          eta1_pp = dataCenter.kernels.get(ii).hDisp_center_pp;
          eta2_pp = dataCenter.kernels.get(jj).hDisp_center_pp;
          ll      = (double)dataCenter.kernels.get(ii).l;
          if (1==dataCenter.kernels.get(ii).l) {
            aux[0] = cnst*(IK2*(xi1_pp-eta1_pp)*(xi1_pp-eta1_pp))
                         /(IK1*(xi2_pp-eta2_pp)*(xi2_pp-eta2_pp));
          } else {
            aux[0] = cnst* IK2*((xi1-eta1)*(xi1-eta1)+(ll*ll+ll-2.0)*eta1*eta1)
                         /(IK1*((xi2-eta2)*(xi2-eta2)+(ll*ll+ll-2.0)*eta2*eta2));
          }
        }

        aPlot.addDataSet(dataCenter.grid,aux);
        aPlot.setIOenabled(!dataCenter.isApplet);
        aPlot.setExtendXrange(true);
        aPlot.findRange();

        title = "("+dataCenter.kernels.get(ii).n+","
                   +dataCenter.kernels.get(ii).l+")/("
                   +dataCenter.kernels.get(jj).n+","
                   +dataCenter.kernels.get(jj).l+")";
        tb.add(title,aPlot);
      }
    }

    JFrame f = new JFrame("Ratio of \u222B K\u03A9/r\u00B2");
    f.setLayout(new BorderLayout());
    f.add("Center",tb);
    f.setSize(700,600);
    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    f.setVisible(true);    

  }

  ////////////////////////////////////////////////////////////////////////////////////
  //  This initialises the endian radio buttons
  ////////////////////////////////////////////////////////////////////////////////////
  private void initEndians() {
    NativeEndianLabel = new JLabel();
    if (ByteOrder.nativeOrder().equals(ByteOrder.BIG_ENDIAN)) {
      NativeEndianLabel.setText(" Native Endian = Big");
    } else {
      NativeEndianLabel.setText(" Native Endian = Little");
    }
    NativeEndian = new JRadioButton("Native");
    OppositeEndian = new JRadioButton("Opposite");
    EndianGroup = new ButtonGroup ();
    EndianGroup.add(NativeEndian);
    EndianGroup.add(OppositeEndian);
    isNative = true;
    NativeEndian.setSelected(true);
    NativeEndian.setEnabled(true);
    OppositeEndian.setEnabled(true);
    ActionListener l = new ActionListener() {
      public void actionPerformed(ActionEvent ev) {
        JRadioButton source = (JRadioButton)ev.getSource();
        isNative = (source == NativeEndian);
      }
    };
    NativeEndian.addActionListener(l);
    OppositeEndian.addActionListener(l);
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This initialises the nfmode radio buttons
  ////////////////////////////////////////////////////////////////////////////////////
  private void initNfRadioButtons(){
    int i;
    ButtonGroup nfGroup = new ButtonGroup();
    nfRadioButtons = new JRadioButton[4];
    for(i=1; i<=3; i++) {
      nfRadioButtons[i] = new JRadioButton(String.format("%d",i));
      nfGroup.add(nfRadioButtons[i]);
    }
    nfmode = 1;
    nfRadioButtons[nfmode].setSelected(true);
    ActionListener l = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        JRadioButton source = (JRadioButton) e.getSource();
        for(int i=1; i<=3; i++) {
          if (source == nfRadioButtons[i]) {
            nfmode = i;
            break;
          }
        }
      }
    };
    for(i=1; i<=3; i++) nfRadioButtons[i].addActionListener(l);
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This initialises the radio buttons which decide the type of model (i.e. model
  // format) to be used when reading a model.
  ////////////////////////////////////////////////////////////////////////////////////
  private void initEFunctionTypes() {
    AMDEtype = new JRadioButton("AMDE");
    FAMDEtype = new JRadioButton("FAMDE");
    FILOUtype = new JRadioButton("FILOU");
    EFunctionGroup = new ButtonGroup();
    EFunctionGroup.add(AMDEtype);
    EFunctionGroup.add(FAMDEtype);
    EFunctionGroup.add(FILOUtype);
    EFunctionType = "AMDE";
    AMDEtype.setSelected(true);
    ActionListener l = new ActionListener() {
      public void actionPerformed(ActionEvent ev) {
        EFunctionType = ((JRadioButton)ev.getSource()).getText();
        if (EFunctionType.equals("AMDE")) {
          NativeEndian.setEnabled(true);
          OppositeEndian.setEnabled(true);
        } else {
          NativeEndian.setEnabled(false);
          OppositeEndian.setEnabled(false);
        }
        if (EFunctionType.endsWith("AMDE")) {
          for(int i=1; i<= 3; i++) nfRadioButtons[i].setEnabled(true);
        } else {
          for(int i=1; i<= 3; i++) nfRadioButtons[i].setEnabled(false);
        }
      }
    };
    AMDEtype.addActionListener(l);
    FAMDEtype.addActionListener(l);
    FILOUtype.addActionListener(l);
  }

  ///////////////////////////////////////////////////////////
  // This initialises the panel which shows the different
  // options for reading eigenfunctions
  ///////////////////////////////////////////////////////////
  private void initEFunctionPanel() {
    EFunctionPanel = new JPanel();
    EFunctionPanel.setLayout(new GridBagLayout());
    GridBagConstraints c = new GridBagConstraints();
    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridx = 0;
    c.gridy = 0;
    c.gridwidth = 6;
    EFunctionPanel.add(EFunctionTypeLabel,c);
    c.gridwidth = 2;
    c.gridy++;
    c.gridx = 0;
    EFunctionPanel.add(AMDEtype,c);
    c.gridx = 2;
    EFunctionPanel.add(FAMDEtype,c);
    c.gridx = 4;
    EFunctionPanel.add(FILOUtype,c);
    c.gridwidth = 6;
    c.gridy++;
    c.gridx = 0;
    EFunctionPanel.add(new JLabel(" "),c);
    c.gridy++;
    EFunctionPanel.add(nfmodeLabel,c);
    c.gridwidth = 2;
    c.gridy++;
    for(int i=1; i<=3; i++) {
      c.gridx = 2*(i-1);
      EFunctionPanel.add(nfRadioButtons[i],c);
    }
    c.gridwidth = 6;
    c.gridy++;
    c.gridx = 0;
    EFunctionPanel.add(new JLabel(" "),c);
    c.gridy++;
    EFunctionPanel.add(Endianness,c);
    c.gridwidth = 3;
    c.gridy++;
    c.gridx = 0;
    EFunctionPanel.add(NativeEndian,c);
    c.gridx = 3;
    EFunctionPanel.add(OppositeEndian,c);
    c.gridwidth = 6;
    c.gridy++;
    c.gridx = 0;
    EFunctionPanel.add(NativeEndianLabel,c);
  }

  /////////////////////////////////////////////////////////////////////
  // initialise readEFunctionsButton:
  /////////////////////////////////////////////////////////////////////
  private void initReadEFunctionsButton() {
    readEFunctionsButton = Utilities.createButton("Load eigenfunctions");
    readEFunctionsButton.setEnabled(!dataCenter.isApplet);
    readEFunctionsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!dataCenter.modelLoaded) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "No model loaded, please load the model first.",
            "No model",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }
        
        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);
        fc.setAccessory(EFunctionPanel);

        // show the filechooser
        int result = fc.showOpenDialog(readEFunctionsButton);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            dataCenter.currentDirectory = aFile.getParentFile();
            if (!readEigenFunctions(aFile.getPath())) return;

            pulsations.findOtherVariables();
            pulsations.buildRotaKernels();
            pulsations.buildStructKernels();
            pulsations.applyVariational();
            kernelTable.updateUI();
          } else {
            JOptionPane.showMessageDialog(
              readEFunctionsButton,
              "File \""+aFile.getName()+"\" doesn't exist.",
              "File not found",
              JOptionPane.ERROR_MESSAGE
            );
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////
  /** This method calls the appropriate routine for reading eigenfunctions
   *  which can be in AMDE, FAMDE our FILOU format. It return "false" when
   *  an error has occured. */
  /////////////////////////////////////////////////////////////////////////
  public boolean readEigenFunctions(String filename) {
    BufferedReader eigInput;
    FortranBinaryInputStream binaryInput;

    try {

      // treat binary case first since the model
      // file cannot be opened in the same way.
      if (EFunctionType.equals("AMDE")) {
        FileInputStream in = new FileInputStream(new File(filename));
        binaryInput = new FortranBinaryInputStream(in,true,isNative);
        if (nfmode == 1) {
          return readAMDEeigenFunctionsNfmode1(binaryInput);
        } else if (nfmode == 2) {
          return readAMDEeigenFunctionsNfmode2(binaryInput);
        } else if (nfmode == 3) {
          return readAMDEeigenFunctionsNfmode3(binaryInput);
        }
      }

      eigInput = Utilities.OpenFile(filename);

      if (EFunctionType.equals("FAMDE")) {
        if (nfmode == 1) {
          return readFAMDEeigenFunctionsNfmode1(eigInput);
        } else if (nfmode == 2) {
          return readFAMDEeigenFunctionsNfmode2(eigInput);
        } else if (nfmode == 3) {
          return readFAMDEeigenFunctionsNfmode3(eigInput);
        }
      }

      if (EFunctionType.equals("FILOU"))
        return readFILOUeigenFunctions(eigInput);

      JOptionPane.showMessageDialog(
        readEFunctionsButton,  // the parent component
        "Unknown or unimplemented format: "+EFunctionType+".\n"+
        "Please contact authors of InversionKit.",
        "Format error",
        JOptionPane.ERROR_MESSAGE
      );
      return false;

    } catch(Exception e) {
      System.out.println(e);
      JOptionPane.showMessageDialog(
        readEFunctionsButton,  // the parent component
        "Unable to read from:\n  "+filename+"\n",
        "Error opening file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
  }

  /////////////////////////////////////////////////////////////////////////
  /** This method reads a file which contains the eigenfunctions in FILOU
  *   format. */
  /////////////////////////////////////////////////////////////////////////
  private boolean readFILOUeigenFunctions(BufferedReader eigInput) {
    int i, nn;
  
    // set up some array lists and variables into which to
    // read different quantities:
    ArrayList<Integer> lvalue  = new ArrayList<Integer>();
    ArrayList<Integer> nvalue  = new ArrayList<Integer>();
    ArrayList<Double> frq      = new ArrayList<Double>();
    ArrayList<Double> frq2     = new ArrayList<Double>();
    ArrayList<double[][]> func = new ArrayList<double[][]>();
    int ngrid=0, ngrid_temp, nfunc, k;
    boolean firstLoop = true;
    double[] Egrid = null;
    double[][] tmp;
    String line = " ";
    StringTokenizer t0;

    try {
      while (true) {
        // Find the degree of the mode:  
        while (-1 == line.indexOf("DEGRE DU MODE L :")) {
              line = eigInput.readLine();
              // check for end of file:
              if (line == null) break;
        }
        if (line == null) break;
        k = line.indexOf(":");
        lvalue.add(Integer.parseInt(line.substring(k+1).trim()));
        
        // Read the number of grid points (and check that this is the
        // same for all eigenfunctions):
        while (-1 == line.indexOf("Nombre de points du reseau du modele :"))
              line = eigInput.readLine();
        k = line.indexOf(":");
        ngrid_temp = Integer.parseInt(line.substring(k+1).trim());

        if ((ngrid_temp<0)||(ngrid_temp>maxNgrid)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "The number of grid points for the\n"+
            "eigenfunctions seems to be erroneous\n"+
            "or the format is incorrect.",
            "Error reading file",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }

        if ((!firstLoop) && (ngrid != ngrid_temp)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "There are different grids for the different eigenfunctions!",
            "Error reading file",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }
        ngrid = ngrid_temp;
        
        // Find non-dimensional squared frequency:
        while (-1 == line.indexOf("requence carree normalisee ="))
              line = eigInput.readLine();
        k = line.indexOf("=");
        frq2.add(Utilities.parseDouble(line.substring(k+1)));
        
        // Find the radial order:
        while (-1 == line.indexOf("noeuds ="))
              line = eigInput.readLine();
        k = line.indexOf("=");
        nvalue.add(Integer.parseInt(line.substring(k+1).trim()));
        
        // Find frequency:
        while (-1 == line.indexOf("requence en micro Hz ="))
              line = eigInput.readLine();
        k = line.indexOf("=");
        frq.add(Utilities.parseDouble(line.substring(k+1)));

        for(i=0; i<3; i++) line = eigInput.readLine();

        // Read main part (I'm assuming the grid is the same
        // for all of the eigenfunctions):
        tmp = new double[4][ngrid];
        if (firstLoop) {
          Egrid = new double[ngrid];
          for(i=0; i<ngrid; i++) {
            line = eigInput.readLine();
            Egrid[i]  = Utilities.parseDouble(line.substring(0,15));
            tmp[0][i] = Utilities.parseDouble(line.substring(15,30));
            tmp[1][i] = Utilities.parseDouble(line.substring(30,45));
            tmp[2][i] = Utilities.parseDouble(line.substring(45,60));
            tmp[3][i] = Utilities.parseDouble(line.substring(60,75));
          }
        } else {
          for(i=0; i<ngrid; i++) {
            line = eigInput.readLine();
            tmp[0][i] = Utilities.parseDouble(line.substring(15,30));
            tmp[1][i] = Utilities.parseDouble(line.substring(30,45));
            tmp[2][i] = Utilities.parseDouble(line.substring(45,60));
            tmp[3][i] = Utilities.parseDouble(line.substring(60,75));
          }
        }
        func.add(tmp);
        firstLoop = false;
      }
      eigInput.close();

    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Format not recognised.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
    
    nfunc = func.size();
    
    // check to see that there is some data:
    if (nfunc == 0) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "No eigenfunctions in file! No data loaded.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
    
    // give warning message about extrapolating the eigenfunctions,
    // if the model's grid is larger than that of the eigenfunctions
    if ((dataCenter.grid[0]<Egrid[0])||
        (dataCenter.grid[dataCenter.grid.length-1]>Egrid[ngrid-1])) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Warning: the eigenfunction are defined on a smaller grid\n"+
        "than the model and will therefore be extrapolated.",
        "Grid warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    // interpolate the eigenfunctions onto the same grid as the model
    // (it is much more efficient to interpolate the entire list in
    // one go rather than to do the different modes individually):
    ArrayList<double[][]> yy = Numerics.interpolateList_r2(Egrid,dataCenter.grid,func,false);
    ngrid = dataCenter.grid.length; // don't forget to change ngrid ...

    // find integration weights:
    double a;

    // variable into which to store the mode information:
    Kernel aKernel;
    
    for(nn = 0; nn < nfunc; nn++) {
      // add kernel to the list of theoretical modes:
      aKernel = new Kernel(nvalue.get(nn).intValue(),lvalue.get(nn).intValue());
      dataCenter.kernels.add(aKernel);

      aKernel.FreqTheo = frq.get(nn).doubleValue();
      aKernel.FreqDimless = Math.sqrt(frq2.get(nn).doubleValue());
      aKernel.equations = Pulsations.FromFile;

      // set up some variables in the DataCenter:
      aKernel.vDisp = new double[ngrid];
      aKernel.hDisp = new double[ngrid];
      aKernel.phi   = new double[ngrid];
      aKernel.phip  = new double[ngrid];
      aKernel.P     = new double[ngrid];
      aKernel.rho   = new double[ngrid];
      aKernel.dP_P  = new double[ngrid];

      // I'm not too sure what happens in the radial case, nor
      // what happens when the variables are divergent in the
      // center (based on the definition given in Suarez and Goupil, 2008)
      a = 1.0/(dataCenter.grid[dataCenter.indSurface]*yy.get(nn)[0][dataCenter.indSurface]);
      for(i=0; i < ngrid; i++) {
        aKernel.vDisp[i] = a*dataCenter.grid[i]*yy.get(nn)[0][i];
        aKernel.hDisp[i] = a*dataCenter.gravity[i]*yy.get(nn)[1][i]/frq2.get(nn).doubleValue();
        aKernel.phi[i]   = a*dataCenter.gravity[i]*dataCenter.grid[i]*yy.get(nn)[2][i];
        aKernel.phip[i]  = a*dataCenter.gravity[i]*yy.get(nn)[3][i];
      }
      pulsations.findVariablesNonRadial(aKernel);

      // tell findOtherVariables() to find remaining variables:
      aKernel.IK = Double.NaN;
    }
    return true;
  }

  /////////////////////////////////////////////////////////////////////////
  /** This method reads a file which contains the eigenfunctions.  The
  *   format of the file is an ASCII version of an output file from JCD's
  *   code plus an additional header as described in the 1996 inversion
  *   workbench. */
  /////////////////////////////////////////////////////////////////////////
  private boolean readFAMDEeigenFunctionsNfmode1(BufferedReader eigInput) {
    double tol = 1.e-06;
    int i, j, nn, ngrid = 0, nfunc;
    String line;
    int nparams = 50, nf = 7;
    StringTokenizer t0;

    // set up some array lists into which to read different quantities:
    ArrayList<double[]> globals = new ArrayList<double[]>();
    ArrayList<double[][]> func  = new ArrayList<double[][]>();

    try {
      // skip the header (look for termination sequence, which is made
      // up of 3 (CTRL+L, CTRL+J) in a row):
      int nterm = 0, oneChar;
      while (nterm < 6) {
        oneChar = eigInput.read();
        if (oneChar < 0) throw new Exception("File input error");
        if (oneChar == termSeq[nterm]) {
          nterm++;
        } else {
          if (oneChar == termSeq[0]) {
            nterm = 1;
          } else {
            nterm = 0;
          }
        }
      }

      // read all available eigenfunctions
      // now read all available eigenfunctions
      line = eigInput.readLine();  // read a line of text
      nfunc = 0;
      while(true) {
        // test for the end of the buffer:
        if (line == null) break;

        // NOW read global params....
        globals.add(new double[nparams]);
        int count = 0;
        while (count < nparams) {
          t0 = new StringTokenizer(line," \t,");
          while ((t0.hasMoreTokens())&&(count < nparams)) {
            globals.get(nfunc)[count] = Utilities.parseDouble(t0.nextToken());
            count++;
          }
          // IMPORTANT: if line is null, then applying a string tokenizer will
          // provoke and error, so will be detected.  There is therefore no need
          // to do an extra test to see if line == null.
          line = eigInput.readLine();
        }

        // check the stellar mass and radius associated with the eigenfunctions
        // to see if they match that of the star:
        if ((Math.abs(1.0-dataCenter.StellarMass/globals.get(nfunc)[1]) > tol)||
            (Math.abs(1.0-dataCenter.StellarRadius/globals.get(nfunc)[2]) > tol)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "Model and eigenfunctions do not\n"+
            "agree or the format is incorrect.\n"+
            "Please chose another set.",
            "Eigenfunction-model error",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }

        // get number of grid points
        t0 = new StringTokenizer(line," \t,");
        ngrid = Integer.parseInt(t0.nextToken());
        if ((ngrid<0)||(ngrid>maxNgrid)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "The number of grid points for the\n"+
            "eigenfunctions seems to be erroneous\n"+
            "or the format is incorrect.",
            "Error reading file",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }
        line = eigInput.readLine();

        // Read grid and eigenfunctions:
        double[][] efunc = new double[nf][ngrid];     
        i = 0; j = 0;
        while (i < ngrid){
          t0 = new StringTokenizer(line, " \t,");
          while ((t0.hasMoreTokens())&&(i<ngrid)) {
            efunc[j][i] = Utilities.parseDouble(t0.nextToken());
            j++;
            if (j == nf) {j = 0; i++;}
          }
          line = eigInput.readLine();
        }
        func.add(efunc);
        nfunc++;
      }
      eigInput.close(); 
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Format not recognised.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }

    // check to see that there is some data:
    if (nfunc == 0) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "No eigenfunctions in file! No data loaded.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
    
    // set up the grid in dataCenter:
    double[] Egrid = new double[ngrid];
    for(i=0;i<ngrid;i++) Egrid[i] = func.get(0)[0][i];

    // give warning message about extrapolating the eigenfunctions,
    // if the model's grid is larger than that of the eigenfunctions
    if ((dataCenter.grid[0]<Egrid[0])||
        (dataCenter.grid[dataCenter.grid.length-1]>Egrid[ngrid-1])) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Warning: the eigenfunction are defined on a smaller grid\n"+
        "than the model and will therefore be extrapolated.",
        "Grid warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    // interpolate the eigenfunctions onto the same grid as the model
    // (it is much more efficient to interpolate the entire list in
    // one go rather than to do the different modes individually):
    ArrayList<double[][]> yy = Numerics.interpolateList_r2(Egrid,dataCenter.grid,func,false);
    ngrid = dataCenter.grid.length; // don't forget to change ngrid ...

    // find integration matrix:
    double[] integrationMatrix = Numerics.findIntegrationMatrix(dataCenter.grid);
    double a, L2, Freq2;

    // variable into which to store the mode information:
    Kernel aKernel;

    for(nn = 0; nn < nfunc; nn++) {
      // add kernel to the list of theoretical modes:
      aKernel = new Kernel(Math.round((float)globals.get(nn)[18]),Math.round((float)globals.get(nn)[17]));
      dataCenter.kernels.add(aKernel);

      aKernel.FreqTheo = 1e3*globals.get(nn)[26];
      aKernel.FreqDimless = aKernel.FreqTheo/dataCenter.FreqRef;
      aKernel.equations = Pulsations.FromFile;

      // set up some variables in the DataCenter:
      aKernel.vDisp = new double[ngrid];
      aKernel.hDisp = new double[ngrid];
      aKernel.phip  = new double[ngrid];
      aKernel.P     = new double[ngrid];
      aKernel.rho   = new double[ngrid];
      aKernel.dP_P  = new double[ngrid];

      // WARNING: yy.get(nn)[0] is simply the original grid.
      //          Hence, true eigenfunctions start at yy.get(nn)[1]
      a = 1.0/yy.get(nn)[1][dataCenter.indSurface];
      if (aKernel.l == 0) {
        Freq2 = aKernel.FreqDimless*aKernel.FreqDimless;
        for(i=0; i<ngrid; i++) {
          aKernel.vDisp[i] = a*yy.get(nn)[1][i];
          aKernel.P[i]     = a*Freq2*dataCenter.rho[i]*yy.get(nn)[2][i];
        }
        pulsations.findVariablesRadial(aKernel);
        aKernel.phi = Numerics.applyIntegrationMatrixReverse(integrationMatrix, aKernel.phip);
      } else {
        L2 = (double)aKernel.l*((double)aKernel.l+1.0);
        aKernel.phi   = new double[ngrid];
        for(i=0; i<ngrid; i++) {
          aKernel.vDisp[i] = a*yy.get(nn)[1][i];
          aKernel.hDisp[i] = a*yy.get(nn)[2][i]/L2;
          aKernel.phi[i]   =-a*dataCenter.gravity[i]*yy.get(nn)[3][i];
          aKernel.phip[i]  =-a*(yy.get(nn)[4][i]*dataCenter.gravity[i]/dataCenter.grid[i]
                           +(dataCenter.gravity[i]/dataCenter.grid[i]
                           + dataCenter.dgravity[i])*yy.get(nn)[3][i]);
        }
        aKernel.phip[0] =-a*(yy.get(nn)[4][0]*dataCenter.dgravity[0]
                        + 2.0*dataCenter.dgravity[0]*yy.get(nn)[3][0]);
        pulsations.findVariablesNonRadial(aKernel);
      }

      // tell findOtherVariables() to find remaining variables:
      aKernel.IK = Double.NaN;
    }

    return true;
  }

  /////////////////////////////////////////////////////////////////////////
  /** This method reads a file which contains the eigenfunctions.  The
  *   format of the file is an ASCII version of an output file from JCD's
  *   code plus an additional header as described in the 1996 inversion
  *   workbench. */
  /////////////////////////////////////////////////////////////////////////
  private boolean readFAMDEeigenFunctionsNfmode2(BufferedReader eigInput) {
    double tol = 1.e-06;
    int i, j, nn, ngrid, nfunc;
    String line;
    int nparams = 50;

    // set up some array lists into which to read different quantities:
    ArrayList<double[]> globals = new ArrayList<double[]>();
    ArrayList<double[][]> func  = new ArrayList<double[][]>();
    double[] Egrid;

    try {
      // skip the header (look for termination sequence, which is made
      // up of 3 (CTRL+L, CTRL+J) in a row):
      int nterm = 0, oneChar;
      while (nterm < 6) {
        oneChar = eigInput.read();
        if (oneChar < 0) throw new Exception("File input error");
        if (oneChar == termSeq[nterm]) {
          nterm++;
        } else {
          if (oneChar == termSeq[0]) {
            nterm = 1;
          } else {
            nterm = 0;
          }
        }
      }

      // get number of grid points
      line = eigInput.readLine();
      StringTokenizer t0 = new StringTokenizer(line," \t,");
      ngrid = Integer.parseInt(t0.nextToken());
      if ((ngrid<0)||(ngrid>maxNgrid)) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "The number of grid points for the\n"+
          "eigenfunctions seems to be erroneous\n"+
          "or the format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return false;
      }

      // set up the grid in dataCenter:
      Egrid = new double[ngrid];
      int count = 0;
      while (count < ngrid) {
        line = eigInput.readLine();
        t0 = new StringTokenizer(line," \t,");
        while (t0.hasMoreTokens()) {
          Egrid[count] = Utilities.parseDouble(t0.nextToken());
          count++;
        }
      }

      // give warning message about extrapolating the eigenfunctions,
      // if the model's grid is larger than that of the eigenfunctions
      if ((dataCenter.grid[0]<Egrid[0])||
          (dataCenter.grid[dataCenter.grid.length-1]>Egrid[ngrid-1])) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "Warning: the eigenfunction are defined on a smaller grid\n"+
          "than the model and will therefore be extrapolated.",
          "Grid warning",
          JOptionPane.WARNING_MESSAGE
        );
      }

      // now read all available eigenfunctions
      nfunc = 0;
      line = eigInput.readLine();  // read a line of text
      while(true) {

        // test for the end of the buffer:
        if (line == null) break;

        // NOW read global params....
        globals.add(new double[nparams]);
        count = 0;
        while (count < nparams) {
          t0 = new StringTokenizer(line, " \t,");
          while (t0.hasMoreTokens()) {
            globals.get(nfunc)[count] = Utilities.parseDouble(t0.nextToken());
            count++;
          }
          // IMPORTANT: if line is null, then applying a string tokenizer will
          // provoke and error, so will be detected.  There is therefore no need
          // to do an extra test to see if line == null.
          line = eigInput.readLine();
        }
        
        // check the stellar mass and radius associated with the eigenfunctions
        // to see if they match that of the star:
        if ((Math.abs(1.0-dataCenter.StellarMass/globals.get(nfunc)[1]) > tol)||
            (Math.abs(1.0-dataCenter.StellarRadius/globals.get(nfunc)[2]) > tol)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "Model and eigenfunctions do not\n"+
            "agree or the format is incorrect.\n"+
            "Please chose another set.",
            "Eigenfunction-model error",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }

        // NOW read eigenfunctions....
        double[][]efunc = new double[2][ngrid];
        count = 0;
        while (count < 2 * ngrid) {
          t0 = new StringTokenizer(line, " \t,");
          while (t0.hasMoreTokens()) {
            efunc[count%2][count/2] = Utilities.parseDouble(t0.nextToken());
            count++;
          }
          line = eigInput.readLine();
        }
        func.add(efunc);
        nfunc++;
      }
      eigInput.close(); 
    } catch (Exception e) {
      System.out.println(e);
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Format not recognised.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }

    // check to see that there is some data:
    if (nfunc == 0) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "No eigenfunctions in file! No data loaded.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
    
    // interpolate the eigenfunctions onto the same grid as the model
    // (it is much more efficient to interpolate the entire list in
    // one go rather than to do the different modes individually):
    ArrayList<double[][]> yy = Numerics.interpolateList_r2(Egrid,dataCenter.grid,func,false);
    ngrid = dataCenter.grid.length; // don't forget to change ngrid ...

    // initialise various variables:
    int order = 1;
    double[] integrationMatrix = Numerics.findIntegrationMatrix(dataCenter.grid);
    double[][] dermat = Numerics.initDeriveFDBand(dataCenter.grid,order,1);
    double a, b, L2, L, Freq2;

    // variable into which to store the mode information:
    Kernel aKernel;

    for(nn = 0; nn < nfunc; nn++) {
      // add kernel to the list of theoretical modes:
      aKernel = new Kernel(Math.round((float)globals.get(nn)[18]),Math.round((float)globals.get(nn)[17]));
      dataCenter.kernels.add(aKernel);

      aKernel.FreqTheo = 1e3*globals.get(nn)[26];
      aKernel.FreqDimless = aKernel.FreqTheo/dataCenter.FreqRef;
      aKernel.equations = Pulsations.FromFile;

      // set up some variables in the DataCenter:
      aKernel.vDisp = new double[ngrid];
      aKernel.hDisp = new double[ngrid];
      aKernel.phip  = new double[ngrid];
      aKernel.P     = new double[ngrid];
      aKernel.rho   = new double[ngrid];
      aKernel.dP_P  = new double[ngrid];

      a = 1.0/yy.get(nn)[0][dataCenter.indSurface];
      if (aKernel.l == 0) {
        Freq2 = aKernel.FreqDimless*aKernel.FreqDimless;
        for(i=0; i<ngrid; i++) {
          aKernel.vDisp[i] = a*yy.get(nn)[0][i];
          aKernel.P[i]     = a*Freq2*dataCenter.rho[i]*yy.get(nn)[1][i];
        }
        pulsations.findVariablesRadial(aKernel);
        aKernel.phi = Numerics.applyIntegrationMatrixReverse(integrationMatrix, aKernel.phip);
      } else {
        L2 = (double)aKernel.l*((double)aKernel.l+1.0);
        aKernel.phi   = null;
        for(i=0; i<ngrid; i++) {
          aKernel.vDisp[i] = a*yy.get(nn)[0][i];
          aKernel.hDisp[i] = a*yy.get(nn)[1][i]/L2;
        }
      }

      // tell findOtherVariables() to find remaining variables:
      aKernel.IK = Double.NaN;
    }
    pulsations.findPhi();  // do this on the entire set, so as to save extra calculations
    return true;
  }

  /////////////////////////////////////////////////////////////////////////
  /** This method reads a file which contains the eigenfunctions.  The
  *   format of the file is an ASCII version of an output file from JCD's
  *   code plus an additional header as described in the 1996 inversion
  *   workbench. */
  /////////////////////////////////////////////////////////////////////////
  private boolean readFAMDEeigenFunctionsNfmode3(BufferedReader eigInput) {
    double tol = 1.e-06;
    int i, j, nn, ngrid, nfunc;
    String line;
    int nparams = 50;

    // set up some array lists into which to read different quantities:
    ArrayList<double[]> globals = new ArrayList<double[]>();
    ArrayList<double[][]> func  = new ArrayList<double[][]>();
    double[] Egrid;

    try {
      // skip the header (look for termination sequence, which is made
      // up of 3 (CTRL+L, CTRL+J) in a row):
      int nterm = 0, oneChar;
      while (nterm < 6) {
        oneChar = eigInput.read();
        if (oneChar < 0) throw new Exception("File input error");
        if (oneChar == termSeq[nterm]) {
          nterm++;
        } else {
          if (oneChar == termSeq[0]) {
            nterm = 1;
          } else {
            nterm = 0;
          }
        }
      }

      // get number of grid points
      line = eigInput.readLine();
      StringTokenizer t0 = new StringTokenizer(line," \t,");
      ngrid = Integer.parseInt(t0.nextToken());
      if ((ngrid<0)||(ngrid>maxNgrid)) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "The number of grid points for the\n"+
          "eigenfunctions seems to be erroneous\n"+
          "or the format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return false;
      }

      // set up the grid in dataCenter:
      Egrid = new double[ngrid];
      int count = 0;
      while (count < ngrid) {
        line = eigInput.readLine();
        t0 = new StringTokenizer(line," \t,");
        while (t0.hasMoreTokens()) {
          Egrid[count] = Utilities.parseDouble(t0.nextToken());
          count++;
        }
      }

      // give warning message about extrapolating the eigenfunctions,
      // if the model's grid is larger than that of the eigenfunctions
      if ((dataCenter.grid[0]<Egrid[0])||
          (dataCenter.grid[dataCenter.grid.length-1]>Egrid[ngrid-1])) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "Warning: the eigenfunction are defined on a smaller grid\n"+
          "than the model and will therefore be extrapolated.",
          "Grid warning",
          JOptionPane.WARNING_MESSAGE
        );
      }

      // now read all available eigenfunctions
      nfunc = 0;
      line = eigInput.readLine();  // read a line of text
      while(true) {

        // test for the end of the buffer:
        if (line == null) break;

        // NOW read global params....
        globals.add(new double[nparams]);
        count = 0;
        while (count < nparams) {
          t0 = new StringTokenizer(line, " \t,");
          while (t0.hasMoreTokens()) {
            globals.get(nfunc)[count] = Utilities.parseDouble(t0.nextToken());
            count++;
          }
          // IMPORTANT: if line is null, then applying a string tokenizer will
          // provoke and error, so will be detected.  There is therefore no need
          // to do an extra test to see if line == null.
          line = eigInput.readLine();
        }
        
        // check the stellar mass and radius associated with the eigenfunctions
        // to see if they match that of the star:
        if ((Math.abs(1.0-dataCenter.StellarMass/globals.get(nfunc)[1]) > tol)||
            (Math.abs(1.0-dataCenter.StellarRadius/globals.get(nfunc)[2]) > tol)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "Model and eigenfunctions do not\n"+
            "agree or the format is incorrect.\n"+
            "Please chose another set.",
            "Eigenfunction-model error",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }

        // NOW read eigenfunctions....
        double[][]efunc = new double[2][ngrid];
        count = 0;
        while (count < 2 * ngrid) {
          t0 = new StringTokenizer(line, " \t,");
          while (t0.hasMoreTokens()) {
            efunc[count%2][count/2] = Utilities.parseDouble(t0.nextToken());
            count++;
          }
          line = eigInput.readLine();
        }
        func.add(efunc);
        nfunc++;
      }
      eigInput.close(); 
    } catch (Exception e) {
      System.out.println(e);
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Format not recognised.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }

    // check to see that there is some data:
    if (nfunc == 0) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "No eigenfunctions in file! No data loaded.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
    
    // interpolate the eigenfunctions onto the same grid as the model
    // (it is much more efficient to interpolate the entire list in
    // one go rather than to do the different modes individually):
    ArrayList<double[][]> yy = Numerics.interpolateList_r2(Egrid,dataCenter.grid,func,false);
    ngrid = dataCenter.grid.length; // don't forget to change ngrid ...

    // initialise various variables:
    int order = 1;
    double[] integrationMatrix = Numerics.findIntegrationMatrix(dataCenter.grid);
    double[][] dermat = Numerics.initDeriveFDBand(dataCenter.grid,order,1);
    double a, b, L2, L;

    // variable into which to store the mode information:
    Kernel aKernel;

    for(nn = 0; nn < nfunc; nn++) {
      // add kernel to the list of theoretical modes:
      aKernel = new Kernel(Math.round((float)globals.get(nn)[18]),Math.round((float)globals.get(nn)[17]));
      dataCenter.kernels.add(aKernel);

      aKernel.FreqTheo = 1e3*globals.get(nn)[26];
      aKernel.FreqDimless = aKernel.FreqTheo/dataCenter.FreqRef;
      aKernel.equations = Pulsations.FromFile;

      // set up some variables in the DataCenter:
      aKernel.vDisp = new double[ngrid];
      aKernel.hDisp = new double[ngrid];
      aKernel.phip  = new double[ngrid];
      aKernel.P     = new double[ngrid];
      aKernel.rho   = new double[ngrid];
      aKernel.dP_P  = new double[ngrid];

      a = Math.sqrt(4.0*Math.PI*Math.pow(dataCenter.grid[dataCenter.indSurface],3.0)*dataCenter.rho[dataCenter.indSurface])
        / yy.get(nn)[0][dataCenter.indSurface];
      if (aKernel.l == 0) {
        for(i=1; i<ngrid; i++) {
          b = Math.sqrt(4.0*Math.PI*Math.pow(dataCenter.grid[i],3.0)*dataCenter.rho[i]);
          aKernel.vDisp[i] = a*yy.get(nn)[0][i]/b;
        }
        aKernel.vDisp[0] = 0.0;  // this is EXTREMELY important
        Dgbmv.dgbmv("N",ngrid,ngrid,order,order,1.0,dermat[1],0,1+2*order,aKernel.vDisp,0,1,0.0,aKernel.P,0,1);
        for(i=1; i<ngrid; i++){
          aKernel.P[i]     = -dataCenter.Gamma1[i]*dataCenter.pressure[i]
                           * (aKernel.P[i] + 2.0*aKernel.vDisp[i]/dataCenter.grid[i])
                           + dataCenter.rho[i]*dataCenter.gravity[i]*aKernel.vDisp[i];
        }
        aKernel.P[0] = -3.0*dataCenter.Gamma1[0]*dataCenter.pressure[0]*aKernel.P[0];
        pulsations.findVariablesRadial(aKernel);
        aKernel.phi = Numerics.applyIntegrationMatrixReverse(integrationMatrix, aKernel.phip);
      } else {
        L2 = (double)aKernel.l*((double)aKernel.l+1.0);
        L  = Math.sqrt(L2);
        aKernel.phi = null;  // this is to tell findPhi() to find phi.
        for(i=1; i<ngrid; i++) {
          b = Math.sqrt(4.0*Math.PI*Math.pow(dataCenter.grid[i],3.0)*dataCenter.rho[i]);
          aKernel.vDisp[i] = a*yy.get(nn)[0][i]/b;
          aKernel.hDisp[i] = a*yy.get(nn)[1][i]/(b*L);
        }
        if (aKernel.l == 1) {
          // extrapolate from above, assuming f(x) = ax^2 + b
          double eta = Math.pow(dataCenter.grid[1],2.0)/(Math.pow(dataCenter.grid[1],2.0)-Math.pow(dataCenter.grid[2],2.0));
          aKernel.vDisp[0] = eta*aKernel.vDisp[2] + (1.0-eta)*aKernel.vDisp[1];
          aKernel.hDisp[0] = eta*aKernel.hDisp[2] + (1.0-eta)*aKernel.hDisp[1];
          // furthermore, vDisp = hDisp in the center, when ell == 1. So we average the two values:
          aKernel.vDisp[0] = (aKernel.hDisp[0]+aKernel.vDisp[0])/2.0;
          aKernel.hDisp[0] =  aKernel.vDisp[0];
        } else {
          aKernel.vDisp[0] = 0.0;
          aKernel.hDisp[0] = 0.0;
        }
      }

      // tell findOtherVariables() to find remaining variables:
      aKernel.IK = Double.NaN;
    }
    pulsations.findPhi();  // do this on the entire set, so as to save extra calculations
    return true;
  }

  /////////////////////////////////////////////////////////////////////////
  /** This method reads a file which contains the eigenfunctions.  The
  *   format of the file is the binary output from JCD's ADIPLS code. */
  /////////////////////////////////////////////////////////////////////////
  private boolean readAMDEeigenFunctionsNfmode1(FortranBinaryInputStream eigInput) {
    double tol = 1.e-06;
    int i, j, nn, ngrid = 0, nfunc;
    int nparams = 50, nf = 6;

    // set up some array lists into which to read different quantities:
    double[] Egrid = null;
    ArrayList<double[]> globals = new ArrayList<double[]>();
    ArrayList<double[][]> func  = new ArrayList<double[][]>();

    try {
      // read all available eigenfunctions
      nfunc = 0;
      while(true) {
        // test for the end of the file:
        eigInput.readRecord();
        if (eigInput.isFinFichier()) break;

        // NOW read global params....
        globals.add(new double[nparams]);
        for (i=0; i<nparams; i++) 
          globals.get(nfunc)[i] = eigInput.readDoublePrecision();

        // check the stellar mass and radius associated with the eigenfunctions
        // to see if they match that of the star:
        if ((Math.abs(1.0-dataCenter.StellarMass/globals.get(nfunc)[1]) > tol)||
            (Math.abs(1.0-dataCenter.StellarRadius/globals.get(nfunc)[2]) > tol)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "Model and eigenfunctions do not\n"+
            "agree or the format is incorrect.\n"+
            "Please chose another set.",
            "Eigenfunction-model error",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }

        // get number of grid points
        ngrid = eigInput.readInteger();
        if ((ngrid<0)||(ngrid>maxNgrid)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "The number of grid points for the\n"+
            "eigenfunctions seems to be erroneous\n"+
            "or the format is incorrect.",
            "Error reading file",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }

        // set up the grid in dataCenter:
        if (nfunc==0) Egrid = new double[ngrid];

        // Read grid and eigenfunctions:
        double[][] efunc = new double[nf][ngrid];
        for(i=0;i<ngrid;i++) {
          Egrid[i] = eigInput.readDoublePrecision();
          for(j=0;j<nf;j++) {
            efunc[j][i] = eigInput.readDoublePrecision();
          }
        }

        func.add(efunc);
        nfunc++;
      }
      eigInput.close(); 
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Format not recognised.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }

    // give warning message about extrapolating the eigenfunctions,
    // if the model's grid is larger than that of the eigenfunctions
    if ((dataCenter.grid[0]<Egrid[0])||
        (dataCenter.grid[dataCenter.grid.length-1]>Egrid[ngrid-1])) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Warning: the eigenfunction are defined on a smaller grid\n"+
        "than the model and will therefore be extrapolated.",
        "Grid warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    // check to see that there is some data:
    if (nfunc == 0) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "No eigenfunctions in file! No data loaded.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
    
    // interpolate the eigenfunctions onto the same grid as the model
    // (it is much more efficient to interpolate the entire list in
    // one go rather than to do the different modes individually):
    ArrayList<double[][]> yy = Numerics.interpolateList_r2(Egrid,dataCenter.grid,func,false);
    ngrid = dataCenter.grid.length; // don't forget to change ngrid ...

    // initialise different variables
    double[] integrationMatrix = Numerics.findIntegrationMatrix(dataCenter.grid);
    double a, L2;

    // variable into which to store the mode information:
    Kernel aKernel;

    for(nn = 0; nn < nfunc; nn++) {
      // add kernel to the list of theoretical modes:
      aKernel = new Kernel(Math.round((float)globals.get(nn)[18]),Math.round((float)globals.get(nn)[17]));
      dataCenter.kernels.add(aKernel);

      aKernel.FreqTheo = 1e3*globals.get(nn)[26];
      aKernel.FreqDimless = aKernel.FreqTheo/dataCenter.FreqRef;
      aKernel.equations = Pulsations.FromFile;

      // set up some variables in the DataCenter:
      aKernel.vDisp = new double[ngrid];
      aKernel.hDisp = new double[ngrid];
      aKernel.phip  = new double[ngrid];
      aKernel.P     = new double[ngrid];
      aKernel.rho   = new double[ngrid];
      aKernel.dP_P  = new double[ngrid];

      a = 1.0/yy.get(nn)[0][dataCenter.indSurface];
      if (aKernel.l == 0) {
        double Freq2 = aKernel.FreqDimless*aKernel.FreqDimless;
        for(i=0; i<ngrid; i++) {
          aKernel.vDisp[i] = a*yy.get(nn)[0][i];
          aKernel.P[i]     = a*Freq2*dataCenter.rho[i]*yy.get(nn)[1][i];
        }
        pulsations.findVariablesRadial(aKernel);
        aKernel.phi = Numerics.applyIntegrationMatrixReverse(integrationMatrix, aKernel.phip);
      } else {
        L2 = (double)aKernel.l*((double)aKernel.l+1.0);
        aKernel.phi   = new double[ngrid];
        for(i=0; i<ngrid; i++) {
          aKernel.vDisp[i] = a*yy.get(nn)[0][i];
          aKernel.hDisp[i] = a*yy.get(nn)[1][i]/L2;
          aKernel.phi[i]   =-a*dataCenter.gravity[i]*yy.get(nn)[2][i];
          aKernel.phip[i]  =-a*(yy.get(nn)[3][i]*dataCenter.gravity[i]/dataCenter.grid[i]
                           +(dataCenter.gravity[i]/dataCenter.grid[i]
                           + dataCenter.dgravity[i])*yy.get(nn)[2][i]);
        }
        aKernel.phip[0] =-a*(yy.get(nn)[3][0]*dataCenter.dgravity[0]
                        + 2.0*dataCenter.dgravity[0]*yy.get(nn)[2][0]);
        pulsations.findVariablesNonRadial(aKernel);
      }

      // tell findOtherVariables() to find remaining variables:
      aKernel.IK = Double.NaN;
    }
    return true;
  }

  /////////////////////////////////////////////////////////////////////////
  /** This method reads a file which contains the eigenfunctions.  The
  *   format of the file is the binary output from JCD's ADIPLS code. */
  /////////////////////////////////////////////////////////////////////////
  private boolean readAMDEeigenFunctionsNfmode2(FortranBinaryInputStream eigInput) {
    double tol = 1.e-06;
    int i, j, nn, ngrid, nfunc;
    int nparams = 50;

    // set up some array lists into which to read different quantities:
    double[] Egrid;
    ArrayList<double[]> globals = new ArrayList<double[]>();
    ArrayList<double[][]> func  = new ArrayList<double[][]>();

    try {
      // get number of grid points
      eigInput.readRecord();
      ngrid = eigInput.readInteger();
      if ((ngrid<0)||(ngrid>maxNgrid)) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "The number of grid points for the\n"+
          "eigenfunctions seems to be erroneous\n"+
          "or the format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return false;
      }

      if (eigInput.isFinFichier()) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "Format not recognised.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return false; 
      }

      // set up the grid:
      Egrid = new double[ngrid];
      for(i=0;i<ngrid;i++) Egrid[i] = eigInput.readDoublePrecision();

      // give warning message about extrapolating the eigenfunctions,
      // if the model's grid is larger than that of the eigenfunctions
      if ((dataCenter.grid[0]<Egrid[0])||
          (dataCenter.grid[dataCenter.grid.length-1]>Egrid[ngrid-1])) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "Warning: the eigenfunction are defined on a smaller grid\n"+
          "than the model and will therefore be extrapolated.",
          "Grid warning",
          JOptionPane.WARNING_MESSAGE
        );
      }

      // now read all available eigenfunctions
      nfunc = 0;
      while(true) {

        // test for the end of the file:
        eigInput.readRecord();
        if (eigInput.isFinFichier()) break;

        // NOW read global params....
        globals.add(new double[nparams]);
        for (i=0; i<nparams; i++) 
          globals.get(nfunc)[i] = eigInput.readDoublePrecision();

        // check the stellar mass and radius associated with the eigenfunctions
        // to see if they match that of the star:
        if ((Math.abs(1.0-dataCenter.StellarMass/globals.get(nfunc)[1]) > tol)||
            (Math.abs(1.0-dataCenter.StellarRadius/globals.get(nfunc)[2]) > tol)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "Model and eigenfunctions do not\n"+
            "agree or the format is incorrect.\n"+
            "Please chose another set.",
            "Eigenfunction-model error",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }

        // NOW read eigenfunctions....
        double[][]tmp = new double[2][ngrid];
        func.add(tmp);
        for (i=0; i<ngrid; i++) {
          tmp[0][i] = eigInput.readDoublePrecision();
          tmp[1][i] = eigInput.readDoublePrecision();
        }
        nfunc++;
      }
      eigInput.close(); 
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Format not recognised.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }

    // check to see that there is some data:
    if (nfunc == 0) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "No eigenfunctions in file! No data loaded.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
    
    // interpolate the eigenfunctions onto the same grid as the model
    // (it is much more efficient to interpolate the entire list in
    // one go rather than to do the different modes individually):
    ArrayList<double[][]> yy = Numerics.interpolateList_r2(Egrid,dataCenter.grid,func,false);
    ngrid = dataCenter.grid.length; // don't forget to change ngrid ...

    // initialise various variables:
    int order = 1;
    double[][] dermat = Numerics.initDeriveFDBand(dataCenter.grid,order,1);
    double[] integrationMatrix = Numerics.findIntegrationMatrix(dataCenter.grid);
    double a, b, L2, L, Freq2;

    // variable into which to store the mode information:
    Kernel aKernel;

    for(nn = 0; nn < nfunc; nn++) {
      // add kernel to the list of theoretical modes:
      aKernel = new Kernel(Math.round((float)globals.get(nn)[18]),Math.round((float)globals.get(nn)[17]));
      dataCenter.kernels.add(aKernel);

      aKernel.FreqTheo = 1e3*globals.get(nn)[26];
      aKernel.FreqDimless = aKernel.FreqTheo/dataCenter.FreqRef;
      aKernel.equations = Pulsations.FromFile;

      // set up some variables in the DataCenter:
      aKernel.vDisp = new double[ngrid];
      aKernel.hDisp = new double[ngrid];
      aKernel.phip  = new double[ngrid];
      aKernel.P     = new double[ngrid];
      aKernel.rho   = new double[ngrid];
      aKernel.dP_P  = new double[ngrid];

      a = 1.0/yy.get(nn)[0][dataCenter.indSurface];
      if (aKernel.l == 0) {
        Freq2 = aKernel.FreqDimless*aKernel.FreqDimless;
        for(i=0; i<ngrid; i++) {
          aKernel.vDisp[i] = a*yy.get(nn)[0][i];
          aKernel.P[i]     = a*Freq2*dataCenter.rho[i]*yy.get(nn)[1][i];
        }
        pulsations.findVariablesRadial(aKernel);
        aKernel.phi = Numerics.applyIntegrationMatrixReverse(integrationMatrix, aKernel.phip);
      } else {
        L2 = (double)aKernel.l*((double)aKernel.l+1.0);
        aKernel.phi = null;  // this is to tell findPhi() to find phi.
        for(i=0; i<ngrid; i++) {
          aKernel.vDisp[i] = a*yy.get(nn)[0][i];
          aKernel.hDisp[i] = a*yy.get(nn)[1][i]/L2;
        }
      }

      // tell findOtherVariables() to find remaining variables:
      aKernel.IK = Double.NaN;
    }
    pulsations.findPhi();  // do this on the entire set, so as to save extra calculations
    return true;
  }

  /////////////////////////////////////////////////////////////////////////
  /** This method reads a file which contains the eigenfunctions.  The
  *   format of the file is the binary output from JCD's ADIPLS code. */
  /////////////////////////////////////////////////////////////////////////
  private boolean readAMDEeigenFunctionsNfmode3(FortranBinaryInputStream eigInput) {
    double tol = 1.e-06;
    int i, j, nn, ngrid, nfunc;
    int nparams = 50;

    // set up some array lists into which to read different quantities:
    double[] Egrid;
    ArrayList<double[]> globals = new ArrayList<double[]>();
    ArrayList<double[][]> func  = new ArrayList<double[][]>();

    try {
      // get number of grid points
      eigInput.readRecord();
      ngrid = eigInput.readInteger();
      if ((ngrid<0)||(ngrid>maxNgrid)) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "The number of grid points for the\n"+
          "eigenfunctions seems to be erroneous\n"+
          "or the format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return false;
      }

      if (eigInput.isFinFichier()) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "Format not recognised.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return false; 
      }

      // set up the grid:
      Egrid = new double[ngrid];
      for(i=0;i<ngrid;i++) Egrid[i] = eigInput.readDoublePrecision();

      // give warning message about extrapolating the eigenfunctions,
      // if the model's grid is larger than that of the eigenfunctions
      if ((dataCenter.grid[0]<Egrid[0])||
          (dataCenter.grid[dataCenter.grid.length-1]>Egrid[ngrid-1])) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "Warning: the eigenfunction are defined on a smaller grid\n"+
          "than the model and will therefore be extrapolated.",
          "Grid warning",
          JOptionPane.WARNING_MESSAGE
        );
      }

      // now read all available eigenfunctions
      nfunc = 0;
      while(true) {

        // test for the end of the file:
        eigInput.readRecord();
        if (eigInput.isFinFichier()) break;

        // NOW read global params....
        globals.add(new double[nparams]);
        for (i=0; i<nparams; i++) 
          globals.get(nfunc)[i] = eigInput.readDoublePrecision();

        // check the stellar mass and radius associated with the eigenfunctions
        // to see if they match that of the star:
        if ((Math.abs(1.0-dataCenter.StellarMass/globals.get(nfunc)[1]) > tol)||
            (Math.abs(1.0-dataCenter.StellarRadius/globals.get(nfunc)[2]) > tol)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "Model and eigenfunctions do not\n"+
            "agree or the format is incorrect.\n"+
            "Please chose another set.",
            "Eigenfunction-model error",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }

        // NOW read eigenfunctions....
        double[][]tmp = new double[2][ngrid];
        func.add(tmp);
        for (i=0; i<ngrid; i++) {
          tmp[0][i] = eigInput.readDoublePrecision();
          tmp[1][i] = eigInput.readDoublePrecision();
        }
        nfunc++;
      }
      eigInput.close(); 
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Format not recognised.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }

    // check to see that there is some data:
    if (nfunc == 0) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "No eigenfunctions in file! No data loaded.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
    
    // interpolate the eigenfunctions onto the same grid as the model
    // (it is much more efficient to interpolate the entire list in
    // one go rather than to do the different modes individually):
    ArrayList<double[][]> yy = Numerics.interpolateList_r2(Egrid,dataCenter.grid,func,false);
    ngrid = dataCenter.grid.length; // don't forget to change ngrid ...

    // initialise various variables:
    int order = 1;
    double[][] dermat = Numerics.initDeriveFDBand(dataCenter.grid,order,1);
    double[] integrationMatrix = Numerics.findIntegrationMatrix(dataCenter.grid);
    double a, b, L2, L;

    // variable into which to store the mode information:
    Kernel aKernel;

    for(nn = 0; nn < nfunc; nn++) {
      // add kernel to the list of theoretical modes:
      aKernel = new Kernel(Math.round((float)globals.get(nn)[18]),Math.round((float)globals.get(nn)[17]));
      dataCenter.kernels.add(aKernel);

      aKernel.FreqTheo = 1e3*globals.get(nn)[26];
      aKernel.FreqDimless = aKernel.FreqTheo/dataCenter.FreqRef;
      aKernel.equations = Pulsations.FromFile;

      // set up some variables in the DataCenter:
      aKernel.vDisp = new double[ngrid];
      aKernel.hDisp = new double[ngrid];
      aKernel.phip  = new double[ngrid];
      aKernel.P     = new double[ngrid];
      aKernel.rho   = new double[ngrid];
      aKernel.dP_P  = new double[ngrid];

      a = Math.sqrt(4.0*Math.PI*Math.pow(dataCenter.grid[dataCenter.indSurface],3.0)*dataCenter.rho[dataCenter.indSurface])
        / yy.get(nn)[0][dataCenter.indSurface];
      if (aKernel.l == 0) {
        for(i=1; i<ngrid; i++) {
          b = Math.sqrt(4.0*Math.PI*Math.pow(dataCenter.grid[i],3.0)*dataCenter.rho[i]);
          aKernel.vDisp[i] = a*yy.get(nn)[0][i]/b;
        }
        aKernel.vDisp[0] = 0.0;  // this is EXTREMELY important
        Dgbmv.dgbmv("N",ngrid,ngrid,order,order,1.0,dermat[1],0,1+2*order,aKernel.vDisp,0,1,0.0,aKernel.P,0,1);
        for(i=1; i<ngrid; i++){
          aKernel.P[i]     = -dataCenter.Gamma1[i]*dataCenter.pressure[i]
                           * (aKernel.P[i] + 2.0*aKernel.vDisp[i]/dataCenter.grid[i])
                           + dataCenter.rho[i]*dataCenter.gravity[i]*aKernel.vDisp[i];
        }
        aKernel.P[0] = -3.0*dataCenter.Gamma1[0]*dataCenter.pressure[0]*aKernel.P[0];
        pulsations.findVariablesRadial(aKernel);
        aKernel.phi = Numerics.applyIntegrationMatrixReverse(integrationMatrix, aKernel.phip);
      } else {
        L2 = (double)aKernel.l*((double)aKernel.l+1.0);
        L  = Math.sqrt(L2);
        aKernel.phi = null;  // this is to tell findPhi() to find phi.
        for(i=1; i<ngrid; i++) {
          b = Math.sqrt(4.0*Math.PI*Math.pow(dataCenter.grid[i],3.0)*dataCenter.rho[i]);
          aKernel.vDisp[i] = a*yy.get(nn)[0][i]/b;
          aKernel.hDisp[i] = a*yy.get(nn)[1][i]/(b*L);
        }
        if (aKernel.l == 1) {
          // extrapolate from above, assuming f(x) = ax^2 + b
          double eta = Math.pow(dataCenter.grid[1],2.0)/(Math.pow(dataCenter.grid[1],2.0)-Math.pow(dataCenter.grid[2],2.0));
          aKernel.vDisp[0] = eta*aKernel.vDisp[2] + (1.0-eta)*aKernel.vDisp[1];
          aKernel.hDisp[0] = eta*aKernel.hDisp[2] + (1.0-eta)*aKernel.hDisp[1];
          // furthermore, vDisp = hDisp in the center, when ell == 1. So we average the two values:
          aKernel.vDisp[0] = (aKernel.hDisp[0]+aKernel.vDisp[0])/2.0;
          aKernel.hDisp[0] =  aKernel.vDisp[0];
        } else {
          aKernel.vDisp[0] = 0.0;
          aKernel.hDisp[0] = 0.0;
        }
      }

      // tell findOtherVariables() to find remaining variables:
      aKernel.IK = Double.NaN;
    }
    pulsations.findPhi();  // do this on the entire set, so as to save extra calculations
    return true;
  }

  ///////////////////////////////////////////////////////////
  // This sets up a cell renderer for a boolean variable.
  ///////////////////////////////////////////////////////////
  private static class EquationRenderer extends DefaultTableCellRenderer {
    public EquationRenderer() {super();}

    public void setValue(Object value) {
      int ivalue = ((Integer)value).intValue();
      if (ivalue == Pulsations.FromFile) {
        setText("File");
      } else if (ivalue == Pulsations.Lagrange) {
        setText("\u03B4P/P");
      } else if (ivalue == Pulsations.Euler) {
        setText("P', \u03C1");
      } else if (ivalue == Pulsations.JCD) {
        setText("JCD");
      } else {
        setText(" ");
      }
      setHorizontalAlignment(SwingConstants.CENTER);
      return;
    }
  }
}
