/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import inversion_applet.Numerics;
import org.netlib.lapack.Dsytrf;
import org.netlib.lapack.Dsytrs;
import org.netlib.lapack.Ilaenv;
import java.util.Random;
import java.util.ArrayList;
import java.util.Vector;
import javax.swing.JOptionPane;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTable;

/*************************************************************************
**
**    MeanDensityEstimate
**                                              Version 2.1, September 2014
**
*************************************************************************/

public class MeanDensityEstimate { 

  private DataCenter dataCenter;
  protected boolean canEstimate; // after initialisation, if the provided information
                                 // is insufficient or erroneous, it is not possible
                                 // to do an inversion, in which case, canEstimate = false
                                 // Otherwise, canEstimate = true
  protected boolean canScale;    // This is true only if it is possible to use a scaling
                                 // law to estimate the mean density. If canEstimate = false, then
                                 // this is also false.
  protected int n_inputKernels;  // the number of kernels used in the inversion

  // variables relating to data, kernels and model:
  private ArrayList<Integer>  dataN = new ArrayList<Integer>(); // radial order of data
  private ArrayList<Integer>  dataL = new ArrayList<Integer>(); // harmonic degree of data
  private ArrayList<Double>   dataF = new ArrayList<Double>();  // frequency from reference model corresponding to data
  private ArrayList<Double>   dataValues   = new ArrayList<Double>();
  private ArrayList<Double>   dataErrors   = new ArrayList<Double>();
  private ArrayList<Integer>  intersection = new ArrayList<Integer>();
  private double[][] ME1Kernel;
  private double[][] ME2Kernel;
  private double[] weights;
  private double[] iGrid;
  private double[] rho;

  // variables relating to surface terms:
  private double[][] surfPoly;
  public static final int n_surf_disp = 0;    // number of surface terms to display;
  public int n_surf;                          // number of surface terms

  // variables needed for matrix inversions:
  private org.netlib.util.intW info = new org.netlib.util.intW(2); // necessary for lapack
  private int lwork;                          // an integer used in the matrix solvers

  // various intermediate matrices used in the SOLA inversions:
  private double[] kernelMat1SOLA;            // matrix used for fitting the data in a SOLA inversion
  private double[] crossTermMat1SOLA;         // matrix used for reducing the cross-term kernel in a SOLA inversion
  private double[] supplMat1SOLA;             // matrix with term relating to supplementary conditions
  private double[] regularisationMatSOLA;     // matrix used for regularisation in a SOLA inversion
  private double[] targetKernelMat1SOLA;      // matrix used for calculating the target kernels
  private double[] SOLAmat1;                  // matrix which contains the coefficients from the inversion
  private int matSizeSOLA;                    // size of left hand side main SOLA matrix
  private int matSizeTargetSOLA;              // size of right hand side SOLA matrix

  // various SOLA inversion results
  public double estimatedMeanDensity;         // this gives the mean density estimate
  public double oneSigmaError;                // this gives the error based on the frequency shift errors
  public double averagingError;               // this gives the error based on averaging kernel mismatch
  public double crossTermError;               // this gives the error from the cross-term kernel
  public double trueMeanDensity;              // this gives the true mean density
  public double trueAveragingError;           // this gives the true error from the averaging kernel mismatch
  public double trueCrossTermError;           // this gives the true error from cross-term effects
  public double[] surfTerms;                  // this keeps track of the surface terms resulting from the inversion

  // variables and results used in scaling relations:
  private double[] SCALINGmat1;               // matrix which contains the coefficients from the Delta nu scaling relation
  private double[] SCALINGmat2;               // matrix which contains the coefficients from Kjeldsen's approach
  private int[] larray;                       // array which contains the number of different radial orders for each ell value
  public double estimatedMeanDensitySCALING;  // this gives the mean density estimate
  public double oneSigmaErrorSCALING;         // this gives the error based on the frequency shift errors
  public double averagingErrorSCALING;        // this gives the error based on averaging kernel mismatch
  public double crossTermErrorSCALING;        // this gives the error from the cross-term kernel
  public double trueAveragingErrorSCALING;    // this gives the true error from the averaging kernel mismatch
  public double trueCrossTermErrorSCALING;    // this gives the true error from cross-term effects
  public double[] surfTermsSCALING;           // this keeps track of the surface terms resulting from the scaling law
  public double estimatedMeanDensitySCALING2; // this gives the mean density estimate
  public double oneSigmaErrorSCALING2;        // this gives the error based on the frequency shift errors
  public double averagingErrorSCALING2;       // this gives the error based on averaging kernel mismatch
  public double crossTermErrorSCALING2;       // this gives the error from the cross-term kernel
  public double trueAveragingErrorSCALING2;   // this gives the true error from the averaging kernel mismatch
  public double trueCrossTermErrorSCALING2;   // this gives the true error from cross-term effects
  public double[] surfTermsSCALING2;          // this keeps track of the surface terms resulting from the scaling law

  /********************************************************
  *  A constructor which does the initialisation
  ********************************************************/
  public MeanDensityEstimate(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
     
    // At the beginning, no data is loaded, so no inversion
    // can be done.  Therefore the inversion is up to date.
    dataCenter.meanDensityEstimateUpToDate = true;
    this.canEstimate = false;
    this.canScale = false;
    
    // This determines the blocksize to be used in the symmetric 
    // matrix solvers. I will just put some average matrix size
    // rather than an exact number.
    this.lwork = Ilaenv.ilaenv(1,"DSYTRF","U",200,200,-1,-1);

    // initialise results (NaN means that the result is not available)
    trueMeanDensity = Double.NaN;
    estimatedMeanDensity = Double.NaN;
    oneSigmaError = Double.NaN;
    averagingError = Double.NaN;
    crossTermError = Double.NaN;
    trueAveragingError = Double.NaN;
    trueCrossTermError = Double.NaN;
    estimatedMeanDensitySCALING = Double.NaN;
    oneSigmaErrorSCALING = Double.NaN;
    averagingErrorSCALING = Double.NaN;
    crossTermErrorSCALING = Double.NaN;
    trueAveragingErrorSCALING = Double.NaN;
    trueCrossTermErrorSCALING = Double.NaN;
    estimatedMeanDensitySCALING2 = Double.NaN;
    oneSigmaErrorSCALING2 = Double.NaN;
    averagingErrorSCALING2 = Double.NaN;
    crossTermErrorSCALING2 = Double.NaN;
    trueAveragingErrorSCALING2 = Double.NaN;
    trueCrossTermErrorSCALING2 = Double.NaN;

    n_surf = (dataCenter.n_surfME>n_surf_disp)?dataCenter.n_surfME:n_surf_disp;
    surfTerms = new double[n_surf];
    surfTermsSCALING = new double[n_surf];
    surfTermsSCALING2 = new double[n_surf];
    for(int i=0; i<n_surf; i++) {
      surfTerms[i] = Double.NaN; 
      surfTermsSCALING[i] = Double.NaN;
      surfTermsSCALING2[i] = Double.NaN;
    }
  }

  ////////////////////////////////////////////////////////////
  /** This calls various initialisation methods */
  ////////////////////////////////////////////////////////////
  public void init(JComponent c) {

    // reinitialise results for a new inversion
    trueMeanDensity = Double.NaN;
    estimatedMeanDensity = Double.NaN;
    oneSigmaError = Double.NaN;
    averagingError = Double.NaN;
    crossTermError = Double.NaN;
    trueAveragingError = Double.NaN;
    trueCrossTermError = Double.NaN;
    estimatedMeanDensitySCALING = Double.NaN;
    oneSigmaErrorSCALING = Double.NaN;
    averagingErrorSCALING = Double.NaN;
    crossTermErrorSCALING = Double.NaN;
    trueAveragingErrorSCALING = Double.NaN;
    trueCrossTermErrorSCALING = Double.NaN;
    estimatedMeanDensitySCALING2 = Double.NaN;
    oneSigmaErrorSCALING2 = Double.NaN;
    averagingErrorSCALING2 = Double.NaN;
    crossTermErrorSCALING2 = Double.NaN;
    trueAveragingErrorSCALING2 = Double.NaN;
    trueCrossTermErrorSCALING2 = Double.NaN;

    canEstimate = false;
    canScale = false;
    dataCenter.meanDensityEstimateUpToDate = true;

    n_inputKernels = 0;
    initIntersection(c);
    if (canEstimate) {
      initMEKernels();
      initModel();
      initNoise();
      initSurfPoly();
      initSOLA();
      initDelta1SOLA();
      initSCALING();
    }
  }

  ////////////////////////////////////////////////////////////
  // This finds the intersection between the set of kernels
  // and the set of frequency shift data and stores it in
  // an ArrayList called intersection
  ////////////////////////////////////////////////////////////
  private void initIntersection(JComponent c) {

    canEstimate = true;

    // make sure frequency shift data is loaded:
    if (dataCenter.nfreq == 0) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No frequency and associated error bar loaded.\n"+
        "Unable to proceed.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canEstimate = false;
      n_inputKernels = 0;
      return;
    }

    // make sure a model is loaded:
    if (!dataCenter.modelLoaded) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No model loaded.  Unable to proceed",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canEstimate = false;
      n_inputKernels = 0;
      return;
    }

    boolean dataWithZeroError = false;
    boolean incorrectDF = false;
    boolean duplicateKernels = false;

    // clear previous data:
    intersection.clear();
    dataValues.clear();
    dataErrors.clear();
    dataN.clear();
    dataL.clear();
    dataF.clear();

    // find index array and frequency shifts
    for(int i=0; i< dataCenter.modes.size();i++) {
      if (dataCenter.modes.get(i).FreqObs == null) continue;
      if (dataCenter.modes.get(i).ErrorFreq == null) continue;
      if (dataCenter.modes.get(i).kernel == null) continue;
      if (dataCenter.modes.get(i).isDuplicate) {
        duplicateKernels = true;
        continue;
      }
      if (dataCenter.modes.get(i).ErrorFreq < 0) {
        dataWithZeroError = true;
        continue;
      }
      if (dataCenter.modes.get(i).l < 0) {
        incorrectDF = true;
        continue;
      }
      
      intersection.add(i);
      dataValues.add(dataCenter.modes.get(i).FreqObs/dataCenter.modes.get(i).kernel.FreqTheo-1.0);
      dataErrors.add(dataCenter.modes.get(i).ErrorFreq/dataCenter.modes.get(i).kernel.FreqTheo);
      dataN.add(dataCenter.modes.get(i).kernel.n);
      dataL.add(dataCenter.modes.get(i).l);
      dataF.add(dataCenter.modes.get(i).kernel.FreqDimless);
    }

    n_inputKernels = intersection.size();

    if (n_inputKernels == 0) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No kernels/data from which to carry\n"+
        "out an inversion.  Unable to proceed.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canEstimate = false;
      n_inputKernels = 0;
      return;
    }

    // Test to see if all the relevant kernels have
    // been calculated and used in the inversion.
    if (n_inputKernels < dataCenter.nfreq) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some kernels might be missing from this inversion.\n"+
        "Pressing the \"Calculate kernels\" button on the\n"+
        "kernel tab would remedy this problem.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (dataWithZeroError) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the frequency shift data has\n"+
        "a 0 error bar, which is not allowed.\n"+
        "This data has consequently been ignored.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
      
    if (incorrectDF) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the frequency shift data seems to be erroneous:\n"+
        "  (\u2113 < 1 for n < 1) or (\u2113 < 0 for n > 0).\n"+
        "This data has consequently not been not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (duplicateKernels) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "There are duplicate kernels.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
  }

  ////////////////////////////////////////////////////////////
  // set up structural kernels before doing inversions:
  ////////////////////////////////////////////////////////////
  private void initMEKernels() {
    ME1Kernel = new double[n_inputKernels][];
    ME2Kernel = new double[n_inputKernels][];
    
    if (dataCenter.kernelTypeME == 0) {
      for(int l=0; l<n_inputKernels; l++) {
        ME1Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.rhoGamma1Kernel;
        ME2Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.c2rhoKernel;
      }
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        ME1Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.c2rhoKernel;
        ME2Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.rhoc2Kernel;
      }
    }
  }

  ////////////////////////////////////////////////////////////
  // set up structural kernels before doing inversions:
  ////////////////////////////////////////////////////////////
  private void initModel() {
    // initialise some integration weights:
    iGrid = dataCenter.grid;
    weights = Numerics.findIntegrationWeights(iGrid);
    rho = dataCenter.rho;
  }

  /********************************************************
  * Initialise the Noise array with a Gauss distribution:
  * Mean = 0.0; std. deviation= 1.0.
  ********************************************************/
  public void initNoise() {
    Random generator = new Random();
    dataCenter.MENoise = new double[intersection.size()];
    for(int l=0; l < intersection.size(); l++) {
      dataCenter.MENoise[l] = generator.nextGaussian();
    }
  }

  /********************************************************
  * This initialises the polynomial expansion used to
  * model the surface terms.
  ********************************************************/
  public void initSurfPoly() {
    n_surf = (dataCenter.n_surfME>n_surf_disp)?dataCenter.n_surfME:n_surf_disp;
    surfTerms = new double[n_surf];
    surfTermsSCALING = new double[n_surf];
    surfTermsSCALING2 = new double[n_surf];
    for(int i=0; i<n_surf; i++) {
      surfTerms[i] = Double.NaN; 
      surfTermsSCALING[i] = Double.NaN;
      surfTermsSCALING2[i] = Double.NaN;
    }

    surfPoly = new double[n_inputKernels][];
    
    for(int l=0; l<n_inputKernels; l++) {
      surfPoly[l] = Numerics.Legendre(dataCenter.modes.get(intersection.get(l)).kernel.FreqDimless,n_surf);
      for(int n=0; n<n_surf; n++) surfPoly[l][n]/=dataCenter.modes.get(intersection.get(l)).kernel.Enl;
    }
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  protected void initSOLA() {
    int n_grid = iGrid.length;
    matSizeSOLA = n_inputKernels+dataCenter.n_surfME+1;

    // initialise matrix associated with ME1 kernels:
    kernelMat1SOLA = new double[matSizeSOLA*matSizeSOLA];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        for(int i=0; i<n_grid; i++){
          kernelMat1SOLA[l+matSizeSOLA*ll]+=weights[i]
               *ME1Kernel[l][i]*ME1Kernel[ll][i];
        }
      }
    }

    // initialise the matrix used for reducing the cross-term from struct 2 on struct 1
    double[] w = new double[n_grid];
    for(int i=0; i<n_grid; i++) w[i] = weights[i];//*Math.pow(1.0+iGrid[i],4.0);
    crossTermMat1SOLA = new double[n_inputKernels*n_inputKernels];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        for(int i=0; i<n_grid; i++){
          crossTermMat1SOLA[l+n_inputKernels*ll]+=w[i]
               *ME2Kernel[l][i]*ME2Kernel[ll][i];
        }
      }
    }

    // initialise matrix associated with supplementary constraints for
    // struct1 kernels:
    double sum;
    supplMat1SOLA = new double[n_inputKernels*(1+dataCenter.n_surfME)];
    for(int l=0; l<n_inputKernels; l++){
      supplMat1SOLA[l] = 1.00;
      for(int n=0; n<dataCenter.n_surfME; n++) {
        supplMat1SOLA[l+n_inputKernels*(1+n)] = surfPoly[l][n]*0.5;
      }
    }

    // initialise regularisation "matrix":
    double mu = 0.0;
    for(int l=0; l<n_inputKernels; l++) mu+=dataErrors.get(l)*dataErrors.get(l);
    mu = (double)n_inputKernels/mu;

    regularisationMatSOLA = new double[n_inputKernels];
    for(int l=0; l<n_inputKernels; l++){
      regularisationMatSOLA[l]=mu*dataErrors.get(l)*dataErrors.get(l);
    }
  }

  /********************************************************
  * Initialise the matrix with the target kernel.
  ********************************************************/
  protected void initDelta1SOLA() {
    matSizeTargetSOLA = n_inputKernels+1;
    int n_grid = iGrid.length;
    double FunctionValue;
    targetKernelMat1SOLA = new double[matSizeTargetSOLA];
    for(int j=0; j<n_grid; j++) {
      FunctionValue = 4.0*Math.PI*rho[j]*iGrid[j]*iGrid[j]*weights[j];
      for(int l=0; l<n_inputKernels; l++) 
        targetKernelMat1SOLA[l]+=(FunctionValue*ME1Kernel[l][j]);
    }
    targetKernelMat1SOLA[n_inputKernels]=2.0;
  }

  /********************************************************
  * Do the inversion, using SOLA.
  ********************************************************/
  protected void SOLAinversion1() {
    int n_grid = iGrid.length;
    int ipiv[] = new int[matSizeSOLA];

    double[] mat = kernelMat1SOLA.clone();
    double tanTheta = Math.tan(dataCenter.METheta);
    for(int l=0; l<n_inputKernels; l++) 
      mat[l+matSizeSOLA*l]+=tanTheta*regularisationMatSOLA[l];
    for(int l=0; l<n_inputKernels; l++) {
      for(int n=0; n<(1+dataCenter.n_surfME); n++)
        mat[l+matSizeSOLA*(n+n_inputKernels)]+=supplMat1SOLA[l+n_inputKernels*n];
    }

    for(int l=0; l<n_inputKernels; l++) {
      for(int ll=l; ll<n_inputKernels; ll++) {
         mat[l+matSizeSOLA*ll]+=dataCenter.MEBeta*crossTermMat1SOLA[l+n_inputKernels*ll];
      }
    }

    SOLAmat1 = new double[matSizeSOLA];
    for(int l=0; l<matSizeTargetSOLA; l++) SOLAmat1[l] = targetKernelMat1SOLA[l];
    double[] work = new double[lwork*matSizeSOLA];
    Dsytrf.dsytrf("U",matSizeSOLA,mat,0,matSizeSOLA,ipiv,0,work,0,lwork*matSizeSOLA,info);
    Dsytrs.dsytrs("U",matSizeSOLA,1,mat,0,matSizeSOLA,ipiv,0,SOLAmat1,0,matSizeSOLA,info);

    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.MENoiseSelected) {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.MENoiseLevel*dataCenter.MENoise[l]*dataErrors.get(l);
      }
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }

    // find the mean density estimate
    estimatedMeanDensity = 0.0;
    for(int l=0; l<n_inputKernels; l++)  estimatedMeanDensity+=dataPlusNoise[l]*SOLAmat1[l];
    estimatedMeanDensity = Math.pow(1.0+estimatedMeanDensity/2.0,2.0);

    // find one sigma error bars:
    oneSigmaError = 0.0;
    for(int l=0; l<n_inputKernels; l++)  oneSigmaError+=Math.pow(SOLAmat1[l]*dataErrors.get(l),2.0);
    oneSigmaError=Math.sqrt(estimatedMeanDensity*oneSigmaError);

    // find error from averaging kernel mismatch:
    averagingError = 0.0;
    double[] Averaging = findAveragingKernel1SOLA();
    for(int i=0; i<n_grid; i++)
      averagingError+=weights[i]*Math.pow(Averaging[i]-4.0*Math.PI*rho[i]*iGrid[i]*iGrid[i],2.0);
    averagingError = Math.sqrt(averagingError);

    // find error from cross-term kernel:
    crossTermError = 0.0;
    double[] CrossTerm = findCrossTermKernel1SOLA();
    for(int i=0; i<n_grid; i++) crossTermError+=weights[i]*CrossTerm[i]*CrossTerm[i];
    crossTermError = Math.sqrt(crossTermError);
    
    // find surface terms
    for(int i=0; i<n_surf; i++) {
      surfTerms[i] = 0.0;
      for(int l=0; l<n_inputKernels; l++) {
        surfTerms[i]+=SOLAmat1[l]*surfPoly[l][i];
      }
    }

    // find true error if true structural variations are provided
    if (dataCenter.gridTargetStruct != null) {
      trueAveragingError = 0.0;
      trueCrossTermError = 0.0;
      double[] struct1 = Numerics.interpolate1D_r2(dataCenter.gridTargetStruct,dataCenter.grid,dataCenter.targetStruct1,false);
      double[] struct2 = Numerics.interpolate1D_r2(dataCenter.gridTargetStruct,dataCenter.grid,dataCenter.targetStruct2,false);

      if (Double.isNaN(trueMeanDensity)) {
        trueMeanDensity = 0.0;
        for(int i=0; i<n_grid; i++) trueMeanDensity += 4.0*Math.PI*iGrid[i]*iGrid[i]*rho[i]*struct1[i]*weights[i];
      }

      double s1,s2;
      for(int i=0; i<n_grid; i++) {
        s1 = (struct1[i]+1.0)/estimatedMeanDensity-1.0;
        trueAveragingError += (Averaging[i]-4.0*Math.PI*rho[i]*iGrid[i]*iGrid[i])*s1*weights[i];
        s2 = (struct2[i]+1.0)/Math.pow(estimatedMeanDensity,dataCenter.scalePowerME)-1.0;
        trueCrossTermError += CrossTerm[i]*s2*weights[i];
      }
    }
  }

  /********************************************************
  * This calculates the SOLA averaging kernel.
  ********************************************************/
  protected double[] findAveragingKernel1SOLA() {
    if (!canEstimate) return null;
    int n_grid = dataCenter.grid.length;

    int i,l;    
    double[] result = new double[n_grid];
    for(i=0; i<n_grid ;i++){
      for(l=0; l<n_inputKernels;l++)
        result[i]+=(ME1Kernel[l][i]*SOLAmat1[l]);
    }
    return result;
  }

  /********************************************************
  * This calculates the SOLA cross-term kernel.
  ********************************************************/
  protected double[] findCrossTermKernel1SOLA() {
    if (!canEstimate) return null;
    int n_grid = dataCenter.grid.length;

    int i,l;    
    double[] result = new double[n_grid];
    for(i=0; i<n_grid ;i++){
      for(l=0; l<n_inputKernels;l++)
        result[i]+=(ME2Kernel[l][i]*SOLAmat1[l]);
    }
    return result;
  }

  /********************************************************
  * Initialise coefficients used in differential scaling law.
  ********************************************************/
  protected void initSCALING() {
    // Determine whether it is possible to apply a scaling law,
    // and if so, up to what harmonic degree ell:
    int lmax = -1;
    for(int l=0; l<n_inputKernels; l++) {
      if (lmax < dataL.get(l)) lmax=dataL.get(l);
    }
    larray = new int[lmax+1];
    for(int l=0;l<=lmax;l++) larray[l] = 0;
    for(int l=0; l<n_inputKernels; l++) { 
      if (dataN.get(l)>=dataCenter.nminME) larray[dataL.get(l)]++;
    }
    canScale = false;
    for(int l=0;l<=lmax;l++) { if (larray[l]>1) {canScale = true; break;} }
    if (!canScale) return;

    // find matrix with coefficients from differential scaling law:
    double ntotal = 0; // total number of data (used for weighting purposes)
    double[] navg  = new double[lmax+1];
    double[] dnavg = new double[lmax+1];
    for(int l=0;l<=lmax;l++) {if (larray[l]>1) ntotal+=((double)(larray[l]-1));}
    for(int l=0;l<=lmax;l++) {navg[l] = 0.0; dnavg[l] = 0.0;}
    for(int l=0; l<n_inputKernels; l++) {
      if (dataN.get(l)>=dataCenter.nminME) {
        navg[dataL.get(l)] +=((double)dataN.get(l));
        dnavg[dataL.get(l)]+=((double)(dataN.get(l)*dataN.get(l)));
      }
    }
    for(int l=0;l<=lmax;l++) {
      if (larray[l]>1) {
        navg[l]/=((double)larray[l]);
        dnavg[l]= dnavg[l]/((double)larray[l]) - navg[l]*navg[l];
      }
    }

    double favg  = 0.0;
    double dfavg = 0.0;
    for(int l=0; l<n_inputKernels; l++) {
      if ((larray[dataL.get(l)]>1)&&(dataN.get(l)>=dataCenter.nminME)) {
        favg += dataF.get(l)*((double)(larray[dataL.get(l)]-1))
              / (((double)larray[dataL.get(l)])*ntotal);
        dfavg+= dataF.get(l)*((double)dataN.get(l)-navg[dataL.get(l)])*((double)(larray[dataL.get(l)]-1))
              / (dnavg[dataL.get(l)]*((double)larray[dataL.get(l)])*ntotal);
      }
    }

    double b = dataCenter.expME;
    SCALINGmat1 = new double[n_inputKernels];
    SCALINGmat2 = new double[n_inputKernels];
    for(int l=0; l<n_inputKernels; l++) {
      if ((larray[dataL.get(l)]>1)&&(dataN.get(l)>=dataCenter.nminME)) {
        SCALINGmat1[l] = 2.0*dataF.get(l)*((double)dataN.get(l)-navg[dataL.get(l)])*((double)(larray[dataL.get(l)]-1))
                       / (dfavg*dnavg[dataL.get(l)]*((double)larray[dataL.get(l)])*ntotal);
        SCALINGmat2[l] = (2.0*dataF.get(l)/(b-1.0))
                       * (b*((double)(larray[dataL.get(l)]-1))/(favg*ntotal*(double)larray[dataL.get(l)])
                       - ((double)dataN.get(l)-navg[dataL.get(l)])*((double)(larray[dataL.get(l)]-1))
                       / (dfavg*dnavg[dataL.get(l)]*((double)larray[dataL.get(l)])*ntotal));
      } else {
        SCALINGmat1[l] = 0.0;
        SCALINGmat2[l] = 0.0;
      }
    }

    // This applies a nu rather than Delta nu scaling law
    /*double den = 0.0;
    for(int l=0; l<n_inputKernels; l++) {
      if (dataN.get(l)<dataCenter.nminME) continue;
      den+=Math.pow(1.0/dataErrors.get(l),2.0);
    }

    SCALINGmat1 = new double[n_inputKernels];
    for(int l=0; l<n_inputKernels; l++) {
      if (dataN.get(l)<dataCenter.nminME) {
        SCALINGmat1[l] = 0.0;
      } else {
        SCALINGmat1[l] = 2.0/(den*Math.pow(dataErrors.get(l),2.0));
      }
    }*/
  }

  /********************************************************
  * Use scaling law in differential form:
  ********************************************************/
  protected void applySCALING() {
    if (!canScale) return; // easy exit if possible

    int n_grid = iGrid.length;
    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.MENoiseSelected) {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.MENoiseLevel*dataCenter.MENoise[l]*dataErrors.get(l);
      }
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }

    // find the mean density estimate
    estimatedMeanDensitySCALING  = 0.0;
    estimatedMeanDensitySCALING2 = 0.0;
    for(int l=0; l<n_inputKernels; l++)  {
      estimatedMeanDensitySCALING +=dataPlusNoise[l]*SCALINGmat1[l];
      estimatedMeanDensitySCALING2+=dataPlusNoise[l]*SCALINGmat2[l];
    }
    estimatedMeanDensitySCALING  = Math.pow(1.0+estimatedMeanDensitySCALING /2.0,2.0);
    estimatedMeanDensitySCALING2 = Math.pow(1.0+estimatedMeanDensitySCALING2/2.0,2.0);

    // find one sigma error bars:
    oneSigmaErrorSCALING  = 0.0;
    oneSigmaErrorSCALING2 = 0.0;
    for(int l=0; l<n_inputKernels; l++)  {
      oneSigmaErrorSCALING +=Math.pow(SCALINGmat1[l]*dataErrors.get(l),2.0);
      oneSigmaErrorSCALING2+=Math.pow(SCALINGmat2[l]*dataErrors.get(l),2.0);
    }
    oneSigmaErrorSCALING =Math.sqrt(estimatedMeanDensitySCALING *oneSigmaErrorSCALING );
    oneSigmaErrorSCALING2=Math.sqrt(estimatedMeanDensitySCALING2*oneSigmaErrorSCALING2);

    // find error from averaging kernel mismatch:
    averagingErrorSCALING  = 0.0;
    averagingErrorSCALING2 = 0.0;
    double[] Averaging  = findAveragingKernel1SCALING();
    double[] Averaging2 = findAveragingKernel2SCALING();
    for(int i=0; i<n_grid; i++) {
      averagingErrorSCALING +=weights[i]*Math.pow(Averaging[i] -4.0*Math.PI*rho[i]*iGrid[i]*iGrid[i],2.0);
      averagingErrorSCALING2+=weights[i]*Math.pow(Averaging2[i]-4.0*Math.PI*rho[i]*iGrid[i]*iGrid[i],2.0);
    }
    averagingErrorSCALING  = Math.sqrt(averagingErrorSCALING);
    averagingErrorSCALING2 = Math.sqrt(averagingErrorSCALING2);

    // find error from cross-term kernel:
    crossTermErrorSCALING  = 0.0;
    crossTermErrorSCALING2 = 0.0;
    double[] CrossTerm  = findCrossTermKernel1SCALING();
    double[] CrossTerm2 = findCrossTermKernel2SCALING();
    for(int i=0; i<n_grid; i++) {
      crossTermErrorSCALING +=weights[i]*CrossTerm[i] *CrossTerm[i] ;
      crossTermErrorSCALING2+=weights[i]*CrossTerm2[i]*CrossTerm2[i];
    }
    crossTermErrorSCALING  = Math.sqrt(crossTermErrorSCALING);
    crossTermErrorSCALING2 = Math.sqrt(crossTermErrorSCALING2);

    // find surface terms
    for(int i=0; i<n_surf; i++) {
      surfTermsSCALING[i]  = 0.0;
      surfTermsSCALING2[i] = 0.0;
      for(int l=0; l<n_inputKernels; l++) {
        surfTermsSCALING[i] +=SCALINGmat1[l]*surfPoly[l][i];
        surfTermsSCALING2[i]+=SCALINGmat2[l]*surfPoly[l][i];
      }
    }

    // find true error if true structural variations are provided
    if (dataCenter.gridTargetStruct != null) {
      trueAveragingErrorSCALING  = 0.0;
      trueCrossTermErrorSCALING  = 0.0;
      trueAveragingErrorSCALING2 = 0.0;
      trueCrossTermErrorSCALING2 = 0.0;
      double[] struct1 = Numerics.interpolate1D_r2(dataCenter.gridTargetStruct,dataCenter.grid,dataCenter.targetStruct1,false);
      double[] struct2 = Numerics.interpolate1D_r2(dataCenter.gridTargetStruct,dataCenter.grid,dataCenter.targetStruct2,false);

      if (Double.isNaN(trueMeanDensity)) {
        trueMeanDensity = 0.0;
        for(int i=0; i<n_grid; i++) trueMeanDensity += 4.0*Math.PI*iGrid[i]*iGrid[i]*rho[i]*struct1[i]*weights[i];
      }

      double s1,s2;
      for(int i=0; i<n_grid; i++) {
        s1 = (struct1[i]+1.0)/estimatedMeanDensitySCALING-1.0;
        trueAveragingErrorSCALING  += (Averaging[i]-4.0*Math.PI*rho[i]*iGrid[i]*iGrid[i])*s1*weights[i];
        s1 = (struct1[i]+1.0)/estimatedMeanDensitySCALING2-1.0;
        trueAveragingErrorSCALING2 += (Averaging2[i]-4.0*Math.PI*rho[i]*iGrid[i]*iGrid[i])*s1*weights[i];
        s2 = (struct2[i]+1.0)/Math.pow(estimatedMeanDensitySCALING,dataCenter.scalePowerME)-1.0;
        trueCrossTermErrorSCALING  += CrossTerm[i]*s2*weights[i];
        s2 = (struct2[i]+1.0)/Math.pow(estimatedMeanDensitySCALING2,dataCenter.scalePowerME)-1.0;
        trueCrossTermErrorSCALING2 += CrossTerm2[i]*s2*weights[i];
      }
    }
  }

  /********************************************************
  * This calculates the SCALING averaging kernel.
  ********************************************************/
  protected double[] findAveragingKernel1SCALING() {
    if (!canScale) return null;
    int n_grid = dataCenter.grid.length;

    int i,l;    
    double[] result = new double[n_grid];
    for(i=0; i<n_grid ;i++){
      for(l=0; l<n_inputKernels;l++)
        result[i]+=(ME1Kernel[l][i]*SCALINGmat1[l]);
    }
    return result;
  }

  /********************************************************
  * This calculates the SCALING cross-term kernel.
  ********************************************************/
  protected double[] findCrossTermKernel1SCALING() {
    if (!canScale) return null;
    int n_grid = dataCenter.grid.length;

    int i,l;    
    double[] result = new double[n_grid];
    for(i=0; i<n_grid ;i++){
      for(l=0; l<n_inputKernels;l++)
        result[i]+=(ME2Kernel[l][i]*SCALINGmat1[l]);
    }
    return result;
  }

  /********************************************************
  * This calculates the SCALING averaging kernel.
  ********************************************************/
  protected double[] findAveragingKernel2SCALING() {
    if (!canScale) return null;
    int n_grid = dataCenter.grid.length;

    int i,l;    
    double[] result = new double[n_grid];
    for(i=0; i<n_grid ;i++){
      for(l=0; l<n_inputKernels;l++)
        result[i]+=(ME1Kernel[l][i]*SCALINGmat2[l]);
    }
    return result;
  }

  /********************************************************
  * This calculates the SCALING cross-term kernel.
  ********************************************************/
  protected double[] findCrossTermKernel2SCALING() {
    if (!canScale) return null;
    int n_grid = dataCenter.grid.length;

    int i,l;    
    double[] result = new double[n_grid];
    for(i=0; i<n_grid ;i++){
      for(l=0; l<n_inputKernels;l++)
        result[i]+=(ME2Kernel[l][i]*SCALINGmat2[l]);
    }
    return result;
  }
}
