/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import graph.IPlot;
import graph.Plot;
import inversion_applet.Utilities;
import inversion_applet.DataCenter;
import inversion_applet.MeanDensityEstimate;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JCheckBox;
import javax.swing.JSlider;
import javax.swing.JComponent;
import javax.swing.JComboBox;
import javax.swing.JRadioButton;
import javax.swing.JFileChooser;
import javax.swing.ButtonGroup;
import javax.swing.SwingConstants;
import java.io.File;
import java.io.BufferedReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import java.util.StringTokenizer;
import java.util.ArrayList;

public class MeanDensityEstimatePage extends JPanel {

  private DataCenter dataCenter;
  private IPlot graphAveragingKernel;
  private IPlot graphCrossTermKernel;
  private JPanel graphContainer;
  private MeanDensityEstimate my_estimate;
  private JPanel SliderContainer;

  private JLabel numberOfKernels;
  private JButton Refresh;
  private JCheckBox calculateWhileSliding;
  private JComboBox kernelTypeComboBox;
  
  private JCheckBox SOLACheckBox;
  private JCheckBox ScalingCheckBox;
  private JCheckBox KjeldsenCheckBox;
  private JCheckBox TargetCheckBox;

  private JCheckBox NoiseCheckBox;
  private JTextField NoiseTextField;
  private JSlider NoiseSlider;
  private JButton NoiseRefresh;

  private JLabel SOLAtitle;
  private JLabel BetaLabel;
  private JTextField BetaTextField;
  private JSlider BetaSlider;
  private JLabel ThetaLabel;
  private JTextField ThetaTextField;
  private JSlider ThetaSlider;
  private JLabel NSurfLabel;
  private JTextField NSurfTextField;

  private JLabel NminLabel;
  private JTextField NminTextField;
  private JLabel ExpMELabel;
  private JTextField ExpMETextField;

  private JLabel  resultLabel;
  private JButton outputButton;

  private JButton LoadVariationsButton;
  private JButton ClearVariationsButton;

  // this keeps track of whether a parameter has been modified in one of the
  // text fields.
  private boolean paramModified;

  private static final double ln10 = 2.302585092994046;
  private static final double pi43 = 4.0*Math.PI/3.0;
  private static final Color DarkGreen = new Color(0.0f,0.5f,0.0f);

  /************************************************************************
  * A constructor.
  ************************************************************************/
  public MeanDensityEstimatePage(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
  }

  //////////////////////////////////////////////////////////////
  // These methods give the relationship between the slider
  // value and that of various parameters of the inversion.
  //////////////////////////////////////////////////////////////
  private int NoiseToSlider(double NoiseLevel) {
    if (NoiseLevel == 0.0) return -60;
    int result = (int)Math.round(Math.log(NoiseLevel)*20.0/ln10);
    if (result < -60) result = -60;
    if (result > 60) result = 60;
    return result;
  }
  private double SliderToNoise() {
    return Math.pow(10.0,NoiseSlider.getValue()/20.0);
  }
  private int BetaToSlider(double Beta) {
    if (Beta == 0.0) return -200;
    int result = (int)Math.round(Math.log(Beta)*20.0/ln10);
    if (result < -200) result = -200;
    if (result > 40) result = 40;
    return result;
  }
  private double SliderToBeta() {
    return Math.pow(10.0,BetaSlider.getValue()/20.0);
  }
  private int ThetaToSlider(double Theta) {
    if (Theta == 0.0) return -200;
    int result = (int)Math.round(Math.log(Theta)*20.0/ln10);
    if (result < -200) result = -200;
    if (result > 0) result = 0;
    return result;
  }
  private double SliderToTheta() {
    return Math.pow(10.0,ThetaSlider.getValue()/20.0);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Setting up the refresh button and the SOLA, Scaling, Kjelsen and target 
  // kernels check boxes.
  /////////////////////////////////////////////////////////////////////////////
  private void initCheckBoxes() {
    String rhobar = "<font style='text-decoration:underline;'>\u03C1</font>";
    Refresh = new JButton("<html> Update "+rhobar+" estimate</html>");
    Refresh.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (dataCenter.MENoiseSelected) {
          if (!checkNoiseTextField()) return;
        }
        if (SOLACheckBox.isSelected()) {
          if (!checkBetaTextField()) return;
          if (!checkThetaTextField()) return;
        }
        if (!checkNSurfTextField()) return;
        if (!checkNminTextField()) return;
        if (!checkExpMETextField()) return;
        updateInversion(Refresh);
        writeMeanDensityEstimate();
        replotAllKernels();
      }
    });

    SOLACheckBox = new JCheckBox("<html> <font color='blue'> SOLA </font></html>",true);
    ScalingCheckBox = new JCheckBox("<html> <font color='red'> \u2039\u0394\u03BD\u203A scaling <font></html>",true);
    KjeldsenCheckBox = new JCheckBox("<html> <font color='green'> Kjeldsen et al. (2008) <font></html>",true);
    TargetCheckBox = new JCheckBox("Target kernels",true);

    SOLACheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        SOLAtitle.setEnabled(SOLACheckBox.isSelected());
        BetaLabel.setEnabled(SOLACheckBox.isSelected());
        BetaTextField.setEnabled(SOLACheckBox.isSelected());
        BetaSlider.setEnabled(SOLACheckBox.isSelected());
        ThetaLabel.setEnabled(SOLACheckBox.isSelected());
        ThetaTextField.setEnabled(SOLACheckBox.isSelected());
        ThetaSlider.setEnabled(SOLACheckBox.isSelected());
        NSurfLabel.setEnabled(SOLACheckBox.isSelected());
        NSurfTextField.setEnabled(SOLACheckBox.isSelected());
        replotAllKernels();
      }
    });

    ScalingCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        NminLabel.setEnabled(ScalingCheckBox.isSelected()||KjeldsenCheckBox.isSelected());
        NminTextField.setEnabled(ScalingCheckBox.isSelected()||KjeldsenCheckBox.isSelected());
        replotAllKernels();
      }
    });

    KjeldsenCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        NminLabel.setEnabled(ScalingCheckBox.isSelected()||KjeldsenCheckBox.isSelected());
        NminTextField.setEnabled(ScalingCheckBox.isSelected()||KjeldsenCheckBox.isSelected());
        ExpMELabel.setEnabled(KjeldsenCheckBox.isSelected());
        ExpMETextField.setEnabled(KjeldsenCheckBox.isSelected());
        replotAllKernels();
      }
    });

    TargetCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        replotAllKernels();
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the noise slider and refresh button
  /////////////////////////////////////////////////////////////////////////////
  private void initNoiseButtons() {

    double NoiseLevel = dataCenter.MENoiseLevel;
    boolean NoiseSelected = dataCenter.MENoiseSelected;
    
    NoiseTextField = new JTextField(String.format("%.4e",NoiseLevel));
    NoiseTextField.setEnabled(NoiseSelected);
    NoiseTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkNoiseTextField()) {
          if (!paramModified) return;
          if (my_estimate.canEstimate) {
            my_estimate.SOLAinversion1();
            my_estimate.applySCALING();
            writeMeanDensityEstimate();
          }
        }
      }
    });

    NoiseSlider = new JSlider(JSlider.HORIZONTAL,-60,60,NoiseToSlider(NoiseLevel));
    NoiseSlider.setEnabled(NoiseSelected);
    NoiseSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double NoiseLevel = SliderToNoise();
        NoiseTextField.setText(String.format("%.4e",NoiseLevel));
        if ((!calculateWhileSliding.isSelected())&&
            (NoiseSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.MENoiseLevel = NoiseLevel;
        if (my_estimate.canEstimate) {
          my_estimate.SOLAinversion1();
          my_estimate.applySCALING();
          writeMeanDensityEstimate();
        }
      }
    });

    NoiseRefresh = new JButton("New noise array");
    NoiseRefresh.setEnabled(NoiseSelected);
    NoiseRefresh.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (my_estimate.canEstimate) {
          my_estimate.initNoise();
          my_estimate.SOLAinversion1();
          my_estimate.applySCALING();
          writeMeanDensityEstimate();
        }
      }
    });
    
    NoiseCheckBox = new JCheckBox("Noise",NoiseSelected);
    NoiseCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.MENoiseSelected = NoiseCheckBox.isSelected();
        NoiseTextField.setEnabled(NoiseCheckBox.isSelected());
        NoiseSlider.setEnabled(NoiseCheckBox.isSelected());
        NoiseRefresh.setEnabled(NoiseCheckBox.isSelected());
        if (my_estimate.canEstimate) {
          my_estimate.SOLAinversion1();
          my_estimate.applySCALING();
          writeMeanDensityEstimate();
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Setting up the SOLA sliders
  /////////////////////////////////////////////////////////////////////////////
  private void initSOLAsliders() {
    SOLAtitle = new JLabel("SOLA",SwingConstants.CENTER);

    // Initialisation
    double Beta = dataCenter.MEBeta;
    double Theta = dataCenter.METheta;

    BetaLabel = new JLabel(" Beta: ");
    BetaTextField = new JTextField(String.format("%.4e",Beta));
    BetaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkBetaTextField()) {
          if (!paramModified) return;
          if (my_estimate.canEstimate) {
            my_estimate.SOLAinversion1();
            writeMeanDensityEstimate();
            replotAllKernels();
          }
        }
      }
    });

    BetaSlider = new JSlider(JSlider.HORIZONTAL,-200,40,BetaToSlider(Beta));
    BetaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Beta = SliderToBeta();
        BetaTextField.setText(String.format("%.4e",Beta));
        if ((!calculateWhileSliding.isSelected())&&
            (BetaSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.MEBeta = Beta;
        if (my_estimate.canEstimate) {
          my_estimate.SOLAinversion1();
          writeMeanDensityEstimate();
          replotAllKernels();
        }
      }
    });
    
    ThetaLabel = new JLabel(" Theta: ");
    ThetaTextField = new JTextField(String.format("%.4e",Theta));
    ThetaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkThetaTextField()) {
          if (!paramModified) return;
          if (my_estimate.canEstimate) {
            my_estimate.SOLAinversion1();
            writeMeanDensityEstimate();
            replotAllKernels();
          }
        }
      }
    });

    ThetaSlider = new JSlider(JSlider.HORIZONTAL,-200,0,ThetaToSlider(Theta));
    ThetaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Theta = SliderToTheta();
        ThetaTextField.setText(String.format("%.4e",Theta));
        if ((!calculateWhileSliding.isSelected())&&
            (ThetaSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.METheta = Theta;
        if (my_estimate.canEstimate){
          my_estimate.SOLAinversion1();
          writeMeanDensityEstimate();
          replotAllKernels();
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise buttons for the selection of kernel pairs, and specific kernels.
  /////////////////////////////////////////////////////////////////////////////
  private void initStructKernelButtons() {
    kernelTypeComboBox = new JComboBox();
    for(int i=0; i<DataCenter.kernelTypesME.length;i++)
      kernelTypeComboBox.addItem(DataCenter.kernelTypesME[i]);
    kernelTypeComboBox.setSelectedItem(DataCenter.kernelTypesME[dataCenter.kernelTypeME]);
    kernelTypeComboBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String newKernelType = (String)kernelTypeComboBox.getSelectedItem();
        if (!DataCenter.kernelTypesME[dataCenter.kernelTypeME].equals(newKernelType)) {
          for(int k=0; k<DataCenter.kernelTypesME.length;k++) {
            if (DataCenter.kernelTypesME[k].equals(newKernelType)) dataCenter.kernelTypeME = k;
          }
          if (!dataCenter.meanDensityEstimateUpToDate||my_estimate.canEstimate) {
            updateInversion(kernelTypeComboBox);
            writeMeanDensityEstimate();
            replotAllKernels();
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the text field and label that allows one to adjust the number
  // of Legendre polynomials to fit/remove surface effects.
  /////////////////////////////////////////////////////////////////////////////
  private void initNSurfPanel() {
    NSurfLabel = new JLabel(" Number of surface poly.: ");
    NSurfTextField = new JTextField(String.format("%d",dataCenter.n_surfME));
    NSurfTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkNSurfTextField()) {
          if (!paramModified) return;
          if (my_estimate.canEstimate) {
            my_estimate.initSurfPoly();
            my_estimate.initSOLA();
            my_estimate.SOLAinversion1();
            writeMeanDensityEstimate();
            replotAllKernels();
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the text field and label that allows one to adjust the minimun
  // radial order which is used in the differential scaling law.
  /////////////////////////////////////////////////////////////////////////////
  private void initNminPanel() {
    NminLabel = new JLabel(" Mininum n in scaling laws: ");
    NminTextField = new JTextField(String.format("%d",dataCenter.nminME));
    NminTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkNminTextField()) {
          if (!paramModified) return;
          if (my_estimate.canScale) {
            my_estimate.initSCALING();
            my_estimate.applySCALING();
            writeMeanDensityEstimate();
            replotAllKernels();
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the text field and label that allows one to adjust the minimun
  // radial order which is used in the differential scaling law.
  /////////////////////////////////////////////////////////////////////////////
  private void initExpMEPanel() {
    ExpMELabel = new JLabel(" Kjeldsen et al. exponent, b: ");
    ExpMETextField = new JTextField(String.format("%.4f",dataCenter.expME));
    ExpMETextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkExpMETextField()) {
          if (!paramModified) return;
          if (my_estimate.canScale) {
            my_estimate.initSCALING();
            my_estimate.applySCALING();
            writeMeanDensityEstimate();
            replotAllKernels();
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Various utility methods which check the entries in the different text
  // fields and returns true if these entries are as they should be.
  // If they are not suitable, an error message will be posted.  Finally,
  // the different sliders will be adjusted according to these entries.
  /////////////////////////////////////////////////////////////////////////////
  private boolean checkNoiseTextField() {
    Double Noise = Utilities.readDouble(NoiseTextField.getText(),
                   "noise level",0e0,1e+30,NoiseTextField);
    if (Noise == null) {
      NoiseTextField.setText(String.format("%.4e",dataCenter.MENoiseLevel));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(NoiseSlider, NoiseToSlider(Noise));
      NoiseTextField.setText(String.format("%.4e",Noise));
      paramModified = (dataCenter.MENoiseLevel != Noise.doubleValue());
      dataCenter.MENoiseLevel = Noise.doubleValue();
      return true;
    }
  }

  private boolean checkBetaTextField() {
    Double Beta = Utilities.readDouble(BetaTextField.getText(),
                   "value of Beta",0.0,1e30,BetaTextField);
    if (Beta == null) {
      BetaTextField.setText(String.format("%.4e",dataCenter.MEBeta));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(BetaSlider, BetaToSlider(Beta));
      BetaTextField.setText(String.format("%.4e",Beta));
      paramModified = (dataCenter.MEBeta != Beta.doubleValue());
      dataCenter.MEBeta = Beta.doubleValue();
      return true;
    }
  }

  private boolean checkThetaTextField() {
    Double Theta = Utilities.readDouble(ThetaTextField.getText(),
                   "value of Theta",0.0,Math.PI/2.0,ThetaTextField);
    if (Theta == null) {
      ThetaTextField.setText(String.format("%.4e",dataCenter.METheta));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(ThetaSlider, ThetaToSlider(Theta));
      ThetaTextField.setText(String.format("%.4e",Theta));
      paramModified = (dataCenter.METheta != Theta.doubleValue());
      dataCenter.METheta = Theta.doubleValue();
      return true;
    }
  }

  private boolean checkNSurfTextField() {
    Integer NSurf = Utilities.readInt(NSurfTextField.getText(),
                            "number of polynomials",
                            0,50,NSurfTextField);
    if (NSurf == null) {
      NSurfTextField.setText(String.format("%d",dataCenter.n_surfME));
      paramModified = false;
      return false;
    } else {
      paramModified = (dataCenter.n_surfME != NSurf.intValue());
      dataCenter.n_surfME = NSurf.intValue();
      return true;
    }
  }

  private boolean checkNminTextField() {
    Integer Nmin = Utilities.readInt(NminTextField.getText(),
                            "minimun radial order",
                            -9999,9999,NminTextField);
    if (Nmin == null) {
      NminTextField.setText(String.format("%d",dataCenter.nminME));
      paramModified = false;
      return false;
    } else {
      paramModified = (dataCenter.nminME != Nmin.intValue());
      dataCenter.nminME = Nmin.intValue();
      return true;
    }
  }

  private boolean checkExpMETextField() {
    Double ExpME = Utilities.readDouble(ExpMETextField.getText(),
                   "exponent from Kjeldsen et al. (2008), b,",-20.0,20.0,ExpMETextField);
    if (ExpME == null) {
      ExpMETextField.setText(String.format("%.4f",dataCenter.expME));
      paramModified = false;
      return false;
    } else { 
      ExpMETextField.setText(String.format("%.4f",ExpME));
      paramModified = (dataCenter.expME != ExpME.doubleValue());
      dataCenter.expME = ExpME.doubleValue();
      return true;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This initialises the output button which allows the user the write the
  // results to a file.
  /////////////////////////////////////////////////////////////////////////////
  private void initOutputButton() {
    outputButton = new JButton("Save results");
    outputButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent ev) {
        File aFile;
        int result;

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        result = fc.showSaveDialog(outputButton);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
           aFile = fc.getSelectedFile();

          // if file exists, check before overwriting:
          if (aFile.exists()) {
            result = JOptionPane.showConfirmDialog(outputButton,
                 "Are you sure you want to overwrite "+aFile.getName()+"?",
                 "File exists",JOptionPane.YES_NO_OPTION);
          } else {
            result = JOptionPane.YES_OPTION;
          }
        
          if(result == JOptionPane.YES_OPTION) {
            writeResults(aFile,outputButton);
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // This writes the results to the file "aFile".
  /////////////////////////////////////////////////////////////////////////////
  private void writeResults(File aFile, JComponent source) {
    try {
      PrintWriter out = new PrintWriter(new FileWriter(aFile));
      out.println("Delta(rho)/rho estimate");
      out.println();
      out.println("Parameters:");
      out.printf("Beta:              %19.12e%n",dataCenter.MEBeta);
      out.printf("Theta:             %19.12e%n",dataCenter.METheta);
      out.printf("N surf. correct.:   %d%n",dataCenter.n_surfME);
      out.printf("Min. n scaling:     %d%n",dataCenter.nminME);
      out.printf("Kjeldsen & al exp: %19.12e%n",dataCenter.expME);
      out.println();
      out.printf("rho_ref (cgs):     %19.12e%n",dataCenter.RhoRef/pi43);
      out.printf("rho_exact (cgs):   %19.12e%n",dataCenter.RhoRef*(1.0+my_estimate.trueMeanDensity)/pi43);
      out.println();
      out.println("Results:                          SOLA             Scaling            Kjeldsen et al. (2008)");
      out.printf("rho_inv (cgs):     %19.12e %19.12e %19.12e%n",dataCenter.RhoRef*my_estimate.estimatedMeanDensity/pi43,
                                                                dataCenter.RhoRef*my_estimate.estimatedMeanDensitySCALING/pi43,
                                                                dataCenter.RhoRef*my_estimate.estimatedMeanDensitySCALING2/pi43);
      out.printf("Sigma(rho) (cgs):  %19.12e %19.12e %19.12e%n",dataCenter.RhoRef*my_estimate.oneSigmaError/pi43,
                                                                dataCenter.RhoRef*my_estimate.oneSigmaErrorSCALING/pi43,
                                                                dataCenter.RhoRef*my_estimate.oneSigmaErrorSCALING2/pi43);
      out.printf("Delta rho/rho:     %19.12e %19.12e %19.12e%n",my_estimate.estimatedMeanDensity-1.0,
                                                                my_estimate.estimatedMeanDensitySCALING-1.0,
                                                                my_estimate.estimatedMeanDensitySCALING2-1.0);
      out.printf("||Delta K_avg||_2: %19.12e %19.12e %19.12e%n",my_estimate.averagingError,
                                                                my_estimate.averagingErrorSCALING,
                                                                my_estimate.averagingErrorSCALING2);
      out.printf("||K_cross||_2:     %19.12e %19.12e %19.12e%n",my_estimate.crossTermError,
                                                                my_estimate.crossTermErrorSCALING,
                                                                my_estimate.crossTermErrorSCALING2);
      out.printf("Int(D K_avg Df/f): %19.12e %19.12e %19.12e%n",my_estimate.trueAveragingError,
                                                                my_estimate.trueAveragingErrorSCALING,
                                                                my_estimate.trueAveragingErrorSCALING2);
      out.printf("Int(K_cross Dg/g): %19.12e %19.12e %19.12e%n",my_estimate.trueCrossTermError,
                                                                my_estimate.trueCrossTermErrorSCALING,
                                                                my_estimate.trueCrossTermErrorSCALING2);
      for(int i=0; i<my_estimate.n_surf; i++) out.printf("Surf[%d]:           %19.12e %19.12e %19.12e%n",i,
                                                                my_estimate.surfTerms[i],
                                                                my_estimate.surfTermsSCALING[i],
                                                                my_estimate.surfTermsSCALING2[i]);
      out.println("NaN = the result is not available");
      out.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        source,
        "Unable to write to file \""+aFile.getName()+"\"",
        "Failure",
        JOptionPane.ERROR_MESSAGE
      );
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // init buttons which deal with a target function
  /////////////////////////////////////////////////////////////////////////////
  private void initVariationsButtons() {
    LoadVariationsButton = new JButton("Load \u03B4s\u2081/s\u2081, \u03B4s\u2082/s\u2082");
    LoadVariationsButton.setEnabled(!dataCenter.isApplet);
    LoadVariationsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showOpenDialog(LoadVariationsButton);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            loadVariations(aFile);
            dataCenter.currentDirectory = aFile.getParentFile();
          } else {
            JOptionPane.showMessageDialog(
              LoadVariationsButton,
              "File \""+aFile.getName()+"\" doesn't exist.",
              "File not found",
              JOptionPane.WARNING_MESSAGE
            );
          }
        }
      }
    });
    
    ClearVariationsButton = new JButton("Clear \u03B4s\u2081/s\u2081, \u03B4s\u2082/s\u2082");
    ClearVariationsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.targetStruct1 = null;
        dataCenter.targetStruct2 = null;
        dataCenter.gridTargetStruct = null;
        my_estimate.trueMeanDensity = Double.NaN;
        my_estimate.trueAveragingError = Double.NaN;
        my_estimate.trueCrossTermError = Double.NaN;
        my_estimate.trueAveragingErrorSCALING = Double.NaN;
        my_estimate.trueCrossTermErrorSCALING = Double.NaN;
        my_estimate.trueAveragingErrorSCALING2 = Double.NaN;
        my_estimate.trueCrossTermErrorSCALING2 = Double.NaN;
        writeMeanDensityEstimate();
      }
    });
  }

  /***************************************************************************
  *  This updates the inversion.
  ***************************************************************************/
  public void updateInversion(JComponent c) {
    my_estimate.init(c);
    numberOfKernels.setText(String.format(" %d kernels",
                            my_estimate.n_inputKernels));
    // no need to give an error message if canEstimate = false, since
    // my_estimate.init(c) already gives error messages.
    if (my_estimate.canEstimate) {
      my_estimate.SOLAinversion1();
      my_estimate.applySCALING();
    }
  }

  /***************************************************************************
  *  This writes the result of the mean density estimate
  ***************************************************************************/
  public void writeMeanDensityEstimate() {
    String space1 = "";
    String space2 = "";
    String space3 = "";
    String rhobar = "<font style='text-decoration:underline;'>\u03C1</font>";
    for(int i=0; i<9; i++)  space1+="&nbsp;";
    for(int i=0; i<6; i++)  space2+="&nbsp;";
    for(int i=0; i<0; i++)  space3+="&nbsp;";
    String fmt = 
      "<html> <table border='1' bordercolor='black'>"+
      "<tr> <td> "+rhobar+"<sub>ref</sub> (g/cm\u00B3)</td> <td> %.3e </td>"+
      "     <td> "+rhobar+"<sub>obs</sub> (g/cm\u00B3)</td> <td> %.3e </td> </tr>"+
      "<tr> <td> </td>"+
      "<td> <font color='blue'>  SOLA           "+space1+" </font> </td>"+
      "<td> <font color='red'>   Scaling        "+space2+" </font> </td>"+
      "<td> <font color='green'> Kjeldsen et al."+space3+" </font> </td> </tr>"+
      "<tr> <td> "+rhobar+"<sub>inv</sub> (g/cm\u00B3)</td> <td> <font color='blue'> %.3e </font> </td>"+
      "<td> <font color='red'> %.3e </font> </td> <td> <font color='green'> %.3e </font> </td> </tr>"+
      "<tr> <td> \u03C3<sub>"+rhobar+"</sub> (g/cm\u00B3)</td> <td> <font color='blue'> %.3e </font> </td>"+
      "<td> <font color='red'> %.3e </font> </td> <td> <font color='green'> %.3e </font> </td> </tr>"+
      "<tr> <td> \u03B4"+rhobar+"/"+rhobar+" </td> <td> <font color='blue'> %.3e </font> </td>"+
      "<td> <font color='red'> %.3e </font> </td> <td> <font color='green'> %.3e </font> </td> </tr>"+
      "<tr> <td> \u03C3<sub>\u03B4"+rhobar+"/"+rhobar+"</sub> </td> <td> <font color='blue'> %.3e </font> </td>"+
      "<td> <font color='red'> %.3e </font> </td> <td> <font color='green'> %.3e </font> </td> </tr>"+
      "<tr> <td> \u2016\u0394K<sub>avg</sub>\u2016<sub>2</sub> </td> <td> <font color='blue'> %.3e </font> </td>"+
      "<td> <font color='red'> %.3e </font> </td> <td> <font color='green'> %.3e </font> </td> </tr>"+
      "<tr> <td> \u2016K<sub>cross</sub>\u2016<sub>2</sub> </td> <td> <font color='blue'> %.3e </font> </td>"+
      "<td> <font color='red'> %.3e </font> </td> <td> <font color='green'> %.3e </font> </td> </tr>"+
      "<tr> <td> \u0283\u0394K<sub>avg</sub>\u03B4\u03C1/\u03C1</td> <td> <font color='blue'> %.3e </font> </td>"+
      "<td> <font color='red'> %.3e </font> </td> <td> <font color='green'> %.3e </font> </td> </tr>"+
      "<tr> <td> \u0283K<sub>cross</sub>\u03B4g/g </td> <td> <font color='blue'> %.3e </font> </td>"+
      "<td> <font color='red'> %.3e </font> </td> <td> <font color='green'> %.3e </font> </td> </tr>";
    for(int i=0; i<MeanDensityEstimate.n_surf_disp; i++) {
      fmt+=String.format("<tr> <td> Surf[%d] </td> <td> <font color='blue'> %.3e </font> </td>"+
                         "<td> <font color='red'> %.3e </font> </td> <td> <font color='green'> %.3e </font> </td> </tr>",
                         i,my_estimate.surfTerms[i],my_estimate.surfTermsSCALING[i],my_estimate.surfTermsSCALING2[i]);
    }
    fmt+="</table> &nbsp;\"NaN\"=result not available </html>";
    resultLabel.setText(String.format(fmt,
      dataCenter.RhoRef/pi43,
      dataCenter.RhoRef*(1.0+my_estimate.trueMeanDensity)/pi43,
      dataCenter.RhoRef*my_estimate.estimatedMeanDensity/pi43,
      dataCenter.RhoRef*my_estimate.estimatedMeanDensitySCALING/pi43,
      dataCenter.RhoRef*my_estimate.estimatedMeanDensitySCALING2/pi43,
      dataCenter.RhoRef*my_estimate.oneSigmaError/pi43,
      dataCenter.RhoRef*my_estimate.oneSigmaErrorSCALING/pi43,
      dataCenter.RhoRef*my_estimate.oneSigmaErrorSCALING2/pi43,
      my_estimate.estimatedMeanDensity-1.0,
      my_estimate.estimatedMeanDensitySCALING-1.0, 
      my_estimate.estimatedMeanDensitySCALING2-1.0, 
      my_estimate.oneSigmaError,
      my_estimate.oneSigmaErrorSCALING,
      my_estimate.oneSigmaErrorSCALING2,
      my_estimate.averagingError,
      my_estimate.averagingErrorSCALING,
      my_estimate.averagingErrorSCALING2,
      my_estimate.crossTermError,
      my_estimate.crossTermErrorSCALING,
      my_estimate.crossTermErrorSCALING2,
      my_estimate.trueAveragingError,
      my_estimate.trueAveragingErrorSCALING,
      my_estimate.trueAveragingErrorSCALING2,
      my_estimate.trueCrossTermError,
      my_estimate.trueCrossTermErrorSCALING,
      my_estimate.trueCrossTermErrorSCALING2));
  }

  /************************************************************************
  * This initialises the program.
  ************************************************************************/
  public void initProgram() {
    graphAveragingKernel = new IPlot("Averaging kernel","Position, r/R", "Kernel",4);
    graphAveragingKernel.setIOenabled(!dataCenter.isApplet);
    graphCrossTermKernel = new IPlot("Cross-term kernel","Position, r/R", "Kernel",4);
    graphCrossTermKernel.setIOenabled(!dataCenter.isApplet);
    graphCrossTermKernel.replaceDataSet(0,null,null,Plot.dashed,Color.black);
    graphContainer = new JPanel();
    graphContainer.setLayout(new GridLayout(2,1));
    graphContainer.add(graphAveragingKernel);
    graphContainer.add(graphCrossTermKernel);

    setLayout(new BorderLayout());
    add("Center", graphContainer);

    // init the inversion
    my_estimate = new MeanDensityEstimate(dataCenter);

    calculateWhileSliding = new JCheckBox("Calculate while sliding",true);
    initVariationsButtons();
    initNoiseButtons();
    initSOLAsliders();
    initStructKernelButtons();
    initCheckBoxes();
    initNSurfPanel();
    initNminPanel();
    initExpMEPanel();
    initOutputButton();
    numberOfKernels = new JLabel(String.format(" %d kernels",
                                 my_estimate.n_inputKernels));
    resultLabel = new JLabel();
    writeMeanDensityEstimate();
    
    SliderContainer = new JPanel();
    SliderContainer.setLayout(new GridBagLayout());
    GridBagConstraints c = new GridBagConstraints();
    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridx = 0;
    c.gridy = 0;
    c.gridwidth = 1;
    SliderContainer.add(Refresh,c);
    c.gridx = 1;
    SliderContainer.add(LoadVariationsButton,c);
    c.gridx = 0;
    c.gridy++;
    SliderContainer.add(outputButton,c);
    c.gridx = 1;
    SliderContainer.add(ClearVariationsButton,c);
    c.gridx = 0;
    c.gridy++;
    SliderContainer.add(numberOfKernels,c);
    c.gridx = 1;
    SliderContainer.add(SOLACheckBox,c);
    c.gridx = 0;
    c.gridy++;
    SliderContainer.add(TargetCheckBox,c);
    c.gridx = 1;
    SliderContainer.add(ScalingCheckBox,c);
    c.gridx = 0;
    c.gridy++;
    SliderContainer.add(calculateWhileSliding,c);
    c.gridx = 1;
    SliderContainer.add(KjeldsenCheckBox,c);
    c.gridx = 0;
    c.gridy++;
    c.gridwidth = 2;
    SliderContainer.add(kernelTypeComboBox,c);
    c.gridx = 0;
    c.gridy++;
    c.gridwidth = 1;
    SliderContainer.add(NoiseCheckBox,c);
    c.gridx = 1;
    SliderContainer.add(NoiseTextField,c);
    c.gridx = 0;
    c.gridy++;
    c.gridwidth = 2;
    SliderContainer.add(NoiseSlider,c);
    c.gridy++;
    SliderContainer.add(NoiseRefresh,c);
    c.gridy++;
    SliderContainer.add(SOLAtitle,c);
    c.gridx = 0;
    c.gridy++;
    c.gridwidth = 1;
    SliderContainer.add(BetaLabel,c);
    c.gridx = 1;
    SliderContainer.add(BetaTextField,c);
    c.gridx = 0;
    c.gridy++;
    c.gridwidth = 2;
    SliderContainer.add(BetaSlider,c);
    c.gridx = 0;
    c.gridy++;
    c.gridwidth = 1;
    SliderContainer.add(ThetaLabel,c);
    c.gridx = 1;
    SliderContainer.add(ThetaTextField,c);
    c.gridx = 0;
    c.gridy++;
    c.gridwidth = 2;
    SliderContainer.add(ThetaSlider,c);
    c.gridx = 0;
    c.gridy++;
    c.gridwidth = 1;
    SliderContainer.add(NSurfLabel,c);
    c.gridx = 1;
    SliderContainer.add(NSurfTextField,c);
    c.gridx = 0;
    c.gridy++;
    SliderContainer.add(NminLabel,c);
    c.gridx = 1;
    SliderContainer.add(NminTextField,c);
    c.gridx = 0;
    c.gridy++;
    SliderContainer.add(ExpMELabel,c);
    c.gridx = 1;
    SliderContainer.add(ExpMETextField,c);
    c.gridx = 0;
    c.gridy++;
    c.gridwidth = 2;
    SliderContainer.add(resultLabel,c);
    add("East",SliderContainer);
       
    graphAveragingKernel.repaint();
  }

  /***************************************************************************
  *  Plot all kernels
  ***************************************************************************/
  private void replotAllKernels() {
    replotAveragingKernel();
    replotCrossTermKernel();
  }

  /***************************************************************************
  *  Plot averaging kernel
  ***************************************************************************/
  private void replotAveragingKernel() {
    if (dataCenter.grid == null) {
      graphAveragingKernel.nullify();
      return;
    }

    double[] y;

    if (TargetCheckBox.isSelected()) {
      y = new double[dataCenter.grid.length];
      for(int i=0;i<dataCenter.grid.length;i++)
        y[i] = 4.0*Math.PI*dataCenter.rho[i]*Math.pow(dataCenter.grid[i],2.0);
      graphAveragingKernel.replaceDataSet(0,dataCenter.grid,y,Plot.dashed,Color.black);
    } else {
      graphAveragingKernel.nullify(0);
    }

    if (ScalingCheckBox.isSelected()) {
      y = my_estimate.findAveragingKernel1SCALING();
      if (y == null) {
        graphAveragingKernel.nullify(1);
      } else {
        graphAveragingKernel.replaceDataSet(1,dataCenter.grid,y,Plot.solid,Color.red);
      }
    } else {
      graphAveragingKernel.nullify(1);
    }

    if (KjeldsenCheckBox.isSelected()) {
      y = my_estimate.findAveragingKernel2SCALING();
      if (y == null) {
        graphAveragingKernel.nullify(2);
      } else {
        graphAveragingKernel.replaceDataSet(2,dataCenter.grid,y,Plot.solid,DarkGreen);
      }
    } else {
      graphAveragingKernel.nullify(2);
    }

    if (SOLACheckBox.isSelected()) {
      y = my_estimate.findAveragingKernel1SOLA();
      if (y == null) {
        graphAveragingKernel.nullify(3);
      } else {
        graphAveragingKernel.replaceDataSet(3,dataCenter.grid,y,Plot.solid,Color.blue);
      }
    } else {
      graphAveragingKernel.nullify(3);
    }

    graphAveragingKernel.findRange();
    graphAveragingKernel.repaint();
  }

  /***************************************************************************
  *  Plot cross-term kernel
  ***************************************************************************/
  private void replotCrossTermKernel() {
    if (dataCenter.grid == null) {
      graphCrossTermKernel.nullify();
      return;
    }

    double[] y;

    if (TargetCheckBox.isSelected()) {
      double[] x = new double[2];
      y = new double[2];
      x[0] = dataCenter.grid[0];
      x[1] = dataCenter.grid[dataCenter.grid.length-1];
      y[0] = 0.0;
      y[1] = 0.0;
      graphCrossTermKernel.replaceDataSet(0,x,y,Plot.dashed,Color.black);
    } else {
      graphCrossTermKernel.nullify(0);
    }

    if (ScalingCheckBox.isSelected()) {
      y = my_estimate.findCrossTermKernel1SCALING();
      if (y == null) {
        graphCrossTermKernel.nullify(1);
      } else {
        graphCrossTermKernel.replaceDataSet(1,dataCenter.grid,y,Plot.solid,Color.red);
      }
    } else {
      graphCrossTermKernel.nullify(1);
    }

    if (KjeldsenCheckBox.isSelected()) {
      y = my_estimate.findCrossTermKernel2SCALING();
      if (y == null) {
        graphCrossTermKernel.nullify(2);
      } else {
        graphCrossTermKernel.replaceDataSet(2,dataCenter.grid,y,Plot.solid,DarkGreen);
      }
    } else {
      graphCrossTermKernel.nullify(2);
    }

    if (SOLACheckBox.isSelected()) {
      y = my_estimate.findCrossTermKernel1SOLA();
      if (y == null) {
        graphCrossTermKernel.nullify(3);
      } else {
        graphCrossTermKernel.replaceDataSet(3,dataCenter.grid,y,Plot.solid,Color.blue);
      }
    } else {
      graphCrossTermKernel.nullify(3);
    }

    graphCrossTermKernel.findRange();
    graphCrossTermKernel.repaint();
  }

  /////////////////////////////////////////////////////////////
  // load a file with a target structural variations function:
  /////////////////////////////////////////////////////////////
  private void loadVariations(File aFile){
    BufferedReader f;
    String line;
    Boolean skipLine = false;
    ArrayList<Double> r = new ArrayList<Double>();
    ArrayList<Double> s1 = new ArrayList<Double>();
    ArrayList<Double> s2 = new ArrayList<Double>();
    try {
      f = Utilities.OpenFile(aFile.getPath());
      while (true) {
        line = f.readLine();     // read a line of text
        if (line == null) break; // test for the end of the buffer
        line = Utilities.trim(line);       // remove comments
        StringTokenizer t = new StringTokenizer(line," \t,");
        // only read lines with EXACTLY 3 numbers
        if (t.countTokens() == 3) {
          r.add(Utilities.parseDouble(t.nextToken()));
          s1.add(Utilities.parseDouble(t.nextToken()));
          s2.add(Utilities.parseDouble(t.nextToken()));
        } else {
          if (t.countTokens() != 0) skipLine = true;
        }
      }
      f.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    LoadVariationsButton,  // the parent component
                    "Unable to read file \""+aFile.getName()+"\"",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    if (r.size() == 0) {
      JOptionPane.showMessageDialog(
                    LoadVariationsButton,  // the parent component
                    "No data in file",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    // perform a basic check on the grid:
    for(int i=1; i<r.size(); i++) {
      if (r.get(i).compareTo(r.get(i-1)) <= 0) {
        JOptionPane.showMessageDialog(
                    LoadVariationsButton,  // the parent component
                    "Please define the target function\n"+
                    "on a strictly ascending grid.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
        return;
      }
    }
    
    if (skipLine) {
      JOptionPane.showMessageDialog(
                    LoadVariationsButton,  // the parent component
                    "Some lines were skipped\nwhen reading the file.",
                    "Warning",
                    JOptionPane.WARNING_MESSAGE
                  );
    }

    // set arrays in data center:
    dataCenter.gridTargetStruct = new double[r.size()];
    dataCenter.targetStruct1 = new double[r.size()];
    dataCenter.targetStruct2 = new double[r.size()];
    for(int i=0;i<r.size();i++) {
      dataCenter.gridTargetStruct[i]  = r.get(i).doubleValue();
      dataCenter.targetStruct1[i] = s1.get(i).doubleValue();
      dataCenter.targetStruct2[i] = s2.get(i).doubleValue();
    }
    dataCenter.targetStructLoaded = true;
  }
}
