/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.Utilities;
import inversion_applet.Kernel;

/************************************************************************
* This class contains all observational values associated with each mode,
* as well as a link to a theoretical mode (Kernel).  By creating such a
* structure, it is more straightforward keeping track of observed modes
* which have been added or removed, due to user manipulations.
*************************************************************************
* NOTE: the use of wrapper classes such as Integer and Double allows us
*       to as assign the value NULL to these fields (this can be useful
*       when adding a new mode via the interactive "Add new line" button
*       on the frequency data page.
************************************************************************/
public class Mode {

  ///////////////////////////////////////////////////////////////////////
  // Identification:
  ///////////////////////////////////////////////////////////////////////
  /** Harmonic degree of the pulsation mode. */
  public Integer l;
  
  ///////////////////////////////////////////////////////////////////////
  // Observations:
  ///////////////////////////////////////////////////////////////////////
  /** Observed frequency. */
  public Double FreqObs;
  /** Error bar on observed frequency. */
  public Double ErrorFreq;
  /** Average rotational splitting.  For prograde modes, the frequency 
   *  is increased by |m|RS */
  public Double RS;
  /** Error on rotational splitting. */
  public Double ErrorRS;

  ///////////////////////////////////////////////////////////////////////
  // Link to theoretical results:
  ///////////////////////////////////////////////////////////////////////
  public Kernel kernel;

  ///////////////////////////////////////////////////////////////////////
  // Miscellaneous fields:
  ///////////////////////////////////////////////////////////////////////
  /** This keeps track of whether a mode is duplicated or not */
  public boolean isDuplicate;

  /**********************************************************************
  * A simple constructor.
  **********************************************************************/
  public Mode() {
    FreqObs = ErrorFreq = RS = ErrorRS = null;
    l = null;
    kernel = null;
    isDuplicate = false;
  }

  /**********************************************************************
  * A more elaborate constructor.
  **********************************************************************/
  public Mode(Integer l, Double FreqObs, Double ErrorFreq) {
    this.kernel = null;
    this.l = l;
    this.FreqObs = FreqObs;
    this.ErrorFreq = ErrorFreq;
    this.RS = this.ErrorRS = null;
    this.isDuplicate = false;
  }

  /**********************************************************************
  * The most elaborate constructor
  **********************************************************************/
  public Mode(Integer l, Double FreqObs, Double ErrorFreq, Double RS, Double ErrorRS) {
    this.kernel = null;
    this.l = l;
    this.FreqObs = FreqObs;
    this.ErrorFreq = ErrorFreq;
    this.RS = RS;
    this.ErrorRS = ErrorRS;
    this.isDuplicate = false;
  }

  /**********************************************************************
  * This is method returns "true" if Mode m1 comes before Mode
  * m2 in terms of their (l,FreqObs) values.
  **********************************************************************/
  public static boolean compareObs(Mode m1, Mode m2) {
    int compareL = Utilities.compareIntegers(m1.l,m2.l);
    if (compareL < 0) return true;
    if (compareL == 0) return (Utilities.compareDoubles(m1.FreqObs,m2.FreqObs) < 0);
    return false;
  }

  /**********************************************************************
  * This is method returns "true" if Mode m1 comes before Mode
  * m2 in terms of their (l,n) values, or (l,FreqObs) if n is not
  * available.
  **********************************************************************/
  public static boolean compareModes(Mode m1, Mode m2) {
    int compareL = Utilities.compareIntegers(m1.l,m2.l);
    if (compareL < 0) return true;
    if (compareL == 0) {
      if (m1.kernel == null) {
        if (m2.kernel == null) {
          return (Utilities.compareDoubles(m1.FreqObs,m2.FreqObs) < 0);
        } else {
          return true;
        }
      } else {
        if (m2.kernel == null) {
          return false;
        } else {
          return (Utilities.compareIntegers(m1.kernel.n,m2.kernel.n) < 0);
        }
      }
    } else {
      return false;
    }
  }
}
