/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import graph.IPlot;
import graph.Plot;
import inversion_applet.DataCenter;
import inversion_applet.Utilities;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.URL;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.zip.GZIPInputStream;
import javax.swing.ButtonGroup;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.SwingConstants;
import javax.swing.filechooser.FileFilter;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import org.fudaa.dodico.fortran.FortranBinaryInputStream;
import org.fudaa.dodico.fortran.FortranBinaryOutputStream;
import org.netlib.blas.Dgbmv;

public class ModelPage extends JPanel {
  private ArrayList<String> models = new ArrayList<String>();
  private ArrayList<Double> lum = new ArrayList<Double>();
  private ArrayList<Double> teff = new ArrayList<Double>();
  private DataCenter dataCenter;
  private JLabel NativeEndianLabel;
  private JLabel sTitle = new JLabel();
  private JLabel sMass = new JLabel();
  private JLabel sRadius = new JLabel();
  private JLabel sOmegaK = new JLabel();
  private JLabel nPoints = new JLabel();
  private JPanel buttons;
  private JTabbedPane tabbedPane;
  private JPanel graphArea;
  private JButton LoadModel;
  private JButton LoadInternetModel;
  private JButton LoadHELASmodel;
  private JButton DownloadHELASmodelButton;
  private JButton LoadSaveHELASmodelButton;
  private JButton HELAScancelButton;
  private JOptionPane HELASdialog;
  private JButton PolytropeButton;
  private JButton WriteModelButton;
  private JButton ClearModel;
  private JButton FindG;
  private JButton Rescale;
  private JButton RemeshButton;
  private JButton TweakButton;
  private JButton ReadListButton;
  private JCheckBox ShowDoublePoints;
  private JCheckBox ShowGridPoints;
  private JTree HELAStree;
  private JTextField HELASTextField;
  private IPlot DensityPlot;
  private IPlot DDensityPlot;
  private IPlot PressurePlot;
  private IPlot Gamma1Plot;
  private IPlot c2Plot;
  private IPlot dcPlot;
  private IPlot N2Plot;
  private IPlot APlot;
  private IPlot UPlot;
  private IPlot etaPlot;
  private IPlot VgPlot;
  private IPlot massPlot;
  private IPlot gPlot;
  private IPlot gpPlot;
  private IPlot diffRhoPlot;
  private IPlot freqPlot;
  private IPlot HRplot;

  // maximal number of grid points in eigenfunctions (this is useful
  // for testing an AMDL file to make sure it actually is in AMDL
  // format):
  private static final int maxNgrid= 300000;

  // order used when calculating the derivative of the density
  // and sound velocity profiles.  This also intervenes when testing
  // whether the models have enough points in each domain.
  private static final int orderDer = 2;

  // list of model formats which InversionKit is able to read:
  // IMPORTANT: 
  //    * if you add new formats, you also need to change the readModel(...) method.
  //    * binary formats must contain the word "endian" in the name
  //    * the keyword "native" is used to determine if the model is in the same
  //      "endianness" as your machine
  private static final String[] readModelFormats = new String[] {
     "AMDL (native endian)", "AMDL (opposite endian)", "FAMDL", "FGONG", "CESAM",
     "MOD (native endian)", "MOD (opposite endian)", "MOD (ascii)","LOSC","CLES"
  };
  // current format (so that InversionKit remembers the previous format used
  // to open a model)
  private String currentReadFormat = "AMDL (native endian)";

  // list of model formats in which InversionKit is able to save a model:
  // IMPORTANT: 
  //    * if you add new formats, you also need to change the writeModel(...) method.
  //    * binary formats must contain the word "endian" in the name
  //    * the keyword "native" is used to determine if the model is in the same
  //      "endianness" as your machine
  private static final String[] saveModelFormats = new String[] {
     "AMDL (native endian)", "AMDL (opposite endian)", "FAMDL",
     "MOD (native endian)", "MOD (opposite endian)", "MOD (ascii)"
  };
  // current format (so that InversionKit remembers the previous format used
  // to save a model)
  private String currentSaveFormat = "AMDL (native endian)";

  /////////////////////////////////////////////////////////////////////////////
  /** A constructor */
  /////////////////////////////////////////////////////////////////////////////
  public ModelPage(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This is to test this out as a stand alone application */
  /////////////////////////////////////////////////////////////////////////////
  public static void main(String[] argv) throws IOException {
    JFrame f = new JFrame("Model reader");
    DataCenter dataCenter = new DataCenter(false);
    ModelPage aModel = new ModelPage(dataCenter);
    aModel.initProgram();

    // add an exit button    
    JButton QuitButton = Utilities.createButton("Exit");
    aModel.buttons.add(QuitButton);
    QuitButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { System.exit(0); }
    });

    f.getContentPane().add("Center",aModel);
    f.setSize(220,200);
    f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    f.setVisible(true);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This is what sets up the buttons and the layout */
  /////////////////////////////////////////////////////////////////////////////
  public void initProgram() {
    initNativeEndianLabel();
    initLoadModel();
    initLoadInternetModel();
    initLoadHELASmodel();
    initPolytropeButton();
    initWriteModelButton();
    initClearModel();
    initFindG();
    initRescale();
    initRemeshButton();
    initTweakButtons();
    initReadListButton();
    initCheckBoxes();
    updateSLabels();
    DensityPlot = new IPlot("Density", "Radius, r/R", "Density (in g/cm\u00B3)");
    DensityPlot.setIDLTitles("Density", "Radius, r/R", "Density (in g/cm!E3!N)");
    DensityPlot.setIOenabled(!dataCenter.isApplet);
    DensityPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(DensityPlot));
    DDensityPlot = new IPlot("d(Density)/dr", "Radius, r/R", "d(Density)/dr (in g/cm\u2074)");
    DDensityPlot.setIDLTitles("d(Density)/dr", "Radius, r/R", "d(Density)/dr (in g/cm!E4!N)");
    DDensityPlot.setIOenabled(!dataCenter.isApplet);
    DDensityPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(DDensityPlot));
    PressurePlot = new IPlot("Pressure", "Radius, r/R", "Pressure (in dyn/cm\u00B2)");
    PressurePlot.setIDLTitles("Pressure", "Radius, r/R", "Pressure (in dyn/cm!E2!N)");
    PressurePlot.setIOenabled(!dataCenter.isApplet);
    PressurePlot.getGraphicsArea().addMouseListener(new ModelMouseListener(PressurePlot));
    Gamma1Plot = new IPlot("\u0393\u2081", "Radius, r/R", "\u0393\u2081");
    Gamma1Plot.setIDLTitles("!7C!17!D1!N", "Radius, r/R", "!7C!17!D1!N");
    Gamma1Plot.setIOenabled(!dataCenter.isApplet);
    Gamma1Plot.getGraphicsArea().addMouseListener(new ModelMouseListener(Gamma1Plot));
    c2Plot = new IPlot("c\u00B2", "Radius, r/R", "c\u00B2 (in cm\u00B2/s\u00B2)");
    c2Plot.setIDLTitles("c!E2!N", "Radius, r/R", "c!E2!N (in cm!E2!N/s!E2!N)");
    c2Plot.setIOenabled(!dataCenter.isApplet);
    c2Plot.getGraphicsArea().addMouseListener(new ModelMouseListener(c2Plot));
    dcPlot = new IPlot("(dc/dr)/r", "Radius, r/R", "(dc/dr)/r (in rad/s)");
    dcPlot.setIOenabled(!dataCenter.isApplet);
    dcPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(dcPlot));
    N2Plot = new IPlot("N\u00B2", "Radius, r/R", "N\u00B2 (in 1/s\u00B2)");
    N2Plot.setIDLTitles("N!E2!N", "Radius, r/R", "N!E2!N (in 1/s!E2!N)");
    N2Plot.setIOenabled(!dataCenter.isApplet);
    N2Plot.getGraphicsArea().addMouseListener(new ModelMouseListener(N2Plot));
    APlot = new IPlot("A", "Radius, r/R", "A");
    APlot.setIOenabled(!dataCenter.isApplet);
    APlot.getGraphicsArea().addMouseListener(new ModelMouseListener(APlot));
    UPlot = new IPlot("U", "Radius, r/R", "U");
    UPlot.setIOenabled(!dataCenter.isApplet);
    UPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(UPlot));
    VgPlot = new IPlot("Vg", "Radius, r/R", "Vg");
    VgPlot.setIOenabled(!dataCenter.isApplet);
    VgPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(VgPlot));
    etaPlot = new IPlot("\u03B7", "Radius, r/R", "\u03B7");
    etaPlot.setIDLTitles("!7g", "Radius, r/R", "!7g");
    etaPlot.setIOenabled(!dataCenter.isApplet);
    etaPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(etaPlot));
    massPlot = new IPlot("Mass", "Radius, r/R", "Mass (in g)");
    massPlot.setIOenabled(!dataCenter.isApplet);
    massPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(massPlot));
    gPlot = new IPlot("Gravity", "Radius, r/R", "Gravity (in cm/s\u00B2)");
    gPlot.setIDLTitles("Gravity", "Radius, r/R", "Gravity (in cm/s!E2!N)");
    gPlot.setIOenabled(!dataCenter.isApplet);
    gPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(gPlot));
    gpPlot = new IPlot("d(Gravity)/dr", "Radius, r/R", "d(Gravity)/dr (in cm/s\u00B3)");
    gpPlot.setIDLTitles("d(Gravity)/dr", "Radius, r/R", "d(Gravity)/dr (in cm/s!E3!N)");
    gpPlot.setIOenabled(!dataCenter.isApplet);
    gpPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(gpPlot));
    diffRhoPlot = new IPlot("(1-\u03C1/\u03C1\u2090)/x\u00B2", "Radius, r/R", "Profile");
    diffRhoPlot.setIDLTitles("!17(1-!7q/q!17!Da!N)/x!U2!N", "Radius, r/R", "Profile");
    diffRhoPlot.setIOenabled(!dataCenter.isApplet);
    diffRhoPlot.getGraphicsArea().addMouseListener(new ModelMouseListener(diffRhoPlot));
    freqPlot = new IPlot("N and S\u2097", "Radius, r/R", "Frequency, \u03BD (in \u03BCHz)");
    freqPlot.setIDLTitles("N and S!D!12l!17!N", "Radius, r/R", "Frequency, !7m!17 (in !7l!17Hz)");
    freqPlot.setIOenabled(!dataCenter.isApplet);
    HRplot = new IPlot("HR diagram","Temperature","L/L\u2609",2);
    HRplot.setYlog(true);
    HRplot.setReverseXrange(true);
    HRplot.setExtendXrange(true);
    HRplot.setExtendYrange(true);
    HRplot.setIOenabled(!dataCenter.isApplet);
    HRplot.getGraphicsArea().addMouseListener(new HRmouseListener());

    tabbedPane = new JTabbedPane();
    tabbedPane.addTab("\u03C1",DensityPlot);
    tabbedPane.addTab("d(\u03C1)/dr",DDensityPlot);
    tabbedPane.addTab("P", PressurePlot);
    tabbedPane.addTab("\u0393\u2081", Gamma1Plot);
    tabbedPane.addTab("c\u00B2", c2Plot);
    tabbedPane.addTab("(dc/dr)/r", dcPlot);
    tabbedPane.addTab("N\u00B2", N2Plot);
    tabbedPane.addTab("A", APlot);
    tabbedPane.addTab("<html>V<sub>g</sub></html>", VgPlot);
    tabbedPane.addTab("U", UPlot);
    tabbedPane.addTab("\u03B7", etaPlot);
    tabbedPane.addTab("mass", massPlot);
    tabbedPane.addTab("gravity", gPlot);
    tabbedPane.addTab("d(gravity)/dr", gpPlot);
    tabbedPane.addTab("(1-\u03C1/\u03C1\u2090)/x\u00B2", diffRhoPlot);
    tabbedPane.addTab("<html>N &amp; S<sub>&#8467;</sub></html>", freqPlot);
    tabbedPane.addTab("HR",HRplot);
    graphArea = new JPanel();
    graphArea.setLayout(new BorderLayout());
    graphArea.add("Center",tabbedPane);

    setLayout(new BorderLayout());
    buttons = new JPanel();
    buttons.setLayout(new GridBagLayout());
    GridBagConstraints c = new GridBagConstraints();
    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridx = 0;
    c.gridy = 0;
    c.gridwidth = 1;
    buttons.add(LoadModel,c);
    c.gridy++;
    buttons.add(LoadInternetModel,c);
    c.gridy++;
    buttons.add(LoadHELASmodel,c);
    c.gridy++;
    buttons.add(PolytropeButton,c);
    c.gridy++;
    buttons.add(WriteModelButton,c);
    c.gridy++;
    buttons.add(ClearModel,c);
    c.gridy++;
    buttons.add(ReadListButton,c);
    c.gridy++;
    buttons.add(new JLabel(" "),c);
    c.gridy++;
    buttons.add(FindG,c);
    c.gridy++;
    buttons.add(Rescale,c);
    c.gridy++;
    buttons.add(RemeshButton,c);
    c.gridy++;
    buttons.add(TweakButton,c);
    c.gridy++;
    buttons.add(NativeEndianLabel,c);
    c.gridy++;
    buttons.add(ShowDoublePoints,c);
    c.gridy++;
    buttons.add(ShowGridPoints,c);
    c.gridy++;
    buttons.add(new JLabel(" "),c);
    c.gridy++;
    buttons.add(sTitle,c);
    c.gridy++;
    buttons.add(sMass,c);
    c.gridy++;
    buttons.add(sRadius,c);
    c.gridy++;
    buttons.add(sOmegaK,c);
    c.gridy++;
    buttons.add(nPoints,c);
    
    add("Center",graphArea);
    add("East",buttons);
  }

  ////////////////////////////////////////////////////////////////////////////////////
  //  This initialises the native endian label
  ////////////////////////////////////////////////////////////////////////////////////
  private void initNativeEndianLabel() {
    NativeEndianLabel = new JLabel();
    if (ByteOrder.nativeOrder().equals(ByteOrder.BIG_ENDIAN)) {
      NativeEndianLabel.setText(" Native Endian = Big");
    } else {
      NativeEndianLabel.setText(" Native Endian = Little");
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////
  //  This initialises some check boxes
  ////////////////////////////////////////////////////////////////////////////////////
  private void initCheckBoxes() {
    ShowDoublePoints = new JCheckBox("show double points",true);
    ShowDoublePoints.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        refreshPlots(false);
      }
    });

    ShowGridPoints = new JCheckBox("show grid points",false);
    ShowGridPoints.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        refreshPlots(false);
      }
    });
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This updates the information given in the sTitle, sMass, sRadius and nPoints.
  ////////////////////////////////////////////////////////////////////////////////////
  private void updateSLabels() {
    if (dataCenter.modelLoaded) {
      sTitle.setText("<html><font color='blue'>&nbsp;The stellar parameters are:</font></html>");
      sMass.setText(String.format("   Mass:   %.3f M\u2299",
                      dataCenter.StellarMass/dataCenter.SolarMass));
      sRadius.setText(String.format("   Radius: %.3f R\u2299",
                      dataCenter.StellarRadius/dataCenter.SolarRadius));
      sOmegaK.setText(String.format("<html>&nbsp;&nbsp;&nbsp;\u03A9<sub>K</sub>: %.3e \u03BCHz",
                      dataCenter.FreqRef));
      nPoints.setText("   Number of points: "+dataCenter.grid.length);
    } else {
      sTitle.setText(" No models loaded");
      sTitle.setText("<html><font color='blue'>&nbsp;No models loaded</font></html>");
      sMass.setText(" ");
      sRadius.setText(" ");
      sOmegaK.setText(" ");
      nPoints.setText(" ");
    }
  }
      
  ////////////////////////////////////////////////////////////////////////////////////
  // this initialises PolytropeButton:
  ////////////////////////////////////////////////////////////////////////////////////
  private void initPolytropeButton() {
    PolytropeButton = Utilities.createButton("Generate Polytrope");
    PolytropeButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent ev) {
        Double N, Mass, Radius;
        Integer nstep;
        Object[] message = new Object[8];
        message[0] = "Enter the polytropic index N (0 \u2264 N \u2264 4.9):";
        message[1] = new JTextField("3.0");
        message[2] = "Enter the number of grid points n (10 \u2264 n \u2264 10000):";
        message[3] = new JTextField("1001");
        message[4] = "Enter the stellar mass M/M\u2299 (0.01 \u2264 M/M\u2299 \u2264 1000):";
        message[5] = new JTextField("1.0");
        message[6] = "Enter the stellar radius R/R\u2299 (0.01 \u2264 R/R\u2299 \u2264 1000): ";
        message[7] = new JTextField("1.0");

        String[] options = new String[2];
        options[0] = "Generate polytrope";
        options[1] = "Cancel";

        int result = JOptionPane.showOptionDialog(
          PolytropeButton,              // the parent that the dialog blocks
          message,                      // the dialog message array
          "Generate a polytrope",       // the title of the dialog window
          JOptionPane.DEFAULT_OPTION,   // option type
          JOptionPane.QUESTION_MESSAGE, // message type
          null,                         // optional icon, use null to use the default icon
          options,                      // options string array, will be made into buttons
          options[0]                    // option that should be made into a default button
        );
        
        switch(result) {
         case 1:  // Cancel
           break;

         default: // Generate polytrope

           // Check to see if different entries are acceptable:

           N     = Utilities.readDouble(((JTextField)message[1]).getText(),
                   "polytropic index",0.0,4.9,PolytropeButton);
           if (N == null) return;
           nstep = Utilities.readInt(((JTextField)message[3]).getText(),
                   "number of grid points",10,10000,PolytropeButton);
           if (nstep == null) return;
           Mass  = Utilities.readDouble(((JTextField)message[5]).getText(),
                   "stellar mass",0.01,1000,PolytropeButton);
           if (Mass == null) return;
           Radius= Utilities.readDouble(((JTextField)message[7]).getText(),
                   "stellar radius",0.01,1000,PolytropeButton);
           if (Radius == null) return;
           
           // set the stellar mass:
           dataCenter.StellarMass  = Mass*dataCenter.SolarMass;

           // set the stellar radius:
           dataCenter.StellarRadius= Radius*dataCenter.SolarRadius;

           // generate the polytrope:
           generatePolytrope(nstep,N);
           break;
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // this initialises the LoadModel button:
  /////////////////////////////////////////////////////////////////////////////
  private void initLoadModel() {
    LoadModel = Utilities.createButton("Load model");
    LoadModel.setEnabled(!dataCenter.isApplet);
    LoadModel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);
        fc.setAcceptAllFileFilterUsed(false);
        FileFilter currentFilter = null;
        for(int i=0; i<readModelFormats.length; i++) {
          FileFilter aFilter = new ModelFormatFilter(readModelFormats[i]);
          fc.addChoosableFileFilter(aFilter);
          if (readModelFormats[i].equals(currentReadFormat)) currentFilter = aFilter;
        }
        fc.setFileFilter(currentFilter);

        // show the filechooser
        int result = fc.showOpenDialog(LoadModel);  // put parent component
        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            currentReadFormat = fc.getFileFilter().getDescription();
            readModel(aFile.getPath(),currentReadFormat,false,LoadModel);
            dataCenter.currentDirectory = aFile.getParentFile();
          } else {
            JOptionPane.showMessageDialog(
              LoadModel,
              "File \""+aFile.getName()+"\" doesn't exist.",
              "File not found",
              JOptionPane.ERROR_MESSAGE
            );
          }
        }
      }
    });
  }
    
  /////////////////////////////////////////////////////////////////////////////
  // this initialises WriteModelButton:
  /////////////////////////////////////////////////////////////////////////////
  private void initWriteModelButton() {
    WriteModelButton = Utilities.createButton("Save model");
    WriteModelButton.setEnabled(!dataCenter.isApplet);
    WriteModelButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // check to see if a model is loaded
        if (!dataCenter.modelLoaded) {
          JOptionPane.showMessageDialog(
            WriteModelButton,  // the parent component
            "Unable to write model because no\n model is currently loaded.",
            "No model",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);
        fc.setAcceptAllFileFilterUsed(false);
        FileFilter currentFilter = null;
        for(int i=0; i<saveModelFormats.length; i++) {
          FileFilter aFilter = new ModelFormatFilter(saveModelFormats[i]);
          fc.addChoosableFileFilter(aFilter);
          if (saveModelFormats[i].equals(currentSaveFormat)) currentFilter = aFilter;
        }
        fc.setFileFilter(currentFilter);

        // show the filechooser
        int result = fc.showSaveDialog(WriteModelButton);  // put parent component

        // get filename
        if(result != JFileChooser.APPROVE_OPTION) return;

        File aFile = fc.getSelectedFile();

        // check to see if file exists:
        if (aFile.exists()) {
          result = JOptionPane.showConfirmDialog(
               WriteModelButton,
               "Are you sure you want to overwrite "+aFile.getName()+"?",
               "File exists",JOptionPane.YES_NO_OPTION);
        } else {
          result = JOptionPane.YES_OPTION;
        }

        if(result == JOptionPane.YES_OPTION) {
          currentSaveFormat = fc.getFileFilter().getDescription();
          writeModel(aFile,currentSaveFormat,WriteModelButton);
        }
        if (aFile.exists()) dataCenter.currentDirectory = aFile.getParentFile();
      }
    });
  }
    
  /////////////////////////////////////////////////////////////////////////////
  // this initialises the LoadInternetModel button
  /////////////////////////////////////////////////////////////////////////////
  private void initLoadInternetModel() {
    LoadInternetModel = Utilities.createButton("Load model from internet");
    LoadInternetModel.setEnabled(!dataCenter.isApplet);
    LoadInternetModel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        Object[] message = new Object[6];
        message[0] = "Please enter the URL of the model";
        message[1] = "you would like to download:";
        message[2] = new JTextField();
        message[3] = " ";
        message[4] = "Model format:";
        message[5] = new JComboBox(readModelFormats);
        ((JComboBox)message[5]).setSelectedItem(currentReadFormat);
      
        Object[] options = new Object[2];
        options[0] = "Load";
        options[1] = "Cancel";

        JOptionPane optionPane = new JOptionPane(
          message,                      // the dialog message array
          JOptionPane.QUESTION_MESSAGE, // message type
          JOptionPane.DEFAULT_OPTION,   // option type
          null,                         // optional icon, use null to use the default icon
          options,                      // options string array, will be made into buttons
          options[0]                    // default option
        );

        JDialog dialog = optionPane.createDialog(LoadInternetModel, "Download model");
        dialog.setSize(400,250);
        dialog.setResizable(true);
        dialog.setVisible(true);
        
        Object selectedValue = optionPane.getValue();
        if (selectedValue == options[0]) {
          String urlAddress = ((JTextField)message[2]).getText();
          currentReadFormat =  (String)((JComboBox)message[5]).getSelectedItem();
          if (urlAddress != null) readModel(urlAddress,currentReadFormat,true,LoadInternetModel);
        }
      }
    });
  }
  
  /////////////////////////////////////////////////////////////////////////////
  // this initialises the LoadHELASmodel button
  /////////////////////////////////////////////////////////////////////////////
  private void initLoadHELASmodel() {
    // Initialise some buttons and a text field:
    LoadHELASmodel = Utilities.createButton("Load HELAS model");
    LoadHELASmodel.setEnabled(!dataCenter.isApplet);
    DownloadHELASmodelButton = new JButton("Load model");
    DownloadHELASmodelButton.setEnabled(false);
    LoadSaveHELASmodelButton = new JButton("Load and save model");
    LoadSaveHELASmodelButton.setEnabled(false);
    HELAScancelButton = new JButton("Cancel");
    HELASTextField = new JTextField();
    HELASTextField.setEditable(false);

    // initialise tree with list of models and assocated scroll pane:
    initHELAStree();
    JScrollPane aScroller = new JScrollPane();
    aScroller.getViewport().add(HELAStree);

    // initialise dialog which allows the user to select and download
    // a helas model:
    Object[] message = new Object[6];
    message[0] = "Please select one of the following models:";
    message[1] = "(hint: using the arrow keys is an\n"+
                 "efficient way to navigate the tree)";
    message[2] = aScroller;
    message[3] = "Selected model:";
    message[4] = HELASTextField;
    message[5] = "<html>NOTE: these models are described in <br/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"+
    "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href='http://ukads.nottingham.ac.uk/abs/2008Ap%26SS.316..173M'>"+
    "Marques et al. (2008, ApSS 316, 173-178)</a></html>";

    Object[] options = new Object[3];
    options[0] = DownloadHELASmodelButton;
    options[1] = LoadSaveHELASmodelButton;
    options[2] = HELAScancelButton;
    HELASdialog = new JOptionPane(
      message,                      // the dialog message array
      JOptionPane.QUESTION_MESSAGE, // message type
      JOptionPane.DEFAULT_OPTION,   // option type
      null,                         // optional icon, use null to use the default icon
      options,                      // options string array, will be made into buttons
      options[0]                    // default option
    );
    
    // add some action listeners to different buttons:    
    DownloadHELASmodelButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        readModel(HELASTextField.getText(),"CESAM",true,DownloadHELASmodelButton);
        HELASdialog.setValue(0);
      }
    });

    // add some action listeners to different buttons:    
    LoadSaveHELASmodelButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showSaveDialog(LoadSaveHELASmodelButton);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            result = JOptionPane.showConfirmDialog(
                 LoadSaveHELASmodelButton,
                 "Are you sure you want to overwrite "+aFile.getName()+"?",
                 "File exists",JOptionPane.YES_NO_OPTION);
          } else {
            result = JOptionPane.YES_OPTION;
          }

          if(result == JOptionPane.YES_OPTION)
            downloadModel(HELASTextField.getText(),aFile.getPath());

          if (aFile.exists()) {
            dataCenter.currentDirectory = aFile.getParentFile();
            readModel(aFile.getPath(),"CESAM",false,LoadSaveHELASmodelButton);
            HELASdialog.setValue(1);
          }
        }
      }
    });

    HELAScancelButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        HELASdialog.setValue(2);
      }
    });

    LoadHELASmodel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        JDialog dialog = HELASdialog.createDialog(LoadHELASmodel, "Download a HELAS model");
        dialog.setSize(500,500);
        dialog.setResizable(true);
        dialog.setVisible(true);
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // This downloads a file from the HELAS platform and saves it on the hard
  // disk.  If the filename doesn't end with ".gz", it is decompressed
  // beforehand.
  /////////////////////////////////////////////////////////////////////////////
  private void downloadModel(String url, String filename) {
    try {
      if (filename.endsWith(".gz")) {
        // just download
        InputStream in = new URL(url).openStream();
        OutputStream out = new FileOutputStream(filename);
        // transfer bytes from in to out:
        byte[] buf = new byte[1024];
        int len;
        while ((len = in.read(buf)) > 0) out.write(buf, 0, len);
        in.close();
        out.close();
      } else {
        // download and decompress
        InputStream aux1 = new URL(url).openStream();
        GZIPInputStream in = new GZIPInputStream(aux1);
        OutputStream out = new FileOutputStream(filename);
        // transfer bytes from in to out:
        byte[] buf = new byte[1024];
        int len;
        while ((len = in.read(buf)) > 0) out.write(buf, 0, len);
        in.close();
        out.close();
      }
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
               LoadSaveHELASmodelButton,  // the parent component
               "Unable to download and/or save HELAS model.",
               "Error",
               JOptionPane.ERROR_MESSAGE
             );
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // this initialises the ClearModel button
  /////////////////////////////////////////////////////////////////////////////
  private void initClearModel() {
    ClearModel = Utilities.createButton("Clear model");
    ClearModel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.modelLoaded = false;
        dataCenter.rotationInversionUpToDate = false;
        dataCenter.structInversionUpToDate = false;
        dataCenter.meanDensityEstimateUpToDate = false;
        dataCenter.acousticTimeEstimateUpToDate = false;
        dataCenter.smallSeparationEstimateUpToDate = false;

        dataCenter.StellarRadius = 0.0;
        dataCenter.StellarMass = 0.0;
        dataCenter.Pref = 0.0;
        dataCenter.RhoRef = 0.0;

        dataCenter.grid = null;
        dataCenter.Gamma1 = null;
        dataCenter.rho = null;
        dataCenter.pressure = null;
        dataCenter.c2 = null;
        dataCenter.N2 = null;
        dataCenter.mass = null;
        dataCenter.gravity = null;
        dataCenter.dgravity = null;
        dataCenter.drho = null;

        dataCenter.clearKernels();
        updateSLabels();
        refreshPlots(true);
      }
    });
  }
   
  /////////////////////////////////////////////////////////////////////////////
  // this initialises the tweak button
  /////////////////////////////////////////////////////////////////////////////
  private void initTweakButtons() {
    TweakButton = Utilities.createButton("Recalculate model");
    TweakButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        tweakModel();
        dataCenter.clearKernels();
        updateSLabels();
        refreshPlots(true);
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // this initialises the read list button
  /////////////////////////////////////////////////////////////////////////////
  private void initReadListButton() {
    ReadListButton = Utilities.createButton("Read list");
    ReadListButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showOpenDialog(ReadListButton);  // put parent component
        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            readList(aFile.getPath());
            dataCenter.currentDirectory = aFile.getParentFile();
          } else {
            JOptionPane.showMessageDialog(
              LoadModel,
              "File \""+aFile.getName()+"\" doesn't exist.",
              "File not found",
              JOptionPane.ERROR_MESSAGE
            );
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // This reads a list of models
  /////////////////////////////////////////////////////////////////////////////
  private void readList(String aFile) {
    BufferedReader list;
    String line;
    StringTokenizer t;

    models.clear();
    teff.clear();
    lum.clear();
    try {
      list = Utilities.OpenFile(aFile);
      while(true) {
        line = list.readLine();
        if (line == null) break;
        line = Utilities.trim(line);
        t = new StringTokenizer(line," \t");
        if (t.countTokens() >= 3) { 
          models.add(t.nextToken());
          teff.add(Utilities.parseDouble(t.nextToken()));
          lum.add(Utilities.parseDouble(t.nextToken())/DataCenter.Lsun);
        }
      }
      list.close();
    } catch (Exception e) {
    
      System.out.println("Disaster!!!");
      System.out.println(e);
    }
    int n = lum.size(),i;
    double[] x = new double[n];
    double[] y = new double[n];
    for(i=0; i<n; i++) {
      x[i] = teff.get(i);
      y[i] = lum.get(i);
    }
    HRplot.replaceDataSet(0,x,y,Plot.noLine,Plot.plus);
    HRplot.nullify(1);
    HRplot.findRange();
    HRplot.repaint();
  }

  /////////////////////////////////////////////////////////////////////////////
  // this initialises the FindG button
  /////////////////////////////////////////////////////////////////////////////
  private void initFindG() {
    FindG = Utilities.createButton("Find G from model");
    FindG.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // first test to see if a model is loaded:
        if (!dataCenter.modelLoaded) {
          JOptionPane.showMessageDialog(
            FindG,  // the parent component
            "Please load model first.",
            "Error",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }

        // find new value for G based on hydrostatic equilibrium:
        double NewG = findGravitationalConstant();
        
        int result = JOptionPane.showConfirmDialog(
                 FindG, "<html>Current value of G:<br />"+
                 String.format("<h2>&nbsp;&nbsp;&nbsp;&nbsp;<font color='red'>%10.8f\u00D710\u207B\u2078 (cgs)</font></h2>",dataCenter.G*1e8)+
                 "Optimal value of G (based on hydrostatic equilibrium):"+
                 String.format("<h2>&nbsp;&nbsp;&nbsp;&nbsp;<font color='blue'>%10.8f\u00D710\u207B\u2078 (cgs)</font></h2>",NewG*1e8)+
                 "Should I update G?</html>", 
                 "The gravitational constant",JOptionPane.YES_NO_OPTION);

        if(result == JOptionPane.YES_OPTION) {
          dataCenter.modifyG(NewG);
          updateSLabels();
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // this initialises the Rescale button
  /////////////////////////////////////////////////////////////////////////////
  private void initRescale() {
    Rescale = Utilities.createButton("Rescale model");
    Rescale.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double MRatio, RRatio;
	final Object[] message = new Object[5];
        message[0] = "<html>Enter mass ratio, M<sub>new</sub>/M<sub>current</sub>:";
        message[1] = new JTextField();
        message[2] = "<html>Enter ratio of radii, R<sub>new</sub>/R<sub>current</sub>:";
        message[3] = new JTextField();
        message[4] = new JButton("Find values based on \u0394\u03BD");
        ((JButton)message[4]).addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            dataCenter.findDFsimple();
            double mratio = Math.pow(dataCenter.DF/dataCenter.SolarDF,2.0)
    	     * Math.pow(dataCenter.StellarRadius/dataCenter.SolarRadius,3.0)
    	     * dataCenter.SolarMass/dataCenter.StellarMass;
    	    ((JTextField)message[1]).setText(String.format("%.15e",mratio));
    	    ((JTextField)message[3]).setText(String.format("%.15e",1.0));
          }
        });

        String[] options = new String[2];
        options[0] = "Rescale model";
        options[1] = "Cancel";

        int result = JOptionPane.showOptionDialog(
          Rescale,                      // the parent that the dialog blocks
          message,                      // the dialog message array
          "Rescale model",              // the title of the dialog window
          JOptionPane.DEFAULT_OPTION,   // option type
          JOptionPane.QUESTION_MESSAGE, // message type
          null,                         // optional icon, use null to use the default icon
          options,                      // options string array, will be made into buttons
          options[0]                    // option that should be made into a default button
        );
        
        switch(result) {
         case 1:  // Cancel
           break;

         default:

           // Check to see if different entries are acceptable:
           MRatio = Utilities.readDouble(((JTextField)message[1]).getText(),
                   "mass ratio",1e-3,1e3,Rescale);
           if (MRatio == null) MRatio = new Double(1.0);
           RRatio = Utilities.readDouble(((JTextField)message[3]).getText(),
                   "ratio of radii",1e-5,1e5,Rescale);
           if (RRatio == null) RRatio = new Double(1.0);

           // Rescale model and update page
           dataCenter.rescale(MRatio,RRatio);
           dataCenter.clearKernels();
           updateSLabels();
           refreshPlots(true);
           break;
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // this initialises RemeshButton
  /////////////////////////////////////////////////////////////////////////////
  private void initRemeshButton() {
    RemeshButton = Utilities.createButton("Remesh model");
    RemeshButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // first test to see if a model is loaded:
        if (!dataCenter.modelLoaded) {
          JOptionPane.showMessageDialog(
            RemeshButton,  // the parent component
            "Please load model first.",
            "Error",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }

        final JTextField c1TextField = new JTextField("10.0");
        final JTextField c2TextField = new JTextField("0.01");
        final JTextField c3TextField = new JTextField("0.015");
        JLabel c1Label = new JLabel("<html>c<sub>1</sub> = </html>",SwingConstants.RIGHT);
        JLabel c2Label = new JLabel("<html>c<sub>2</sub> = </html>",SwingConstants.RIGHT);
        JLabel c3Label = new JLabel("<html>c<sub>3</sub> = </html>",SwingConstants.RIGHT);
        JPanel c1Panel = new JPanel(new GridLayout(1,2));
        JPanel c2Panel = new JPanel(new GridLayout(1,2));
        JPanel c3Panel = new JPanel(new GridLayout(1,2));
        c1Panel.add(c1Label);
        c1Panel.add(c1TextField);
        c2Panel.add(c2Label);
        c2Panel.add(c2TextField);
        c3Panel.add(c3Label);
        c3Panel.add(c3TextField);
        JLabel ngridLabel = new JLabel("<html>N<sub>mesh</sub> = </html>",SwingConstants.RIGHT);
        final JTextField ngridTextField = new JTextField("2400");
        JPanel ngridPanel = new JPanel(new GridLayout(1,2));
        ngridPanel.add(ngridLabel);
        ngridPanel.add(ngridTextField);

        final Object[] message = new Object[13];
        message[0] = "<html>Remesh stellar model according to even<br />increments of the following function:</html>";
        message[1] = new JLabel(new ImageIcon(getClass().getResource("/images/formula.png")));
        message[2] = "<html>(see Christensen-Dalsgaard, 2003, <a href='http://astro.phys.au.dk/~jcd/oscilnotes/'>Lecture Notes on Stellar Oscillations</a>, p. 124)</html>";
        message[3] = ngridPanel;
        message[4] = c1Panel;
        message[5] = c2Panel;
        message[6] = c3Panel;
        message[7] = new JRadioButton("Combine grids",false);
        message[8] = new JButton("<html>p-mode grid (c<sub>1</sub> = 10.0, c<sub>2</sub>=0.01, c<sub>3</sub>=0.015)</html>");
        message[9] = new JButton("<html>g-mode grid (c<sub>1</sub> = 0.025, c<sub>2</sub>=0.1, c<sub>3</sub>=0.0001)</html>");
        message[10] = new JButton("<html>p/g-mode grid (c<sub>1</sub> = 0.5, c<sub>2</sub>=0.03, c<sub>3</sub>=0.001)</html>");
        message[11] = " ";
        message[12] = new JButton("Preview");
        ((JButton)message[12]).addActionListener(new ActionListener(){
          public void actionPerformed(ActionEvent ev) {
            Double c1 = Utilities.readDouble(c1TextField.getText(),
                    "value of c1",0.0,1e9,c1TextField);
            if (c1 == null) return;
            Double c2 = Utilities.readDouble(c2TextField.getText(),
                    "value of c2",0.0,1e9,c2TextField);
            if (c2 == null) return;
            Double c3 = Utilities.readDouble(c3TextField.getText(),
                    "value of c3",0.0,1e9,c3TextField);
            if (c3 == null) return;
            Integer ngridNew = Utilities.readInt(ngridTextField.getText(),
                    "number of grid points",2,maxNgrid,ngridTextField);
            if (ngridNew == null) return;

            double[] newGrid =  findRadialGrid(ngridNew.intValue(), c1.doubleValue(),
                                 c2.doubleValue(), c3.doubleValue(),
                                 ((JRadioButton)message[7]).isSelected());
            IPlot aPlot = new IPlot("Grid","Grid point number","Radius, r/R", 2);
            double[] x1 = new double[dataCenter.grid.length];
            for(int i=0; i<dataCenter.grid.length; i++) x1[i] = (double)i;
            double[] x2 = new double[newGrid.length];
            for(int i=0; i<newGrid.length; i++) x2[i] = (double)i;
            aPlot.replaceDataSet(0, x1, dataCenter.grid, Plot.solid, Plot.noSymbol, Color.red);
            aPlot.replaceDataSet(1, x2, newGrid, Plot.solid, Plot.noSymbol, Color.blue);
            aPlot.findRange();
            JPanel aPanel = new JPanel(new BorderLayout());
            aPanel.add("Center",aPlot);
            aPanel.add("South",new JLabel("<html><h2><font color='red'>\u2015 Old grid</font><br /><font color='blue'>\u2015 New grid</font></h2></html>",SwingConstants.CENTER));
            aPanel.setPreferredSize(new Dimension(500,500));
            JOptionPane.showMessageDialog(
              (Component)message[12],  // the parent component
              aPanel,
              "Old versus new grid",
              JOptionPane.INFORMATION_MESSAGE
            );
          }
        });



        // The following constants come from JCD's course, p. 124.
        // Thanks to Remi and Jerome for pointing out this method.
        ActionListener l = new ActionListener(){
          public void actionPerformed(ActionEvent ev) {
            JButton source = (JButton)ev.getSource();
            if (source.equals(message[8])) {
              c1TextField.setText("10.0"); 
              c2TextField.setText("0.01"); 
              c3TextField.setText("0.015"); 
            } else if (source.equals(message[9])) {
              c1TextField.setText("0.025"); 
              c2TextField.setText("0.1"); 
              c3TextField.setText("0.0001"); 
            } else if (source.equals(message[10])) {
              c1TextField.setText("0.5"); 
              c2TextField.setText("0.03"); 
              c3TextField.setText("0.001"); 
            }
          }
        };
        ((JButton)message[8]).addActionListener(l);
        ((JButton)message[9]).addActionListener(l);
        ((JButton)message[10]).addActionListener(l);

        String[] options = new String[2];
        options[0] = "Remesh model";
        options[1] = "Cancel";

        int result = JOptionPane.showOptionDialog(
          RemeshButton,                 // the parent that the dialog blocks
          message,                      // the dialog message array
          "Remesh model",               // the title of the dialog window
          JOptionPane.DEFAULT_OPTION,   // option type
          JOptionPane.QUESTION_MESSAGE, // message type
          null,                         // optional icon, use null to use the default icon
          options,                      // options string array, will be made into buttons
          options[0]                    // option that should be made into a default button
        );
        
        switch(result) {
          case 1:  // Cancel
            break;

          default: // remesh model:

            // Check to see if different entries are acceptable:
            Double c1 = Utilities.readDouble(c1TextField.getText(),
                    "value of c1",0.0,1e9,c1TextField);
            if (c1 == null) break;
            Double c2 = Utilities.readDouble(c2TextField.getText(),
                    "value of c2",0.0,1e9,c2TextField);
            if (c2 == null) break;
            Double c3 = Utilities.readDouble(c3TextField.getText(),
                    "value of c3",0.0,1e9,c3TextField);
            if (c3 == null) break;
            Integer ngridNew = Utilities.readInt(ngridTextField.getText(),
                    "number of grid points",2,maxNgrid,ngridTextField);
            if (ngridNew == null) break;

            // Calculate new grid:
            double[] newGrid =  findRadialGrid(ngridNew.intValue(), c1.doubleValue(),
                                c2.doubleValue(), c3.doubleValue(),
                                 ((JRadioButton)message[7]).isSelected());

            // Interpolate model:
            interpolateModel(newGrid);
            dataCenter.clearKernels();
            updateSLabels();
            refreshPlots(true);
            break;
        }
      }
    });
  }

  /**********************************************************
  * This subroutine takes in a a file "aFile" and determines
  * which format it's in and whether or not it should be
  * decompressed.  It then call up the appropriate method
  * to read it (depending on the format).
  **********************************************************/ 
  private void readModel(String fileOrURL, String modelType, boolean fromInternet, Component source) {
    BufferedReader modelInput;
    FortranBinaryInputStream binaryInput;
 
    try {

      // treat binary case first since the model
      // file cannot be opened in the same way.
      if (modelType.contains("endian")) {
        InputStream in = null;
        if (fromInternet) {
          in = new URL(fileOrURL).openStream();
        } else {
          in = new FileInputStream(new File(fileOrURL));
        }
        boolean isNative = modelType.contains("native");
        binaryInput = new FortranBinaryInputStream(in,true,isNative);
        
        if (modelType.startsWith("AMDL")) {
          readAMDLmodel(binaryInput);
        } else if (modelType.startsWith("MOD")) {
          readBinaryMODmodel(binaryInput);
        } else {
          JOptionPane.showMessageDialog(
            source,  // the parent component
            "Unknown or unimplemented format: "+modelType+".\n"+
            "Please contact authors of InversionKit.",
            "Format error",
            JOptionPane.ERROR_MESSAGE
          );
        }
        dataCenter.clearKernels();
        return;
      }

      // now for non-binary files:
      if (fromInternet) {
        modelInput = Utilities.OpenURL(fileOrURL);
      } else {
        modelInput = Utilities.OpenFile(fileOrURL);
      }

      if (modelType.equals("CESAM")) {
        // This next part determines whether the model is in
        // "CESAM2k" or "CESAM" format.

        // The argument is the limit on the number of characters
        // that may be read while still preserving the mark.
        // Putting a large number may require increasing the size
        // of the buffer.
        modelInput.mark(500);

        String line;
        line = modelInput.readLine();
        line = modelInput.readLine();
        modelInput.reset();
        if (line.indexOf("CESAM2k")>=0) {
          readOSCmodel(modelInput,1);
        } else {
          readOSCmodel(modelInput,0);
        }
        dataCenter.clearKernels();
        return;
      }

      if (modelType.equals("MOD (ascii)")) {
        readAsciiMODmodel(modelInput);
        dataCenter.clearKernels();
        return;
      }

      if (modelType.equals("FAMDL")) {
        readFAMDLmodel(modelInput);
        dataCenter.clearKernels();
        return;
      }

      if (modelType.equals("FGONG")) {
        readFGONGmodel(modelInput);
        dataCenter.clearKernels();
        return;
      }

      if (modelType.equals("LOSC")) {
        readLOSCmodel(modelInput);
        dataCenter.clearKernels();
        return;
      }

      if (modelType.equals("CLES")) {
        readCLESmodel(modelInput, fileOrURL);
        dataCenter.clearKernels();
        return;
      }

      JOptionPane.showMessageDialog(
        source,  // the parent component
        "Unknown or unimplemented format: "+modelType+".\n"+
        "Please contact authors of InversionKit.",
        "Format error",
        JOptionPane.ERROR_MESSAGE
      );

    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        source,  // the parent component
        "Unable to read: "+fileOrURL+"\n",
        "Error opening file",
        JOptionPane.ERROR_MESSAGE
      );
    }
  }    

  /**********************************************************
   * This reads an AMDL binary model.
   **********************************************************
   * @param modelInput input stream which contains the amdl
   *                   model in fortran binary format
   *********************************************************/
  public void readAMDLmodel(FortranBinaryInputStream modelInput) {
    int i, j;
    int nmod, nr, ivar;
    double[] xdatain = new double[9];
    double[][] xin;

    // read the model
    try {
      modelInput.readRecord();
      nmod = modelInput.readInteger();
      nr = modelInput.readInteger();
      if ((nr<0)||(nr>maxNgrid)) {
        JOptionPane.showMessageDialog(
          LoadModel,
          "The number of grid points for the\n"+
          "model seems to be too large, or the\n"+
          "format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return;
      }

      for(i=1; i<=8; i++) xdatain[i] = modelInput.readDoublePrecision();

      // test the flag to see how many structural functions there are:
      int idata8 = (int)Math.round(xdatain[7]);
      if (idata8 >= 100) {
        ivar = 8;
      } else {
        if (idata8 >= 10) {
          ivar = 6;
        } else {
          ivar = 5;
        }
      }

      xin = new double[ivar+1][nr];
      for(i=0; i<nr; i++) {
        for(j=0; j<=ivar; j++) {
          xin[j][i] = modelInput.readDoublePrecision();
        }
      }
      modelInput.close();
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "File format not recognised",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    // check for multiple points and non-monotonic grid:
    ArrayList<Integer> mpoints = Numerics.findMpoints(xin[0]);
    if (mpoints == null) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "Grid in reverse order or non-monotonic",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (Numerics.findMinDomainSize(mpoints,nr) < orderDer) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "This model contains multiple points which are too\n"+
        "close to allow an accurate calculation of d\u03C1/dr.\n"+
        "Try increasing the number of significant digits\n"+
        "on your radial coordinate.", 
        "Error in model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (mpoints.size() > 0) System.out.println("Removing multiple points from model");


    // Stellar mass in g:
    dataCenter.StellarMass = xdatain[1];
    
    // Stellar radius in cm:
    dataCenter.StellarRadius = xdatain[2];

    findReferenceQuantities();

    // second derivative of density in center:
    dataCenter.ddrho  = -xdatain[4]*xdatain[6]/dataCenter.RhoRef; 

    // find JCD variables:
    dataCenter.grid   = Numerics.removeMpoints(mpoints,xin[0]);
    dataCenter.eta    = Numerics.removeMpoints(mpoints,xin[1]);
    dataCenter.Vg     = Numerics.removeMpoints(mpoints,xin[2]);
    dataCenter.Gamma1 = Numerics.removeMpoints(mpoints,xin[3]);
    dataCenter.A      = Numerics.removeMpoints(mpoints,xin[4]);
    dataCenter.U      = Numerics.removeMpoints(mpoints,xin[5]);

    // find remaining variables:
    completeAMDL(xdatain[3]);
    findOtherVariables();
 
    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.modelLoaded = true;
    updateSLabels();
    refreshPlots(true);
  }

  /**********************************************************
  * This subroutine writes the current model in AMDL format.
  **********************************************************/ 
  public void writeAMDLmodel(FortranBinaryOutputStream out) throws IOException {

    int i, j;
    int[] ndimout = new int[3];
    double[] xdataout = new double[9];
    double[][] xout = new double[6][];

    ndimout[1] = 0;
    ndimout[2] = dataCenter.grid.length;
    
    xdataout[1] = dataCenter.StellarMass;
    xdataout[2] = dataCenter.StellarRadius;
    xdataout[3] = dataCenter.pressure[0]*dataCenter.Pref;
    xdataout[4] = dataCenter.rho[0]*dataCenter.RhoRef;
    xdataout[5] = 4.0*Math.PI*dataCenter.rho[0]*dataCenter.rho[0]
                /(3.0*dataCenter.Gamma1[0]*dataCenter.pressure[0]);
    xdataout[6] = -dataCenter.ddrho/dataCenter.rho[0];
    xdataout[7] = -1.0;
    xdataout[8] = 0.0;

    xout[0] = dataCenter.grid;
    xout[1] = dataCenter.eta;
    xout[2] = dataCenter.Vg;
    xout[3] = dataCenter.Gamma1;
    xout[4] = dataCenter.A;
    xout[5] = dataCenter.U;

    out.writeInteger(ndimout[1]);
    out.writeInteger(ndimout[2]);
    for(i=1; i<=8; i++) out.writeDoublePrecision(xdataout[i]);
    for(i=0; i<dataCenter.grid.length; i++) {
      for(j=0; j<6; j++) out.writeDoublePrecision(xout[j][i]);
    }
    out.writeRecord();
    out.close();
  }

  /**********************************************************
  * This subroutine reads an FAMDL model from ASTEC
  * and stores the information in different arrays.
  **********************************************************/ 
  public void readFAMDLmodel(BufferedReader modelInput) {

    int[] ndimin = new int[4];
    double[] xdatain = new double[9];
    double[][] xin;
    String line;

    try{
      line = modelInput.readLine();
      StringTokenizer t = new StringTokenizer(line," \t,");
      for(int i=1; i<4; i++) {ndimin[i] = Integer.parseInt(t.nextToken());}
      if ((ndimin[2]<0)||(ndimin[2]>maxNgrid)) {
        JOptionPane.showMessageDialog(
          LoadModel,
          "The number of grid points for the\n"+
          "model seems to be too large, or the\n"+
          "format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return;
      }

      line = modelInput.readLine();
      xdatain[1] = Utilities.parseDouble(line.substring( 0,20));
      xdatain[2] = Utilities.parseDouble(line.substring(20,40));
      xdatain[3] = Utilities.parseDouble(line.substring(40,60));
      xdatain[4] = Utilities.parseDouble(line.substring(60,80));

      line = modelInput.readLine();
      xdatain[5] = Utilities.parseDouble(line.substring( 0,20));
      xdatain[6] = Utilities.parseDouble(line.substring(20,40));
      xdatain[7] = Utilities.parseDouble(line.substring(40,60));
      xdatain[8] = Utilities.parseDouble(line.substring(60,80));

      xin = new double[ndimin[3]+1][ndimin[2]];
      int k = 0;
      for(int i=0; i<ndimin[2]; i++) {
        for(int j=0; j<(ndimin[3]+1); j++) {
          if (k==0) line = modelInput.readLine();
          xin[j][i] = Utilities.parseDouble(line.substring(k*20,(k+1)*20));
          k++;
          if (k==4) k=0;
        }
      }
      modelInput.close();
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "File format not recognised",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    // check for multiple points and non-monotonic grid:
    ArrayList<Integer> mpoints = Numerics.findMpoints(xin[0]);
    if (mpoints == null) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "Grid in reverse order or non-monotonic",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (Numerics.findMinDomainSize(mpoints,ndimin[2]) < orderDer) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "This model contains multiple points which are too\n"+
        "close to allow an accurate calculation of d\u03C1/dr.\n"+
        "Try increasing the number of significant digits\n"+
        "on your radial coordinate.", 
        "Error in model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (mpoints.size() > 0) System.out.println("Removing multiple points from model");

    // Stellar mass in g:
    dataCenter.StellarMass = xdatain[1];
    
    // Stellar radius in cm:
    dataCenter.StellarRadius = xdatain[2];

    findReferenceQuantities();

    // second derivative of density in center:
    dataCenter.ddrho  = -xdatain[4]*xdatain[6]/dataCenter.RhoRef; 

    // find JCD variables:
    dataCenter.grid   = Numerics.removeMpoints(mpoints,xin[0]);
    dataCenter.eta    = Numerics.removeMpoints(mpoints,xin[1]);
    dataCenter.Vg     = Numerics.removeMpoints(mpoints,xin[2]);
    dataCenter.Gamma1 = Numerics.removeMpoints(mpoints,xin[3]);
    dataCenter.A      = Numerics.removeMpoints(mpoints,xin[4]);
    dataCenter.U      = Numerics.removeMpoints(mpoints,xin[5]);

    // find remaining variables:
    completeAMDL(xdatain[3]);
    findOtherVariables();

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.modelLoaded = true;
    updateSLabels();
    refreshPlots(true);
  }    
  
  /**********************************************************
  * This subroutine writes the current model in FAMDL format.
  **********************************************************/ 
  public void writeFAMDLmodel(PrintWriter out) {

    int i, j, k;
    int[] ndimout = new int[4];
    double[] xdataout = new double[9];
    double[][] xout = new double[6][];

    ndimout[1] = 0;
    ndimout[2] = dataCenter.grid.length;
    ndimout[3] = 5;
    
    xdataout[1] = dataCenter.StellarMass;
    xdataout[2] = dataCenter.StellarRadius;
    xdataout[3] = dataCenter.pressure[0]*dataCenter.Pref;
    xdataout[4] = dataCenter.rho[0]*dataCenter.RhoRef;
    xdataout[5] = 4.0*Math.PI*dataCenter.rho[0]*dataCenter.rho[0]
                /(3.0*dataCenter.Gamma1[0]*dataCenter.pressure[0]);
    xdataout[6] = -dataCenter.ddrho/dataCenter.rho[0];
    xdataout[7] = -1.0;
    xdataout[8] = 0.0;

    xout[0] = dataCenter.grid;
    xout[1] = dataCenter.eta;
    xout[2] = dataCenter.Vg;
    xout[3] = dataCenter.Gamma1;
    xout[4] = dataCenter.A;
    xout[5] = dataCenter.U;

    out.printf("%10d%10d%10d%n",ndimout[1],ndimout[2],ndimout[3]);
    out.printf("%20.13E%20.13E%20.13E%20.13E%n",xdataout[1],xdataout[2],xdataout[3],xdataout[4]);
    out.printf("%20.13E%20.13E%20.13E%20.13E%n",xdataout[5],xdataout[6],xdataout[7],xdataout[8]);

    k = 0;
    for(i=0; i<dataCenter.grid.length; i++) {
      for(j=0; j<6; j++) {
        out.printf("%20.13E",xout[j][i]);
        k++;
        if (k == 4) {k=0;out.println();}
      }
    }
    out.close();
  }

  /**********************************************************
  * This subroutine reads model in FGONG format
  * and stores the information in different arrays.
  **********************************************************/ 
  public void readFGONGmodel(BufferedReader modelInput) {

    int[] ndimin = new int[5];
    double[] xdatain;
    double[][] xin;
    String line;
    int i, j, k;

    try{
      // skip header:
      for(i=0; i<4; i++) line = modelInput.readLine();

      // read various dimensions:
      line = modelInput.readLine();
      StringTokenizer t = new StringTokenizer(line," \t,");
      for(i=1; i<=4; i++) {ndimin[i] = Integer.parseInt(t.nextToken());}
      if ((ndimin[1]<0)||(ndimin[1]>maxNgrid)) {
        JOptionPane.showMessageDialog(
          LoadModel,
          "The number of grid points for the\n"+
          "model seems to be too large, or the\n"+
          "format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return;
      }

      // read global variables:      
      xdatain = new double[ndimin[2]+1];
      k = 0;
      for(i=1; i<= ndimin[2]; i++) {
        if (k==0) line = modelInput.readLine();
        xdatain[i] = Utilities.parseDouble(line.substring(k,k+16));
        k+=16;
        if (k>=80) k=0;
      }

      // read local variables (WARNING: they're in reverse order):
      xin = new double[ndimin[3]][ndimin[1]];
      k = 0;
      for(i=ndimin[1]-1;i>=0; i--) {
        for(j=0; j<ndimin[3]; j++) {
          if (k==0) line = modelInput.readLine();
          xin[j][i] = Utilities.parseDouble(line.substring(k,k+16));
          k+=16;
          if (k>=80) k=0;
        }
      }
      modelInput.close();
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "File format not recognised",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    // check for multiple points and non-monotonic grid:
    ArrayList<Integer> mpoints = Numerics.findMpoints(xin[0]);
    if (mpoints == null) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "Grid in reverse order or non-monotonic",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (Numerics.findMinDomainSize(mpoints,ndimin[1]) < orderDer) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "This model contains multiple points which are too\n"+
        "close to allow an accurate calculation of d\u03C1/dr.\n"+
        "Try increasing the number of significant digits\n"+
        "on your radial coordinate.", 
        "Error in model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (mpoints.size() > 0) System.out.println("Removing multiple points from model");

    // Stellar mass in g:
    dataCenter.StellarMass = xdatain[1];
    
    // Stellar radius in cm:
    dataCenter.StellarRadius = xdatain[2];

    findReferenceQuantities();

    // the grid:
    dataCenter.grid = Numerics.removeMpoints(mpoints,xin[0]);
    int ngrid = dataCenter.grid.length;
    for(i=0; i<ngrid; i++) dataCenter.grid[i]/=dataCenter.StellarRadius;

    // the normalised pressure:
    dataCenter.pressure = Numerics.removeMpoints(mpoints,xin[3]);
    for(i=0; i<ngrid; i++) dataCenter.pressure[i]/=dataCenter.Pref;

    // the normalised density:
    dataCenter.rho = Numerics.removeMpoints(mpoints,xin[4]);
    for(i=0; i<ngrid; i++) dataCenter.rho[i]/=dataCenter.RhoRef;
    dataCenter.ddrho = xdatain[12]*dataCenter.rho[0];

    // the adiabatic exponent:
    dataCenter.Gamma1 = Numerics.removeMpoints(mpoints,xin[9]);

    // the cumulative mass:
    dataCenter.mass = Numerics.removeMpoints(mpoints,xin[1]);
    for(i=1; i<ngrid; i++) dataCenter.mass[i] = Math.exp(dataCenter.mass[i]);
    dataCenter.mass[0] = 0.0;

    // the normalised adiabatic discrimenant (related to N2):
    dataCenter.A = Numerics.removeMpoints(mpoints,xin[14]);

    findJCDvariables();
    findOtherVariables();

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.modelLoaded = true;
    updateSLabels();
    refreshPlots(true);
  }
  
  /**********************************************************
  * This subroutine reads a .OSCMOD model from CLES
  * and stores the information in different arrays.
  **********************************************************
  * modelInput = buffer with model
  **********************************************************/
  public void readCLESmodel(BufferedReader modelInput, String filename) {
    
    int ndimin;                        // number of points describing the model (before removing double points)
    int ngrid;                         // number of points describing the model (after removing double points)
    int ndomains;                       // number of convective and radiative zones in the model
    double[] xdatain = new double[3];  // the dimensions are set by the fixed number of global parameters given=3.
    double[][] xin;                    // the dimensions are set when the model is read (model WITH double points)
    int[] idomain;
    String line, subline;
    int i, ii, j;

    try{
      // Skip the header
      line = modelInput.readLine();
      while (line.indexOf("%%beginoscdata") < 0) {
        line = modelInput.readLine();
      }

      // Reading global parameters
      line = modelInput.readLine();
      StringTokenizer t = new StringTokenizer(line," \t");
      for(i=0;i<3; i++) xdatain[i]=Double.parseDouble(t.nextToken());
      
      // Reading information about the double points: (Note: reading the tabular is useless)
      line = modelInput.readLine();
      t = new StringTokenizer(line," \t");
      ndomains = Integer.parseInt(t.nextToken());
      idomain = new int[3*ndomains];
      for(i=0; i<(3*ndomains); i++){
        ii=i%3;
        if(ii==0) line = modelInput.readLine();
        idomain[i]=Integer.parseInt(line.substring(12*ii,12*(ii+1)).trim());
      }

      // Reading the local variables of the model
      line = modelInput.readLine();
      t = new StringTokenizer(line," \t");
      ndimin = Integer.parseInt(t.nextToken());

      xin = new double[6][ndimin];
      for(i=0; i<(ndimin); i++){
        for(j=0; j<6; j++){
          ii=j%6;
          if(ii==0) line = modelInput.readLine();
          subline = line.substring(12,156);
          xin[j][i]=Double.parseDouble(subline.substring(24*ii,24*(ii+1)));
        }
      }
      modelInput.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "File format not recognised",
        "Error reading model",
        JOptionPane.WARNING_MESSAGE
      );
      return;
    }

    // check for multiple points and non-monotonic grid, using the grid:
    ArrayList<Integer> mpoints = Numerics.findMpoints(xin[0]);
    if (mpoints == null) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "Grid in reverse order or non-monotonic",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (Numerics.findMinDomainSize(mpoints,ndimin) < orderDer) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "This model contains multiple points which are too\n"+
        "close to allow an accurate calculation of d\u03C1/dr.\n"+
        "Try increasing the number of significant digits\n"+
        "on your radial coordinate.", 
        "Error in model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (mpoints.size() > 0) System.out.println("Removing multiple points from model");

    // Stellar mass in g:
    dataCenter.StellarRadius = xdatain[0];

    // Stellar radius in cm (does not correspond exactly to last grid point):
    dataCenter.StellarMass = xdatain[1];

    findReferenceQuantities();

    // the radial grid, normalised by the stellar radius:
    dataCenter.grid = Numerics.removeMpoints(mpoints,xin[0]);
    ngrid = dataCenter.grid.length;
    for(i=0; i<ngrid; i++) dataCenter.grid[i]/=dataCenter.StellarRadius;

    // Normalised pressure profile and reference value for the pressure
    dataCenter.pressure = Numerics.removeMpoints(mpoints,xin[2]);
    for(i=0; i<ngrid; i++) dataCenter.pressure[i]/=dataCenter.Pref;

    // Normalised density profile and reference value for the density:
    dataCenter.rho = Numerics.removeMpoints(mpoints,xin[3]);
    for(i=0; i<ngrid; i++) dataCenter.rho[i]/=dataCenter.RhoRef;
    dataCenter.ddrho = Double.NaN;

    // Gamma1 profile:                                   
    dataCenter.Gamma1 = Numerics.removeMpoints(mpoints,xin[4]);

    // the cumulative mass:
    double[] aux = new double[ndimin];
    for(i=0; i<ndimin; i++) aux[i] = xin[1][i]*Math.pow(xin[0][i],3.0)/dataCenter.StellarMass;
    dataCenter.mass = Numerics.removeMpoints(mpoints,aux);

    // Normalised squared Brunt-Vaisala frequency:
    for(i=0; i<ndimin; i++) aux[i] = -xin[0][i]*xin[0][i]*xin[5][i];
    dataCenter.A = Numerics.removeMpoints(mpoints,aux);

    findJCDvariables();
    findOtherVariables();

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.modelLoaded = true;
    updateSLabels();
    refreshPlots(true);
  }

  /**********************************************************
  * This subroutine reads model in LOSC format
  * and stores the information in different arrays.
  **********************************************************/ 
  public void readLOSCmodel(BufferedReader modelInput) {

    int nr;
    double[] xdatain = new double[3];
    double[][] xin;
    String line;
    int i, j, k;

    try{
      // skip header:
      for(i=0; i<5; i++) line = modelInput.readLine();

      // read various dimensions:
      line = modelInput.readLine();
      StringTokenizer t = new StringTokenizer(line," \t,");
      nr = Integer.parseInt(t.nextToken());
      if ((nr<0)||(nr>maxNgrid)) {
        JOptionPane.showMessageDialog(
          LoadModel,
          "The number of grid points for the\n"+
          "model seems to be too large, or the\n"+
          "format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return;
      }

      // read global variables:      
      for(i=0; i<3; i++) {xdatain[i] = Utilities.parseDouble(t.nextToken());}
      line = modelInput.readLine(); // skip a line
      line = modelInput.readLine(); // skip a line
      line = modelInput.readLine(); // skip a line

      // read local variables:
      xin = new double[6][nr];
      for(i=0; i<nr; i++) {
        line = modelInput.readLine();
        t = new StringTokenizer(line," \t,");
        t.nextToken(); // skip a token
        for(j=0;j<6;j++) xin[j][i] = Utilities.parseDouble(t.nextToken());
      }
      modelInput.close();
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "File format not recognised",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    // check for multiple points and non-monotonic grid:
    ArrayList<Integer> mpoints = Numerics.findMpoints(xin[0]);
    if (mpoints == null) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "Grid in reverse order or non-monotonic",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (Numerics.findMinDomainSize(mpoints,nr) < orderDer) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "This model contains multiple points which are too\n"+
        "close to allow an accurate calculation of d\u03C1/dr.\n"+
        "Try increasing the number of significant digits\n"+
        "on your radial coordinate.", 
        "Error in model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (mpoints.size() > 0) System.out.println("Removing multiple points from model");

    // Stellar mass in g:
    dataCenter.StellarMass = xdatain[1];
    
    // Stellar radius in cm:
    dataCenter.StellarRadius = xdatain[0];

    findReferenceQuantities();

    // the grid:
    dataCenter.grid = Numerics.removeMpoints(mpoints,xin[0]);
    int ngrid = dataCenter.grid.length;

    // the normalised pressure:
    double[] aux = new double[nr];
    for(i=0; i<nr; i++) aux[i] = xin[2][i]*xin[3][i]/(4.0*Math.PI);
    dataCenter.pressure = Numerics.removeMpoints(mpoints,aux);

    // the normalised density:
    for(i=0; i<nr; i++) aux[i] = xin[3][i]/(4.0*Math.PI);
    dataCenter.rho = Numerics.removeMpoints(mpoints,aux);
    dataCenter.ddrho = Double.NaN;

    // the adiabatic exponent:
    dataCenter.Gamma1 = Numerics.removeMpoints(mpoints,xin[4]);

    // the cumulative mass:
    for(i=0; i<nr; i++) aux[i] = xin[1][i]*Math.pow(xin[0][i],3.0);
    dataCenter.mass = Numerics.removeMpoints(mpoints,aux);

    // the normalised adiabatic discriminant (related to N2):
    dataCenter.A = Numerics.removeMpoints(mpoints,xin[5]);
    for(i=0; i<ngrid; i++) dataCenter.A[i] *= -(dataCenter.grid[i]*dataCenter.grid[i]);

    findJCDvariables();
    findOtherVariables();

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.modelLoaded = true;
    updateSLabels();
    refreshPlots(true);
  }

  /**********************************************************
   * This reads a binary MOD model.
   **********************************************************
   * @param modelInput input stream which contains the MOD
   *                   model in fortran binary format
   *********************************************************/
  public void readBinaryMODmodel(FortranBinaryInputStream modelInput) {
    int i, j, nvar, np, len;
    double radius, mass, Grav;
    double[][] star = null;
    String[] descript = new String[5];
    try {
      modelInput.readRecord();
      for(i=0;i<5;i++) descript[i] = modelInput.readCharacter(80);
      len = modelInput.readRecord();

      // number of grid points:
      np = modelInput.readInteger();
      if ((np<0)||(np>maxNgrid)) {
        JOptionPane.showMessageDialog(
          LoadModel,
          "The number of grid points for the\n"+
          "model seems to be too large, or the\n"+
          "format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return;
      }

      radius = modelInput.readDoublePrecision();
      mass = modelInput.readDoublePrecision();

      // determine whether there's a fourth variable
      if (len == 20) {
        Grav = dataCenter.G;
      } else {
        Grav = modelInput.readDoublePrecision();
        if (Math.abs(Grav/dataCenter.G - 1.0) > 2e-8) {
          int result = JOptionPane.showConfirmDialog(
            LoadModel,
            "<html>WARNING: your model uses a different value of G:"+
            String.format("<h2>&nbsp;&nbsp;&nbsp;&nbsp;<font color='blue'>G<sub>model</sub>&nbsp;&nbsp;=%10.8f\u00D710\u207B\u2078 (cgs)</font></h2>",Grav*1e8)+
            String.format("<h2>&nbsp;&nbsp;&nbsp;&nbsp;<font color='red'>G<sub>current</sub>=%10.8f\u00D710\u207B\u2078 (cgs)</font></h2>",dataCenter.G*1e8)+
            "Should I update G?</html>",
            "The gravitational constant",JOptionPane.YES_NO_OPTION
          );
          if(result == JOptionPane.YES_OPTION) dataCenter.modifyG(Grav);
        }
      }

      // determine the number of columns
      len = modelInput.readRecord();
      if (len == 40) {
        nvar = 5;
      } else {
        nvar = 6;
      }

      // read model:
      star = new double[nvar][np];
      for(j=0;j<nvar;j++) star[j][0] = modelInput.readDoublePrecision();
      for(i=1;i<np;i++) {
        modelInput.readRecord();
        for(j=0;j<nvar;j++) star[j][i] = modelInput.readDoublePrecision();
      }
      modelInput.close();

    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "File format not recognised",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    // check for multiple points and non-monotonic grid:
    ArrayList<Integer> mpoints = Numerics.findMpoints(star[0]);
    if (mpoints == null) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "Grid in reverse order or non-monotonic",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (Numerics.findMinDomainSize(mpoints,np) < orderDer) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "This model contains multiple points which are too\n"+
        "close to allow an accurate calculation of d\u03C1/dr.\n"+
        "Try increasing the number of significant digits\n"+
        "on your radial coordinate.", 
        "Error in model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (mpoints.size() > 0) System.out.println("Removing multiple points from model");

    // Stellar mass in g:
    dataCenter.StellarMass = mass;
    
    // Stellar radius in cm:
    dataCenter.StellarRadius = radius;

    findReferenceQuantities();

    // the grid:
    dataCenter.grid = Numerics.removeMpoints(mpoints,star[0]);
    int ngrid = dataCenter.grid.length;
    for(i=0;i<ngrid;i++) dataCenter.grid[i]/=dataCenter.StellarRadius;

    // the normalised pressure:
    dataCenter.pressure = Numerics.removeMpoints(mpoints,star[2]);
    for(i=0;i<ngrid;i++) dataCenter.pressure[i]/=dataCenter.Pref;

    // the normalised density:
    dataCenter.rho = Numerics.removeMpoints(mpoints,star[3]);
    for(i=0;i<ngrid;i++) dataCenter.rho[i]/=dataCenter.RhoRef;
    dataCenter.ddrho = Double.NaN;

    // the adiabatic exponent:
    dataCenter.Gamma1 = Numerics.removeMpoints(mpoints,star[4]);

    // the cumulative mass:
    dataCenter.mass = Numerics.removeMpoints(mpoints,star[1]);
    for(i=0;i<ngrid;i++) dataCenter.mass[i]/=dataCenter.StellarMass;

    // tell findOtherVariables() to calculate A:
    dataCenter.A = null;

    findJCDvariables();
    findOtherVariables();

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.modelLoaded = true;
    updateSLabels();
    refreshPlots(true);
  }

  /**********************************************************
  * This subroutine writes the current model in MOD format
  * (binary version).
  **********************************************************/ 
  public void writeBinaryMODmodel(FortranBinaryOutputStream out) throws IOException {

    int i, j;
    int ngrid = dataCenter.grid.length;
    double[] xdataout = new double[9];
    double[][] xout = new double[5][ngrid];
    String[] descript = new String[5];
    
    String hostname = "'localhost'";
    try {
      hostname = InetAddress.getLocalHost().getHostName();
    } catch (Exception e) {
      hostname = "'localhost'";
    }

    descript[0] = String.format("%-80s","# Model converted to MOD format by InversionKit 2.1");
    descript[1] = String.format("%-80s","# Date: "+(new Date()).toString());
    descript[2] = String.format("%-80s","# Created on "+hostname+" by "+System.getProperty("user.name","'user'"));
    descript[3] = String.format("%-80s","#");
    descript[4] = String.format("%-80s","#");

    xdataout[0] = dataCenter.StellarRadius;
    xdataout[1] = dataCenter.StellarMass;
    xdataout[2] = dataCenter.G;

    for(i=0; i<ngrid; i++) {
      xout[0][i] = dataCenter.grid[i]*dataCenter.StellarRadius;
      xout[1][i] = dataCenter.mass[i]*dataCenter.StellarMass;
      xout[2][i] = dataCenter.pressure[i]*dataCenter.Pref;
      xout[3][i] = dataCenter.rho[i]*dataCenter.RhoRef;
      xout[4][i] = dataCenter.Gamma1[i];
    }

    for(i=0;i<5;i++) out.writeCharacter(descript[i]);
    out.writeRecord();
    out.writeInteger(dataCenter.grid.length);
    for(i=0;i<3;i++) out.writeDoublePrecision(xdataout[i]);
    out.writeRecord();
    for(i=0; i<ngrid; i++) {
      for(j=0; j<5; j++) out.writeDoublePrecision(xout[j][i]);
      out.writeRecord();
    }
    out.close();
  }

  /**********************************************************
   * This reads an ascii MOD model.
   **********************************************************
   * @param modelInput input stream which contains the MOD
   *                   model in fortran binary format
   *********************************************************/
  public void readAsciiMODmodel(BufferedReader modelInput) {
    int i, j, nvar, np, len;
    double radius, mass, Grav;
    double[][] star = null;
    String[] descript = new String[5];
    String line;
    StringTokenizer t;
    try {
      for(i=0;i<5;i++) descript[i] = modelInput.readLine();
      line = modelInput.readLine();
      t = new StringTokenizer(line," \t");

      // number of grid points:
      np = Integer.parseInt(t.nextToken());
      if ((np<0)||(np>maxNgrid)) {
        JOptionPane.showMessageDialog(
          LoadModel,
          "The number of grid points for the\n"+
          "model seems to be too large, or the\n"+
          "format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return;
      }

      radius = Utilities.parseDouble(t.nextToken());
      mass   = Utilities.parseDouble(t.nextToken());

      // determine whether there's a fourth variable
      if (t.hasMoreTokens()) {
        Grav = Utilities.parseDouble(t.nextToken());
        if (Math.abs(Grav/dataCenter.G - 1.0) > 2e-8) {
          int result = JOptionPane.showConfirmDialog(
            LoadModel,
            "<html>WARNING: your model uses a different value of G:"+
            String.format("<h2>&nbsp;&nbsp;&nbsp;&nbsp;<font color='blue'>G<sub>model</sub>&nbsp;&nbsp;=%10.8f\u00D710\u207B\u2078 (cgs)</font></h2>",Grav*1e8)+
            String.format("<h2>&nbsp;&nbsp;&nbsp;&nbsp;<font color='red'>G<sub>current</sub>=%10.8f\u00D710\u207B\u2078 (cgs)</font></h2>",dataCenter.G*1e8)+
            "Should I update G?</html>",
            "The gravitational constant",JOptionPane.YES_NO_OPTION
          );
          if(result == JOptionPane.YES_OPTION) dataCenter.modifyG(Grav);
        }
      } else {
        Grav = dataCenter.G;
      }

      // determine the number of columns
      line = modelInput.readLine();
      t = new StringTokenizer(line," \t");
      nvar = t.countTokens();
      nvar--; // subtract 1 for index column

      // read model:
      star = new double[nvar][np];
      t.nextToken(); // skip grid point index
      for(j=0;j<nvar;j++) star[j][0] = Utilities.parseDouble(t.nextToken());
      for(i=1;i<np;i++) {
        line = modelInput.readLine();
        t = new StringTokenizer(line," \t");
        t.nextToken(); // skip grid point index
        for(j=0;j<nvar;j++) star[j][i] = Utilities.parseDouble(t.nextToken());
      }
      modelInput.close();
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "File format not recognised",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    // check for multiple points and non-monotonic grid:
    ArrayList<Integer> mpoints = Numerics.findMpoints(star[0]);
    if (mpoints == null) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "Grid in reverse order or non-monotonic",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (Numerics.findMinDomainSize(mpoints,np) < orderDer) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "This model contains multiple points which are too\n"+
        "close to allow an accurate calculation of d\u03C1/dr.\n"+
        "Try increasing the number of significant digits\n"+
        "on your radial coordinate.", 
        "Error in model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (mpoints.size() > 0) System.out.println("Removing multiple points from model");

    // Stellar mass in g:
    dataCenter.StellarMass = mass;
    
    // Stellar radius in cm:
    dataCenter.StellarRadius = radius;

    findReferenceQuantities();

    // the grid:
    dataCenter.grid = Numerics.removeMpoints(mpoints,star[0]);
    int ngrid = dataCenter.grid.length;
    for(i=0;i<ngrid;i++) dataCenter.grid[i]/=dataCenter.StellarRadius;

    // the normalised pressure:
    dataCenter.pressure = Numerics.removeMpoints(mpoints,star[2]);
    for(i=0;i<ngrid;i++) dataCenter.pressure[i]/=dataCenter.Pref;

    // the normalised density:
    dataCenter.rho = Numerics.removeMpoints(mpoints,star[3]);
    for(i=0;i<ngrid;i++) dataCenter.rho[i]/=dataCenter.RhoRef;
    dataCenter.ddrho = Double.NaN;

    // the adiabatic exponent:
    dataCenter.Gamma1 = Numerics.removeMpoints(mpoints,star[4]);

    // the cumulative mass:
    dataCenter.mass = Numerics.removeMpoints(mpoints,star[1]);
    for(i=0;i<ngrid;i++) dataCenter.mass[i]/=dataCenter.StellarMass;

    // tell findOtherVariables() to calculate A:
    dataCenter.A = null;

    findJCDvariables();
    findOtherVariables();

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.modelLoaded = true;
    updateSLabels();
    refreshPlots(true);
  }

  /**********************************************************
  * This subroutine writes the current model in MOD format
  * (ascii version).
  **********************************************************/ 
  public void writeAsciiMODmodel(PrintWriter out) {

    int i, j;
    int ngrid = dataCenter.grid.length;
    double[] xdataout = new double[9];
    double[][] xout = new double[5][ngrid];
    String[] descript = new String[5];
    
    String hostname = "'localhost'";
    try {
      hostname = InetAddress.getLocalHost().getHostName();
    } catch (Exception e) {
      hostname = "'localhost'";
    }

    descript[0] = "# Model converted to MOD format by InversionKit 2.1";
    descript[1] = "# Date: "+(new Date()).toString();
    descript[2] = "# Created on "+hostname+" by "+System.getProperty("user.name","'user'");
    descript[3] = "#";
    descript[4] = "#";

    xdataout[0] = dataCenter.StellarRadius;
    xdataout[1] = dataCenter.StellarMass;
    xdataout[2] = dataCenter.G;

    for(i=0; i<ngrid; i++) {
      xout[0][i] = dataCenter.grid[i]*dataCenter.StellarRadius;
      xout[1][i] = dataCenter.mass[i]*dataCenter.StellarMass;
      xout[2][i] = dataCenter.pressure[i]*dataCenter.Pref;
      xout[3][i] = dataCenter.rho[i]*dataCenter.RhoRef;
      xout[4][i] = dataCenter.Gamma1[i];
    }

    for(i=0;i<5;i++) out.println(descript[i]);
    out.print(ngrid);
    for(i=0;i<3;i++) out.printf(" %22.15e",xdataout[i]);
    out.println();
    for(i=0; i<ngrid; i++) {
      out.printf("%5d",i+1);
      for(j=0; j<5; j++) out.printf(" %22.15e",xout[j][i]);
      out.println();
    }
    out.close();
  }

  /**********************************************************
  * This subroutine reads a OSC model from CESAM
  * and stores the information in different arrays.
  **********************************************************
  * modelInput = buffer with model
  * nVers      = a number which gives the version:
  *              nVers = 1 <=> CESAM2k version
  *              nVers = 0 <=> older version
  **********************************************************/
  public void readOSCmodel(BufferedReader modelInput, int nVers) {
    
    int[] ndimin = new int[5];
    double[] xdatain;  // the dimensions are set when the model is read
    double[][] xin;    // the dimensions are set when the model is read
    String line;
    int iVers;

    try{
      // Skip the header
      for(int i=0; i<5; i++) {line = modelInput.readLine();}

      // Now read lines of text:
      line = modelInput.readLine();
      StringTokenizer t = new StringTokenizer(line," \t,");
      if (nVers == 1) {
        for(int i=0; i<5; i++) {ndimin[i] = Integer.parseInt(t.nextToken());}
      } else {
        for(int i=0; i<3; i++) {ndimin[i] = Integer.parseInt(t.nextToken());}
        iVers = Integer.parseInt(t.nextToken());
        for(int i=3; i<5; i++) {ndimin[i] = Integer.parseInt(t.nextToken());}
      }
      if ((ndimin[0]<0)||(ndimin[0]>maxNgrid)) {
        JOptionPane.showMessageDialog(
          LoadModel,
          "The number of grid points for the\n"+
          "model seems to be too large, or the\n"+
          "format is incorrect.",
          "Error reading file",
          JOptionPane.ERROR_MESSAGE
        );
        return;
      }

      // I couldn't find any readf methods, so I cut the string up into
      // substrings of 19 characters and then try to read a value.
      //
      // Note: the StringTokenizer method won't work because there
      //       are no white spaces preceding negative values (since
      //       the minus sign occupies this space).

      int jj;
      // I add an extra element to xdatain so as to agree with CESAM
      // documentation.  xdatain[0] corresponds to nothing.
      xdatain = new double[ndimin[1]+1];
      for(int i=0; i<ndimin[1];i++) {
        jj=i%5;
        if(jj==0) line = modelInput.readLine();
        xdatain[i+1] = Utilities.parseDouble(line.substring(jj*19,(jj+1)*19));
      }

      // I increase the first dimension of xin by 1, so as to
      // have the same first index as what is given in the Cesam
      // documentation.  Therefore, xin[0][*] corresponds to nothing:
      xin = new double[ndimin[2]+ndimin[3]+1][ndimin[0]];

      // Cesam Models start from the surface and go to the center,
      // hence the reverse direction on the i loop.
      for(int i=ndimin[0]-1; i>=0;i--) {
        for(int j=0; j<(ndimin[2]+ndimin[3]); j++) {
          jj=j%5;
          if(jj==0) line = modelInput.readLine();
          // The "j+1" coordinate insures that the first index of
          // of xin corresponds to Cesam documentation.
          xin[j+1][i] = Utilities.parseDouble(line.substring(jj*19,(jj+1)*19));
        }
      }
      modelInput.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "File format not recognised",
        "Error reading model",
        JOptionPane.WARNING_MESSAGE
      );
      return;
    }

     // check for multiple points and non-monotonic grid:
    ArrayList<Integer> mpoints = Numerics.findMpoints(xin[1]);
    if (mpoints == null) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "Grid in reverse order or non-monotonic",
        "Error reading model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (Numerics.findMinDomainSize(mpoints,ndimin[0]) < orderDer) {
      JOptionPane.showMessageDialog(
        LoadModel,  // the parent component
        "This model contains multiple points which are too\n"+
        "close to allow an accurate calculation of d\u03C1/dr.\n"+
        "Try increasing the number of significant digits\n"+
        "on your radial coordinate.", 
        "Error in model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }
    if (mpoints.size() > 0) System.out.println("Removing multiple points from model");

    // Stellar mass in g:
    dataCenter.StellarMass = xdatain[1];

    // Stellar radius in cm (does not correspond exactly to last grid point):
    dataCenter.StellarRadius = xdatain[2];

    findReferenceQuantities();
    
    // Radial grid, normalised by stellar radius:
    dataCenter.grid = Numerics.removeMpoints(mpoints,xin[1]);
    int ngrid = dataCenter.grid.length;
    for(int i=0; i<ngrid; i++) dataCenter.grid[i]/=dataCenter.StellarRadius;

    // Normalised pressure profile and reference value for the pressure
    dataCenter.pressure = Numerics.removeMpoints(mpoints,xin[4]);
    for (int i=0; i < ngrid; i++) dataCenter.pressure[i]/=dataCenter.Pref;

    // Normalised density profile and reference value for the density:
    dataCenter.rho = Numerics.removeMpoints(mpoints,xin[5]);
    for (int i=0; i < ngrid; i++) dataCenter.rho[i]/=dataCenter.RhoRef;
    dataCenter.ddrho = xdatain[10]*dataCenter.rho[0];

    // Gamma1 profile:                                    
    dataCenter.Gamma1 = Numerics.removeMpoints(mpoints,xin[10]);

    // Cumulative mass:
    dataCenter.mass = Numerics.removeMpoints(mpoints,xin[2]);
    for(int i=1; i<ngrid; i++) dataCenter.mass[i]=Math.exp(dataCenter.mass[i]);
    dataCenter.mass[0] = 0.0;

    // the normalised adiabatic discriminant (related to N2):
    dataCenter.A = Numerics.removeMpoints(mpoints,xin[15]);

    findJCDvariables();
    findOtherVariables();

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.modelLoaded = true;
    updateSLabels();
    refreshPlots(true);
  }

  /**********************************************************
  * This generates a polytropic model (the actual
  * calculations are done in the Polytrope class):
  **********************************************************/ 
  private void generatePolytrope(int nstep, double N) {

    // this generates the polytrope:
    Polytrope p = new Polytrope(nstep, N);
    N = p.getN(); // just in case N is modified in the previous step
    
    // put the results in the dataCenter:

    findReferenceQuantities();

    // the radial grid:
    dataCenter.grid = new double[nstep];
    for(int i=0; i<nstep; i++) dataCenter.grid[i] = p.r[i];
    
    // the adiabatic exponent (= 5/3 throughout):
    double Gamma1 = 5.0/3.0;
    dataCenter.Gamma1 = new double[nstep];
    for(int i=0; i<nstep; i++) dataCenter.Gamma1[i] = Gamma1;
    
    // the normalised density profile:
    double RhoC = 3.0*p.Alpha/(4.0*Math.PI);
    dataCenter.rho = new double[nstep];
    for(int i=0; i<nstep; i++) dataCenter.rho[i] = RhoC*Math.pow(Math.abs(p.h[i]),N);
    dataCenter.ddrho = -N*p.Lambda*RhoC/3.0;

    // find cumulative mass:
    dataCenter.mass = findCumulativeMass();

    // the normalised pressure profile:
    double PC = 9.0*p.Alpha*p.Alpha/(4.0*Math.PI*(N+1.0)*p.Lambda);
    dataCenter.pressure = new double[nstep];
    for(int i=0; i<nstep; i++) dataCenter.pressure[i] = PC*Math.pow(Math.abs(p.h[i]),N+1);

    // the normalised adiabatic discriminant (note: this diverges at the surface):
    double AA = ((N+1.0)/Gamma1 - N);
    dataCenter.A = new double[nstep];
    for(int i=0; i<nstep-1; i++) dataCenter.A[i] = AA*dataCenter.grid[i]*p.hp[i]/p.h[i];
    dataCenter.A[nstep-1] = Double.NaN;

    findJCDvariables();
    findOtherVariables();

    // set the correct surface value of the normalised sound velocity:
    dataCenter.c2[nstep-1] = 0.0;

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.modelLoaded = true;
    updateSLabels();
    refreshPlots(true);
  }

  /**********************************************************
  * This finds the reference values for the pressure, density
  * and frequency, based on the stellar mass and radius.
  **********************************************************/ 
  private void findReferenceQuantities() {
    // Reference pressure:
    dataCenter.Pref = dataCenter.G*Math.pow(dataCenter.StellarMass,2.0)
                                  /Math.pow(dataCenter.StellarRadius,4.0);

    // Reference density:
    dataCenter.RhoRef = dataCenter.StellarMass
                       /Math.pow(dataCenter.StellarRadius,3.0);

    // Reference frequency:
    dataCenter.FreqRef = Math.sqrt(dataCenter.G*dataCenter.RhoRef)*5e5/Math.PI;

    // Dynamical Time:
    dataCenter.TauRef = 1.0/Math.sqrt(dataCenter.G*dataCenter.RhoRef);
  }

  /***********************************************************
  * This finds the cumulative mass from the density profile.
  ***********************************************************/
  private double[] findCumulativeMass() {
    int ngrid = dataCenter.grid.length;
    int order = 1;
    double[] intMat = Numerics.findIntegrationMatrix(dataCenter.grid,order,2);
    double[] aux = new double[ngrid];
    for(int i=0; i<ngrid; i++) aux[i] = 4.0*Math.PI*dataCenter.rho[i];
    return Numerics.applyIntegrationMatrix(intMat,aux,0.0,order);
  }

  /***********************************************************
  * This finds the cumulative mass from the density profile.
  ***********************************************************/
  private double[] findCumulativeMassOld() {
    int ngrid = dataCenter.grid.length;

    // analytical integration of rho treated as trapezoidal function in r^2
    // this gives good results for diffrho_r2:
    double ri, rip, cnst = 4.0*Math.PI/15.0;
    double [] mass = new double[ngrid];
    rip = dataCenter.grid[0];
    mass[0] = 0.0; 
    for(int i=1; i<ngrid; i++) {
      ri = rip;
      rip = dataCenter.grid[i];
      mass[i] = mass[i-1] + (cnst*(rip-ri)/(rip+ri))
              * ((2.0*rip*rip*rip+4.0*rip*rip*ri+6.0*rip*ri*ri+3.0*ri*ri*ri)*dataCenter.rho[i-1]
              +  (2.0*ri*ri*ri+4.0*ri*ri*rip+6.0*ri*rip*rip+3.0*rip*rip*rip)*dataCenter.rho[i]);
    }
    return mass;
  }

  /**********************************************************
  * This subroutine calculates the density, pressure and
  * cumulative mass for AMDL and FAMDL models.
  **********************************************************/ 
  private void completeAMDL(double Pc) {
    int ngrid = dataCenter.grid.length, i;

    // the normalised density:
    dataCenter.rho = new double[ngrid];
    for(i=0; i<ngrid; i++)
      dataCenter.rho[i] = dataCenter.eta[i]*dataCenter.U[i]/(4.0*Math.PI);

    // the normalised pressure:
    dataCenter.pressure = new double[ngrid];
    for(i=1; i<ngrid; i++) 
      dataCenter.pressure[i] = Math.pow(dataCenter.eta[i]*dataCenter.grid[i],2.0)*dataCenter.U[i]
                             / (4.0*Math.PI*dataCenter.Vg[i]*dataCenter.Gamma1[i]);
    dataCenter.pressure[0] = Pc/dataCenter.Pref; // special treatment for the center

    // the cumulative mass:
    dataCenter.mass = findCumulativeMass();
  }

  /***********************************************************
  * This finds the variables U, Vg and eta in FGONG, CESAM,
  * MOD, LOSC, and polytropic models.
  ************************************************************
  * IMPORTANT: the cumulative mass has to be known beforehand.
  ***********************************************************/
  private void findJCDvariables() {
    int ngrid = dataCenter.grid.length, i, j;

    dataCenter.Vg  = new double[ngrid];
    dataCenter.U   = new double[ngrid];
    dataCenter.eta = new double[ngrid];
    for(i=1;i<ngrid;i++) {
      dataCenter.Vg[i]  = dataCenter.mass[i]*dataCenter.rho[i]
                        /(dataCenter.Gamma1[i]*dataCenter.pressure[i]*dataCenter.grid[i]);
      dataCenter.U[i]   = 4.0*Math.PI*dataCenter.rho[i]
                        * Math.pow(dataCenter.grid[i],3.0)
                        / dataCenter.mass[i];
      dataCenter.eta[i] = dataCenter.mass[i]
                        / Math.pow(dataCenter.grid[i],3.0);
    }
    dataCenter.eta[0] = 4.0*Math.PI*dataCenter.rho[0]/3.0;
    dataCenter.U[0]   = 3.0;
    dataCenter.Vg[0]  = 0.0;
  }

  /***********************************************************
  * This finds the following variables: c2, gravity, dgravity,
  * drho/dr, A (if needed), N2, diffrho_r2, indSurface
  ************************************************************
  * IMPORTANT: the cumulative mass and eta need to be known.
  * NOTE: this is where to add other quantities which need
  *       to be calculated for all of the models.
  ***********************************************************/
  private void findOtherVariables() {
    int ngrid = dataCenter.grid.length, i, j;

    // normalised sound velocity profile:
    dataCenter.c2 = new double[ngrid];
    for(i=0; i<ngrid; i++) dataCenter.c2[i] = dataCenter.Gamma1[i]*dataCenter.pressure[i]/dataCenter.rho[i];

    // the gravity profile:
    dataCenter.gravity = new double[ngrid];
    for(i=1; i<ngrid; i++)
      dataCenter.gravity[i] = dataCenter.mass[i]/(dataCenter.grid[i]*dataCenter.grid[i]);
    dataCenter.gravity[0] = 0.0;

    // the derivative of the gravity profile
    dataCenter.dgravity = new double[ngrid];
    for(i=1; i<ngrid; i++)
      dataCenter.dgravity[i] = -2.0*dataCenter.gravity[i]/dataCenter.grid[i]
                             + 4.0*Math.PI*dataCenter.rho[i];
    dataCenter.dgravity[0] = 4.0*Math.PI*dataCenter.rho[0]/3.0;

    // the derivative of the density profile:
    double[] aux = new double[ngrid];
    for(i=0; i<ngrid; i++) aux[i] = dataCenter.grid[i]*dataCenter.grid[i];
    double[][] dermat = Numerics.initDeriveFDBand(aux,orderDer,1);
    dataCenter.drho = new double[ngrid];
    Dgbmv.dgbmv("N",ngrid,ngrid,orderDer,orderDer,1.0,dermat[1],0,2*orderDer+1,dataCenter.rho,0,1,0.0,dataCenter.drho,0,1);
    // I'm assuming dataCenter.grid[0] == 0.0:
    if (Double.isNaN(dataCenter.ddrho)) dataCenter.ddrho = 2.0*dataCenter.drho[0];
    for(i=0; i<ngrid; i++) dataCenter.drho[i]*=2.0*dataCenter.grid[i];

    // find the normalised adiabatic discriminant if needed:
    if (dataCenter.A == null) {
      dataCenter.A = new double[ngrid];
      for(i=0;i<ngrid;i++) {
        dataCenter.A[i] = -dataCenter.grid[i]
        * (dataCenter.gravity[i]/dataCenter.c2[i]
        +  dataCenter.drho[i]/dataCenter.rho[i]);
      }
    }

    // find the normalised squared Brunt-Vaisala frequency:
    dataCenter.N2 = new double[ngrid];
    for(i=0;i<ngrid;i++) dataCenter.N2[i] = dataCenter.eta[i]*dataCenter.A[i];

    // find the relative difference between the local and average densities, divided by the square of
    // the position:
    double[] mass = findCumulativeMass();  // recalculate the cummulative mass using a method suitable for diffrho_r2
    dataCenter.diffrho_r2 = new double[ngrid];
    for(i=1; i<ngrid; i++) dataCenter.diffrho_r2[i] = (1.0-4.0*Math.PI*Math.pow(dataCenter.grid[i],3.0)*dataCenter.rho[i]
                                                    / (3.0*mass[i]))/Math.pow(dataCenter.grid[i],2.0);
    dataCenter.diffrho_r2[0] = -dataCenter.ddrho/(5.0*dataCenter.rho[0]);

    // variables related to the age indicator 
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);
    double[] c = new double[ngrid];
    for(i=0; i<ngrid; i++) c[i]= Math.sqrt(dataCenter.c2[i]);
    c[0] = c[1] - (c[2]-c[1])*aux[1]/(aux[2]-aux[1]); // ad-hoc correction to center value
    dataCenter.agesmallsep2 = new double[ngrid];
    Dgbmv.dgbmv("N",ngrid,ngrid,orderDer,orderDer,1.0,dermat[1],0,2*orderDer+1,c,0,1,0.0,dataCenter.agesmallsep2,0,1);
    dataCenter.agesmallsep = 0.0;
    for(i=0; i<ngrid; i++) {
      dataCenter.agesmallsep2[i] *= 2.0;
      dataCenter.agesmallsep += (dataCenter.agesmallsep2[i]*weights[i]);
    }

    // the normalised acoustic radius (beware of polytropes...):
    dataCenter.Tau = 0.0; 
    for(i=0; i<ngrid; i++) dataCenter.Tau +=weights[i]/Math.sqrt(dataCenter.c2[i]);

    // find the surface index for the grid.
    double surface=1.0;
    dataCenter.indSurface=0;
    double err=Math.abs(surface - dataCenter.grid[0]);
    for (i=1;i<ngrid;i++) {
      if (Math.abs(surface - dataCenter.grid[i])< err) {
        dataCenter.indSurface=i;
        err=Math.abs(surface - dataCenter.grid[i]);
      }
    }
  }

  /**********************************************************
  * This creates initialises HELAS tree with the stellar
  * models from the HELAS website.
  **********************************************************/ 
  private void initHELAStree() {
    HELAStree = new JTree(readHELASlist());
    HELAStree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
    HELAStree.addTreeSelectionListener(new TreeSelectionListener() {
      public void valueChanged(TreeSelectionEvent e) {
        TreePath aPath = HELAStree.getSelectionPath();
        if (aPath == null) return;
        HELASTextField.setText(Path2URL(aPath));
        DownloadHELASmodelButton.setEnabled(((MutableTreeNode)aPath.getLastPathComponent()).isLeaf());
        LoadSaveHELASmodelButton.setEnabled(((MutableTreeNode)aPath.getLastPathComponent()).isLeaf());
      }
    });
  }

  /**********************************************************
  * This reads the list of HELAS models from a file called
  * model_list in the resource_files directory.
  **********************************************************/ 
  private DefaultMutableTreeNode readHELASlist() {
    DefaultMutableTreeNode root = new DefaultMutableTreeNode("HELAS stellar model");
    DefaultMutableTreeNode previousNode;
    DefaultMutableTreeNode currentNode;
    DefaultMutableTreeNode nextNode;
    String anElement;
    Enumeration en;
    String line;
    BufferedReader list;

    try {
      // Open file with the list of HELAS models
      URL url = getClass().getResource("/resource_files/model_list");
      InputStream is = url.openStream();
      InputStreamReader isr = new InputStreamReader(is);
      list = new BufferedReader(isr);
    } catch (Exception e) {
      return root;
    }
    try {
      while (true) {
        line = list.readLine();  // read a line of text
        if (line == null) break; // test for the end of the buffer
        StringTokenizer t =  new StringTokenizer(line,"/");
        previousNode = root;

        while(t.hasMoreTokens()) {
          anElement = t.nextToken();

          // test to see if previousNode already has current node:
          nextNode = null;
          for (en = previousNode.children();en.hasMoreElements() ;) {
            currentNode = (DefaultMutableTreeNode)en.nextElement();
            if (currentNode.toString().equals(anElement)) {
              nextNode = currentNode;
              break;
            }
          }

          // node if it doesn't exist yet:
          if (nextNode == null) {
            nextNode = new DefaultMutableTreeNode(anElement);
            previousNode.add(nextNode);
          }

          previousNode = nextNode;
        }
      }
      list.close();
    } catch (Exception e) {
      return root;
    }
    return root;
  }

  /**********************************************************
  * This converts a tree path from the HELAS tree into the
  * correct URL address.
  **********************************************************/ 
  private static String Path2URL(TreePath aPath) {
    String result = "http://www.astro.up.pt/helas/stars/cesam/";
    if (aPath == null) return result;
    Object[] pathArray = aPath.getPath();
    if (pathArray.length < 2) return result;
    result = result + pathArray[1].toString() + "/data";
    for(int i=2; i<pathArray.length; i++)
      result = result + "/"+pathArray[i].toString();
    return result;
  }

  /**********************************************************
  * This refreshes the density and pressure plots and writes
  * some information in the output log.
  **********************************************************/ 
  private void refreshPlots(boolean findRange) {
    DensityPlot.clear();
    DDensityPlot.clear();
    PressurePlot.clear();
    Gamma1Plot.clear();
    c2Plot.clear();
    dcPlot.clear();
    N2Plot.clear();
    APlot.clear();
    UPlot.clear();
    VgPlot.clear();
    etaPlot.clear();
    massPlot.clear();
    gPlot.clear();
    gpPlot.clear();
    diffRhoPlot.clear();
    freqPlot.clear();
      
    if (!dataCenter.modelLoaded) {
      DensityPlot.repaint();
      DDensityPlot.repaint();
      PressurePlot.repaint();
      Gamma1Plot.repaint();
      c2Plot.repaint();
      dcPlot.repaint();
      N2Plot.repaint();
      APlot.repaint();
      UPlot.repaint();
      VgPlot.repaint();
      etaPlot.repaint();
      massPlot.repaint();
      gPlot.repaint();
      gpPlot.repaint();
      diffRhoPlot.repaint();
      freqPlot.repaint();
      return;
    }
    int n;
    double factor;
    double[] aux = new double[dataCenter.grid.length];
    ArrayList<Integer> domains = Numerics.findDomains(dataCenter.grid);
    double[][] doublePoints = new double[domains.size()/2-1][2];
    for(n=1;n<domains.size()-2;n+=2) {
      doublePoints[(n-1)/2][0] = dataCenter.grid[domains.get(n)];
      doublePoints[(n-1)/2][1] = dataCenter.grid[domains.get(n+1)];
    }
    double[] range;
    boolean showDoublePoints = (ShowDoublePoints.isSelected() && (doublePoints.length > 0));
    boolean showGridPoints = ShowGridPoints.isSelected();
    int symbol = Plot.times;

    for(int i=0; i<aux.length; i++) aux[i] = dataCenter.RhoRef*dataCenter.rho[i];
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        DensityPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    DensityPlot.addDataSet(dataCenter.grid,aux);
    if (showGridPoints) DensityPlot.addDataSet(dataCenter.grid,aux,Plot.noLine,symbol);
    if (findRange) DensityPlot.findRange();
    DensityPlot.repaint();

    for(int i=0; i<aux.length; i++) aux[i] = dataCenter.RhoRef*dataCenter.drho[i]/dataCenter.StellarRadius;
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        DDensityPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    DDensityPlot.addDataSet(dataCenter.grid,aux);
    if (showGridPoints) DDensityPlot.addDataSet(dataCenter.grid,aux,Plot.noLine,symbol);
    if (findRange) DDensityPlot.findRange();
    DDensityPlot.repaint();

    for(int i=0; i<aux.length; i++) aux[i] = dataCenter.Pref*dataCenter.pressure[i];
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        PressurePlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    PressurePlot.addDataSet(dataCenter.grid,aux);
    if (showGridPoints) PressurePlot.addDataSet(dataCenter.grid,aux,Plot.noLine,symbol);
    if (findRange) PressurePlot.findRange();
    PressurePlot.repaint();

    if (showDoublePoints) {
      range = new double[]{Numerics.minval(dataCenter.Gamma1),Numerics.maxval(dataCenter.Gamma1)};
      for(n=0;n<doublePoints.length;n++) {
        Gamma1Plot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    Gamma1Plot.addDataSet(dataCenter.grid,dataCenter.Gamma1);
    if (showGridPoints) Gamma1Plot.addDataSet(dataCenter.grid,dataCenter.Gamma1,Plot.noLine,symbol);
    if (findRange) Gamma1Plot.findRange();
    Gamma1Plot.repaint();

    factor = dataCenter.Pref/dataCenter.RhoRef;
    for(int i=0; i<aux.length; i++) aux[i] = factor*dataCenter.c2[i];
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        c2Plot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    c2Plot.addDataSet(dataCenter.grid,aux);
    if (showGridPoints) c2Plot.addDataSet(dataCenter.grid,aux,Plot.noLine,symbol);
    if (findRange) c2Plot.findRange();
    c2Plot.repaint();

    factor = 1.0/dataCenter.TauRef;
    for(int i=0; i<aux.length; i++) aux[i] = factor*dataCenter.agesmallsep2[i];
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        dcPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    dcPlot.addDataSet(dataCenter.grid,aux);
    if (showGridPoints) dcPlot.addDataSet(dataCenter.grid,aux,Plot.noLine,symbol);
    if (findRange) dcPlot.findRange();
    dcPlot.repaint();

    factor = Math.pow(Math.PI*2e-6*dataCenter.FreqRef,2.0);
    for(int i=0; i<aux.length; i++) aux[i] = factor*dataCenter.N2[i];
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        N2Plot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    N2Plot.addDataSet(dataCenter.grid,aux);
    if (showGridPoints) N2Plot.addDataSet(dataCenter.grid,aux,Plot.noLine,symbol);
    if (findRange) N2Plot.findRange();
    N2Plot.repaint();

    if (showDoublePoints) {
      range = new double[]{Numerics.minval(dataCenter.A),Numerics.maxval(dataCenter.A)};
      for(n=0;n<doublePoints.length;n++) {
        APlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    APlot.addDataSet(dataCenter.grid,dataCenter.A);
    if (showGridPoints) APlot.addDataSet(dataCenter.grid,dataCenter.A,Plot.noLine,symbol);
    if (findRange) APlot.findRange();
    APlot.repaint();

    if (showDoublePoints) {
      range = new double[]{Numerics.minval(dataCenter.U),Numerics.maxval(dataCenter.U)};
      for(n=0;n<doublePoints.length;n++) {
        UPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    UPlot.addDataSet(dataCenter.grid,dataCenter.U);
    if (showGridPoints) UPlot.addDataSet(dataCenter.grid,dataCenter.U,Plot.noLine,symbol);
    if (findRange) UPlot.findRange();
    UPlot.repaint();

    if (showDoublePoints) {
      range = new double[]{Numerics.minval(dataCenter.Vg),Numerics.maxval(dataCenter.Vg)};
      for(n=0;n<doublePoints.length;n++) {
        VgPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    VgPlot.addDataSet(dataCenter.grid,dataCenter.Vg);
    if (showGridPoints) VgPlot.addDataSet(dataCenter.grid,dataCenter.Vg,Plot.noLine,symbol);
    if (findRange) VgPlot.findRange();
    VgPlot.repaint();

    if (showDoublePoints) {
      range = new double[]{Numerics.minval(dataCenter.eta),Numerics.maxval(dataCenter.eta)};
      for(n=0;n<doublePoints.length;n++) {
        etaPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    etaPlot.addDataSet(dataCenter.grid,dataCenter.eta);
    if (showGridPoints) etaPlot.addDataSet(dataCenter.grid,dataCenter.eta,Plot.noLine,symbol);
    if (findRange) etaPlot.findRange();
    etaPlot.repaint();

    for(int i=0; i<aux.length; i++) aux[i] = dataCenter.StellarMass*dataCenter.mass[i];
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        massPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    massPlot.addDataSet(dataCenter.grid,aux);
    if (showGridPoints) massPlot.addDataSet(dataCenter.grid,aux,Plot.noLine,symbol);
    if (findRange) massPlot.findRange();
    massPlot.repaint();

    factor = dataCenter.G*dataCenter.StellarMass/Math.pow(dataCenter.StellarRadius,2.0);
    for(int i=0; i<aux.length; i++) aux[i] = factor*dataCenter.gravity[i];
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        gPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    gPlot.addDataSet(dataCenter.grid,aux);
    if (showGridPoints) gPlot.addDataSet(dataCenter.grid,aux,Plot.noLine,symbol);
    if (findRange) gPlot.findRange();
    gPlot.repaint();

    factor = dataCenter.G*dataCenter.StellarMass/Math.pow(dataCenter.StellarRadius,3.0);
    for(int i=0; i<aux.length; i++) aux[i] = factor*dataCenter.dgravity[i];
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        gpPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    gpPlot.addDataSet(dataCenter.grid,aux);
    if (showGridPoints) gpPlot.addDataSet(dataCenter.grid,aux,Plot.noLine,symbol);
    if (findRange) gpPlot.findRange();
    gpPlot.repaint();

    if (showDoublePoints) {
      range = new double[]{Numerics.minval(dataCenter.diffrho_r2),Numerics.maxval(dataCenter.diffrho_r2)};
      for(n=0;n<doublePoints.length;n++) {
        diffRhoPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    diffRhoPlot.addDataSet(dataCenter.grid,dataCenter.diffrho_r2);
    if (showGridPoints) diffRhoPlot.addDataSet(dataCenter.grid,dataCenter.diffrho_r2,Plot.noLine,symbol);
    if (findRange) diffRhoPlot.findRange();
    diffRhoPlot.repaint();

    for(int i=0; i<aux.length; i++) aux[i] = dataCenter.FreqRef*Math.sqrt(dataCenter.N2[i]);
    if (showDoublePoints) {
      range = new double[]{Numerics.minval(aux),Numerics.maxval(aux)};
      for(n=0;n<doublePoints.length;n++) {
        freqPlot.addDataSet(doublePoints[n],range,Plot.dotted,Plot.noSymbol,Color.LIGHT_GRAY);
      }
    }
    freqPlot.addDataSet(dataCenter.grid,aux);
    // ell = 1;
    for(int i=0; i<aux.length; i++) aux[i] = dataCenter.FreqRef*Math.sqrt(2.0*dataCenter.c2[i])/dataCenter.grid[i];
    freqPlot.addDataSet(dataCenter.grid,aux,Plot.dotted,Plot.noSymbol,Color.blue);
    // ell = 2;
    for(int i=0; i<aux.length; i++) aux[i] = dataCenter.FreqRef*Math.sqrt(6.0*dataCenter.c2[i])/dataCenter.grid[i];
    freqPlot.addDataSet(dataCenter.grid,aux,Plot.dotted,Plot.noSymbol,Color.blue);
    // ell = 3;
    for(int i=0; i<aux.length; i++) aux[i] = dataCenter.FreqRef*Math.sqrt(12.0*dataCenter.c2[i])/dataCenter.grid[i];
    freqPlot.addDataSet(dataCenter.grid,aux,Plot.dotted,Plot.noSymbol,Color.blue);
    freqPlot.findRange(new int[]{0});
    freqPlot.setXmin(dataCenter.grid[0]);
    freqPlot.setXmax(dataCenter.grid[dataCenter.grid.length-1]);
    freqPlot.repaint();
  }

  /***********************************************************
  * This tries to determine the value of G using the equation
  * of hydrostatic equilibrium on the calculated model.
  ***********************************************************/
  private double findGravitationalConstant() {
    int ngrid = dataCenter.grid.length;

    // find derivative of pressure
    int orderFD = 2; // 2 leads to 4th order accuracy
    double[][] dermat = Numerics.initDeriveFDBand(dataCenter.grid,orderFD,1);
    double[] dpressure = new double[ngrid];
    Dgbmv.dgbmv("N",ngrid,ngrid,orderFD,orderFD,1.0,dermat[1],0,1+2*orderFD,dataCenter.pressure,0,1,0.0,dpressure,0,1);

    // find integration weights    
    int orderInt = 1; // 1 leads to 4th order accuracy
    double[] w = Numerics.findIntegrationWeights(dataCenter.grid,orderInt);

    double num = 0.0, den = 0.0;
    for(int i=0; i<ngrid; i++) {
      if (dataCenter.grid[i] > 1.00) continue; // skip atmosphere
      num-=w[i]*dataCenter.gravity[i]*(dpressure[i]/dataCenter.rho[i]+dataCenter.gravity[i]);
      den+=w[i]*dataCenter.gravity[i]*dataCenter.gravity[i];
    }
    return dataCenter.G*(1.0+num/den);
  }

  /***********************************************************************
  * This subroutine calculates a radial grid appropriate for p or g modes,
  * depending on the values of c1, c2 and c3.  ngridNew is the new
  * resolution.
  ***********************************************************************/
  private double[] findRadialGrid(int ngridNew, double c1, double c2, double c3, boolean combineGrids) {
    double T_total, T_target, mu, A;
    int i, j, k;

    // Squared sound velocity:
    int ngrid = dataCenter.grid.length;
    double[] c_aux = dataCenter.c2.clone();
    double c_max = 0.0;
    for(i=0; i<ngrid; i++) {
      if (c_aux[i] > c_max) c_max = c_aux[i];
    }
    // Possible saturation on sound velocity:
    /*for(i=0; i<ngrid; i++) {
      if (c_aux[i] < (1e-8*c_max)) c_aux[i] = 1e-8*c_max;
    }*/

    // find mesh distribution function:
    double dlnP, dlnRho;
    double[] H = new double[ngrid];
    for(i=1; i<ngrid; i++) {
      dlnP = dataCenter.rho[i]*dataCenter.gravity[i]/dataCenter.pressure[i];
      dlnRho = dataCenter.grid[i]*dataCenter.drho[i]/dataCenter.rho[i];
      H[i] = Math.sqrt(1.0+c1/c_aux[i]+c3*dlnP*dlnP
    	   + c2*Math.abs(dataCenter.N2[i])/(dataCenter.grid[i]*dataCenter.grid[i]));
    }
    mu = (dataCenter.grid[2]-dataCenter.grid[0])
       / (dataCenter.grid[2]-dataCenter.grid[1]);
    H[0]= mu*H[1]+(1.0-mu)*H[2]; // linear extrapolation

    // find antiderivative of H (using trapezoidal rule):
    double[] TT = new double[ngrid];
    TT[0] = 0.0;
    for(i=1; i<ngrid; i++)
      TT[i]=TT[i-1]+(dataCenter.grid[i]-dataCenter.grid[i-1])*(H[i-1]+H[i])/2.0;

    // find total integral:
    T_total = TT[ngrid-1]-TT[0];

    // define new grid:
    double[] r = new double[ngridNew];
    r[0] = dataCenter.grid[0];
    r[ngridNew-1] = dataCenter.grid[ngrid-1];
    j = 0;
    for(i=1; i<ngridNew-1; i++) {
      T_target = T_total*(double)(i)/(double)ngridNew;
      while (TT[j] < T_target) j++;
      A = T_target-TT[j-1];
      r[i] = dataCenter.grid[j-1] + 2.0*A
           / (H[j-1]+Math.sqrt(H[j-1]*H[j-1] 
           + 2.0*A*(H[j]-H[j-1])/(dataCenter.grid[j]-dataCenter.grid[j-1])));
    }

    // sanity check:
    for (i=1; i<ngridNew; i++) {
      if (r[i] < r[i-1]) {
    	System.out.println("WARNING: new grid is non-monotonic. Re-ordering");
        Arrays.sort(r);
        break;
      }
    }

    // combine with previous grid (is combineGrids==true)
    if (combineGrids) {
      double[] r_aux = new double[ngridNew+ngrid-2];  // the end points are duplicated
      r_aux[0] = r[0];
      r_aux[ngridNew+ngrid-3] = r[ngridNew-1];
      j = k = 1;
      for(i=1; i<ngridNew-1; i++){
        while(dataCenter.grid[j] < r[i]) {
          r_aux[k] = dataCenter.grid[j];
          j++;
          k++;
        }
        r_aux[k] = r[i];
        k++;
      }
      // the last part:
      for(i=j; i<ngrid-1; i++) {
        r_aux[k] = dataCenter.grid[i];
        k++;
      }

      r = r_aux;
      // sanity check:
      for (i=1; i<r.length; i++) {
        if (r[i] < r[i-1]) {
    	  System.out.println("WARNING: new grid is non-monotonic. Re-ordering");
          Arrays.sort(r);
          break;
        }
      }
    }

    // find multiple points and remove them:
    r = Numerics.removeMpoints(Numerics.findMpoints(r),r);

    // return result:
    return r;
  }

  /***************************************************************************
  *  This writes the model in the format given by the GUI.
  ***************************************************************************/
  private void writeModel(File aFile, String modelType, JComponent c) {
    FortranBinaryOutputStream binaryOutput;
 
    try {

      // binary case:
      if (modelType.contains("endian")) {
        boolean isNative = modelType.contains("native");
        FileOutputStream out = new FileOutputStream(aFile);
        binaryOutput = new FortranBinaryOutputStream(out,true,isNative);
        if (modelType.startsWith("AMDL")) {
          writeAMDLmodel(binaryOutput);
        } else if (modelType.startsWith("MOD")) {
          writeBinaryMODmodel(binaryOutput);
        } else {
          // This should never happen:
          JOptionPane.showMessageDialog(
            c,  // the parent component
            "Unknown or unimplemented format: "+modelType+".\n"+
            "Please contact authors of InversionKit.",
            "Format error",
            JOptionPane.ERROR_MESSAGE
          );
        }
        return;
      }

      // ascii case:
      if (modelType.equals("FAMDL")) {
        writeFAMDLmodel(new PrintWriter(new FileWriter(aFile)));
        return;
      }

      if (modelType.equals("MOD (ascii)")) {
        writeAsciiMODmodel(new PrintWriter(new FileWriter(aFile)));
        return;
      }

      // This should never happen:
      JOptionPane.showMessageDialog(
        c,  // the parent component
        "Unknown or unimplemented format: "+modelType+".\n"+
        "Please contact authors of InversionKit.",
        "Format error",
        JOptionPane.ERROR_MESSAGE
      );

    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        c,  // the parent component
        "Unable to write to: "+aFile.getName()+"\n",
        "Error writing to file",
        JOptionPane.ERROR_MESSAGE
      );
    }
  }    

  /***************************************************************************
  *  This interpolates and reconstructs the model from A and rho the model.
  ***************************************************************************/
  private void interpolateModel(double[] newGrid) {
    int ngrid = newGrid.length, i;
    int ngridOld = dataCenter.grid.length;
    dataCenter.rho = Numerics.interpolate1D_r2(dataCenter.grid,newGrid,dataCenter.rho,true);
    dataCenter.Gamma1 = Numerics.interpolate1D_r2(dataCenter.grid,newGrid,dataCenter.Gamma1,false);
    dataCenter.A = Numerics.interpolate1D_r2(dataCenter.grid,newGrid,dataCenter.A,false);

    // update the grid (it's rather important ...):
    dataCenter.grid = newGrid;

    // find mass:
    dataCenter.mass = findCumulativeMass();

    // find pressure through hydrostatic equilibrium:
    double[] aux = new double[ngrid];
    for(i=1; i<ngrid; i++) aux[i] = -dataCenter.rho[i]*dataCenter.mass[i]/(newGrid[i]*newGrid[i]);
    aux[0] = 0.0;
    int orderInt = 1;
    double[] intMatrix = Numerics.findIntegrationMatrix(newGrid, orderInt);
    dataCenter.pressure = Numerics.applyIntegrationMatrixReverse(intMatrix,aux,dataCenter.pressure[ngridOld-1],orderInt);

    findJCDvariables();
    findOtherVariables();

    // show mass difference:
    System.out.println(dataCenter.mass[ngrid-1]);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This recalculates stellar equilibrium profiles from the rho and Gamma_1
   *  profiles.  This can lead to improved results with the variational
   *  principle, but also irregularities in the A and N2 profiles. */
  /////////////////////////////////////////////////////////////////////////////
  public void tweakModel(){
    int ngrid = dataCenter.grid.length, i;
    dataCenter.mass = findCumulativeMass();

    // find pressure through hydrostatic equilibrium:
    double[] aux = new double[ngrid];
    for(i=1; i<ngrid; i++) {
      aux[i] = -dataCenter.rho[i]*dataCenter.mass[i]/(dataCenter.grid[i]*dataCenter.grid[i]);
    }
    aux[0] = 0.0;
    int orderInt = 2;
    double[] intMatrix = Numerics.findIntegrationMatrix(dataCenter.grid, orderInt);
    dataCenter.pressure = Numerics.applyIntegrationMatrixReverse(intMatrix,aux,dataCenter.pressure[ngrid-1],orderInt);

    // tell findOtherVariables() to recalculate A
    dataCenter.A = null;

    findJCDvariables();
    findOtherVariables();
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This inserts a double point at the specified index. */
  /////////////////////////////////////////////////////////////////////////////
  public void addDoublePoint(int index){

    // the order used to extrapolate from either side:
    int order = 3;  // = degree + 1;

    // sanity checks:
    if ((index < order)||(index >= dataCenter.grid.length-order)) {
      JOptionPane.showMessageDialog(
        DensityPlot,
        "I cannot insert a double point here - it is too\n"+
        "close to the edge.  Please try somewhere else.",
        "Unsuitable point",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    int ngrid = dataCenter.grid.length, i, j;
    for(i=index-order; i<index-1;i++) {
      if (dataCenter.grid[i] == dataCenter.grid[i+1]) {
        JOptionPane.showMessageDialog(
          DensityPlot,
          "I cannot insert a double point here - it is too close\n"+
          "to another double point.  Please try somewhere else.",
          "Unsuitable point",
          JOptionPane.ERROR_MESSAGE
        );
        return;
      }
    }
    for(i=index+1; i<index+order-1;i++) {
      if (dataCenter.grid[i] == dataCenter.grid[i+1]) {
        JOptionPane.showMessageDialog(
          DensityPlot,
          "I cannot insert a double point here - it is too close\n"+
          "to another double point.  Please try somewhere else.",
          "Unsuitable point",
          JOptionPane.ERROR_MESSAGE
        );
        return;
      }
    }

    // find interpolation coefficients:
    double[] coeffBelow = new double[order+1];
    double[] coeffAbove = new double[order+1];
    for(i=1; i<=order; i++) {
      coeffBelow[i] = coeffAbove[i] = 1.0;
      for(j=1; j<=order; j++) {
        if (j==i) continue;
        coeffBelow[i]*=(dataCenter.grid[index]-dataCenter.grid[index-j])/(dataCenter.grid[index-i]-dataCenter.grid[index-j]);
        coeffAbove[i]*=(dataCenter.grid[index]-dataCenter.grid[index+j])/(dataCenter.grid[index+i]-dataCenter.grid[index+j]);
      }
    }
    
    // insert supplementary point:
    int ngridNew = ngrid + 1;

    double[] newGrid = new double[ngridNew];
    for(i=0;i<=index;i++) newGrid[i]=dataCenter.grid[i];
    for(i=index;i<ngrid;i++) newGrid[i+1]=dataCenter.grid[i];
    
    double[] newRho = new double[ngridNew];
    for(i=0;i<index;i++) newRho[i]=dataCenter.rho[i];
    for(i=index+1;i<ngrid;i++) newRho[i+1]=dataCenter.rho[i];
    newRho[index] = newRho[index+1] = 0.0;
    for(i=1;i<=order;i++) {
      newRho[index]+=dataCenter.rho[index-i]*coeffBelow[i];
      newRho[index+1]+=dataCenter.rho[index+i]*coeffAbove[i];
    }
    dataCenter.rho = newRho;

    double[] newGamma1 = new double[ngridNew];
    for(i=0;i<index;i++) newGamma1[i]=dataCenter.Gamma1[i];
    for(i=index+1;i<ngrid;i++) newGamma1[i+1]=dataCenter.Gamma1[i];
    newGamma1[index] = newGamma1[index+1] = 0.0;
    for(i=1;i<=order;i++) {
      newGamma1[index]+=dataCenter.Gamma1[index-i]*coeffBelow[i];
      newGamma1[index+1]+=dataCenter.Gamma1[index+i]*coeffAbove[i];
    }
    dataCenter.Gamma1 = newGamma1;

    double[] dpressure = new double[ngridNew];
    for(i=0;i<=index;i++) dpressure[i]=dataCenter.pressure[i];
    for(i=index;i<ngrid;i++) dpressure[i+1]=dataCenter.pressure[i];
    double Psurf = dataCenter.pressure[ngrid-1];
    dataCenter.grid = newGrid;
    ngrid = ngridNew;

    // recalculate cumulative mass (this should have a very small impact): 
    dataCenter.mass = findCumulativeMass();

    // find pressure through hydrostatic equilibrium:
    double[] aux = new double[ngrid];
    for(i=1; i<ngrid; i++) {
      aux[i] = -dataCenter.rho[i]*dataCenter.mass[i]/(dataCenter.grid[i]*dataCenter.grid[i]);
    }
    aux[0] = 0.0;
    int orderInt = 2;
    double[] intMatrix = Numerics.findIntegrationMatrix(dataCenter.grid, orderInt);
    dataCenter.pressure = Numerics.applyIntegrationMatrixReverse(intMatrix,aux,Psurf,orderInt);

    // display difference on pressure profile:
    for(i=0; i<ngrid; i++) dpressure[i] = dpressure[i]/dataCenter.pressure[i] - 1.0;
    JFrame f = new JFrame("Model reader");
    IPlot aPlot = new IPlot("Pressure modification","Radius, r/R","Relative error");
    aPlot.addDataSet(dataCenter.grid,dpressure);
    aPlot.findRange();
    f.getContentPane().add("Center",aPlot);
    f.setSize(600,600);
    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    f.setVisible(true);
   
    // tell findOtherVariables() to recalculate A
    dataCenter.A = null;

    findJCDvariables();
    findOtherVariables();

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.clearKernels();
    updateSLabels();
    refreshPlots(true);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This removes the specified point. */
  /////////////////////////////////////////////////////////////////////////////
  public void removePoint(int index){

    int ngrid = dataCenter.grid.length;

    // sanity checks:
    if ((index < 0)||(index >= ngrid)) {
      JOptionPane.showMessageDialog(
        DensityPlot,
        "An unexpected error occured.  I wasn't able to remove\n"+
        "this point.  Please contact the authors of InversionKit.",
        "Unexpected error",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    // insert supplementary point:
    int ngridNew = ngrid - 1, i;

    double[] newGrid = new double[ngridNew];
    for(i=0;i<index;i++) newGrid[i]=dataCenter.grid[i];
    for(i=index+1;i<ngrid;i++) newGrid[i-1]=dataCenter.grid[i];
    
    double[] newRho = new double[ngridNew];
    for(i=0;i<index;i++) newRho[i]=dataCenter.rho[i];
    for(i=index+1;i<ngrid;i++) newRho[i-1]=dataCenter.rho[i];
    dataCenter.rho = newRho;

    double[] newGamma1 = new double[ngridNew];
    for(i=0;i<index;i++) newGamma1[i]=dataCenter.Gamma1[i];
    for(i=index+1;i<ngrid;i++) newGamma1[i-1]=dataCenter.Gamma1[i];
    dataCenter.Gamma1 = newGamma1;

    double[] dpressure = new double[ngridNew];
    for(i=0;i<index;i++) dpressure[i]=dataCenter.pressure[i];
    for(i=index+1;i<ngrid;i++) dpressure[i-1]=dataCenter.pressure[i];
    double Psurf = dataCenter.pressure[ngrid-1];
    dataCenter.grid = newGrid;
    ngrid = ngridNew;

    // recalculate cumulative mass (this should have a very small impact): 
    dataCenter.mass = findCumulativeMass();

    // find pressure through hydrostatic equilibrium:
    double[] aux = new double[ngrid];
    for(i=1; i<ngrid; i++) {
      aux[i] = -dataCenter.rho[i]*dataCenter.mass[i]/(dataCenter.grid[i]*dataCenter.grid[i]);
    }
    aux[0] = 0.0;
    int orderInt = 2;
    double[] intMatrix = Numerics.findIntegrationMatrix(dataCenter.grid, orderInt);
    dataCenter.pressure = Numerics.applyIntegrationMatrixReverse(intMatrix,aux,Psurf,orderInt);

    // display difference on pressure profile:
    for(i=0; i<ngrid; i++) dpressure[i] = dpressure[i]/dataCenter.pressure[i] - 1.0;
    JFrame f = new JFrame("Model reader");
    IPlot aPlot = new IPlot("Pressure modification","Radius, r/R","Relative error");
    aPlot.addDataSet(dataCenter.grid,dpressure);
    aPlot.findRange();
    f.getContentPane().add("Center",aPlot);
    f.setSize(600,600);
    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    f.setVisible(true);
   
    // tell findOtherVariables() to recalculate A
    dataCenter.A = null;

    findJCDvariables();
    findOtherVariables();

    dataCenter.rotationInversionUpToDate = false;
    dataCenter.structInversionUpToDate = false;
    dataCenter.meanDensityEstimateUpToDate = false;
    dataCenter.acousticTimeEstimateUpToDate = false;
    dataCenter.smallSeparationEstimateUpToDate = false;
    dataCenter.clearKernels();
    updateSLabels();
    refreshPlots(true);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Private class which describes an additional mouse listener for the HR
  // plot.
  /////////////////////////////////////////////////////////////////////////////
  private class HRmouseListener implements MouseListener, MouseMotionListener {

    /** Unused method which still needs to be
    *   implemented. */
    public void mouseClicked(MouseEvent e) {
    }

    /** Unused method which still needs to be
    *   implemented. */
    public void mouseEntered(MouseEvent e) {
    }

    /** Unused method which still needs to be
    *   implemented. */
    public void mouseMoved(MouseEvent e) {
    }

    /** Method which erases the previous rectangle
    *   and draws a new one when the mouse is being
    *   dragged across the plot. */
    public void mouseDragged(MouseEvent e) {
    }

    /** Method which erases the last rectangle
    *   when the mouse leaves the plot area,
    *   and cancels the zoom. */
    public void mouseExited(MouseEvent e) {
    }

    /** Method which keeps track of the point
    *   from where the mouse is being dragged */
    public void mousePressed(MouseEvent e) {
    }

    /** Method which performs the zoom-in once
    *   the mouse button has been released. */
    public void mouseReleased(MouseEvent e) {
      if (e.getButton() == MouseEvent.BUTTON3) {
        float x = (float)e.getX();
        float y = (float)e.getY();
        float xx, yy;
        float distmin = Float.MAX_VALUE, dist;
        int imin = -1;
        for(int i=0; i<teff.size(); i++) {
          xx = HRplot.getGraphicsArea().xCoor(teff.get(i));
          yy = HRplot.getGraphicsArea().yCoor(lum.get(i));
          dist = (x-xx)*(x-xx) + (y-yy)*(y-yy);
          if (dist < distmin) {
            distmin = dist;
            imin = i;
          }
        }

        Object[] message = new Object[5];
        message[0] = "Do you want to load the following model?";
        message[1] = "   "+models.get(imin);
        message[2] = " ";
        message[3] = "Model format:";
        message[4] = new JComboBox(readModelFormats);
        ((JComboBox)message[4]).setSelectedItem(currentReadFormat);
      
        Object[] options = new Object[2];
        options[0] = "Load";
        options[1] = "Cancel";

        JOptionPane optionPane = new JOptionPane(
          message,                      // the dialog message array
          JOptionPane.QUESTION_MESSAGE, // message type
          JOptionPane.DEFAULT_OPTION,   // option type
          null,                         // optional icon, use null to use the default icon
          options,                      // options string array, will be made into buttons
          options[0]                    // default option
        );

        JDialog dialog = optionPane.createDialog(LoadInternetModel, "Download model");
        dialog.setSize(400,200);
        dialog.setResizable(true);
        dialog.setVisible(true);
        
        Object selectedValue = optionPane.getValue();
        if (selectedValue == options[0]) {
          currentReadFormat =  (String)((JComboBox)message[4]).getSelectedItem();
          File aFile = new File(models.get(imin));
          if (aFile.exists()) {
            readModel(models.get(imin),currentReadFormat,false,HRplot);
            HRplot.replaceDataSet(1,new double[]{teff.get(imin)},new double[]{lum.get(imin)},
                                  Plot.noLine,Plot.square);
            HRplot.repaint();
          } else {
            JOptionPane.showMessageDialog(
              null,
              "File \""+models.get(imin)+"\" doesn't exist.",
              "File not found",
              JOptionPane.ERROR_MESSAGE
            );
          }
        }
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Private class which describes an additional mouse listener for the 
  // density profile plot.  This can be used to add in double points.
  /////////////////////////////////////////////////////////////////////////////
  private class ModelMouseListener implements MouseListener, MouseMotionListener {

    /** A local copy of the plot window to which this mouse listener is added: */
    IPlot myPlot;

    /** A constructor */
    ModelMouseListener(IPlot aPlot) {
      myPlot = aPlot;
    }

    /** Unused method which still needs to be
    *   implemented. */
    public void mouseClicked(MouseEvent e) {
    }

    /** Unused method which still needs to be
    *   implemented. */
    public void mouseEntered(MouseEvent e) {
    }

    /** Unused method which still needs to be
    *   implemented. */
    public void mouseMoved(MouseEvent e) {
    }

    /** Method which erases the previous rectangle
    *   and draws a new one when the mouse is being
    *   dragged across the plot. */
    public void mouseDragged(MouseEvent e) {
    }

    /** Method which erases the last rectangle
    *   when the mouse leaves the plot area,
    *   and cancels the zoom. */
    public void mouseExited(MouseEvent e) {
    }

    /** Method which keeps track of the point
    *   from where the mouse is being dragged */
    public void mousePressed(MouseEvent e) {
    }

    /** Method which performs the zoom-in once
    *   the mouse button has been released. */
    public void mouseReleased(MouseEvent e) {
      if (e.getButton() == MouseEvent.BUTTON3) {
        if (!dataCenter.modelLoaded) return;
        int nn = myPlot.getNDataSets();
        if (nn == 0) return;
        double[] xarray = myPlot.getDataSetX(nn-1);
        double[] yarray = myPlot.getDataSetY(nn-1);
        float x = (float)e.getX();
        float y = (float)e.getY();
        float xx, yy;
        float distmin = Float.MAX_VALUE, dist;
        int imin = -1;
        for(int i=0; i<xarray.length; i++) {
          xx = myPlot.getGraphicsArea().xCoor(xarray[i]);
          yy = myPlot.getGraphicsArea().yCoor(yarray[i]);
          dist = (x-xx)*(x-xx) + (y-yy)*(y-yy);
          if (dist < distmin) {
            distmin = dist;
            imin = i;
          }
        }
        // check to see what the user wants to do with the selected point:
        String[] options = new String[3];
        options[0] = "Add double point";
        options[1] = "Remove point";
        options[2] = "Cancel";
        int result = JOptionPane.showOptionDialog(
               DensityPlot,                  // parent component
               "You have selected the following point:\n"+
               "   r["+imin+"] = "+dataCenter.grid[imin]+" R\n"+
               "What should I do to this point?",  // message
               "Selected point",             // title
               JOptionPane.DEFAULT_OPTION,   // option type
               JOptionPane.QUESTION_MESSAGE, // message type
               null,                         // optional icon, use null to use the default icon
               options,                      // different options
               options[0]                    // default option
        );
        if (result == 0) {
          addDoublePoint(imin);
        } else if (result == 1) {
          removePoint(imin);
        } else {
          return;
        }
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Private class used to select the file format of a model which is being
  // read or saved.  The description field will typically contain:
  // "AMDL", "FAMDL", etc.
  /////////////////////////////////////////////////////////////////////////////
  private class ModelFormatFilter extends FileFilter {
    private String description;

    public ModelFormatFilter(String description) {
        this.description = description;
    }

    public boolean accept(File file) {
        return true;
    }

    public String getDescription() {
        return description;
    }
  }
}
