/**********************************************************************
* Some revisions were made to this file: the native endianness is
* determined using ByteOrder.nativeOrder() rather than going through
* the machine architecture.  As a result, some of the comments below
* are no longer applicable.  To get an original version, see:
*   http://docjar.com/docs/api/org/fudaa/dodico/fortran/NativeBinaryInputStream.html
*
* Furthermore, the conversion from byte arrays to various types is now
* done using the ByteBuffer class, which allows a straightforward way
* of handling endianness.
*
* D. Reese, 24/9/2009
**********************************************************************/

/*
 * @file         NativeBinaryInputStream.java
 * @creation     1998-08-19
 * @modification $Date: 2002/12/17 16:47:12 $
 * @license      GNU General Public License 2
 * @copyright    (c)1998-2001 CETMEF 2 bd Gambetta F-60231 Compiegne
 * @mail         devel@fudaa.org
 */

package org.fudaa.dodico.fortran;

import java.io.*;
import java.nio.ByteOrder;
import java.nio.ByteBuffer;

/**
 * Une classe etendant DataInputStream et permettant de facilement lire
 * des fichiers binaires dependants de la machine (sparc, i386, ...)
 * Attention: cette classe est utilisee par FortranBinaryInputStream. A chaque
 * appel de fonction le compteur de FortranBinaryInputStream est incremente.
 * Pour l'instant ,il ne faut pas chainer les appels de methodes.<br/>
 * Ne pas faire:<br/>
 * public int readToto()
 * {
 *   readToto2();
 * }
 * il y aura eu 2 increments dans FortranBinaryInputStream ( pour readToto
 * et pour readToto2)) ce qui fausse la lecture.
 *
 * @version      $Revision: 1.8 $ $Date: 2002/12/17 16:47:12 $ by $Author: deniger $
 * @author       Axel von Arnim
 */
public class NativeBinaryInputStream
             extends DataInputStream
{

  /** ceci stocke l'"endianness" utilise dans les operations d'ecriture */
  private ByteOrder byteOrder;

  /** entier renvoye par la methode read */
  private int nbByteLus_;

  /** ceci stocke les bytes obtenus avant conversions vers types primitives */
  byte buf[];

  /** ceci permet la conversion de bytes en types primitives */
  ByteBuffer bBuffer;

  public NativeBinaryInputStream(InputStream _in) throws IOException
  {
    super(_in);
    this.buf=null;
    this.byteOrder = ByteOrder.nativeOrder();
  }

  public NativeBinaryInputStream(InputStream _in, ByteOrder byteOrder) throws IOException
  {
    super(_in);
    this.buf=null;
    this.byteOrder = byteOrder;
  }

  public NativeBinaryInputStream(InputStream _in, boolean isNativeEndian) throws IOException
  {
    super(_in);
    this.buf=null;
    if (isNativeEndian) {
      this.byteOrder = ByteOrder.nativeOrder();
    } else {
      if (ByteOrder.nativeOrder().equals(ByteOrder.BIG_ENDIAN)) {
        this.byteOrder = ByteOrder.LITTLE_ENDIAN;
      } else {
        this.byteOrder = ByteOrder.BIG_ENDIAN;
      }
    }
  }

  /**
   * Indique si la fin du fichier est atteinte. Si  <code>true</code>
   * cela signifie que la derniere lecture est erronnee.
   * @return si fin du fichier.
   */
  public boolean isFinFichier()
  {
    return nbByteLus_==-1;
  }

  public byte readInt_8() throws IOException
  {
    buf=new byte[1];
    nbByteLus_=read(buf);
    return buf[0];
  }

  public short readInt_16() throws IOException
  {
    buf=new byte[2];
    nbByteLus_=read(buf);
    bBuffer = ByteBuffer.wrap(buf).order(byteOrder);
    return bBuffer.getShort();
  }

  public int readInt_32() throws IOException
  {
    buf=new byte[4];
    nbByteLus_=read(buf);
    bBuffer = ByteBuffer.wrap(buf).order(byteOrder);
    return bBuffer.getInt();
  }

  public long readInt_64() throws IOException
  {
    buf=new byte[8];
    nbByteLus_=read(buf);
    bBuffer = ByteBuffer.wrap(buf).order(byteOrder);
    return bBuffer.getLong();
  }

  public float readFloat_32() throws IOException
  {
    buf=new byte[4];
    nbByteLus_=read(buf);
    bBuffer = ByteBuffer.wrap(buf).order(byteOrder);
    return bBuffer.getFloat();
  }

  public double readFloat_64() throws IOException
  {
    buf=new byte[8];
    nbByteLus_=read(buf);
    bBuffer = ByteBuffer.wrap(buf).order(byteOrder);
    return bBuffer.getDouble();
  }
}
