/**********************************************************************
* Some revisions were made to this file: the native endianness is
* determined using ByteOrder.nativeOrder() rather than going through
* the machine architecture.  As a result, some of the comments below
* are no longer applicable.  To get an original version, see:
*   http://docjar.com/docs/api/org/fudaa/dodico/fortran/NativeBinaryOutputStream.html
*
* Furthermore, the conversion from various types to byte arrays seemed
* to be incorrect.  Consequently, this has been replaced with conversions
* using the ByteBuffer class, which also handles endianness in a
* transpanrent way.
*
* D. Reese, 24/9/2009
**********************************************************************/

/*
 * @file         NativeBinaryOutputStream.java
 * @creation     1998-08-19
 * @modification $Date: 2001/09/11 14:29:51 $
 * @license      GNU General Public License 2
 * @copyright    (c)1998-2001 CETMEF 2 bd Gambetta F-60231 Compiegne
 * @mail         devel@fudaa.org
 */

package org.fudaa.dodico.fortran;

import java.io.*;
import java.nio.ByteOrder;
import java.nio.ByteBuffer;

/**
 * Une classe etendant DataOutputStream et permettant de facilement
 * ecrire des fichiers binaires dependants de la machine (sparc, i386, ...)
 *
 * @version      $Revision: 1.11 $ $Date: 2001/09/11 14:29:51 $ by $Author: deniger $
 * @author       Axel von Arnim 
 */
public class NativeBinaryOutputStream
       extends DataOutputStream
{
  /** ceci stocke l'"endianness" utilise dans les operations d'ecriture */
  private ByteOrder byteOrder;

  /** ceci stocke les bytes obtenus apres conversions de types primitives */
  byte buf[];

  /** ceci permet la conversion de types primitives en bytes */
  ByteBuffer bBuffer;

  public NativeBinaryOutputStream(OutputStream _out) throws IOException
  {
    super(_out);
    this.buf=null;
    this.byteOrder = ByteOrder.nativeOrder();
  }

  public NativeBinaryOutputStream(OutputStream _out, ByteOrder byteOrder) throws IOException
  {
    super(_out);
    this.buf=null;
    this.byteOrder = byteOrder;
  }

  public NativeBinaryOutputStream(OutputStream _out, boolean isNativeEndian) throws IOException
  {
    super(_out);
    this.buf=null;
    if (isNativeEndian) {
      this.byteOrder = ByteOrder.nativeOrder();
    } else {
      if (ByteOrder.nativeOrder().equals(ByteOrder.BIG_ENDIAN)) {
        this.byteOrder = ByteOrder.LITTLE_ENDIAN;
      } else {
        this.byteOrder = ByteOrder.BIG_ENDIAN;
      }
    }
  }

  public void writeInt_8(byte int_8) throws IOException
  {
    buf=new byte[1];
    buf[0]=int_8;
    write(buf);
  }
  
  public void writeInt_16(short int_16) throws IOException
  {
    buf= new byte[2];
    bBuffer = ByteBuffer.wrap(buf).order(byteOrder);
    bBuffer.putShort(int_16);
    write(buf);
  }

  public void writeInt_32(int int_32) throws IOException
  {
    buf= new byte[4];
    bBuffer = ByteBuffer.wrap(buf).order(byteOrder);
    bBuffer.putInt(int_32);
    write(buf);
  }
  
  public void writeInt_64(long int_64) throws IOException
  {
    buf = new byte[8];
    bBuffer = ByteBuffer.wrap(buf).order(byteOrder);
    bBuffer.putLong(int_64);
    write(buf);
  }
  
  public void writeFloat_32(float float_32) throws IOException
  {
    // B.M. Java autorise la valeur -0.0 qui perturbe le codage en bytes.
    // L'interet de cette valeur etant tres relative, on la remplace par 0.0
    if (float_32==-0.0f) float_32=0.0f;
    writeInt_32(Float.floatToIntBits(float_32));
  }

  public void writeFloat_64(double float_64) throws IOException
  {
    // B.M. Java autorise la valeur -0.0 qui perturbe le codage en bytes.
    // L'interet de cette valeur etant tres relative, on la remplace par 0.0
    if (float_64==-0.0) float_64=0.0;
    writeInt_64(Double.doubleToLongBits(float_64));
  }
}
