/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import org.netlib.blas.Dgemm;
import org.netlib.blas.Dgbmv;
import org.netlib.lapack.Dgbtrf;
import org.netlib.lapack.Dgbtrs;
import java.util.ArrayList;
import java.util.Arrays;

public class Numerics {

  /********************************************************
  * Initialise Chebyshev spectral transformation real2spec
  * for the Chebyshev-Gauss collocation grid. The endpoints
  * are excluded.
  ********************************************************/
  private static double[] initChebReal2Spec(int nr) {
    double[] real2spec = new double[nr*nr];
    double[] temp = new double[4*nr];
    
    for(int i=0; i<4*nr; i++) {
      temp[i] = Math.cos((Math.PI*(double)i)/(double)(2*nr));
    }

    double aux = 2.0/(double)nr;
    for (int i=0; i<nr; i++){
      for (int j=0; j<nr; j++){
        real2spec[i+nr*j] = aux*temp[(i*(2*j+1))%(4*nr)];
      }
    }

    for(int i=0; i<nr; i++) real2spec[0+nr*i]*=0.5;
    
    return real2spec;
  }

  /********************************************************
  * Initialise Chebyshev spectral transformation spec2real
  * for the Chebyshev-Gauss collocation grid. The endpoints
  * are excluded.
  ********************************************************/
  private static double[] initChebSpec2Real(int nr) {
    double[] spec2real = new double[nr*nr];
    double[] temp = new double[4*nr];
    
    for(int i=0; i<4*nr; i++) {
      temp[i] = Math.cos((Math.PI*(double)i)/(double)(2*nr));
    }

    for (int i=0; i<nr; i++){
      for (int j=0; j<nr; j++){
        spec2real[i+nr*j] = temp[(j*(2*i+1))%(4*nr)];
      }
    }

    return spec2real;
  }

  /********************************************************
  * Initialise Chebyshev spectral transformation real2spec
  * for the Chebyshev-Gauss-Lobatto collocation grid. The
  * endpoints are included.
  ********************************************************/
  private static double[] initChebReal2SpecLobatto(int nr) {
    double[] real2spec = new double[nr*nr];
    double[] temp = new double[2*nr];
    
    for(int i=0; i<2*nr-2; i++) {
      temp[i] = Math.cos((Math.PI*i)/(nr-1));
    }

    double aux = 2.0/(double)(nr-1);
    for (int i=0; i<nr; i++){
      for (int j=0; j<nr; j++){
        real2spec[i+nr*j] = aux*temp[(i*j)%(2*nr-2)];
      }
    }
    
    for(int i=0; i<nr; i++) {
      real2spec[0+nr*i]   *=0.5;
      real2spec[nr-1+nr*i]*=0.5;
      real2spec[i+nr*0]   *=0.5;
      real2spec[i+nr*(nr-1)]*=0.5;
    }
    
    return real2spec;
  }

  /********************************************************
  * Initialise Chebyshev spectral transformation spec2real
  * for the Chebyshev-Gauss-Lobatto collocation grid. The
  * endpoints are included.
  ********************************************************/
  private static double[] initChebSpec2RealLobatto(int nr) {
    double[] spec2real = new double[nr*nr];
    double[] temp = new double[2*nr];
    
    for(int i=0; i<2*nr-2; i++) {
      temp[i] = Math.cos((Math.PI*i)/(nr-1));
    }

    for (int i=0; i<nr; i++){
      for (int j=0; j<nr; j++){
        spec2real[i+nr*j] = temp[(i*j)%(2*nr-2)];
      }
    }

    return spec2real;
  }

  /********************************************************
  * Initialise Chebyshev-based 1st order derivation matrix:
  ********************************************************/
  public static double[] initDerive(int nr) {
    double[] derive    = new double[nr*nr];
    double[] derive_spec = new double[nr*nr];
    double[] derive_temp = new double[nr*nr];
    double[] real2spec = initChebReal2Spec(nr);
    double[] spec2real = initChebSpec2Real(nr);
    
    for(int i=0; i<nr; i++) {
      for(int j=i+1; j<nr; j+=2) {
        derive_spec[i+nr*j] = -(double)(4*j);
      }
    }
    
    for(int i=0; i<nr; i++){derive_spec[0+nr*i]*=0.5;}

    // Use BLAS matrix multiplication
    // (add an offset of 0 to all array arguments)
    Dgemm.dgemm("N","N",nr,nr,nr,1.0,derive_spec,0,nr,real2spec,0,nr,
                0.0,derive_temp,0,nr);
    Dgemm.dgemm("N","N",nr,nr,nr,1.0,spec2real,0,nr,derive_temp,0,nr,
                0.0,derive,0,nr);
    return derive;
  }

  /********************************************************
  * Initialise Chebyshev-based 2nd order derivation matrix:
  ********************************************************/
  public static double[] initDDerive(int nr) {
    double[] derive    = initDerive(nr);
    double[] dderive   = new double[nr*nr];
    Dgemm.dgemm("N","N",nr,nr,nr,1.0,derive,0,nr,derive,0,nr,
                0.0,dderive,0,nr);
    return dderive;
  }

  /**************************************************************************
  *  This subroutine sets up a derivation matrix for functions expressed on
  *  the Gauss-Lobatto Chebyshev collocation grid.
  ***************************************************************************
  * Variables:
  *
  * rspan    = the span of the underlying grid (rspan = grid[nr-1]-grid[0])
  * nr       = the number of points in the grid (the grid is assumed to be the
  *            Gauss-Lobatto Chebyshev collocation grid)
  * der_max  = highest order derivative
  ***************************************************************************
  * The output: derive[0:der_max][nr*nr] (where nr = grid.length)
  *
  * derive = a two dimensional matrix which contains the derivation matrices
  *          in general (non-banded) form.  The first coordinate corresponds
  *          to the derivation order.  The second coordinate gives the
  *          position (i+nr*j) in the matrix  (where i=row, j=column).
  **************************************************************************/
  
  public static double[][] initDeriveCheb(double rspan, int nr, int der_max) {
  
    // check parameters:
    if (der_max < 1)
      throw new IllegalArgumentException("Please set der_max > 0 in initDeriveCheb");
    if (nr < 2)
      throw new IllegalArgumentException("Please set nr > 1 in initDeriveCheb");

    // set up derivation matrix:
    double[][] derive    = new double[der_max+1][nr*nr];
    derive[1] = initDerive(nr);
    for(int i=0; i< nr; i++) {
      derive[0][i+nr*i] = 1.0;
      for (int j=0; j<nr; j++) derive[1][i+nr*j]/=rspan;
    }
    for(int i=2; i<= der_max; i++)
      Dgemm.dgemm("N","N",nr,nr,nr,1.0,derive[1],0,nr,derive[i-1],0,nr,
                  0.0,derive[i],0,nr);
    return derive;
  }

  /********************************************************
  * Initialise integration weights for a Chebyshev 
  * Gauss-Lobatto collocation grid (i.e. using
  * spectral methods):
  ********************************************************/
  public static double[] initWeightsCheb(int n_grid) {
    double[] weights = new double [n_grid];
    double[] real2spec = initChebReal2Spec(n_grid);
    for(int i=0; i<n_grid; i++) {
      weights[i] = 0.0;
      for (int j=0; j<n_grid; j+=2) {
        weights[i]+=real2spec[j+n_grid*i]/(double)(1-j*j);
      }
    }
    return weights;
  }

  /**************************************************************************
  *  This subroutine sets up a banded derivation matrices (which is stored in
  *  banded form) using nth order finite difference schemes.  The
  *  derivatives are calculated with respect to an (almost) arbitrary grid.
  *  
  *  It calculates the derivation coefficients based on Lagrange interpolation
  *  polynomials (cf. Fornberg, 1988).
  ***************************************************************************
  * Variables:
  *
  * grid     = underlying on which the derivatives are calculated (grid(1:nr))
  *            Warning: this grid must not contain two identical points.
  * order    = order of the scheme, divided by 2.  Note:  (2.order >= der_max)
  * der_max  = highest order derivative
  ***************************************************************************
  * The output: derive[0:der_max][nr*ndiagonals] (where nr = grid.length)
  *
  * derive = a two dimensional matrix which contains the derivation matrices
  *          in general (non-banded) form.  The first coordinate corresponds
  *          to the derivation order.  The second coordinate gives the
  *          position (i+nr*j) in the matrix  (where i=row, j=column).
  **************************************************************************/
  public static double[][] initDeriveFDBand(double[] grid, int order, int der_max) {

    int start, finish, i, j, k, l, n;
    double prdct;

    // check parameters:
    if (order < der_max) 
      throw new IllegalArgumentException("Please set order >= der_max in initDeriveFD");
    if (order < 0) 
      throw new IllegalArgumentException("Please set order > 0 in initDeriveFD");
    if (der_max < 1)
      throw new IllegalArgumentException("Please set der_max > 0 in initDeriveFD");

    // look for multiple points and determine domains:
    ArrayList<Integer> domains = findDomains(grid);

    // check to make sure domains are large enough:
    if (findMinDomainSize(domains) < order)
      throw new IllegalArgumentException("Please set order <= min domain size");

    int nr = grid.length;
    
    // This is rather confusing but derive[der][i][j] = derive[der][i-j+order+(2*order+1)*j]
    // This is a banded matrix stored in banded form, which itself is stored as a vector,
    // according to its columns.
    double[][] derive = new double[der_max+1][nr*(2*order+1)];
    
    double[] a = new double[1+der_max];
    double[] mu = new double[2*order+1];

    for(i=0; i<nr; i++) derive[0][order+(2*order+1)*i] = 1.0;

    for(n=0;n<domains.size();n+=2) {
      for(i=domains.get(n);i<=domains.get(n+1);i++) {
        start  = -order;
        finish =  order;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the
        // edges, but preserves the banded form of the derivation matrix.
        if ((i+start) < domains.get(n)) start = domains.get(n) - i;
        if ((i+finish) > domains.get(n+1)) finish = domains.get(n+1) - i;


        for(j=start; j<=finish; j++) mu[j+order]=grid[j+i]-grid[i];
        
        for(j=start; j<=finish; j++) {

          // Initialise Lagrange polynomial
          a[0] = 1.0;
          for(l=1; l <= der_max; l++) a[l] = 0.0;
          
          prdct = 1.0;
          for(k=start; k<j; k++)     prdct*=(mu[j+order]-mu[k+order]);
          for(k=j+1; k<=finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]/=prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(k=start; k<j; k++) {
            for(l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
            a[0]*=(-mu[k+order]);
          }
          for(k=j+1; k<=finish; k++) {
            for(l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
            a[0]*=(-mu[k+order]);
          }
          
          // The coeffecients a(l) of the Lagrange polynomial are the
          // derivation coefficients divided by l!.
          prdct = 1.0;
          for(l=1; l<= der_max; l++) {
            prdct*=l;

            // Reminder: derive[der][i][j] = derive[der][i-j+order+(2*order+1)*j]
            // Here we want derive[l][i][i+j]
            derive[l][order-j+(2*order+1)*(i+j)]=a[l]*prdct;
          }
        }
      }
    }
    return derive;
  }

  /**************************************************************************
  *  This subroutine sets up a banded derivation matrices (which is stored in
  *  banded form) using nth order finite difference schemes.  The
  *  derivatives are calculated with respect to an (almost) arbitrary grid.
  *  As opposed to the previous method, the derivatives are calculated at
  *  carefully selected intermediate points which increase the order of
  *  accuracy by 1 for 1st order derivatives.
  *  
  *  It calculates the derivation coefficients based on Lagrange interpolation
  *  polynomials (cf. Fornberg, 1988).
  ***************************************************************************
  * Variables:
  *
  * grid     = underlying on which the derivatives are calculated (grid(1:nr))
  *            Warning: this grid must not contain two identical points.
  * order    = order of the scheme, divided by 2.  Note:  (2.order >= der_max)
  * der_max  = highest order derivative
  ***************************************************************************
  * The output: derive[0:der_max][nr*ndiagonals] (where nr = grid.length)
  *
  * derive = a two dimensional matrix which contains the derivation matrices
  *          in general (non-banded) form.  The first coordinate corresponds
  *          to the derivation order.  The second coordinate gives the
  *          position (i+nr*j) in the matrix  (where i=row, j=column).
  **************************************************************************/
  public static double[][] initDeriveFDBandRoots(double[] grid, int order, int der_max) {

    double[] aa = new double[2*order+1];
    double prdct,acoef, bcoef, xx, dx, num, den;
    int start, finish,i,j,k,l,n;
    double eps = 1e-14;
    int n_iter_max = 100;

    // check parameters:
    if (order < der_max) 
      throw new IllegalArgumentException("Please set order >= der_max in initDeriveFD");
    if (order < 0) 
      throw new IllegalArgumentException("Please set order > 0 in initDeriveFD");
    if (der_max < 1)
      throw new IllegalArgumentException("Please set der_max > 0 in initDeriveFD");

    // look for multiple points and determine domains:
    ArrayList<Integer> domains = findDomains(grid);

    // check to make sure domains are large enough:
    if (findMinDomainSize(domains) < order)
      throw new IllegalArgumentException("Please set order <= min domain size");

    // define new grid:
    int nr = grid.length;
    double[] new_grid = new double[nr-1];

    // This is rather confusing but derive[der][i][j] = derive[der][i-j+order+(2*order)*j]
    // This is a banded matrix stored in banded form, which itself is stored as a vector,
    // according to its columns.
    double[][] derive = new double[der_max+1][2*nr*order];
    
    double[] a = new double[1+der_max];
    double[] mu = new double[2*order+1];

    for(n=0;n<domains.size();n+=2) {
      for(i=domains.get(n);i<domains.get(n+1);i++) {
        start  = -order+1;
        finish = order;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the
        // edges, but preserves the banded form of the derivation matrix.
        if ((i+start) < domains.get(n)) start = domains.get(n) - i;
        if ((i+finish) > domains.get(n+1)) finish = domains.get(n+1) - i;

        // scale polynomial
        acoef = 2.0/(grid[i+1]-grid[i]);
        bcoef = (grid[i]+grid[i+1])/(grid[i]-grid[i+1]);

        // find polynomial coefficients:
        for(j=1;j<=2*order;j++) aa[j] = 0.0;
        aa[0] = 1.0;
        for(j=start;j<=finish;j++) {
          for(k=2*order;k>0;k--) aa[k] = aa[k-1]-(acoef*grid[i+j]+bcoef)*aa[k];
          aa[0]*=-(acoef*grid[i+j]+bcoef);
        }

        // find a root of the polynomial: this will become the new grid point
        xx = 0.0;
        dx = 1.0;

        // Newton type iteration
        j = 0;
        while((Math.abs(dx)>eps)&&(j<n_iter_max)) {
          num = aa[2*order]*(double)(2*order);
          for(k=2*order-1;k>=1;k--) num = num*xx+aa[k]*(double)k;
          den = aa[2*order]*(double)(2*order*(2*order-1));
          for(k=2*order-1;k>=2;k--) den = den*xx+aa[k]*(double)(k*(k-1));
          dx = num/den;
          xx -= dx;
          j++;
        }
        xx = (xx-bcoef)/acoef;

        // sanity check (just in case)
        if ((xx<=grid[i])||(xx>=grid[i+1])) {
          System.out.println("Stray grid point in initDeriveFDBandRoots");
          System.exit(1);
        }

        // assign new grid point
        new_grid[i] = xx;
      }

      for(i=domains.get(n);i<domains.get(n+1);i++) {

        start  = -order+1;
        finish = order;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the
        // edges, but preserves the banded form of the derivation matrix.
        if ((i+start) < domains.get(n)) start = domains.get(n) - i;
        if ((i+finish) > domains.get(n+1)) finish = domains.get(n+1) - i;

        for(j=start; j<=finish; j++) mu[j+order]=grid[j+i]-new_grid[i];
        
        for(j=start; j<=finish; j++) {

          // Initialise Lagrange polynomial
          a[0] = 1.0;
          for(l=1; l <= der_max; l++) a[l] = 0.0;
          
          prdct = 1.0;
          for(k=start; k<j; k++)     prdct*=(mu[j+order]-mu[k+order]);
          for(k=j+1; k<=finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]/=prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(k=start; k<j; k++) {
            for(l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
            a[0]*=(-mu[k+order]);
          }
          for(k=j+1; k<=finish; k++) {
            for(l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
            a[0]*=(-mu[k+order]);
          }
          
          // The coeffecients a(l) of the Lagrange polynomial are the
          // derivation coefficients divided by l!.
          prdct = 1.0;
          derive[0][order-j+(2*order)*(i+j)]=a[0];
          for(l=1; l<= der_max; l++) {
            prdct*=(double)l;

            // Reminder: derive[der][i][j] = derive[der][i-j+order+(2*order)*j]
            // Here we want derive[l][i][i+j]
            derive[l][order-j+(2*order)*(i+j)]=a[l]*prdct;
          }
        }
      }
    }
    return derive;
  }

  /**************************************************************************
  *  This subroutine sets up a banded derivation matrices (which is stored in
  *  general form) using nth order finite difference schemes.  The
  *  derivatives are calculated with respect to an (almost) arbitrary grid.
  *  
  *  It calculates the derivation coefficients based on Lagrange interpolation
  *  polynomials (cf. Fornberg, 1988).
  ***************************************************************************
  * Variables:
  *
  * grid     = underlying on which the derivatives are calculated (grid(1:nr))
  *            Warning: this grid must not contain two identical points.
  * order    = order of the scheme, divided by 2.  Note:  (2.order >= der_max)
  * der_max  = highest order derivative
  ***************************************************************************
  * The output: derive[0:der_max][nr*nr] (where nr = grid.length)
  *
  * derive = a two dimensional matrix which contains the derivation matrices
  *          in general (non-banded) form.  The first coordinate corresponds
  *          to the derivation order.  The second coordinate gives the
  *          position (i+nr*j) in the matrix  (where i=row, j=column).
  **************************************************************************/
  public static double[][] initDeriveFD(double[] grid, int order, int der_max) {

    int start, finish;
    double prdct;

    // check parameters:
    if (order < der_max) 
      throw new IllegalArgumentException("Please set order >= der_max in initDeriveFD");
    if (order < 0) 
      throw new IllegalArgumentException("Please set order > 0 in initDeriveFD");
    if (der_max < 1)
      throw new IllegalArgumentException("Please set der_max > 0 in initDeriveFD");

    // look for multiple points and determine domains:
    ArrayList<Integer> domains = findDomains(grid);

    // check to make sure domains are large enough:
    if (findMinDomainSize(domains) < order)
      throw new IllegalArgumentException("Please set order <= min domain size");

    int nr = grid.length, i, j, k, l, n;
    double[][] derive = new double[der_max+1][nr*nr];
    
    double[] a = new double[1+der_max];
    double[] mu = new double[2*order+1];

    for(i=0; i<nr; i++) derive[0][i+nr*i] = 1.0;

    for(n=0;n<domains.size();n+=2) {
      for(i=domains.get(n); i<=domains.get(n+1); i++) {

        start  = -order;
        finish =  order;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the
        // edges, but preserves the banded form of the derivation matrix.
        if ((i+start) < domains.get(n)) start = domains.get(n) - i;
        if ((i+finish) > domains.get(n+1)) finish = domains.get(n+1) - i;

        for(j=start; j<=finish; j++) mu[j+order]=grid[j+i]-grid[i];
      
        for(j=start; j<=finish; j++) {

          // Initialise Lagrange polynomial
          a[0] = 1.0;
          for(l=1; l <= der_max; l++) a[l] = 0.0;
        
          prdct = 1.0;
          for(k=start; k<j; k++)     prdct*=(mu[j+order]-mu[k+order]);
          for(k=j+1; k<=finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]/=prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(k=start; k<j; k++) {
            for(l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
            a[0]*=(-mu[k+order]);
          }
          for(k=j+1; k<=finish; k++) {
            for(l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
            a[0]*=(-mu[k+order]);
          }
        
          // The coeffecients a(l) of the Lagrange polynomial are the
          // derivation coefficients divided by l!.
          prdct = 1.0;
          for(l=1; l<= der_max; l++) {
            prdct*=l;
            derive[l][i+nr*(i+j)]=a[l]*prdct;
          }
        }
      }
    }

    return derive;
  }

  /******************************************************************************** 
  *  This subroutine sets up an integration matrix in banded form for an arbitrary
  *  grid.  This matrix can then be applied to a function using the 
  *  applyIntegrationMatrix method in order to produce the anti-derivative of the
  *  function.  The conficients are based on the integrals of Lagrange
  *  interpolation polynomials calculated over a sliding window which spans
  *  [i-order, i+order+1].
  ******************************************************************************* 
  * description of variables:
  *
  * grid[]    = grid on which the integration is carried out
  * order     = positive int which gives half the size of the window
  * power     = multiply the function by r^power before carrying out the
  *             integration
  *
  * Output:
  *
  * matrix[]  = matrix which enables one to calculate the anti-derivative of a
  *             function.
  ******************************************************************************/
  public static double[] findIntegrationMatrix(double[] grid) {
    return findIntegrationMatrix(grid, 1, 0);
  }

  public static double[] findIntegrationMatrix(double[] grid, int order) {
    return findIntegrationMatrix(grid, order, 0);
  }

  public static double[] findIntegrationMatrix(double[] grid, int order, int power) {

    if (grid == null) return null;
    int ngrid = grid.length;
    if (ngrid == 0) return new double[0];
    // check parameters:
    if (order < 0)
      throw new IllegalArgumentException("Please set order >= 0");
    if (power < 0)
      throw new IllegalArgumentException("Please set power >= 0");

    // look for multiple points and determine domains:
    ArrayList<Integer> domains = findDomains(grid);

    // check to make sure domains are large enough:
    if (findMinDomainSize(domains) < 1+order)
      throw new IllegalArgumentException("Please set order <= min domain size");

    int i, j, k, l, n;
    int ld = 2*order+2;
    double[] matrix = new double[ngrid*ld];
    
    double[] mu = new double[2*order+2];
    double[] a = new double[2*order+3];

    int start, finish;
    double  prdct, my_sum;

    // make use of Pascal's triangle to calculate binomial coefficients
    // (when power > 0):
    double[] binome = new double[power+1];
    binome[0] = 1.0;
    for(i=1; i<= power; i++){
      binome[i] = 0.0;
      for(j=i; j>0; j--) binome[j]+=binome[j-1];
    }

    for(n=0; n<domains.size(); n+=2) {
      for(i=domains.get(n); i < domains.get(n+1); i++) {

        start  = -order;
        finish = order+1;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the edges
        if ((i+start) < domains.get(n)) start = domains.get(n) - i;
        if ((i+finish) > domains.get(n+1)) finish = domains.get(n+1) - i;

        for(j=start; j <= finish; j++) mu[j+order] = grid[j+i] - grid[i];

        for(j=start; j <= finish; j++) {

          // Initialise Lagrange polynomial
          for(l=1; l<= 2*(order+1); l++) a[l]=0.0;
        
          prdct = 1.0;
          for(k=start; k < j; k++) prdct*=(mu[j+order]-mu[k+order]);
          for(k=j+1; k <= finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]=1.0/prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(k=start; k < j; k++) {
            for(l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }
          for(k=j+1; k <= finish; k++) {
            for(l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }

          // Integrate the Lagrange polynomials, and add the results
          // as weights at the appropriate points.

          for(l=0; l <= (2*order+2); l++)  {
            for(k=0; k<= power; k++) {
              matrix[1+order-j+ld*(i+j)]+=
                binome[k]*a[l]*Math.pow(mu[1+order],(double)(l+k+1))/(double)(l+k+1)
                              *Math.pow(grid[i],power-k);
              // this corresponds to matrix(i+1,i+j)
            }
          }
        }
      }
    }
    return matrix;
  }

  /******************************************************************************** 
  *  This applies an integration matrix to a function to find the anti-derivative
  *  of the function.  It starts from the first grid point and works its way to
  *  the last grid point.  See below for a description of the integration matrix.
  ******************************************************************************* 
  * description of variables:
  *
  * matrix[]  = integration matrix as supplied by findIntegrationMatrix
  * f[]       = function for which we want an anti-derivative
  * offset    = starting point for the anti-derivative (= value at the first grid point)
  * order     = positive int which gives half the size of the window
  *
  * Output:
  *
  * result[]  = contains the anti-derivative
  ******************************************************************************/
  public static double[] applyIntegrationMatrix(double[] matrix, double[] f) {
    return applyIntegrationMatrix(matrix,f,0.0,1);
  }

  public static double[] applyIntegrationMatrix(double[] matrix, double[] f, int order) {
    return applyIntegrationMatrix(matrix,f,0.0,order);
  }

  public static double[] applyIntegrationMatrix(double[] matrix, double[] f, double offset) {
    return applyIntegrationMatrix(matrix,f,offset,1);
  }

  public static double[] applyIntegrationMatrix(double[] matrix, double[] f, double offset, int order) {
    if ((f == null)||(matrix == null)) return null;
    int ngrid = f.length;
    if (ngrid == 0) return new double[0];

    // check parameters:
    if (order < 0)
      throw new IllegalArgumentException("Please set order >= 0");
    if (matrix.length < ((2*order+2)*ngrid))
      throw new IllegalArgumentException("The dimension of matrix is incorrect");

    double[] result = new double[ngrid];
    Dgbmv.dgbmv("N",ngrid,ngrid,order+1,order,1.0,matrix,0,2*order+2,f,0,1,0.0,result,0,1);
    result[0] = offset;
    for(int i=1; i<ngrid; i++) result[i]+=result[i-1];
    return result;
  }
  
  /******************************************************************************** 
  *  This applies an integration matrix to a function to find the anti-derivative of
  *  the function.  It starts from the last grid point and works its way the the
  *  first grid point (this is particularly useful for integrating functions which
  *  diverge in the center).  See below for a description of the integration matrix.
  ******************************************************************************* 
  * description of variables:
  *
  * matrix[]  = integration matrix as supplied by findIntegrationMatrix
  * f[]       = function for which we want an anti-derivative
  * offset    = starting point for the anti-derivative (= value at the last grid point)
  * order     = positive int which gives half the size of the window
  *
  * Output:
  *
  * result[]  = contains the anti-derivative
  ******************************************************************************/
  public static double[] applyIntegrationMatrixReverse(double[] matrix, double[] f) {
    return applyIntegrationMatrixReverse(matrix,f,0.0,1);
  }

  public static double[] applyIntegrationMatrixReverse(double[] matrix, double[] f, int order) {
    return applyIntegrationMatrixReverse(matrix,f,0.0,order);
  }

  public static double[] applyIntegrationMatrixReverse(double[] matrix, double[] f, double offset) {
    return applyIntegrationMatrixReverse(matrix,f,offset,1);
  }

  public static double[] applyIntegrationMatrixReverse(double[] matrix, double[] f, double offset, int order) {
    if ((f == null)||(matrix == null)) return null;
    int ngrid = f.length;
    if (ngrid == 0) return new double[0];

    // check parameters:
    if (order < 0)
      throw new IllegalArgumentException("Please set order >= 0");
    if (matrix.length < ((2*order+2)*ngrid))
      throw new IllegalArgumentException("The dimension of matrix is incorrect");

    double[] result = new double[ngrid];
    Dgbmv.dgbmv("N",ngrid,ngrid,order+1,order,1.0,matrix,0,2*order+2,f,0,1,0.0,result,0,1);
    for(int i=1;i<ngrid;i++) result[i-1] = result[i];
    result[ngrid-1] = offset;
    for(int i=ngrid-2; i>=0; i--) result[i]=result[i+1]-result[i];
    return result;
  }
  
  /******************************************************************************** 
  *  This subroutine sets up an integration matrix in banded form for an arbitrary
  *  grid.  It assumes that the function that needs to be integrated is a piecewise
  *  continuous functions made of segments of the form:
  *      f(r) = (r_{i+1}^n - r^n) f(r_i) + (r^n - r_i^n) f(r_{i+1}) / (r_{i+1}^n - r_i^n)
  *
  *  This matrix can then be applied to a function using the 
  *  applyPiecewiseIntegrationMatrix method in order to produce
  *  the anti-derivative of the function.
  ******************************************************************************* 
  * description of variables:
  *
  * grid[]    = grid on which the integration is carried out
  * n         = exponent used to describe function
  *
  * Output:
  *
  * matrix[]  = matrix which enables one to calculate the anti-derivative of a
  *             function.
  ******************************************************************************/
  public static double[] findPiecewiseIntegrationMatrix(double[] grid, int n) {
    if (grid == null) return null;
    int ngrid = grid.length;
    if (ngrid == 0) return new double[0];
    // check parameters:
    if (n < 0) throw new IllegalArgumentException("Please set n >= 0");

    double xm, xp, a, b, den, sum;
    int ld = 2, i, j;
    double[] matrix = new double[ngrid*ld];

    xp  = grid[0];
    for(i=1; i < ngrid; i++) {
      xm = xp;
      xp = grid[i];

      den = 0.0;
      for(j=0; j<n; j++) den+=Math.pow(xm,j)*Math.pow(xp,n-1-j);

      sum = 0.0;
      for(j=0; j<=n; j++) sum+=Math.pow(xm,j)*Math.pow(xp,n-j);
      sum /= (double)(n+1);

      a = (Math.pow(xp,n) - sum)/den;
      b = (sum - Math.pow(xm,n))/den;

      matrix[1+ld*(i-1)] = a;  // a(i-1,i);
      matrix[ld*i] = b;        // a(i,i);
    }
    
    return matrix;
  }

  /******************************************************************************** 
  *  This applies an integration matrix to a function to find the anti-derivative
  *  of the function.  It starts from the first grid point and works its way to
  *  the last grid point.  See below for a description of the integration matrix.
  ******************************************************************************* 
  * description of variables:
  *
  * matrix[]  = integration matrix as supplied by findIntegrationMatrix
  * f[]       = function for which we want an anti-derivative
  * offset    = starting point for the anti-derivative (= value at the first grid point)
  *
  * Output:
  *
  * result[]  = contains the anti-derivative
  ******************************************************************************/
  public static double[] applyPiecewiseIntegrationMatrix(double[] matrix, double[] f) {
    return applyIntegrationMatrix(matrix,f,0.0);
  }

  public static double[] applyPiecewiseIntegrationMatrix(double[] matrix, double[] f, double offset) {
    if ((f == null)||(matrix == null)) return null;
    int ngrid = f.length;
    if (ngrid == 0) return new double[0];

    // check parameters:
    if (matrix.length < (2*ngrid))
      throw new IllegalArgumentException("The dimension of matrix is incorrect");

    double[] result = new double[ngrid];
    Dgbmv.dgbmv("N",ngrid,ngrid,1,0,1.0,matrix,0,2,f,0,1,0.0,result,0,1);
    result[0] = offset;
    for(int i=1; i<ngrid; i++) result[i]+=result[i-1];
    return result;
  }
  
  /******************************************************************************** 
  *  This subroutine sets up integration weights for an arbitrary grid.  It does 
  *  this by integrating Lagrange interpolation polynomials calculated over a
  *  sliding window which spans [i-order, i+order+1].
  ******************************************************************************* 
  * description of variables:
  *
  * grid[]    = grid on which the integration is carried out
  * order     = positive int which gives half the size of the window
  *
  * Output:
  *
  * weights[] = contains the integration weights.  This is actually an ngrid x ngrid
  *             matrix stored as a vector, to be used with Dgemm.
  ******************************************************************************/
  public static double[] findIntegrationWeights(double[] grid) {
    return findIntegrationWeights(grid, 1);
  }

  public static double[] findIntegrationWeights(double[] grid, int order) {

    if (grid == null) return null;
    int ngrid = grid.length;
    if (ngrid == 0) return new double[0];
    double[] weights = new double[ngrid];
    
    // check parameters:
    if (order < 0)
      throw new IllegalArgumentException("Please set order >= 0 in Integral.findWeigths");

    // look for multiple points and determine domains:
    ArrayList<Integer> domains = findDomains(grid);

    // check to make sure domains are large enough:
    if (findMinDomainSize(domains) < 1+order)
      throw new IllegalArgumentException("Please set order <= min domain size");

    double[] mu = new double[2*order+2];
    double[] a = new double[2*order+3];

    int start, finish, i, j, k, l, n;
    double  prdct, my_sum;

    for(n=0; n<domains.size(); n+=2) {
      for(i=domains.get(n); i < domains.get(n+1); i++) {

        start  = -order;
        finish = order+1;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the edges
        if ((i+start) < domains.get(n)) start = domains.get(n) - i;
        if ((i+finish) > domains.get(n+1)) finish = domains.get(n+1) - i;

        for(j=start; j <= finish; j++) mu[j+order] = grid[j+i] - grid[i];

        for(j=start; j <= finish; j++) {

          // Initialise Lagrange polynomial
          for(l=1; l<= 2*(order+1); l++) a[l]=0.0;
        
          prdct = 1.0;
          for(k=start; k < j; k++) prdct*=(mu[j+order]-mu[k+order]);
          for(k=j+1; k <= finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]=1.0/prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(k=start; k < j; k++) {
            for(l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }
          for(k=j+1; k <= finish; k++) {
            for(l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }

          // Integrate the Lagrange polynomials, and add the results
          // as weights at the appropriate points.

          for(l=0; l <= (2*order+2); l++) 
            weights[i+j]+=a[l]*Math.pow(mu[1+order],(double)(l+1))/(double)(l+1);
        }
      }
    }
    
    return weights;
  }

  /******************************************************************************** 
  *  This subroutine finds an antiderivative of a function defined over an
  *  arbitrary grid.  It does this by integrating Lagrange interpolation
  *  polynomials calculated over a sliding window which spans [i-order,
  *  i+order+1].
  *********************************************************************************
  * description of variables:
  *
  * grid[]    = grid on which the integration is carried out
  * order     = positive int which gives half the size of the window
  *
  * Output:
  *
  * weights[] = contains the integration weights.  This is actually an ngrid x ngrid
  *             matrix stored as a vector, to be used with Dgemm.
  ******************************************************************************/
  public static double[] findAntiDerivative(double[] grid, double[] f) {
    return findAntiDerivative(grid, f, 0.0, 1);
  }

  public static double[] findAntiDerivative(double[] grid, double[] f, double offset) {
    return findAntiDerivative(grid, f, offset, 1);
  }

  public static double[] findAntiDerivative(double[] grid, double[] f, int order) {
    return findAntiDerivative(grid, f, 0.0, order);
  }

  public static double[] findAntiDerivative(double[] grid, double[] f, double offset, int order) {

    if (grid == null) return null;
    if (f == null) return null;
    int ngrid = grid.length;
    int nf = f.length;
    ngrid = (ngrid>nf?nf:ngrid);
    if (ngrid == 0) return new double[0];
    double[] antiDerivative = new double[ngrid];
    
    // check parameters:
    if (order < 0)
      throw new IllegalArgumentException("Please set order >= order in antiDerivative.findWeigths");

    // look for multiple points and determine domains:
    ArrayList<Integer> domains = findDomains(grid);

    // check to make sure domains are large enough:
    if (findMinDomainSize(domains) < 1+order)
      throw new IllegalArgumentException("Please set order <= min domain size");

    double[] mu = new double[2*order+2];
    double[] a = new double[2*order+3];

    int start, finish, i, j, k, l, n;
    double  prdct, my_sum;

    for(n=0;n<domains.size();n+=2) {
      if (n == 0) {
        // set up the initial point
        antiDerivative[0] = offset;
      } else {
        // ensure continuity across domains
        antiDerivative[domains.get(n)] = antiDerivative[domains.get(n-1)];
      }
      for(i=domains.get(n); i < domains.get(n+1); i++) {
        antiDerivative[i+1] = antiDerivative[i];

        start  = -order;
        finish = order+1;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the edges
        if ((i+start) < domains.get(n))    start  = domains.get(n) - i;
        if ((i+finish) > domains.get(n+1)) finish = domains.get(n+1) - i;

        for(j=start; j <= finish; j++) mu[j+order] = grid[j+i] - grid[i];

        for(j=start; j <= finish; j++) {

          // Initialise Lagrange polynomial
          for(l=1; l<= 2*(order+1); l++) a[l]=0.0;
          
          prdct = 1.0;
          for(k=start; k < j; k++) prdct*=(mu[j+order]-mu[k+order]);
          for(k=j+1; k <= finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]=1.0/prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(k=start; k < j; k++) {
            for(l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }
          for(k=j+1; k <= finish; k++) {
            for(l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }

          // Integrate the Lagrange polynomials, and add the results
          // as weights at the appropriate points.

          for(l=0; l <= (2*order+2); l++)
            antiDerivative[i+1]+=f[i+j]*a[l]*Math.pow(mu[1+order],(double)(l+1))/(double)(l+1);
        }
      }
    }
    return antiDerivative;
  }

  /******************************************************************************** 
  *  This subroutine finds the integral of a function defined over an
  *  arbitrary grid.  It does this by integrating Lagrange interpolation
  *  polynomials calculated over a sliding window which spans [i-order,
  *  i+order+1].
  *********************************************************************************
  * description of variables:
  *
  * grid[]    = grid on which the integration is carried out
  * f[]       = the function to be integrated
  * order     = positive int which gives half the size of the window
  *
  * Output:
  *
  * the integral
  ******************************************************************************/
  public static double findIntegral(double[] grid, double[] f) {
    return findIntegral(grid, f, 1);
  }

  public static double findIntegral(double[] grid, double[] f, int order) {

    if (grid == null) return 0.0;
    if (f == null) return 0.0;
    int ngrid = grid.length;
    int nf = f.length;
    ngrid = (ngrid>nf?nf:ngrid);
    if (ngrid == 0) return 0.0;
    
    // check parameters:
    if (order < 0)
      throw new IllegalArgumentException("Please set order >= order in Integral.findWeigths");

    // look for multiple points and determine domains:
    ArrayList<Integer> domains = findDomains(grid);

    // check to make sure domains are large enough:
    if (findMinDomainSize(domains) < 1+order)
    throw new IllegalArgumentException("Please set order <= min domain size");

    double[] mu = new double[2*order+2];
    double[] a = new double[2*order+3];

    int start, finish, i, j, k, l, n;
    double  prdct, my_sum;
    double integral = 0.0;

    for(n=0;n<domains.size();n+=2) {
      for(i=domains.get(n); i < domains.get(n+1); i++) {

        start  = -order;
        finish = order+1;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the edges
        if ((i+start) < domains.get(n))    start  = domains.get(n) - i;
        if ((i+finish) > domains.get(n+1)) finish = domains.get(n+1) - i;

        for(j=start; j <= finish; j++) mu[j+order] = grid[j+i] - grid[i];

        for(j=start; j <= finish; j++) {

          // Initialise Lagrange polynomial
          for(l=1; l<= 2*(order+1); l++) a[l]=0.0;
          
          prdct = 1.0;
          for(k=start; k < j; k++) prdct*=(mu[j+order]-mu[k+order]);
          for(k=j+1; k <= finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]=1.0/prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(k=start; k < j; k++) {
            for(l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }
          for(k=j+1; k <= finish; k++) {
            for(l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }

          // Integrate the Lagrange polynomials, and add the results
          // as weights at the appropriate points.

          for(l=0; l <= (2*order+2); l++)
            integral+=f[i+j]*a[l]*Math.pow(mu[1+order],(double)(l+1))/(double)(l+1);
        }
      }
    }
    return integral;
  }

  ////////////////////////////////////////////////////////////////////////////
  /** This method gives the value of the Legendre polynomials
  *   of degrees 0 to n-1, at the point x.  This is done using
  *   an efficient recursion formula. */
  ////////////////////////////////////////////////////////////////////////////
  public static double[] Legendre(double x, int n) {
    if (n < 1) return null;
    double[] result = new double[n];
    result[0] = 1.0;
    if (n > 1) result[1] = x;
    for(int i=2; i<n; i++)
      result[i] = ((2.0*i-1.0)*x*result[i-1]-(i-1.0)*result[i-2])/(double)i;
    return result;
  }

  /*********************************************************************
  * This finds an adequate grid for setting up a basis of B-splines,
  * based on a formula by P. Morel (+ a minor correction).
  **********************************************************************
  * @param grid   original grid
  * @param order  order of the B-splines (= degree + 1)
  * @return output grid which will be used in setting up B-splines
  *********************************************************************/
  public static double[] find_knots(double[] grid, int order) {
    int ngrid = grid.length;
    double[] knots = new double[ngrid+order];
    int i,j;

    for(i = 0; i< order; i++) knots[i] = grid[0];

    for(i = order; i<ngrid; i++) {
      knots[i] = 0.0;
      for(j = -order+1; j<0; j++)
        knots[i]+=grid[i+j];
      knots[i]/=(double)(order-1);
    }
    
    for(i = 0; i< order; i++) knots[i+ngrid] = grid[ngrid-1];
    return knots;
  }

  /*********************************************************************
  *  This evaluates a B-spline at a point x.
  **********************************************************************
  *  @param x      point at which to evaluate the B-spline
  *  @param knots  points used in constructing the B-spline.  The points are
  *                assumed to be sorted. (the size of knots determines the
  *                order:  order = size(knots) - 1 )
  *  @param left_boundary  left boundary of the domain.  This intervenes when
  *                calculating B-splines on that boundary.
  *
  *  @return the value of the B-spline at point x. (Note: if x lies outside
  *                the knots grid, then the return value is 0)
  *********************************************************************/
  public static double find_b_spline(double x, double[] knots, double left_boundary) {
    int order = knots.length - 1;
    double[] y = new double[order];
    int i,j;
    
    // check for easy exit:
    if ((x < knots[0])||(x > knots[order])) return 0.0;

    // setting up initial conditions:
    for(i=0; i<order; i++) {

      // The strict inequality in the left-hand condition
      // eliminates spurious peaks around the different
      // knots.  However, B-splines on the left boundary
      // need to be treated differently:

      if (knots[i] == left_boundary) {
        if ((knots[i] <= x)&&(x <= knots[i+1])) {
          y[i] = 1.0;
        } else {
          y[i] = 0.0;
        }
      } else {
        if ((knots[i] < x)&&(x <= knots[i+1])) {
          y[i] = 1.0;
        } else {
          y[i] = 0.0;
        }
      }
    }

    // Use recursion formula for B-splines.
    for(i=1; i< order; i++) {
      for(j=0; j<= order-1-i; j++) {
        if (knots[j] != knots[j+i]) {
          y[j] = y[j]*(x-knots[j])/(knots[j+i]-knots[j]);
        } else {
          y[j] = 0.0;
        }
        if (knots[j+1] != knots[j+i+1]) 
          y[j] = y[j] + y[j+1]*(knots[j+i+1]-x)/(knots[j+i+1]-knots[j+1]);
      }
    }

    return y[0];
  }

  /*********************************************************************
  *  This evaluates a B-spline and its derivatives at a point x.
  **********************************************************************
  *  @param x      point at which to evaluate the B-spline and its derivatives
  *  @param knots  points used in constructing the B-spline.  The points are
  *                assumed to be sorted. (the size of knots determines the
  *                order: order = size(knots) - 1 )
  *  @param dermax  maximum derivative (should obey dermax <= order-2)
  *  @param left_boundary  left boundary of the domain.  This intervenes when
  *                calculating B-splines on that boundary.
  *
  *  @return array containing the value of the B-spline and its derivatives
  *                at point x. (Note: if x lies outside the knots grid, then
  *                the return value is 0)
  *********************************************************************/
  public static double[] find_b_spline(double x, double[] knots, int dermax, double left_boundary) {
    int order = knots.length - 1;
    
    // reduce dermax if too large:
    if (dermax > order-2) {
      dermax = order-2;
      System.out.println("Warning: value of dermax has been modified");
    }
    
    double[][] y = new double[order][dermax+1];
    int i,j;
    
    // check for easy exit:
    if ((x < knots[0])||(x > knots[order])) return y[0];

    // setting up initial conditions:
    for(i=0; i<order; i++) {

      // The strict inequality in the left-hand condition
      // eliminates spurious peaks around the different
      // knots.  However, B-splines on the left boundary
      // need to be treated differently:

      if (knots[i] == left_boundary) {
        if ((knots[i] <= x)&&(x <= knots[i+1])) {
          y[i][0] = 1.0;
        } else {
          y[i][0] = 0.0;
        }
      } else {
        if ((knots[i] < x)&&(x <= knots[i+1])) {
          y[i][0] = 1.0;
        } else {
          y[i][0] = 0.0;
        }
      }
    }

    // Use recursion formula for B-splines.
    // For derivatives: combine recursion with Leibniz' formula.
    int der;
    for(i=1; i< order; i++) {
      for(j=0; j<= order-1-i; j++) {
        if (knots[j] != knots[j+i]) {
          for (der = dermax; der >= 1; der--) {
            y[j][der] = (y[j][der]*(x-knots[j])+(double)der*y[j][der-1])
                      / (knots[j+i]-knots[j]);
          }
          y[j][0] = y[j][0]*(x-knots[j])/(knots[j+i]-knots[j]);
        } else {
          for (der = 0; der <= dermax; der++) y[j][der] = 0.0;
        }
        if (knots[j+1] != knots[j+i+1])  {
          for (der = dermax; der >= 1; der--) {
            y[j][der] = y[j][der] + (y[j+1][der]*(knots[j+i+1]-x)
                      - (double)der*y[j+1][der-1])
                      / (knots[j+i+1]-knots[j+1]);
          }
          y[j][0] = y[j][0] + y[j+1][0]*(knots[j+i+1]-x)/(knots[j+i+1]-knots[j+1]);
        }
      }
    }

    return y[0];
  }

  /*********************************************************************
  * This subroutine interpolates a 1D function defined on a strictly
  * monotonic grid to another grid using polynomial splines.
  *
  * NOTE: this assumes that xin AND xout are sorted in ascending order
  **********************************************************************
  * @param xin    input grid.
  * @param xout   output grid
  * @param y      input function
  * @return a one dimensional array result[nout] containing the interpolated y
  *               function where nout = size(xout)
  *********************************************************************/
  public static double[] interpolate1D_simple(double[] xin, double[] xout, double[] y) {

    int nin  = xin.length;
    int nout = xout.length;

    // sanity checks
    if ((nout==0)||(nin < 2)) return null;
    if (y.length < nin) return null;

    // simple linear interpolation
    double[] result = new double[nout];
    int i, j;
    double eta;
    for(i=0; i<nout; i++) {
      for(j=1; j<nin-1; j++) {
        if (xout[i] < xin[j]) break; 
      }
      if (xin[j] != xin[j-1]) {
        eta = (xout[i]-xin[j-1])/(xin[j]-xin[j-1]);
      } else {
        eta = 0.5; // a quick and approximate solution
      }
      result[i] = (1.0-eta)*y[j-1] + eta*y[j];
    }

    return result;
  }

  /*********************************************************************
  * This subroutine interpolates a 1D function defined on an increasing
  * grid to another grid using polynomial splines.  The input grid may
  * contain double points, in which case, an interpolation on each domain
  * is carried out.
  *
  * NOTE: this assumes that xin AND xout are sorted in ascending order
  **********************************************************************
  * @param xin    input grid.
  * @param xout   output grid
  * @param y      input function
  * @param isLog  specifies whether to interpolate the function or its logarithm
  * @return a one dimensional array result[nout] containing the interpolated y
  *               function where nout = size(xout)
  *********************************************************************/
  public static double[] interpolate1D(double[] xin, double[] xout, double[] y, boolean isLog) {

    int nin  = xin.length;
    int nout = xout.length;

    // sanity checks
    if ((nout==0)||(nin < 2)) return null;
    if (y.length < nin) return null;

    int defaultOrder = 4, order;                     // Cubic splines
 
    double[] kn;                                     // This will correspond to
                                                     // subarrays taken from knots.
    double[] mat_band = new double[(3*defaultOrder+1)*nin]; // A linear system, the solution of
                                                     // which is the B-spline coefficients.
    int[] ipiv = new int[nin];                       // Used by lapack
    org.netlib.util.intW info = new org.netlib.util.intW(2); // Used by lapack
    int i, j, k, n;

    // adjust input grid (this is necessary for the placement of knots):
    double[] xtemp = xin.clone();
    xtemp[0] = xin[0]>xout[0]?xout[0]:xin[0];
    xtemp[nin-1] = xin[nin-1]<xout[nout-1]?xout[nout-1]:xin[nin-1];

    // Find domains for the input grid - the interpolation
    // will be carried out on each domain separately.
    ArrayList<Integer> domains = findDomains(xin);

    double lb;
    double[][] knots = new double[domains.size()/2][];

    for(n=0; n<domains.size(); n+=2) {

      // reduce order if need be:      
      order = Math.min(1+domains.get(n+1)-domains.get(n),defaultOrder);
      kn = new double[order+1];
      
      // find the knots
      lb = xtemp[domains.get(n)];
      knots[n/2] = find_knots(Arrays.copyOfRange(xtemp,domains.get(n),domains.get(n+1)+1),order);

      // set up matrix
      for(i=domains.get(n);i<=domains.get(n+1);i++) {

        // We will consider groups of order+1 knots
        // starting at j.  j will also index the
        // B-splines in the matrix.
        for(j=Math.max(domains.get(n),i-order); j<=Math.min(domains.get(n+1),i+order); j++) {

          if ((xin[i] >= knots[n/2][j-domains.get(n)])&&(xin[i] <= knots[n/2][j-domains.get(n)+order])) {

            // Create subarray kn using elements from knots:
            System.arraycopy(knots[n/2],j-domains.get(n),kn,0,order+1);
          
            // calculate B-spline and derivatives and
            // setup matrix in banded form:
            //   mat_band(kl+ku+i-j,j) = mat(i,j)
            mat_band[(2*defaultOrder+i-j)+(3*defaultOrder+1)*j] = find_b_spline(xin[i],kn,lb);

          }
        }
      }
    }

    // use LAPACK to factorise matrix:
    Dgbtrf.dgbtrf(nin,nin,defaultOrder,defaultOrder,mat_band,0,3*defaultOrder+1,ipiv,0,info);

    // store B-spline function values:
    double[][] b_splines = new double[nin][];
    int[] istart = new int[nin];
    int[] ifinish = new int[nin];
    int iinit = 0;
    for(n=0; n<domains.size(); n+=2) {

      // reduce order if need be:      
      order = Math.min(1+domains.get(n+1)-domains.get(n),defaultOrder);
      kn = new double[order+1];
      lb = xtemp[domains.get(n)];

      for(i=domains.get(n);i<=domains.get(n+1);i++) {

        // find starting point:
        while((iinit<nout)&&(xout[iinit]<knots[n/2][i-domains.get(n)])) iinit++;
        istart[i] = iinit;
 
        // find finishing point:
        ifinish[i] = istart[i];
        while((ifinish[i]<nout)&&(xout[ifinish[i]]<=knots[n/2][i-domains.get(n)+order])) ifinish[i]++;
        
        // reduce ifinish if need be:
        while (ifinish[i] > 2) {
          if ((xout[ifinish[i]-1] == xin[domains.get(n+1)])&&
              (xout[ifinish[i]-1] == xout[ifinish[i]-2])) {
            ifinish[i]--;
          } else {
            break;
          }
        }

        // find appropriate b-splines:
        if (ifinish[i] > istart[i]) {
          b_splines[i] = new double[ifinish[i]-istart[i]];
          System.arraycopy(knots[n/2],i-domains.get(n),kn,0,order+1);
          for(j=istart[i];j<ifinish[i];j++)
            b_splines[i][j-istart[i]] = find_b_spline(xout[j],kn,lb);
        } else {
          b_splines[i] = new double[0];
        }
      }
      iinit = ifinish[domains.get(n+1)]; 
                // If the output grid contains the same double points as the
                // input grid, this ensures that one of the points will be in
                // the previous domains, and the next one in the next domain.
    }

    // find interpolated function:
    double[] result = new double[nout];
    double[] bcoeff = new double[nin];
    if (isLog) {
      for(j=0;j<nin;j++) bcoeff[j] = Math.log(y[j]);
    } else {
      for(j=0;j<nin;j++) bcoeff[j] = y[j];
    }
    // use LAPACK to find b-spline coefficients:
    Dgbtrs.dgbtrs("N",nin,defaultOrder,defaultOrder,1,mat_band,0,3*defaultOrder+1,ipiv,0,bcoeff,0,nin,info);
    // find interpolated function
    for(n=0; n<domains.size(); n+=2) {
      for(j=domains.get(n);j<=domains.get(n+1);j++) {
        for(k=istart[j];k<ifinish[j];k++) {
          result[k]+=b_splines[j][k-istart[j]]*bcoeff[j];
        }
      }
    }
    if (isLog) {
      for(j=0;j<nout;j++) result[j] = Math.exp(result[j]);
    }

    return result;
  }

  /*********************************************************************
  * This subroutine interpolates a 1D function defined on an increasing
  * grid to another grid using polynomial splines.  The input grid may
  * contain double points, in which case, an interpolation on each domain
  * is carried out.
  *
  * It squares each element of the grid before carrying out the
  * interpolation.  Hence, if the grid starts at 0, then it will
  * behave as a polynomial in x^2 around that point.  Of course,
  * both grids MUST be positive.
  *
  * NOTE: this assumes that xin AND xout are sorted in ascending order
  **********************************************************************
  * @param xin    input grid.
  * @param xout   output grid
  * @param y      input function
  * @param isLog  specifies whether to interpolate the function or its logarithm
  * @return a one dimensional array result[nout] containing the interpolated y
  *               function where nout = size(xout)
  *********************************************************************/
  public static double[] interpolate1D_r2(double[] xin, double[] xout, double[] y, boolean isLog) {

    int i;
    int nin  = xin.length;
    int nout = xout.length;

    // sanity checks
    if ((nout==0)||(nin < 2)) return null;
    if (y.length < nin) return null;
    
    double[] xin2 = new double[nin];
    double[] xout2 = new double[nout];
    
    for(i=0; i<nin; i++) xin2[i] = xin[i]*xin[i];
    for(i=0; i<nout; i++) xout2[i] = xout[i]*xout[i];
    
    return interpolate1D(xin2,xout2,y,isLog);
  }

  /*********************************************************************
  * These subroutines interpolate a list of 1D functions defined on an
  * increasing grid to another grid using polynomial splines.  The input
  * grid may contain double points, in which case, an interpolation on each
  * domain is carried out.
  **********************************************************************
  * @param xin    input grid.
  * @param xout   output grid
  * @param y      2D array of the form y[index][nin] containing a list of
  *               1D functions to be interpolated
  * @param index  list of indices for which to do the interpolation
  * @param isLog  specifies whether to interpolate the function or its logarithm
  * @return a two dimensional array of the form result[size(index)][nout]
  *               containing the interpolated functions.
  *********************************************************************/

  public static ArrayList<double[][]> interpolateList(double[] xin, double[] xout, ArrayList<double[][]> y, boolean isLog) {
    int ntotal = 0, i, j, k;
    for(i=0; i<y.size(); i++) ntotal+=y.get(i).length;
    ArrayList<Integer> index = new ArrayList<Integer>();
    for(i=0; i<ntotal; i++) index.add(i);
    double[][] yy = new double[ntotal][];
    k = 0;
    for(i=0; i<y.size(); i++) {
      for(j=0; j<y.get(i).length; j++) {
        yy[k] = y.get(i)[j];
        k++;
      }
    }
    double[][] aux = interpolateList(xin,xout,yy,index,isLog);
    ArrayList<double[][]> result = new ArrayList<double[][]>();
    k = 0;
    double[][] tmp;
    for(i=0; i<y.size(); i++) {
      tmp = new double[y.get(i).length][];
      for(j=0; j<y.get(i).length; j++) {
        tmp[j]=aux[k];
        k++;
      }
      result.add(tmp);
    }
    return result;
  }

  public static ArrayList<double[][]> interpolateList_r2(double[] xin, double[] xout, ArrayList<double[][]> y, boolean isLog) {
    int ntotal = 0, i, j, k;
    for(i=0; i<y.size(); i++) ntotal+=y.get(i).length;
    ArrayList<Integer> index = new ArrayList<Integer>();
    for(i=0; i<ntotal; i++) index.add(i);
    double[][] yy = new double[ntotal][];
    k = 0;
    for(i=0; i<y.size(); i++) {
      for(j=0; j<y.get(i).length; j++) {
        yy[k] = y.get(i)[j];
        k++;
      }
    }
    double[] xin2 = new double[xin.length];
    double[] xout2 = new double[xout.length];
    for(i=0; i<xin.length; i++) xin2[i] = xin[i]*xin[i];
    for(i=0; i<xout.length; i++) xout2[i] = xout[i]*xout[i];
    double[][] aux = interpolateList(xin2,xout2,yy,index,isLog);
    ArrayList<double[][]> result = new ArrayList<double[][]>();
    k = 0;
    double[][] tmp;
    for(i=0; i<y.size(); i++) {
      tmp = new double[y.get(i).length][];
      for(j=0; j<y.get(i).length; j++) {
        tmp[j]=aux[k];
        k++;
      }
      result.add(tmp);
    }
    return result;
  }

  public static double[][] interpolateList(double[] xin, double[] xout, double[][] y, boolean isLog) {
    int n = y.length;
    ArrayList<Integer> index = new ArrayList<Integer>();
    for(int i=0; i<n; i++) index.add(i);
    return interpolateList(xin,xout,y,index,isLog);
  }

  public static double[][] interpolateList_r2(double[] xin, double[] xout, double[][] y, boolean isLog) {
    int i;
    double[] xin2 = new double[xin.length];
    double[] xout2 = new double[xout.length];
    for(i=0; i<xin.length; i++) xin2[i] = xin[i]*xin[i];
    for(i=0; i<xout.length; i++) xout2[i] = xout[i]*xout[i];
    int n = y.length;
    ArrayList<Integer> index = new ArrayList<Integer>();
    for(i=0; i<n; i++) index.add(i);
    return interpolateList(xin2,xout2,y,index,isLog);
  }

  public static double[][] interpolateList_r2(double[] xin, double[] xout, double[][] y, ArrayList<Integer> index, boolean isLog) {
    int i;
    double[] xin2 = new double[xin.length];
    double[] xout2 = new double[xout.length];
    for(i=0; i<xin.length; i++) xin2[i] = xin[i]*xin[i];
    for(i=0; i<xout.length; i++) xout2[i] = xout[i]*xout[i];
    return interpolateList(xin2,xout2,y,index,isLog);
  }

  public static double[][] interpolateList(double[] xin, double[] xout, double[][] y, ArrayList<Integer> index, boolean isLog) {

    int nin  = xin.length;
    int nout = xout.length;

    // sanity checks
    if ((nout==0)||(nin < 2)) return null;

    int defaultOrder = 4, order;                     // Cubic splines
 
    double[] kn;                                     // This will correspond to
                                                     // subarrays taken from knots.
    double[] mat_band = new double[(3*defaultOrder+1)*nin]; // A linear system, the solution of
                                                     // which is the B-spline coefficients.
    int[] ipiv = new int[nin];                       // Used by lapack
    org.netlib.util.intW info = new org.netlib.util.intW(2); // Used by lapack
    int i, j, k, n;

    // adjust input grid (this is necessary for the placement of knots):
    double[] xtemp = xin.clone();
    xtemp[0] = xin[0]>xout[0]?xout[0]:xin[0];
    xtemp[nin-1] = xin[nin-1]<xout[nout-1]?xout[nout-1]:xin[nin-1];

    // Find domains for the input grid - the interpolation
    // will be carried out on each domain separately.
    ArrayList<Integer> domains = findDomains(xin);

    double lb;
    double[][] knots = new double[domains.size()/2][];

    for(n=0; n<domains.size(); n+=2) {

      // reduce order if need be:      
      order = Math.min(1+domains.get(n+1)-domains.get(n),defaultOrder);
      kn = new double[order+1];
      
      // find the knots
      lb = xtemp[domains.get(n)];
      knots[n/2] = find_knots(Arrays.copyOfRange(xtemp,domains.get(n),domains.get(n+1)+1),order);

      // set up matrix
      for(i=domains.get(n);i<=domains.get(n+1);i++) {

        // We will consider groups of order+1 knots
        // starting at j.  j will also index the
        // B-splines in the matrix.
        for(j=Math.max(domains.get(n),i-order); j<=Math.min(domains.get(n+1),i+order); j++) {

          if ((xin[i] >= knots[n/2][j-domains.get(n)])&&(xin[i] <= knots[n/2][j-domains.get(n)+order])) {

            // Create subarray kn using elements from knots:
            System.arraycopy(knots[n/2],j-domains.get(n),kn,0,order+1);
          
            // calculate B-spline and derivatives and
            // setup matrix in banded form:
            //   mat_band(kl+ku+i-j,j) = mat(i,j)
            mat_band[(2*defaultOrder+i-j)+(3*defaultOrder+1)*j] = find_b_spline(xin[i],kn,lb);

          }
        }
      }
    }

    // use LAPACK to factorise matrix:
    Dgbtrf.dgbtrf(nin,nin,defaultOrder,defaultOrder,mat_band,0,3*defaultOrder+1,ipiv,0,info);

    // store B-spline function values:
    double[][] b_splines = new double[nin][];
    int[] istart = new int[nin];
    int[] ifinish = new int[nin];
    int iinit = 0;
    for(n=0; n<domains.size(); n+=2) {

      // reduce order if need be:      
      order = Math.min(1+domains.get(n+1)-domains.get(n),defaultOrder);
      kn = new double[order+1];
      lb = xtemp[domains.get(n)];

      for(i=domains.get(n);i<=domains.get(n+1);i++) {

        // find starting point:
        while((iinit<nout)&&(xout[iinit]<knots[n/2][i-domains.get(n)])) iinit++;
        istart[i] = iinit;
 
        // find finishing point:
        ifinish[i] = istart[i];
        while((ifinish[i]<nout)&&(xout[ifinish[i]]<=knots[n/2][i-domains.get(n)+order])) ifinish[i]++;
        
        // reduce ifinish if need be:
        while (ifinish[i] > 2) {
          if (xout[ifinish[i]-1] == xout[ifinish[i]-2]) {
            ifinish[i]--;
          } else {
            break;
          }
        }

        // find appropriate b-splines:
        if (ifinish[i] > istart[i]) {
          b_splines[i] = new double[ifinish[i]-istart[i]];
          System.arraycopy(knots[n/2],i-domains.get(n),kn,0,order+1);
          for(j=istart[i];j<ifinish[i];j++)
            b_splines[i][j-istart[i]] = find_b_spline(xout[j],kn,lb);
        } else {
          b_splines[i] = new double[0];
        }
      }
      iinit = ifinish[domains.get(n+1)]; 
                // If the output grid contains the same double points as the
                // input grid, this ensures that one of the points will be in
                // the previous domains, and the next one in the next domain.
    }

    // find interpolated functions:
    double[][] result = new double[index.size()][nout];
    double[] bcoeff = new double[nin];
    for(i=0;i<index.size();i++) {
      // sanity check:
      if (y[index.get(i)] == null) {
        result[i] = null;
        continue;
      }
      if (isLog) {
        for(j=0;j<nin;j++) bcoeff[j] = Math.log(y[index.get(i)][j]);
      } else {
        for(j=0;j<nin;j++) bcoeff[j] = y[index.get(i)][j];
      }
      // use LAPACK to find b-spline coefficients:
      Dgbtrs.dgbtrs("N",nin,defaultOrder,defaultOrder,1,mat_band,0,3*defaultOrder+1,ipiv,0,bcoeff,0,nin,info);
      // find interpolated function
      for(n=0; n<domains.size(); n+=2) {
        for(j=domains.get(n);j<=domains.get(n+1);j++) {
          for(k=istart[j];k<ifinish[j];k++) {
            result[i][k]+=b_splines[j][k-istart[j]]*bcoeff[j];
          }
        }
      }
      if (isLog) {
        for(j=0;j<nout;j++) result[i][j] = Math.exp(result[i][j]);
      }
    }
    return result;
  }

  /****************************************************************
  * Routine numerical recipes computing the weights and points for
  * a Gauss-Legendre integration.
  *****************************************************************
  * input:
  * x1 = lower bound
  * x2 = upper bound
  * n  = number of points
  *
  * output:
  * xw[0][0:n-1] = Gauss-Legendre collocation points
  * xw[1][0:n-1] = Gauss-Legendre collocation weights
  ****************************************************************/
  public static double[][] gauleg(double x1, double x2, int n) {
    double[][] xw  = new double[2][n];

    double eps = 3e-15;
    int m=(n+1)/2;
    double xm=0.5*(x2+x1);
    double xl=0.5*(x2-x1);
    double pi=Math.PI;
    double p1, p2, p3, z, z1, pp = 0.0;
    for (int i=1; i<=m; i++) {
      z=Math.cos(pi*((double)i-0.25)/((double)n+0.5));
      z1=z+1.0;
      while (Math.abs(z-z1) > eps) {
        p1=1.0;
        p2=0.0;
        for (int j=1; j<=n; j++) {
          p3=p2;
          p2=p1;
          p1=((2.0*(double)j-1.0)*z*p2-((double)j-1.0)*p3)/(double)j;
        }
        pp=(double)n*(z*p1-p2)/(z*z-1.0);
        z1=z;
        z=z1-p1/pp;
      }
      xw[0][i-1]=xm-xl*z;
      xw[0][n-i]=xm+xl*z;
      xw[1][i-1]=2.0*xl/((1.0-z*z)*pp*pp);
      xw[1][n-i]=xw[1][i-1];
    }
    return xw;
  }

  /***************************************************************************
  *  This searches for and identifies multiple points in a grid.  It also
  *  checks the grid and return null if the grid is not increasing.
  ***************************************************************************/
  public static ArrayList<Integer> findMpoints(double[] grid) {
    ArrayList<Integer> result = new ArrayList<Integer>();

    int i = 1, j;
    while(i<grid.length) {
      if (grid[i-1] > grid[i]) {
        // grid descreases: return null
        return null;
      } else if (grid[i-1] == grid[i]) {
        result.add(i-1);
        j=1;
        while(i<grid.length){
          if (grid[i-1] != grid[i]) break;
          i++;
          j++;
        }
        result.add(j);
      } else {
        i++;
      }
    }
    return result;
  }

  /***************************************************************************
  *  This removes multiple points with 3 or more points by removing
  *  intermediate points.  It also makes sure the endpoints are not multiple
  *  points.  This then allows a straightforward multi-domain approach.
  ***************************************************************************/
  public static double[] removeMpoints(ArrayList<Integer> mpoints, double[] array) {
    // easy exit
    if (mpoints.size() == 0) return array.clone();

    // find the number of points that need to be removed:
    int nremove = 0, i, j, k, l;
    for(k=1;k<mpoints.size();k+=2) nremove+=(mpoints.get(k).intValue()-2);
    if (mpoints.get(0).equals(0)) nremove++;
    if ((mpoints.get(mpoints.size()-2)+mpoints.get(mpoints.size()-1)) == array.length) nremove++;

    // easy exit
    if (nremove == 0) return array.clone();
    
    // sanity check:
    if (nremove == array.length) {
      double[] result = new double[1];
      result[0] = array[0];
      return result;
    }

    // more complicated approach:
    double[] result = new double[array.length-nremove];

    // copy "array" into "result" while removing extra multiple points
    i = j = k = 0;
    double sum;
    while(i<array.length) {
      if (k>=mpoints.size()) {
        // ordinary point
        result[j] = array[i]; j++; i++;
      } else {
        if (i<mpoints.get(k)) {
          // ordinary point
          result[j] = array[i]; j++; i++;
        } else {
          // multiple point
          if (mpoints.get(k) > 0) {
            result[j] = array[i];
            j++;
          }
          if ((mpoints.get(k) + mpoints.get(k+1)) < array.length) {
            i+=(mpoints.get(k+1)-1);
          } else {
            i+=mpoints.get(k+1);
          }
          k+=2;
        }
      }
    }
    
    return result;
  }

  /***************************************************************************
  *  This finds the endpoints of domains contained within a grid, and delimited
  * by double/end points.
  ***************************************************************************/
  public static ArrayList<Integer> findDomains(double[] grid) {
    // easy exit:
    if (grid == null) return null;
    int i, ngrid = grid.length;
    if (ngrid == 0) return null;
    
    ArrayList<Integer> result = new ArrayList<Integer>();
    result.add(0);
    for(i=1; i<ngrid; i++) {
      if (grid[i] == grid[i-1]) {
        result.add(i-1);
        result.add(i);
      }
    }
    result.add(ngrid-1);
    return result;
  }
  
  /***************************************************************************
  *  This finds the minimum domain size from a set of domains.
  *  NOTE: this assumes that domains.size() >= 2
  ***************************************************************************/
  public static int findMinDomainSize(ArrayList<Integer> domains) {
    int minSize = domains.get(1) - domains.get(0), domainSize;
    for(int i=2; i<domains.size(); i+=2) {
      domainSize = domains.get(i+1) - domains.get(i);
      if (domainSize < minSize) minSize = domainSize;
    }
    return minSize;
  }

  /***************************************************************************
  *  This finds the minimum domain size from a set of mpoints and the size of
  *  the grid.
  *  IMPORTANT: this assumes that mpoints != null.
  ***************************************************************************/
  public static int findMinDomainSize(ArrayList<Integer> mpoints, int ngrid) {
    if (mpoints.size() == 0) {
      return ngrid-1;
    } else {
      int minSize, domainSize;
      if (mpoints.get(0) == 0) {
        minSize = ngrid-1;
      } else {
        minSize = mpoints.get(0);
      }
      for(int i=2; i<mpoints.size(); i+=2) {
        domainSize = mpoints.get(i) - mpoints.get(i-1) - mpoints.get(i-2) + 1;
        if (domainSize < minSize) minSize = domainSize;
      }
      // The last domain:
      domainSize = ngrid - mpoints.get(mpoints.size()-1) - mpoints.get(mpoints.size()-2);
      if (domainSize > 0) {
        if (domainSize < minSize) minSize = domainSize;
      }
      return minSize;
    } 
  }

  /***************************************************************************
  * Finds the minimun value of an array.
  ****************************************************************************
  * @param   array  array for which we're searching for the minimum value
  * @return  the minimun value of the array
  ***************************************************************************/
  public static double minval(double[] array) {
    if (array == null) return Double.NaN;
    if (array.length == 0) return Double.NaN;
    double result = Double.POSITIVE_INFINITY;
    for(int i=0; i<array.length; i++) {
      if (array[i] < result) result = array[i];
    }
    return result;
  }

  /***************************************************************************
  * Finds the maximum value of an array.
  ****************************************************************************
  * @param   array  array for which we're searching for the minimum value
  * @return  the maximum value of the array
  ***************************************************************************/
  public static double maxval(double[] array) {
    if (array == null) return Double.NaN;
    if (array.length == 0) return Double.NaN;
    double result = Double.NEGATIVE_INFINITY;
    for(int i=0; i<array.length; i++) {
      if (array[i] > result) result = array[i];
    }
    return result;
  }

  /***************************************************************************
  * Adds two arrays.  If the two arrays don't have the same size, then an array
  * with the smaller dimension containing the sum is returned.
  ****************************************************************************
  * @param   array1  first array
  * @param   array2  second array
  * @return  the sum of array1 and array2
  ***************************************************************************/
  public static double[] addArrays(double[] array1, double[] array2) {
    if (array1 == null) return null;
    if (array2 == null) return null;
    int len1 = array1.length; 
    int len2 = array2.length; 
    int len=len1<len2?len1:len2;
    double[] result = new double[len];
    for(int i=0; i<len; i++) result[i] = array1[i] + array2[i];
    return result;
  }

  /***************************************************************************
  * Subtract two arrays (the second one from the first).  If the two arrays
  * don't have the same size, then an array with the smaller dimension
  * containing the sum is returned.
  ****************************************************************************
  * @param   array1  first array
  * @param   array2  second array
  * @return  the difference array1 - array2
  ***************************************************************************/
  public static double[] subtractArrays(double[] array1, double[] array2) {
    if (array1 == null) return null;
    if (array2 == null) return null;
    int len1 = array1.length; 
    int len2 = array2.length; 
    int len=len1<len2?len1:len2;
    double[] result = new double[len];
    for(int i=0; i<len; i++) result[i] = array1[i] - array2[i];
    return result;
  }
}
