/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package graph;

import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.awt.Color;
import java.awt.BasicStroke;
import java.awt.Stroke;
import java.awt.Graphics2D;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Font;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.geom.AffineTransform;
import java.awt.font.LineMetrics;
import java.awt.font.FontRenderContext;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import java.lang.Cloneable;

/****************************************************************************
* The purpose of this class is to draw plots in Java.
****************************************************************************/

public class Plot extends JComponent implements Cloneable
{
  // Arrays containing data sets to be plotted:
  // (x.get(0)[*],y.get(0)[*]) is the first data set to be plotted.
  // leftError,  rightError = lower and upper bounds on the error bars on x values
  // lowerError, upperError = lower and upper bounds on the error bars on y values
  private ArrayList<double[]> x = new ArrayList<double[]>();
  private ArrayList<double[]> y = new ArrayList<double[]>();
  private ArrayList<double[]> leftError  = new ArrayList<double[]>();
  private ArrayList<double[]> rightError = new ArrayList<double[]>();
  private ArrayList<double[]> lowerError = new ArrayList<double[]>();
  private ArrayList<double[]> upperError = new ArrayList<double[]>();
  
  // Integer containing the number of data sets:
  private int nDataSets = 0;

  // Array which says what type of stroke to use for plotting
  // the different data sets. See section "Various line patterns"
  // below for acceptable values and their meanings.
  private ArrayList<Integer> linestyleArray = new ArrayList<Integer>();

  // Array which says what symbols to use for plotting
  // the different data sets. See section "Various symbols"
  // below for acceptable values and their meanings.
  private ArrayList<Integer> psymArray = new ArrayList<Integer>();

  // Array which contains the different colors used for the different
  // data sets:
  private ArrayList<Color> colorArray = new ArrayList<Color>();

  // Array which specifics whether or not to plot the horizontal error bars:
  // (true = plot error bars, false = do not plot error bars)
  private ArrayList<Boolean> hErrorArray = new ArrayList<Boolean>();
 
  // Array which specifics whether or not to plot the vertical error bars:
  // (true = plot error bars, false = do not plot error bars)
  private ArrayList<Boolean> vErrorArray = new ArrayList<Boolean>();
 
  // Margins around the plot area.  These
  // need to be big enough to allow tick
  // marks, annotations and various titles.
  private float leftMinMargin= 50F;
  private float leftMargin   = 50F;
  private float rightMargin  = 30F;
  private float topMargin    = 50F;
  private float bottomMargin = 50F;
  
  // These submargins are for the x and y axis titles:
  private float leftSubMargin   = 30F;
  private float bottomSubMargin = 30F;

  // These are the lengths of the tick marks
  private float xLongTick  = 8F;
  private float xShortTick = 4F;
  private float yLongTick  = 8F;
  private float yShortTick = 4F;
  
  // These are variables in which to store information
  // about the ticks and their labels;
  private int nticks;
  private double[] ticks;
  private boolean[] isMajor;
  private int nDecimal;
  private boolean withExp;
  private boolean withMantissa;

  // The x and y bounds, and their logarithms:
  private double xmin=0.0, xmax=1.0, ymin=0.0, ymax=1.0;
  private double xlmin=Double.NEGATIVE_INFINITY, xlmax=0.0;
  private double ylmin=Double.NEGATIVE_INFINITY, ylmax=0.0;

  // Various booleans which affect the behaviour of
  // findRange():
  private boolean extendXrange = false;
  private boolean extendYrange = true;
  private boolean reverseXrange = false;
  private boolean reverseYrange = false;

  // Booleans which specify whether to use a linear or logarithmic X/Y axis:
  private boolean Xlog = false;
  private boolean Ylog = false;

  // Various titles:
  private String xtitle = "x axis";
  private String ytitle = "y axis";
  /** The title of the plot */
  protected String title = "A plot";
  
  // IDL counterparts to title (IDL doesn't always use the same type of
  // coding as java):
  private String xIDLtitle = "x axis";
  private String yIDLtitle = "y axis";
  /** The title of the plot used when writing an IDL file */
  protected String IDLtitle = "A plot";

  // Various fonts:
  private Font annotationFont = new Font("Sans Serif", Font.PLAIN, 12);
  private Font exponentFont = annotationFont.deriveFont(9F);
  private Font xyTitleFont = annotationFont.deriveFont(18F);
  private Font titleFont   = annotationFont.deriveFont(Font.BOLD,22F);
  private Font timesFont   = annotationFont.deriveFont(10F);

  // These strings are used when plotting numbers using scientific notation
  private String timesString="x";
  private String tenString="10";

  // These keep track of the size of the plot area:
  private float hspan, vspan;

  // These are the coordinates of the four corners
  // of the plot area.  They are calculated using
  // the margins given above and the bounds of the
  // component.  By declaring xleft as a method, it
  // automatically adjusts to any changes in the left
  // margin.
  private float xleft() {return leftMargin;}
  private float xright() {return hspan-rightMargin;}
  private float yup() {return topMargin;}
  private float ydown() {return vspan-bottomMargin;}

  /////////////////////////////////////////////////////////////////////////////
  // Various line patterns:
  // The conventions used here are similar to IDL.
  /** No line style. This is useful for plotting only symbols. */
  public static final int noLine = -1;
  /** Solid line style. This is the default value. */
  public static final int solid = 0;      // the default
  /** Dotted line style. */
  public static final int dotted = 1;
  /** Dashed line style. */
  public static final int dashed = 2;
  /** Dot-dashed line style. */
  public static final int dotDash = 3;
  /** Dot-dot-dashed line style. */
  public static final int dotDotDash = 4;
  /** Line style with long dashes. */
  public static final int longDash = 5;
  
  private static final Stroke[] linestyle = new Stroke[]
    {
      // Stroke pattern corresponding to a solid line
      new BasicStroke(2F),
      // Stroke pattern corresponding to a dotted line 
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {2F,2F},             // Dash pattern
          0.0F),                           // Dash phase
      // Stroke pattern corresponding to a dashed line 
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {6F,6F},             // Dash pattern
          0.0F),                           // Dash phase
      // Stroke pattern corresponding to a dot-dash line
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {6F,6F,2F,6F},       // Dash pattern
          0.0F),                           // Dash phase
      // Stroke pattern corresponding to a dot-dot-dash line
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {6F,6F,2F,2F,2F,6F}, // Dash pattern
          0.0F),                           // Dash phase
      // Stroke pattern corresponding to a line with long dashes
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {12F,12F},           // Dash pattern
          0.0F)                            // Dash phase
    };

  /////////////////////////////////////////////////////////////////////////////
  // Various symbols:
  // The conventions used here are similar to IDL.
  /** No symbols. This is useful for plotting only with a linestyle.
      This is the default value. */
  /** No symbols. */
  public static final int noSymbol = 0;      // the default
  /** Plus symbols. */
  public static final int plus = 1;
  /** Asterisks. */
  public static final int asterisk = 2;
  /** Single dots. */
  public static final int singleDot = 3;
  /** Diamonds. */
  public static final int diamond = 4;
  /** Triangles. */
  public static final int triangle = 5;
  /** Squares. */
  public static final int square = 6;
  /** Time symbols. */
  public static final int times = 7;
  
  // This part codes the different segments as a series
  // of line segments (x0,y0,x1,y1,x2,y2 ...) where:
  // (x0,y0)-(x1,y1) = first segment
  // (x2,y2)-(x3,y3) = second segment
  // ...
  // Note the y axis is inverted (this affects the triangle).
  private static final float[][] psym = new float[][]
  {
    null,
    new float[] {0F,5F,0F,-5F,-5F,0F,5F,0F},
    new float[] {0F,5F,0F,-5F,-5F,0F,5F,0F,-5F,-5F,5F,5F,-5F,5F,5F,-5F},
    new float[] {-1F,1F,1F,1F,1F,1F,1F,-1F,1F,-1F,-1F,-1F,-1F,-1F,-1F,1F},
    new float[] {-5F,0F,0F,5F,0F,5F,5F,0F,5F,0F,0F,-5F,0F,-5F,-5F,0F},
    new float[] {0F,-6F,5.2F,3F,5.2F,3F,-5.2F,3F,-5.2F,3F,0F,-6F},    
    new float[] {-5F,5F,5F,5F,5F,5F,5F,-5F,5F,-5F,-5F,-5F,-5F,-5F,-5F,5F},
    new float[] {-5F,-5F,5F,5F,-5F,5F,5F,-5F}
  };

  // The stroke pattern used for creating the symbols:
  private static final Stroke psymStroke = new BasicStroke(1F);

  // Square root of 10 (this intervenes when finding the x or y range in the logarithmic case):
  private static final double sqrt10 = 3.1622776601683793;

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y, int ls, int ps, Color aColor) {

    // Check input parameters:
    if ((x == null) || (y == null)) return;

    int xsize = x.length;
    int ysize = y.length;

    if ((ls < -1)||(ls >5)) ls = 0;

    if ((ps < 0)||(ps > 7)) ps = 0;

    // input (x,y) arrays, linestyle ls and symbol style ps
    int length = (xsize>ysize)?ysize:xsize;
    
    nDataSets = 1;
    this.x.add(new double[length]);
    this.y.add(new double[length]);
    this.leftError.add(new double[length]);
    this.rightError.add(new double[length]);
    this.lowerError.add(new double[length]);
    this.upperError.add(new double[length]);
    this.linestyleArray.add(ls);
    this.psymArray.add(ps);
    this.colorArray.add(aColor);
    this.hErrorArray.add(false);
    this.vErrorArray.add(false);

    System.arraycopy(x,0,this.x.get(this.x.size()-1),0,length);
    System.arraycopy(y,0,this.y.get(this.y.size()-1),0,length);
    
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y, int ls, int ps) {this(x,y,ls,ps,Color.black);}

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y, int ls, Color aColor) {this(x,y,ls,0,aColor);}

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y, int ls) {this(x,y,ls,0,Color.black);}

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor for a single data set. If the arrays x and y do not have
   * the same lengths, the longer one is truncated.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y) {this(x,y,0,0,Color.black);}
  
  /////////////////////////////////////////////////////////////////////////////
  /** A constructor based on the different titles and number of data sets.
   *
   * @param title  title above the plot
   * @param xtitle title for the x-axis
   * @param ytitle title for the y-axis
   * @param nDataSets number of data sets (i.e. x[], y[] pairs)
   */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(String title, String xtitle, String ytitle, int nDataSets) {
    setTitles(title,xtitle,ytitle);

    if (nDataSets < 0) {
      this.nDataSets = 0;
      return;
    }

    this.nDataSets = nDataSets;
    for(int i=0; i<nDataSets; i++) {
      x.add(null);
      y.add(null);
      leftError.add(null);
      rightError.add(null);
      lowerError.add(null);
      upperError.add(null);
      linestyleArray.add(0);
      psymArray.add(0);
      colorArray.add(Color.black);  // It is important to put a true color
                                    // rather than null because of the way
                                    // replaceDataSet is implemented.  The
                                    // same applies to linestyleArray and
                                    // psymArray.
      hErrorArray.add(false);
      vErrorArray.add(false);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** A constructor based on the different titles.
   *
   * @param title  title above the plot
   * @param xtitle title for the x-axis
   * @param ytitle title for the y-axis
   */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(String title, String xtitle, String ytitle) {
    setTitles(title,xtitle,ytitle);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the titles to be modified.  Both the normal and the IDL
   *  titles are modified.
   *
   *  @param title   new title for the plot
   *  @param xtitle  new title for the x-axis
   *  @param ytitle  new title for the y-axis
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setTitles(String title, String xtitle, String ytitle) {
    this.title = title;
    this.xtitle = xtitle;
    this.ytitle = ytitle;
    this.IDLtitle = title;
    this.xIDLtitle = xtitle;
    this.yIDLtitle = ytitle;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the IDL titles to be modified.
   *
   *  @param title   new IDL title for the plot
   *  @param xtitle  new IDL title for the x-axis
   *  @param ytitle  new IDL title for the y-axis
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setIDLTitles(String title, String xtitle, String ytitle) {
    this.IDLtitle = title;
    this.xIDLtitle = xtitle;
    this.yIDLtitle = ytitle;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the x-axis titles to be modified.  Both the normal and the IDL
   *  x-axis titles are modified.
   *
   *  @param xtitle  new title for the x-axis
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setXtitle(String xtitle) {
    this.xtitle = xtitle;
    this.xIDLtitle = xtitle;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the IDL x-axis title to be modified.
   *
   *  @param xtitle  new IDL title for the x-axis
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setIDLXtitle(String xtitle) {
    this.xIDLtitle = xtitle;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle, choice of 
   *  symbols and color.
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y, int ls, int ps, Color aColor) {

    // Check parameters:
    if (ds < 0) return;
    if (ds > this.x.size()) return;
    if ((x == null) || (y == null)) return;

    int xsize = x.length;
    int ysize = y.length;

    if ((ls < -1)||(ls >5)) ls = 0;

    if ((ps < 0)||(ps > 7)) ps = 0;

    // Add data set:
    int length = (xsize>ysize)?ysize:xsize;
    this.x.add(ds,new double[length]);
    this.y.add(ds,new double[length]);
    leftError.add(ds,new double[length]);
    rightError.add(ds,new double[length]);
    lowerError.add(ds,new double[length]);
    upperError.add(ds,new double[length]);
    System.arraycopy(x,0,this.x.get(ds),0,length);
    System.arraycopy(y,0,this.y.get(ds),0,length);
    linestyleArray.add(ds,ls);
    psymArray.add(ds,ps);
    colorArray.add(ds,aColor);
    hErrorArray.add(ds,false);
    vErrorArray.add(ds,false);
    nDataSets++;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of 
   *  symbols.
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y, int ls, int ps) {
    addDataSet(ds,x,y,ls,ps,Color.black);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of 
   *  color.
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y, int ls, Color aColor) {
    addDataSet(ds,x,y,ls,0,aColor);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle.
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y, int ls) {
    addDataSet(ds,x,y,ls,0,Color.black);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y).
   *
   * @param  ds     location where to add the data set
   * @param  x      array of x values
   * @param  y      array of y values
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(int ds, double[] x, double[] y) {
    addDataSet(ds,x,y,0,0,Color.black);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle, choice of 
   *  symbols and color.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls, int ps, Color aColor) {

    // Check parameters:
    if ((x == null) || (y == null)) return;

    int xsize = x.length;
    int ysize = y.length;

    if ((ls < -1)||(ls >5)) ls = 0;

    if ((ps < 0)||(ps > 7)) ps = 0;

    // Add data set:
    int length = (xsize>ysize)?ysize:xsize;
    this.x.add(new double[length]);
    this.y.add(new double[length]);
    leftError.add(new double[length]);
    rightError.add(new double[length]);
    lowerError.add(new double[length]);
    upperError.add(new double[length]);
    System.arraycopy(x,0,this.x.get(this.x.size()-1),0,length);
    System.arraycopy(y,0,this.y.get(this.y.size()-1),0,length);
    linestyleArray.add(ls);
    psymArray.add(ps);
    colorArray.add(aColor);
    hErrorArray.add(false);
    vErrorArray.add(false);
    nDataSets++;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of 
   *  symbols.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  ps     integer giving the choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls, int ps) {
    addDataSet(x,y,ls,ps,Color.black);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of 
   *  color.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   * @param  aColor a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls, Color aColor) {
    addDataSet(x,y,ls,0,aColor);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle.
   *
   * @param  x      array of x values
   * @param  y      array of y values
   * @param  ls     integer giving the line style
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls) {
    addDataSet(x,y,ls,0,Color.black);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y).
   *
   * @param  x      array of x values
   * @param  y      array of y values
   */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y) {
    addDataSet(x,y,0,0,Color.black);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This removes data set number ds from arrays x and y.
   *
   * @param ds index of the data set to be removed (0 <= ds < nDataSets)
   */
  /////////////////////////////////////////////////////////////////////////////
  public void removeDataSet(int ds) {
    if ((ds >= nDataSets)||(ds < 0)) return;
   
    x.remove(ds);
    y.remove(ds);
    leftError.remove(ds);
    rightError.remove(ds);
    lowerError.remove(ds);
    upperError.remove(ds);
    linestyleArray.remove(ds);
    psymArray.remove(ds);
    colorArray.remove(ds);
    hErrorArray.remove(ds);
    vErrorArray.remove(ds);
    nDataSets--;
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This removes all data sets.
   */
  /////////////////////////////////////////////////////////////////////////////
  public void clearDataSets() {
    x.clear();
    y.clear();
    leftError.clear();
    rightError.clear();
    lowerError.clear();
    upperError.clear();
    linestyleArray.clear();
    psymArray.clear();
    colorArray.clear();
    hErrorArray.clear();
    vErrorArray.clear();
    nDataSets=0;
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls,
   * symbol choice ps, and color choice aColor.  The error bars are left 
   * unchanged.
   *
   * @param ds     data set that will be replaced (0 <= ds < nDataSets)
   * @param x      new array of x values
   * @param y      new array of y values
   * @param ls     new linestyle
   * @param ps     new choice of symbols
   * @param aColor new choice of color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls, int ps,
                             Color aColor) {

    // check parameters:
    if ((ds >= nDataSets)||(ds < 0)) return;
    
    if ((x == null) || (y == null)) return;

    int xsize = x.length;
    int ysize = y.length;

    if ((ls < -1)||(ls >5)) ls = linestyleArray.get(ds);

    if ((ps < 0)||(ps > 7)) ps = psymArray.get(ds);

    // replace data set:
    int length = (xsize>ysize)?ysize:xsize;
    this.x.set(ds,new double[length]);
    this.y.set(ds,new double[length]);
    System.arraycopy(x,0,this.x.get(ds),0,length);
    System.arraycopy(y,0,this.y.get(ds),0,length);
    linestyleArray.set(ds,ls);
    psymArray.set(ds,ps);
    colorArray.set(ds,aColor);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls,
   * and color choice aColor.  The error bars are left unchanged.
   *
   * @param ds     data set that will be replaced (0 <= ds < nDataSets)
   * @param x      new array of x values
   * @param y      new array of y values
   * @param ls     new linestyle
   * @param aColor new choice of color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls, Color aColor) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    replaceDataSet(ds,x,y,ls,psymArray.get(ds),aColor);
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls and
   * symbol choice ps.  The error bars are left unchanged.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param x  new array of x values
   * @param y  new array of y values
   * @param ls new linestyle
   * @param ps new choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls, int ps) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    replaceDataSet(ds,x,y,ls,ps,colorArray.get(ds));
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y and linestyle ls.
   * The error bars are left unchanged.
   *
   * @param ds  data set that will be replaced (0 <= ds < nDataSets)
   * @param x   new array of x values
   * @param y   new array of y values
   * @param ls  new linestyle
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    replaceDataSet(ds,x,y,ls,psymArray.get(ds),colorArray.get(ds));
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y.
   * The error bars are left unchanged.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param x  new array of x values
   * @param y  new array of y values
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    replaceDataSet(ds,x,y,linestyleArray.get(ds),psymArray.get(ds),colorArray.get(ds));
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces the lower x error bar for data set ds.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param leftError new lower x error bar array
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceLeftError(int ds, double[] leftError) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    if (leftError == null) return;
    if (x.get(ds) == null) return;
    int length = x.get(ds).length;
    if (leftError.length != length) return;
    this.leftError.set(ds,new double[length]);
    for(int i=0;i<length;i++) this.leftError.get(ds)[i] = Math.abs(leftError[i]);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces the upper x error bar for data set ds.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param rightError new upper x error bar array
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceRightError(int ds, double[] rightError) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    if (rightError == null) return;
    if (x.get(ds) == null) return;
    int length = x.get(ds).length;
    if (rightError.length != length) return;
    this.rightError.set(ds,new double[length]);
    for(int i=0;i<length;i++) this.rightError.get(ds)[i] = Math.abs(rightError[i]);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces the lower y error bar for data set ds.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param lowerError new lower y error bar array
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceLowerError(int ds, double[] lowerError) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    if (lowerError == null) return;
    if (y.get(ds) == null) return;
    int length = y.get(ds).length;
    if (lowerError.length != length) return;
    this.lowerError.set(ds,new double[length]);
    for(int i=0;i<length;i++) this.lowerError.get(ds)[i] = Math.abs(lowerError[i]);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces the upper y error bar for data set ds.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param upperError new upper y error bar array
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceUpperError(int ds, double[] upperError) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    if (upperError == null) return;
    if (y.get(ds) == null) return;
    int length = y.get(ds).length;
    if (upperError.length != length) return;
    this.upperError.set(ds,new double[length]);
    for(int i=0;i<length;i++) this.upperError.get(ds)[i] = Math.abs(upperError[i]);
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This sets hErrorArray(ds) with a new boolean value.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param hErrorValue new boolean value
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setHErrorValue(int ds, boolean hErrorValue) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    hErrorArray.set(ds,hErrorValue);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This sets vErrorArray(ds) with a new boolean value.
   *
   * @param ds data set that will be replaced (0 <= ds < nDataSets)
   * @param vErrorValue new boolean value
   */
  /////////////////////////////////////////////////////////////////////////////
  public void setVErrorValue(int ds, boolean vErrorValue) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    vErrorArray.set(ds,vErrorValue);
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of extendXrange */
  /////////////////////////////////////////////////////////////////////////////
  public void setExtendXrange(boolean value) {this.extendXrange = value;}
 
  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of extendYrange */
  /////////////////////////////////////////////////////////////////////////////
  public void setExtendYrange(boolean value) {this.extendYrange = value;}

  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of reverseXrange */
  /////////////////////////////////////////////////////////////////////////////
  public void setReverseXrange(boolean value) {this.reverseXrange = value;}
 
  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of reverseYrange */
  /////////////////////////////////////////////////////////////////////////////
  public void setReverseYrange(boolean value) {this.reverseYrange = value;}

  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of Xlog */
  /////////////////////////////////////////////////////////////////////////////
  public void setXlog(boolean value) {this.Xlog = value;}
 
  /////////////////////////////////////////////////////////////////////////////
  /** This sets the value of Ylog */
  /////////////////////////////////////////////////////////////////////////////
  public void setYlog(boolean value) {this.Ylog = value;}

  /////////////////////////////////////////////////////////////////////////////
  /** This gets the number of data sets */
  /////////////////////////////////////////////////////////////////////////////
  public int getNDataSets() {return this.nDataSets;}

  /////////////////////////////////////////////////////////////////////////////
  /** This gets the x values of a particular data set. It returns a clone so
   *  as to avoid contaminating the data set.*/
  /////////////////////////////////////////////////////////////////////////////
  public double[] getDataSetX(int i) {return x.get(i).clone();}

  /////////////////////////////////////////////////////////////////////////////
  /** This gets the y values of a particular data set. It returns a clone so
   *  as to avoid contaminating the data set. */
  /////////////////////////////////////////////////////////////////////////////
  public double[] getDataSetY(int i) {return y.get(i).clone();}

  /////////////////////////////////////////////////////////////////////////////
  /** This gets the value of Xlog */
  /////////////////////////////////////////////////////////////////////////////
  public boolean getXlog() {return this.Xlog;}
 
  /////////////////////////////////////////////////////////////////////////////
  /** This gets the value of Ylog */
  /////////////////////////////////////////////////////////////////////////////
  public boolean getYlog() {return this.Ylog;}

  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for xmin, xmax, ymin and ymax, for a
   *  list of data sets.
   * 
   *  @param selection list containing the indices of the data sets for which
   *                   to adjust the values of xmin, xmax, ymin and ymax */
  /////////////////////////////////////////////////////////////////////////////
  public void findRange(int[] selection) {
    if (selection == null) return;
    if (selection.length == 0) return;
    if (nDataSets == 0) return;

    if (Xlog) {
      findXRangeLog(selection);
    } else {
      findXRangeLinear(selection);
    }
    if (Ylog) {
      findYRangeLog(selection);
    } else {
      findYRangeLinear(selection);
    }

    // I do this afterwards using the final values of xmin, xmax, ymin, and ymax:
    if (reverseXrange) reverseXaxis();
    if (reverseYrange) reverseYaxis();
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for xmin, xmax, ymin and ymax. */
  /////////////////////////////////////////////////////////////////////////////
  public void findRange() {
    int[] selection = new int[nDataSets];
    for(int i=0; i<nDataSets; i++) selection[i] = i;
    findRange(selection);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This finds checks to see if Xlog is selected, and if so, modifies
   *  xmin and xmax if at least one of the two has a negative value. */
  /////////////////////////////////////////////////////////////////////////////
  protected void correctXrange() {
    if (!Xlog) return;
    if ((xmin <= 0.0)||(xmax <= 0.0)) {
      int[] selection = new int[nDataSets];
      for(int i=0; i<nDataSets; i++) selection[i] = i;
      findXRangeLog(selection);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This finds checks to see if Ylog is selected, and if so, modifies
   *  ymin and ymax if at least one of the two has a negative value. */
  /////////////////////////////////////////////////////////////////////////////
  protected void correctYrange() {
    if (!Ylog) return;
    if ((ymin <= 0.0)||(ymax <= 0.0)) {
      int[] selection = new int[nDataSets];
      for(int i=0; i<nDataSets; i++) selection[i] = i;
      findYRangeLog(selection);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for xmin and xmax when Xlog = false
   * 
   *  @param selection list containing the indices of the data sets for which
   *                   to adjust the values of xmin and xmax */
  /////////////////////////////////////////////////////////////////////////////
  private void findXRangeLinear(int[] selection) {
    int i = 0, ii, ns = selection.length;
    
    for(ii=0; ii<ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) 
        if ((x.get(i) != null)&&(x.get(i).length > 0)) break;
    }
    if (ii == ns) return;

    xmin = Double.POSITIVE_INFINITY;
    xmax = Double.NEGATIVE_INFINITY;

    // exclude infinities in range search
    for(ii = 0; ii < ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) {
        if (x.get(i) != null) {
          if (hErrorArray.get(i)) {
            for(int j = 0; j < x.get(i).length; j++) {
              if (Double.isInfinite(x.get(i)[j])) continue;
              if (Double.isInfinite(y.get(i)[j])) continue;
              if (Ylog && (y.get(i)[j] <= 0.0)) continue;
              if (xmin > (x.get(i)[j]-leftError.get(i)[j]))  xmin = x.get(i)[j]-leftError.get(i)[j];
              if (xmax < (x.get(i)[j]+rightError.get(i)[j])) xmax = x.get(i)[j]+rightError.get(i)[j];
            }
          } else {
            for(int j = 0; j < x.get(i).length; j++) {
              if (Double.isInfinite(x.get(i)[j])) continue;
              if (Double.isInfinite(y.get(i)[j])) continue;
              if (Ylog && (y.get(i)[j] <= 0.0)) continue;
              if (xmin > x.get(i)[j]) xmin = x.get(i)[j];
              if (xmax < x.get(i)[j]) xmax = x.get(i)[j];
            }
          }
        }
      }
    }
   
    // check for infinities (this happens when no finite values where found in the data sets):
    if (Double.isInfinite(xmin)) xmin = 0.0;
    if (Double.isInfinite(xmax)) xmax = 0.0;

    if (extendXrange) { 
      // This adds a margin to the left a right of the plotted functions.
      double dx = 0.1*xmax - 0.1*xmin;
      xmax+=dx;
      xmin-=dx;
    }

    // check for infinities (this happens when extending the range causes an overflow):
    if (Double.isInfinite(xmin)) xmin =-Double.MAX_VALUE; 
    if (Double.isInfinite(xmax)) xmax = Double.MAX_VALUE;

    // if xmin = xmax, then add space to the left and right:
    if (xmin==xmax) {xmin-=0.5; xmax+=0.5;}
    xlmin = Math.log(xmin);
    xlmax = Math.log(xmax);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for xmin and xmax when Xlog = true
   * 
   *  @param selection list containing the indices of the data sets for which
   *                   to adjust the values of xmin and xmax */
  /////////////////////////////////////////////////////////////////////////////
  private void findXRangeLog(int[] selection) {
    int i = 0, ii, ns = selection.length;
    
    for(ii=0; ii<ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) 
        if ((x.get(i) != null)&&(x.get(i).length > 0)) break;
    }
    if (ii == ns) return;

    xmin = Double.POSITIVE_INFINITY;
    xmax = 0.0;

    // exclude infinities in range search
    for(ii = 0; ii < ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) {
        if (x.get(i) != null) {
          if (hErrorArray.get(i)) {
            for(int j = 0; j < x.get(i).length; j++) {
              if (Double.isInfinite(x.get(i)[j])) continue;
              if (x.get(i)[j] <= 0.0) continue;
              if (Double.isInfinite(y.get(i)[j])) continue;
              if (Ylog && (y.get(i)[j] <= 0.0)) continue;
              if (xmin > (x.get(i)[j]-leftError.get(i)[j])) {
                if (x.get(i)[j]-leftError.get(i)[j] > 0.0) xmin = x.get(i)[j]-leftError.get(i)[j];
              } else {
                if (xmin > x.get(i)[j]) xmin = x.get(i)[j];
              }
              if (xmax < (x.get(i)[j]+rightError.get(i)[j])) xmax = x.get(i)[j]+rightError.get(i)[j];
            }
          } else {
            for(int j = 0; j < x.get(i).length; j++) {
              if (Double.isInfinite(x.get(i)[j])) continue;
              if (x.get(i)[j] <= 0.0) continue;
              if (Double.isInfinite(y.get(i)[j])) continue;
              if (Ylog && (y.get(i)[j] <= 0.0)) continue;
              if (xmin > x.get(i)[j]) xmin = x.get(i)[j];
              if (xmax < x.get(i)[j]) xmax = x.get(i)[j];
            }
          }
        }
      }
    }
   
    // check for infinities (this happens when no strictly positive values where found in the data sets):
    if (Double.isInfinite(xmin)) xmin = 1.0;
    if (xmax <= 0.0) xmax = 1.0;

    if (extendXrange) { 
      // This adds a margin to the left a right of the plotted functions.
      double dx = Math.exp(0.1*(Math.log(xmax)-Math.log(xmin)));
      xmax*=dx;
      xmin/=dx;
    }

    // check for infinities (this happens when extending the range causes an overflow):
    if (xmin <= 0.0) xmin = Double.MIN_VALUE;
    if (Double.isInfinite(xmax)) xmax = Double.MAX_VALUE;

    // if xmin = xmax, then add space to the left and right:
    if (xmin==xmax) {xmin/=sqrt10; xmax*=sqrt10;}
    xlmin = Math.log(xmin);
    xlmax = Math.log(xmax);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for ymin and ymax when Ylog = false.
   * 
   *  @param selection list containing the indices of the data sets for which
   *                   to adjust the values of ymin and ymax */
  /////////////////////////////////////////////////////////////////////////////
  private void findYRangeLinear(int[] selection) {
    int i = 0, ii, ns = selection.length;
    
    for(ii=0; ii<ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) 
        if ((y.get(i) != null)&&(y.get(i).length > 0)) break;
    }
    if (ii == ns) return;

    ymin = Double.POSITIVE_INFINITY;
    ymax = Double.NEGATIVE_INFINITY;

    // exclude infinities in range search
    for(ii = 0; ii < ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) {
        if (y.get(i) != null) {
          if (vErrorArray.get(i)) {
            for(int j = 0; j < y.get(i).length; j++) {
              if (Double.isInfinite(y.get(i)[j])) continue;
              if (Double.isInfinite(x.get(i)[j])) continue;
              if (Xlog && (x.get(i)[j] <= 0.0)) continue;
              if (ymin > (y.get(i)[j]-lowerError.get(i)[j])) ymin = y.get(i)[j]-lowerError.get(i)[j];
              if (ymax < (y.get(i)[j]+upperError.get(i)[j])) ymax = y.get(i)[j]+upperError.get(i)[j];
            }
          } else {
            for(int j = 0; j < y.get(i).length; j++) {
              if (Double.isInfinite(y.get(i)[j])) continue;
              if (Double.isInfinite(x.get(i)[j])) continue;
              if (Xlog && (x.get(i)[j] <= 0.0)) continue;
              if (ymin > y.get(i)[j]) ymin = y.get(i)[j];
              if (ymax < y.get(i)[j]) ymax = y.get(i)[j];
            }
          }
        }
      }
    }
   
    // check for infinities (this happens when no finite values is found in the data sets):
    if (Double.isInfinite(ymin)) ymin = 0.0;
    if (Double.isInfinite(ymax)) ymax = 0.0;

    if (extendYrange) { 
      // This adds a margin above and below the plotted functions.
      double dy = 0.1*ymax-0.1*ymin;
      ymax+=dy;
      ymin-=dy;
    }

    // check for infinities (this happens when the range overflows):
    if (Double.isInfinite(ymin)) ymin =-Double.MAX_VALUE;
    if (Double.isInfinite(ymax)) ymax = Double.MAX_VALUE;

    // if ymin = ymax, then add space below and above:
    if (ymin==ymax) {ymin-=0.5; ymax+=0.5;}
    ylmin = Math.log(ymin);
    ylmax = Math.log(ymax);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for ymin and ymax when Ylog = true.
   * 
   *  @param selection list containing the indices of the data sets for which
   *                   to adjust the values of ymin and ymax */
  /////////////////////////////////////////////////////////////////////////////
  private void findYRangeLog(int[] selection) {
    int i = 0, ii, ns = selection.length;
    
    for(ii=0; ii<ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) 
        if ((y.get(i) != null)&&(y.get(i).length > 0)) break;
    }
    if (ii == ns) return;

    ymin = Double.POSITIVE_INFINITY;
    ymax = 0.0;

    // exclude infinities in range search
    for(ii = 0; ii < ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) {
        if (y.get(i) != null) {
          if (vErrorArray.get(i)) {
            for(int j = 0; j < y.get(i).length; j++) {
              if (Double.isInfinite(y.get(i)[j])) continue;
              if (y.get(i)[j] <= 0.0) continue;
              if (Double.isInfinite(x.get(i)[j])) continue;
              if (Xlog && (x.get(i)[j] <= 0.0)) continue;
              if (ymin > (y.get(i)[j]-lowerError.get(i)[j])) {
                if (y.get(i)[j]-lowerError.get(i)[j] > 0.0) ymin = y.get(i)[j]-lowerError.get(i)[j];
              } else {
                if (ymin > y.get(i)[j]) ymin = y.get(i)[j];
              }
              if (ymax < (y.get(i)[j]+upperError.get(i)[j])) ymax = y.get(i)[j]+upperError.get(i)[j];
            }
          } else {
            for(int j = 0; j < y.get(i).length; j++) {
              if (Double.isInfinite(y.get(i)[j])) continue;
              if (y.get(i)[j] <= 0.0) continue;
              if (Double.isInfinite(x.get(i)[j])) continue;
              if (Xlog && (x.get(i)[j] <= 0.0)) continue;
              if (ymin > y.get(i)[j]) ymin = y.get(i)[j];
              if (ymax < y.get(i)[j]) ymax = y.get(i)[j];
            }
          }
        }
      }
    }
   
    // check for infinities (this happens when no strictly positive values where found in the data sets):
    if (Double.isInfinite(ymin)) ymin = 1.0;
    if (ymax <= 0.0) ymax = 1.0;

    if (extendYrange) { 
      // This adds a margin above and below the plotted functions.
      double dy = Math.exp(0.1*(Math.log(ymax)-Math.log(ymin)));
      ymax*=dy;
      ymin/=dy;
    }

    // check for infinities (this happens when the range overflows):
    if (ymin <= 0.0) ymin = Double.MIN_VALUE;
    if (Double.isInfinite(ymax)) ymax = Double.MAX_VALUE;

    // if ymin = ymax, then add space below and above:
    if (ymin==ymax) {ymin/=sqrt10; ymax*=sqrt10;}
    ylmin = Math.log(ymin);
    ylmax = Math.log(ymax);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This reverses the x-axis by swapping xmin and xmax */
  /////////////////////////////////////////////////////////////////////////////
  public void reverseXaxis() {
    double temp = xmin;
    xmin  = xmax;
    xmax  = temp;
    temp  = xlmin;
    xlmin = xlmax;
    xlmax = temp;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This reverses the y-axis by swapping ymin and ymax */
  /////////////////////////////////////////////////////////////////////////////
  public void reverseYaxis() {
    double temp = ymin;
    ymin  = ymax;
    ymax  = temp;
    temp  = ylmin;
    ylmin = ylmax;
    ylmax = temp;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set xmin.  If Xlog = true, then setting a
   *  negative or zero value will set xmin to the default value 1.0.
   *
   *  @param xmin new value for xmin. */
  /////////////////////////////////////////////////////////////////////////////
  public void setXmin(double xmin) {
    if (Xlog && xmin <=0) {
      this.xmin  = 1.0;
      this.xlmin = 0.0;
    } else {
      this.xmin  = xmin;
      this.xlmin = Math.log(xmin);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set xmax.  If Xlog = true, then setting a
   *  negative or zero value will set xmax to the default value 1.0.
   *
   *  @param xmax new value for xmax. */
  /////////////////////////////////////////////////////////////////////////////
  public void setXmax(double xmax) {
    if (Xlog && xmax <=0) {
      this.xmax  = 1.0;
      this.xlmax = 0.0;
    } else {
      this.xmax  = xmax;
      this.xlmax  = Math.log(xmax);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set ymin.  If Ylog = true, then setting a
   *  negative or zero value will set ymin to the default value 1.0.
   *
   *  @param ymin new value for ymin. */
  /////////////////////////////////////////////////////////////////////////////
  public void setYmin(double ymin) {
    if (Ylog && ymin <=0) {
      this.ymin  = 1.0;
      this.ylmin = 0.0;
    } else {
      this.ymin  = ymin;
      this.ylmin = Math.log(ymin);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set ymax.  If Ylog = true, then setting a
   *  negative or zero value will set ymax to the default value 1.0.
   *
   *  @param ymax new value for ymax. */
  /////////////////////////////////////////////////////////////////////////////
  public void setYmax(double ymax) {
    if (Ylog && ymax <=0) {
      this.ymax  = 1.0;
      this.ylmax = 0.0;
    } else {
      this.ymax  = ymax;
      this.ylmax = Math.log(ymax);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to obtain xmin.
   *
   *  @return the value of xmin. */
  /////////////////////////////////////////////////////////////////////////////
  public double getXmin() {return xmin;}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to obtain xmax.
   *
   *  @return the value of xmax. */
  /////////////////////////////////////////////////////////////////////////////
  public double getXmax() {return xmax;}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to obtain ymin
   *
   *  @return the value of ymin. */
  /////////////////////////////////////////////////////////////////////////////
  public double getYmin() {return ymin;}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to obtain ymax
   *
   *  @return the value of ymax. */
  /////////////////////////////////////////////////////////////////////////////
  public double getYmax() {return ymax;}

  /////////////////////////////////////////////////////////////////////////////
  /** This zooms to the specified coordinates.  IMPORTANT: these coordinates
  *   are given in terms of the graphics window rather than in terms of
  *   the coordinate system associated with the data sets.
  *
  *   @param xleft   coordinate of left bound of the zoom window
  *   @param xright  coordinate of right bound of the zoom window
  *   @param ybottom coordinate of lower bound of the zoom window
  *   @param ytop    coordinate of upper bound of the zoom window */
  /////////////////////////////////////////////////////////////////////////////
  protected void zoom(int xleft, int xright, int ybottom, int ytop) {
    double xxmin, xxmax, yymin, yymax;

    // Important: xValue and yValue depend on xmin, xmax, ymin and ymax.
    // Therefore, these should be modified only after calling xValue
    // and yValue.

    xxmin = xValue(xleft);
    xxmax = xValue(xright);
    yymin = yValue(ybottom);
    yymax = yValue(ytop);
    
    xmin  = xxmin;
    xmax  = xxmax;
    ymin  = yymin;
    ymax  = yymax;
    xlmin = Math.log(xxmin);
    xlmax = Math.log(xxmax);
    ylmin = Math.log(yymin);
    ylmax = Math.log(yymax);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This shifts to the specified coordinates.  IMPORTANT: these coordinates
  *   are given in terms of the graphics window rather than in terms of
  *   the coordinate system associated with the data sets.
  *
  *   @param xx0  initial real x coordinate from which window is shifting
  *   @param yy0  initial real y coordinate from which window is shifting
  *   @param xl0  log of initial real x coordinate from which window is shifting
  *   @param yl0  log of initial real y coordinate from which window is shifting
  *   @param x1   current x coordinate to which window is shifting
  *   @param y1   current y coordinate to which window is shifting */
  /////////////////////////////////////////////////////////////////////////////
  protected void shift(double xx0, double yy0, double xl0, double yl0, int x1, int y1) {
    double DeltaX, DeltaY;

    if (Xlog) {
      DeltaX = xlmax - xlmin;
      xlmin = xl0 - DeltaX*(double)(x1-xleft())/(double)(xright()-xleft());
      xlmax = xlmin + DeltaX;
      xmin = Math.exp(xlmin);
      xmax = Math.exp(xlmax);
    } else {
      DeltaX = xmax - xmin;
      xmin = xx0 - DeltaX*(double)(x1-xleft())/(double)(xright()-xleft());
      xmax = xmin + DeltaX;
      xlmin = Math.log(xmin);
      xlmax = Math.log(xmax);
    }

    if (Ylog) {
      DeltaY = ylmax - ylmin;
      ylmin = yl0 - DeltaY*(double)(y1-ydown())/(double)(yup()-ydown());
      ylmax = ylmin + DeltaY;
      ymin = Math.exp(ylmin);
      ymax = Math.exp(ylmax);
    } else {
      DeltaY = ymax - ymin;
      ymin = yy0 - DeltaY*(double)(y1-ydown())/(double)(yup()-ydown());
      ymax = ymin + DeltaY;
      ylmin = Math.log(ymin);
      ylmax = Math.log(ymax);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This creates the plot.
   *
   *  @param g graphics frame in which to create the plot */
  /////////////////////////////////////////////////////////////////////////////
  public void paint(Graphics g) {
    Graphics2D g2D = (Graphics2D)g;
    Rectangle r = this.getBounds();

    hspan = (float) r.getWidth();
    vspan = (float) r.getHeight();
    
    // I think that r.x, r.y are the offset from the
    // complete frame, and should therefore be removed:
    r.x = 0;
    r.y = 0;
    g2D.setColor(Color.white);
    g2D.fill(r);
    
    // Make a nice frame around the plot:
    r.height--; // this is to see the lower part of the frame
    r.width--;  // this is to see the right part of the frame
    g2D.setColor(Color.black);
    g2D.draw(r);

    // Plot tick marks and labels:
    plotYTicks(g2D);
    plotXTicks(g2D);

    // Create a rectangle which corresponds to the axes of the plot:
    Rectangle2D rect = new Rectangle2D.Float(xleft(),yup(),xright()-xleft(),ydown()-yup());

    // Do the rest of the plot:
    plotFrame(g2D,rect);
    g2D.clip(rect);
    for(int i=0; i < nDataSets; i++) {
      if ((x.get(i) != null) && (x.get(i).length > 0)) {
        g2D.setColor(colorArray.get(i));
        plotXY(x.get(i),y.get(i),linestyleArray.get(i),g2D);
        plotXYpsym(x.get(i),y.get(i),psymArray.get(i),g2D);
        if (hErrorArray.get(i)) plotHError(x.get(i),y.get(i),leftError.get(i),rightError.get(i),g2D);
        if (vErrorArray.get(i)) plotVError(x.get(i),y.get(i),lowerError.get(i),upperError.get(i),g2D);
      }
    }
    g2D.setClip(null);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This draws a rectangle which delimits the zoom window.
   *  @param x0 x coordinate of one corner of the zoom window
   *  @param y0 corresponding y coordinate (x0, y0)
   *  @param x1 x coordinate of opposite corner
   *  @param y1 corresponding y coordinate (x1, y1) */
  /////////////////////////////////////////////////////////////////////////////
  protected void drawRect(int x0, int y0, int x1, int y1) {

    // get current graphics context:
    Graphics2D g2D = (Graphics2D) this.getGraphics();
    
    GeneralPath f = new GeneralPath();
    f.moveTo(x0,y0);
    f.lineTo(x0,y1);
    f.lineTo(x1,y1);
    f.lineTo(x1,y0);
    f.lineTo(x0,y0);
    
    g2D.setXORMode(Color.white);
    g2D.setStroke(new BasicStroke(1F));
    g2D.draw(f);
  }
    
  
  /////////////////////////////////////////////////////////////////////////////
  /** This writes an IDL file which can reproduce the plot.
   *
   *  @param f file which will contain the IDL program
   *  @param c graphics component which launches this method (useful
   *           for error message) */
  /////////////////////////////////////////////////////////////////////////////
  protected void writeIDLfile(File f, JComponent c) {
    PrintWriter out;
    boolean skipData = false;
    try {
      out = new PrintWriter(new FileWriter(f));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Unable to write to "+f.getName(),
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    int maxLength = 0;
    for(int i = 0; i < nDataSets; i++) {
      if (x.get(i) != null) {
        if (x.get(i).length > maxLength) maxLength = x.get(i).length;
      }
    }

    if (maxLength == 0) {
      out.println("; Warning: no data - plot not possible.");
      out.close();
      return;
    }

    out.println("; To run me type \"idl "+f.getName()+"\"");
    out.println(";------------------------------------------------------");
    out.println("; This part contains different variables which can be customised");
    out.println("!p.charsize = 1.5 ; change this to change the font size");
    out.println();
    out.println("filename = '"+f.getName().replaceAll("idl","eps")+"' ; this is the name of the output file");
    out.println("title = '!17"+IDLtitle+"' ; the title of the plot");
    out.println("xtitle = '!17"+xIDLtitle+"' ; the title of the x axis");
    out.println("ytitle = '!17"+yIDLtitle+"' ; the title of the y axis");
    out.println();
    out.println(";------------------------------------------------------");
    out.println("; This part contains various settings for the eps file");
    out.println("; Comment this part out if you want to visualise the plot");
    out.println("; directly from IDL.");
    out.println();
    out.println("set_plot,'ps'");
    out.println("device,file=filename,xsize=15,ysize=12,xoffset=0,yoffset=0,$");
    out.println("       /color,encapsulated=1,Bits_per_Pixel=8");
    out.println("print,'Writing in file ',filename");
    out.println(";------------------------------------------------------");
    out.println();
    out.println("nDataSets = "+nDataSets);
    out.println("xmin = "+xmin);
    out.println("xmax = "+xmax);
    out.println("ymin = "+ymin);
    out.println("ymax = "+ymax);
    out.println("xrange = [xmin, xmax]");
    out.println("yrange = [ymin, ymax]");
    out.println("x = dblarr("+maxLength+","+nDataSets+")");
    out.println("y = dblarr("+maxLength+","+nDataSets+")");
    out.println("leftError = dblarr("+maxLength+","+nDataSets+")");
    out.println("rightError = dblarr("+maxLength+","+nDataSets+")");
    out.println("upperError = dblarr("+maxLength+","+nDataSets+")");
    out.println("lowerError = dblarr("+maxLength+","+nDataSets+")");
    out.println("x(*,*) = 0.0");
    out.println("y(*,*) = 0.0");
    out.println("leftError(*,*) = 0.0");
    out.println("rightError(*,*) = 0.0");
    out.println("upperError(*,*) = 0.0");
    out.println("lowerError(*,*) = 0.0");
    int[] setLengths = new int[nDataSets];
    int jj;
    for(int i=0; i<nDataSets; i++) {
      if (x.get(i) != null) {
        jj = 0;
        for (int j=0; j<x.get(i).length; j++) {
          if (Double.isNaN(x.get(i)[j])||Double.isNaN(y.get(i)[j])) {
            skipData = true;
          } else {
            out.printf("x(%d,%d)=%19.12e%n",jj,i,x.get(i)[j]);
            out.printf("y(%d,%d)=%19.12e%n",jj,i,y.get(i)[j]);
            if (hErrorArray.get(i)) {
              out.printf("leftError(%d,%d)=%19.12e%n",jj,i,Math.abs(leftError.get(i)[j]));
              out.printf("rightError(%d,%d)=%19.12e%n",jj,i,Math.abs(rightError.get(i)[j]));
            }
            if (vErrorArray.get(i)) {
              if (lowerError.get(i) != null) 
                out.printf("lowerError(%d,%d)=%19.12e%n",jj,i,Math.abs(lowerError.get(i)[j]));
              if (upperError.get(i) != null) 
                out.printf("upperError(%d,%d)=%19.12e%n",jj,i,Math.abs(upperError.get(i)[j]));
            }
            jj++;
          }
        }
        setLengths[i] = jj;
      }
    }
    out.println();
    out.println("; the plot:");
    out.println("plot,xrange,yrange,/nodata,xrange=xrange,yrange=yrange,$");
    out.println("     xstyle=1,ystyle=1,title=title,$");
    out.print("     xtitle=xtitle,ytitle=ytitle");
    if (Xlog) out.print(",/xlog");
    if (Ylog) out.print(",/ylog");
    out.println();
    out.println();
    out.println("; This next part defines the different colors used in");
    out.println("; the plot by creating a custom color table:");
    out.println("red =   intarr("+nDataSets+")");
    out.println("green = intarr("+nDataSets+")");
    out.println("blue =  intarr("+nDataSets+")");
    for(int i=0; i<nDataSets; i++) {
      out.printf("red(%d) = %d%n",i,colorArray.get(i).getRed());
      out.printf("green(%d) = %d%n",i,colorArray.get(i).getGreen());
      out.printf("blue(%d) = %d%n",i,colorArray.get(i).getBlue());
    }
    out.println("tvlct,red,green,blue");
    out.println();
    out.println("; The style used for plotting the different lines can");
    out.println("; be modified by changing the value of linestyle.");
    out.println("; The choice of symbols can be modified by changing");
    out.println("; the value of psym.");
    out.println("dx = (xmax-xmin)/75.0");
    out.println("dy = (ymax-ymin)/75.0");
    for(int i=0; i<nDataSets; i++) {
      if ((x.get(i) != null)&&(x.get(i).length>0)) {
        if (linestyleArray.get(i) != -1)
          out.printf("oplot,x(0:%d,%d),y(0:%d,%d),linestyle=%d,color=%d,thick=2.0%n",
              setLengths[i]-1,i,setLengths[i]-1,i,linestyleArray.get(i),i);
        if (psymArray.get(i) != 0)
          out.printf("oplot,x(0:%d,%d),y(0:%d,%d),psym=%d,color=%d%n",
              setLengths[i]-1,i,setLengths[i]-1,i,psymArray.get(i),i);
        if (hErrorArray.get(i)) {
          out.println("i = "+i);
          out.println("for j=0,"+(setLengths[i]-1)+" do $");
          out.println("  oplot,x(j,i)-leftError(j,i)*[1,1,1,0,0,0]+rightError(j,i)*[0,0,0,1,1,1],$");
          out.println("        y(j,i)+dy*[1,-1,0,0,1,-1],linestyle=0,color="+i+",thick=2.0");
        } 
        if (vErrorArray.get(i)) {
          out.println("i = "+i);
          out.println("for j=0,"+(setLengths[i]-1)+" do $");
          out.println("  oplot,x(j,i)+dx*[1,-1,0,0,1,-1],$");
          out.println("        y(j,i)-lowerError(j,i)*[1,1,1,0,0,0]+upperError(j,i)*[0,0,0,1,1,1],$");
          out.println("        linestyle=0,color="+i+",thick=2.0");
        } 
      }
    }
    out.println("device,/close_file  ; this is necessary to make sure the entire eps file is outputed");
    out.println("exit");
    out.close();
    
    if (skipData) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the data was NaN\nand was therefore skipped",
        "Error",
        JOptionPane.WARNING_MESSAGE
      );
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This writes a TXT file which contains the data from the plot.
   *
   *  @param f file which will contain the plot data in ascii format
   *  @param c graphics component which launches this method (useful
   *           for error message) */
  /////////////////////////////////////////////////////////////////////////////
  protected void writeTXTfile(File f, JComponent c) {
    PrintWriter out;
    boolean skipData = false;
    boolean allNAN;
    try {
      out = new PrintWriter(new FileWriter(f));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Unable to write to "+f.getName(),
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    int maxLength = 0;
    for(int i = 0; i < nDataSets; i++) {
      if (x.get(i) != null) {
        if (x.get(i).length > maxLength) maxLength = x.get(i).length;
      }
    }

    if (maxLength == 0) {
      out.println("Warning: no data.");
      out.close();
      return;
    }

    for(int i=0; i<nDataSets; i++) {
      if (x.get(i) != null) {
        allNAN = true;
        for (int j=0; j<x.get(i).length; j++) {
          if (Double.isNaN(x.get(i)[j])||Double.isNaN(y.get(i)[j])) {
            skipData = true;
          } else {
            out.printf("%19.12e %19.12e%n",x.get(i)[j],y.get(i)[j]);
            allNAN = false;
          }
        }
        if (!allNAN) out.println("&");
      }
    }
    out.close();
    
    if (skipData) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the data was NaN\nand was therefore skipped",
        "Error",
        JOptionPane.WARNING_MESSAGE
      );
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots the axis (in box format) and the plot titles
  //
  // @param g2D graphics area
  // @param rect axis in box format
  /////////////////////////////////////////////////////////////////////////////
  private void plotFrame(Graphics2D g2D, Rectangle2D rect) {
    float width, height;

    g2D.setStroke(linestyle[Plot.solid]);
    g2D.draw(rect);

    // Insert title to x-axis:
    g2D.setFont(xyTitleFont);
    FontRenderContext frc = g2D.getFontRenderContext();
    Rectangle2D bounds = xyTitleFont.getStringBounds(xtitle,frc);
    width = (float) bounds.getWidth(); 
    height = getStringHeight(xtitle,xyTitleFont,frc);
    g2D.drawString(xtitle,(xleft()+xright()-width)/2F,vspan-(bottomSubMargin-height)/2F);

    // Insert title to y-axis (including rotating the text font):
    bounds = xyTitleFont.getStringBounds(ytitle,frc);
    width = (float) bounds.getWidth();
    height = getStringHeight(ytitle,xyTitleFont,frc);
    AffineTransform t = g2D.getTransform();
    g2D.rotate(-Math.PI/2.0);
    g2D.drawString(ytitle,-(ydown()+yup()+width)/2F,(leftSubMargin+height)/2F);
    g2D.setTransform(t);

    // Insert top title:
    g2D.setFont(titleFont);
    frc = g2D.getFontRenderContext();
    bounds = titleFont.getStringBounds(title,frc);
    width = (float) bounds.getWidth();
    height = getStringHeight(title,titleFont,frc);
    g2D.drawString(title,(xleft()+xright()-width)/2F,(height+bottomMargin)/2F);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This calculates the tick arrays for either axis in the linear case.
  //
  // @param z1 first bound to domain
  // @param z2 second bound to domain
  /////////////////////////////////////////////////////////////////////////////
  private void findTickArraysLinear(double z1, double z2) {
    double zspan = Math.abs(z2-z1);
    double zexp = Math.floor(Math.log10(zspan));
    double r = zspan/Math.pow(10.0,zexp);
    
    // dz  = the large increment (i.e. between major ticks)
    // ddz = the small increment (i.e. between minor ticks)
    // n   = the ratio dx/ddx
    double dz, ddz; int n;

    // this is quite similar to what is done in IDL:
         if (r <= 1.5) {dz = 0.2; ddz = 0.05; n = 4; }
    else if (r <= 3.5) {dz = 0.5; ddz = 0.1;  n = 5; }
    else if (r <= 7.0) {dz = 1.0; ddz = 0.1;  n = 10;}
    else               {dz = 2.0; ddz = 0.5;  n = 4; }

    dz  = dz*Math.pow(10.0,zexp);
    ddz = ddz*Math.pow(10.0,zexp);
    
    // In case "zmax" is smaller than "zmin":
    double zmax, zmin;
    if (z1>z2) {zmax=z1; zmin=z2;}
          else {zmax=z2; zmin=z1;}
              
    // Determine where the ticks start, using the small
    // increment ddz:
    double nstart = Math.ceil(zmin/ddz);
    double zstart = nstart*ddz;

    // The number of ticks (the 1.1 is to avoid truncation errors):
    nticks = (int)(Math.floor(zmax/ddz) - nstart + 1.1);
    
    // Allocate space for the tick arrays:
    ticks = new double[nticks];
    isMajor = new boolean[nticks];
    
    // i is just a counter to let me know when to put a
    // large tick.
    int i = (int)(((long)nstart)%(long)n);
    double zz = zstart;
    
    for (int j = 0; j < nticks; j++) {
      if (i == 0) {
        // we have a major tick
        isMajor[j] = true;
        ticks[j] = zz;
        // Perhaps this is unecessary, but it is there to make sure I
        // get 0 exactly rather than some stupid value like 10e-320:
        if (Math.abs(zz) < dz/2.0) {ticks[j] = 0.0;}
      } else {
        // we have a minor tick
        isMajor[j] = false;
        ticks[j] = zz;
      }
      i++;
      if (i==n) {i=0;}
      zz+=ddz;
    }

    /////////////////////////////////////
    // This next part is tricky:
    //
    // Determining the number of decimal
    // points and whether or not to keep
    // the exponent.
    /////////////////////////////////////
    
    // a new definition for zmax
    zmax = (Math.abs(z1)>Math.abs(z2)?Math.abs(z1):Math.abs(z2));

    // This gives the number of significant digits - 1 (the 0.1 is
    // to avoid truncation errors):
    int nDigits = (int)floor(Math.log10(zmax)) - (int)floor(Math.log10(dz));
    
    int maxExp = (int)floor(Math.log10(zmax));

    // This part is really complicated, but the main idea
    // is to try to find the shortest representation of
    // the numbers.
    if (maxExp >= 0) {
      if (maxExp > (nDigits+3)) {
        withExp = true;
        nDecimal = nDigits;
      } else {
        withExp = false;
        nDecimal = nDigits-maxExp;
        if (nDecimal < 0) {nDecimal = 0;}
      }
    } else {
      if (maxExp < -3) {
        withExp = true;
        nDecimal = nDigits;
      } else {
        withExp = false;
        nDecimal = nDigits-maxExp;
      }
    }
    withMantissa = true;
  }

  /////////////////////////////////////////////////////////////////////////////
  // This calculates the tick arrays for either axis in the logarithmic case.
  // IMPORTANT: I'm assuming z1 and z2 are strictly positive.
  //
  // @param z1 first bound to domain
  // @param z2 second bound to domain
  /////////////////////////////////////////////////////////////////////////////
  private void findTickArraysLog(double z1, double z2) {
    double zmin = z2>z1?z1:z2;
    double zmax = z2>z1?z2:z1;
    double zspan = zmax/zmin;

    // 1st case: use linear approach
    if (zspan < 10.0) {
      findTickArraysLinear(z1,z2);
      return;
    }

    // In the remaining cases, it is obvious how the major tick labels 
    // should be written.  Therefore, we set up the appropriate variables
    // straight away.
    withExp = true;
    withMantissa = false; 
    nDecimal = 0; // this doesn't even come into account

    // Starting and stopping exponents (the 0.1 is to avoid truncation errors):
    int expStart = (int)floor(Math.log10(zmin));
    int expStop  = (int)floor(Math.log10(zmax));
    int r = expStop-expStart;

    // 2nd case: logarithmic scale with linear minor increments:
    if (r < 6) {
      // The 0.1 is to avoid truncation errors:
      int nStart   = (int)ceil(zmin/Math.pow(10.0,expStart));
      int nStop    = (int)floor(zmax/Math.pow(10.0,expStop));
      if (nStart==10) {nStart=1; expStart++; r--;} // don't forget this ...
      nticks = (10-nStart) + nStop + 9*(r-1);
      ticks = new double[nticks];
      isMajor = new boolean[nticks];
      int nn = nStart;
      int exp = expStart;
      for(int i=0; i<nticks; i++) {
        ticks[i] = (double)nn*Math.pow(10.0,exp);
        if (nn == 1) {
          isMajor[i] = true;
        } else {
          isMajor[i] = false;
        }
        nn++;
        if (nn==10) {
          nn=1;
          exp++;
        }
      }
      return;
    }

    // 3rd case: logarithmic scale with logarithmic minor increments:
    if (zmin > Math.pow(10,expStart)) expStart++; 
    // dexp  = the large increment on exponent
    // ddexp = the small increment on exponent (i.e. between minor ticks)
    int dexp, ddexp;  // Note: 632 is the maximal value for r
         if (r <=  14) {dexp =   2; ddexp =   1;}
    else if (r <=  35) {dexp =   5; ddexp =   1;}
    else if (r <=  70) {dexp =  10; ddexp =   2;}
    else if (r <= 140) {dexp =  20; ddexp =   5;}
    else if (r <= 350) {dexp =  50; ddexp =  10;}
    else               {dexp = 100; ddexp =  20;}
    int n = dexp/ddexp;

    // find out where minor ticks start and stop (beware of integer division):
    int nStop  = (int)floor((double)expStop/(double)ddexp);
    int nStart = (int)ceil((double)expStart/(double)ddexp);

    // find the number of ticks:
    nticks = nStop - nStart + 1;
    
    // Allocate space for the tick arrays:
    ticks = new double[nticks];
    isMajor = new boolean[nticks];

    // create tick arrays:    
    int exp  = nStart*ddexp;
    for(int i = 0; i < nticks; i++) {
      ticks[i] = Math.pow(10.0,exp);
      if (exp%dexp == 0) {
        isMajor[i] = true;
      } else {
        isMajor[i] = false;
      }
      exp+=ddexp;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots the tick marks (major and minor) on the y-axis
  //
  // @param g2D graphics area
  /////////////////////////////////////////////////////////////////////////////
  private void plotYTicks(Graphics2D g2D) {
    if (Ylog) {
      findTickArraysLog(ymin,ymax);
    } else {
      findTickArraysLinear(ymin,ymax);
    }
    
    // find the maximum width for the y-axis tick labels
    float yLabelWidth = 0;
    float a_width;
    for(int j=0; j < nticks; j++) {
      if (isMajor[j]) {
        a_width = findNumberWidth(ticks[j],g2D);
        yLabelWidth = (a_width>yLabelWidth?a_width:yLabelWidth);
      }
    }

    // adjust left margin accordingly:    
    leftMargin = Math.max(34F + yLabelWidth,leftMinMargin);

    // A few preliminary declarations and preparations:
    GeneralPath f = new GeneralPath();
    double yvalue;
    float yy, vAlign;

    for (int j=0; j < nticks; j++) {
      yvalue = ticks[j];
      yy = yCoor(yvalue);
      if (isMajor[j]) {
        f.moveTo(xleft(),yy);
        f.lineTo(xleft()+yLongTick,yy);
        f.moveTo(xright(),yy);
        f.lineTo(xright()-yLongTick,yy);
        if (Math.abs(yy-ydown()) < 3F) {vAlign = 0F;} else {vAlign = 0.5F;}
        plotNumber(yvalue,nDecimal,withExp,withMantissa,xleft()-4F,yy,1F,vAlign,g2D);
      } else {
        f.moveTo(xleft(),yy);
        f.lineTo(xleft()+yShortTick,yy);
        f.moveTo(xright(),yy);
        f.lineTo(xright()-yShortTick,yy);
      }
    }
    g2D.setStroke(linestyle[Plot.solid]);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots the tick marks (major and minor) on the x-axis
  //
  // @param g2D graphics area
  /////////////////////////////////////////////////////////////////////////////
  private void plotXTicks(Graphics2D g2D) {
    if (Xlog) {
      findTickArraysLog(xmin,xmax);
    } else {
      findTickArraysLinear(xmin,xmax);
    }
    
    // A few preliminary declarations and preparations:
    GeneralPath f = new GeneralPath();
    double xvalue;
    float xx;

    for (int j=0; j < nticks; j++) {
      xvalue = ticks[j];
      xx = xCoor(xvalue);
      if (isMajor[j]) {
        f.moveTo(xx,ydown());
        f.lineTo(xx,ydown()-xLongTick);
        f.moveTo(xx,yup());
        f.lineTo(xx,yup()+xLongTick);
        plotNumber(xvalue,nDecimal,withExp,withMantissa,xx,ydown()+17F,0.5F,0F,g2D);
      } else {
        f.moveTo(xx,ydown());
        f.lineTo(xx,ydown()-xShortTick);
        f.moveTo(xx,yup());
        f.lineTo(xx,yup()+xShortTick);
      }
    }
    g2D.setStroke(linestyle[Plot.solid]);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots a number using scientific notation, and centers it around
  // a particular position.
  /////////////////////////////////////////////////////////////////////////////
  // @param z     the number to be plotted
  // @param g2D   graphics onto which to plot the number
  /////////////////////////////////////////////////////////////////////////////
  private float findNumberWidth(double z, Graphics2D g2D) {
    FontRenderContext frc = g2D.getFontRenderContext();
    Rectangle2D bounds;
    float  width;
    String labelCoef, labelExp;
    String sformat;

    if (z == 0) {
      bounds = annotationFont.getStringBounds("0",frc);
      width = (float) bounds.getWidth(); 
      return width;
    }
    
    if (withExp) {
      ////////////////////////////////////
      // The number uses an exponent
      ////////////////////////////////////
      double zexp,r;
      
      // separate number into mantissa (r) and exponent (zexp):
      // find width of exponent (zexp):
      zexp = Math.floor(Math.log10(Math.abs(z)));
      labelExp = String.format("%.0f",zexp);

      bounds = annotationFont.getStringBounds(tenString,frc);
      width  = (float) bounds.getWidth();
    
      bounds = exponentFont.getStringBounds(labelExp,frc);
      width += (float) bounds.getWidth();

      if (withMantissa) {
        // find width of mantissa (r):
        r = z/Math.pow(10.0,zexp);
        sformat = "%."+nDecimal+"f";
        labelCoef = String.format(sformat,r);

        // get dimensions of different components:
        bounds = annotationFont.getStringBounds(labelCoef,frc);
        width += (float) bounds.getWidth(); 

        bounds = timesFont.getStringBounds(timesString,frc);
        width += (float) bounds.getWidth();
      }

      // return the result:
      return width-1F;

    } else {
      //////////////////////////////////////
      // The number is without an exponent
      //////////////////////////////////////
      sformat = "%."+nDecimal+"f";
      labelCoef = String.format(sformat,z);

      // get dimensions of different components:
      bounds = annotationFont.getStringBounds(labelCoef,frc);
      width = (float) bounds.getWidth(); 
      return width;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots a number using scientific notation at a particular location
  // (xxpos, yypos), with a horizontal and vertical alignment determined by
  // (hAlign, vAlign).
  /////////////////////////////////////////////////////////////////////////////
  // @param z            the number to be plotted
  // @param n            the number of digits after the decimal point
  // @param wthExp       if true, use exponent, otherwise, omit exponent
  // @param wthMantissa  if true, include mantissa, otherwise, omit mantissa
  // @param xxpos        x position around which to center the number
  // @param yypos        y position around which to center the number
  // @param hAlign       this specifies how to align the number horizontally:
  //                     0f = flushleft, 1f = flushright
  // @param vAlign       this specifies how to align the number vertically:
  //                     0f = [baseline = yypos], 1f = [topline = yypos]
  // @param g2D          graphics onto which to plot the number
  /////////////////////////////////////////////////////////////////////////////
  private void plotNumber(double z, int n, boolean wthExp, boolean wthMantissa,
                          float xxpos, float yypos, float hAlign, float vAlign,
                          Graphics2D g2D) {
    FontRenderContext frc = g2D.getFontRenderContext();
    Rectangle2D bounds;
    float  widthCoef=0f, widthTimes=0f, widthTen, widthExp, width, height, xx, yy;
    String labelCoef="", labelExp;
    String sformat;

    // Treat zero separatly.  It will have 0 decimal places and
    // no exponent.
    if (z==0) {
      labelCoef = "0";

      // get dimensions of different components:
      bounds = annotationFont.getStringBounds(labelCoef,frc);
      widthCoef = (float) bounds.getWidth(); 
      height = getStringHeight(labelCoef, annotationFont, frc);

      // find starting position:
      xx = xxpos - hAlign*widthCoef;
      yy = yypos + vAlign*height;

      // plot the number:
      g2D.setFont(annotationFont);
      g2D.drawString(labelCoef,xx,yy);
      return;
    }
      
    if (wthExp) {
      ////////////////////////////////////
      // Plot the number using an exponent
      ////////////////////////////////////
      double zexp,r;
      
      // treat exponent (zexp):
      zexp = Math.floor(Math.log10(Math.abs(z)));
      labelExp = String.format("%.0f",zexp);

      bounds = annotationFont.getStringBounds(tenString,frc);
      widthTen = (float) bounds.getWidth();
    
      // I'm assuming that the exponent determines the maximum
      // height:
      bounds = exponentFont.getStringBounds(labelExp,frc);
      widthExp = (float) bounds.getWidth();
      height = 5F + getStringHeight(labelExp,exponentFont,frc);

      // treat mantissa (r):
      if (withMantissa) {
        r = z/Math.pow(10.0,zexp);
        sformat = "%."+n+"f";
        labelCoef = String.format(sformat,r);

        bounds = annotationFont.getStringBounds(labelCoef,frc);
        widthCoef = (float) bounds.getWidth(); 

        bounds = timesFont.getStringBounds(timesString,frc);
        widthTimes = (float) bounds.getWidth();
      }

      // find starting position:
      width = widthCoef+widthTimes+widthTen+widthExp-1F;
      
      // this is to avoid having the number go off the right edge:
      if ((xxpos + (1F-hAlign)*width) > (hspan-2F)) {
        xx = hspan - width - 2F;
      } else {
        xx = xxpos - hAlign*width;
      }
      yy = yypos + vAlign*height;

      // plot the number:
      g2D.setFont(annotationFont);
      if (withMantissa) g2D.drawString(labelCoef,xx,yy);
      g2D.drawString(tenString,xx+widthCoef+widthTimes,yy);
      if (withMantissa) { 
        g2D.setFont(timesFont);
        g2D.drawString(timesString,xx+widthCoef+1,yy-1);
      }
      g2D.setFont(exponentFont);
      g2D.drawString(labelExp,xx+widthCoef+widthTimes+widthTen-1F,yy-5F);
    } else {
      //////////////////////////////////////
      // Plot the number without an exponent
      //////////////////////////////////////
      sformat = "%."+n+"f";
      labelCoef = String.format(sformat,z);

      // get dimensions of different components:
      bounds = annotationFont.getStringBounds(labelCoef,frc);
      widthCoef = (float) bounds.getWidth(); 
      height = getStringHeight(labelCoef, annotationFont, frc);

      // find starting position:
      width = widthCoef;
      // this is to avoid having the number go off the right edge:
      if ((xxpos + (1F-hAlign)*width) > (hspan-2F)) {
        xx = hspan - width - 2F;
      } else {
        xx = xxpos - hAlign*width;
      }
      yy = yypos + vAlign*height;

      // plot the number:
      g2D.setFont(annotationFont);
      g2D.drawString(labelCoef,xx,yy);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This function calculates the height of a string:
  //
  // @param s   string to be plotted
  // @param f   font with which to plot the string
  // @param frc font render context (see java documentation)
  /////////////////////////////////////////////////////////////////////////////
  private float getStringHeight(String s, Font f, FontRenderContext frc) {
    LineMetrics metrics = f.getLineMetrics(s, frc);
    float ascent = metrics.getAscent();         // from baseline to top of text
    float descent = metrics.getDescent();       // from baseline to bottom of text
    //float leading = metrics.getLeading();     // additional space used in
                                                // between lines of text
    //float lineheight = metrics.getHeight();   // The total height of our text
                                                // = ascent + descent + leading

    // This may look incorrect, but it seems that in most cases, the ascent
    // actually overshoots the top of the text by an amount that's close to the
    // descent, hence the formula. Furthermore, I only want to center the ascent,
    // which is more aesthetic.
    // Final remark: the leading usually seems to be 0, so I do not use it.
    return ascent-descent;
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots a data set using the linestyle aLinestyle:
  /////////////////////////////////////////////////////////////////////////////
  // There is probably a simpler way of doing this, but I do not know it.
  // This function not only plots the function but clips the GeneralPath f
  // to a circular region which encompases the strict plotting area.  This
  // is to avoid spending a long time plotting line segments significantly
  // outside the plot area.
  /////////////////////////////////////////////////////////////////////////////
  // @param x          x values of data set
  // @param y          y values of data set
  // @param aLinestyle the line style with which to plot the data set
  // @param g2D        graphics area in which to plot the data set
  /////////////////////////////////////////////////////////////////////////////
  private void plotXY(double[] x, double[] y, int aLinestyle, Graphics2D g2D) {

    if (aLinestyle == -1) {return;}

    GeneralPath f = new GeneralPath();
    boolean isInside1, isInside2 = false, isNaN = false;
    float x1, y1, x2 = 0f, y2 = 0f, xTemp, yTemp, tt, dt;
    float xcenter, ycenter, radius2;
    int i, istart;

    // Rather than simply plotting and clipping, it is
    // necessary to clip beforehand to avoid wasting a
    // lot of time plotting outside the plot area.

    // the center of the plot area:
    xcenter = 0.5F*(xleft()+xright());
    ycenter = 0.5F*(yup()+ydown());

    // set radius2 slightly larger than half the diagonal squared
    radius2 = 0.3F*norm2(xright()-xleft(),ydown()-yup());

    for(istart=0; istart<x.length;istart++) {
      if (Double.isNaN(x[istart])) continue;
      if (Double.isNaN(y[istart])) continue;
      if (Double.isInfinite(x[istart])) continue;
      if (Double.isInfinite(y[istart])) continue;
      if (Xlog && (x[istart] <= 0.0)) continue;
      if (Ylog && (y[istart] <= 0.0)) continue;
      x2 = xCoor(x[istart]);
      y2 = yCoor(y[istart]); 
      isInside2 = isInside(x2,y2,xcenter,ycenter,radius2);
      if (isInside2) f.moveTo(x2,y2);
      break;
    }

    for (i = istart+1; i < x.length; i++) {
      if (Double.isNaN(x[i])) {isNaN = true; continue;}
      if (Double.isNaN(y[i])) {isNaN = true; continue;}
      if (Double.isInfinite(x[i])) {isNaN = true; continue;}
      if (Double.isInfinite(y[i])) {isNaN = true; continue;}
      if (Xlog && (x[i] <= 0.0)) {isNaN = true; continue;}
      if (Ylog && (y[i] <= 0.0)) {isNaN = true; continue;}

      x1 = x2;
      y1 = y2;
      isInside1 = isInside2;

      x2 = xCoor(x[i]);
      y2 = yCoor(y[i]);
      isInside2 = isInside(x2,y2,xcenter,ycenter,radius2);

      // Previous point was either NaN or infinite.
      // Avoid trying to connect to it.
      /////////////////////////////////////////////
      if (isNaN) {
        if (isInside2) f.moveTo(x2,y2);
        isNaN = false;
        continue;
      }

      // check to see if both points are identitical:
      if (0F == norm2(x2-x1,y2-y1)) {continue;}

      // Both points inside:
      //////////////////////
      if (isInside1&&isInside2) {
        f.lineTo(x2,y2);
        continue;
      }

      // First point inside, second point outside:
      ////////////////////////////////////////////
      if (isInside1) {
        // an approximate upper bound for tt:
        tt = 2F * (float)Math.sqrt(radius2/norm2(x2-x1,y2-y1));
        xTemp = x1 + tt*(x2-x1);
        yTemp = y1 + tt*(y2-y1);
        f.lineTo(xTemp,yTemp);
        continue;
      }
      
      // First point outside, second point inside:
      ////////////////////////////////////////////
      if (isInside2) {
        // an approximate upper bound for tt:
        tt = 2F * (float)Math.sqrt(radius2/norm2(x2-x1,y2-y1));
        xTemp = x2 + tt*(x1-x2);
        yTemp = y2 + tt*(y1-y2);
        f.moveTo(xTemp,yTemp);
        f.lineTo(x2,y2);
        continue;
      }

      // Both points outside, the most complicated case:
      //////////////////////////////////////////////////
      
      // check to see if the segment (x1,y1)-(x2,y2) intersects the circular
      // plot region:
      tt = ((xcenter-x1)*(x2-x1)+(ycenter-y1)*(y2-y1))/norm2(x2-x1,y2-y1);
      // no intersetion so go to next iteration:
      if ((tt < 0F)||(tt > 1F)) {continue;} 
      // the segment might intersect the plot region:
      dt = (float)Math.sqrt(radius2/norm2(x2-x1,y2-y1));
      xTemp = x1 + (tt-dt)*(x2-x1);
      yTemp = y1 + (tt-dt)*(y2-y1);
      f.moveTo(xTemp,yTemp);
      xTemp = x1 + (tt+dt)*(x2-x1);
      yTemp = y1 + (tt+dt)*(y2-y1);
      f.lineTo(xTemp,yTemp);
    }
         
    g2D.setStroke(linestyle[aLinestyle]);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots a data set using the symbols designated by aPsym.
  /////////////////////////////////////////////////////////////////////////////
  // @param x       x values of data set
  // @param y       y values of data set
  // @param aPsym   the type of symbols with which to plot the data set
  // @param g2D     graphics area in which to plot the data set
  /////////////////////////////////////////////////////////////////////////////
  private void plotXYpsym(double[] x, double[] y, int aPsym, Graphics2D g2D) {

    if (aPsym == 0) {return;}

    GeneralPath f = new GeneralPath();
    float xx, yy;

    for (int i = 0; i < x.length; i++) {
      if (Double.isNaN(x[i])) continue;
      if (Double.isNaN(y[i])) continue;
      if (Double.isInfinite(x[i])) continue;
      if (Double.isInfinite(y[i])) continue;
      if (Xlog && (x[i] <= 0.0)) continue;
      if (Ylog && (y[i] <= 0.0)) continue;
      xx = xCoor(x[i]);
      yy = yCoor(y[i]);
      for (int j = 0; j < psym[aPsym].length; j+=4) {
        f.moveTo(xx+psym[aPsym][j],yy+psym[aPsym][j+1]);
        f.lineTo(xx+psym[aPsym][j+2],yy+psym[aPsym][j+3]);
      }
    }
    
    g2D.setStroke(Plot.psymStroke);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots horizontal error bars for a data set.
  /////////////////////////////////////////////////////////////////////////////
  // @param x          x values of data set
  // @param y          y values of data set
  // @param leftError  the lower bound of the error bar on the x value
  // @param rightError the upper bound of the error bar on the x value
  // @param g2D        graphics area in which to plot the data set
  /////////////////////////////////////////////////////////////////////////////
  private void plotHError(double[] x, double[] y, double[] leftError, double[] rightError, Graphics2D g2D) {

    GeneralPath f = new GeneralPath();
    float x1, x2, yy;

    for(int i=0; i<x.length; i++) {
      if (Double.isNaN(x[i])) continue;
      if (Double.isNaN(y[i])) continue;
      if (Double.isInfinite(x[i])) continue;
      if (Double.isInfinite(y[i])) continue;
      if (Xlog && (x[i] <= 0.0)) continue;
      if (Ylog && (y[i] <= 0.0)) continue;
      yy = yCoor(y[i]);
      if (yy<(yup()-5F))   continue;
      if (yy>(ydown()+5F)) continue;
      // NOTE: x1 <= x2, regardless of reverseXrange
      if (reverseXrange) {
        x1 = xCoor(x[i]+rightError[i]);
        if (Xlog && (x[i]-leftError[i] <= 0.0)) {
          x2 = xright()+5F;
        } else {
          x2 = xCoor(x[i]-leftError[i]);
        }
      } else {
        if (Xlog && (x[i]-leftError[i] <= 0.0)) {
          x1 = xleft()-5F;
        } else {
          x1 = xCoor(x[i]-leftError[i]);
        }
        x2 = xCoor(x[i]+rightError[i]);
      }
      if (x2<(xleft()-5F))  continue;
      if (x1>(xright()+5F)) continue;
      if (x1<(xleft()-5F)) {
        f.moveTo(xleft()-5F,yy);
      } else {
        f.moveTo(x1,yy-5F);
        f.lineTo(x1,yy+5F);
        f.moveTo(x1,yy);
      }
      if (x2>(xright()+5F)){
        f.lineTo(xright()+5F,yy);
      } else {
        f.lineTo(x2,yy);
        f.moveTo(x2,yy-5F);
        f.lineTo(x2,yy+5F);
      }
    }
    g2D.setStroke(linestyle[solid]);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots vertical error bars for a data set.
  /////////////////////////////////////////////////////////////////////////////
  // @param x          x values of data set
  // @param y          y values of data set
  // @param lowerError the lower bound of the error bar on the y value
  // @param upperError the upper bound of the error bar on the y value
  // @param g2D        graphics area in which to plot the data set
  /////////////////////////////////////////////////////////////////////////////
  private void plotVError(double[] x, double[] y, double[] lowerError, double[] upperError, Graphics2D g2D) {

    GeneralPath f = new GeneralPath();
    float xx, y1, y2;

    for(int i=0; i<x.length; i++) {
      if (Double.isNaN(x[i])) continue;
      if (Double.isNaN(y[i])) continue;
      if (Double.isInfinite(x[i])) continue;
      if (Double.isInfinite(y[i])) continue;
      if (Xlog && (x[i] <= 0.0)) continue;
      if (Ylog && (y[i] <= 0.0)) continue;
      xx = xCoor(x[i]);
      if (xx<(xleft()-5F))  continue;
      if (xx>(xright()+5F)) continue;
      // NOTE: y1 >= y2, regardless of reverseYrange
      if (reverseYrange) {
        y1 = yCoor(y[i]+upperError[i]);
        if (Ylog && (y[i]-lowerError[i] <= 0.0)) {
          y2 = yup()-5F;
        } else {
          y2 = yCoor(y[i]-lowerError[i]);
        }
      } else {
        if (Ylog && (y[i]-lowerError[i] <= 0.0)) {
          y1 = ydown()+5F;
        } else {
          y1 = yCoor(y[i]-lowerError[i]);
        }
        y2 = yCoor(y[i]+upperError[i]);
      }
      if (y1<(yup()-5F))   continue;
      if (y2>(ydown()+5F)) continue;
      if (y1>(ydown()+5F)) {
        f.moveTo(xx,ydown()+5F);
      } else {
        f.moveTo(xx-5F,y1);
        f.lineTo(xx+5F,y1);
        f.moveTo(xx,y1);
      }
      if (y2<(yup()-5F)) {
        f.lineTo(xx,yup()-5F);
      } else {
        f.lineTo(xx,y2);
        f.moveTo(xx-5F,y2);
        f.lineTo(xx+5F,y2);
      }
    }
    g2D.setStroke(linestyle[solid]);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This function returns "true" when a specific point lies inside the plot
  // area (which is a circle centered on (xcenter,ycenter) and of radius
  // sqrt(radius2)), and "false" otherwise.
  /////////////////////////////////////////////////////////////////////////////
  // @param  xx       x coordinate of the point being tested
  // @param  yy       y coordinate of the point being tested
  // @param  xcenter  x coordinate of the center of the plot area
  // @param  ycenter  y coordinate of the center of the plot area
  // @param  radius2  squared radius off plot area
  // @return true = point is inside, false = point is outside
  /////////////////////////////////////////////////////////////////////////////
  private static boolean isInside(float xx, float yy, float xcenter,
                                  float ycenter, float radius2) {
    float distance2 = norm2(xx-xcenter,yy-ycenter);
    return (distance2 <= radius2);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This function returns the squared norm of vector (xx,yy).
  /////////////////////////////////////////////////////////////////////////////
  // @param  xx       x coordinate of vector
  // @param  yy       y coordinate of vector
  // @return squared norm of vector
  /////////////////////////////////////////////////////////////////////////////
  private static float norm2(float xx, float yy) {
    return xx*xx+yy*yy;
  }

  /////////////////////////////////////////////////////////////////////////////
  // This converts x values to x coordinates:
  //
  // @param  xvalue x value in the same units as the data sets
  // @return x value in terms of the graphics area
  /////////////////////////////////////////////////////////////////////////////
  public float xCoor(double xvalue) {
    if (Xlog) {
      return (float)(xleft() + (xright()-xleft())*(Math.log(xvalue)-xlmin)/(xlmax-xlmin));
    } else {
      return (float)(xleft() + (xright()-xleft())*(xvalue-xmin)/(xmax-xmin));
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This converts y values to y coordinates:
  //
  // @param  yvalue y value in the same units as the data sets
  // @return y value in terms of the graphics area
  /////////////////////////////////////////////////////////////////////////////
  public float yCoor(double yvalue) {
    if (Ylog) {
      return (float)(ydown() + (yup()-ydown())*(Math.log(yvalue)-ylmin)/(ylmax-ylmin));
    } else {
      return (float)(ydown() + (yup()-ydown())*(yvalue-ymin)/(ymax-ymin));
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This converts x coordinates to x values:
  //
  // @param  xcoor x value in terms of the graphics area
  // @return x value in the same units as the data sets
  /////////////////////////////////////////////////////////////////////////////
  protected double xValue(int xcoor) {
    double lambda = ((double)xcoor-(double)xleft())/(double)(xright()-xleft());
    if (Xlog) {
      return Math.exp((1.0-lambda)*xlmin+lambda*xlmax);
    } else {
      return (1.0-lambda)*xmin+lambda*xmax;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This converts y coordinates to y values:
  //
  // @param  ycoor y value in terms of the graphics area
  // @return y value in the same units as the data sets
  /////////////////////////////////////////////////////////////////////////////
  protected double yValue(int ycoor) {
    double lambda = ((double)ycoor-(double)ydown())/(double)(yup()-ydown());
    if (Ylog) {
      return Math.exp((1.0-lambda)*ylmin+lambda*ylmax);
    } else {
      return (1.0-lambda)*ymin+lambda*ymax;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This removes all data sets. */
  /////////////////////////////////////////////////////////////////////////////
  public void clear() {
    // clear various arrays
    x.clear();
    y.clear();
    leftError.clear();
    rightError.clear();
    lowerError.clear();
    upperError.clear();
    linestyleArray.clear();
    psymArray.clear();
    colorArray.clear();
    hErrorArray.clear();
    vErrorArray.clear();
    // update the number of data sets:
    nDataSets = 0;
    // change the bounds
    /*xmin=0; xmax=1; ymin=0; ymax=1;
    xlmin=Double.NEGATIVE_INFINITY; xmax = 0.0;
    ylmin=Double.NEGATIVE_INFINITY; ymax = 0.0;*/
  }

  /////////////////////////////////////////////////////////////////////////////
  /** Sets all arrays contained in x and y to null, but keeps the same number
   *  of data sets. */
  /////////////////////////////////////////////////////////////////////////////
  public void nullify() {
    for(int i=0; i<x.size(); i++) {
      x.set(i,null);
      y.set(i,null);
      leftError.set(i,null);
      rightError.set(i,null);
      lowerError.set(i,null);
      upperError.set(i,null);
    }
    // change the bounds
    xmin=0; xmax=1; ymin=0; ymax=1;
    xlmin=Double.NEGATIVE_INFINITY; xmax = 0.0;
    ylmin=Double.NEGATIVE_INFINITY; ymax = 0.0;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This nullifies a specific data set in a plot by setting its arrays 
   *  contained in this.x and this.y to null.
   *
   *  @param i index of data set to be nullified */
  /////////////////////////////////////////////////////////////////////////////
  public void nullify(int i) {
    if ((i >= 0)&&(i < x.size())) {
      x.set(i,null);
      y.set(i,null);
      leftError.set(i,null);
      rightError.set(i,null);
      lowerError.set(i,null);
      upperError.set(i,null);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This clones the plot */
  /////////////////////////////////////////////////////////////////////////////
  public Plot clone() {
    Plot newPlot = new Plot(title,xtitle,ytitle,nDataSets);
    newPlot.setIDLTitles(IDLtitle,xIDLtitle,yIDLtitle);
    for(int i=0; i<nDataSets; i++) {
      double[] x, y, leftError, rightError, lowerError, upperError;
      if (this.x.get(i) == null) {
        x = null;
      } else {
        x = this.x.get(i).clone();
      }
      if (this.y.get(i) == null) {
        y = null;
      } else {
        y = this.y.get(i).clone();
      }
      if (this.leftError.get(i) == null) {
        leftError = null;
      } else {
        leftError = this.leftError.get(i).clone();
      }
      if (this.rightError.get(i) == null) {
        rightError = null;
      } else {
        rightError = this.rightError.get(i).clone();
      }
      if (this.lowerError.get(i) == null) {
        lowerError = null;
      } else {
        lowerError = this.lowerError.get(i).clone();
      }
      if (this.upperError.get(i) == null) {
        upperError = null;
      } else {
        upperError = this.upperError.get(i).clone();
      }
      Color aColor = new Color(colorArray.get(i).getRGB());
      newPlot.replaceDataSet(i,x,y,linestyleArray.get(i),psymArray.get(i),aColor);
      newPlot.replaceLeftError(i,leftError);
      newPlot.replaceRightError(i,rightError);
      newPlot.replaceLowerError(i,lowerError);
      newPlot.replaceUpperError(i,upperError);
      newPlot.setHErrorValue(i,this.hErrorArray.get(i));
      newPlot.setVErrorValue(i,this.vErrorArray.get(i));
    }
    newPlot.setXmin(xmin);
    newPlot.setXmax(xmax);
    newPlot.setYmin(ymin);
    newPlot.setYmax(ymax);
    newPlot.extendXrange = extendXrange;
    newPlot.extendYrange = extendYrange;
    newPlot.reverseXrange = reverseXrange;
    newPlot.reverseYrange = reverseYrange;
    newPlot.Xlog = Xlog;
    newPlot.Ylog = Ylog;
    return newPlot;
  }

  /////////////////////////////////////////////////////////////////////////////
  // This duplicates the ceil of java but reterns a long rather than a double.
  // Care is taken to avoid round-off problems both for positive and negative
  // numbers.
  //
  // @param  ycoor y value in terms of the graphics area
  // @return y value in the same units as the data sets
  /////////////////////////////////////////////////////////////////////////////
  private long ceil(double x) {
    double y = Math.ceil(x);
    if (y < 0.0) {
      return (long)(y-0.1);
    } else {
      return (long)(y+0.1);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This duplicates the floor of java but reterns a long rather than a double.
  // Care is taken to avoid round-off problems both for positive and negative
  // numbers.
  //
  // @param  ycoor y value in terms of the graphics area
  // @return y value in the same units as the data sets
  /////////////////////////////////////////////////////////////////////////////
  private long floor(double x) {
    double y = Math.floor(x);
    if (y < 0.0) {
      return (long)(y-0.1);
    } else {
      return (long)(y+0.1);
    }
  }
}
