/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import inversion_applet.DataCenter;
import inversion_applet.Numerics;
import inversion_applet.Mode;
import inversion_applet.Kernel;
import org.netlib.arpack.Dnaupd;
import org.netlib.arpack.Dneupd;
import org.netlib.blas.Dgbmv;
import org.netlib.blas.Dgemm;
import org.netlib.lapack.Dgbtrf;
import org.netlib.lapack.Dgbtrs;
import org.netlib.lapack.Dgbsv;

public class Pulsations {
  private DataCenter dataCenter;

  // order of accuracy on pulsation calculations:
  protected int order = 2; // order = 2 corresponds to 4th order accuracy

  // variables which control the Arnoldi method:
  private final int nev = 2; // number of wanted eigenvalues;
  private final int ncv = 20;   // number of Arnoldi vectors

  // variable which determines whether to use lagrangian or eulerian pressure
  // perturbations in the pulsation equations
  protected int equations;
  static public final int FromFile = 0;
  static public final int Lagrange = 1;
  static public final int Euler = 2;
  static public final int JCD = 3;
  static public final String[] equationName = new String[]{"File","dP/P","P',rho'","JCD"};
  
  // This is what determines the order of the equations:
  private int eqEt, eqdP_P, eqP, eqRho, eqEr, eqPhiP, eqPhi,
              eqEr_radial, eqdP_P_radial, eqP_radial,
              eqY1, eqY2, eqY3, eqY4, eqY1_radial, eqY2_radial;

  // This is what determines the order of the variables:
  private int varEt, vardP_P, varP, varRho, varEr, varPhiP, varPhi,
              varEr_radial, vardP_P_radial, varP_radial,
              varY1, varY2, varY3, varY4, varY1_radial, varY2_radial;

  // This gives the number of lower and upper bands for order = 0 in the pulsation matrix,
  // as determined experimentally through TOP (1D version)
  private int kl0, ku0, kl0_radial, ku0_radial;

  // The number of variables and equations in the pulsation equations
  private int nvar, nvar_radial;

  // This specifies if an equation has been shifted due to a boundary condition
  private int[] shifted, shifted_radial;

  /******************************************************************************
  *  A constructor.
  ******************************************************************************/
  public Pulsations(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
    equations = JCD;
  }

  /******************************************************************************
  * This sets the order of the equations and variables
  ******************************************************************************/
  private void initEquations() {
    if (equations == Euler)    {initEquationsEuler();    initEquationsEulerRadial();}
    if (equations == Lagrange) {initEquationsLagrange(); initEquationsLagrangeRadial();}
    if (equations == JCD)      {initEquationsJCD();      initEquationsJCDRadial();}
  }
  
  /******************************************************************************
  * This finds the pulsation modes using the java translation of ARPACK.  A shift-
  * invert treatment is needed to target eigenvalues around each given shift.
  ******************************************************************************/
  public void findPulsations() {
    // initialise different variables according to the type of equations to be solved:
    initEquations();

    // various variables which specify the size of the problem:
    int ngrid = dataCenter.grid.length;
    int nfreq = dataCenter.modes.size();
    if (nfreq == 0) return;  // easy exit if possible
    dataCenter.sortObs();
    
    // variables needed for matrix operations:
    int ell = -1;
    int ntotal = 0;
    int kl = 0;
    int ku = 0;
    int ldmat = 0;
    int[] ipiv = null;
    double[][] mat = null;

    ///////////////////////////////////////////////////
    // variables (and their initial values) for ARPACK:
    // variables needed for Dnaupd:
    org.netlib.util.intW ido = new org.netlib.util.intW(0); // reverse communication parameter
    org.netlib.util.intW nevW = new org.netlib.util.intW(nev); // wrapper for nev
    org.netlib.util.intW info = new org.netlib.util.intW(0); // gives the status of the calculations
    org.netlib.util.doubleW tol = new org.netlib.util.doubleW(0.0); // desired precision (0.0 corresponds to machine precision)
    int[] iparam = new int[11];
    int[] ipntr = new int[14];
    int lworkl = 3*ncv*ncv + 6*ncv;
    double[] workl = new double[lworkl];
    double[] resid = null;
    double[] v = null;
    double[] workd = null;

    // supplementary variables needed for Dneupd:
    boolean[] select = new boolean[ncv];
    double[] dr = new double[nev+2]; // sometimes an extra eigenvalue appears and must be accounted for
    double[] di = new double[nev+2]; // sometimes an extra eigenvalue appears and must be accounted for
    double[] workev = new double[3*ncv];

    // variables needed to find Ik and modes.Enl:
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);
    double L2 = 0.0;

    double shift, diff, diffmin, num, den;
    double[] eig = new double[nev];
    double[] y1 = null;
    double[] y2 = null;
    double[] asigmab = null;
    int i, ii, k, l, kmin;
    double[] rlm = new double[ngrid];
    double[] rl = new double[ngrid];
    double[] integrationMatrix = Numerics.findIntegrationMatrix(dataCenter.grid);

    for(l=0;l<nfreq;l++) {
      if (dataCenter.modes.get(l).kernel != null) continue; // avoid recalculating the same mode
      if (ell != dataCenter.modes.get(l).l.intValue()) {
        ell = dataCenter.modes.get(l).l.intValue();
        L2 = (double)(ell*(ell+1));
        if (ell == 0) {
          ntotal = nvar_radial*ngrid;
          kl = order*nvar_radial+kl0_radial;
          ku = order*nvar_radial+ku0_radial;
          mat = createMatricesRadial(order);
        } else {
          ntotal = nvar*ngrid;
          kl = order*nvar+kl0;
          ku = order*nvar+ku0;
          mat = createMatrices(order,ell);
        }

        // various initialisations:
        ldmat = 2*kl+ku+1;
        ipiv = new int[ntotal];
        resid = new double[ntotal];
        v = new double[ncv*ntotal];
        workd = new double[3*ntotal];
        y1 = new double[ntotal];
        y2 = new double[ntotal];
        asigmab = new double[ntotal*ldmat];

        if (ell > 0) {
          for(i=0; i<ngrid; i++) {
            rlm[i] = Math.pow(dataCenter.grid[i],ell-1);
            rl[i] = rlm[i]*dataCenter.grid[i];
          }
        }
      }

      // Set up matrices and shift; factorise matrices
      shift = dataCenter.modes.get(l).FreqObs/dataCenter.FreqRef;
      for(i=0;i<ntotal*ldmat;i++) asigmab[i] = mat[0][i] - shift*shift*mat[1][i];
      Dgbtrf.dgbtrf(ntotal,ntotal,kl,ku,asigmab,0,ldmat,ipiv,0,info);
      if (info.val != 0) {
        System.out.println("Error with dgbtrf, info = " + info.val);
        System.exit(1);
      }

      ///////////////////////////////////////////////////////////
      // Set up various parameters for Dnaupd (this has to be
      // redone each time within the loop):
      ///////////////////////////////////////////////////////////
      ido.val = 0;
      nevW.val = nev;
      info.val = 0;
      tol.val = 0.0;
      iparam[0] = 1;  // shift strategy (use 1)
      iparam[2] = 100;  // maximum number of iterations
      iparam[3] = 1;  // block size (only 1 is implemented)
      iparam[6] = 1;  // selected mode: 1 = simple eigenvalue problem

      ///////////////////////////////////////////////////////////
      // Reverse communication loop:
      ///////////////////////////////////////////////////////////
      while (ido.val != 99) {
        Dnaupd.dnaupd(ido,"I",ntotal,"LM",nev,tol,resid,0,
                      ncv,v,0,ntotal,iparam,0,ipntr,0,workd,0,
                      workl,0,lworkl,info);
        if (Math.abs(ido.val) == 1) {
          Dgbmv.dgbmv("N",ntotal,ntotal,kl,kl+ku,1.0,mat[1],0,ldmat,workd,ipntr[0]-1,1,0.0,workd,ipntr[1]-1,1);
          Dgbtrs.dgbtrs("N",ntotal,kl,ku,1,asigmab,0,ldmat,ipiv,0,workd,ipntr[1]-1,ntotal,info);
        }
      }
      // Check exit status:
      if (info.val != 0) {
        System.out.println("Error with dnaupd, info = " + info.val);
        System.exit(1);
      }

      ///////////////////////////////////////////////////////////
      // Retrieve eigenvectors using Dneupd:
      ///////////////////////////////////////////////////////////
      // sanity check (in case there's 2 or more extra eigenvalues):
      if (iparam[4] > (nev+1)) {
        System.out.println("Warning: increased number of eigenvalues");
        dr = new double[iparam[4]+1];
        di = new double[iparam[4]+1];
      }
      // retrieve eigenvectors:
      Dneupd.dneupd(true,"A",select,0,dr,0,di,0,v,0,ntotal,0.0,
                    0.0,workev,0,"I",ntotal,"LM",nevW,tol.val,resid,
                    0,ncv,v,0,ntotal,iparam,0,ipntr,0,workd,0,workl,
                    0,lworkl,info);

      // Check exit status:
      if (info.val != 0) {
        System.out.println("Error with dneupd, info = " + info.val);
        System.exit(1);
      } 

      ///////////////////////////////////////////////////////////
      // Retrieve eigenvalues:
      ///////////////////////////////////////////////////////////
      kmin = -1; diffmin = 1e39;
      for (k=0; k<nev; k++) {
        // Option 1: use values supplied by Dneupd:
        eig[k] = shift*shift + 1.0/dr[k];

        // look for the eigenvalue which is closest to the shift: 
        diff = Math.abs(shift*shift-eig[k]);
        if (diffmin > diff) {
          diffmin = diff;
          kmin = k;
        }
      }

      if (di[kmin] != 0.0)
          System.out.println("Warning: complex eigenvalue "
            + dataCenter.modes.get(l).l + " " + shift + ": "+ di[kmin]);

      // create theoretical mode:
      Kernel aKernel = new Kernel(ell);
      dataCenter.kernels.add(aKernel);
      dataCenter.modes.get(l).kernel = aKernel;

      aKernel.mode = dataCenter.modes.get(l);
      aKernel.FreqDimless = Math.sqrt(eig[kmin]);
      aKernel.FreqTheo = aKernel.FreqDimless*dataCenter.FreqRef;

      // extract mode components and rescale:
      aKernel.equations = equations;
      aKernel.vDisp = new double[ngrid];
      aKernel.hDisp = new double[ngrid];
      aKernel.dP_P  = new double[ngrid];
      aKernel.P     = new double[ngrid];
      aKernel.rho   = new double[ngrid];
      aKernel.phip  = new double[ngrid];
      if (ell == 0) {
        if (equations == Lagrange) extractVariablesLagrangeRadial(aKernel,v,kmin*ntotal,rlm,rl,ngrid);
        if (equations == Euler)    extractVariablesEulerRadial(aKernel,v,kmin*ntotal,rlm,rl,ngrid);
        if (equations == JCD)      extractVariablesJCDRadial(aKernel,v,kmin*ntotal,rlm,rl,ngrid);
        aKernel.phi = Numerics.applyIntegrationMatrixReverse(integrationMatrix, aKernel.phip);
      } else {
        aKernel.phi = new double[ngrid];
        if (equations == Lagrange) extractVariablesLagrange(aKernel,v,kmin*ntotal,rlm,rl,ngrid);
        if (equations == Euler)    extractVariablesEuler(aKernel,v,kmin*ntotal,rlm,rl,ngrid);
        if (equations == JCD)      extractVariablesJCD(aKernel,v,kmin*ntotal,rlm,rl,ngrid);
      }
      aKernel.IK = Double.NaN;  // this tells findOtherVariables() to find IK, Enl, etc.
      aKernel.n  = Integer.MIN_VALUE; // this tell findRadialOrder() to find n.
    }
    findOtherVariables();
    findRadialOrders();
  }
  
  ///////////////////////////////////////////////////////////////////////////////
  // This initialises the matrices which contain the pulsation equations.
  ///////////////////////////////////////////////////////////////////////////////
  private double[][] createMatrices(int order, int ell) {
    int ngrid = dataCenter.grid.length;
    int kl = order*nvar+kl0;
    int ku = order*nvar+ku0;
    int offset = kl+ku;
    int ldmat = 2*kl+ku+1;
    double[][][][] matA  = new double[2*order][][][];
    double[][][][] matB  = new double[2*order][][][];
    double[][][] matAvgA = new double[2][nvar][nvar];
    double[][][] matAvgB = new double[2][nvar][nvar];
    double[][] result = new double[2][nvar*ngrid*ldmat];
    double LL = (double)(ell*(ell+1)), LS = (double)ell;
    int ii, i, j, ibis, jbis, k, nstart, nstop, n;

    // initialise derivation matrix:
    double[] aux = new double[ngrid];
    for(i=0; i<ngrid; i++) aux[i] = dataCenter.grid[i]*dataCenter.grid[i];
    double[][] dermat = Numerics.initDeriveFDBandRoots(aux,order,1);

    ArrayList<Integer> domains = Numerics.findDomains(dataCenter.grid);
    
    for(n=0; n<domains.size(); n+=2) {
      // main equations:
      if (equations == Euler) {
        for(ii=0; ii<order;ii++) matA[ii+order] = createSingleLevelMatrixAEuler(LS,LL,ii+domains.get(n));
        for(ii=0; ii<order;ii++) matB[ii+order] = createSingleLevelMatrixBEuler(ell,ii+domains.get(n));
      } else if (equations == Lagrange) {
        for(ii=0; ii<order;ii++) matA[ii+order] = createSingleLevelMatrixALagrange(LS,LL,ii+domains.get(n));
        for(ii=0; ii<order;ii++) matB[ii+order] = createSingleLevelMatrixBLagrange(ell,ii+domains.get(n));
      } else {
        for(ii=0; ii<order;ii++) matA[ii+order] = createSingleLevelMatrixAJCD(LS,LL,ii+domains.get(n));
        for(ii=0; ii<order;ii++) matB[ii+order] = createSingleLevelMatrixBJCD(LS,LL,ii+domains.get(n));
      }
      for(ii=domains.get(n); ii<domains.get(n+1); ii++) {
        nstart = -order+1;
        nstop  =  order+1;
        if ((nstart+ii) < domains.get(n)) nstart = domains.get(n)-ii;
        if ((nstop+ii) > (domains.get(n+1)+1))  nstop = domains.get(n+1)+1-ii;

        // find new single level matrices
        for (i=1;i<2*order;i++) {
          matA[i-1] = matA[i];
          matB[i-1] = matB[i];
        }
        if (equations == Euler) {
          matA[2*order-1] = createSingleLevelMatrixAEuler(LS,LL,ii+nstop-1);
          matB[2*order-1] = createSingleLevelMatrixBEuler(ell,ii+nstop-1);
        } else if (equations == Lagrange) {
          matA[2*order-1] = createSingleLevelMatrixALagrange(LS,LL,ii+nstop-1);
          matB[2*order-1] = createSingleLevelMatrixBLagrange(ell,ii+nstop-1);
        } else {
          matA[2*order-1] = createSingleLevelMatrixAJCD(LS,LL,ii+nstop-1);
          matB[2*order-1] = createSingleLevelMatrixBJCD(LS,LL,ii+nstop-1);
        }
        
        // find averaged matrices
        for(i=0;i<nvar;i++) {
          for(j=0;j<nvar;j++) {
            matAvgA[0][i][j]=matAvgA[1][i][j]=0;
            matAvgB[0][i][j]=matAvgB[1][i][j]=0;
            for(k=nstart; k<nstop; k++) {
              matAvgA[0][i][j]+=dermat[0][-k+order*(2*(ii+k)+1)]*matA[k+order-1][0][i][j];
              matAvgA[1][i][j]+=dermat[0][-k+order*(2*(ii+k)+1)]*matA[k+order-1][1][i][j];
              matAvgB[0][i][j]+=dermat[0][-k+order*(2*(ii+k)+1)]*matB[k+order-1][0][i][j];
              matAvgB[1][i][j]+=dermat[0][-k+order*(2*(ii+k)+1)]*matB[k+order-1][1][i][j];
            }
          }
        }

        // Set up matrices in banded form:
        for(i=0;i<nvar;i++) {
          ibis = nvar*(ii+shifted[i])+i;
          for(j=0;j<nvar;j++) {
            jbis = nvar*ii+j;
            result[0][offset+ibis-jbis+ldmat*jbis]+=matA[order-1][2][i][j];
            result[1][offset+ibis-jbis+ldmat*jbis]+=matB[order-1][2][i][j];
            for(k=nstart; k<nstop; k++) {
              jbis = nvar*(ii+k)+j;
              result[0][offset+ibis-jbis+ldmat*jbis]+=
                (matAvgA[0][i][j]*dermat[0][-k+order*(2*(ii+k)+1)]
                +matAvgA[1][i][j]*dermat[1][-k+order*(2*(ii+k)+1)]);
              result[1][offset+ibis-jbis+ldmat*jbis]+=
                (matAvgB[0][i][j]*dermat[0][-k+order*(2*(ii+k)+1)]
                +matAvgB[1][i][j]*dermat[1][-k+order*(2*(ii+k)+1)]);
            }
          }
        }
      }
      ii = domains.get(n+1);
      for(i=0;i<nvar;i++) {
        if (shifted[i] == 0) {
          ibis = nvar*ii+i;
          for(j=0;j<nvar;j++) {
            jbis = nvar*ii+j;
            result[0][offset+ibis-jbis+ldmat*jbis]+=matA[order-1][2][i][j];
            result[1][offset+ibis-jbis+ldmat*jbis]+=matB[order-1][2][i][j];
          }
        }
      }
    }

    // insert interface conditions:
    for(n=1; n<domains.size()-2; n+=2) {
      if (equations == Euler)    insertInterfaceConditionsEuler(result,offset,ldmat,kl,ku,ngrid,domains.get(n));
      if (equations == Lagrange) insertInterfaceConditionsLagrange(result,offset,ldmat,kl,ku,ngrid,domains.get(n));
      if (equations == JCD)      insertInterfaceConditionsJCD(result,offset,ldmat,kl,ku,ngrid,LL,domains.get(n));
    }    

    // insert boundary counditions:
    if (equations == Euler)     insertBoundaryConditionsEuler(result,offset,ldmat,kl,ku,ngrid,LS);
    if (equations == Lagrange)  insertBoundaryConditionsLagrange(result,offset,ldmat,kl,ku,ngrid,LS);
    if (equations == JCD)       insertBoundaryConditionsJCD(result,offset,ldmat,kl,ku,ngrid,LS);

    return result;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This initialises the matrices which contain the pulsation equations.
  ///////////////////////////////////////////////////////////////////////////////
  private double[][] createMatricesRadial(int order) {
    int ngrid = dataCenter.grid.length;
    int kl = order*nvar_radial+kl0_radial;
    int ku = order*nvar_radial+ku0_radial;
    int offset = kl+ku;
    int ldmat = 2*kl+ku+1;
    double[][][][] matA  = new double[2*order][][][];
    double[][][][] matB  = new double[2*order][][][];
    double[][][] matAvgA = new double[2][nvar_radial][nvar_radial];
    double[][][] matAvgB = new double[2][nvar_radial][nvar_radial];
    double[][] result = new double[2][nvar_radial*ngrid*ldmat];
    int ii, i, j, ibis, jbis, k, nstart, nstop, n;

    // initialise derivation matrix:
    double[] aux = new double[ngrid];
    for(i=0; i<ngrid; i++) aux[i] = dataCenter.grid[i]*dataCenter.grid[i];
    double[][] dermat = Numerics.initDeriveFDBandRoots(aux,order,1);

    ArrayList<Integer> domains = Numerics.findDomains(dataCenter.grid);
    
    for(n=0; n<domains.size(); n+=2) {
      // main equations:
      if (equations == Euler) {
        for(ii=0; ii<order;ii++) matA[ii+order] = createSingleLevelMatrixAEulerRadial(ii+domains.get(n));
        for(ii=0; ii<order;ii++) matB[ii+order] = createSingleLevelMatrixBEulerRadial(ii+domains.get(n));
      } else if (equations == Lagrange) {
        for(ii=0; ii<order;ii++) matA[ii+order] = createSingleLevelMatrixALagrangeRadial(ii+domains.get(n));
        for(ii=0; ii<order;ii++) matB[ii+order] = createSingleLevelMatrixBLagrangeRadial(ii+domains.get(n));
      } else {
        for(ii=0; ii<order;ii++) matA[ii+order] = createSingleLevelMatrixAJCDRadial(ii+domains.get(n));
        for(ii=0; ii<order;ii++) matB[ii+order] = createSingleLevelMatrixBJCDRadial(ii+domains.get(n));
      }
      for(ii=domains.get(n); ii<domains.get(n+1); ii++) {
        nstart = -order+1;
        nstop  =  order+1;
        if ((nstart+ii) < domains.get(n))      nstart = domains.get(n)-ii;
        if ((nstop+ii) > (domains.get(n+1)+1)) nstop  = domains.get(n+1)+1-ii;

        // find new single level matrices
        for (i=1;i<2*order;i++) {
          matA[i-1] = matA[i];
          matB[i-1] = matB[i];
        }
        if (equations == Euler) {
          matA[2*order-1] = createSingleLevelMatrixAEulerRadial(ii+nstop-1);
          matB[2*order-1] = createSingleLevelMatrixBEulerRadial(ii+nstop-1);
        } else if (equations == Lagrange) {
          matA[2*order-1] = createSingleLevelMatrixALagrangeRadial(ii+nstop-1);
          matB[2*order-1] = createSingleLevelMatrixBLagrangeRadial(ii+nstop-1);
        } else {
          matA[2*order-1] = createSingleLevelMatrixAJCDRadial(ii+nstop-1);
          matB[2*order-1] = createSingleLevelMatrixBJCDRadial(ii+nstop-1);
        }

        // find averaged matrices
        for(i=0;i<nvar_radial;i++) {
          for(j=0;j<nvar_radial;j++) {
            matAvgA[0][i][j]=matAvgA[1][i][j]=0;
            matAvgB[0][i][j]=matAvgB[1][i][j]=0;
            for(k=nstart; k<nstop; k++) {
              matAvgA[0][i][j]+=dermat[0][-k+order*(2*(ii+k)+1)]*matA[k+order-1][0][i][j];
              matAvgA[1][i][j]+=dermat[0][-k+order*(2*(ii+k)+1)]*matA[k+order-1][1][i][j];
              matAvgB[0][i][j]+=dermat[0][-k+order*(2*(ii+k)+1)]*matB[k+order-1][0][i][j];
              matAvgB[1][i][j]+=dermat[0][-k+order*(2*(ii+k)+1)]*matB[k+order-1][1][i][j];
            }
          }
        }

        // Set up matrices in banded form:
        for(i=0;i<nvar_radial;i++) {
          ibis = nvar_radial*(ii+shifted_radial[i])+i;
          for(j=0;j<nvar_radial;j++) {
            jbis = nvar_radial*ii+j;
            result[0][offset+ibis-jbis+ldmat*jbis]+=matA[order-1][2][i][j];
            result[1][offset+ibis-jbis+ldmat*jbis]+=matB[order-1][2][i][j];
            for(k=nstart; k<nstop; k++) {
              jbis = nvar_radial*(ii+k)+j;
              result[0][offset+ibis-jbis+ldmat*jbis]+=
                (matAvgA[0][i][j]*dermat[0][-k+order*(2*(ii+k)+1)]
                +matAvgA[1][i][j]*dermat[1][-k+order*(2*(ii+k)+1)]);
              result[1][offset+ibis-jbis+ldmat*jbis]+=
                (matAvgB[0][i][j]*dermat[0][-k+order*(2*(ii+k)+1)]
                +matAvgB[1][i][j]*dermat[1][-k+order*(2*(ii+k)+1)]);
            }
          }
        }
      }
      ii = ngrid-1;
      for(i=0;i<nvar_radial;i++) {
        if (shifted[i] == 0) {
          ibis = nvar_radial*ii+i;
          for(j=0;j<nvar_radial;j++) {
            jbis = nvar_radial*ii+j;
            result[0][offset+ibis-jbis+ldmat*jbis]+=matA[order-1][2][i][j];
            result[1][offset+ibis-jbis+ldmat*jbis]+=matB[order-1][2][i][j];
          }
        }
      }
    }

    // insert interface conditions:
    for(n=1; n<domains.size()-2; n+=2) {
      if (equations == Euler)    insertInterfaceConditionsEulerRadial(result,offset,ldmat,kl,ku,ngrid,domains.get(n));
      if (equations == Lagrange) insertInterfaceConditionsLagrangeRadial(result,offset,ldmat,kl,ku,ngrid,domains.get(n));
      if (equations == JCD)      insertInterfaceConditionsJCDRadial(result,offset,ldmat,kl,ku,ngrid,domains.get(n));
    }    

    // Boundary conditions
    if (equations == Euler)     insertBoundaryConditionsEulerRadial(result,offset,ldmat,kl,ku,ngrid);
    if (equations == Lagrange)  insertBoundaryConditionsLagrangeRadial(result,offset,ldmat,kl,ku,ngrid);
    if (equations == JCD)       insertBoundaryConditionsJCDRadial(result,offset,ldmat,kl,ku,ngrid);

    return result;
  }

  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////  Euler  --  non-radial  ///////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////

  /******************************************************************************
  * This sets the order of the equations and variables using eulerian pressure
  * and density perturbations.
  ******************************************************************************/
  private void initEquationsEuler() {
    eqRho = 0;
    eqP = 1;
    eqEt = 2;
    eqEr = 3;
    eqPhi = 4;
    eqPhiP = 5;

    varEr = 0;
    varEt = 1;
    varRho = 2;
    varP = 3;
    varPhi = 4;
    varPhiP = 5;

    kl0 = 0;
    ku0 = 2;
    nvar = 6;
    shifted = new int[]{1, 0, 0, 0, 1, 0}; // equations shifted due to boundary conditions
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixAEuler(double LS, double LL, int i) {
    double[][][] mat = new double[3][nvar][nvar];  // I'll assume everything is initialised to zero

    // Equation on Rho:
    mat[0][eqRho][varRho]   = dataCenter.grid[i]*dataCenter.grid[i];
    mat[1][eqRho][varEr]    = 2.0*dataCenter.rho[i]*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqRho][varEr]    = (LS+1.0)*dataCenter.rho[i] + dataCenter.grid[i]*dataCenter.drho[i];
    mat[0][eqRho][varEt]    = -LL*dataCenter.rho[i];

    // Equation on Er:
    mat[1][eqEr][varP]      = 2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqEr][varP]      = LS;
    mat[0][eqEr][varRho]    = dataCenter.gravity[i]*dataCenter.grid[i];
    mat[0][eqEr][varPhiP]   = dataCenter.rho[i];
    
    // Equation on Et:
    mat[2][eqEt][varP]      = 1.0;
    mat[2][eqEt][varPhi]    = dataCenter.rho[i];

    // Equation on P:
    mat[2][eqP][varP]       = 1.0;
    mat[2][eqP][varRho]     = -dataCenter.c2[i];

    // terms with singularities in the center:
    if (i==0) {
      mat[2][eqP][varEr]    = -4.0*Math.PI*dataCenter.rho[0]*dataCenter.rho[0]/3.0
                              -dataCenter.c2[0]*dataCenter.ddrho;
    } else {
      mat[2][eqP][varEr]    =  dataCenter.rho[i]*dataCenter.N2[i]*dataCenter.c2[i]
                            / (dataCenter.grid[i]*dataCenter.gravity[i]);
    }

    // Equation on Phi:
    mat[1][eqPhi][varPhi]   =  2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqPhi][varPhi]   =  LS;
    mat[0][eqPhi][varPhiP]  = -1.0;
    
    // Equation on PhiP:
    mat[1][eqPhiP][varPhiP] = 2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqPhiP][varPhiP] = LS+1.0;
    mat[0][eqPhiP][varPhi]  = -LL;
    mat[0][eqPhiP][varRho]  = -4.0*Math.PI*dataCenter.grid[i]*dataCenter.grid[i];
    
    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixBEuler(int l, int i) {
    double[][][] mat = new double[3][nvar][nvar];  // I'll assume everything is initialised to zero

    // Equation on Er:
    mat[0][eqEr][varEr] = dataCenter.rho[i];

    // Equation on Et:
    mat[2][eqEt][varEt] = dataCenter.rho[i];
    
    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert interface conditions in the Eulerian case:
  ///////////////////////////////////////////////////////////////////////////////
  // NOTE: the interface is between points i and i+1
  ///////////////////////////////////////////////////////////////////////////////
  private void insertInterfaceConditionsEuler(double[][] result, int offset, int ldmat,
                                             int kl, int ku, int ngrid, int i) {
    int ibis, jbis;
    // Interface condition on eqPhi (shifted):
    ibis = nvar*(i+1) + eqPhi;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varPhi;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
    jbis = nvar*i + varPhi;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;

    // Interface condition on eqPhiP (normal):
    ibis = nvar*i+eqPhiP;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varPhiP;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = nvar*(i+1) + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] = 4.0*Math.PI*dataCenter.rho[i+1];
    jbis = nvar*i + varPhiP;
    result[0][offset+ibis-jbis+ldmat*jbis] =-1.0;
    jbis = nvar*i + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] =-4.0*Math.PI*dataCenter.rho[i];
    
    // Interface condition on eqRho (shifted):
    ibis = nvar*(i+1) + eqRho;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varP;
    result[0][offset+ibis-jbis+ldmat*jbis] =  dataCenter.grid[i+1];
    jbis = nvar*(i+1) + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] = -dataCenter.rho[i+1]*dataCenter.gravity[i+1];
    jbis = nvar*i + varP;
    result[0][offset+ibis-jbis+ldmat*jbis] = -dataCenter.grid[i];
    jbis = nvar*i + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] =  dataCenter.rho[i]*dataCenter.gravity[i];

    // Interface condition on eqEr (normal):
    ibis = nvar*i + eqEr;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
    jbis = nvar*i + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert boundary conditions in the Eulerian case:
  ///////////////////////////////////////////////////////////////////////////////
  private void insertBoundaryConditionsEuler(double[][] result, int offset, int ldmat,
                                             int kl, int ku, int ngrid, double LS) {
    int ibis, jbis;
    // Boundary condition on eqPhi (shifted):
    ibis = eqPhi;
    for(jbis=0; jbis<=ibis+ku; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = varPhiP;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;
    jbis = varPhi;
    result[0][offset+ibis-jbis+ldmat*jbis] = LS;

    // Boundary condition on eqPhiP (normal):
    ibis = nvar*(ngrid-1)+eqPhiP;
    for(jbis=ibis-kl; jbis<nvar*ngrid; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(ngrid-1)+varPhiP;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = nvar*(ngrid-1)+varPhi;
    result[0][offset+ibis-jbis+ldmat*jbis] = LS+1.0; // for scaled variables, otherwise /grid[ngrid-1];
    jbis = nvar*(ngrid-1)+varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] = 4.0*Math.PI*dataCenter.rho[ngrid-1];
    
    // Boundary condition on eqRho (shifted):
    ibis = eqRho;
    for(jbis=0; jbis<=ibis+ku; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = varEt;
    result[0][offset+ibis-jbis+ldmat*jbis] = -LS;

    // Boundary condition on eqEr (normal):
    ibis = nvar*(ngrid-1)+eqEr;
    for(jbis=ibis-kl; jbis<nvar*ngrid; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(ngrid-1)+varP;
    result[0][offset+ibis-jbis+ldmat*jbis] = dataCenter.grid[ngrid-1];
    jbis = nvar*(ngrid-1)+varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] =-dataCenter.rho[ngrid-1]*dataCenter.gravity[ngrid-1];

  }

  ///////////////////////////////////////////////////////////////////////////////
  // This extracts a standard set of pulsation variables from the pulsation
  // calculations.
  ///////////////////////////////////////////////////////////////////////////////
  private void extractVariablesEuler(Kernel aKernel, double[] v, int offset, double[] rlm,
                                     double[] rl, int ngrid) {
    double cnst  = 1.0/(v[offset+nvar*dataCenter.indSurface+varEr]*rlm[dataCenter.indSurface]);
    int i, ii;
    for(i=0; i<ngrid; i++) {
      ii = offset+nvar*i;
      aKernel.vDisp[i] = v[ii+varEr]*rlm[i]*cnst;
      aKernel.hDisp[i] = v[ii+varEt]*rlm[i]*cnst;
      aKernel.phi[i]   = v[ii+varPhi]*rl[i]*cnst;
      aKernel.phip[i]  = v[ii+varPhiP]*rlm[i]*cnst;
      aKernel.P[i]     = v[ii+varP]*rl[i]*cnst;
      aKernel.rho[i]   = v[ii+varRho]*rl[i]*cnst;
      aKernel.dP_P[i]  =(aKernel.P[i]
                       - dataCenter.rho[i]
                       * dataCenter.gravity[i]
                       * aKernel.vDisp[i])
                       / dataCenter.pressure[i];
    }
    ii = offset;
    aKernel.vDisp_center = v[ii+varEr]*cnst;
    aKernel.hDisp_center = v[ii+varEt]*cnst;
    cnst/=Math.pow(dataCenter.grid[2],2.0);
    aKernel.vDisp_center_pp = cnst*(v[ii+varEr+2*nvar]-v[ii+varEr]);
    aKernel.hDisp_center_pp = cnst*(v[ii+varEt+2*nvar]-v[ii+varEt]);
  }

  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////  Euler  --  radial  ///////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////

  /******************************************************************************
  * This sets the order of the equations and variables using eulerian pressure
  * and density perturbations.
  ******************************************************************************/
  private void initEquationsEulerRadial() {
    eqP_radial = 0;
    eqEr_radial = 1;
    
    varP_radial = 0;
    varEr_radial = 1;
    
    kl0_radial = 0;
    ku0_radial = 0;
    nvar_radial = 2;
    shifted_radial = new int[]{1, 0};
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixAEulerRadial(int i) {
    double[][][] mat = new double[3][nvar_radial][nvar_radial];  // I'll assume everything is initialised to zero

    // Equation on P:
    mat[0][eqP_radial][varP_radial]    = 1.0/(dataCenter.Gamma1[i]*dataCenter.pressure[i]);
    mat[1][eqP_radial][varEr_radial]   = 2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqP_radial][varEr_radial]   = 3.0 - dataCenter.grid[i]*dataCenter.gravity[i]/dataCenter.c2[i];

    // Equation on Er:
    mat[1][eqEr_radial][varP_radial]   = 2.0;
    mat[0][eqEr_radial][varEr_radial]  = -(4.0*Math.PI*Math.pow(dataCenter.rho[i],2.0)
                                + Math.pow(dataCenter.gravity[i],2.0)*dataCenter.rho[i]/dataCenter.c2[i]
                                + dataCenter.gravity[i]*dataCenter.drho[i]);

    // special treatment in the center:
    if (i == 0) {
      mat[0][eqEr_radial][varP_radial]   = dataCenter.dgravity[0]/dataCenter.c2[0];
    } else {
      mat[0][eqEr_radial][varP_radial]   = dataCenter.gravity[i]/(dataCenter.grid[i]*dataCenter.c2[i]);
    }
    
    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixBEulerRadial(int i) {
    double[][][] mat = new double[3][nvar_radial][nvar_radial];

    // Equation on Er:
    mat[0][eqEr_radial][varEr_radial] = dataCenter.rho[i];

    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert interface conditions in the radial Eulerian case:
  ///////////////////////////////////////////////////////////////////////////////
  // NOTE: the interface is between points i and i+1
  ///////////////////////////////////////////////////////////////////////////////
  private void insertInterfaceConditionsEulerRadial(double[][] result, int offset, int ldmat,
                                             int kl, int ku, int ngrid, int i) {
    int ibis, jbis;
    
    // Interface condition on eqP (shifted):
    ibis = nvar_radial*(i+1) + eqP_radial;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar_radial*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar_radial*(i+1) + varP_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
    jbis = nvar_radial*(i+1) + varEr_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = -dataCenter.rho[i+1]*dataCenter.gravity[i+1]*dataCenter.grid[i+1];
    jbis = nvar_radial*i + varP_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;
    jbis = nvar_radial*i + varEr_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] =  dataCenter.rho[i]*dataCenter.gravity[i]*dataCenter.grid[i];

    // Interface condition on eqEr (normal):
    ibis = nvar_radial*i + eqEr_radial;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar_radial*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar_radial*(i+1) + varEr_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
    jbis = nvar_radial*i + varEr_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert boundary conditions in the radial Eulerian case:
  ///////////////////////////////////////////////////////////////////////////////
  private void insertBoundaryConditionsEulerRadial(double[][] result, int offset, int ldmat,
                                                   int kl, int ku, int ngrid) {
    int ibis, jbis;

    // Boundary condition on eqP (shifted):
    ibis = eqP_radial;
    for(jbis=0; jbis<=ibis+ku; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = varP_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0/(dataCenter.Gamma1[1]*dataCenter.pressure[1]);
    jbis = varEr_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = 3.0;

    // Boundary condition on eqEr (normal):
    ibis = nvar_radial*(ngrid-1)+eqEr_radial;
    for(jbis=ibis-kl; jbis<nvar_radial*ngrid; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar_radial*(ngrid-1)+varP_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = nvar_radial*(ngrid-1)+varEr_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = -dataCenter.gravity[ngrid-1]*dataCenter.rho[ngrid-1]*dataCenter.grid[ngrid-1];
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This extracts a standard set of pulsation variables from the pulsation
  // calculations.
  ///////////////////////////////////////////////////////////////////////////////
  private void extractVariablesEulerRadial(Kernel aKernel, double[] v, int offset, double[] rlm,
                                           double[] rl, int ngrid) {
    double cnst = 1.0/(v[offset+nvar_radial*dataCenter.indSurface+varEr_radial]*dataCenter.grid[dataCenter.indSurface]);
    int i, ii;
    for(i=0; i<ngrid; i++) {
      ii = offset+nvar_radial*i;
      aKernel.vDisp[i] = v[ii+varEr_radial]*dataCenter.grid[i]*cnst;
      aKernel.hDisp[i] = 0.0;
      aKernel.phip[i]  = -4.0*Math.PI*dataCenter.rho[i]
                       * aKernel.vDisp[i];
      aKernel.P[i]     = v[ii+varP_radial]*cnst;
      aKernel.dP_P[i]  =(aKernel.P[i]
                       - dataCenter.rho[i]
                       * dataCenter.gravity[i]
                       * aKernel.vDisp[i])
                       / dataCenter.pressure[i];
      aKernel.rho[i]   = aKernel.dP_P[i]
                       * dataCenter.rho[i]
                       / dataCenter.Gamma1[i]
                       - dataCenter.drho[i]
                       * aKernel.vDisp[i];
    }
    ii = offset;
    aKernel.vDisp_center = v[ii+varEr_radial]*cnst;
    aKernel.hDisp_center = 0.0;
    cnst/=Math.pow(dataCenter.grid[2],2.0);
    aKernel.vDisp_center_pp = cnst*(v[ii+varEr_radial+2*nvar_radial]-v[ii+varEr_radial]);
    aKernel.hDisp_center_pp = 0.0;
  } 

  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////  Lagrange -- non-radial  //////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////

  /******************************************************************************
  * This sets the order of the equations and variables using lagragian pressure
  * perturbations.
  ******************************************************************************/
  private void initEquationsLagrange() {
    eqEt = 0;
    eqdP_P = 1;
    eqEr = 2;
    eqPhiP = 3;
    eqPhi = 4;

    varEt = 0;
    vardP_P = 1;
    varEr = 2;
    varPhiP = 3;
    varPhi = 4;

    kl0 = 1;
    ku0 = 1;
    nvar = 5;
    shifted = new int[]{0, 1, 0, 0, 1}; // equations shifted due to boundary conditions
  }
    
  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixALagrange(double LS, double LL, int i) {
    double[][][] mat = new double[3][nvar][nvar];  // I'll assume everything is initialised to zero

    // Equation on dP/P:
    mat[0][eqdP_P][vardP_P] = dataCenter.grid[i]*dataCenter.grid[i]/dataCenter.Gamma1[i];
    mat[1][eqdP_P][varEr]   = 2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqdP_P][varEr]   = LS+1.0;
    mat[0][eqdP_P][varEt]   = -LL;

    // Equation on Er:
    mat[1][eqEr][vardP_P]   = 2.0*dataCenter.grid[i]*dataCenter.grid[i]*dataCenter.c2[i]/dataCenter.Gamma1[i];
    mat[0][eqEr][vardP_P]   = dataCenter.gravity[i]*dataCenter.grid[i]*(1.0/dataCenter.Gamma1[i]-1.0)
                            + LS*dataCenter.c2[i]/dataCenter.Gamma1[i];
    mat[0][eqEr][varPhiP]   = 1.0;
    mat[1][eqEr][varEr]     = 2.0*dataCenter.grid[i]*dataCenter.gravity[i];
    
    // Equation on Et:
    mat[2][eqEt][vardP_P]   = dataCenter.c2[i]/dataCenter.Gamma1[i];
    mat[2][eqEt][varPhi]    = 1.0;

    // terms with singularities in the center:
    if (i==0) {
      mat[0][eqEr][varEr]   = LS*dataCenter.dgravity[i];
      mat[2][eqEt][varEr]   = dataCenter.dgravity[i];
    } else {
      mat[0][eqEr][varEr]   = dataCenter.dgravity[i]+(LS-1.0)*dataCenter.gravity[i]/dataCenter.grid[i];
      mat[2][eqEt][varEr]   = dataCenter.gravity[i]/dataCenter.grid[i];
    }
    
    // Equation on Phi:
    mat[1][eqPhi][varPhi]   = 2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqPhi][varPhi]   = LS;
    mat[0][eqPhi][varPhiP]  = -1.0;
    
    // Equation on PhiP:
    mat[1][eqPhiP][varPhiP] = 2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqPhiP][varPhiP] = LS+1.0;
    mat[0][eqPhiP][varPhi]  = -LL;
    mat[0][eqPhiP][vardP_P] = -4.0*Math.PI*dataCenter.grid[i]*dataCenter.grid[i]*dataCenter.rho[i]/dataCenter.Gamma1[i];
    mat[0][eqPhiP][varEr]   =  4.0*Math.PI*dataCenter.grid[i]*dataCenter.drho[i];
    
    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixBLagrange(int l, int i) {
    double[][][] mat = new double[3][nvar][nvar];  // I'll assume everything is initialised to zero

    // Equation on Er:
    mat[0][eqEr][varEr] = 1.0;

    // Equation on Et:
    mat[2][eqEt][varEt] = 1.0;
    
    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert interface conditions in the Lagrangian case:
  ///////////////////////////////////////////////////////////////////////////////
  // NOTE: the interface is between points i and i+1
  ///////////////////////////////////////////////////////////////////////////////
  private void insertInterfaceConditionsLagrange(double[][] result, int offset, int ldmat,
                                             int kl, int ku, int ngrid, int i) {
    int ibis, jbis;
    // Interface condition on eqPhi (shifted):
    ibis = nvar*(i+1) + eqPhi;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varPhi;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
    jbis = nvar*i + varPhi;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;

    // Interface condition on eqPhiP (normal):
    ibis = nvar*i+eqPhiP;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varPhiP;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = nvar*(i+1) + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] = 4.0*Math.PI*dataCenter.rho[i+1];
    jbis = nvar*i + varPhiP;
    result[0][offset+ibis-jbis+ldmat*jbis] =-1.0;
    jbis = nvar*i + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] =-4.0*Math.PI*dataCenter.rho[i];
    
    // Interface condition on eqdP_P (shifted):
    ibis = nvar*(i+1) + eqdP_P;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + vardP_P;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
    jbis = nvar*i + vardP_P;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;

    // Interface condition on eqEr (normal):
    ibis = nvar*i + eqEr;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
    jbis = nvar*i + varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert boundary conditions in the Lagrangian case:
  ///////////////////////////////////////////////////////////////////////////////
  private void insertBoundaryConditionsLagrange(double[][] result, int offset, int ldmat,
                                                int kl, int ku, int ngrid, double LS) {
    int ibis, jbis;

    // Boundary condition on eqPhi (shifted):
    ibis = eqPhi;
    for(jbis=0; jbis<=ibis+ku; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = varPhiP;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;
    jbis = varPhi;
    result[0][offset+ibis-jbis+ldmat*jbis] = LS;

    // Boundary condition on eqPhiP (normal):
    ibis = nvar*(ngrid-1)+eqPhiP;
    for(jbis=ibis-kl; jbis<nvar*ngrid; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(ngrid-1)+varPhiP;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = nvar*(ngrid-1)+varPhi;
    result[0][offset+ibis-jbis+ldmat*jbis] = LS+1.0; // for scaled variables, otherwise /grid[ngrid-1];
    jbis = nvar*(ngrid-1)+varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] = 4.0*Math.PI*dataCenter.rho[ngrid-1];
    
    // Boundary condition on eqdP_P (shifted):
    ibis = eqdP_P;
    for(jbis=0; jbis<=ibis+ku; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = varEr;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = varEt;
    result[0][offset+ibis-jbis+ldmat*jbis] = -LS;

    // Boundary condition on eqEr (normal):
    ibis = nvar*(ngrid-1)+eqEr;
    for(jbis=ibis-kl; jbis<nvar*ngrid; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(ngrid-1)+vardP_P;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This extracts a standard set of pulsation variables from the pulsation
  // calculations.
  ///////////////////////////////////////////////////////////////////////////////
  private void extractVariablesLagrange(Kernel aKernel, double[] v, int offset, double[] rlm,
                                        double[] rl, int ngrid) {
    double cnst = 1.0/(v[offset+nvar*dataCenter.indSurface+varEr]*rlm[dataCenter.indSurface]);
    int i, ii;
    for(i=0; i<ngrid; i++) {
      ii = offset+nvar*i;
      aKernel.vDisp[i] = v[ii+varEr]*rlm[i]*cnst;
      aKernel.hDisp[i] = v[ii+varEt]*rlm[i]*cnst;
      aKernel.phi[i]   = v[ii+varPhi]*rl[i]*cnst;
      aKernel.phip[i]  = v[ii+varPhiP]*rlm[i]*cnst;
      aKernel.dP_P[i]  = v[ii+vardP_P]*rl[i]*cnst;
      aKernel.P[i]     = aKernel.dP_P[i]
                       * dataCenter.pressure[i]
                       + dataCenter.rho[i]
                       * dataCenter.gravity[i]
                       * aKernel.vDisp[i];
      aKernel.rho[i]   = aKernel.dP_P[i]
                       * dataCenter.rho[i]
                       / dataCenter.Gamma1[i]
                       - dataCenter.drho[i]
                       * aKernel.vDisp[i];
    }
    ii = offset;
    aKernel.vDisp_center = v[ii+varEr]*cnst;
    aKernel.hDisp_center = v[ii+varEt]*cnst;
    cnst/=Math.pow(dataCenter.grid[2],2.0);
    aKernel.vDisp_center_pp = cnst*(v[ii+varEr+2*nvar]-v[ii+varEr]);
    aKernel.hDisp_center_pp = cnst*(v[ii+varEt+2*nvar]-v[ii+varEt]);
  }

  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////  Lagrange -- radial  //////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////

  /******************************************************************************
  * This sets the order of the equations and variables using lagragian pressure
  * perturbations.
  ******************************************************************************/
  private void initEquationsLagrangeRadial() {
    eqdP_P_radial = 0;
    eqEr_radial = 1;
    
    vardP_P_radial = 0;
    varEr_radial = 1;
    
    kl0_radial = 0;
    ku0_radial = 0;
    nvar_radial = 2;
    shifted_radial = new int[]{1, 0};
  }
   
  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixALagrangeRadial(int i) {
    double[][][] mat = new double[3][nvar_radial][nvar_radial];  // I'll assume everything is initialised to zero

    // Equation on dP/P:
    mat[0][eqdP_P_radial][vardP_P_radial] = 1.0/dataCenter.Gamma1[i];
    mat[1][eqdP_P_radial][varEr_radial]   = 2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqdP_P_radial][varEr_radial]   = 3.0;

    // Equation on Er:
    mat[1][eqEr_radial][vardP_P_radial]   = 2.0*dataCenter.c2[i]/dataCenter.Gamma1[i];

    // special treatment in the center:
    if (i == 0) {
      mat[0][eqEr_radial][vardP_P_radial]   = -dataCenter.dgravity[0];
      mat[0][eqEr_radial][varEr_radial]     = -4.0*dataCenter.gravity[0];
    } else {
      mat[0][eqEr_radial][vardP_P_radial]   = -dataCenter.gravity[i]/dataCenter.grid[i];
      mat[0][eqEr_radial][varEr_radial]     = -4.0*dataCenter.gravity[i]/dataCenter.grid[i];
    }
    
    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixBLagrangeRadial(int i) {
    double[][][] mat = new double[3][nvar_radial][nvar_radial];

    // Equation on Er:
    mat[0][eqEr_radial][varEr_radial] = 1.0;

    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert interface conditions in the radial Lagrangian case:
  ///////////////////////////////////////////////////////////////////////////////
  // NOTE: the interface is between points i and i+1
  ///////////////////////////////////////////////////////////////////////////////
  private void insertInterfaceConditionsLagrangeRadial(double[][] result, int offset, int ldmat,
                                             int kl, int ku, int ngrid, int i) {
    int ibis, jbis;
    // Interface condition on eqdP_P (shifted):
    ibis = nvar_radial*(i+1) + eqdP_P_radial;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar_radial*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar_radial*(i+1) + vardP_P_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
    jbis = nvar_radial*i + vardP_P_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;

    // Interface condition on eqEr (normal):
    ibis = nvar_radial*i + eqEr_radial;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar_radial*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar_radial*(i+1) + varEr_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
    jbis = nvar_radial*i + varEr_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert boundary conditions in the radial Lagrangian case:
  ///////////////////////////////////////////////////////////////////////////////
  private void insertBoundaryConditionsLagrangeRadial(double[][] result, int offset, int ldmat,
                                                      int kl, int ku, int ngrid) {
    int ibis, jbis;

    // Boundary condition on eqdP_P (shifted):
    ibis = eqdP_P_radial;
    for(jbis=0; jbis<=ibis+ku; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = vardP_P_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0/dataCenter.Gamma1[1];
    jbis = varEr_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = 3.0;

    // Boundary condition on eqEr (normal):
    ibis = nvar_radial*(ngrid-1)+eqEr_radial;
    for(jbis=ibis-kl; jbis<nvar_radial*ngrid; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar_radial*(ngrid-1)+vardP_P_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This extracts a standard set of pulsation variables from the pulsation
  // calculations.
  ///////////////////////////////////////////////////////////////////////////////
  private void extractVariablesLagrangeRadial(Kernel aKernel, double[] v, int offset, double[] rlm,
                                              double[] rl, int ngrid) {
    double cnst = 1.0/(v[offset+nvar_radial*dataCenter.indSurface+varEr_radial]*dataCenter.grid[dataCenter.indSurface]);
    int i, ii;
    for(i=0; i<ngrid; i++) {
      ii = offset+nvar_radial*i;
      aKernel.vDisp[i] = v[ii+varEr_radial]*dataCenter.grid[i]*cnst;
      aKernel.hDisp[i] = 0.0;
      aKernel.phip[i]  = -4.0*Math.PI*dataCenter.rho[i]
                       * aKernel.vDisp[i];
      aKernel.dP_P[i]  = v[ii+vardP_P_radial]*cnst;
      aKernel.P[i]     = aKernel.dP_P[i]
                       * dataCenter.pressure[i]
                       + dataCenter.rho[i]
                       * dataCenter.gravity[i]
                       * aKernel.vDisp[i];
      aKernel.rho[i]   = aKernel.dP_P[i]
                       * dataCenter.rho[i]
                       / dataCenter.Gamma1[i]
                       - dataCenter.drho[i]
                       * aKernel.vDisp[i];
    }
    ii = offset;
    aKernel.vDisp_center = v[ii+varEr_radial]*cnst;
    aKernel.hDisp_center = 0.0;
    cnst/=Math.pow(dataCenter.grid[2],2.0);
    aKernel.vDisp_center_pp = cnst*(v[ii+varEr_radial+2*nvar_radial]-v[ii+varEr_radial]);
    aKernel.hDisp_center_pp = 0.0;
  } 

  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////  JCD -- non-radial  ///////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////

  /******************************************************************************
  * This sets the order of the equations and variables using JCD type equations.
  ******************************************************************************/
  private void initEquationsJCD() {
      eqY1 = 0;
      eqY2 = 1;
      eqY3 = 2;
      eqY4 = 3;

      varY1 = 0;
      varY2 = 1;
      varY3 = 2;
      varY4 = 3;

      kl0 = 1;
      ku0 = 1;
      nvar = 4;
      shifted = new int[]{1, 1, 0, 0}; // equations shifted due to boundary conditions
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixAJCD(double LS, double LL, int i) {
    double[][][] mat = new double[3][nvar][nvar];  // I'll assume everything is initialised to zero

    // Equation on Y1:
    mat[1][eqY1][varY1] = -2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqY1][varY1] = dataCenter.Vg[i]-LS-1.0;
    mat[0][eqY1][varY2] = 1.0;
    mat[0][eqY1][varY3] = -dataCenter.Vg[i];

    // Equation on Y2:
    mat[0][eqY2][varY1] = -LL*dataCenter.eta[i]*dataCenter.A[i];
    mat[0][eqY2][varY3] = LL*dataCenter.eta[i]*dataCenter.A[i];
    
    // Equation on Y3:
    mat[1][eqY3][varY3] = -2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqY3][varY3] = 2.0-LS;
    mat[0][eqY3][varY4] = 1.0;

    // Equation on Y4:
    mat[1][eqY4][varY4] = -2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqY4][varY1] = -dataCenter.A[i]*dataCenter.U[i];
    mat[0][eqY4][varY3] = LL+dataCenter.U[i]*(dataCenter.A[i]-2.0);
    mat[0][eqY4][varY4] = 3.0-LS-2.0*dataCenter.U[i];
    
    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixBJCD(double LS, double LL, int i) {
    double[][][] mat = new double[3][nvar][nvar];  // I'll assume everything is initialised to zero

    // Equation on Y1:
    mat[0][eqY1][varY2] = dataCenter.Vg[i]/(LL*dataCenter.eta[i]);

    // Equation on Y2:
    mat[1][eqY2][varY2] = 2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqY2][varY1] = -LL;
    mat[0][eqY2][varY2] = -dataCenter.A[i]+LS;

    // Equation on Y4:
    mat[0][eqY4][varY2] = dataCenter.U[i]*dataCenter.Vg[i]
                        / (LL*dataCenter.eta[i]);
    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert interface conditions when using JCD-type equations:
  ///////////////////////////////////////////////////////////////////////////////
  // NOTE: the double point is located on points i and i+1.
  ///////////////////////////////////////////////////////////////////////////////
  private void insertInterfaceConditionsJCD(double[][] result, int offset, int ldmat,
                                           int kl, int ku, int ngrid, double LL,
                                           int i) {
    int ibis, jbis;

    // Interface condition on eqY1 (shifted):
    ibis = nvar*(i+1) + eqY1;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varY1;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = nvar*i + varY1;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;

    // Interface condition on eqY2 (shifted):
    ibis = nvar*(i+1) + eqY2;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varY1;
    result[0][offset+ibis-jbis+ldmat*jbis] = -dataCenter.eta[i+1]*dataCenter.U[i+1];
    jbis = nvar*(i+1) + varY3;
    result[0][offset+ibis-jbis+ldmat*jbis] =  dataCenter.eta[i+1]*dataCenter.U[i+1];
    jbis = nvar*(i+1) + varY2;
    result[1][offset+ibis-jbis+ldmat*jbis] = -dataCenter.U[i+1]/LL;
    jbis = nvar*i + varY1;
    result[0][offset+ibis-jbis+ldmat*jbis] =  dataCenter.eta[i]*dataCenter.U[i];
    jbis = nvar*i + varY3;
    result[0][offset+ibis-jbis+ldmat*jbis] = -dataCenter.eta[i]*dataCenter.U[i];
    jbis = nvar*i + varY2;
    result[1][offset+ibis-jbis+ldmat*jbis] =  dataCenter.U[i]/LL;

    // Interface condition on eqY3 (normal):
    ibis = nvar*i + eqY3;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varY3;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = nvar*i + varY3;
    result[0][offset+ibis-jbis+ldmat*jbis] =-1.0;

    // Interface condition on eqY4 (normal):
    ibis = nvar*i + eqY4;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(i+1) + varY1;
    result[0][offset+ibis-jbis+ldmat*jbis] =  dataCenter.U[i+1];
    jbis = nvar*(i+1) + varY3;
    result[0][offset+ibis-jbis+ldmat*jbis] = -dataCenter.U[i+1];
    jbis = nvar*(i+1) + varY4;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;
    jbis = nvar*i + varY1;
    result[0][offset+ibis-jbis+ldmat*jbis] = -dataCenter.U[i];
    jbis = nvar*i + varY3;
    result[0][offset+ibis-jbis+ldmat*jbis] =  dataCenter.U[i];
    jbis = nvar*i + varY4;
    result[0][offset+ibis-jbis+ldmat*jbis] =  1.0;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert boundary conditions when using JCD-type equations:
  ///////////////////////////////////////////////////////////////////////////////
  private void insertBoundaryConditionsJCD(double[][] result, int offset, int ldmat,
                                           int kl, int ku, int ngrid, double LS) {
    int ibis, jbis;

    // Boundary condition on eqY1 (shifted):
    ibis = eqY1;
    for(jbis=0; jbis<=ibis+ku; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = varY1;
    result[0][offset+ibis-jbis+ldmat*jbis] = LS+1.0;
    jbis = varY2;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;

    // Boundary condition on eqY2 (shifted):
    ibis = eqY2;
    for(jbis=0; jbis<=ibis+ku; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = varY3;
    result[0][offset+ibis-jbis+ldmat*jbis] = LS-2.0;
    jbis = varY4;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;

    // Boundary condition on eqY3 (normal):
    ibis = nvar*(ngrid-1)+eqY3;
    for(jbis=ibis-kl; jbis<nvar*ngrid; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(ngrid-1)+varY1;
    result[0][offset+ibis-jbis+ldmat*jbis] = LS*(LS+1)*dataCenter.eta[ngrid-1];
    jbis = nvar*(ngrid-1)+varY2;
    result[1][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = nvar*(ngrid-1)+varY3;
    result[0][offset+ibis-jbis+ldmat*jbis] =-LS*(LS+1)*dataCenter.eta[ngrid-1];

    // Boundary condition on eqY4 (normal):
    ibis = nvar*(ngrid-1)+eqY4;
    for(jbis=ibis-kl; jbis<nvar*ngrid; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar*(ngrid-1)+varY1;
    result[0][offset+ibis-jbis+ldmat*jbis] = dataCenter.U[ngrid-1];
    jbis = nvar*(ngrid-1)+varY3;
    result[0][offset+ibis-jbis+ldmat*jbis] = -LS-dataCenter.U[ngrid-1];
    jbis = nvar*(ngrid-1)+varY4;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This extracts a standard set of pulsation variables from the pulsation
  // calculations.
  ///////////////////////////////////////////////////////////////////////////////
  private void extractVariablesJCD(Kernel aKernel, double[] v, int offset, double[] rlm,
                                   double[] rl, int ngrid) {
    double cnst = 1.0/(v[offset+nvar*dataCenter.indSurface+varY1]*rlm[dataCenter.indSurface]);
    double L2 = (double)aKernel.l*(aKernel.l+1);
    double sigma2 = aKernel.FreqDimless*aKernel.FreqDimless;
    int i, ii;
    for(i=0; i<ngrid; i++) {
      ii = offset+nvar*i;
      aKernel.vDisp[i] = v[ii+varY1]*rlm[i]*cnst;
      aKernel.hDisp[i] = v[ii+varY2]*rlm[i]*cnst/L2;
      aKernel.phi[i]   = -v[ii+varY3]*rlm[i]*cnst*dataCenter.gravity[i];
      aKernel.phip[i]  =-(v[ii+varY4]*dataCenter.gravity[i]/dataCenter.grid[i]
                       +(dataCenter.gravity[i]/dataCenter.grid[i]
                       + dataCenter.dgravity[i])*v[ii+varY3])*rlm[i]*cnst;
      aKernel.P[i]     =(aKernel.hDisp[i]*dataCenter.grid[i]*sigma2
                       - aKernel.phi[i])*dataCenter.rho[i];
      aKernel.rho[i]   = aKernel.P[i]/dataCenter.c2[i]
                       + aKernel.vDisp[i]*dataCenter.A[i]
                       * dataCenter.rho[i]/dataCenter.grid[i];
      aKernel.dP_P[i]  =(aKernel.P[i]
                       - dataCenter.rho[i]
                       * dataCenter.gravity[i]
                       * aKernel.vDisp[i])
                       / dataCenter.pressure[i];
    }
    ii = offset;
    aKernel.phip[0]  =-(v[ii+varY4]*dataCenter.dgravity[0]
                     + 2.0*dataCenter.dgravity[0]*v[ii+varY3])
                     * rlm[0]*cnst;
    aKernel.rho[0]   = aKernel.P[0]/dataCenter.c2[0];

    ii = offset;
    aKernel.vDisp_center = v[ii+varY1]*cnst;
    aKernel.hDisp_center = v[ii+varY2]*cnst/L2;
    cnst/=Math.pow(dataCenter.grid[2],2.0);
    aKernel.vDisp_center_pp = cnst*(v[ii+varY1+2*nvar]-v[ii+varY1]);
    aKernel.hDisp_center_pp = cnst*(v[ii+varY2+2*nvar]-v[ii+varY2])/L2;
  }

  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////  JCD  --  radial  /////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////

  /******************************************************************************
  * This sets the order of the equations and variables using eulerian pressure
  * and density perturbations.
  ******************************************************************************/
  private void initEquationsJCDRadial() {
    eqY1_radial = 0;
    eqY2_radial = 1;
    
    varY1_radial = 0;
    varY2_radial = 1;
    
    kl0_radial = 0;
    ku0_radial = 0;
    nvar_radial = 2;
    shifted_radial = new int[]{1, 0};
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixAJCDRadial(int i) {
    double[][][] mat = new double[3][nvar_radial][nvar_radial];  // I'll assume everything is initialised to zero

    // Equation on Y1:
    mat[1][eqY1_radial][varY1_radial] = -2.0*dataCenter.grid[i]*dataCenter.grid[i];
    mat[0][eqY1_radial][varY1_radial] = dataCenter.Vg[i]-3.0;

    // Equation on Y2:
    mat[0][eqY2_radial][varY1_radial] = (dataCenter.U[i]-dataCenter.A[i])*dataCenter.eta[i];
    
    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Create matrix on level i
  ///////////////////////////////////////////////////////////////////////////////
  private double[][][] createSingleLevelMatrixBJCDRadial(int i) {
    double[][][] mat = new double[3][nvar_radial][nvar_radial];

    // Equation on Y1:
    mat[0][eqY1_radial][varY2_radial] = 1.0/dataCenter.c2[i];

    // Equation on Y2:
    mat[1][eqY2_radial][varY2_radial] =  2.0;
    mat[0][eqY2_radial][varY1_radial] = -1.0;

    // special treatment in the center:
    if (i==0) {
      mat[0][eqY2_radial][varY2_radial] = dataCenter.dgravity[0]/dataCenter.c2[0] 
                                        + dataCenter.ddrho/dataCenter.rho[0];
    } else {
      mat[0][eqY2_radial][varY2_radial] = -dataCenter.A[i]/(dataCenter.grid[i]*dataCenter.grid[i]);
    }

    return mat;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert interface conditions when using radial JCD-type equations:
  ///////////////////////////////////////////////////////////////////////////////
  // NOTE: the double point is located on points i and i+1.
  ///////////////////////////////////////////////////////////////////////////////
  private void insertInterfaceConditionsJCDRadial(double[][] result, int offset, int ldmat,
                                           int kl, int ku, int ngrid, int i) {
    int ibis, jbis;

    // Interface condition on eqY1 (shifted):
    ibis = nvar_radial*(i+1) + eqY1_radial;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar_radial*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar_radial*(i+1) + varY1_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = 1.0;
    jbis = nvar_radial*i + varY1_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = -1.0;

    // Interface condition on eqY2 (normal):
    ibis = nvar_radial*i + eqY2_radial;
    for(jbis=Math.max(ibis-kl,0); jbis<=Math.min(ibis+ku,nvar_radial*ngrid-1); jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar_radial*(i+1) + varY1_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] =  Math.pow(dataCenter.grid[i+1],2.0)*dataCenter.eta[i+1]*dataCenter.U[i+1];
    jbis = nvar_radial*(i+1) + varY2_radial;
    result[1][offset+ibis-jbis+ldmat*jbis] =  dataCenter.U[i+1];
    jbis = nvar_radial*i + varY1_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = -Math.pow(dataCenter.grid[i],2.0)*dataCenter.eta[i]*dataCenter.U[i];
    jbis = nvar_radial*i + varY2_radial;
    result[1][offset+ibis-jbis+ldmat*jbis] = -dataCenter.U[i];
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Insert boundary conditions in the radial JCD case:
  ///////////////////////////////////////////////////////////////////////////////
  private void insertBoundaryConditionsJCDRadial(double[][] result, int offset, int ldmat,
                                                   int kl, int ku, int ngrid) {
    int ibis, jbis;

    // Boundary condition on eqY1 (shifted):
    ibis = eqY1_radial;
    for(jbis=0; jbis<=ibis+ku; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = varY1_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] =  3.0;
    jbis = varY2_radial;
    result[1][offset+ibis-jbis+ldmat*jbis] = -1.0/dataCenter.c2[0];

    // Boundary condition on eqY2 (normal):
    ibis = nvar_radial*(ngrid-1)+eqY2_radial;
    for(jbis=ibis-kl; jbis<nvar_radial*ngrid; jbis++) {
      result[0][offset+ibis-jbis+ldmat*jbis]=0.0;
      result[1][offset+ibis-jbis+ldmat*jbis]=0.0;
    }
    jbis = nvar_radial*(ngrid-1)+varY1_radial;
    result[0][offset+ibis-jbis+ldmat*jbis] = dataCenter.grid[ngrid-1]*dataCenter.gravity[ngrid-1];
    jbis = nvar_radial*(ngrid-1)+varY2_radial;
    result[1][offset+ibis-jbis+ldmat*jbis] = 1.0;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This extracts a standard set of pulsation variables from the pulsation
  // calculations.
  ///////////////////////////////////////////////////////////////////////////////
  private void extractVariablesJCDRadial(Kernel aKernel, double[] v, int offset, double[] rlm,
                                         double[] rl, int ngrid) {
    double cnst = 1.0/(v[offset+nvar_radial*dataCenter.indSurface+varY1_radial]*dataCenter.grid[dataCenter.indSurface]);
    double sigma2 = aKernel.FreqDimless*aKernel.FreqDimless;
    int i, ii;
    for(i=0; i<ngrid; i++) {
      ii = offset+nvar_radial*i;
      aKernel.vDisp[i] = v[ii+varY1_radial]*dataCenter.grid[i]*cnst;
      aKernel.hDisp[i] = 0.0;
      aKernel.P[i]     = v[ii+varY2_radial]*dataCenter.rho[i]*sigma2*cnst;
      aKernel.phip[i]  = -4.0*Math.PI*dataCenter.rho[i]
                       * aKernel.vDisp[i];
      aKernel.dP_P[i]  =(aKernel.P[i]
                       - dataCenter.rho[i]
                       * dataCenter.gravity[i]
                       * aKernel.vDisp[i])
                       / dataCenter.pressure[i];
      aKernel.rho[i]   = aKernel.dP_P[i]
                       * dataCenter.rho[i]
                       / dataCenter.Gamma1[i]
                       - dataCenter.drho[i]
                       * aKernel.vDisp[i];
    }
    ii = offset;
    aKernel.vDisp_center = v[ii+varY1_radial]*cnst;
    aKernel.hDisp_center = 0.0;
    cnst/=Math.pow(dataCenter.grid[2],2.0);
    aKernel.vDisp_center_pp = cnst*(v[ii+varY1_radial+2*nvar_radial]-v[ii+varY1_radial]);
    aKernel.hDisp_center_pp = 0.0;
  } 

  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  /////////////////  Other eigenfunction extraction procedures  /////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////

  ///////////////////////////////////////////////////////////////////////////////
  // This finds phi, phip, P, rho, dP_P from the Lagrangian displacement
  ///////////////////////////////////////////////////////////////////////////////
  // REQUIREMENTS: vDisp, hDisp (from the eigenmode)
  ///////////////////////////////////////////////////////////////////////////////
  public void findPhi() {
    int i, j, nn, lprev=-1;
    int ngrid = dataCenter.grid.length;
    Kernel aKernel;
    double L,L2,a;
    double[] ksi   = new double[ngrid];

    int order = 1;
    double[][] dermat = Numerics.initDeriveFDBand(dataCenter.grid,order,1);

    // variables used to solve Poisson's equation:
    int orderFD = 2, klu = 2*orderFD, ldmat = 3*klu+1, offset = 2*klu;
    double[][] dermatBis = Numerics.initDeriveFDBandRoots(dataCenter.grid,orderFD,1);
    double[] mat = new double[ldmat*ngrid*2];
    double[] v = new double[2*ngrid];
    int[] ipiv = new int[2*ngrid];
    org.netlib.util.intW info = new org.netlib.util.intW(2);

    for(nn=0; nn<dataCenter.kernels.size(); nn++){
      if (dataCenter.kernels.get(nn).phi != null) continue;

      aKernel = dataCenter.kernels.get(nn); // just to simplify the notation
      L  = (double)aKernel.l;
      L2 = L*(L+1.0);

      // find ksi (= div displ):
      Dgbmv.dgbmv("N",ngrid,ngrid,order,order,1.0,dermat[1],0,1+2*order,aKernel.vDisp,0,1,0.0,ksi,0,1);
      for(i=1; i<ngrid; i++) ksi[i]+=(2.0*aKernel.vDisp[i]-L2*aKernel.hDisp[i])/dataCenter.grid[i];
      ksi[0] = (aKernel.l==0)?3.0*ksi[0]:0.0;

      // find dP_P, eulerian density and pressure perturbations:
      for(i=0;i<ngrid;i++) {
        aKernel.dP_P[i] =-dataCenter.Gamma1[i]*ksi[i];
        aKernel.rho[i]  =-dataCenter.drho[i]*aKernel.vDisp[i]-dataCenter.rho[i]*ksi[i];
        aKernel.P[i]    = dataCenter.rho[i]*dataCenter.gravity[i]*aKernel.vDisp[i]
                        - dataCenter.Gamma1[i]*dataCenter.pressure[i]*ksi[i];
      }

      // find phi and phip by solving Poisson's equation (more stable numerically
      // than applying a numerical anti-derivative to the analytical formula)
      if (lprev != aKernel.l) {
        for(i=0;i<2*ngrid*ldmat;i++) mat[i] = 0.0;
        for(i=0;i<ngrid-1;i++) {
          for(j=Math.max(0,i-orderFD+1);j<=Math.min(ngrid-1,i+orderFD);j++) {
            // The indices are highly complicated, so don't modify them.
            // They correspond to a band matrix stored as a 1D array (ouch!!).
            mat[offset+2*(i-j)+1+ldmat*2*j]     = dermatBis[1][orderFD+i-j+2*orderFD*j]*dataCenter.grid[j]*dataCenter.grid[j];
            mat[offset+2*(i-j)+ldmat*(2*j+1)]   = -L2*dermatBis[0][orderFD+i-j+2*orderFD*j];
            mat[offset+2*(i-j)+2+ldmat*2*j]     = -dermatBis[0][orderFD+i-j+2*orderFD*j];
            mat[offset+2*(i-j)+1+ldmat*(2*j+1)] = dermatBis[1][orderFD+i-j+2*orderFD*j];
          }
        }
        // Boundary conditions:
        if (aKernel.l==0) {
          mat[offset] = 1.0;
        } else {
          mat[offset-1+ldmat] = 1.0;
        }
        mat[offset+1+ldmat*(2*ngrid-2)] = dataCenter.grid[ngrid-1];
        mat[offset+ldmat*(2*ngrid-1)] = L+1.0;
        Dgbtrf.dgbtrf(2*ngrid,2*ngrid,klu,klu,mat,0,ldmat,ipiv,0,info);
        if (info.val != 0) {
          System.out.println("ERROR: in dgbtrf in findPhi(), info = " + info.val);
          System.out.println("Modes not calculated properly - this will cause many more errors");
          return;
        }
        lprev = aKernel.l;
      }
      for(i=0;i<2*ngrid;i++) v[i] = 0.0;
      for(i=0;i<ngrid-1;i++) {
        for(j=Math.max(0,i-orderFD+1);j<=Math.min(ngrid-1,i+orderFD);j++) {
          v[2*i+1]+=4.0*Math.PI*dermatBis[0][orderFD+i-j+2*orderFD*j]*Math.pow(dataCenter.grid[j],2.0)*aKernel.rho[j];
        }
      }
      v[0] = 0.0;
      v[2*ngrid-1] = -4.0*Math.PI*dataCenter.rho[ngrid-1]*dataCenter.grid[ngrid-1]*aKernel.vDisp[ngrid-1];
      Dgbtrs.dgbtrs("N",2*ngrid,klu,klu,1,mat,0,ldmat,ipiv,0,v,0,2*ngrid,info);
      if (info.val != 0) {
        System.out.println("WARNING: in dgbtrs in findPhi(), info = " + info.val);
        System.out.println("Modes not calculated properly - this will cause many more errors");
      }
      aKernel.phi = new double[ngrid]; // don't forget this...
      for(i=0;i<ngrid;i++) {aKernel.phi[i] = v[2*i+1]; aKernel.phip[i] = v[2*i];}
    }
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This finds the variables P, rho and dP_P
  ///////////////////////////////////////////////////////////////////////////////
  // REQUIREMENTS: vDisp, hDisp (from the eigenmode)
  //               gravity, rho, pressure, Gamma1 (from the model)
  // NOTE: this only applies for radial (l = 0) modes
  ///////////////////////////////////////////////////////////////////////////////
  public void findVariablesRadial(Kernel aKernel){
    int ngrid = dataCenter.grid.length, i;
    for(i=0; i < ngrid; i++) {
      aKernel.hDisp[i] = 0.0;
      aKernel.phip[i]  =-4.0*Math.PI*dataCenter.rho[i]*aKernel.vDisp[i];
      aKernel.dP_P[i]  =(aKernel.P[i]
                       - dataCenter.rho[i]
                       * dataCenter.gravity[i]
                       * aKernel.vDisp[i])
                       / dataCenter.pressure[i];
      aKernel.rho[i]   = aKernel.dP_P[i]
                       * dataCenter.rho[i]
                       / dataCenter.Gamma1[i]
                       - dataCenter.drho[i]
                       * aKernel.vDisp[i];
    }
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This finds the variables P, rho and dP_P
  ///////////////////////////////////////////////////////////////////////////////
  // REQUIREMENTS: vDisp, hDisp, phi, FreqDimless (from the eigenmode)
  //               grid, c2, A, gravity, rho, pressure (from the model)
  // NOTE: this only applies for non-radial (l > 0) modes
  ///////////////////////////////////////////////////////////////////////////////
  public void findVariablesNonRadial(Kernel aKernel){
    int ngrid = dataCenter.grid.length, i;
    double Freq2 = aKernel.FreqDimless*aKernel.FreqDimless;
    for(i=0; i < ngrid; i++) {
      aKernel.P[i]    = dataCenter.rho[i]
                      *(Freq2*dataCenter.grid[i]*aKernel.hDisp[i]
                      - aKernel.phi[i]);
      aKernel.rho[i]  = aKernel.P[i]/dataCenter.c2[i]
                      + aKernel.vDisp[i]*dataCenter.A[i]
                      * dataCenter.rho[i]/dataCenter.grid[i];
      aKernel.dP_P[i] =(aKernel.P[i]
                      - dataCenter.rho[i]
                      * dataCenter.gravity[i]
                      * aKernel.vDisp[i])
                      / dataCenter.pressure[i];
    }
    aKernel.rho[0]    = aKernel.P[0]/dataCenter.c2[0];
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This finds IK, Enl, and Y1 and Y2 from Takata (2006).
  ///////////////////////////////////////////////////////////////////////////////
  // REQUIREMENTS: l, vDisp, hDisp, P, phi, phip (from the eigenmode)
  //               grid, diffrho_r2, gravity, rho (from the model)
  // NOTE: this is where to add new quantities which need to be calculated for
  //       all of the theoretical modes.
  ///////////////////////////////////////////////////////////////////////////////
  public void findOtherVariables() {
    Kernel aKernel;
    int ngrid = dataCenter.grid.length, i, nn;
    double[] w = Numerics.findIntegrationWeights(dataCenter.grid);
    double L2;
    
    for(nn=0; nn<dataCenter.kernels.size(); nn++){
      aKernel = dataCenter.kernels.get(nn);

      // test to see if these have alread by calculated
      if (!Double.isNaN(aKernel.IK)) continue;

      L2 = (double)aKernel.l*(1.0+(double)aKernel.l);
      aKernel.IK = 0.0;
      for(i=1; i < ngrid; i++) {
        aKernel.IK+=w[i]*dataCenter.rho[i]*dataCenter.grid[i]*dataCenter.grid[i]
                      * ( aKernel.vDisp[i]*aKernel.vDisp[i]
                      +L2*aKernel.hDisp[i]*aKernel.hDisp[i]);
      }

      aKernel.Enl = aKernel.IK /  (aKernel.vDisp[dataCenter.indSurface]*aKernel.vDisp[dataCenter.indSurface]
                               +L2*aKernel.hDisp[dataCenter.indSurface]*aKernel.hDisp[dataCenter.indSurface]);

      aKernel.Y1 = new double[ngrid];
      aKernel.Y2 = new double[ngrid];
      if (aKernel.l == 1) {
        for(i=1; i < ngrid; i++) {
          aKernel.Y1[i] = dataCenter.diffrho_r2[i]
               * dataCenter.grid[i]*aKernel.vDisp[i]
               +(aKernel.phi[i]/dataCenter.grid[i]
               - aKernel.phip[i])/(3.0*dataCenter.gravity[i]);
          aKernel.Y2[i] = dataCenter.diffrho_r2[i]
               * dataCenter.grid[i]*aKernel.P[i]
               /(dataCenter.rho[i]*dataCenter.gravity[i])
               +(aKernel.phi[i]/dataCenter.grid[i]
               - aKernel.phip[i])/(3.0*dataCenter.gravity[i]);
        }
        aKernel.Y1[0] = 0.0;
        aKernel.Y2[0] = 0.0;
      }
    }
  }

  ///////////////////////////////////////////////////////////////////////////////
  // This finds the radial orders of the modes, using the following approaches:
  //   * for l=1 modes, assign radial order using the Takata (2006) approach
  //     (using the variables Y1 and Y2)
  //   * for other modes, assign the radial order using the ESO
  //     (Eckart (1960) - Scuflaire (1974) - Osaki (1975)) approach
  ///////////////////////////////////////////////////////////////////////////////
  // REQUIREMENTS: l, vDisp, hDisp, P, phi, phip (from the eigenmode)
  //               grid, diffrho_r2, gravity, rho (from the model)
  // NOTE: this is where to add new quantities which need to be calculated for
  //       all of the theoretical modes.
  ///////////////////////////////////////////////////////////////////////////////
  public void findRadialOrders() {
    Kernel aKernel;
    int ngrid = dataCenter.grid.length, i, nn;
    
    for(nn=0; nn<dataCenter.kernels.size(); nn++){
      aKernel = dataCenter.kernels.get(nn);
      if (aKernel.n != Integer.MIN_VALUE) continue;
      if (aKernel.equations == FromFile) continue;  // this can be changed if need be

      // for dipole modes, assign radial order using the Takata (2006) approach:
      // Note: the variable Y2 (rather than dP_P) works a lot better for red giants.
      if (aKernel.l == 1) {
        aKernel.n = 0;
        for(i=2; i<ngrid-1; i++) {
          // find zeros of the Y1 function
          if (aKernel.Y1[i]*aKernel.Y1[i-1]< 0) {
            // find in which direction the phase curve crosses the Y2 axis:
            if (aKernel.Y1[i]*aKernel.Y2[i] < 0) {
              aKernel.n++;
            } else {
              aKernel.n--;
            }
          }
        }
        if (aKernel.n >= 0) aKernel.n++;
        continue;
      }
 
      // for other modes, assign radial order using the
      // ESO (Eckart (1960) - Scuflaire (1974) - Osaki (1975)) approach:
      if (aKernel.l == 0) {
        aKernel.n = 1;
      } else {
        aKernel.n = 0;
      }
      for(i=2; i<ngrid; i++) { // exclude center point as vDisp[0] = 0
        // find zero of Lagrangian pressure perturbation:
        if (aKernel.vDisp[i]*aKernel.vDisp[i-1] < 0) {
          // find in which direction the phase curve crosses the P axis:
          if (aKernel.vDisp[i]*aKernel.P[i] < 0) {
            aKernel.n++;
          } else {
            aKernel.n--;
          }
        }
      }
    }
  }

  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////  Kernels and variational principle  ///////////////////
  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////

  /******************************************************************************
  * This calculates the rotation kernels from the scaled eigenfunctions.
  ******************************************************************************/
  public void buildRotaKernels() {
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);
    double[] intMat = null, intMatR2 = null;
    int ngrid = dataCenter.grid.length;
    int nfunc = dataCenter.kernels.size();
    int i, fnum, degree, nprev = 0, n;
    double L2, total, totalR2, aux;

    for (fnum = 0; fnum < nfunc; fnum++){
      // avoid recalculating a rotation kernel:
      if (dataCenter.kernels.get(fnum).rotaKernel != null) continue;

      dataCenter.kernels.get(fnum).rotaKernel = new double[ngrid];
      degree = dataCenter.kernels.get(fnum).l;
      L2 = (double)degree * ((double)degree + 1.0);
      if (degree > 0) {
        n = 2*degree;
        if (degree == 1) n+=4;
        if (n != nprev) {
          intMat = Numerics.findPiecewiseIntegrationMatrix(dataCenter.grid,n);
          intMatR2 = Numerics.findPiecewiseIntegrationMatrix(dataCenter.grid,n-2);
        }
        nprev = n;
        dataCenter.kernels.get(fnum).rotaKernel[0] = 0.0;
        for (i=1; i<ngrid; i++) {
          dataCenter.kernels.get(fnum).rotaKernel[i] = dataCenter.rho[i]* 
                     (dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.kernels.get(fnum).vDisp[i]
            +(L2-1.0)*dataCenter.kernels.get(fnum).hDisp[i]*dataCenter.kernels.get(fnum).hDisp[i]
                 -2.0*dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.kernels.get(fnum).hDisp[i])
                     /dataCenter.kernels.get(fnum).IK;
        }

        // find anti-derivative of rotation kernel:
        dataCenter.kernels.get(fnum).intRotaKernelR2 =
            Numerics.applyPiecewiseIntegrationMatrix(intMatR2,
            dataCenter.kernels.get(fnum).rotaKernel,0.0);

        // find average radius from rotation kernel, divided by r^2:
        dataCenter.kernels.get(fnum).rRotaR2 = 0.0;
        totalR2 = 0.0; 
        for(i=1; i<ngrid; i++) {
          dataCenter.kernels.get(fnum).rRotaR2 += weights[i]*dataCenter.grid[i]
              *dataCenter.kernels.get(fnum).rotaKernel[i];
          totalR2 += weights[i]*dataCenter.kernels.get(fnum).rotaKernel[i];
        }
        dataCenter.kernels.get(fnum).rRotaR2 /= totalR2;

        // find correct rotation kernel by multiplying it by r^2:
        for (i=1; i<ngrid; i++) {
          dataCenter.kernels.get(fnum).rotaKernel[i]*=(dataCenter.grid[i]*dataCenter.grid[i]);
        }

        dataCenter.kernels.get(fnum).intRotaKernel =
            Numerics.applyPiecewiseIntegrationMatrix(intMat,
            dataCenter.kernels.get(fnum).rotaKernel,0.0);

        // find average radius from rotation kernel:
        dataCenter.kernels.get(fnum).rRota = 0.0;
        total = 0.0; 
        for(i=1; i<ngrid; i++) {
          dataCenter.kernels.get(fnum).rRota += weights[i]*dataCenter.grid[i]
              *dataCenter.kernels.get(fnum).rotaKernel[i];
          total += weights[i]*dataCenter.kernels.get(fnum).rotaKernel[i];
        }
        dataCenter.kernels.get(fnum).rRota /= total;

      } else {
        dataCenter.kernels.get(fnum).intRotaKernel = new double[ngrid];
        dataCenter.kernels.get(fnum).intRotaKernelR2 = new double[ngrid];
        for (i=1; i<ngrid; i++) {
          dataCenter.kernels.get(fnum).rotaKernel[i] = 0.0;
          dataCenter.kernels.get(fnum).intRotaKernel[i] = 0.0;
          dataCenter.kernels.get(fnum).intRotaKernelR2[i] = 0.0;
        }
        dataCenter.kernels.get(fnum).rRota = 0.0;
        dataCenter.kernels.get(fnum).rRotaR2 = 0.0;
        total = 0.0; 
        totalR2 = 0.0; 
        for(i=1; i<ngrid; i++) {
          aux = weights[i]*dataCenter.rho[i]
              * dataCenter.kernels.get(fnum).vDisp[i]
              * dataCenter.kernels.get(fnum).vDisp[i];
          dataCenter.kernels.get(fnum).rRotaR2 += aux*dataCenter.grid[i];
          totalR2 += aux;
          aux *= (dataCenter.grid[i]*dataCenter.grid[i]);
          dataCenter.kernels.get(fnum).rRota += aux*dataCenter.grid[i];
          total += aux;
        }
        dataCenter.kernels.get(fnum).rRota /= total;
        dataCenter.kernels.get(fnum).rRotaR2 /= totalR2;
      }
    }
  }

  /******************************************************************************
  * This calculates the different structural kernels from the scaled
  * eigenfunctions.
  ******************************************************************************/
  public void buildStructKernels() {
    int ngrid = dataCenter.grid.length;
    int nfunc = dataCenter.kernels.size();
    int i, j, fnum, degree;
    double L,L2,LL,a;
    double eta = (dataCenter.grid[2]-dataCenter.grid[0])/(dataCenter.grid[2]-dataCenter.grid[1]);
    double[] rho   = new double[ngrid];
    double[] dvDisp= new double[ngrid];
    double[] ksi   = new double[ngrid];

    double[][] dermat = Numerics.initDeriveFDBand(dataCenter.grid,1,1);
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);
    double[] mat = Numerics.findIntegrationMatrix(dataCenter.grid);
    double[] intg_aux = new double[ngrid];
    double[] aux = new double[ngrid];
    double[] intg1, intg2;
    double AA, intg_r3p;


    // find integral_0^R r^3*p(r) dr:
    intg_r3p = 0.0;
    for(i=1;i<ngrid;i++) intg_r3p+=(weights[i]*Math.pow(dataCenter.grid[i],3.0)*dataCenter.pressure[i]);

    for(fnum=0; fnum<nfunc; fnum++){

      // avoid recalculating kernel:
      if (dataCenter.kernels.get(fnum).rhoc2Kernel != null) continue;

      // initialise kernels:
      dataCenter.kernels.get(fnum).c2rhoKernel     = new double[ngrid];
      dataCenter.kernels.get(fnum).rhoc2Kernel     = new double[ngrid];
      dataCenter.kernels.get(fnum).rhoGamma1Kernel = new double[ngrid];

      degree = dataCenter.kernels.get(fnum).l;
      L  = (double)dataCenter.kernels.get(fnum).l;
      L2 = L*(L+1.0);
      LL = Math.sqrt(L2);

      // find derivative of vertical component: 
      Dgbmv.dgbmv("N",ngrid,ngrid,1,1,1.0,dermat[1],0,3,dataCenter.kernels.get(fnum).vDisp,0,1,0.0,dvDisp,0,1);

      // find ksi (= div displ):
      for (i = 1; i<ngrid; i++) {
        ksi[i] = dvDisp[i] + (2.0*dataCenter.kernels.get(fnum).vDisp[i]-L2*dataCenter.kernels.get(fnum).hDisp[i])/dataCenter.grid[i];
      }
      ksi[0] = (degree==0)?3.0*dvDisp[0]:0.0;

      // find eulerian density perturbation:
      for(i=0;i<ngrid;i++) rho[i] = -dataCenter.drho[i]*dataCenter.kernels.get(fnum).vDisp[i]-dataCenter.rho[i]*ksi[i];

      // find an antiderivative:
      for(i=0;i<ngrid;i++) intg_aux[i] = 2.0*dataCenter.rho[i]*dataCenter.kernels.get(fnum).vDisp[i]*ksi[i]
                                       + dataCenter.drho[i]*dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.kernels.get(fnum).vDisp[i];
      intg1 = Numerics.applyIntegrationMatrixReverse(mat, intg_aux);
      for(i=0;i<ngrid;i++) intg1[i]=4.0*Math.PI*(intg1[ngrid-1]-intg1[i]);

      ////////////////////////////////////////////////////////////////////////////////////
      // find {c^2,rho} Kernel (these are also the same for {Gamma1,rho})
      a = 1.0/(2.0*dataCenter.kernels.get(fnum).IK*Math.pow(dataCenter.kernels.get(fnum).FreqDimless,2.0));
      for(i=1;i<ngrid;i++) {
        dataCenter.kernels.get(fnum).c2rhoKernel[i] = a*dataCenter.grid[i]*dataCenter.grid[i]
                                        * dataCenter.c2[i]*dataCenter.rho[i]*ksi[i]*ksi[i];
      }
      dataCenter.kernels.get(fnum).c2rhoKernel[0] = 0.0;
      dataCenter.kernels.get(fnum).center_c2rhoKernel = a*dataCenter.c2[0]*ksi[0]*ksi[0];

      ////////////////////////////////////////////////////////////////////////////////////
      // find {rho,c^2} Kernel
      for(i=1;i<ngrid;i++) {
        dataCenter.kernels.get(fnum).rhoc2Kernel[i] = a*dataCenter.grid[i]*dataCenter.grid[i]*dataCenter.rho[i]*
              (dataCenter.c2[i]*ksi[i]*ksi[i]
               -Math.pow(dataCenter.kernels.get(fnum).FreqDimless,2.0)*(dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.kernels.get(fnum).vDisp[i]
               +L2*dataCenter.kernels.get(fnum).hDisp[i]*dataCenter.kernels.get(fnum).hDisp[i])
               -2.0*dataCenter.gravity[i]*dataCenter.kernels.get(fnum).vDisp[i]*ksi[i]-intg1[i]
               +2.0*dataCenter.gravity[i]*dataCenter.kernels.get(fnum).vDisp[i]*dvDisp[i]
               +4.0*Math.PI*dataCenter.rho[i]*dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.kernels.get(fnum).vDisp[i]
               +2.0*(dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.kernels.get(fnum).phip[i]
               +L2*dataCenter.kernels.get(fnum).hDisp[i]*dataCenter.kernels.get(fnum).phi[i]/dataCenter.grid[i]));
      }
      dataCenter.kernels.get(fnum).rhoc2Kernel[0] = 0.0;
      dataCenter.kernels.get(fnum).center_rhoc2Kernel = a*(dataCenter.c2[0]*ksi[0]*ksi[0]
               -Math.pow(dataCenter.kernels.get(fnum).FreqDimless,2.0)*(dataCenter.kernels.get(fnum).vDisp[0]*dataCenter.kernels.get(fnum).vDisp[0]
               +L2*dataCenter.kernels.get(fnum).hDisp[0]*dataCenter.kernels.get(fnum).hDisp[0])
               -intg1[0]+4.0*Math.PI*dataCenter.rho[0]*dataCenter.kernels.get(fnum).vDisp[0]*dataCenter.kernels.get(fnum).vDisp[0]
               +2.0*(dataCenter.kernels.get(fnum).vDisp[0]*dataCenter.kernels.get(fnum).phip[0]
               +L2*dataCenter.kernels.get(fnum).hDisp[0]*dataCenter.kernels.get(fnum).phip[0]));

      ////////////////////////////////////////////////////////////////////////////////////
      // find {rho,Gamma1} Kernel
      // 1st antiderivative;
      for(i=0;i<ngrid;i++) intg_aux[i]=a*dataCenter.Gamma1[i]*Math.pow(ksi[i]*dataCenter.grid[i],2.0);
      intg1 = Numerics.applyIntegrationMatrix(mat, intg_aux);
      for(i=ngrid-1;i>=0;i--) intg1[i]-=intg1[0];
      for(i=1;i<ngrid;i++) intg1[i]*=(dataCenter.rho[i]/Math.pow(dataCenter.grid[i],2.0));

      // 2nd antiderivative;
      for(i=1;i<ngrid;i++) intg_aux[i]=4.0*Math.PI*intg1[i];
      intg2 = Numerics.applyIntegrationMatrixReverse(mat, intg_aux);
      for(i=0;i<ngrid;i++) intg2[i]=intg2[ngrid-1]-intg2[i];

      // I'm assuming that the surface variation of p (between the reference and
      // observed models) is zero, when setting up the kernel:
      for(i=0; i<ngrid; i++) {
        dataCenter.kernels.get(fnum).rhoGamma1Kernel[i] = dataCenter.kernels.get(fnum).rhoc2Kernel[i]
                                            - dataCenter.kernels.get(fnum).c2rhoKernel[i]
                                            + intg1[i]*dataCenter.mass[i]
                                            + intg2[i]*Math.pow(dataCenter.grid[i],2.0)*dataCenter.rho[i];
      }
      dataCenter.kernels.get(fnum).center_rhoGamma1Kernel = dataCenter.kernels.get(fnum).center_rhoc2Kernel
                                              - dataCenter.kernels.get(fnum).center_c2rhoKernel
                                              + intg2[0];
    }
  }

  /******************************************************************************
  * This method recalculates the frequencies, using the variational principle.
  ******************************************************************************/
  public void applyVariational() {
    applyVariationalEuler();
  }

  /******************************************************************************
  * This method recalculates the frequencies, using the variational principle.
  ******************************************************************************/
  public void applyVariationalEuler() {
    int ngrid = dataCenter.grid.length;
    int nfunc = dataCenter.kernels.size();
    int i, fnum, n;
    double L,L2,sigma2;
    double[] p = new double[ngrid];
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);

    ArrayList<Integer> domains = Numerics.findDomains(dataCenter.grid);

    for(fnum=0; fnum<nfunc; fnum++){
      // avoid recalculating FreqVar:
      //if (!Double.isNaN(dataCenter.kernels.get(fnum).FreqVar)) continue;

      L  = (double)dataCenter.kernels.get(fnum).l;
      L2 = L*(L+1.0);

      // calculate variational principle:
      sigma2 = 0.0;
      for(i=1;i<ngrid;i++) {
        sigma2+=weights[i]*(dataCenter.kernels.get(fnum).P[i]*dataCenter.kernels.get(fnum).P[i]/(dataCenter.c2[i]*dataCenter.rho[i])
                          + dataCenter.rho[i]*dataCenter.N2[i]*dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.kernels.get(fnum).vDisp[i]
                          - dataCenter.kernels.get(fnum).phip[i]*dataCenter.kernels.get(fnum).phip[i]/(4.0*Math.PI)
                          - L2*Math.pow(dataCenter.kernels.get(fnum).phi[i]/dataCenter.grid[i],2.0)/(4.0*Math.PI))
                          * dataCenter.grid[i]*dataCenter.grid[i];
      }

      // terms arising from internal discontinuities:
      for(n=1; n<(domains.size()-2); n++) {
        i = domains.get(n);
        sigma2+=(dataCenter.rho[i]-dataCenter.rho[i+1])*dataCenter.gravity[i]
               * Math.pow(dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.grid[i],2.0);
      }

      // surface term when delta p = 0:
      sigma2+=dataCenter.rho[ngrid-1]*dataCenter.gravity[ngrid-1]
             *Math.pow(dataCenter.kernels.get(fnum).vDisp[ngrid-1]*dataCenter.grid[ngrid-1],2.0);

      // surface term in general case:
      //sigma2+=(dataCenter.kernels.get(fnum).dP_P[ngrid-1]
      //       +dataCenter.rho[ngrid-1]*dataCenter.gravity[ngrid-1]
      //       *dataCenter.kernels.get(fnum).vDisp[ngrid-1])
      //       *dataCenter.kernels.get(fnum).vDisp[ngrid-1];  // general case
             
      sigma2+=(L+1.0)*dataCenter.grid[ngrid-1]*Math.pow(dataCenter.kernels.get(fnum).phi[ngrid-1],2.0)/(4.0*Math.PI);

      // final result:
      dataCenter.kernels.get(fnum).FreqVar=Math.sqrt(sigma2/dataCenter.kernels.get(fnum).IK)*dataCenter.FreqRef;
    }
  }

  /******************************************************************************
  * This method recalculates the frequencies, using the variational principle.
  ******************************************************************************/
  public void applyVariationalLagrange() {
    int ngrid = dataCenter.grid.length;
    int nfunc = dataCenter.kernels.size();
    int i, fnum, n;
    double L,L2,sigma2;
    double[] p = new double[ngrid];
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);

    ArrayList<Integer> domains = Numerics.findDomains(dataCenter.grid);

    for(fnum=0; fnum<nfunc; fnum++){
      // avoid recalculating FreqVar:
      //if (!Double.isNaN(dataCenter.kernels.get(fnum).FreqVar)) continue;

      L  = (double)dataCenter.kernels.get(fnum).l;
      L2 = L*(L+1.0);

      // calculate variational principle:
      sigma2 = 0.0;
      for(i=1;i<ngrid;i++) {
        sigma2+=weights[i]*(dataCenter.kernels.get(fnum).dP_P[i]*dataCenter.kernels.get(fnum).dP_P[i]*dataCenter.pressure[i]/dataCenter.Gamma1[i]
                          + 2.0*dataCenter.rho[i]*dataCenter.gravity[i]*dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.kernels.get(fnum).dP_P[i]/dataCenter.Gamma1[i]
                          - dataCenter.gravity[i]*dataCenter.drho[i]*Math.pow(dataCenter.kernels.get(fnum).vDisp[i],2.0)
                          - dataCenter.kernels.get(fnum).phip[i]*dataCenter.kernels.get(fnum).phip[i]/(4.0*Math.PI)
                          - L2*Math.pow(dataCenter.kernels.get(fnum).phi[i]/dataCenter.grid[i],2.0)/(4.0*Math.PI))
                          * dataCenter.grid[i]*dataCenter.grid[i];
      }

      // terms arising from internal discontinuities:
      for(n=1; n<(domains.size()-2); n++) {
        i = domains.get(n);
        sigma2+=(dataCenter.rho[i]-dataCenter.rho[i+1])*dataCenter.gravity[i]
               * Math.pow(dataCenter.kernels.get(fnum).vDisp[i]*dataCenter.grid[i],2.0);
      }

      // surface term when delta p = 0:
      sigma2+=dataCenter.rho[ngrid-1]*dataCenter.gravity[ngrid-1]
             *Math.pow(dataCenter.kernels.get(fnum).vDisp[ngrid-1]*dataCenter.grid[ngrid-1],2.0);

      sigma2+=(L+1.0)*dataCenter.grid[ngrid-1]*Math.pow(dataCenter.kernels.get(fnum).phi[ngrid-1],2.0)/(4.0*Math.PI);

      // final result:
      dataCenter.kernels.get(fnum).FreqVar=Math.sqrt(sigma2/dataCenter.kernels.get(fnum).IK)*dataCenter.FreqRef;
    }
  }
}
