/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import inversion_applet.Numerics;
import org.netlib.blas.Dgemv;
import org.netlib.lapack.Dsytrf;
import org.netlib.lapack.Dsytrs;
import org.netlib.lapack.Ilaenv;
import java.util.Random;
import java.util.ArrayList;
import java.util.Vector;
import javax.swing.JOptionPane;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTable;

/*************************************************************************
**
**    RotationInversion
**                                              Version 2.1, September 2014
**
*************************************************************************/

public class RotationInversion { 

  private DataCenter dataCenter;
  protected boolean canInvert;   // after initialisation, if the provided information
                                 // is insufficient or erroneous, it is not possible
                                 // to do an inversion, in which case, canInvert = false
                                 // Otherwise, canInvert = true
  protected int n_inputKernels;  // the number of kernels used in the inversion
  public double[] f_invertedRLS; // function given by the RLS inversion (f_inverted[n_grid])
  public double[] RLS_errorBars; // this gives the uncertainty in the inversion, based on
                                 // the error bars of the rotational splittings
  public double[] f_invertedSOLA;// function given by the SOLA inversion (f_inverted[n_grid])
  public double[] SOLA_errorBars;// this gives the uncertainty in the inversion, based on
                                 // the error bars of the rotational splittings
  public double[] SOLA_leftErrorBars; // this gives the position of the 1st quartile point of the averaging kernels
  public double[] SOLA_newGrid;       // this gives the position of the 2nd quartile point of the averaging kernels
  public double[] SOLA_rightErrorBars;// this gives the position of the 3rd quartile point of the averaging kernels
  public double[] f_invertedMOLA;// function given by the SOLA inversion (f_inverted[n_grid])
  public double[] MOLA_errorBars;// this gives the uncertainty in the inversion, based on
                                 // the error bars of the rotational splittings
  public double[] MOLA_leftErrorBars; // this gives the position of the 1st quartile point of the averaging kernels
  public double[] MOLA_newGrid;       // this gives the position of the 2nd quartile point of the averaging kernels
  public double[] MOLA_rightErrorBars;// this gives the position of the 3rd quartile point of the averaging kernels

  // variables and constants used in the choice of method:
  protected static final int Chebyshev = 0;
  protected static final int SlidingWindow = 1; // the default
  protected int integrationMethod = SlidingWindow;
  
  // different quantities used in setting up B-splines
  private int order = 4;                 // order of the B-splines (4 = cubic b-splines)
  private double[] knots;                // knots used to determine the B-spline basis
  
  // various intermediate matrices used in the inverstions:
  private double[] fitDataMatRLS;          // matrix used for fitting the data in a RLS inversion
  private double[] regularisationMatRLS;   // matrix used for regularisation in a RLS inversion
  private double[] kernelMatRLS;           // matrix used for calculating output kernels
  private double[] RLSmat;                 // matrix which contains the coefficients from the inversion
  private double[] fitDataMatSOLA;         // matrix used for fitting the data in a SOLA inversion
  private double[] regularisationMatSOLA;  // matrix used for regularisation in a SOLA inversion
  private double[] targetKernelMatSOLA;    // matrix used for calculating the target kernels
  private double[][] averagingKernelsSOLA; // matrix which contains the averaging kernels from the inversion
  private double[] MatMOLA1;               // matrix which contains the constant part of the MOLA inversion matrix
  private double[] MatMOLA2;               // matrix which contains the part of the MOLA inversion matrix multiplied by x0
  private double[] MatMOLA3;               // matrix which contains the part of the MOLA inversion matrix multiplied by x0*x0
  private double[] regularisationMatMOLA;  // matrix used for regularisation in a MOLA inversion
  private double[][] averagingKernelsMOLA; // matrix which contains the averaging kernels from the inversion
  private org.netlib.util.intW info = new org.netlib.util.intW(2); // necessary for lapack
  private ArrayList<Integer> intersection = new ArrayList<Integer>();
  private ArrayList<Double>  dataValues   = new ArrayList<Double>();
  private ArrayList<Double>  dataErrors   = new ArrayList<Double>();
  private double[][] rotaKernel;
  private int lwork;        // an integer used in the matrix solvers
  private double[] intMatrix; // coefficients used in anti-derivation
  private double[] weights; // weights used in integration
  private double[] iGrid;   // grid used in integration (it changes according to the method)
  private double rspan;     // span of dataCenter.grid
  
  /********************************************************
  *  A constructor which does the initialisation
  ********************************************************/
  public RotationInversion(DataCenter dataCenter) {
    this.dataCenter = dataCenter;

    // At the beginning, no data is loaded, so no inversion
    // can be done.  Therefore the inversion is up to date.
    dataCenter.rotationInversionUpToDate = true;
    this.canInvert = false;

    // This determines the blocksize to be used in the symmetric
    // matrix solvers. I will just put some average matrix size
    // rather than an exact number.
    this.lwork = Ilaenv.ilaenv(1,"DSYTRF","U",100,100,-1,-1);
  }

  ////////////////////////////////////////////////////////////
  /** This calls various initialisation methods */
  ////////////////////////////////////////////////////////////
  public void init(JComponent c) {
    dataCenter.rotationInversionUpToDate = true;
    n_inputKernels = 0;
    initIntersection(c);
    if (canInvert) {
      initGridInversion();
      initRotaKernel();
      initNoise();
      initRLS();
      initSOLA();
      initDeltaSOLA();
      initMOLA();
    }
  }

  ////////////////////////////////////////////////////////////
  // This finds the intersection between the set of kernels
  // and the set of rotational splittings data and stores it in
  // an ArrayList called intersection
  ////////////////////////////////////////////////////////////
  private void initIntersection(JComponent c) {

    canInvert = true;

    // make sure rotational splittings are loaded:
    if (dataCenter.nRS == 0) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No rotational splittings loaded.  Unable to proceed",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      n_inputKernels = 0;
      canInvert = false;
      return;
    }

    // make sure a model is loaded
    if (!dataCenter.modelLoaded) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No model loaded.  Unable to proceed",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      n_inputKernels = 0;
      canInvert = false;
      return;
    }

    boolean dataWithZeroError = false;
    boolean incompleteRS = false;
    boolean incorrectRS = false;
    boolean duplicateKernels = false;

    // clear previous data:
    intersection.clear();
    dataValues.clear();
    dataErrors.clear();

    // find index array with rotational splittings:
    for(int i=0; i<dataCenter.modes.size();i++) {
      if (dataCenter.modes.get(i).FreqObs == null) continue;
      if (dataCenter.modes.get(i).RS == null) continue;
      if (dataCenter.modes.get(i).kernel == null) continue;
      if (dataCenter.modes.get(i).isDuplicate) {
        duplicateKernels = true;
        continue;
      }
      if (dataCenter.modes.get(i).ErrorRS == null) {
        incompleteRS = true;
        continue;
      }
      if (dataCenter.modes.get(i).ErrorRS < 0.0) {
        dataWithZeroError = true;
        continue;
      }
      if (dataCenter.modes.get(i).l.intValue() < 1) {
        incorrectRS = true;
        continue;
      }

      intersection.add(i);
      dataValues.add(dataCenter.modes.get(i).RS);
      dataErrors.add(dataCenter.modes.get(i).ErrorRS);
    }

    n_inputKernels = intersection.size();

    if (n_inputKernels == 0) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No kernels/data from which to carry\n"+
        "out an inversion.  Unable to proceed.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canInvert = false;
      return;
    }

    // Test to see if all the relevant kernels have
    // been calculated and used in the inversion.
    if (n_inputKernels < dataCenter.nRS) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some kernels might be missing from this inversion.\n"+
        "Pressing the \"Calculate kernels\" button on the\n"+
        "kernel tab would remedy this problem.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (dataWithZeroError) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the rotational splittings have\n"+
        "a 0 error bar, which is not allowed.\n"+
        "This data has consequently been ignored.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
      
    if (incompleteRS) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the rotational splittings were\n"+
        "incomplete or contained infinite/NaN\n"+
        "values and were therefore not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
    
    if (incorrectRS) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the rotational splittings seemed to be erroneous:\n"+
        "  (\u2113 < 1).\n"+
        "This data has consequently not been not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (duplicateKernels) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "There are duplicate kernels.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
  }

  ////////////////////////////////////////////////////////////
  // set up the grid on which to calculate the inverted
  // function:
  ////////////////////////////////////////////////////////////
  private void initGridInversion() {
    if (integrationMethod == Chebyshev) initGridInversionCheb();
    if (integrationMethod == SlidingWindow)  initGridInversionSplines();
  }

  ////////////////////////////////////////////////////////////
  // This initialises a Chebyshev collocation grid and spectral
  // integration weights to be used in the inversion.  For a
  // given number of grid points, it is more accurate than
  // initGridInversion() above.
  ////////////////////////////////////////////////////////////
  private void initGridInversionCheb() {
    int n_grid = dataCenter.n_gridR;
    rspan = dataCenter.grid[dataCenter.grid.length-1]-
            dataCenter.grid[0];

    dataCenter.gridRInversion = new double[n_grid];
    for(int i=0; i<n_grid;i++)
      dataCenter.gridRInversion[i] = 
         0.5*(1.0-Math.cos(Math.PI*(double)(2*i+1)/(double)(2*n_grid)))
        *rspan+dataCenter.grid[0];

    // initialise some integration weights and grid:
    weights = Numerics.initWeightsCheb(n_grid);
    iGrid = dataCenter.gridRInversion;
    for(int i=0; i< n_grid; i++) weights[i]*=rspan;
    intMatrix = Numerics.findIntegrationMatrix(dataCenter.grid);
  }

  ////////////////////////////////////////////////////////////
  // This initialises a grid to be used in the inversion.
  // It works with a spline based method and spaces the
  // grid points according to sound velocity.
  ////////////////////////////////////////////////////////////
  private void initGridInversionSplines() {
    int n_grid = dataCenter.n_gridR;
    rspan = dataCenter.grid[dataCenter.grid.length-1]-
            dataCenter.grid[0];

    // obtain the sound velocity:
    double[] c = new double[dataCenter.c2.length];
    for(int i=0; i<c.length; i++) c[i] = Math.sqrt(dataCenter.c2[i]);

    // this is to avoid division by zero later on:
    double max_c = 0.0;
    for(int i=0; i<c.length; i++) if (max_c < c[i]) max_c = c[i];
    for(int i=0; i<c.length; i++) if ((c[i]/max_c) < 1e-3) c[i] = max_c*1e-3;

    // find the inverse sound velocity:
    for(int i=0; i<c.length; i++) c[i] = 1.0/Math.sqrt(c[i]);

    double[] CC = Numerics.findAntiDerivative(dataCenter.grid,c);
    double cInt = CC[CC.length-1]-CC[0];

    dataCenter.gridRInversion = new double[n_grid];
    int j = 0;
    double cTarget, lambda;
    //for(int i=0; i<n_grid; i++) {
    for(int i=1; i<n_grid-1; i++) {
      cTarget = cInt*(double)(i)/(double)(n_grid-1);  // a target value
      //cTarget = cInt*(double)(2*i+1)/(double)(2*n_grid);  // a target value
      while(CC[j]<cTarget) j++;
      lambda = (CC[j]-cTarget)/(CC[j]-CC[j-1]);
      dataCenter.gridRInversion[i] = dataCenter.grid[j-1]*lambda 
                                   + dataCenter.grid[j]*(1.0-lambda);
      j--;
    }
    dataCenter.gridRInversion[0] = dataCenter.grid[0];
    dataCenter.gridRInversion[n_grid-1] = dataCenter.grid[dataCenter.grid.length-1];

    // initialise some knots:
    knots = Numerics.find_knots(dataCenter.gridRInversion, order);
    
    // initialise some integration weights and grid:
    iGrid = dataCenter.grid;
    weights = Numerics.findIntegrationWeights(iGrid);
    intMatrix = Numerics.findIntegrationMatrix(dataCenter.grid);
  }

  ////////////////////////////////////////////////////////////
  // set up rotational kernels before doing inversions:
  ////////////////////////////////////////////////////////////
  private void initRotaKernel() {
    if (integrationMethod == Chebyshev) initRotaKernelCheb();
    if (integrationMethod == SlidingWindow)  initRotaKernelSplines();
  }
  
  ////////////////////////////////////////////////////////////
  // interpolate all rotational kernels before doing inversions:
  ////////////////////////////////////////////////////////////
  private void initRotaKernelCheb() {
    double[][] aux = new double[dataCenter.modes.size()][];
    for(int i=0; i<dataCenter.modes.size(); i++)
      aux[i] = dataCenter.modes.get(i).kernel.rotaKernel;
    rotaKernel = Numerics.interpolateList_r2(dataCenter.grid,
                 dataCenter.gridRInversion,
                 aux,intersection,false);
  }

  ////////////////////////////////////////////////////////////
  // set up rotational kernels for spline based method:
  ////////////////////////////////////////////////////////////
  private void initRotaKernelSplines() {
    rotaKernel = new double[n_inputKernels][];
    for(int l=0; l<n_inputKernels; l++)
      rotaKernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.rotaKernel;
  }

  /********************************************************
  * Initialise the RNoise array with a Gauss distribution:
  * Mean = 0.0; std. deviation= 1.0.
  ********************************************************/
  protected void initNoise() {
    Random generator = new Random();
    dataCenter.RNoise = new double[intersection.size()];
    for(int l=0; l < intersection.size(); l++) {
      dataCenter.RNoise[l] = generator.nextGaussian();
    }
  }

  ////////////////////////////////////////////////////////////
  // Initialise various matrices used in a RLS inversion.
  ////////////////////////////////////////////////////////////
  private void initRLS() {
    if (integrationMethod == Chebyshev) initRLScheb();
    if (integrationMethod == SlidingWindow)  initRLSsplines();
  }

  /********************************************************
  * Initialise various matrices used in a RLS inversion
  * for a method based on Chebyshev polynomials.
  ********************************************************/
  private void initRLScheb() {
    int n_grid = dataCenter.gridRInversion.length;
    int i, j, k, l;

    // initialise the matrix used for calculating the output kernels:
    kernelMatRLS = new double[n_grid*n_inputKernels];
    for(l=0; l<n_inputKernels; l++){
      for(i=0; i<n_grid; i++)
        kernelMatRLS[i+n_grid*l] = weights[i]*rotaKernel[l][i]/dataErrors.get(l);
    }

    // initialise matrix used for fitting the data:
    fitDataMatRLS = new double[n_grid*n_grid];
    for(i=0; i<n_grid; i++){
      for(j=i; j<n_grid; j++){
        fitDataMatRLS[i+n_grid*j]=0.0;
        for(l=0; l<n_inputKernels; l++){
          fitDataMatRLS[i+n_grid*j]+=kernelMatRLS[i+n_grid*l]*
                                     kernelMatRLS[j+n_grid*l];
        }
      }
    }

    // finish initialising kernelMatRLS:
    for(i=0; i<n_grid; i++) {
      for(l=0; l<n_inputKernels; l++) 
        kernelMatRLS[i+n_grid*l]/=dataErrors.get(l);
    }

    // initialise a derivation matrix:
    double[] derive = Numerics.initDDerive(n_grid);
    double cnst = 1.0/(rspan*rspan);
    for(i=0; i<(n_grid*n_grid); i++) derive[i]*=cnst;

    // initialise the regularisation matrix:
    regularisationMatRLS = new double[n_grid*n_grid];
    for(i=0; i<n_grid; i++){
      for(j=i; j<n_grid; j++){
        regularisationMatRLS[i+n_grid*j]=0.0;
        for(k=0; k<n_grid; k++){
          regularisationMatRLS[i+n_grid*j]+=(weights[k]*derive[k+n_grid*i]*derive[k+n_grid*j]);
        }
      }
    }
  }

  /********************************************************
  * Initialise various matrices used in a RLS inversion
  * for a spline-based method.
  ********************************************************/
  private void initRLSsplines() {
    int n_grid = dataCenter.gridRInversion.length;
    int n_big_grid = iGrid.length;
    double lb = knots[0];
    double[] subknots = new double[order+1];
    int i, j, k, l;

    // initialise the matrix used for calculating the output kernels:
    int istart = 0;
    double bs_value;
    kernelMatRLS = new double[n_grid*n_inputKernels];
    for(i=0; i<n_grid; i++){
      for(j=0; j<= order; j++) subknots[j] = knots[i+j];
      while (iGrid[istart] < subknots[0]) istart++;
      j = istart;
      while ((j < n_big_grid)&&(iGrid[j] <= subknots[order])) {
        bs_value = Numerics.find_b_spline(iGrid[j],subknots,lb);
        for(l=0; l<n_inputKernels; l++)
          kernelMatRLS[i+n_grid*l]+=weights[j]*rotaKernel[l][j]*bs_value;
        j++;
      }
    }
    for(i=0; i<n_grid; i++) {
      for(l=0; l<n_inputKernels; l++) {
        kernelMatRLS[i+n_grid*l]/=dataErrors.get(l);
      }
    }

    // initialise matrix used for fitting the data:
    fitDataMatRLS = new double[n_grid*n_grid];
    for(i=0; i<n_grid; i++){
      for(j=i; j<n_grid; j++){
        fitDataMatRLS[i+n_grid*j]=0.0;
        for(l=0; l<n_inputKernels; l++){
          fitDataMatRLS[i+n_grid*j]+=kernelMatRLS[i+n_grid*l]*
                                     kernelMatRLS[j+n_grid*l];
        }
      }
    }

    // finish initialising kernelMatRLS:
    for(i=0; i<n_grid; i++) {
      for(l=0; l<n_inputKernels; l++) 
        kernelMatRLS[i+n_grid*l]/=dataErrors.get(l);
    }

    // find Gauss-Legendre collocation points and weigths:
    int ngauleg = order - 2;
    double[][] xw = Numerics.gauleg(0.0, 1.0, ngauleg);

    // find b-spline values before calculating regularisation matrix:
    double[][][] bs = new double[n_grid][order][ngauleg];
    double[] temp = new double[3];
    double xp;
    for(i=0; i<n_grid; i++) {
      for(j=0; j<=order; j++) subknots[j] = knots[i+j];
      for(j=0; j<order; j++) {
        for(k=0; k<ngauleg; k++) {
          xp = subknots[j]+(subknots[j+1]-subknots[j])*xw[0][k];
          temp = Numerics.find_b_spline(xp,subknots,2,lb);
          bs[i][j][k] = temp[2];
        }
      }
    }

    // initialise the regularisation matrix:
    regularisationMatRLS = new double[n_grid*n_grid];
    int jfinish;
    for(i=0; i<n_grid; i++){
      jfinish = i+order;
      if (n_grid < jfinish) jfinish = n_grid;
      for(j=i; j<jfinish; j++){
        for(k=j; k<i+order; k++){
          for(l=0; l<ngauleg; l++){
            regularisationMatRLS[i+n_grid*j]+=
              xw[1][l]*
              bs[i][k-i][l]*
              bs[j][k-j][l]*
              (knots[k+1]-knots[k]);
          }
        }
      }
    }
  }

  /********************************************************
  * Do the inversion, using RLS
  ********************************************************/
  protected void RLSinversion() {  
    int n_grid = dataCenter.gridRInversion.length;
    int n_grid2 = n_grid*n_grid;
    int ipiv[] = new int[n_grid];
    int i,j,l;
    
    double[] mat = fitDataMatRLS.clone();

    double Lambda = 0.0;
    for(l=0; l<n_inputKernels; l++)
      Lambda+=1.0/(dataErrors.get(l)*dataErrors.get(l));
    Lambda*=(dataCenter.RLambda/(double)n_inputKernels);
    
    for(i=0; i<n_grid2; i++)
      mat[i]+=Lambda*regularisationMatRLS[i];
    RLSmat = kernelMatRLS.clone();

    double[] work = new double[lwork*n_grid];
    Dsytrf.dsytrf("U",n_grid,mat,0,n_grid,ipiv,0,work,0,lwork*n_grid,info);
    Dsytrs.dsytrs("U",n_grid,n_inputKernels,mat,0,n_grid,ipiv,0,RLSmat,0,n_grid,info);

    if (integrationMethod == SlidingWindow) {
      // transform RLSmat by going from a spline coefficient
      // representation to a grid based representation:
      double[] new_RLSmat = new double[n_grid*n_inputKernels];
      double[] subknots = new double[order+1];
      double bs_value, lb = dataCenter.gridRInversion[0];
      int istart = 0;
      for(i=0; i<n_grid; i++){
        for(j=0; j<= order; j++) subknots[j] = knots[i+j];
        while (dataCenter.gridRInversion[istart] < subknots[0]) istart++;
        j = istart;
        while ((j < n_grid)&&(dataCenter.gridRInversion[j] <= subknots[order])) {
          bs_value = Numerics.find_b_spline(dataCenter.gridRInversion[j],subknots,lb);
          for(l = 0; l<n_inputKernels; l++)
            new_RLSmat[j+n_grid*l]+=RLSmat[i+n_grid*l]*bs_value;
          j++;
        }
      }
      RLSmat = new_RLSmat;
    }

    // find inverted function:
    f_invertedRLS = new double[n_grid];
    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.RNoiseSelected) {
      for(l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.RNoiseLevel*dataCenter.RNoise[l]*dataErrors.get(l);
      }
    } else {
      for(l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }
    Dgemv.dgemv("N",n_grid,n_inputKernels,1.0,RLSmat,0,n_grid,
                dataPlusNoise,0,1,0.0,f_invertedRLS,0,1);

    // find the error bars:
    RLS_errorBars = new double[n_grid];
    for(i=0; i<n_grid; i++) {
      RLS_errorBars[i] = 0.0;
      for(l=0; l<n_inputKernels; l++) {
        RLS_errorBars[i]+=Math.pow(RLSmat[i+n_grid*l]*dataErrors.get(l),2.0);
      }
      RLS_errorBars[i] = Math.sqrt(RLS_errorBars[i]);
    }
  }

  /********************************************************
  * This calculates the RLS averaging kernel at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findAveragingKernelRLS(int iLocation) {
    int n_grid = dataCenter.gridRInversion.length;
    int n_big_grid = dataCenter.grid.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;

    int i,l;    
    double[] result = new double[n_big_grid];
    for(i=0; i<n_big_grid ;i++){
      for(l=0; l<n_inputKernels;l++){
        result[i]+=(
          dataCenter.modes.get(intersection.get(l)).kernel.rotaKernel[i]
          *RLSmat[iLocation+n_grid*l]);
      }
    }
    return result;
    //return Numerics.applyIntegrationMatrix(intMatrix,result,0.0);
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion
  * for the spline-based method.
  ********************************************************/
  private void initSOLA() {
    int n_grid = iGrid.length;

    // initialise regularisation matrix:
    double mu = 0.0;
    for(int l=0; l<n_inputKernels; l++) mu+=dataErrors.get(l)*dataErrors.get(l);
    mu = n_inputKernels/mu;
    
    regularisationMatSOLA = new double[n_inputKernels];
    for(int l=0; l<n_inputKernels; l++){
      regularisationMatSOLA[l]=mu*dataErrors.get(l)*dataErrors.get(l);
    }

    // initialise matrix used for fitting the data:
    fitDataMatSOLA = new double[(n_inputKernels+1)*(n_inputKernels+1)];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        fitDataMatSOLA[l+(n_inputKernels+1)*ll]=0.0;
        for(int i=0; i<n_grid; i++){
          fitDataMatSOLA[l+(n_inputKernels+1)*ll]+=weights[i]
               *rotaKernel[l][i]*rotaKernel[ll][i];
        }
      }
    }
    
    double sum;
    for(int l=0; l<n_inputKernels; l++){
      sum = 0.0;
      for(int i=0; i<n_grid; i++) sum+=(weights[i]*rotaKernel[l][i]);
      fitDataMatSOLA[l+(n_inputKernels+1)*n_inputKernels] = sum*0.5;
    }
    fitDataMatSOLA[n_inputKernels+(n_inputKernels+1)*n_inputKernels]=0.0;
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  protected void initDeltaSOLA() {
    int n_grid = dataCenter.gridRInversion.length;
    int n_big_grid = iGrid.length;

    // obtain the sound velocity:
    double[] c = Numerics.interpolate1D_r2(dataCenter.grid,dataCenter.gridRInversion,dataCenter.c2,false);
    for(int i=0; i<c.length; i++) c[i] = Math.sqrt(c[i]);

    // this is to avoid division by zero later on:
    double max_c = 0.0;
    for(int i=0; i<c.length; i++) if (max_c < c[i]) max_c = c[i];
    for(int i=0; i<c.length; i++) if ((c[i]/max_c) < 1e-3) c[i] = max_c*1e-3;

    // initialise the matrix with the target kernels:
    double Norm, Delta, Cnst;
    double FunctionValue;
    targetKernelMatSOLA = new double[(n_inputKernels+1)*n_grid];
    for(int i=0; i<n_grid; i++) {
      Norm = 0.0;
      Delta = dataCenter.RDelta*c[i];
      Cnst = Math.log(1.0/(Delta*dataCenter.gridRInversion[i]));
      for(int j=0; j<n_big_grid; j++) {
        //FunctionValue = weights[j]*iGrid[j]*Math.exp(-Math.pow((iGrid[j]-dataCenter.gridRInversion[i])/Delta,2.0)
        //              - (iGrid[j]-dataCenter.gridRInversion[i])/dataCenter.gridRInversion[i] + Cnst);
        FunctionValue = weights[j]*Math.exp(-Math.pow((iGrid[j]-dataCenter.gridRInversion[i])/Delta,2.0));
        Norm+=FunctionValue;
        for(int l=0; l<n_inputKernels; l++)
          targetKernelMatSOLA[l+(n_inputKernels+1)*i]+=(FunctionValue*rotaKernel[l][j]);
      }
      Norm = 1.0/Norm; 
      for(int l=0; l<n_inputKernels; l++)
        targetKernelMatSOLA[l+(n_inputKernels+1)*i]*=Norm;
      targetKernelMatSOLA[n_inputKernels+(n_inputKernels+1)*i]=0.5;
    }
  }

  /********************************************************
  * Do the inversion, using SOLA
  ********************************************************/
  protected void SOLAinversion() {
    int n_grid = dataCenter.gridRInversion.length;
    int n_big_grid = dataCenter.grid.length;
    int ipiv[] = new int[n_inputKernels+1];

    double[] mat = fitDataMatSOLA.clone();
    double tanTheta = Math.tan(dataCenter.RTheta);
    for(int l=0; l<n_inputKernels; l++)
      mat[(n_inputKernels+2)*l]+=tanTheta*regularisationMatSOLA[l];

    double[] SOLAmat = targetKernelMatSOLA.clone();
    double[] work = new double[lwork*(n_inputKernels+1)];
    Dsytrf.dsytrf("U",n_inputKernels+1,mat,0,n_inputKernels+1,ipiv,0,work,0,lwork*(n_inputKernels+1),info);
    Dsytrs.dsytrs("U",n_inputKernels+1,n_grid,mat,0,n_inputKernels+1,ipiv,0,SOLAmat,0,n_inputKernels+1,info);

    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.RNoiseSelected) {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.RNoiseLevel*dataCenter.RNoise[l]*dataErrors.get(l);
      }
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }
    
    // find the inteverted function:
    f_invertedSOLA = new double[n_grid];
    Dgemv.dgemv("T",n_inputKernels,n_grid,1.0,SOLAmat,0,n_inputKernels+1,
                dataPlusNoise,0,1,0.0,f_invertedSOLA,0,1);

    // find the error bars:
    SOLA_errorBars = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      SOLA_errorBars[i] = 0.0;
      for(int l=0; l<n_inputKernels; l++) {
        SOLA_errorBars[i]+=Math.pow(SOLAmat[l+(n_inputKernels+1)*i]*dataErrors.get(l),2.0);
      }
      SOLA_errorBars[i]=Math.sqrt(SOLA_errorBars[i]);
    }

    // Averaging kernels obtained from the inversion:
    // 1st coordinate: target position
    // 2nd coordinate: grid position
    averagingKernelsSOLA = new double[n_grid][n_big_grid];
    for(int i=0; i<n_big_grid; i++) {
      for(int j=0; j<n_grid; j++) {
        averagingKernelsSOLA[j][i]=0.0;
        for(int k=0; k<n_inputKernels; k++) {
          averagingKernelsSOLA[j][i]+=(dataCenter.modes.get(intersection.get(k)).kernel.rotaKernel[i]*SOLAmat[k+(n_inputKernels+1)*j]);
        }
      }
    }

    // find new output grid and horizontal error bars for the SOLA method:
    double[] matIntg = Numerics.findIntegrationMatrix(dataCenter.grid);
    double[] intg;
    double lambda;
    SOLA_newGrid = new double[n_grid];
    SOLA_leftErrorBars  = new double[n_grid];
    SOLA_rightErrorBars = new double[n_grid];
    for(int j=0; j<n_grid; j++) {
      intg = Numerics.applyIntegrationMatrix(matIntg, averagingKernelsSOLA[j]);
      int i=1;
      while(i<n_big_grid) {if (intg[i]>0.25) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.25-intg[i-1])/(intg[i]-intg[i-1]);
      SOLA_leftErrorBars[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      while(i<n_big_grid) {if (intg[i]>0.5) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.5-intg[i-1])/(intg[i]-intg[i-1]);
      SOLA_newGrid[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      while(i<n_big_grid) {if (intg[i]>0.75) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.75-intg[i-1])/(intg[i]-intg[i-1]);
      SOLA_rightErrorBars[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      SOLA_leftErrorBars[j] = SOLA_newGrid[j]-SOLA_leftErrorBars[j];
      SOLA_rightErrorBars[j] = SOLA_rightErrorBars[j]-SOLA_newGrid[j];
    }
  }

  /********************************************************
  * This calculates the SOLA averaging kernel at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findAveragingKernelSOLA(int iLocation) {
    int n_grid = dataCenter.gridRInversion.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;
    return averagingKernelsSOLA[iLocation];
    //return Numerics.applyIntegrationMatrix(intMatrix,averagingKernelsSOLA[iLocation],0.0);
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion
  * for the spline-based method.
  ********************************************************/
  private void initMOLA() {
    int n_grid = iGrid.length, l, ll, i;

    // initialise regularisation matrix:
    double mu = 0.0;
    for(l=0; l<n_inputKernels; l++) mu+=dataErrors.get(l)*dataErrors.get(l);
    mu = n_inputKernels/mu;
    
    regularisationMatMOLA = new double[n_inputKernels];
    for(l=0; l<n_inputKernels; l++){
      regularisationMatMOLA[l]=mu*dataErrors.get(l)*dataErrors.get(l);
    }

    // initialise matrix used for fitting the data:
    double cnst;
    MatMOLA1 = new double[(n_inputKernels+1)*(n_inputKernels+1)];
    MatMOLA2 = new double[(n_inputKernels+1)*(n_inputKernels+1)];
    MatMOLA3 = new double[(n_inputKernels+1)*(n_inputKernels+1)];
    for(l=0; l<n_inputKernels; l++){
      for(ll=l; ll<n_inputKernels; ll++){
        MatMOLA1[l+(n_inputKernels+1)*ll]=0.0;
        MatMOLA2[l+(n_inputKernels+1)*ll]=0.0;
        MatMOLA3[l+(n_inputKernels+1)*ll]=0.0;
        for(i=0; i<n_grid; i++){
          cnst = weights[i]*rotaKernel[l][i]*rotaKernel[ll][i];
          MatMOLA1[l+(n_inputKernels+1)*ll]+=cnst*iGrid[i]*iGrid[i];
          MatMOLA2[l+(n_inputKernels+1)*ll]+=cnst*iGrid[i];
          MatMOLA3[l+(n_inputKernels+1)*ll]+=cnst;
        }
        MatMOLA1[l+(n_inputKernels+1)*ll]*=  12.0;
        MatMOLA2[l+(n_inputKernels+1)*ll]*=(-24.0);
        MatMOLA3[l+(n_inputKernels+1)*ll]*=  12.0;
      }
    }
    
    double sum;
    for(l=0; l<n_inputKernels; l++){
      sum = 0.0;
      for(i=0; i<n_grid; i++) sum+=(weights[i]*rotaKernel[l][i]);
      MatMOLA1[l+(n_inputKernels+1)*n_inputKernels] = sum*0.5;
    }
    MatMOLA1[n_inputKernels+(n_inputKernels+1)*n_inputKernels]=0.0;

    // Symmetrise the matrices (just in case ...):
    for(l=1; l<n_inputKernels; l++) {
      for(ll=0;ll<l;ll++) {
        MatMOLA1[l+(n_inputKernels+1)*ll] = MatMOLA1[ll+(n_inputKernels+1)*l];
        MatMOLA2[l+(n_inputKernels+1)*ll] = MatMOLA2[ll+(n_inputKernels+1)*l];
        MatMOLA3[l+(n_inputKernels+1)*ll] = MatMOLA3[ll+(n_inputKernels+1)*l];
      }
    }
  }

  /********************************************************
  * Do the inversion, using SOLA
  ********************************************************/
  protected void MOLAinversion() {
    int n_grid = dataCenter.gridRInversion.length;
    int n_big_grid = dataCenter.grid.length;
    int ipiv[] = new int[n_inputKernels+1];
    int i, j, k, l, size2 = (n_inputKernels+1)*(n_inputKernels+1);

    double tanTheta = Math.tan(dataCenter.RThetaM), x0;
    double[] work = new double[lwork*(n_inputKernels+1)];
    double[] mat = new double[size2];
    double[] vector = new double[n_inputKernels+1];
    double[] MOLAmat = new double[n_grid*n_inputKernels];
    for(i=0;i<n_grid;i++) {
      x0 = dataCenter.gridRInversion[i];
      for(l=0;l<size2;l++) mat[l] = MatMOLA1[l]+x0*(MatMOLA2[l]+x0*MatMOLA3[l]);
      for(l=0; l<n_inputKernels; l++) mat[(n_inputKernels+2)*l]+=tanTheta*regularisationMatMOLA[l];
      for(l=0; l<n_inputKernels; l++) vector[l] = 0.0;
      vector[n_inputKernels] = 0.5;
      Dsytrf.dsytrf("U",n_inputKernels+1,mat,0,n_inputKernels+1,ipiv,0,work,0,lwork*(n_inputKernels+1),info);
      Dsytrs.dsytrs("U",n_inputKernels+1,1,mat,0,n_inputKernels+1,ipiv,0,vector,0,n_inputKernels+1,info);
      for(l=0; l<n_inputKernels;l++) MOLAmat[i+n_grid*l] = vector[l];
    }

    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.RNoiseSelected) {
      for(l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.RNoiseLevel*dataCenter.RNoise[l]*dataErrors.get(l);
      }
    } else {
      for(l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }
    
    // find the inteverted function:
    f_invertedMOLA = new double[n_grid];
    Dgemv.dgemv("N",n_grid,n_inputKernels,1.0,MOLAmat,0,n_grid,
                dataPlusNoise,0,1,0.0,f_invertedMOLA,0,1);

    // find the error bars:
    MOLA_errorBars = new double[n_grid];
    for(i=0; i<n_grid; i++) {
      MOLA_errorBars[i] = 0.0;
      for(l=0; l<n_inputKernels; l++) {
        MOLA_errorBars[i]+=Math.pow(MOLAmat[i+n_grid*l]*dataErrors.get(l),2.0);
      }
      MOLA_errorBars[i]=Math.sqrt(MOLA_errorBars[i]);
    }

    // Averaging kernels obtained from the inversion:
    // 1st coordinate: target position
    // 2nd coordinate: grid position
    averagingKernelsMOLA = new double[n_grid][n_big_grid];
    for(i=0; i<n_big_grid; i++) {
      for(j=0; j<n_grid; j++) {
        averagingKernelsMOLA[j][i]=0.0;
        for(k=0; k<n_inputKernels; k++) {
          averagingKernelsMOLA[j][i]+=(dataCenter.modes.get(intersection.get(k)).kernel.rotaKernel[i]*MOLAmat[j+n_grid*k]);
        }
      }
    }

    // find new output grid and horizontal error bars for the MOLA method:
    double[] matIntg = Numerics.findIntegrationMatrix(dataCenter.grid);
    double[] intg;
    double lambda;
    MOLA_newGrid = new double[n_grid];
    MOLA_leftErrorBars  = new double[n_grid];
    MOLA_rightErrorBars = new double[n_grid];
    for(j=0; j<n_grid; j++) {
      intg = Numerics.applyIntegrationMatrix(matIntg, averagingKernelsMOLA[j]);
      i=1;
      while(i<n_big_grid) {if (intg[i]>0.25) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.25-intg[i-1])/(intg[i]-intg[i-1]);
      MOLA_leftErrorBars[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      while(i<n_big_grid) {if (intg[i]>0.5) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.5-intg[i-1])/(intg[i]-intg[i-1]);
      MOLA_newGrid[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      while(i<n_big_grid) {if (intg[i]>0.75) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.75-intg[i-1])/(intg[i]-intg[i-1]);
      MOLA_rightErrorBars[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      MOLA_leftErrorBars[j] = MOLA_newGrid[j]-MOLA_leftErrorBars[j];
      MOLA_rightErrorBars[j] = MOLA_rightErrorBars[j]-MOLA_newGrid[j];
    }
  }

  /********************************************************
  * This calculates the MOLA averaging kernel at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findAveragingKernelMOLA(int iLocation) {
    int n_grid = dataCenter.gridRInversion.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;
    return averagingKernelsMOLA[iLocation];
    //return Numerics.applyIntegrationMatrix(intMatrix,averagingKernelsMOLA[iLocation],0.0);
  }

}
