/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import graph.IPlot;
import graph.Plot;
import inversion_applet.Utilities;
import inversion_applet.DataCenter;
import inversion_applet.RotationInversion;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JSeparator;
import javax.swing.JCheckBox;
import javax.swing.JSlider;
import javax.swing.JComponent;
import javax.swing.JComboBox;
import javax.swing.SwingConstants;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import java.util.StringTokenizer;
import java.util.ArrayList;

/*************************************************************************
**
**    RotationInversionPage
**                                              Version 2.1, September 2014
**
*************************************************************************/

public class RotationInversionPage extends JPanel {

  private DataCenter dataCenter;
  private IPlot graphInvertedFunction;
  private IPlot graphAveragingKernel;
  private JPanel graphContainer;
  private RotationInversion my_inversion;
  private JPanel SliderContainer;

  private JLabel numberOfKernels;
  private JButton Refresh;
  private JCheckBox SOLACheckBox;
  private JCheckBox MOLACheckBox;
  private JCheckBox RLSCheckBox;
  private JCheckBox TargetFunctionCheckBox;

  private JCheckBox calculateWhileSliding;
  private JCheckBox showErrorBars;
  
  private JPanel NoiseTitle;
  private JCheckBox NoiseCheckBox;
  private JTextField NoiseTextField;
  private JSlider NoiseSlider;
  private JButton NoiseRefresh;

  private JLabel RLStitle;
  private JPanel LambdaTitle;
  private JLabel LambdaLabel;
  private JTextField LambdaTextField;
  private JSlider LambdaSlider;

  private JLabel SOLAtitle;
  private JPanel ThetaTitle;
  private JLabel ThetaLabel;
  private JTextField ThetaTextField;
  private JSlider ThetaSlider;
  private JPanel DeltaTitle;
  private JLabel DeltaLabel;
  private JTextField DeltaTextField;
  private JSlider DeltaSlider;

  private JLabel MOLAtitle;
  private JPanel MOLAThetaTitle;
  private JLabel MOLAThetaLabel;
  private JTextField MOLAThetaTextField;
  private JSlider MOLAThetaSlider;

  private JPanel KernelTitle;
  private JLabel KernelLabel;
  private JTextField KernelTextField;
  private JSlider KernelSlider;
  private JPanel NGridPanel;
  private JLabel NGridLabel;
  private JTextField NGridTextField;

  private JComboBox integrationMethodComboBox;

  private JButton LoadTarget;
  private JButton ClearTarget;
  private JButton CalculateRS;

  // this keeps track of whether a parameter has been modified in one of the
  // text fields.
  private boolean paramModified;

  private static final double ln10 = 2.302585092994046;
  private static final Color DarkGreen = new Color(0.0f,0.5f,0.0f);

  /************************************************************************
  * A constructor.
  ************************************************************************/
  public RotationInversionPage(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
    this.paramModified = false;
  }

  //////////////////////////////////////////////////////////////
  //  These methods give the relationship between the slider
  //  value and that of various parameters of the inversion.
  //////////////////////////////////////////////////////////////
  private int LambdaToSlider(double RLambda) {
    if (RLambda == 0.0) return -180;
    int result = (int)Math.round(Math.log(RLambda)*20.0/ln10);
    if (result < -180) result = -180;
    if (result > 20) result = 20;
    return result;
  }
  private double SliderToLambda() {
    return Math.pow(10.0,LambdaSlider.getValue()/20.0);
  }
  private int NoiseToSlider(double RNoiseLevel) {
    if (RNoiseLevel == 0.0) return -60;
    int result = (int)Math.round(Math.log(RNoiseLevel)*20.0/ln10);
    if (result < -60) result = -60;
    if (result > 60) result = 60;
    return result;
  }
  private double SliderToNoise() {
    return Math.pow(10.0,NoiseSlider.getValue()/20.0);
  }
  private int ThetaToSlider(double RTheta) {
    if (RTheta == 0.0) return -200;
    int result = (int)Math.round(Math.log(RTheta/1.5)*50.0/ln10);
    if (result < -200) result = -200;
    if (result > 0) result = 0;
    return result;
  }
  private double SliderToTheta() {
    return 1.5*Math.pow(10.0,ThetaSlider.getValue()/50.0);
  }
  private double SliderToThetaM() {
    return 1.5*Math.pow(10.0,MOLAThetaSlider.getValue()/50.0);
  }
  private int DeltaToSlider(double RDelta) {
    if (RDelta == 0.0) return -200;
    int result = (int)Math.round(Math.log(RDelta)*50.0/ln10);
    if (result < -200) result = -200;
    if (result > 0) result = 0;
    return result;
  }
  private double SliderToDelta() {
    return Math.pow(10.0,DeltaSlider.getValue()/50.0);
  }

  private int XKernelToSlider(double xKernel) {
    if (dataCenter.gridRInversion == null) return 0;
    int ngridR = dataCenter.gridRInversion.length;
    if (ngridR == 0) return 0;

    if (xKernel < dataCenter.gridRInversion[0]) return 0;
    if (xKernel > dataCenter.gridRInversion[ngridR-1]) return ngridR-1;

    int istart = 0;
    int ifinish = ngridR - 1;
    int inext;
    while((istart+1) < ifinish) {
      inext = (istart+ifinish)/2;
      if (dataCenter.gridRInversion[inext] > xKernel) {
        ifinish = inext;
      } else {
        istart = inext;
      }
    }

    if ((xKernel-dataCenter.gridRInversion[istart]) < (dataCenter.gridRInversion[ifinish]-xKernel)) {
      return istart;
    } else {
      return ifinish;
    }
  }

  private double SliderToXKernel () {
    if (dataCenter.gridRInversion == null) return 0.0;
    return dataCenter.gridRInversion[KernelSlider.getValue()];
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the noise slider and refresh button
  /////////////////////////////////////////////////////////////////////////////
  private void initNoiseButtons() {

    double RNoiseLevel = dataCenter.RNoiseLevel;
    boolean RNoiseSelected = dataCenter.RNoiseSelected;
    
    NoiseTextField = new JTextField(String.format("%.4e",RNoiseLevel));
    NoiseTextField.setEnabled(RNoiseSelected);
    NoiseTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkNoiseTextField()) {
          if (!paramModified) return;
          if (my_inversion.canInvert) {
            my_inversion.RLSinversion();
            my_inversion.SOLAinversion();
            my_inversion.MOLAinversion();
            replotInvertedFunctions();
          }
        }
      }
    });

    NoiseSlider = new JSlider(JSlider.HORIZONTAL,-60,60,NoiseToSlider(RNoiseLevel));
    NoiseSlider.setEnabled(RNoiseSelected);
    NoiseSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double RNoiseLevel = SliderToNoise();
        NoiseTextField.setText(String.format("%.4e",RNoiseLevel));
        if ((!calculateWhileSliding.isSelected())&&
            (NoiseSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.RNoiseLevel = RNoiseLevel;
        if (my_inversion.canInvert) {
          my_inversion.RLSinversion();
          my_inversion.SOLAinversion();
          my_inversion.MOLAinversion();
          replotInvertedFunctions();
        }
      }
    });

    NoiseRefresh = new JButton("New noise array");
    NoiseRefresh.setEnabled(RNoiseSelected);
    NoiseRefresh.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (my_inversion.canInvert) {
          my_inversion.initNoise();
          my_inversion.RLSinversion();
          my_inversion.SOLAinversion();
          my_inversion.MOLAinversion();
          replotInvertedFunctions();
        }
      }
    });
    
    NoiseCheckBox = new JCheckBox("Noise",RNoiseSelected);
    NoiseCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.RNoiseSelected = NoiseCheckBox.isSelected();
        NoiseTextField.setEnabled(NoiseCheckBox.isSelected());
        NoiseSlider.setEnabled(NoiseCheckBox.isSelected());
        NoiseRefresh.setEnabled(NoiseCheckBox.isSelected());
        if (my_inversion.canInvert) {
          my_inversion.RLSinversion();
          my_inversion.SOLAinversion();
          my_inversion.MOLAinversion();
          replotInvertedFunctions();
        }
      }
    });
    NoiseTitle = new JPanel();
    NoiseTitle.setLayout(new GridLayout(1,2));
    NoiseTitle.add(NoiseCheckBox);
    NoiseTitle.add(NoiseTextField);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the RLS slider
  /////////////////////////////////////////////////////////////////////////////
  private void initRLSslider() {
    double RLambda = dataCenter.RLambda;
    RLStitle = new JLabel("RLS",SwingConstants.CENTER);
    LambdaSlider = new JSlider(JSlider.HORIZONTAL,-180,20,LambdaToSlider(RLambda));
    LambdaLabel = new JLabel(" Lambda: ");
    
    LambdaTextField = new JTextField(String.format("%.4e",RLambda));
    LambdaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkLambdaTextField()) {
          if (!paramModified) return;
          if (my_inversion.canInvert) {
            my_inversion.RLSinversion();
            replotInvertedFunctions();
            replotAveragingKernels();
          }
        }
      }
    });
    
    LambdaTitle = new JPanel();
    LambdaTitle.setLayout(new GridLayout(1,2));
    LambdaTitle.add(LambdaLabel);
    LambdaTitle.add(LambdaTextField);
    
    LambdaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double RLambda = SliderToLambda();
        LambdaTextField.setText(String.format("%.4e",RLambda));
        if ((!calculateWhileSliding.isSelected())&&
            (LambdaSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.RLambda = RLambda;
        if (my_inversion.canInvert) {
          my_inversion.RLSinversion();
          replotInvertedFunctions();
          replotAveragingKernels();
        }
      }
    });
  }
    
  /////////////////////////////////////////////////////////////////////////////
  // Setting up the SOLA sliders
  /////////////////////////////////////////////////////////////////////////////
  private void initSOLAsliders() {
    double RTheta = dataCenter.RTheta;
    SOLAtitle = new JLabel("SOLA",SwingConstants.CENTER);
    ThetaLabel = new JLabel(" Theta: ");
    ThetaTextField = new JTextField(String.format("%.4e",RTheta));
    ThetaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkThetaTextField()) {
          if (!paramModified) return;
          if (my_inversion.canInvert) {
            my_inversion.SOLAinversion();
            replotInvertedFunctions();
            replotAveragingKernels();
          }
        }
      }
    });

    ThetaTitle = new JPanel();
    ThetaTitle.setLayout(new GridLayout(1,2));
    ThetaTitle.add(ThetaLabel);
    ThetaTitle.add(ThetaTextField);

    ThetaSlider = new JSlider(JSlider.HORIZONTAL,-200,0,ThetaToSlider(RTheta));
    ThetaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double RTheta = SliderToTheta();
        ThetaTextField.setText(String.format("%.4e",RTheta));
        if ((!calculateWhileSliding.isSelected())&&
            (ThetaSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.RTheta = RTheta;
        if (my_inversion.canInvert) {
          my_inversion.SOLAinversion();
          replotInvertedFunctions();
          replotAveragingKernels();
        }
      }
    });
        
    double RDelta = dataCenter.RDelta;
    DeltaSlider = new JSlider(JSlider.HORIZONTAL,-200,0,DeltaToSlider(RDelta));
    DeltaLabel = new JLabel(" Delta: ");
    DeltaTextField = new JTextField(String.format("%.4e",RDelta));
    DeltaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkDeltaTextField()) {
          if (!paramModified) return;
          // make sure a model is loaded (because you need c)
          if (my_inversion.canInvert&&!dataCenter.modelLoaded) {
            JOptionPane.showMessageDialog(
              DeltaTextField, // parent component
              "No model loaded.  Unable to proceed",
              "Error",
              JOptionPane.ERROR_MESSAGE
            );
            return;
          }
          if (my_inversion.canInvert) {
            my_inversion.initDeltaSOLA();
            my_inversion.SOLAinversion();
            replotInvertedFunctions();
            replotAveragingKernels();
          }
        }
      }
    });

    DeltaTitle = new JPanel();
    DeltaTitle.setLayout(new GridLayout(1,2));
    DeltaTitle.add(DeltaLabel);
    DeltaTitle.add(DeltaTextField);

    DeltaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double RDelta = SliderToDelta();
        DeltaTextField.setText(String.format("%.4e",RDelta));
        if (!calculateWhileSliding.isSelected()&&
            DeltaSlider.getModel().getValueIsAdjusting()) return;
        dataCenter.RDelta = RDelta;
        // make sure a model is loaded (because you need c)
        if (my_inversion.canInvert&&!dataCenter.modelLoaded) {
          JOptionPane.showMessageDialog(
            DeltaSlider, // parent component
            "No model loaded.  Unable to proceed",
            "Error",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }
        if (my_inversion.canInvert) {
          my_inversion.initDeltaSOLA();
          my_inversion.SOLAinversion();
          replotInvertedFunctions();
          replotAveragingKernels();
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Setting up the MOLA slider
  /////////////////////////////////////////////////////////////////////////////
  private void initMOLAslider() {
    double RThetaM = dataCenter.RThetaM;
    MOLAtitle = new JLabel("MOLA",SwingConstants.CENTER);
    MOLAThetaLabel = new JLabel(" Theta: ");
    MOLAThetaTextField = new JTextField(String.format("%.4e",RThetaM));
    MOLAThetaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkMOLAThetaTextField()) {
          if (!paramModified) return;
          if (my_inversion.canInvert) {
            my_inversion.MOLAinversion();
            replotInvertedFunctions();
            replotAveragingKernels();
          }
        }
      }
    });

    MOLAThetaTitle = new JPanel();
    MOLAThetaTitle.setLayout(new GridLayout(1,2));
    MOLAThetaTitle.add(MOLAThetaLabel);
    MOLAThetaTitle.add(MOLAThetaTextField);

    MOLAThetaSlider = new JSlider(JSlider.HORIZONTAL,-200,0,ThetaToSlider(RThetaM));
    MOLAThetaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double RThetaM = SliderToThetaM();
        MOLAThetaTextField.setText(String.format("%.4e",RThetaM));
        if ((!calculateWhileSliding.isSelected())&&
            (MOLAThetaSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.RThetaM = RThetaM;
        if (my_inversion.canInvert) {
          my_inversion.MOLAinversion();
          replotInvertedFunctions();
          replotAveragingKernels();
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Setting up the Kernel slider:   
  /////////////////////////////////////////////////////////////////////////////
  private void initKernelSlider() {
    KernelLabel = new JLabel(" Position, r/R: ");
    KernelTextField = new JTextField(String.format("%.4f",0.0));
    KernelTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkKernelTextField()) {
          if (!paramModified) return;
          replotAveragingKernels();
        }
      }
    });

    KernelTitle = new JPanel();
    KernelTitle.setLayout(new GridLayout(1,2));
    KernelTitle.add(KernelLabel);
    KernelTitle.add(KernelTextField);

    // VERY IMPORTANT: the number of positions on the Kernel Slider MUST
    // correspond to the number of points in gridInversion
    KernelSlider = new JSlider(JSlider.HORIZONTAL,0,dataCenter.n_gridR-1,0);
    KernelSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        KernelTextField.setText(String.format("%.4f",SliderToXKernel()));
        if (!calculateWhileSliding.isSelected()&&
            KernelSlider.getModel().getValueIsAdjusting()) return;
        replotAveragingKernels();
      }
    });

    NGridLabel = new JLabel(" # of grid points: ");
    NGridTextField = new JTextField(String.format("%d",dataCenter.n_gridR));
    NGridTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkNGridTextField()) {
          if (!paramModified) return;
          if (!dataCenter.rotationInversionUpToDate||my_inversion.canInvert) {
            updateInversion(NGridTextField);
            replotInvertedFunctions();
            replotAveragingKernels();
          }
        }
      }
    });
    NGridPanel = new JPanel();
    NGridPanel.setLayout(new GridLayout(1,2));
    NGridPanel.add(NGridLabel);
    NGridPanel.add(NGridTextField);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Various utility methods which check the entries in the different text
  // fields and returns true if these entries are as they should be.
  // If they are not suitable, an error message will be posted.  Finally,
  // the different sliders will be adjusted according to these entries.
  /////////////////////////////////////////////////////////////////////////////
  private boolean checkNoiseTextField() {
    Double Noise = Utilities.readDouble(NoiseTextField.getText(),
                   "noise level",0.0,1e30,NoiseTextField);
    if (Noise == null) {
      NoiseTextField.setText(String.format("%.4e",dataCenter.RNoiseLevel));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(NoiseSlider, NoiseToSlider(Noise));
      NoiseTextField.setText(String.format("%.4e",Noise));
      paramModified = (dataCenter.RNoiseLevel != Noise.doubleValue());
      dataCenter.RNoiseLevel = Noise.doubleValue();
      return true;
    }
  }

  private boolean checkLambdaTextField() {
    Double RLambda = Utilities.readDouble(LambdaTextField.getText(),
                   "value of Lambda",0.0,1e30,LambdaTextField);
    if (RLambda == null) {
      LambdaTextField.setText(String.format("%.4e",dataCenter.RLambda));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(LambdaSlider, LambdaToSlider(RLambda));
      LambdaTextField.setText(String.format("%.4e",RLambda));
      paramModified = (dataCenter.RLambda != RLambda.doubleValue());
      dataCenter.RLambda = RLambda.doubleValue();
      return true;
    }
  }

  private boolean checkMOLAThetaTextField() {
    Double RThetaM = Utilities.readDouble(MOLAThetaTextField.getText(),
                   "value of Theta(MOLA)",0.0,Math.PI/2.0,MOLAThetaTextField);
    if (RThetaM == null) {
      MOLAThetaTextField.setText(String.format("%.4e",dataCenter.RThetaM));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(MOLAThetaSlider, ThetaToSlider(RThetaM));
      MOLAThetaTextField.setText(String.format("%.4e",RThetaM));
      paramModified = (dataCenter.RThetaM != RThetaM.doubleValue());
      dataCenter.RThetaM = RThetaM.doubleValue();
      return true;
    }
  }

  private boolean checkThetaTextField() {
    Double RTheta = Utilities.readDouble(ThetaTextField.getText(),
                   "value of Theta(SOLA)",0.0,Math.PI/2.0,ThetaTextField);
    if (RTheta == null) {
      ThetaTextField.setText(String.format("%.4e",dataCenter.RTheta));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(ThetaSlider, ThetaToSlider(RTheta));
      ThetaTextField.setText(String.format("%.4e",RTheta));
      paramModified = (dataCenter.RTheta != RTheta.doubleValue());
      dataCenter.RTheta = RTheta.doubleValue();
      return true;
    }
  }

  private boolean checkDeltaTextField() {
    Double RDelta = Utilities.readDouble(DeltaTextField.getText(),
                   "value of Delta(SOLA)",0.0,1e2,DeltaTextField);
    if (RDelta == null) {
      DeltaTextField.setText(String.format("%.4e",dataCenter.RDelta));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(DeltaSlider, DeltaToSlider(RDelta));
      DeltaTextField.setText(String.format("%.4e",RDelta));
      paramModified = (dataCenter.RDelta != RDelta.doubleValue());
      dataCenter.RDelta = RDelta.doubleValue();
      return true;
    }
  }

  private boolean checkKernelTextField() {
    // checking the bounds is useless - I will be modifying the value
    // anyways afterwards:
    Double XKernel = Utilities.readDouble(KernelTextField.getText(),
                     "position value",Double.NEGATIVE_INFINITY,
                     Double.POSITIVE_INFINITY,KernelTextField);
    if (XKernel == null) {
      KernelTextField.setText(String.format("%.4f",SliderToXKernel()));
      paramModified = false;
      return false;
    } else {
      double oldValue = SliderToXKernel();
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(KernelSlider, XKernelToSlider(XKernel));
      double newValue = SliderToXKernel();
      KernelTextField.setText(String.format("%.4f",newValue));
      paramModified = (oldValue != newValue);
      return true;
    }
  }

  private boolean checkNGridTextField() {
    Integer NGrid = Utilities.readInt(NGridTextField.getText(), "number of grid points",
                            2,9999,NGridTextField);
    if (NGrid == null) {
      NGridTextField.setText(String.format("%d",dataCenter.n_gridR));
      paramModified = false;
      return false;
    } else {
      Utilities.ChangeJSliderMaximum(KernelSlider,NGrid-1);
      paramModified = (dataCenter.n_gridR != NGrid.intValue());
      dataCenter.n_gridR = NGrid.intValue();
      return true;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise buttons for the selection of the numerical method.
  /////////////////////////////////////////////////////////////////////////////
  private void initIntegrationMethodComboBox() {
    integrationMethodComboBox = new JComboBox();
    integrationMethodComboBox.addItem("Sliding window");
    integrationMethodComboBox.addItem("Chebyshev");
    if (my_inversion.integrationMethod == RotationInversion.Chebyshev)
      integrationMethodComboBox.setSelectedItem("Chebyshev");
    if (my_inversion.integrationMethod == RotationInversion.SlidingWindow)
      integrationMethodComboBox.setSelectedItem("Sliding window");
    integrationMethodComboBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        int newIntegrationMethod = my_inversion.integrationMethod;
        String newChoice = (String)integrationMethodComboBox.getSelectedItem();
        if (newChoice.equals("Chebyshev")) newIntegrationMethod = RotationInversion.Chebyshev;
        if (newChoice.equals("Sliding window")) newIntegrationMethod = RotationInversion.SlidingWindow;
        if (newIntegrationMethod != my_inversion.integrationMethod) {
          my_inversion.integrationMethod = newIntegrationMethod;
          if (!dataCenter.rotationInversionUpToDate||my_inversion.canInvert) {
            updateInversion(integrationMethodComboBox);
            replotInvertedFunctions();
            replotAveragingKernels();
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Setting up the refresh button and the RLS, SOLA and Target function check
  // boxes.
  /////////////////////////////////////////////////////////////////////////////
  private void initCheckBoxes() {
    Refresh = new JButton("Update Inversion");
    RLSCheckBox = new JCheckBox("<html> <font color='red'> RLS (red lines) </font> </html>",true);
    SOLACheckBox = new JCheckBox("<html> <font color='blue'> SOLA (blue lines) </font> </html>",true);
    MOLACheckBox = new JCheckBox("<html> <font color='green'> MOLA (green lines) </font> </html>",true);
    TargetFunctionCheckBox = new JCheckBox("Target function (black line)",true);
    showErrorBars = new JCheckBox("Show error bars",true);

    Refresh.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (dataCenter.RNoiseSelected) {
          if (!checkNoiseTextField()) return;
        }
        if (RLSCheckBox.isSelected()) {
          if (!checkLambdaTextField()) return;
        }
        if (SOLACheckBox.isSelected()) {
          if (!checkThetaTextField()) return;
          if (!checkDeltaTextField()) return;
        }
        if (MOLACheckBox.isSelected()) {
          if (!checkMOLAThetaTextField()) return;
        }
        if (!checkKernelTextField()) return;
        if (!checkNGridTextField()) return;
        updateInversion(Refresh);
        replotInvertedFunctions();
        replotAveragingKernels();
      }
    });

    RLSCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        RLStitle.setEnabled(RLSCheckBox.isSelected());
        LambdaLabel.setEnabled(RLSCheckBox.isSelected());
        LambdaTextField.setEnabled(RLSCheckBox.isSelected());
        LambdaSlider.setEnabled(RLSCheckBox.isSelected());
        replotInvertedFunctions();
        replotAveragingKernels();
      }
    });
          
    SOLACheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        SOLAtitle.setEnabled(SOLACheckBox.isSelected());
        ThetaLabel.setEnabled(SOLACheckBox.isSelected());
        ThetaTextField.setEnabled(SOLACheckBox.isSelected());
        ThetaSlider.setEnabled(SOLACheckBox.isSelected());
        DeltaLabel.setEnabled(SOLACheckBox.isSelected());
        DeltaTextField.setEnabled(SOLACheckBox.isSelected());
        DeltaSlider.setEnabled(SOLACheckBox.isSelected());
        replotInvertedFunctions();
        replotAveragingKernels();
      }
    });

    MOLACheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        MOLAtitle.setEnabled(MOLACheckBox.isSelected());
        MOLAThetaLabel.setEnabled(MOLACheckBox.isSelected());
        MOLAThetaTextField.setEnabled(MOLACheckBox.isSelected());
        MOLAThetaSlider.setEnabled(MOLACheckBox.isSelected());
        replotInvertedFunctions();
        replotAveragingKernels();
      }
    });

    TargetFunctionCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        replotInvertedFunctions();
      }
    });
    
    showErrorBars.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        replotInvertedFunctions();
      }
    });
          
  }

  /////////////////////////////////////////////////////////////////////////////
  // init buttons which deal with a target function
  /////////////////////////////////////////////////////////////////////////////
  private void initTargetFunctionButtons() {
    LoadTarget = new JButton("Load target \u03A9 function");
    LoadTarget.setEnabled(!dataCenter.isApplet);
    LoadTarget.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showOpenDialog(LoadTarget);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            loadRotaTarget(aFile);
            replotInvertedFunctions();
            dataCenter.currentDirectory = aFile.getParentFile();
          } else {
            JOptionPane.showMessageDialog(
              LoadTarget,
              "File \""+aFile.getName()+"\" doesn't exist.",
              "File not found",
              JOptionPane.WARNING_MESSAGE
            );
          }
        }
      }
    });
    
    ClearTarget = new JButton("Clear target \u03A9 function");
    ClearTarget.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.targetRota = null;
        dataCenter.gridTargetRota = null;
        dataCenter.targetRotaLoaded = false;
        replotInvertedFunctions();
      }
    });

    CalculateRS = new JButton("Find R.S. from target \u03A9 function");
    CalculateRS.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.frequenciesData.calculateRS(CalculateRS);
        dataCenter.frequenciesTable.updateUI();
      }
    });
  }

  /***************************************************************************
  *  This updates the inversion.
  ***************************************************************************/
  public void updateInversion(JComponent c) {
    my_inversion.init(c);
      numberOfKernels.setText(String.format(" This inversion uses %d kernels",
                              my_inversion.n_inputKernels));
    // no need to give an error message if canInvert = false, since
    // my_inversion.init(c) already gives error messages.
    if (my_inversion.canInvert) {
      my_inversion.RLSinversion();
      my_inversion.SOLAinversion();
      my_inversion.MOLAinversion();
    }
  }

  /************************************************************************
  * This initialises the program.
  ************************************************************************/
  public void initProgram() {
    graphInvertedFunction = new IPlot("Inversion results","Position, r/R", "Profile",6);
    graphInvertedFunction.setIOenabled(!dataCenter.isApplet);
    graphAveragingKernel = new IPlot("Averaging kernels","Position, r/R", "Kernel",3);
    graphAveragingKernel.setIOenabled(!dataCenter.isApplet);
    graphContainer = new JPanel();
    graphContainer.setLayout(new GridLayout(2,1));
    graphContainer.add(graphInvertedFunction);
    graphContainer.add(graphAveragingKernel);

    setLayout(new BorderLayout());
    add("Center", graphContainer);

    // init the inversion
    my_inversion = new RotationInversion(dataCenter);

    calculateWhileSliding = new JCheckBox("Calculate while using sliders",false);
    initNoiseButtons();
    initRLSslider();
    initSOLAsliders();
    initMOLAslider();
    initKernelSlider();
    initIntegrationMethodComboBox();
    initCheckBoxes();
    initTargetFunctionButtons();
    numberOfKernels = new JLabel(String.format(" This inversion uses %d kernels",
                                 my_inversion.n_inputKernels));
    
    SliderContainer = new JPanel();
    SliderContainer.setLayout(new GridLayout(36,1));
    SliderContainer.add(numberOfKernels);
    SliderContainer.add(Refresh);
    SliderContainer.add(RLSCheckBox);
    SliderContainer.add(SOLACheckBox);
    SliderContainer.add(MOLACheckBox);
    SliderContainer.add(TargetFunctionCheckBox);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(showErrorBars);
    SliderContainer.add(calculateWhileSliding);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(NoiseTitle);
    SliderContainer.add(NoiseSlider);
    SliderContainer.add(NoiseRefresh);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(RLStitle);
    SliderContainer.add(LambdaTitle);
    SliderContainer.add(LambdaSlider);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(SOLAtitle);
    SliderContainer.add(ThetaTitle);
    SliderContainer.add(ThetaSlider);
    SliderContainer.add(DeltaTitle);
    SliderContainer.add(DeltaSlider);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(MOLAtitle);
    SliderContainer.add(MOLAThetaTitle);
    SliderContainer.add(MOLAThetaSlider);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(KernelTitle);
    SliderContainer.add(KernelSlider);
    SliderContainer.add(NGridPanel);
    SliderContainer.add(integrationMethodComboBox);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(LoadTarget);
    SliderContainer.add(ClearTarget);
    SliderContainer.add(CalculateRS);
    add("East",SliderContainer);
       
    graphInvertedFunction.repaint();
    graphAveragingKernel.repaint();
  }

  /////////////////////////////////////////////////////////
  // load a file with a target rota function:
  /////////////////////////////////////////////////////////
  private void loadRotaTarget(File aFile){
    BufferedReader f;
    String line;
    Boolean skipLine = false;
    ArrayList<Double> x = new ArrayList<Double>();
    ArrayList<Double> y = new ArrayList<Double>();
    try {
      f = Utilities.OpenFile(aFile.getPath());
      while (true) {
        line = f.readLine();     // read a line of text
        if (line == null) break; // test for the end of the buffer
        line = Utilities.trim(line);       // remove comments
        StringTokenizer t = new StringTokenizer(line," \t,");
        // only read lines with EXACTLY 2 numbers
        if (t.countTokens() == 2) {
          x.add(Utilities.parseDouble(t.nextToken()));
          y.add(Utilities.parseDouble(t.nextToken()));
        } else {
          if (t.countTokens() != 0) skipLine = true;
        }
      }
      f.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "Unable to read file \""+aFile.getName()+"\"",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    if (x.size() == 0) {
      JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "No data in file",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    // perform a basic check on the grid:
    for(int i=1; i<x.size(); i++) {
      if (x.get(i).compareTo(x.get(i-1)) <= 0) {
        JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "Please define the target function\n"+
                    "on a strictly ascending grid.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
        return;
      }
    }
    
    if (skipLine) {
      JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "Some lines were skipped\nwhen reading the file.",
                    "Warning",
                    JOptionPane.WARNING_MESSAGE
                  );
    }
    dataCenter.gridTargetRota = new double[x.size()];
    dataCenter.targetRota = new double[y.size()];
    for(int i=0; i<x.size(); i++) {
      dataCenter.gridTargetRota[i] = (double)(x.get(i));
      dataCenter.targetRota[i] = (double)(y.get(i));
    }
    dataCenter.targetRotaLoaded = true;
  }

  /***************************************************************************
  *  Plot the inverted and target rotation functions
  ***************************************************************************/
  private void replotInvertedFunctions() {
    if (TargetFunctionCheckBox.isSelected()&&dataCenter.targetRotaLoaded)
      graphInvertedFunction.replaceDataSet(0,dataCenter.gridTargetRota,dataCenter.targetRota,Plot.solid);
    else
      graphInvertedFunction.nullify(0);

    if (my_inversion.canInvert) {
      int n_grid = dataCenter.gridRInversion.length;
      double[] aux = new double[n_grid];

      if (RLSCheckBox.isSelected()) { 
        graphInvertedFunction.replaceDataSet(1,dataCenter.gridRInversion,my_inversion.f_invertedRLS,
                                             Plot.solid,Color.red);
      } else {
        graphInvertedFunction.nullify(1);
      }

      if (RLSCheckBox.isSelected()&&showErrorBars.isSelected()) {
        for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f_invertedRLS[i]+my_inversion.RLS_errorBars[i];
        graphInvertedFunction.replaceDataSet(2,dataCenter.gridRInversion,aux,Plot.dotted,Color.red);
        for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f_invertedRLS[i]-my_inversion.RLS_errorBars[i];
        graphInvertedFunction.replaceDataSet(3,dataCenter.gridRInversion,aux,Plot.dotted,Color.red);
      } else {
        graphInvertedFunction.nullify(2);
        graphInvertedFunction.nullify(3);
      }

      if (SOLACheckBox.isSelected()) {
        if (showErrorBars.isSelected()) {
          graphInvertedFunction.replaceDataSet(4,my_inversion.SOLA_newGrid,my_inversion.f_invertedSOLA,
                                               Plot.noLine,Color.blue);
          graphInvertedFunction.replaceLeftError(4,my_inversion.SOLA_leftErrorBars);
          graphInvertedFunction.replaceRightError(4,my_inversion.SOLA_rightErrorBars);
          graphInvertedFunction.replaceLowerError(4,my_inversion.SOLA_errorBars);
          graphInvertedFunction.replaceUpperError(4,my_inversion.SOLA_errorBars);
          graphInvertedFunction.setHErrorValue(4,true);
          graphInvertedFunction.setVErrorValue(4,true);
        } else {
          graphInvertedFunction.replaceDataSet(4,my_inversion.SOLA_newGrid,my_inversion.f_invertedSOLA,
                                               Plot.solid,Color.blue);
          graphInvertedFunction.setHErrorValue(4,false);
          graphInvertedFunction.setVErrorValue(4,false);
        }
      } else {
        graphInvertedFunction.nullify(4);
      }

      if (MOLACheckBox.isSelected()) {
        if (showErrorBars.isSelected()) {
          graphInvertedFunction.replaceDataSet(5,my_inversion.MOLA_newGrid,my_inversion.f_invertedMOLA,
                                               Plot.noLine,DarkGreen);
          graphInvertedFunction.replaceLeftError(5,my_inversion.MOLA_leftErrorBars);
          graphInvertedFunction.replaceRightError(5,my_inversion.MOLA_rightErrorBars);
          graphInvertedFunction.replaceLowerError(5,my_inversion.MOLA_errorBars);
          graphInvertedFunction.replaceUpperError(5,my_inversion.MOLA_errorBars);
          graphInvertedFunction.setHErrorValue(5,true);
          graphInvertedFunction.setVErrorValue(5,true);
        } else {
          graphInvertedFunction.replaceDataSet(5,my_inversion.MOLA_newGrid,my_inversion.f_invertedMOLA,
                                               Plot.solid,DarkGreen);
          graphInvertedFunction.setHErrorValue(5,false);
          graphInvertedFunction.setVErrorValue(5,false);
        }
      } else {
        graphInvertedFunction.nullify(5);
      }
    } else {
      for(int i=1; i<=5; i++) graphInvertedFunction.nullify(i);
    }

    graphInvertedFunction.findRange();

    if (my_inversion.canInvert) {
      graphInvertedFunction.setXmin(dataCenter.gridRInversion[0]);
      graphInvertedFunction.setXmax(dataCenter.gridRInversion[dataCenter.gridRInversion.length-1]);
    }
    graphInvertedFunction.repaint();
  }

  /***************************************************************************
  *  Plot averaging kernels
  ***************************************************************************/
  private void replotAveragingKernels() {
    // quick exit if possible:
    if (!my_inversion.canInvert) {
      graphAveragingKernel.nullify();
      graphAveragingKernel.repaint();
      return;
    }
    
    // number of grid points
    int n_grid = dataCenter.gridRInversion.length;
    int n_big_grid = dataCenter.grid.length;
    
    // this specifies which kernel to plot:
    int iKernel = KernelSlider.getValue();
    
    // update KernelTextField (just in case):
    KernelTextField.setText(String.format("%.4f",SliderToXKernel()));

    // RLS averaging kernel:
    if (RLSCheckBox.isSelected()) {
      double[] y = my_inversion.findAveragingKernelRLS(iKernel);
      graphAveragingKernel.replaceDataSet(0,dataCenter.grid,y,Plot.solid,Color.red);
    } else {
      graphAveragingKernel.nullify(0);
    }
    
    // SOLA averaging kernel:
    if (SOLACheckBox.isSelected()) {
      double[] y = my_inversion.findAveragingKernelSOLA(iKernel);
      graphAveragingKernel.replaceDataSet(1,dataCenter.grid,y,Plot.solid,Color.blue);
    } else {
      graphAveragingKernel.nullify(1);
    }

    // MOLA averaging kernel:
    if (MOLACheckBox.isSelected()) {
      double[] y = my_inversion.findAveragingKernelMOLA(iKernel);
      graphAveragingKernel.replaceDataSet(2,dataCenter.grid,y,Plot.solid,DarkGreen);
    } else {
      graphAveragingKernel.nullify(2);
    }

    graphAveragingKernel.findRange();
    graphAveragingKernel.repaint();
  }
}
