/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import inversion_applet.Numerics;
import org.netlib.lapack.Dsytrf;
import org.netlib.lapack.Dsytrs;
import org.netlib.lapack.Ilaenv;
import java.util.Random;
import java.util.ArrayList;
import java.util.Vector;
import javax.swing.JOptionPane;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTable;

/*************************************************************************
**
**    StructuralInversion
**                                              Version 2.1, September 2014
**
*************************************************************************/

public class StructuralInversion { 

  private DataCenter dataCenter;
  protected boolean canInvert;   // after initialisation, if the provided information
                                 // is insufficient or erroneous, it is not possible
                                 // to do an inversion, in which case, canInvert = false
                                 // Otherwise, canInvert = true
  protected int n_inputKernels;  // the number of kernels used in the inversion

  // variables and constants used in the choice of method:
  protected static final int Chebyshev = 0;
  protected static final int SlidingWindow = 1; // the default
  protected int integrationMethod = SlidingWindow;
  
  // various inversion results
  public double[] f1_invertedRLS;       // function given by the RLS inversion (f_inverted[n_grid])
  public double[] f2_invertedRLS;       // function given by the RLS inversion (f_inverted[n_grid])
  public double[] RLS_errorBars1;       // this gives the uncertainty in the inversion, based on
                                        // the error bars of the frequency shifts
  public double[] RLS_errorBars2;       // this gives the uncertainty in the inversion, based on
                                        // the error bars of the frequency shifts
  public double[] f1_invertedSOLA;      // function given by the SOLA inversion (f_inverted[n_grid])
  public double[] f2_invertedSOLA;      // function given by the SOLA inversion (f_inverted[n_grid])
  public double[] SOLA_errorBars1;      // this gives the uncertainty in the inversion, based on
                                        // the error bars of the frequency shifts
  public double[] SOLA_errorBars2;      // this gives the uncertainty in the inversion, based on
                                        // the error bars of the frequency shifts
  public double[] SOLA_newGrid1;        // this gives the 2nd quartile point for each averaging kernel
  public double[] SOLA_newGrid2;        // this gives the 2nd quartile point for each averaging kernel
  public double[] SOLA_leftErrorBars1;  // this gives the 1st quartile point for each averaging kernel
  public double[] SOLA_leftErrorBars2;  // this gives the 1st quartile point for each averaging kernel
  public double[] SOLA_rightErrorBars1; // this gives the 3re quartile point for each averaging kernel
  public double[] SOLA_rightErrorBars2; // this gives the 3re quartile point for each averaging kernel
  
  // various intermediate matrices used in the inverstions:
  private double[] fitDataMatRLS;           // matrix used for fitting the data in a RLS inversion
  private double[] regularisationMatRLS;    // matrix used for regularisation in a RLS inversion
  private double[] kernelMatRLS;            // matrix used for calculating output kernels
  private double[] RLSmat;                  // matrix which contains the coefficients from the inversion
  private double[] kernelMat1SOLA;          // matrix used for fitting the data in a SOLA inversion
  private double[] kernelMat2SOLA;          // matrix used for fitting the data in a SOLA inversion
  private double[] crossTermMat1SOLA;       // matrix used for reducing the cross-term kernel in a SOLA inversion
  private double[] crossTermMat2SOLA;       // matrix used for reducing the cross-term kernel in a SOLA inversion
  private double[] supplMat1SOLA;           // matrix with term relating to supplementary conditions
  private double[] supplMat2SOLA;           // matrix with term relating to supplementary conditions
  private double[] regularisationMatSOLA;   // matrix used for regularisation in a SOLA inversion
  private double[] targetKernelMat1SOLA;    // matrix used for calculating the target kernels
  private double[] targetKernelMat2SOLA;    // matrix used for calculating the target kernels
  private double[] SOLAmat1;                // matrix which contains the coefficients from the inversion
  private double[] SOLAmat2;                // matrix which contains the coefficients from the inversion
  private double[][] averagingKernels1SOLA; // matrix which contains the averaging kernels from the inversion
  private double[][] averagingKernels2SOLA; // matrix which contains the averaging kernels from the inversion
  private int matSizeRLS, matSizeSOLA, matSizeTargetSOLA;   // sizes of certain matrices in both types of inversions
  private org.netlib.util.intW info = new org.netlib.util.intW(2); // necessary for lapack
  private ArrayList<Integer> intersection = new ArrayList<Integer>();
  private ArrayList<Double>  dataValues   = new ArrayList<Double>();
  private ArrayList<Double>  dataErrors   = new ArrayList<Double>();
  private ArrayList<Double>  freq   = new ArrayList<Double>();
  private ArrayList<Double>  Enl    = new ArrayList<Double>();
  private double[][] struct1Kernel;
  private double[][] struct2Kernel;
  private double[][] surfPoly;
  private int lwork;  // an integer used in the matrix solvers
  private double[] weights;
  private double[] iGrid;
  private double rspan;     // span of dataCenter.grid
  
  // different quantities used in setting up B-splines
  private int order = 4;                 // order of the B-splines (4 = cubic b-splines)
  private double[] knots;                // knots used to determine the B-spline basis
  
  /********************************************************
  *  A constructor which does the initialisation
  ********************************************************/
  public StructuralInversion(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
     
    // At the beginning, no data is loaded, so no inversion
    // can be done.  Therefore the inversion is up to date.
    dataCenter.structInversionUpToDate = true;
    this.canInvert = false;
    
    // This determines the blocksize to be used in the symmetric 
    // matrix solvers. I will just put some average matrix size
    // rather than an exact number.
    this.lwork = Ilaenv.ilaenv(1,"DSYTRF","U",200,200,-1,-1);
  }

  ////////////////////////////////////////////////////////////
  /** This calls various initialisation methods */
  ////////////////////////////////////////////////////////////
  public void init(JComponent c) {
    dataCenter.structInversionUpToDate = true;
    n_inputKernels = 0;
    initIntersection(c);
    if (canInvert) {
      initGridInversion();
      initStructKernels();
      initNoise();
      initSurfPoly();
      initRLS();
      initSOLA();
      initDelta1SOLA();
      initDelta2SOLA();
    }
  }

  ////////////////////////////////////////////////////////////
  // This finds the intersection between the set of kernels
  // and the set of frequency shift data and stores it in
  // an ArrayList called intersection
  ////////////////////////////////////////////////////////////
  private void initIntersection(JComponent c) {

    canInvert = true;

    // make sure frequencies and associated error bars are loaded:
    if (dataCenter.nfreq == 0) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No frequencies and associated error bars loaded.\n"+
        "Unable to proceed.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      n_inputKernels = 0;
      canInvert = false;
      return;
    }
    
    // make sure a model is loaded:
    if (!dataCenter.modelLoaded) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No model loaded.  Unable to proceed",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canInvert = false;
      n_inputKernels = 0;
      return;
    }

    boolean dataWithZeroError = false;
    boolean incorrectDF = false;
    boolean duplicateKernels = false;

    // clear previous data:
    intersection.clear();
    dataValues.clear();
    dataErrors.clear();
    freq.clear();
    Enl.clear();

    // find index array and frequency shifts
    for(int i=0; i< dataCenter.modes.size();i++) {
      if (dataCenter.modes.get(i).FreqObs == null) continue;
      if (dataCenter.modes.get(i).ErrorFreq == null) continue;
      if (dataCenter.modes.get(i).kernel == null) continue;
      if (dataCenter.modes.get(i).isDuplicate) {
        duplicateKernels = true;
        continue;
      }
      if (dataCenter.modes.get(i).ErrorFreq < 0) {
        dataWithZeroError = true;
        continue;
      }
      if (dataCenter.modes.get(i).l < 0) {
        incorrectDF = true;
        continue;
      }
      
      intersection.add(i);
      dataValues.add(dataCenter.modes.get(i).FreqObs/dataCenter.modes.get(i).kernel.FreqTheo-1.0);
      dataErrors.add(dataCenter.modes.get(i).ErrorFreq/dataCenter.modes.get(i).kernel.FreqTheo);
      freq.add(dataCenter.modes.get(i).kernel.FreqDimless);
      Enl.add(dataCenter.modes.get(i).kernel.Enl);
    }

    n_inputKernels = intersection.size();

    if (n_inputKernels == 0) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No kernels/data from which to carry\n"+
        "out an inversion.  Unable to proceed.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canInvert = false;
      return;
    }
 
    // Test to see if all the relevant kernels have
    // been calculated and used in the inversion.
    if (n_inputKernels < dataCenter.nfreq) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some kernels might be missing from this inversion.\n"+
        "Pressing the \"Calculate kernels\" button on the\n"+
        "kernel tab would remedy this problem.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (dataWithZeroError) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the frequency shift data has\n"+
        "a 0 error bar, which is not allowed.\n"+
        "This data has consequently been ignored.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
      
    if (incorrectDF) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the frequency shift data seems to be erroneous:\n"+
        "  (\u2113 < 1 for n < 1) or (\u2113 < 0 for n > 0).\n"+
        "This data has consequently not been not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (duplicateKernels) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "There are duplicate kernels.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
  }

  ////////////////////////////////////////////////////////////
  // set up the grid on which to calculate the inverted
  // function:
  ////////////////////////////////////////////////////////////
  private void initGridInversion() {
    if (integrationMethod == Chebyshev) initGridInversionCheb();
    if (integrationMethod == SlidingWindow)  initGridInversionSplines();
  }

  ////////////////////////////////////////////////////////////
  // This initialises a Chebyshev collocation grid and spectral
  // integration weights to be used in the inversion.  For a
  // given number of grid points, it is more accurate than
  // initGridInversion() above.
  ////////////////////////////////////////////////////////////
  private void initGridInversionCheb() {
    int n_grid = dataCenter.n_gridS;
    rspan = dataCenter.grid[dataCenter.grid.length-1]-
            dataCenter.grid[0];

    dataCenter.gridSInversion = new double[n_grid];
    for(int i=0; i<n_grid;i++)
      dataCenter.gridSInversion[i] = 
         0.5*(1.0-Math.cos(Math.PI*(double)(2*i+1)/(double)(2*n_grid)))
        *rspan+dataCenter.grid[0];

    // initialise some integration weights:
    weights = Numerics.initWeightsCheb(n_grid);
    iGrid = dataCenter.gridSInversion;
    for(int i=0; i< n_grid; i++) weights[i]*=rspan;
  }

  ////////////////////////////////////////////////////////////
  // This initialises a grid to be used in the inversion.
  // For a given number of grid points, it is not as
  // accurate as initGridInversionCheb() below.
  ////////////////////////////////////////////////////////////
  private void initGridInversionSplines() {
    int n_grid = dataCenter.n_gridS;
    rspan = dataCenter.grid[dataCenter.grid.length-1]-
            dataCenter.grid[0];

    // obtain the sound velocity:
    double[] c = new double[dataCenter.c2.length];
    for(int i=0; i<c.length; i++) c[i] = Math.sqrt(dataCenter.c2[i]);

    // this is to avoid division by zero later on:
    double max_c = 0.0;
    for(int i=0; i<c.length; i++) if (max_c < c[i]) max_c = c[i];
    for(int i=0; i<c.length; i++) if ((c[i]/max_c) < 1e-3) c[i] = max_c*1e-3;

    // find the inverse sound velocity:
    for(int i=0; i<c.length; i++) c[i] = 1.0/Math.sqrt(c[i]);

    double[] CC = Numerics.findAntiDerivative(dataCenter.grid,c);
    double cInt = CC[CC.length-1]-CC[0];

    dataCenter.gridSInversion = new double[n_grid];
    int j = 0;
    double cTarget, lambda;
    for(int i=0; i<n_grid; i++) {
      cTarget = cInt*(double)(2*i+1)/(double)(2*n_grid);  // a target value
      while(CC[j]<cTarget) j++;
      lambda = (CC[j]-cTarget)/(CC[j]-CC[j-1]);
      dataCenter.gridSInversion[i] = dataCenter.grid[j-1]*lambda 
                                  + dataCenter.grid[j]*(1.0-lambda);
      j--;
    }

    // initialise some knots:
    knots = Numerics.find_knots(dataCenter.gridSInversion, order);

    // initialise some integration weights:
    iGrid = dataCenter.grid;
    weights = Numerics.findIntegrationWeights(iGrid);
  }

  ////////////////////////////////////////////////////////////
  // set up structural kernels before doing inversions:
  ////////////////////////////////////////////////////////////
  private void initStructKernels() {
    struct1Kernel = new double[n_inputKernels][];
    struct2Kernel = new double[n_inputKernels][];
    
    if (dataCenter.kernelType == 0) {
      for(int l=0; l<n_inputKernels; l++) {
        struct1Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.rhoc2Kernel;
        struct2Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.c2rhoKernel;
      }
      dataCenter.constantMass = 0; // the mass can be modified by the inversion
    } else if (dataCenter.kernelType == 1) {
      for(int l=0; l<n_inputKernels; l++) {
        struct1Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.rhoGamma1Kernel;
        struct2Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.c2rhoKernel;
      }
      dataCenter.constantMass = 0; // the mass can be modified by the inversion
    } else if (dataCenter.kernelType == 2) {
      for(int l=0; l<n_inputKernels; l++) {
        struct1Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.rhoc2Kernel;
        struct2Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.c2rhoKernel;
      }
      dataCenter.constantMass = 1; // this imposes keeping a constant mass on the inversion
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        struct1Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.rhoGamma1Kernel;
        struct2Kernel[l] = dataCenter.modes.get(intersection.get(l)).kernel.c2rhoKernel;
      }
      dataCenter.constantMass = 1; // this imposes keeping a constant mass on the inversion
    }
    if (integrationMethod == Chebyshev) {
      struct1Kernel = Numerics.interpolateList_r2(dataCenter.grid,dataCenter.gridSInversion,struct1Kernel,false);
      struct2Kernel = Numerics.interpolateList_r2(dataCenter.grid,dataCenter.gridSInversion,struct2Kernel,false);
    }
  }

  /********************************************************
  * Initialise the Noise array with a Gauss distribution:
  * Mean = 0.0; std. deviation= 1.0.
  ********************************************************/
  public void initNoise() {
    Random generator = new Random();
    dataCenter.SNoise = new double[intersection.size()];
    for(int l=0; l < intersection.size(); l++) {
      dataCenter.SNoise[l] = generator.nextGaussian();
    }
  }

  /********************************************************
  * This initialises the polynomial expansion used to
  * model the surface terms.
  ********************************************************/
  public void initSurfPoly() {
    surfPoly = new double[n_inputKernels][];
    
    for(int l=0; l<n_inputKernels; l++) {
      surfPoly[l] = Numerics.Legendre(freq.get(l),dataCenter.n_surf);
      for(int n=0; n<dataCenter.n_surf; n++)
        surfPoly[l][n]/=Enl.get(l);
    }
  }

  ////////////////////////////////////////////////////////////
  // Initialise various matrices used in a RLS inversion.
  ////////////////////////////////////////////////////////////
  protected void initRLS() {
    if (integrationMethod == Chebyshev) initRLScheb();
    if (integrationMethod == SlidingWindow)  initRLSsplines();
  }

  /********************************************************
  * Initialise various matrices used in a RLS inversion.
  ********************************************************/
  private void initRLScheb() {
    int n_grid = dataCenter.n_gridS;
    matSizeRLS = 2*n_grid+dataCenter.n_surf;
    double[] rho = null;
    int i, j, k, l, m;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      matSizeRLS++;
    if (dataCenter.constantMass==3)
      matSizeRLS+=2;
    if (dataCenter.constantMass!=0)
      rho = Numerics.interpolate1D_r2(dataCenter.grid,iGrid,dataCenter.rho,true);

    // initialise the matrix used for calculating the output kernels:
    kernelMatRLS = new double[matSizeRLS*n_inputKernels];
    for(l=0; l<n_inputKernels; l++){
      for(i=0; i<n_grid; i++) {
        kernelMatRLS[i+matSizeRLS*l]=
            weights[i]*struct1Kernel[l][i]/dataErrors.get(l);
        kernelMatRLS[i+n_grid+matSizeRLS*l]= 
            weights[i]*struct2Kernel[l][i]/dataErrors.get(l);
      }
      for(m=0; m<dataCenter.n_surf; m++) {
        kernelMatRLS[m+2*n_grid+matSizeRLS*l]=
            surfPoly[l][m]/dataErrors.get(l);
      }
    }

    // initialise matrix used for fitting the data:
    fitDataMatRLS = new double[matSizeRLS*matSizeRLS];
    for(i=0; i<(2*n_grid+dataCenter.n_surf); i++){
      for(j=i; j<(2*n_grid+dataCenter.n_surf); j++){
        for(l=0; l<n_inputKernels; l++){
          fitDataMatRLS[i+matSizeRLS*j]+=
              kernelMatRLS[i+matSizeRLS*l]
             *kernelMatRLS[j+matSizeRLS*l];
        }
      }
    }

    // supplementary constraints for keeping the mass constant after inversion:  
    if (dataCenter.constantMass == 1) {
      for(i=0; i<n_grid; i++) {
        fitDataMatRLS[i+matSizeRLS*(2*n_grid+dataCenter.n_surf)]=
          dataCenter.gridSInversion[i]*dataCenter.gridSInversion[i]*rho[i]*weights[i]*0.5;
      }
    }

    if (dataCenter.constantMass == 2) {
      for(i=0; i<n_grid; i++) {
        fitDataMatRLS[i+n_grid+matSizeRLS*(2*n_grid+dataCenter.n_surf)]=
          dataCenter.gridSInversion[i]*dataCenter.gridSInversion[i]*rho[i]*weights[i]*0.5;
      }
    }

    if (dataCenter.constantMass == 3) {
      for(i=0; i<n_grid; i++) {
        fitDataMatRLS[i+matSizeRLS*(2*n_grid+dataCenter.n_surf)]=
          dataCenter.gridSInversion[i]*dataCenter.gridSInversion[i]*rho[i]*weights[i]*0.5;
        fitDataMatRLS[i+n_grid+matSizeRLS*(2*n_grid+dataCenter.n_surf+1)]=
          dataCenter.gridSInversion[i]*dataCenter.gridSInversion[i]*rho[i]*weights[i]*0.5;
      }
    }

    // finish initialising the matrix used for calculating the output kernels:
    for(i=0; i<(2*n_grid+dataCenter.n_surf); i++) {
      for(l=0; l<n_inputKernels; l++) {
        kernelMatRLS[i+matSizeRLS*l]/=dataErrors.get(l);
      }
    }

    // initialise a derivation matrix:
    double[] derive = Numerics.initDDerive(n_grid);
    double cnst = 1.0/(rspan*rspan);
    for(i=0; i<(n_grid*n_grid); i++) derive[i]*=cnst;

    // initialise the regularisation matrix:
    regularisationMatRLS = new double[n_grid*n_grid];
    for(i=0; i<n_grid; i++){
      for(j=i; j<n_grid; j++){
        for(k=0; k<n_grid; k++){
          regularisationMatRLS[i+n_grid*j]+=
              weights[k]*derive[k+n_grid*i]
                        *derive[k+n_grid*j];
        }
      }
    }
  }

  /********************************************************
  * Initialise various matrices used in a RLS inversion.
  ********************************************************/
  private void initRLSsplines() {
    int n_grid = dataCenter.n_gridS;
    int n_big_grid = iGrid.length;
    matSizeRLS = 2*n_grid+dataCenter.n_surf;
    double lb = knots[0];
    double[] rho = null;
    double[] subknots = new double[order+1];
    double[] massArray = new double[n_grid];
    int i, j, k, l, m;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      matSizeRLS++;
    if (dataCenter.constantMass==3)
      matSizeRLS+=2;
    if (dataCenter.constantMass!=0)
      rho = Numerics.interpolate1D_r2(dataCenter.grid,iGrid,dataCenter.rho,true);

    // initialise the matrix used for calculating the output kernels:
    int istart = 0;
    double bs_value;
    kernelMatRLS = new double[matSizeRLS*n_inputKernels];
    for(i=0; i<n_grid; i++){
      for(j=0; j<= order; j++) subknots[j] = knots[i+j];
      while (iGrid[istart] < subknots[0]) istart++;
      j = istart;
      while ((j < n_big_grid)&&(iGrid[j] <= subknots[order])) {
        bs_value = Numerics.find_b_spline(iGrid[j],subknots,lb);
        for(l=0; l<n_inputKernels; l++) {
          kernelMatRLS[i+matSizeRLS*l]+=weights[j]*struct1Kernel[l][j]*bs_value;
          kernelMatRLS[i+n_grid+matSizeRLS*l]+=weights[j]*struct2Kernel[l][j]*bs_value;
        }
        if (dataCenter.constantMass!=0)
          massArray[i]+=iGrid[j]*iGrid[j]*rho[j]*weights[j]*bs_value;
        j++;
      }
    }
    for(l=0; l<n_inputKernels; l++){
      for(i=0; i<2*n_grid; i++) {
        kernelMatRLS[i+matSizeRLS*l]/=dataErrors.get(l);
      }
      for(m=0; m<dataCenter.n_surf; m++) {
        kernelMatRLS[m+2*n_grid+matSizeRLS*l]=
            surfPoly[l][m]/dataErrors.get(l);
      }
    }

    // initialise matrix used for fitting the data:
    fitDataMatRLS = new double[matSizeRLS*matSizeRLS];
    for(i=0; i<(2*n_grid+dataCenter.n_surf); i++){
      for(j=i; j<(2*n_grid+dataCenter.n_surf); j++){
        for(l=0; l<n_inputKernels; l++){
          fitDataMatRLS[i+matSizeRLS*j]+=
              kernelMatRLS[i+matSizeRLS*l]
             *kernelMatRLS[j+matSizeRLS*l];
        }
      }
    }

    // supplementary constraints for keeping the mass constant after inversion:  
    if (dataCenter.constantMass == 1) {
      for(i=0; i<n_grid; i++) {
        fitDataMatRLS[i+matSizeRLS*(2*n_grid+dataCenter.n_surf)]=massArray[i]*0.5;
      }
    }

    if (dataCenter.constantMass == 2) {
      for(i=0; i<n_grid; i++) {
        fitDataMatRLS[i+n_grid+matSizeRLS*(2*n_grid+dataCenter.n_surf)]=massArray[i]*0.5;
      }
    }

    if (dataCenter.constantMass == 3) {
      for(i=0; i<n_grid; i++) {
        fitDataMatRLS[i+matSizeRLS*(2*n_grid+dataCenter.n_surf)]=massArray[i]*0.5;
        fitDataMatRLS[i+n_grid+matSizeRLS*(2*n_grid+dataCenter.n_surf+1)]=massArray[i]*0.5;
      }
    }

    // finish initialising the matrix used for calculating the output kernels:
    for(i=0; i<(2*n_grid+dataCenter.n_surf); i++) {
      for(l=0; l<n_inputKernels; l++) {
        kernelMatRLS[i+matSizeRLS*l]/=dataErrors.get(l);
      }
    }

    // find Gauss-Legendre collocation points and weigths:
    int ngauleg = order - 2;
    double[][] xw = Numerics.gauleg(0.0, 1.0, ngauleg);

    // find b-spline values before calculating regularisation matrix:
    double[][][] bs = new double[n_grid][order][ngauleg];
    double[] temp = new double[3];
    double xp;
    for(i=0; i<n_grid; i++) {
      for(j=0; j<=order; j++) subknots[j] = knots[i+j];
      for(j=0; j<order; j++) {
        for(k=0; k<ngauleg; k++) {
          xp = subknots[j]+(subknots[j+1]-subknots[j])*xw[0][k];
          temp = Numerics.find_b_spline(xp,subknots,2,lb);
          bs[i][j][k] = temp[2];
        }
      }
    }

    // initialise the regularisation matrix:
    regularisationMatRLS = new double[n_grid*n_grid];
    int jfinish;
    for(i=0; i<n_grid; i++){
      jfinish = i+order;
      if (n_grid < jfinish) jfinish = n_grid;
      for(j=i; j<jfinish; j++){
        for(k=j; k<i+order; k++){
          for(l=0; l<ngauleg; l++){
            regularisationMatRLS[i+n_grid*j]+=
              xw[1][l]*
              bs[i][k-i][l]*
              bs[j][k-j][l]*
              (knots[k+1]-knots[k]);
          }
        }
      }
    }
  }

  /********************************************************
  * Do the inversion, using RLS
  ********************************************************/
  protected void RLSinversion() {  
    int n_grid = dataCenter.n_gridS;
    int ipiv[] = new int[matSizeRLS];
    int i, j, l;
    
    double Lambda = 0.0;
    for(l=0; l<n_inputKernels; l++) Lambda+=1.0/(dataErrors.get(l)*dataErrors.get(l));
    Lambda*=(dataCenter.SLambda/(double)n_inputKernels);
    
    double[] mat = fitDataMatRLS.clone();
    for(i=0; i<n_grid; i++) {
      for(j=i; j<n_grid; j++) {
        mat[i+matSizeRLS*j]+=Lambda*regularisationMatRLS[i+n_grid*j];
        mat[i+n_grid+matSizeRLS*(j+n_grid)]+=Lambda*regularisationMatRLS[i+n_grid*j];
      }
    }

    RLSmat = kernelMatRLS.clone();
    
    double[] work = new double[lwork*matSizeRLS];
    Dsytrf.dsytrf("U",matSizeRLS,mat,0,matSizeRLS,ipiv,0,work,0,lwork*matSizeRLS,info);
    Dsytrs.dsytrs("U",matSizeRLS,n_inputKernels,mat,0,matSizeRLS,ipiv,0,RLSmat,0,matSizeRLS,info);

    if (integrationMethod == SlidingWindow) {
      // transform RLSmat by going from a spline coefficient
      // representation to a grid based representation:
      double[] new_RLSmat = new double[matSizeRLS*n_inputKernels];
      double[] subknots = new double[order+1];
      double bs_value, lb = dataCenter.gridSInversion[0];
      int istart = 0;
      for(i=0; i<n_grid; i++){
        for(j=0; j<= order; j++) subknots[j] = knots[i+j];
        while (dataCenter.gridSInversion[istart] < subknots[0]) istart++;
        j = istart;
        while ((j < n_grid)&&(dataCenter.gridSInversion[j] <= subknots[order])) {
          bs_value = Numerics.find_b_spline(dataCenter.gridSInversion[j],subknots,lb);
          for(l = 0; l<n_inputKernels; l++) {
            new_RLSmat[j+matSizeRLS*l]+=RLSmat[i+matSizeRLS*l]*bs_value;
            new_RLSmat[j+n_grid+matSizeRLS*l]+=RLSmat[i+n_grid+matSizeRLS*l]*bs_value;
          }
          j++;
        }
      }
      RLSmat = new_RLSmat;
    }

    // create data with or without noise:
    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.SNoiseSelected) {
      for(l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.SNoiseLevel*dataCenter.SNoise[l]*dataErrors.get(l);
      }
    } else {
      for(l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }

    // find the inverted functions:
    f1_invertedRLS = new double[n_grid];
    f2_invertedRLS = new double[n_grid];
    for(i=0; i<n_grid; i++) {
      for(l=0; l<n_inputKernels; l++) {
        f1_invertedRLS[i]+=RLSmat[i+matSizeRLS*l]*dataPlusNoise[l];
        f2_invertedRLS[i]+=RLSmat[i+n_grid+matSizeRLS*l]*dataPlusNoise[l];
      }
    }

    // find the error bars:
    RLS_errorBars1 = new double[n_grid];
    RLS_errorBars2 = new double[n_grid];
    for(i=0; i<n_grid; i++) {
      for(l=0; l<n_inputKernels; l++) {
        RLS_errorBars1[i]+=Math.pow(RLSmat[i+matSizeRLS*l]*dataErrors.get(l),2.0);
        RLS_errorBars2[i]+=Math.pow(RLSmat[i+n_grid+matSizeRLS*l]*dataErrors.get(l),2.0);
      }
      RLS_errorBars1[i] = Math.sqrt(RLS_errorBars1[i]);
      RLS_errorBars2[i] = Math.sqrt(RLS_errorBars2[i]);
    }
  }
  
  /********************************************************
  * This calculates the RLS averaging kernel 1 at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findAveragingKernel1RLS(int iLocation) {
    int n_grid = dataCenter.gridSInversion.length;
    int n_big_grid = dataCenter.grid.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;

    int i,l;    
    double[] result = new double[n_big_grid];
    if ((dataCenter.kernelType==0)||(dataCenter.kernelType==2)) {
      for(i=0; i<n_big_grid ;i++){
        for(l=0; l<n_inputKernels;l++){
          result[i]+=(dataCenter.modes.get(intersection.get(l)).kernel.rhoc2Kernel[i]
            *RLSmat[iLocation+matSizeRLS*l]);
        }
      }
    } else if ((dataCenter.kernelType==1)||(dataCenter.kernelType==3)) {
      for(i=0; i<n_big_grid ;i++){
        for(l=0; l<n_inputKernels;l++){
          result[i]+=(dataCenter.modes.get(intersection.get(l)).kernel.rhoGamma1Kernel[i]
            *RLSmat[iLocation+matSizeRLS*l]);
        }
      }
    } else {
      // This should never happen
      JOptionPane.showMessageDialog(
        null, // parent component
        "An unexpected error occured when calculating an averaging kernel.\n"+
        "Please contact the authors of InversionKit.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      return null;
    } 
    return result;
  }

  /********************************************************
  * This calculates the RLS averaging kernel 2 at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findAveragingKernel2RLS(int iLocation) {
    int n_grid = dataCenter.gridSInversion.length;
    int n_big_grid = dataCenter.grid.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;

    int i,l;    
    double[] result = new double[n_big_grid];
    for(i=0; i<n_big_grid ;i++){
      for(l=0; l<n_inputKernels;l++){
        result[i]+=(dataCenter.modes.get(intersection.get(l)).kernel.c2rhoKernel[i]
          *RLSmat[iLocation+n_grid+matSizeRLS*l]);
      }
    }
    return result;
  }

  /********************************************************
  * This calculates the RLS averaging kernel 2 at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findCrossTermKernel1RLS(int iLocation) {
    int n_grid = dataCenter.gridSInversion.length;
    int n_big_grid = dataCenter.grid.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;

    int i,l;    
    double[] result = new double[n_big_grid];
    for(i=0; i<n_big_grid ;i++){
      for(l=0; l<n_inputKernels;l++){
        result[i]+=(dataCenter.modes.get(intersection.get(l)).kernel.c2rhoKernel[i]
          *RLSmat[iLocation+matSizeRLS*l]);
      }
    }
    return result;
  }

  /********************************************************
  * This calculates the RLS averaging kernel 2 at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findCrossTermKernel2RLS(int iLocation) {
    int n_grid = dataCenter.gridSInversion.length;
    int n_big_grid = dataCenter.grid.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;

    int i,l;    
    double[] result = new double[n_big_grid];
    if ((dataCenter.kernelType==0)||(dataCenter.kernelType==2)) {
      for(i=0; i<n_big_grid ;i++){
        for(l=0; l<n_inputKernels;l++){
          result[i]+=(dataCenter.modes.get(intersection.get(l)).kernel.rhoc2Kernel[i]
            *RLSmat[iLocation+n_grid+matSizeRLS*l]);
        }
      }
    } else if ((dataCenter.kernelType==1)||(dataCenter.kernelType==3)) {
      for(i=0; i<n_big_grid ;i++){
        for(l=0; l<n_inputKernels;l++){
          result[i]+=(dataCenter.modes.get(intersection.get(l)).kernel.rhoGamma1Kernel[i]
            *RLSmat[iLocation+n_grid+matSizeRLS*l]);
        }
      }
    } else {
      // This should never happen
      JOptionPane.showMessageDialog(
        null, // parent component
        "An unexpected error occured when calculating a cross-term kernel.\n"+
        "Please contact the authors of InversionKit.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      return null;
    } 
    return result;
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  protected void initSOLA() {
    int n_grid = iGrid.length;
    matSizeSOLA = n_inputKernels+dataCenter.n_surf+1;
    double[] rho = null;
    int suppl = 0;

    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2)) suppl = 1;
    if (dataCenter.constantMass==3) suppl = 2;
    if (dataCenter.constantMass!=0)
      rho = Numerics.interpolate1D_r2(dataCenter.grid,iGrid,dataCenter.rho,true);
    matSizeSOLA+=suppl;

    // initialise matrix associated with struct1 kernels:
    kernelMat1SOLA = new double[matSizeSOLA*matSizeSOLA];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        for(int i=0; i<n_grid; i++){
          kernelMat1SOLA[l+matSizeSOLA*ll]+=weights[i]
               *struct1Kernel[l][i]*struct1Kernel[ll][i];
        }
      }
    }
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==3)) {
      for(int l=0; l<n_inputKernels; l++){
        for(int i=0; i<n_grid; i++) {
          kernelMat1SOLA[l+matSizeSOLA*n_inputKernels]+=weights[i]
              *struct1Kernel[l][i]*iGrid[i]*iGrid[i]*rho[i];
        }
      }
      for(int i=0; i<n_grid; i++) {
        kernelMat1SOLA[(1+matSizeSOLA)*n_inputKernels]+=
           weights[i]*rho[i]*rho[i]*Math.pow(iGrid[i],4.0);
      }
    }

    // initialise matrix associated with struct2 kernels:
    kernelMat2SOLA = new double[matSizeSOLA*matSizeSOLA];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        for(int i=0; i<n_grid; i++){
          kernelMat2SOLA[l+matSizeSOLA*ll]+=weights[i]
               *struct2Kernel[l][i]*struct2Kernel[ll][i];
        }
      }
    }
    if ((dataCenter.constantMass==2)||(dataCenter.constantMass==3)) {
      int ndx = dataCenter.constantMass-2;
      for(int l=0; l<n_inputKernels; l++){
        for(int i=0; i<n_grid; i++) {
          kernelMat2SOLA[l+matSizeSOLA*(n_inputKernels+ndx)]+=weights[i]
              *struct2Kernel[l][i]*iGrid[i]*iGrid[i]*rho[i];
        }
      }
      for(int i=0; i<n_grid; i++) {
        kernelMat2SOLA[(1+matSizeSOLA)*(n_inputKernels+ndx)]+=
           weights[i]*rho[i]*rho[i]*Math.pow(iGrid[i],4.0);
      }
    }

    // initialise the matrix used for reducing the cross-term from struct 2 on struct 1
    double[] w = new double[n_grid];
    for(int i=0; i<n_grid; i++) w[i] = weights[i]*Math.pow(1.0+iGrid[i],4.0);
    crossTermMat1SOLA = new double[(n_inputKernels+suppl)*(n_inputKernels+suppl)];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        for(int i=0; i<n_grid; i++){
          crossTermMat1SOLA[l+(n_inputKernels+suppl)*ll]+=w[i]
               *struct2Kernel[l][i]*struct2Kernel[ll][i];
        }
      }
    }
    if ((dataCenter.constantMass==2)||(dataCenter.constantMass==3)) {
      for(int l=0; l<n_inputKernels; l++){
        for(int i=0; i<n_grid; i++) {
          crossTermMat1SOLA[l+(n_inputKernels+suppl)*(n_inputKernels+suppl-1)]+=w[i]
              *struct2Kernel[l][i]*iGrid[i]*iGrid[i]*rho[i];
        }
      }
      for(int i=0; i<n_grid; i++) {
        crossTermMat1SOLA[(n_inputKernels+suppl+1)*(n_inputKernels+suppl-1)]+=
           w[i]*rho[i]*rho[i]*Math.pow(iGrid[i],4.0);
      }
    }

    // initialise the matrix used for reducing the cross-term from struct 1 on struct 2
    crossTermMat2SOLA = new double[(n_inputKernels+suppl)*(n_inputKernels+suppl)];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        for(int i=0; i<n_grid; i++){
          crossTermMat2SOLA[l+(n_inputKernels+suppl)*ll]+=w[i]
               *struct1Kernel[l][i]*struct1Kernel[ll][i];
        }
      }
    }
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==3)) {
      for(int l=0; l<n_inputKernels; l++){
        for(int i=0; i<n_grid; i++) {
          crossTermMat2SOLA[l+(n_inputKernels+suppl)*n_inputKernels]+=w[i]
              *struct1Kernel[l][i]*iGrid[i]*iGrid[i]*rho[i];
        }
      }
      for(int i=0; i<n_grid; i++) {
        crossTermMat2SOLA[(n_inputKernels+suppl+1)*n_inputKernels]+=
           w[i]*rho[i]*rho[i]*Math.pow(iGrid[i],4.0);
      }
    }


    // initialise matrix associated with supplementary constraints for
    // struct1 kernels:
    double sum;
    supplMat1SOLA = new double[n_inputKernels*(1+dataCenter.n_surf)];
    for(int l=0; l<n_inputKernels; l++){
      sum = 0.0;
      for(int i=0; i<n_grid; i++) sum+=(weights[i]*struct1Kernel[l][i]);
      supplMat1SOLA[l] = sum*0.5;
      for(int n=0; n<dataCenter.n_surf; n++) {
        supplMat1SOLA[l+n_inputKernels*(1+n)] = surfPoly[l][n]*0.5;
      }
    }

    // initialise matrix associated with supplementary constraints for
    // struct2 kernels:
    supplMat2SOLA = new double[n_inputKernels*(1+dataCenter.n_surf)];
    for(int l=0; l<n_inputKernels; l++){
      sum = 0.0;
      for(int i=0; i<n_grid; i++) sum+=(weights[i]*struct2Kernel[l][i]);
      supplMat2SOLA[l] = sum*0.5;
      for(int n=0; n<dataCenter.n_surf; n++) {
        supplMat2SOLA[l+n_inputKernels*(1+n)] = surfPoly[l][n]*0.5;
      }
    }

    // initialise regularisation "matrix":
    double mu = 0.0;
    for(int l=0; l<n_inputKernels; l++) mu+=dataErrors.get(l)*dataErrors.get(l);
    mu = (double)n_inputKernels/mu;
    
    regularisationMatSOLA = new double[n_inputKernels];
    for(int l=0; l<n_inputKernels; l++){
      regularisationMatSOLA[l]=mu*dataErrors.get(l)*dataErrors.get(l);
    }
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  protected void initDelta1SOLA() {
    int n_grid = dataCenter.n_gridS;
    int n_big_grid = iGrid.length;
    int suppl = 0;
    double[] rho = null;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      suppl=1;
    if (dataCenter.constantMass==3)
      suppl=2;
    if (dataCenter.constantMass!=0)
      rho = Numerics.interpolate1D_r2(dataCenter.grid,iGrid,dataCenter.rho,true);
    matSizeTargetSOLA = n_inputKernels+suppl+1;

    // obtain the sound velocity:
    double[] c = Numerics.interpolate1D_r2(dataCenter.grid,dataCenter.gridSInversion,dataCenter.c2,false);
    for(int i=0; i<c.length; i++) c[i] = Math.sqrt(c[i]);

    // this is to avoid division by zero later on:
    double max_c = 0.0;
    for(int i=0; i<c.length; i++) if (max_c < c[i]) max_c = c[i];
    for(int i=0; i<c.length; i++) if ((c[i]/max_c) < 1e-3) c[i] = max_c*1e-3;

    // initialise the matrix with the target kernels:
    double Norm, Delta, Cnst;
    double FunctionValue;
    targetKernelMat1SOLA = new double[matSizeTargetSOLA*n_grid];
    for(int i=0; i<n_grid; i++) {
      Norm = 0.0;
      Delta = dataCenter.SDelta1*c[i];
      Cnst = Math.log(1.0/(Delta*dataCenter.gridSInversion[i]));
      for(int j=0; j<n_big_grid; j++) {
        FunctionValue = weights[j]*iGrid[j]*Math.exp(-Math.pow((iGrid[j]-dataCenter.gridSInversion[i])/Delta,2.0)
                      - (iGrid[j]-dataCenter.gridSInversion[i])/dataCenter.gridSInversion[i] + Cnst);
        Norm+=FunctionValue;
        for(int l=0; l<n_inputKernels; l++) 
          targetKernelMat1SOLA[l+matSizeTargetSOLA*i]+=(FunctionValue*struct1Kernel[l][j]);
        if ((dataCenter.constantMass==1)||(dataCenter.constantMass==3)) 
          targetKernelMat1SOLA[n_inputKernels+matSizeTargetSOLA*i]+=(FunctionValue*rho[j]
             *iGrid[j]*iGrid[j]);
      }
      Norm = 1.0/Norm;
      for(int l=0; l<n_inputKernels; l++) 
        targetKernelMat1SOLA[l+matSizeTargetSOLA*i]*=Norm;
      if ((dataCenter.constantMass==1)||(dataCenter.constantMass==3)) 
        targetKernelMat1SOLA[n_inputKernels+matSizeTargetSOLA*i]*=Norm;
      targetKernelMat1SOLA[n_inputKernels+suppl+matSizeTargetSOLA*i]=0.5;
    }
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  protected void initDelta2SOLA() {
    int n_grid = dataCenter.n_gridS;
    int n_big_grid = iGrid.length;
    int suppl = 0, ndx = 0;
    double[] rho = null;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      suppl=1;
    if (dataCenter.constantMass==3)
      suppl=2;
    if (dataCenter.constantMass!=0)
      rho = Numerics.interpolate1D_r2(dataCenter.grid,iGrid,dataCenter.rho,true);
    if (dataCenter.constantMass>1) ndx = dataCenter.constantMass-2;
    matSizeTargetSOLA = n_inputKernels+suppl+1;

    // obtain the sound velocity:
    double[] c = Numerics.interpolate1D_r2(dataCenter.grid,dataCenter.gridSInversion,dataCenter.c2,false);
    for(int i=0; i<c.length; i++) c[i] = Math.sqrt(c[i]);

    // this is to avoid division by zero later on:
    double max_c = 0.0;
    for(int i=0; i<c.length; i++) if (max_c < c[i]) max_c = c[i];
    for(int i=0; i<c.length; i++) if ((c[i]/max_c) < 1e-3) c[i] = max_c*1e-3;

    // initialise the matrix with the target kernels:
    double Norm, Delta, Cnst;
    double FunctionValue;
    targetKernelMat2SOLA = new double[matSizeTargetSOLA*n_grid];
    for(int i=0; i<n_grid; i++) {
      Norm = 0.0;
      Delta = dataCenter.SDelta2*c[i];
      Cnst = Math.log(1.0/(Delta*dataCenter.gridSInversion[i]));
      for(int j=0; j<n_big_grid; j++) {
        FunctionValue = weights[j]*iGrid[j]*Math.exp(-Math.pow((iGrid[j]-dataCenter.gridSInversion[i])/Delta,2.0)
                      - (iGrid[j]-dataCenter.gridSInversion[i])/dataCenter.gridSInversion[i] + Cnst);
        Norm+=FunctionValue;
        for(int l=0; l<n_inputKernels; l++) 
          targetKernelMat2SOLA[l+matSizeTargetSOLA*i]+=(FunctionValue*struct2Kernel[l][j]);
        if ((dataCenter.constantMass==2)||(dataCenter.constantMass==3)) 
          targetKernelMat2SOLA[n_inputKernels+ndx+matSizeTargetSOLA*i]+=(FunctionValue*rho[j]
             *iGrid[j]*iGrid[j]);
      }
      Norm = 1.0/Norm;
      for(int l=0; l<n_inputKernels; l++) 
        targetKernelMat2SOLA[l+matSizeTargetSOLA*i]*=Norm;
      if ((dataCenter.constantMass==2)||(dataCenter.constantMass==3)) 
        targetKernelMat2SOLA[n_inputKernels+ndx+matSizeTargetSOLA*i]*=Norm;
      targetKernelMat2SOLA[n_inputKernels+suppl+matSizeTargetSOLA*i]=0.5;
    }
  }

  /********************************************************
  * Do the inversion, using SOLA
  ********************************************************/
  protected void SOLAinversion1() {
    int n_grid = dataCenter.n_gridS;
    int n_big_grid = dataCenter.grid.length;
    int suppl = 0;

    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      suppl=1;
    if (dataCenter.constantMass==3)
      suppl=2;

    int ipiv[] = new int[matSizeSOLA];

    double[] mat = kernelMat1SOLA.clone();
    double tanTheta = Math.tan(dataCenter.STheta1);
    for(int l=0; l<n_inputKernels; l++) 
      mat[l+matSizeSOLA*l]+=tanTheta*regularisationMatSOLA[l];
    for(int l=0; l<n_inputKernels; l++) {
      for(int n=0; n<(1+dataCenter.n_surf); n++)
        mat[l+matSizeSOLA*(n+suppl+n_inputKernels)]+=supplMat1SOLA[l+n_inputKernels*n];
    }

    for(int l=0; l<(n_inputKernels+suppl); l++) {
      for(int ll=l; ll<(n_inputKernels+suppl); ll++) {
         mat[l+matSizeSOLA*ll]+=dataCenter.Beta1*crossTermMat1SOLA[l+(n_inputKernels+suppl)*ll];
      }
    }

    SOLAmat1 = new double[matSizeSOLA*n_grid];
    for(int i=0; i<matSizeTargetSOLA; i++) {
      for(int j=0; j<n_grid; j++) {
        SOLAmat1[i+matSizeSOLA*j] = targetKernelMat1SOLA[i+matSizeTargetSOLA*j];
      }
    }
    double[] work = new double[lwork*matSizeSOLA];
    Dsytrf.dsytrf("U",matSizeSOLA,mat,0,matSizeSOLA,ipiv,0,work,0,lwork*matSizeSOLA,info);
    Dsytrs.dsytrs("U",matSizeSOLA,n_grid,mat,0,matSizeSOLA,ipiv,0,SOLAmat1,0,matSizeSOLA,info);

    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.SNoiseSelected) {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.SNoiseLevel*dataCenter.SNoise[l]*dataErrors.get(l);
      }
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }

    // find the inverted function:
    f1_invertedSOLA = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        f1_invertedSOLA[i]+=dataPlusNoise[l]*SOLAmat1[l+matSizeSOLA*i];
      }
    }

    // find the error bars:
    SOLA_errorBars1 = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        SOLA_errorBars1[i]+=Math.pow(SOLAmat1[l+matSizeSOLA*i]*dataErrors.get(l),2.0);
      }
      SOLA_errorBars1[i]=Math.sqrt(SOLA_errorBars1[i]);
    }

    // Averaging kernels obtained from the inversion:
    // 1st coordinate: target position
    // 2nd coordinate: grid position
    averagingKernels1SOLA = new double[n_grid][n_big_grid];
    if ((dataCenter.kernelType==0)||(dataCenter.kernelType==2)) {
      for(int i=0; i<n_big_grid ;i++){
        for(int j=0; j<n_grid; j++) {
          averagingKernels1SOLA[j][i]=0.0;
          for(int k=0; k<n_inputKernels;k++){
            averagingKernels1SOLA[j][i]+=(
               dataCenter.modes.get(intersection.get(k)).kernel.rhoc2Kernel[i]
              *SOLAmat1[k+matSizeSOLA*j]);
          }
        }
      }
    } else {
      for(int i=0; i<n_big_grid ;i++){
        for(int j=0; j<n_grid; j++) {
          averagingKernels1SOLA[j][i]=0.0;
          for(int k=0; k<n_inputKernels;k++){
            averagingKernels1SOLA[j][i]+=(
               dataCenter.modes.get(intersection.get(k)).kernel.rhoGamma1Kernel[i]
              *SOLAmat1[k+matSizeSOLA*j]);
          }
        }
      }
    }

    // find new output grid and horizontal error bars for the SOLA method:
    double[] matIntg = Numerics.findIntegrationMatrix(dataCenter.grid);
    double[] intg;
    double lambda;
    SOLA_newGrid1 = new double[n_grid];
    SOLA_leftErrorBars1  = new double[n_grid];
    SOLA_rightErrorBars1 = new double[n_grid];
    for(int j=0; j<n_grid; j++) {
      intg = Numerics.applyIntegrationMatrix(matIntg, averagingKernels1SOLA[j]);
      int i=1;
      while(i<n_big_grid) {if (intg[i]>0.25) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.25-intg[i-1])/(intg[i]-intg[i-1]);
      SOLA_leftErrorBars1[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      while(i<n_big_grid) {if (intg[i]>0.5) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.5-intg[i-1])/(intg[i]-intg[i-1]);
      SOLA_newGrid1[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      while(i<n_big_grid) {if (intg[i]>0.75) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.75-intg[i-1])/(intg[i]-intg[i-1]);
      SOLA_rightErrorBars1[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      SOLA_leftErrorBars1[j] = SOLA_newGrid1[j]-SOLA_leftErrorBars1[j];
      SOLA_rightErrorBars1[j] = SOLA_rightErrorBars1[j]-SOLA_newGrid1[j];
    }
  }

  /********************************************************
  * This calculates the SOLA averaging kernel 1 at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findAveragingKernel1SOLA(int iLocation) {
    int n_grid = dataCenter.gridSInversion.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;
    return averagingKernels1SOLA[iLocation];
  }

  /********************************************************
  * This calculates the SOLA cross-term kernel 1 at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findCrossTermKernel1SOLA(int iLocation) {
    int n_grid = dataCenter.gridSInversion.length;
    int n_big_grid = dataCenter.grid.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;

    int i,l;    
    double[] result = new double[n_big_grid];
    for(i=0; i<n_big_grid ;i++){
      for(l=0; l<n_inputKernels;l++){
        result[i]+=(
          dataCenter.modes.get(intersection.get(l)).kernel.c2rhoKernel[i]
          *SOLAmat1[l+matSizeSOLA*iLocation]);
      }
    }
    return result;
  }

  /********************************************************
  * Do the inversion, using SOLA
  ********************************************************/
  protected void SOLAinversion2() {
    int n_grid = dataCenter.n_gridS;
    int n_big_grid = dataCenter.grid.length;
    int suppl = 0;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      suppl=1;
    if (dataCenter.constantMass==3)
      suppl=2;

    int ipiv[] = new int[matSizeSOLA];

    double[] mat = kernelMat2SOLA.clone();
    double tanTheta = Math.tan(dataCenter.STheta2);
    for(int l=0; l<n_inputKernels; l++) 
      mat[l+matSizeSOLA*l]+=tanTheta*regularisationMatSOLA[l];
    for(int l=0; l<n_inputKernels; l++) {
      for(int n=0; n<(1+dataCenter.n_surf); n++)
        mat[l+matSizeSOLA*(n+suppl+n_inputKernels)]+=supplMat2SOLA[l+n_inputKernels*n];
    }

    for(int l=0; l<(n_inputKernels+suppl); l++) {
      for(int ll=l; ll<(n_inputKernels+suppl); ll++) {
         mat[l+matSizeSOLA*ll]+=dataCenter.Beta2*crossTermMat2SOLA[l+(n_inputKernels+suppl)*ll];
      }
    }

    SOLAmat2 = new double[matSizeSOLA*n_grid];
    for(int i=0; i<matSizeTargetSOLA; i++) {
      for(int j=0; j<n_grid; j++) {
        SOLAmat2[i+matSizeSOLA*j] = targetKernelMat2SOLA[i+matSizeTargetSOLA*j];
      }
    }
    double[] work = new double[lwork*matSizeSOLA];
    Dsytrf.dsytrf("U",matSizeSOLA,mat,0,matSizeSOLA,ipiv,0,work,0,lwork*matSizeSOLA,info);
    Dsytrs.dsytrs("U",matSizeSOLA,n_grid,mat,0,matSizeSOLA,ipiv,0,SOLAmat2,0,matSizeSOLA,info);

    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.SNoiseSelected) {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.SNoiseLevel*dataCenter.SNoise[l]*dataErrors.get(l);
      }
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }
    
    // find the inteverted function:
    f2_invertedSOLA = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        f2_invertedSOLA[i]+=dataPlusNoise[l]*SOLAmat2[l+matSizeSOLA*i];
      }
    }

    // find the error bars:
    SOLA_errorBars2 = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        SOLA_errorBars2[i]+=Math.pow(SOLAmat2[l+matSizeSOLA*i]*dataErrors.get(l),2.0);
      }
      SOLA_errorBars2[i]=Math.sqrt(SOLA_errorBars2[i]);
    }

    // Averaging kernels obtained from the inversion:
    // 1st coordinate: target position
    // 2nd coordinate: grid position
    averagingKernels2SOLA = new double[n_grid][n_big_grid];
    for(int i=0; i<n_big_grid; i++) {
      for(int j=0; j<n_grid; j++) {
        averagingKernels2SOLA[j][i]=0.0;
        for(int k=0; k<n_inputKernels; k++) {
          averagingKernels2SOLA[j][i]+=(
             dataCenter.modes.get(intersection.get(k)).kernel.c2rhoKernel[i]
            *SOLAmat2[k+matSizeSOLA*j]);
        }
      }
    }

    // find new output grid and horizontal error bars for the SOLA method:
    double[] matIntg = Numerics.findIntegrationMatrix(dataCenter.grid);
    double[] intg;
    double lambda;
    SOLA_newGrid2 = new double[n_grid];
    SOLA_leftErrorBars2  = new double[n_grid];
    SOLA_rightErrorBars2 = new double[n_grid];
    for(int j=0; j<n_grid; j++) {
      intg = Numerics.applyIntegrationMatrix(matIntg, averagingKernels2SOLA[j]);
      int i=1;
      while(i<n_big_grid) {if (intg[i]>0.25) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.25-intg[i-1])/(intg[i]-intg[i-1]);
      SOLA_leftErrorBars2[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      while(i<n_big_grid) {if (intg[i]>0.5) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.5-intg[i-1])/(intg[i]-intg[i-1]);
      SOLA_newGrid2[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      while(i<n_big_grid) {if (intg[i]>0.75) break; i++;}
      if (i==n_big_grid) continue;
      lambda = (0.75-intg[i-1])/(intg[i]-intg[i-1]);
      SOLA_rightErrorBars2[j] = (1.0-lambda)*dataCenter.grid[i-1]+lambda*dataCenter.grid[i];
      SOLA_leftErrorBars2[j] = SOLA_newGrid2[j]-SOLA_leftErrorBars2[j];
      SOLA_rightErrorBars2[j] = SOLA_rightErrorBars2[j]-SOLA_newGrid2[j];
    }
  }

  /********************************************************
  * This calculates the SOLA averaging kernel 2 at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findAveragingKernel2SOLA(int iLocation) {
    int n_grid = dataCenter.gridSInversion.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;
    return averagingKernels2SOLA[iLocation];
  }

  /********************************************************
  * This calculates the SOLA cross-term kernel 2 at a
  * particular location (given by iLocation).
  ********************************************************/
  protected double[] findCrossTermKernel2SOLA(int iLocation) {
    int n_grid = dataCenter.gridSInversion.length;
    int n_big_grid = dataCenter.grid.length;
    if (iLocation < 0) return null;
    if (iLocation >= n_grid) return null;

    int i,l;    
    double[] result = new double[n_big_grid];
    if ((dataCenter.kernelType==0)||(dataCenter.kernelType==2)) {
      for(i=0; i<n_big_grid ;i++){
        for(l=0; l<n_inputKernels;l++){
          result[i]+=(dataCenter.modes.get(intersection.get(l)).kernel.rhoc2Kernel[i]
            *SOLAmat2[l+matSizeSOLA*iLocation]);
        }
      }
    } else if ((dataCenter.kernelType==1)||(dataCenter.kernelType==3)) {
      for(i=0; i<n_big_grid ;i++){
        for(l=0; l<n_inputKernels;l++){
          result[i]+=(dataCenter.modes.get(intersection.get(l)).kernel.rhoGamma1Kernel[i]
            *SOLAmat2[l+matSizeSOLA*iLocation]);
        }
      }
    } else {
      // This should never happen
      JOptionPane.showMessageDialog(
        null, // parent component
        "An unexpected error occured when calculating a cross-term kernel.\n"+
        "Please contact the authors of InversionKit.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      return null;
    } 
    return result;
  }
}
