/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import graph.IPlot;
import graph.Plot;
import inversion_applet.Utilities;
import inversion_applet.DataCenter;
import inversion_applet.StructuralInversion;
import inversion_applet.FrequenciesDataModel;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JSeparator;
import javax.swing.JCheckBox;
import javax.swing.JSlider;
import javax.swing.JComponent;
import javax.swing.JComboBox;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;
import javax.swing.SwingConstants;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import java.util.StringTokenizer;
import java.util.ArrayList;

public class StructuralInversionPage extends JPanel {

  private DataCenter dataCenter;
  private IPlot graphInvertedFunction;
  private IPlot graphAveragingKernel;
  private IPlot graphCrossTermKernel;
  private JPanel graphContainer;
  private StructuralInversion my_inversion;
  private JPanel SliderContainer;

  private JLabel numberOfKernels;
  private JButton Refresh;
  private JCheckBox SOLACheckBox;
  private JCheckBox RLSCheckBox;
  private JCheckBox TargetFunctionCheckBox;

  private JCheckBox showErrorBars;
  private JCheckBox calculateWhileSliding;
  
  private JComboBox kernelTypeComboBox;
  private JPanel structPanel;
  private JRadioButton struct1Inversion;
  private JRadioButton struct2Inversion;
  
  private JPanel NoiseTitle;
  private JCheckBox NoiseCheckBox;
  private JTextField NoiseTextField;
  private JSlider NoiseSlider;
  private JButton NoiseRefresh;

  private JLabel RLStitle;
  private JPanel LambdaTitle;
  private JLabel LambdaLabel;
  private JTextField LambdaTextField;
  private JSlider LambdaSlider;

  private JLabel SOLAtitle;
  private JPanel BetaTitle;
  private JLabel BetaLabel;
  private JTextField BetaTextField;
  private JSlider BetaSlider;
  private JPanel ThetaTitle;
  private JLabel ThetaLabel;
  private JTextField ThetaTextField;
  private JSlider ThetaSlider;
  private JPanel DeltaTitle;
  private JLabel DeltaLabel;
  private JTextField DeltaTextField;
  private JSlider DeltaSlider;

  private JPanel KernelTitle;
  private JLabel KernelLabel;
  private JTextField KernelTextField;
  private JSlider KernelSlider;
  private JPanel NGridPanel;
  private JLabel NGridLabel;
  private JTextField NGridTextField;

  private JPanel NSurfPanel;
  private JLabel NSurfLabel;
  private JTextField NSurfTextField;

  private JComboBox integrationMethodComboBox;

  private JButton LoadTarget;
  private JButton ClearTarget;
  private JButton CalculateFreq;

  // this keeps track of whether a parameter has been modified in one of the
  // text fields.
  private boolean paramModified;

  private static final double ln10 = 2.302585092994046;

  /************************************************************************
  * A constructor.
  ************************************************************************/
  public StructuralInversionPage(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
  }

  //////////////////////////////////////////////////////////////
  // These methods give the relationship between the slider
  // value and that of various parameters of the inversion.
  //////////////////////////////////////////////////////////////
  private int LambdaToSlider(double Lambda) {
    if (Lambda == 0.0) return -180;
    int result = (int)Math.round(Math.log(Lambda)*20.0/ln10);
    if (result < -180) result = -180;
    if (result > 20) result = 20;
    return result;
  }
  private double SliderToLambda() {
    return Math.pow(10.0,LambdaSlider.getValue()/20.0);
  }
  private int NoiseToSlider(double NoiseLevel) {
    if (NoiseLevel == 0.0) return -60;
    int result = (int)Math.round(Math.log(NoiseLevel)*20.0/ln10);
    if (result < -60) result = -60;
    if (result > 60) result = 60;
    return result;
  }
  private double SliderToNoise() {
    return Math.pow(10.0,NoiseSlider.getValue()/20.0);
  }
  private int BetaToSlider(double Beta) {
    if (Beta == 0.0) return -100;
    int result = (int)Math.round(Math.log(Beta)*25.0/ln10);
    if (result < -100) result = -100;
    if (result > 100) result = 100;
    return result;
  }
  private double SliderToBeta() {
    return Math.pow(10.0,BetaSlider.getValue()/25.0);
  }
  private int ThetaToSlider(double Theta) {
    if (Theta == 0.0) return -200;
    int result = (int)Math.round(Math.log(Theta/1.5)*50.0/ln10);
    if (result < -200) result = -200;
    if (result > 0) result = 0;
    return result;
  }
  private double SliderToTheta() {
    return 1.5*Math.pow(10.0,ThetaSlider.getValue()/50.0);
  }
  private int DeltaToSlider(double Delta) {
    if (Delta == 0.0) return -200;
    int result = (int)Math.round(Math.log(Delta)*50.0/ln10);
    if (result < -200) result = -200;
    if (result > 0) result = 0;
    return result;
  }
  private double SliderToDelta() {
    return Math.pow(10.0,DeltaSlider.getValue()/50.0);
  }

  private int XKernelToSlider(double xKernel) {
    if (dataCenter.gridSInversion == null) return 0;
    int ngridS = dataCenter.gridSInversion.length;
    if (ngridS == 0) return 0;
    if (xKernel < dataCenter.gridSInversion[0]) return 0;
    if (xKernel > dataCenter.gridSInversion[ngridS - 1]) return ngridS - 1;

    int istart = 0;
    int ifinish = ngridS - 1;
    int inext;
    while((istart+1) < ifinish) {
      inext = (istart+ifinish)/2;
      if (dataCenter.gridSInversion[inext] > xKernel) {
        ifinish = inext;
      } else {
        istart = inext;
      }
    }

    if ((xKernel-dataCenter.gridSInversion[istart]) < (dataCenter.gridSInversion[ifinish]-xKernel)) {
      return istart;
    } else {
      return ifinish;
    }
  }

  private double SliderToXKernel () {
    if (dataCenter.gridSInversion == null) return 0.0;
    return dataCenter.gridSInversion[KernelSlider.getValue()];
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the noise slider and refresh button
  /////////////////////////////////////////////////////////////////////////////
  private void initNoiseButtons() {

    double NoiseLevel = dataCenter.SNoiseLevel;
    boolean NoiseSelected = dataCenter.SNoiseSelected;
    
    NoiseTextField = new JTextField(String.format("%.4e",NoiseLevel));
    NoiseTextField.setEnabled(NoiseSelected);
    NoiseTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkNoiseTextField()) {
          if (!paramModified) return;
          if (my_inversion.canInvert) {
            my_inversion.RLSinversion();
            my_inversion.SOLAinversion1();
            my_inversion.SOLAinversion2();
            replotInvertedFunctions();
          }
        }
      }
    });

    NoiseSlider = new JSlider(JSlider.HORIZONTAL,-60,60,NoiseToSlider(NoiseLevel));
    NoiseSlider.setEnabled(NoiseSelected);
    NoiseSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double NoiseLevel = SliderToNoise();
        NoiseTextField.setText(String.format("%.4e",NoiseLevel));
        if ((!calculateWhileSliding.isSelected())&&
            (NoiseSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.SNoiseLevel = NoiseLevel;
        if (my_inversion.canInvert) {
          my_inversion.RLSinversion();
          my_inversion.SOLAinversion1();
          my_inversion.SOLAinversion2();
          replotInvertedFunctions();
        }
      }
    });

    NoiseRefresh = new JButton("New noise array");
    NoiseRefresh.setEnabled(NoiseSelected);
    NoiseRefresh.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (my_inversion.canInvert) {
          my_inversion.initNoise();
          my_inversion.RLSinversion();
          my_inversion.SOLAinversion1();
          my_inversion.SOLAinversion2();
          replotInvertedFunctions();
        }
      }
    });
    
    NoiseCheckBox = new JCheckBox("Noise",NoiseSelected);
    NoiseCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.SNoiseSelected = NoiseCheckBox.isSelected();
        NoiseTextField.setEnabled(NoiseCheckBox.isSelected());
        NoiseSlider.setEnabled(NoiseCheckBox.isSelected());
        NoiseRefresh.setEnabled(NoiseCheckBox.isSelected());
        if (my_inversion.canInvert) {
          my_inversion.RLSinversion();
          my_inversion.SOLAinversion1();
          my_inversion.SOLAinversion2();
          replotInvertedFunctions();
        }
      }
    });

    NoiseTitle = new JPanel();
    NoiseTitle.setLayout(new GridLayout(1,2));
    NoiseTitle.add(NoiseCheckBox);
    NoiseTitle.add(NoiseTextField);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the RLS slider
  /////////////////////////////////////////////////////////////////////////////
  private void initRLSslider() {
    double Lambda = dataCenter.SLambda;
    RLStitle = new JLabel("RLS",SwingConstants.CENTER);
    LambdaSlider = new JSlider(JSlider.HORIZONTAL,-180,20,LambdaToSlider(Lambda));
    LambdaLabel = new JLabel(" Lambda: ");
    
    LambdaTextField = new JTextField(String.format("%.4e",Lambda));
    LambdaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkLambdaTextField()) {
          if (!paramModified) return;
          if (my_inversion.canInvert) {
            my_inversion.RLSinversion();
            replotInvertedFunctions();
            replotAllKernels();
          }
        }
      }
    });
    
    LambdaTitle = new JPanel();
    LambdaTitle.setLayout(new GridLayout(1,2));
    LambdaTitle.add(LambdaLabel);
    LambdaTitle.add(LambdaTextField);
    
    LambdaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Lambda = SliderToLambda();
        LambdaTextField.setText(String.format("%.4e",Lambda));
        if ((!calculateWhileSliding.isSelected())&&
            (LambdaSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.SLambda = Lambda;
        if (my_inversion.canInvert) {
          my_inversion.RLSinversion();
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });
  }
    
  /////////////////////////////////////////////////////////////////////////////
  // Setting up the SOLA sliders
  /////////////////////////////////////////////////////////////////////////////
  private void initSOLAsliders() {
    SOLAtitle = new JLabel("SOLA",SwingConstants.CENTER);

    // later on, struct1Inversion will be initialised to true, hence this choice:
    double Beta = dataCenter.Beta1;
    double Theta = dataCenter.STheta1;
    double Delta = dataCenter.SDelta1;

    BetaLabel = new JLabel(" Beta: ");
    BetaTextField = new JTextField(String.format("%.4e",Beta));
    BetaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkBetaTextField()) {
          if (!paramModified) return;
          if (my_inversion.canInvert) {
            if (struct1Inversion.isSelected()) {
              my_inversion.SOLAinversion1();
            } else {
              my_inversion.SOLAinversion2();
            }
            replotInvertedFunctions();
            replotAllKernels();
          }
        }
      }
    });

    BetaTitle = new JPanel();
    BetaTitle.setLayout(new GridLayout(1,2));
    BetaTitle.add(BetaLabel);
    BetaTitle.add(BetaTextField);

    BetaSlider = new JSlider(JSlider.HORIZONTAL,-100,100,BetaToSlider(Beta));
    BetaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Beta = SliderToBeta();
        BetaTextField.setText(String.format("%.4e",Beta));
        if ((!calculateWhileSliding.isSelected())&&
            (BetaSlider.getModel().getValueIsAdjusting())) return;
        if (struct1Inversion.isSelected()) {
          dataCenter.Beta1 = Beta;
        } else {
          dataCenter.Beta2 = Beta;
        }
        if (my_inversion.canInvert) {
          if (struct1Inversion.isSelected()) {
            my_inversion.SOLAinversion1();
          } else {
            my_inversion.SOLAinversion2();
          }
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });
    
    ThetaLabel = new JLabel(" Theta: ");
    ThetaTextField = new JTextField(String.format("%.4e",Theta));
    ThetaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkThetaTextField()) {
          if (!paramModified) return;
          if (my_inversion.canInvert) {
            if (struct1Inversion.isSelected()) {
              my_inversion.SOLAinversion1();
            } else {
              my_inversion.SOLAinversion2();
            }
            replotInvertedFunctions();
            replotAllKernels();
          }
        }
      }
    });

    ThetaTitle = new JPanel();
    ThetaTitle.setLayout(new GridLayout(1,2));
    ThetaTitle.add(ThetaLabel);
    ThetaTitle.add(ThetaTextField);

    ThetaSlider = new JSlider(JSlider.HORIZONTAL,-200,0,ThetaToSlider(Theta));
    ThetaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Theta = SliderToTheta();
        ThetaTextField.setText(String.format("%.4e",Theta));
        if ((!calculateWhileSliding.isSelected())&&
            (ThetaSlider.getModel().getValueIsAdjusting())) return;
        if (struct1Inversion.isSelected()) {
          dataCenter.STheta1 = Theta;
        } else {
          dataCenter.STheta2 = Theta;
        }
        if (my_inversion.canInvert){
          if (struct1Inversion.isSelected()) {
            my_inversion.SOLAinversion1();
          } else {
            my_inversion.SOLAinversion2();
          }
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });
        
    DeltaSlider = new JSlider(JSlider.HORIZONTAL,-200,0,DeltaToSlider(Delta));
    DeltaLabel = new JLabel(" Delta: ");
    DeltaTextField = new JTextField(String.format("%.4e",Delta));
    DeltaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkDeltaTextField()) {
          if (!paramModified) return;

          // make sure a model is loaded (because you need c)
          if (my_inversion.canInvert&&!dataCenter.modelLoaded) {
            JOptionPane.showMessageDialog(
              DeltaTextField, // parent component
              "No model loaded.  Unable to proceed",
              "Error",
              JOptionPane.ERROR_MESSAGE
            );
            return;
          }

          if (my_inversion.canInvert) {
            if (struct1Inversion.isSelected()) {
              my_inversion.initDelta1SOLA();
              my_inversion.SOLAinversion1();
            } else {
              my_inversion.initDelta2SOLA();
              my_inversion.SOLAinversion2();
            }
            replotInvertedFunctions();
            replotAllKernels();
          }
        }
      }
    });

    DeltaTitle = new JPanel();
    DeltaTitle.setLayout(new GridLayout(1,2));
    DeltaTitle.add(DeltaLabel);
    DeltaTitle.add(DeltaTextField);

    DeltaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Delta = SliderToDelta();
        DeltaTextField.setText(String.format("%.4e",Delta));
        if ((!calculateWhileSliding.isSelected())&&
            (DeltaSlider.getModel().getValueIsAdjusting())) return;
        if (struct1Inversion.isSelected()) {
          dataCenter.SDelta1 = Delta;
        } else {
          dataCenter.SDelta2 = Delta;
        }

        // make sure a model is loaded (because you need c)
        if (my_inversion.canInvert&&!dataCenter.modelLoaded) {
          JOptionPane.showMessageDialog(
            DeltaSlider, // parent component
            "No model loaded.  Unable to proceed",
            "Error",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }

        if (my_inversion.canInvert) {
          if (struct1Inversion.isSelected()) {
            my_inversion.initDelta1SOLA();
            my_inversion.SOLAinversion1();
          } else {
            my_inversion.initDelta2SOLA();
            my_inversion.SOLAinversion2();
          }
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise buttons for the selection of kernel pairs, and specific kernels.
  /////////////////////////////////////////////////////////////////////////////
  private void initStructKernelButtons() {
    kernelTypeComboBox = new JComboBox();
    for(int i=0; i<DataCenter.kernelTypes.length;i++)
      kernelTypeComboBox.addItem(DataCenter.kernelTypes[i]);
    kernelTypeComboBox.setSelectedItem(DataCenter.kernelTypes[dataCenter.kernelType]);
    kernelTypeComboBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String newKernelType = (String)kernelTypeComboBox.getSelectedItem();
        if (!DataCenter.kernelTypes[dataCenter.kernelType].equals(newKernelType)) {
          for(int k=0; k<DataCenter.kernelTypes.length;k++) {
            if (DataCenter.kernelTypes[k].equals(newKernelType)) dataCenter.kernelType = k;
          }
          struct1Inversion.setText(DataCenter.structNames[dataCenter.kernelType][0]);
          struct2Inversion.setText(DataCenter.structNames[dataCenter.kernelType][1]);
          if (!dataCenter.structInversionUpToDate||my_inversion.canInvert) {
            updateInversion(kernelTypeComboBox);
            replotInvertedFunctions();
            replotAllKernels();
          }
        }
      }
    });

    struct1Inversion = new JRadioButton(DataCenter.structNames[dataCenter.kernelType][0],true);
    struct2Inversion = new JRadioButton(DataCenter.structNames[dataCenter.kernelType][1]);
    ActionListener l = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (struct1Inversion.isSelected()) {
          Utilities.ChangeJSliderValue(BetaSlider,BetaToSlider(dataCenter.Beta1));
          BetaTextField.setText(String.format("%.4e",dataCenter.Beta1));
          Utilities.ChangeJSliderValue(ThetaSlider,ThetaToSlider(dataCenter.STheta1));
          ThetaTextField.setText(String.format("%.4e",dataCenter.STheta1));
          Utilities.ChangeJSliderValue(DeltaSlider,DeltaToSlider(dataCenter.SDelta1));
          DeltaTextField.setText(String.format("%.4e",dataCenter.SDelta1));
        } else {
          Utilities.ChangeJSliderValue(BetaSlider,BetaToSlider(dataCenter.Beta2));
          BetaTextField.setText(String.format("%.4e",dataCenter.Beta2));
          Utilities.ChangeJSliderValue(ThetaSlider,ThetaToSlider(dataCenter.STheta2));
          ThetaTextField.setText(String.format("%.4e",dataCenter.STheta2));
          Utilities.ChangeJSliderValue(DeltaSlider,DeltaToSlider(dataCenter.SDelta2));
          DeltaTextField.setText(String.format("%.4e",dataCenter.SDelta2));
        }
        replotInvertedFunctions();
        replotAllKernels();
      }
    };
    struct1Inversion.addActionListener(l);
    struct2Inversion.addActionListener(l);
    ButtonGroup group2 = new ButtonGroup();
    group2.add(struct1Inversion);
    group2.add(struct2Inversion);

    structPanel = new JPanel();
    structPanel.setLayout(new GridLayout(1,2));
    structPanel.add(struct1Inversion);
    structPanel.add(struct2Inversion);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Setting up the Kernel slider:   
  /////////////////////////////////////////////////////////////////////////////
  private void initKernelSlider() {
    KernelLabel = new JLabel(" Position, r/R: ");
    KernelTextField = new JTextField(String.format("%.4f",0.0));
    KernelTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkKernelTextField()) {
          if (!paramModified) return;
          replotAllKernels();
        }
      }
    });

    KernelTitle = new JPanel();
    KernelTitle.setLayout(new GridLayout(1,2));
    KernelTitle.add(KernelLabel);
    KernelTitle.add(KernelTextField);


    // VERY IMPORTANT: the number of positions on the Kernel Slider MUST
    // correspond to the number of points in gridInversion
    KernelSlider = new JSlider(JSlider.HORIZONTAL,0,dataCenter.n_gridS-1,0);
    KernelSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        KernelTextField.setText(String.format("%.4f",SliderToXKernel()));
        if ((!calculateWhileSliding.isSelected())&&
            (KernelSlider.getModel().getValueIsAdjusting())) return;
        replotAllKernels();
      }
    });

    NGridLabel = new JLabel(" # of grid points: ");
    NGridTextField = new JTextField(String.format("%d",dataCenter.n_gridS));
    NGridTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkNGridTextField()) {
          if (!paramModified) return;
          if (!dataCenter.structInversionUpToDate||my_inversion.canInvert) {
            updateInversion(NGridTextField);
            replotInvertedFunctions();
            replotAllKernels();
          }
        }
      }
    });
    NGridPanel = new JPanel();
    NGridPanel.setLayout(new GridLayout(1,2));
    NGridPanel.add(NGridLabel);
    NGridPanel.add(NGridTextField);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise buttons for the selection of the numerical method.
  /////////////////////////////////////////////////////////////////////////////
  private void initIntegrationMethodComboBox() {
    integrationMethodComboBox = new JComboBox();
    integrationMethodComboBox.addItem("Sliding window");
    integrationMethodComboBox.addItem("Chebyshev");
    if (my_inversion.integrationMethod == StructuralInversion.Chebyshev)
      integrationMethodComboBox.setSelectedItem("Chebyshev");
    if (my_inversion.integrationMethod == StructuralInversion.SlidingWindow)
      integrationMethodComboBox.setSelectedItem("Sliding window");
    integrationMethodComboBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        int newIntegrationMethod = my_inversion.integrationMethod;
        String newChoice = (String)integrationMethodComboBox.getSelectedItem();
        if (newChoice.equals("Chebyshev")) newIntegrationMethod = StructuralInversion.Chebyshev;
        if (newChoice.equals("Sliding window")) newIntegrationMethod = StructuralInversion.SlidingWindow;
        if (newIntegrationMethod != my_inversion.integrationMethod) {
          my_inversion.integrationMethod = newIntegrationMethod;
          if (!dataCenter.structInversionUpToDate||my_inversion.canInvert) {
            updateInversion(integrationMethodComboBox);
            replotInvertedFunctions();
            replotAllKernels();
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the text field and label that allows one to adjust the number
  // of Legendre polynomials to fit/remove surface effects.
  /////////////////////////////////////////////////////////////////////////////
  private void initNSurfPanel() {
    NSurfLabel = new JLabel(" # of surf. poly.: ");
    NSurfTextField = new JTextField(String.format("%d",dataCenter.n_surf));
    NSurfTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (checkNSurfTextField()) {
          if (!paramModified) return;
          if (my_inversion.canInvert) {
            my_inversion.initSurfPoly();
            my_inversion.initRLS();
            my_inversion.initSOLA();
            my_inversion.RLSinversion();
            if (struct1Inversion.isSelected()) {
              my_inversion.SOLAinversion1();
            } else {
              my_inversion.SOLAinversion2();
            }
            replotInvertedFunctions();
            replotAllKernels();
          }
        }
      }
    });
    NSurfPanel = new JPanel();
    NSurfPanel.setLayout(new GridLayout(1,2));
    NSurfPanel.add(NSurfLabel);
    NSurfPanel.add(NSurfTextField);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Various utility methods which check the entries in the different text
  // fields and returns true if these entries are as they should be.
  // If they are not suitable, an error message will be posted.  Finally,
  // the different sliders will be adjusted according to these entries.
  /////////////////////////////////////////////////////////////////////////////
  private boolean checkNoiseTextField() {
    Double Noise = Utilities.readDouble(NoiseTextField.getText(),
                   "noise level",0.0,1e30,NoiseTextField);
    if (Noise == null) {
      NoiseTextField.setText(String.format("%.4e",dataCenter.SNoiseLevel));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(NoiseSlider, NoiseToSlider(Noise));
      NoiseTextField.setText(String.format("%.4e",Noise));
      paramModified = (dataCenter.SNoiseLevel != Noise.doubleValue());
      dataCenter.SNoiseLevel = Noise.doubleValue();
      return true;
    }
  }

  private boolean checkLambdaTextField() {
    Double Lambda = Utilities.readDouble(LambdaTextField.getText(),
                   "value of Lambda",0.0,1e30,LambdaTextField);
    if (Lambda == null) {
      LambdaTextField.setText(String.format("%.4e",dataCenter.SLambda));
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(LambdaSlider, LambdaToSlider(Lambda));
      LambdaTextField.setText(String.format("%.4e",Lambda));
      paramModified = (dataCenter.SLambda != Lambda.doubleValue());
      dataCenter.SLambda = Lambda.doubleValue();
      return true;
    }
  }

  private boolean checkBetaTextField() {
    Double Beta = Utilities.readDouble(BetaTextField.getText(),
                   "value of Beta",0.0,1e30,BetaTextField);
    if (Beta == null) {
      if (struct1Inversion.isSelected()) {
        BetaTextField.setText(String.format("%.4e",dataCenter.Beta1));
      } else {
        BetaTextField.setText(String.format("%.4e",dataCenter.Beta2));
      }
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(BetaSlider, BetaToSlider(Beta));
      BetaTextField.setText(String.format("%.4e",Beta));
      if (struct1Inversion.isSelected()) {
        paramModified = (dataCenter.Beta1 != Beta.doubleValue());
        dataCenter.Beta1 = Beta.doubleValue();
      } else {
        paramModified = (dataCenter.Beta2 != Beta.doubleValue());
        dataCenter.Beta2 = Beta.doubleValue();
      }
      return true;
    }
  }

  private boolean checkThetaTextField() {
    Double Theta = Utilities.readDouble(ThetaTextField.getText(),
                   "value of Theta",0.0,Math.PI/2.0,ThetaTextField);
    if (Theta == null) {
      if (struct1Inversion.isSelected()) {
        ThetaTextField.setText(String.format("%.4e",dataCenter.STheta1));
      } else {
        ThetaTextField.setText(String.format("%.4e",dataCenter.STheta2));
      }
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(ThetaSlider, ThetaToSlider(Theta));
      ThetaTextField.setText(String.format("%.4e",Theta));
      if (struct1Inversion.isSelected()) {
        paramModified = (dataCenter.STheta1 != Theta.doubleValue());
        dataCenter.STheta1 = Theta.doubleValue();
      } else {
        paramModified = (dataCenter.STheta2 != Theta.doubleValue());
        dataCenter.STheta2 = Theta.doubleValue();
      }
      return true;
    }
  }

  private boolean checkDeltaTextField() {
    Double Delta = Utilities.readDouble(DeltaTextField.getText(),
                   "value of Delta",0.0,1e2,DeltaTextField);
    if (Delta == null) {
      if (struct1Inversion.isSelected()) {
        DeltaTextField.setText(String.format("%.4e",dataCenter.SDelta1));
      } else {
        DeltaTextField.setText(String.format("%.4e",dataCenter.SDelta2));
      }
      paramModified = false;
      return false;
    } else {
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(DeltaSlider, DeltaToSlider(Delta));
      DeltaTextField.setText(String.format("%.4e",Delta));
      if (struct1Inversion.isSelected()) {
        paramModified = (dataCenter.SDelta1 != Delta.doubleValue());
        dataCenter.SDelta1 = Delta.doubleValue();
      } else {
        paramModified = (dataCenter.SDelta2 != Delta.doubleValue());
        dataCenter.SDelta2 = Delta.doubleValue();
      }
      return true;
    }
  }

  private boolean checkKernelTextField() {
    // checking the bounds is useless - I will be modifying the value
    // anyways afterwards:
    Double XKernel = Utilities.readDouble(KernelTextField.getText(),
                     "position value",Double.NEGATIVE_INFINITY,
                     Double.POSITIVE_INFINITY,KernelTextField);
    if (XKernel == null) {
      KernelTextField.setText(String.format("%.4f",SliderToXKernel()));
      paramModified = false;
      return false;
    } else {
      double oldValue = SliderToXKernel();
      // I use this home-made method to change the slider without
      // triggering the ChangeListener:
      Utilities.ChangeJSliderValue(KernelSlider, XKernelToSlider(XKernel));
      double newValue = SliderToXKernel();
      KernelTextField.setText(String.format("%.4f",newValue));
      paramModified = (oldValue != newValue);
      return true;
    }
  }

  private boolean checkNGridTextField() {
    Integer NGrid = Utilities.readInt(NGridTextField.getText(), "number of grid points",
                            2,9999,NGridTextField);
    if (NGrid == null) {
      NGridTextField.setText(String.format("%d",dataCenter.n_gridS));
      paramModified = false;
      return false;
    } else {
      Utilities.ChangeJSliderMaximum(KernelSlider,NGrid-1);
      paramModified = (dataCenter.n_gridS != NGrid.intValue());
      dataCenter.n_gridS = NGrid.intValue();
      return true;
    }
  }

  private boolean checkNSurfTextField() {
    Integer NSurf = Utilities.readInt(NSurfTextField.getText(),
                            "number of polynomials",0,50,NSurfTextField);
    if (NSurf == null) {
      NSurfTextField.setText(String.format("%d",dataCenter.n_surf));
      paramModified = false;
      return false;
    } else {
      paramModified = (dataCenter.n_surf != NSurf.intValue());
      dataCenter.n_surf = NSurf.intValue();
      return true;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Setting up the refresh button and the RLS, SOLA and Target function check
  // boxes.
  /////////////////////////////////////////////////////////////////////////////
  private void initCheckBoxes() {
    Refresh = new JButton("Update Inversion");
    RLSCheckBox = new JCheckBox("<html> <font color='red'>RLS</font></html>",true);
    SOLACheckBox = new JCheckBox("<html> <font color='blue'>SOLA</font></html>",true);
    TargetFunctionCheckBox = new JCheckBox("Target functions",true);
    showErrorBars = new JCheckBox("Show error bars",true);

    Refresh.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (dataCenter.SNoiseSelected) {
          if (!checkNoiseTextField()) return;
        }
        if (RLSCheckBox.isSelected()) {
          if (!checkLambdaTextField()) return;
        }
        if (SOLACheckBox.isSelected()) {
          if (!checkBetaTextField()) return;
          if (!checkThetaTextField()) return;
          if (!checkDeltaTextField()) return;
        }
        if (!checkKernelTextField()) return;
        if (!checkNGridTextField()) return;
        if (!checkNSurfTextField()) return;
        updateInversion(Refresh);
        replotInvertedFunctions();
        replotAllKernels();
      }
    });

    RLSCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        RLStitle.setEnabled(RLSCheckBox.isSelected());
        LambdaLabel.setEnabled(RLSCheckBox.isSelected());
        LambdaTextField.setEnabled(RLSCheckBox.isSelected());
        LambdaSlider.setEnabled(RLSCheckBox.isSelected());
        replotInvertedFunctions();
        replotAllKernels();
      }
    });
          
    SOLACheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        SOLAtitle.setEnabled(SOLACheckBox.isSelected());
        BetaLabel.setEnabled(SOLACheckBox.isSelected());
        BetaTextField.setEnabled(SOLACheckBox.isSelected());
        BetaSlider.setEnabled(SOLACheckBox.isSelected());
        ThetaLabel.setEnabled(SOLACheckBox.isSelected());
        ThetaTextField.setEnabled(SOLACheckBox.isSelected());
        ThetaSlider.setEnabled(SOLACheckBox.isSelected());
        DeltaLabel.setEnabled(SOLACheckBox.isSelected());
        DeltaTextField.setEnabled(SOLACheckBox.isSelected());
        DeltaSlider.setEnabled(SOLACheckBox.isSelected());
        replotInvertedFunctions();
        replotAllKernels();
      }
    });

    TargetFunctionCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        replotInvertedFunctions();
      }
    });

    showErrorBars.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        replotInvertedFunctions();
      }
    });

  }

  /////////////////////////////////////////////////////////////////////////////
  // init buttons which deal with a target function
  /////////////////////////////////////////////////////////////////////////////
  private void initTargetFunctionButtons() {
    LoadTarget = new JButton("Load \u03B4f/f, \u03B4g/g functions");
    LoadTarget.setEnabled(!dataCenter.isApplet);
    LoadTarget.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showOpenDialog(LoadTarget);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            loadStructTargets(aFile);
            replotInvertedFunctions();
            dataCenter.currentDirectory = aFile.getParentFile();
          } else {
            JOptionPane.showMessageDialog(
              LoadTarget,
              "File \""+aFile.getName()+"\" doesn't exist.",
              "File not found",
              JOptionPane.WARNING_MESSAGE
            );
          }
        }
      }
    });
    
    ClearTarget = new JButton("Clear \u03B4f/f, \u03B4g/g functions");
    ClearTarget.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.targetStruct1 = null;
        dataCenter.targetStruct2 = null;
        dataCenter.gridTargetStruct = null;
        dataCenter.targetStructLoaded = false;
        replotInvertedFunctions();
      }
    });

    CalculateFreq = new JButton("<html>Find \u03BD<sub>obs</sub> from target functions</html>");
    CalculateFreq.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.frequenciesData.calculateFreq(CalculateFreq);
        dataCenter.frequenciesTable.updateUI();
      }
    });
  }

  /***************************************************************************
  *  This updates the inversion.
  ***************************************************************************/
  public void updateInversion(JComponent c) {
    my_inversion.init(c);
    numberOfKernels.setText(String.format(" This inversion uses %d kernels",
                            my_inversion.n_inputKernels));
    // no need to give an error message if canInvert = false, since
    // my_inversion.init(c) already gives error messages.
    if (my_inversion.canInvert) {
      my_inversion.RLSinversion();
      my_inversion.SOLAinversion1();
      my_inversion.SOLAinversion2();
    }
  }

  /************************************************************************
  * This initialises the program.
  ************************************************************************/
  public void initProgram() {
    graphInvertedFunction = new IPlot("Inversion results","Position, r/R", "Profile",5);
    graphInvertedFunction.setIOenabled(!dataCenter.isApplet);
    graphAveragingKernel = new IPlot("Averaging kernels","Position, r/R", "Kernel",2);
    graphAveragingKernel.setIOenabled(!dataCenter.isApplet);
    graphCrossTermKernel = new IPlot("Cross-term kernels","Position, r/R", "Kernel",2);
    graphCrossTermKernel.setIOenabled(!dataCenter.isApplet);
    graphContainer = new JPanel();
    graphContainer.setLayout(new GridLayout(3,1));
    graphContainer.add(graphInvertedFunction);
    graphContainer.add(graphAveragingKernel);
    graphContainer.add(graphCrossTermKernel);

    setLayout(new BorderLayout());
    add("Center", graphContainer);

    // init the inversion
    my_inversion = new StructuralInversion(dataCenter);

    calculateWhileSliding = new JCheckBox("Calculate while using sliders",false);
    initNoiseButtons();
    initRLSslider();
    initSOLAsliders();
    initStructKernelButtons();
    initKernelSlider();
    initCheckBoxes();
    initTargetFunctionButtons();
    initNSurfPanel();
    initIntegrationMethodComboBox();
    numberOfKernels = new JLabel(String.format(" This inversion uses %d kernels",
                                 my_inversion.n_inputKernels));
    
    SliderContainer = new JPanel();
    SliderContainer.setLayout(new GridLayout(31,1));
    SliderContainer.add(numberOfKernels);
    SliderContainer.add(Refresh);
    JPanel checkBoxes1 = new JPanel(new GridLayout(1,2));
    checkBoxes1.add(RLSCheckBox);
    checkBoxes1.add(TargetFunctionCheckBox);
    SliderContainer.add(checkBoxes1);
    JPanel checkBoxes2 = new JPanel(new GridLayout(1,2));
    checkBoxes2.add(SOLACheckBox);
    checkBoxes2.add(showErrorBars);
    SliderContainer.add(checkBoxes2);
    SliderContainer.add(calculateWhileSliding);
    SliderContainer.add(kernelTypeComboBox);
    SliderContainer.add(structPanel);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(NoiseTitle);
    SliderContainer.add(NoiseSlider);
    SliderContainer.add(NoiseRefresh);
    SliderContainer.add(RLStitle);
    SliderContainer.add(LambdaTitle);
    SliderContainer.add(LambdaSlider);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(SOLAtitle);
    SliderContainer.add(BetaTitle);
    SliderContainer.add(BetaSlider);
    SliderContainer.add(ThetaTitle);
    SliderContainer.add(ThetaSlider);
    SliderContainer.add(DeltaTitle);
    SliderContainer.add(DeltaSlider);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(KernelTitle);
    SliderContainer.add(KernelSlider);
    SliderContainer.add(NGridPanel);
    SliderContainer.add(integrationMethodComboBox);
    SliderContainer.add(NSurfPanel);
    SliderContainer.add(LoadTarget);
    SliderContainer.add(ClearTarget);
    SliderContainer.add(CalculateFreq);
    add("East",SliderContainer);
       
    graphInvertedFunction.repaint();
    graphAveragingKernel.repaint();
  }

  /////////////////////////////////////////////////////////
  // load a file with target structural functions:
  /////////////////////////////////////////////////////////
  private void loadStructTargets(File aFile){
    BufferedReader f;
    String line;
    Boolean skipLine = false;
    ArrayList<Double> x = new ArrayList<Double>();
    ArrayList<Double> y = new ArrayList<Double>();
    ArrayList<Double> z = new ArrayList<Double>();
    try {
      f = Utilities.OpenFile(aFile.getPath());
      while (true) {
        line = f.readLine();     // read a line of text
        if (line == null) break; // test for the end of the buffer
        line = Utilities.trim(line);       // remove comments
        StringTokenizer t = new StringTokenizer(line," \t,");
        // only read lines with EXACTLY 3 numbers
        if (t.countTokens() == 3) {
          x.add(Utilities.parseDouble(t.nextToken()));
          y.add(Utilities.parseDouble(t.nextToken()));
          z.add(Utilities.parseDouble(t.nextToken()));
        } else {
          if (t.countTokens() != 0) skipLine = true;
        }
      }
      f.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "Unable to read file \""+aFile.getName()+"\"",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    if (x.size() == 0) {
      JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "No data in file",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    // perform a basic check on the grid:
    for(int i=1; i<x.size(); i++) {
      if (x.get(i).compareTo(x.get(i-1)) < 0) {
        JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "Please define the target functions\n"+
                    "on a strictly ascending grid.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
        return;
      }
    }
    
    if (skipLine) {
      JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "Some lines were skipped\nwhen reading the file.",
                    "Warning",
                    JOptionPane.WARNING_MESSAGE
                  );
    }
    dataCenter.gridTargetStruct = new double[x.size()];
    dataCenter.targetStruct1 = new double[y.size()];
    dataCenter.targetStruct2 = new double[z.size()];
    for(int i=0; i<x.size(); i++) {
      dataCenter.gridTargetStruct[i] = (double)(x.get(i));
      dataCenter.targetStruct1[i] = (double)(y.get(i));
      dataCenter.targetStruct2[i] = (double)(z.get(i));
    }
    dataCenter.targetStructLoaded = true;
  }

  /***************************************************************************
  *  Plot the inverted and target rotation functions
  ***************************************************************************/
  private void replotInvertedFunctions() {
    if (TargetFunctionCheckBox.isSelected()&&dataCenter.targetStructLoaded) {
      if (struct1Inversion.isSelected()) {
        graphInvertedFunction.replaceDataSet(4,dataCenter.gridTargetStruct,
                                               dataCenter.targetStruct1,Plot.solid);
      } else {
        graphInvertedFunction.replaceDataSet(4,dataCenter.gridTargetStruct,
                                               dataCenter.targetStruct2,Plot.solid);
      } 
    } else {
      graphInvertedFunction.nullify(4);
    }

    if (my_inversion.canInvert) {
      int n_grid = dataCenter.gridSInversion.length;
      double[] aux = new double[n_grid];

      if (RLSCheckBox.isSelected()) {
        if (struct1Inversion.isSelected()) {
          graphInvertedFunction.replaceDataSet(0,dataCenter.gridSInversion,my_inversion.f1_invertedRLS,
                                               Plot.solid,Color.red);
        } else {
          graphInvertedFunction.replaceDataSet(0,dataCenter.gridSInversion,my_inversion.f2_invertedRLS,
                                               Plot.solid,Color.red);
        }
      } else {
        graphInvertedFunction.nullify(0);
      }
      
      if (RLSCheckBox.isSelected()&&showErrorBars.isSelected()) {
        if (struct1Inversion.isSelected()) {
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f1_invertedRLS[i]+my_inversion.RLS_errorBars1[i];
          graphInvertedFunction.replaceDataSet(1,dataCenter.gridSInversion,aux,Plot.dotted,Color.red);
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f1_invertedRLS[i]-my_inversion.RLS_errorBars1[i];
          graphInvertedFunction.replaceDataSet(2,dataCenter.gridSInversion,aux,Plot.dotted,Color.red);
        } else {
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f2_invertedRLS[i]+my_inversion.RLS_errorBars2[i];
          graphInvertedFunction.replaceDataSet(1,dataCenter.gridSInversion,aux,Plot.dotted,Color.red);
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f2_invertedRLS[i]-my_inversion.RLS_errorBars2[i];
          graphInvertedFunction.replaceDataSet(2,dataCenter.gridSInversion,aux,Plot.dotted,Color.red);
        }
      } else {
        graphInvertedFunction.nullify(1);
        graphInvertedFunction.nullify(2);
      }

      if (SOLACheckBox.isSelected()) {
        if (struct1Inversion.isSelected()) {
          if (showErrorBars.isSelected()) {
            graphInvertedFunction.replaceDataSet(3,my_inversion.SOLA_newGrid1,my_inversion.f1_invertedSOLA,
                                                 Plot.noLine,Color.blue);
            graphInvertedFunction.replaceLeftError(3,my_inversion.SOLA_leftErrorBars1);
            graphInvertedFunction.replaceRightError(3,my_inversion.SOLA_rightErrorBars1);
            graphInvertedFunction.replaceLowerError(3,my_inversion.SOLA_errorBars1);
            graphInvertedFunction.replaceUpperError(3,my_inversion.SOLA_errorBars1);
            graphInvertedFunction.setHErrorValue(3,true);
            graphInvertedFunction.setVErrorValue(3,true);
          } else {
            graphInvertedFunction.replaceDataSet(3,my_inversion.SOLA_newGrid1,my_inversion.f1_invertedSOLA,
                                                 Plot.solid,Color.blue);
            graphInvertedFunction.setHErrorValue(3,false);
            graphInvertedFunction.setVErrorValue(3,false);
          }
        } else {
          if (showErrorBars.isSelected()) {
            graphInvertedFunction.replaceDataSet(3,my_inversion.SOLA_newGrid2,my_inversion.f2_invertedSOLA,
                                                 Plot.noLine,Color.blue);
            graphInvertedFunction.replaceLeftError(3,my_inversion.SOLA_leftErrorBars2);
            graphInvertedFunction.replaceRightError(3,my_inversion.SOLA_rightErrorBars2);
            graphInvertedFunction.replaceLowerError(3,my_inversion.SOLA_errorBars2);
            graphInvertedFunction.replaceUpperError(3,my_inversion.SOLA_errorBars2);
            graphInvertedFunction.setHErrorValue(3,true);
            graphInvertedFunction.setVErrorValue(3,true);
          } else {
            graphInvertedFunction.replaceDataSet(3,my_inversion.SOLA_newGrid2,my_inversion.f2_invertedSOLA,
                                                 Plot.solid,Color.blue);
            graphInvertedFunction.setHErrorValue(3,false);
            graphInvertedFunction.setVErrorValue(3,false);
          }
        }
      } else {
        graphInvertedFunction.nullify(3);
      }
    } else {
      for(int i=0; i<=3; i++) graphInvertedFunction.nullify(i);
    }

    graphInvertedFunction.findRange();

    if (my_inversion.canInvert) {
      graphInvertedFunction.setXmin(dataCenter.gridSInversion[0]);
      graphInvertedFunction.setXmax(dataCenter.gridSInversion[dataCenter.gridSInversion.length-1]);
    }
    graphInvertedFunction.repaint();
  }

  /***************************************************************************
  *  Plot all kernels
  ***************************************************************************/
  private void replotAllKernels() {
    replotAveragingKernels();
    replotCrossTermKernels();
  }

  /***************************************************************************
  *  Plot averaging kernels
  ***************************************************************************/
  private void replotAveragingKernels() {
    // quick exit if possible:
    if (!my_inversion.canInvert) {
      graphAveragingKernel.nullify();
      graphAveragingKernel.repaint();
      return;
    }
    
    // number of grid points
    int n_grid = dataCenter.gridSInversion.length;
    
    // this specifies which kernel to plot:
    int iKernel = KernelSlider.getValue();
    
    // update KernelTextField (just in case):
    KernelTextField.setText(String.format("%.4f",SliderToXKernel()));

    // RLS averaging kernel:
    if (RLSCheckBox.isSelected()) {
      double[] y = new double[n_grid];
      if (struct1Inversion.isSelected()) {
        y = my_inversion.findAveragingKernel1RLS(iKernel);
      } else {
        y = my_inversion.findAveragingKernel2RLS(iKernel);
      }
      graphAveragingKernel.replaceDataSet(0,dataCenter.grid,y,Plot.solid,Color.red);
    } else {
      graphAveragingKernel.nullify(0);
    }
    
    // SOLA averaging kernel:
    if (SOLACheckBox.isSelected()) {
      double[] y = new double[n_grid];
      if (struct1Inversion.isSelected()) {
        y = my_inversion.findAveragingKernel1SOLA(iKernel);
      } else {
        y = my_inversion.findAveragingKernel2SOLA(iKernel);
      }
      graphAveragingKernel.replaceDataSet(1,dataCenter.grid,y,Plot.solid,Color.blue);
    } else {
      graphAveragingKernel.nullify(1);
    }

    graphAveragingKernel.findRange();
    graphAveragingKernel.repaint();
  }

  /***************************************************************************
  *  Plot cross-term kernels
  ***************************************************************************/
  private void replotCrossTermKernels() {
    // quick exit if possible:
    if (!my_inversion.canInvert) {
      graphCrossTermKernel.nullify();
      graphCrossTermKernel.repaint();
      return;
    }
    
    // number of grid points
    int n_grid = dataCenter.gridSInversion.length;
    
    // this specifies which kernel to plot:
    int iKernel = KernelSlider.getValue();
    
    // update KernelTextField (just in case):
    KernelTextField.setText(String.format("%.4f",SliderToXKernel()));

    // RLS crossTerm kernel:
    if (RLSCheckBox.isSelected()) {
      double[] y = new double[n_grid];
      if (struct1Inversion.isSelected()) {
        y = my_inversion.findCrossTermKernel1RLS(iKernel);
      } else {
        y = my_inversion.findCrossTermKernel2RLS(iKernel);
      }
      graphCrossTermKernel.replaceDataSet(0,dataCenter.grid,y,Plot.solid,Color.red);
    } else {
      graphCrossTermKernel.nullify(0);
    }
    
    // SOLA crossTerm kernel:
    if (SOLACheckBox.isSelected()) {
      double[] y = new double[n_grid];
      if (struct1Inversion.isSelected()) {
        y = my_inversion.findCrossTermKernel1SOLA(iKernel);
      } else {
        y = my_inversion.findCrossTermKernel2SOLA(iKernel);
      }
      graphCrossTermKernel.replaceDataSet(1,dataCenter.grid,y,Plot.solid,Color.blue);
    } else {
      graphCrossTermKernel.nullify(1);
    }

    graphCrossTermKernel.findRange();
    graphCrossTermKernel.repaint();
  }
}
